/*  QT Abstraction for Frame and Signal handling
    Copyright (C) 2018 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <QDateTime>
#include <QDebug>

#include <string.h>
#include <time.h>

#include "QCANDecode.h"
#include "restbus/cs_dbc.h"

/****************************************************************************/
/****************************************************************************/

QCANSocket::QCANSocket(QObject * parent):
  QObject(parent),
  rsocket(NULL),
  id(0)
{
  sk = cs_socketcan_init();
  if(sk == NULL){
    qDebug()<< "Failed to int socket structure";
  }
  memset(rawBuffer,0,sizeof(uint8_t)*CS_SOCKETCAN_MAXDLC);
  dt.setTime_t(0);
  dlc = CS_SOCKETCAN_MAXDLC;
}

/****************************************************************************/
/****************************************************************************/

QCANSocket::~QCANSocket()
{
  if(rsocket){
    disconnect(rsocket, 0, 0, 0);
    delete rsocket;
  }
  if(sk != NULL){
    cs_socketcan_free(sk);
  }
}

/****************************************************************************/
/****************************************************************************/

int QCANSocket::open(QString interface)
{
  if(cs_socketcan_config_native(sk,(char*)interface.toLocal8Bit().constData())){
    qDebug() << "Failed to configure native socket";
    emit error(CANSocket_CONFIG,QString("Failed to configure native socket"));
    return 1;
  }
  
  if(cs_socketcan_open(sk)){
    qDebug() << "Failed to open native socket";
    emit error(CANSocket_CONFIG,QString("Failed to open native socket"));
    return 1;
  }
  rsocket = new QSocketNotifier((qintptr)cs_socketcan_getsocket(sk),QSocketNotifier::Read,0);  
  connect(rsocket, SIGNAL(activated(int)), this, SLOT(read()));
  return 0;
}

/****************************************************************************/
/****************************************************************************/

int QCANSocket::close()
{
  if(rsocket) {
    disconnect(rsocket, 0, 0, 0);
    delete rsocket;
    rsocket = NULL;
  }
  if(cs_socketcan_close(sk)){
    qDebug() << "Failed to close native socket";
    emit error(CANSocket_CONFIG,QString("Failed to close native socket"));
    return 1;
  }
  return 0;
}

/****************************************************************************/
/****************************************************************************/

void QCANSocket::enableDebug()
{
  cs_socketcan_debuglevel(SOCKETCAN_DEBUG_LEVEL_DEBUG);
}

/****************************************************************************/
/****************************************************************************/

void QCANSocket::disableDebug()
{
  cs_socketcan_debuglevel(SOCKETCAN_DEBUG_LEVEL_NONE);
}

/****************************************************************************/
/****************************************************************************/

int QCANSocket::socket_receive()
{
  uint32_t lid;
  uint8_t lbuffer[CS_SOCKETCAN_MAXDLC];
  int llen;
  struct timeval tv;
  int rv = CS_SOCKETCAN_OK;
  
  llen = CS_SOCKETCAN_MAXDLC;
  rv = cs_socketcan_recv_nonblock(sk,&lid,lbuffer,&llen,&tv);
  switch(rv){
  case CS_SOCKETCAN_OK:
    //We received a valid frame
    memcpy(rawBuffer,lbuffer,sizeof(uint8_t)*CS_SOCKETCAN_MAXDLC);
    dt.setTime_t(tv.tv_sec);
    dt = dt.addMSecs(1.0E-3*tv.tv_usec);
    id = lid;
    dlc = llen;
    return 1;
    break;
  case CS_SOCKETCAN_FAIL:
    //Nothing received or failure
    if(cs_socketcan_getlasterror(sk) != CS_SOCKETCAN_ERROR_NOERROR){
      //There was a failure
      emit error(CANSocket_IO,QString(cs_socketcan_getlasterrorstring(sk)));
    } 
    break;
  case CS_SOCKETCAN_ERROR:
    //Error Frame
    //Nothing todo
    rv = CS_SOCKETCAN_OK;
    break;
  case CS_SOCKETCAN_RTR:
    //RTR Frame
    //Nothing todo
    break;
  case CS_SOCKETCAN_TIMEOUT:
    //Receive timeout
    //Should never be reached for this function call
    break;
  default:
    //Should never be reached
    break;
  }
  return 0;
}

/****************************************************************************/
/****************************************************************************/

void QCANSocket::read()
{
  int rv = 0;
  while(socket_receive()){
    rv = 1;
  }
  if(rv == 1){
    emit received(id,dt);
  }
}

/****************************************************************************/
/****************************************************************************/

int QCANSocket::send(unsigned int id, unsigned char *buffer, int len)
{
  int rv;
  
  rv = cs_socketcan_send(sk,(uint32_t)id,(uint8_t*)buffer,len);
  
  if(rv == CS_SOCKETCAN_OK){
    //everything is ok
    return 0;
  }
  emit error(CANSocket_IO,QString(cs_socketcan_getlasterrorstring(sk)));
  return 1;
}


/****************************************************************************/
/****************************************************************************/

QVector<unsigned char> QCANSocket::getRAWBuffer()
{
  QVector<unsigned char> rv;
  int i;

  for(i = 0; i< dlc; i++){
    rv.append((unsigned char)rawBuffer[i]);
  }
  return rv;
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

/****************************************************************************/
/****************************************************************************/

QCANSocketMessage::QCANSocketMessage(QObject * parent):
  QCANSocket(parent)
{
  
}

/****************************************************************************/
/****************************************************************************/

QCANSocketMessage::~QCANSocketMessage()
{
  QHashIterator<QString, cs_signal_t*> i(shash);
  
  while (i.hasNext()) {
    i.next();
    cs_signal_config_free(i.value(),1);
    free(i.value());
  }
  shash.clear();
}

/****************************************************************************/
/****************************************************************************/

int QCANSocketMessage::open(QString interface,QString dbc, unsigned int id)
{
  uint32_t lid = (uint32_t) id;
  //Get DBC pointer structure
  cs_dbc_t *sdbc = cs_dbc_init();
  cs_message_t *m = NULL;
  int i;
  int numsignals = 0;

  qDebug() << "Open interface: "<< interface << " DBC: " << dbc << " ID: "<< id;

  QCANSocket::id = id; //FIXME Hm ist das richtig hier ???  
  //Parse DBC file
  if(cs_dbc_read(sdbc,(char*)dbc.toLocal8Bit().constData())){
    qDebug() << "Failed to read DBC file";
    cs_dbc_free(sdbc);
    emit error(CANSocket_CONFIG,QString("Failed to read DBC file"));
    return 1;
  }
  //Get message from DBC file
  if(cs_dbc_getmessage(sdbc,lid,&m) || (m == NULL)){
    cs_dbc_free(sdbc);
    qDebug() << "Failed to get Message";
    emit error(CANSocket_CONFIG,QString("Failed to get Message"));
    return 1;
  }
  //Get number of signals for message
  numsignals = cs_message_getnumsignals(m);

  dlc = cs_message_getdlc(m);
  
  if(numsignals <= 0){
    cs_dbc_free(sdbc);
    qDebug() << "Failed to get number of signals";
    emit error(CANSocket_CONFIG,QString("Failed to get number of signals"));
    return 1;
  }
  
  for(i = 0; i < numsignals; i++){
    cs_signal_t *ds = NULL;
    cs_signal_t *cps = NULL;
    //Get signal pointer from DBC Message signal
    ds = cs_message_getsignal(m,i);
    //Make a copy of the signal because we want to free the DBC structure and its content
    cps = cs_signal_config_cp(ds);
    //Add Name and signal pointer to hash
    shash.insert(QString(cps->name),cps);
  }
  //Free DBC structure
  cs_dbc_free(sdbc);
  
  if(cs_socketcan_config_native(sk,(char*)interface.toLocal8Bit().constData())){
    qDebug() << "Failed to configure native socket";
    emit error(CANSocket_CONFIG,QString("Failed to configure native socket"));
    return 1;
  }

  if(cs_socketcan_set_filter(sk,&lid,1)){
    qDebug() << "Failed to configure native socket filter";
    emit error(CANSocket_CONFIG,QString("Failed to configure native socket filter"));
    return 1;
  }
  
  if(cs_socketcan_open(sk)){
    qDebug() << "Failed to open native socket";
    emit error(CANSocket_CONFIG,QString("Failed to open native socket"));
    return 1;
  }
  rsocket = new QSocketNotifier((qintptr)cs_socketcan_getsocket(sk),QSocketNotifier::Read,0);  
  connect(rsocket, SIGNAL(activated(int)), this, SLOT(read()));
  return 0;
}

/****************************************************************************/
/****************************************************************************/

int QCANSocketMessage::close()
{
  QHashIterator<QString, cs_signal_t*> i(shash);

  disconnect(rsocket, 0, 0, 0);
  delete rsocket;
  rsocket = NULL;  
  while (i.hasNext()) {
    i.next();
    cs_signal_config_free(i.value(),1);
    free(i.value());
  }
  shash.clear();

  
  
  if(cs_socketcan_close(sk)){
    qDebug() << "Failed to close native socket";
    emit error(CANSocket_CONFIG,QString("Failed to close native socket"));
    return 1;
  }
  return 0;
}

/****************************************************************************/
/****************************************************************************/

double QCANSocketMessage::getValue(QString signal)
{
  double rv;
  cs_signal_t *s = NULL;

  if(!shash.contains(signal)){
    qDebug() << "Signal is not configured";
    emit error(CANSocket_SIGNAL,QString("Signal is not configured"));
    return 0.0;
  }

  s = shash.value(signal);
  if(cs_signal_runtime_decode(s,&rv,NULL,rawBuffer,CS_SOCKETCAN_MAXDLC)){
    qDebug() << "Signal decoding failed";
    emit error(CANSocket_SIGNAL,QString("Signal decoding failed"));
    return 0.0;
  }

  return rv;
}


/****************************************************************************/
/****************************************************************************/

void QCANSocketMessage::setValue(QString signal, double value)
{
  cs_signal_t *s = NULL;

  if(!shash.contains(signal)){
    emit error(CANSocket_SIGNAL,QString("Signal is not configured"));
    return;
  }
  
  s = shash.value(signal);

  if(cs_signal_runtime_encode(s,value,rawBuffer,dlc,0)){
    qDebug() << "Signal encoding failed";
    emit error(CANSocket_SIGNAL,QString("Signal encoding failed"));
    return;
  }
}

/****************************************************************************/
/****************************************************************************/

QStringList QCANSocketMessage::getSignals()
{
  QStringList rv = QStringList(shash.keys());
  return rv;
}

/****************************************************************************/
/****************************************************************************/

int QCANSocketMessage::send()
{
  int rv;
  
  rv = cs_socketcan_send(sk,id,rawBuffer,dlc);
  
  if(rv == CS_SOCKETCAN_OK){
    //everything is ok
    return 0;
  }
  emit error(CANSocket_IO,QString(cs_socketcan_getlasterrorstring(sk)));
  return 1;
}
