/*  QT Abstraction for Frame and Signal handling
    Copyright (C) 2018 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

/**
 * @file QCANDecode
 * @author Andreas Stewering-Bone
 * @date 2 July 2018
 * @brief QT Abstraction for Frame and Signal handling
 *
 * QT Abstraction for Frame and Signal handling
 */

#ifndef QCANDecode_H
#define QCANDecode_H

#include <QObject>
#include <QVector>
#include <QDateTime>
#include <QStringList>
#include <QSocketNotifier>

#include <inttypes.h>

#include "restbus/cs_socketcan.h"
#include "restbus/cs_signal.h"

/** \class QCANSocket
    Base class for SocketCAN IO handling
*/
class QCANSocket: public QObject
{
  Q_OBJECT
 public:
  /**Constructor*/
  QCANSocket(QObject * parent = 0);
  /**Destructor*/
  ~QCANSocket();
  /** Open CAN interface
   \retval 1 Failure
   \retval 0 Success
  */
  int open(QString interface);
  /** Close CAN Interface
   \retval 1 Failure
   \retval 0 Success
   */
  int close();
  /** Enable debugging
   */
  void enableDebug();
  /** Disable debugging
   */
  void disableDebug();
  /** Send raw buffer
      \retval 1 Failure
      \retval 0 Success
  */
  int send(unsigned int id, unsigned char *buffer, int len);
  /** Get RAW buffer received
      \retval 1 Failure
      \retval 0 Success
  */
  QVector<unsigned char> getRAWBuffer();
  typedef enum{
    CANSocket_CONFIG = 0, /**< Configuration failure*/
    CANSocket_IO,         /**< General IO failure*/
    CANSocket_SIGNAL      /**< General Signal failure*/
  } CANSocket_errorno;
 signals:
  void error(int error, QString errorstring);
  void received(unsigned int id, QDateTime &ts);
 private slots:
   void read();
 protected:
   cs_socketcan_t *sk;
   QSocketNotifier *rsocket;
   uint8_t rawBuffer[CS_SOCKETCAN_MAXDLC];
   QDateTime dt;
   uint32_t id;
   int dlc;
   /** Low level receive routine*/
   int socket_receive();
};

/** \class QCANSocketMessage
    Message data class for SocketCAN handling
*/
class QCANSocketMessage:  public QCANSocket
{
  Q_OBJECT
 public:
  /**Constructor*/
  QCANSocketMessage(QObject * parent = 0);
  /**Destructor*/
  ~QCANSocketMessage();
  /** Open interface and handle message defined by DBC file and message id*/
  int open(QString interface,QString dbc,unsigned int id);
  /** Close CAN Interface
      \retval 1 Failure
      \retval 0 Success
  */
  int close();
  /** Send signal buffer
      \retval 1 Failure
      \retval 0 Success
  */
  int send();
  double getValue(QString signal);
  void setValue(QString signal, double value);
  QStringList getSignals();
 private:
  QHash<QString,cs_signal_t *> shash;
};

#endif
