/*  QT Abstraction for Frame and Signal handling
    Copyright (C) 2018 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <QDebug>
#include <QMessageBox>
#include <QDateTime>

#include "MainWindow.h"

SignalEntry::SignalEntry(QString name, double value):
  value(value),
  name(name)
{

}

SignalEntry::~SignalEntry()
{

}

void SignalEntry::setValue(double value)
{
  SignalEntry::value = value;
}

SignalModel::SignalModel(QObject * parent) :
  QAbstractTableModel(parent)
{
  
}
int SignalModel::rowCount(const QModelIndex &) const
{
  return m_data.count();
}
int SignalModel::columnCount(const QModelIndex &) const
{
  return 2;
}
QVariant SignalModel::data(const QModelIndex &index, int role) const
{
  if (role != Qt::DisplayRole && role != Qt::EditRole)
    return QVariant();
  SignalEntry signal = m_data[index.row()];
  switch (index.column()) {
  case 0: return signal.getName();
  case 1: return signal.getValue();
  default: return QVariant();
  }
}
QVariant SignalModel::headerData(int section, Qt::Orientation orientation, int role) const
{
  if (orientation != Qt::Horizontal || role != Qt::DisplayRole)
    return QVariant();
  switch (section) {
  case 0: return "Name";
  case 1: return "Value";
  default: return QVariant();
  }
}
void SignalModel::append(const SignalEntry & signal)
{
  beginInsertRows(QModelIndex(), m_data.count(), m_data.count());
  m_data.append(signal);
  endInsertRows();
}

void SignalModel::clear()
{
  removeRows(0, m_data.count());
}

double SignalModel::getValue(QString name)
{
  int i;
  for(i = 0; i < m_data.count();i++){
    if(m_data[i].getName() == name){
      return m_data[i].getValue();
    }
  }
  return 0.0;
}


bool SignalModel::setData(const QModelIndex &index,
			       const QVariant &value, int role)
{
  if((index.row() < 0) || (m_data.count() < (index.row()+1))){
    return false;
  }
  if (index.isValid() && role == Qt::EditRole) {
    SignalEntry &signal = m_data[index.row()];
    signal.setValue(value.toDouble());
    qDebug() << "SetValue " << value;
    emit dataChanged(index, index);
    return true;
  }
  return false;
}

Qt::ItemFlags SignalModel::flags(const QModelIndex &index) const
{
  Qt::ItemFlags f;
  
  if (!index.isValid())
    return 0;
  
  f = Qt::ItemIsEnabled | Qt::ItemIsSelectable;
  
  if (index.column() == 1)
    f |= Qt::ItemIsEditable;
  
  return f;

}


MainWindow::MainWindow(
		       QWidget *parent):
  QMainWindow(parent)
{
    setupUi(this);
    can = new QCANSocketMessage(this);
    smdl = new SignalModel(this);
    tableView->setModel(smdl);
    started = false;
    connect(pushButton_start, SIGNAL(pressed()), this, SLOT(start()));
    connect(pushButton_stop, SIGNAL(pressed()), this, SLOT(stop()));
    connect(can, SIGNAL(received(unsigned int,QDateTime&)), this, SLOT(receive(unsigned int,QDateTime&)));
    connect(can, SIGNAL(error(int,QString)), this, SLOT(error(int,QString)));
    pushButton_start->setEnabled(true);
    pushButton_stop->setEnabled(false);
    sendtimer = new QTimer(this);
    connect(sendtimer, SIGNAL(timeout()), this, SLOT(send()));
}

/****************************************************************************/

MainWindow::~MainWindow()
{
}


/****************************************************************************/

void MainWindow::start()
{
  int i;
  if(can->open(lineEdit_iface->text(),lineEdit_dbc->text(),lineEdit_id->text().toInt())){
    QMessageBox msgBox;
    msgBox.setText("Please check configuration.");
    msgBox.exec();
    return;
  }
  textEdit->clear();
  started = true;
  slist = can->getSignals();
  smdl->clear();
  for (i = 0; i < slist.size(); i++){
    smdl->append(SignalEntry(slist[i],0.0));
  }
  if(doubleSpinBox_interval->value() > 0.0){
    sendtimer->setInterval(1000*doubleSpinBox_interval->value());
    sendtimer->start();
  }
  pushButton_start->setEnabled(false);
  pushButton_stop->setEnabled(true);
}

/****************************************************************************/

void MainWindow::stop()
{
  can->close();
  started = false;
  sendtimer->stop();
  pushButton_start->setEnabled(true);
  pushButton_stop->setEnabled(false);
}

/****************************************************************************/

void MainWindow::receive(unsigned int id, QDateTime &ts)
{
  QVector<unsigned char> rv = can->getRAWBuffer();
  int i;
  QString out = "";
  for(i = 0; i < rv.size(); i++){
    out.append(QString("%1 ").arg(rv[i]));
  }
  textEdit->append(QString("%1: %2 (%3): %4").arg(ts.toString("yy-MM-dd hh:mm:ss.zzz")).arg(id).arg(rv.size()).arg(out));
  for (i = 0; i < slist.size(); i++){
    textEdit->append(QString("    %1: %2").arg(slist[i]).arg(can->getValue(slist[i])));
  }
  
}

/****************************************************************************/

void MainWindow::error(int error, QString errorstr)
{
  Q_UNUSED(error)
  QMessageBox msgBox;
  msgBox.setText(errorstr);
  msgBox.exec();
  return;
}

/****************************************************************************/

void MainWindow::send()
{
  int i;
  for (i = 0; i < slist.size(); i++){
    can->setValue(slist[i],smdl->getValue(slist[i]));
  }
  can->send();
}

   
