/*  
    A2L Parser utils
    Copyright (C) 2015 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/mman.h>
#include <unistd.h>
#include <math.h>

#include "cs_a2l.h"
#include "cs_calibration_utils.h"
#include "cs_a2l_private.h"

/********************************************************************/
cs_debug_level_t cs_a2l_debuglevel(
				   cs_debug_level_t level
				   )
/********************************************************************/
{
  return cs_debuglevel(CS_DEBUG_A2L,level);
}

/********************************************************************/
void cs_a2l_debug_setcallback(
			      void *ctx,
			      void (*callb)(void *ctx,
					    cs_debug_sub_t sub,
					    cs_debug_level_t dl,
					    const char *func,
					    const int line,
					    const char *fmt,
					    va_list ap)
			      )
/********************************************************************/
{
  cs_debug_setcallback(CS_DEBUG_A2L,ctx,callb);
}


/** Cleanup A2L file
 */
/**************************************************************************/
int cs_a2l_cleanup(char *in, char **out, int size)
/**************************************************************************/
{
  // remove comments => /*jfhjgfhkgh*/
  // remove \r
  // replace \t by ' '
  // replace \n by ' '

  int i,j;
  int position;
  int fcp;
  char replace;
  enum {
	NORMAL = 0,
	POSSIBLE_START,
	POSSIBLE_END,
	INCOMMENT
  }stm = NORMAL;

  if((in == NULL)||(out == NULL)||(size <= 0)){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }

  //Malloc space
  *out = malloc(sizeof(char)*size);
  if(*out == NULL){
    _a2l_error("Malloc failure\n");
    return CS_FAIL;
  }

  memset(*out,0,size);
  position = 0;
  for(i = 0; i < size; i++){
    //printf("i: %d, position: %d, state: %d\n",i,position,stm);
    fcp = -1;
    //Remove comments
    switch(stm){
    case NORMAL:
      switch(in[i]){
      case '/':
	stm = POSSIBLE_START;
	break;
      case '\r':
	break;
      case '\n':
      case '\t':
	fcp = -2;
	replace = ' ';
	break;
      default:
	fcp = 0;
      }
      break;
    case POSSIBLE_START:
      if(in[i] == '*'){
	//_a2l_verbose("Detect comment start\n");
	stm = INCOMMENT;
      } else {
	//_a2l_verbose("No comment, back to normal\n");
	stm = NORMAL;
	fcp = 1;
      }
      break;
    case POSSIBLE_END:
      if(in[i] == '/'){
	//_a2l_verbose("Detect comment end\n");
	stm = NORMAL;
      } else {
	//_a2l_verbose("No end of comment, back to comment\n");
	stm = INCOMMENT;
      }
      break;
    case INCOMMENT:
      if(in[i] == '*'){
	stm = POSSIBLE_END;
      }
      break;
    }
    
    //Copy data
    if(fcp > -1){
      for(j = fcp; j >=0; j--){
	//printf("Copy i: %d, position: %d, fcp: %d, content: %c\n",i,position,fcp,in[i-j]);
	(*out)[position] = in[i-j];
	position++;
      }
    } else {
      switch(fcp){
      case -2:
	(*out)[position] = replace;
	position++;
	break;
      }
    }
  }
  return CS_OK;
}

/********************************************************************/
uint16_t cs_a2l_libgetversion()
/********************************************************************/
{
  return CS_A2L_VERSION_MAGIC;
}

/**************************************************************************/
uint32_t cs_a2l_getuintvalue(
			     cs_parser_result_t result,
			     char *name,
			     uint8_t *error
			     )
/**************************************************************************/
{
  pcre *re = NULL;
  int rv = 0;
  int noresults = 4;
  int ovector[CS_REGEX_RESULTSIZE*noresults];
  char regex[CS_REGEX_BUFFER];
  uint32_t value = 0;
  char namedvalue[CS_PARSER_CHARBUF];
  char namedhex[CS_PARSER_CHARBUF];

  if(error == NULL)
    return value;

  *error = CS_OK;
  
  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER);
  if(name != NULL){
    snprintf(regex,CS_REGEX_BUFFER,"%s[\\s]+(?'hex'(0x)?)(?'value'[0-9a-fA-F]+)",name);
  } else {
    snprintf(regex,CS_REGEX_BUFFER,"[\\s]+(?'hex'(0x)?)(?'value'[0-9a-fA-F]+)");
  }
  //Precompile regex
  if((re = cs_parser_precompile((const char*)regex)) == NULL){
    _a2l_error("Regex precompilation failed\n");
    *error = CS_FAIL;
    return value;
  }

  rv = pcre_exec(re, NULL, (const char*)result.searchin,result.searchlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE*noresults);

  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _a2l_error("No match\n"); 
      *error = CS_FAIL;
      pcre_free(re);
      return value;
      break;
    default: 
      _a2l_error("Matching error %d\n", rv);
      pcre_free(re);
      *error = CS_FAIL;
      return value;
      break;
    }
  } else {
    if(cs_parser_getsubstring(re,result.searchin,ovector,CS_REGEX_RESULTSIZE*noresults,"value",namedvalue)){
      _a2l_error("Cannot find value section\n");
      *error = CS_FAIL;
      pcre_free(re);
      return value;
    }
    _a2l_verbose("Found value: %s\n",namedvalue);
    if(!cs_parser_getsubstring(re,result.searchin,ovector,CS_REGEX_RESULTSIZE*noresults,"hex",namedhex)&&(strcmp(namedhex,"0x")==0)){
      // It is a hex value
      _a2l_verbose("Found hex value\n");
      if(!sscanf(namedvalue,"%X",&value)){
	_a2l_error("Failed to parse substring value: %s\n",namedvalue);
	pcre_free(re);
	*error = CS_FAIL;
	return value;
      }
    } else {
      _a2l_verbose("Found dec value\n");
      if(!sscanf(namedvalue,"%u",&value)){
	_a2l_error("Failed to parse substring value: %s\n",namedvalue);
	pcre_free(re);
	*error = CS_FAIL;
	return value;
      }
    }   
  }
 
  pcre_free(re);   
  return value;
}

/**************************************************************************/
int cs_a2l_getdoublevaluevector(
				cs_parser_result_t result,
				char *name,
				double *value,
				int lenmax
				)
/**************************************************************************/
{
  pcre *re = NULL;
  int rv = 0;
  int noresults = lenmax +2;
  int ovector[CS_REGEX_RESULTSIZE*noresults];
  char regex[CS_REGEX_BUFFER*5];
  char regexcp[CS_REGEX_BUFFER*5];
  char search[CS_PARSER_CHARBUF];
  char namedvalue[CS_PARSER_CHARBUF];
  int i;

  if((name == NULL) || (value == NULL)) {
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER*5);
  if(name != NULL){
    snprintf(regex,CS_REGEX_BUFFER*5,"%s",name);
  }
  for(i = 0; i < lenmax; i++){
    memcpy(regexcp,regex,sizeof(char)*CS_REGEX_BUFFER*5);
    snprintf(regex,CS_REGEX_BUFFER*5,"%s[\\s]+(?'value%d'[0-9\\.-]+)",regexcp,i);
  }

  //Precompile regex
  if((re = cs_parser_precompile((const char*)regex)) == NULL){
    _a2l_error("Regex precompilation failed\n");
    return CS_FAIL;
  }

  rv = pcre_exec(re, NULL, (const char*)result.searchin,result.searchlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE*noresults);

  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _a2l_error("No match\n");
      pcre_free(re);
      return CS_FAIL;
      break;
    default: 
      _a2l_error("Matching error %d\n", rv);
      pcre_free(re);
      return CS_FAIL;
      break;
    }
  } else {
    for(i = 0; i < lenmax; i++){
      //prepare searchstring
      memset(search,0,sizeof(char)*CS_PARSER_CHARBUF);
      snprintf(search,CS_PARSER_CHARBUF,"value%d",i);
      if(cs_parser_getsubstring(re,result.searchin,ovector,CS_REGEX_RESULTSIZE*noresults,search,namedvalue)){
	_a2l_error("Cannot find value %s\n",search);
	pcre_free(re);
	return CS_FAIL;
      }
      if(!sscanf(namedvalue,"%lf",&value[i])){
	_a2l_error("Failed to parse substring value: %s\n",namedvalue);
	pcre_free(re);
	return CS_FAIL;
      }
    }
  }
 
  pcre_free(re);   
  return CS_OK;
}

/**************************************************************************/
int cs_a2l_getstringvalue(
			  cs_parser_result_t result,
			  char *name,
			  char* value,
			  int maxlen
			  )
/**************************************************************************/
{
  pcre *re = NULL;
  int rv = 0;
  int noresults = 4;
  int ovector[CS_REGEX_RESULTSIZE*noresults];
  char regex[CS_REGEX_BUFFER];
  char namedvalue[CS_PARSER_CHARBUF];

  if((name == NULL) || (value == NULL)){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }

  //reset value
  memset(value,0,maxlen*sizeof(char));

  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER);
  snprintf(regex,CS_REGEX_BUFFER,"%s (?'value'[^\\s]+)",name);

  //Precompile regex
  if((re = cs_parser_precompile((const char*)regex)) == NULL){
    _a2l_error("Regex precompilation failed\n");
    return CS_FAIL;
  }

  rv = pcre_exec(re, NULL, (const char*)result.searchin,result.searchlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE*noresults);

  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _a2l_error("No match\n"); 
      pcre_free(re);
      return CS_FAIL;
      break;
    default: 
      _a2l_error("Matching error %d\n", rv);
      pcre_free(re);
      return CS_FAIL;
      break;
    }
  } else {
    if(cs_parser_getsubstring(re,result.searchin,ovector,CS_REGEX_RESULTSIZE*noresults,"value",namedvalue)){
      _a2l_error("Cannot find value section\n");
      pcre_free(re);
      return CS_FAIL;
    }
    _a2l_verbose("Found value: %s\n",namedvalue);
  }
  strncpy(value,namedvalue,maxlen);
  pcre_free(re);   
  return CS_OK;
}

/**************************************************************************/
int cs_a2l_getstringvaluequoted(
				cs_parser_result_t result,
				char *name,
				char* value,
				int maxlen
				)
/**************************************************************************/
{
  pcre *re = NULL;
  int rv = 0;
  int noresults = 4;
  int ovector[CS_REGEX_RESULTSIZE*noresults];
  char regex[CS_REGEX_BUFFER];
  char namedvalue[CS_PARSER_CHARBUF];

  if((name == NULL) || (value == NULL)){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }

  //reset value
  memset(value,0,maxlen*sizeof(char));

  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER);
  snprintf(regex,CS_REGEX_BUFFER,"%s (?'value'\"[^\"]+\")",name);

  //Precompile regex
  if((re = cs_parser_precompile((const char*)regex)) == NULL){
    _a2l_error("Regex precompilation failed\n");
    return CS_FAIL;
  }

  rv = pcre_exec(re, NULL, (const char*)result.searchin,result.searchlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE*noresults);

  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _a2l_error("No match\n"); 
      pcre_free(re);
      return CS_FAIL;
      break;
    default: 
      _a2l_error("Matching error %d\n", rv);
      pcre_free(re);
      return CS_FAIL;
      break;
    }
  } else {
    if(cs_parser_getsubstring(re,result.searchin,ovector,CS_REGEX_RESULTSIZE*noresults,"value",namedvalue)){
      _a2l_error("Cannot find value section\n");
      pcre_free(re);
      return CS_FAIL;
    }
    _a2l_verbose("Found value: %s\n",namedvalue);
  }
  strncpy(value,namedvalue,maxlen);
  pcre_free(re);   
  return CS_OK;
}

/**************************************************************************/
int cs_a2l_getlookupvalue(
			  cs_parser_result_t *result,
			  cs_a2l_compu_tab_type_t ttype,
			  cs_a2l_compu_vtab_elem_t *elem
			  )
/**************************************************************************/
{
  pcre *re = NULL;
  int rv = 0;
  int noresults = 5;
  int ovector[CS_REGEX_RESULTSIZE*noresults];
  char regex[CS_REGEX_BUFFER];
  char namedvalue[CS_PARSER_CHARBUF];

  if((result == NULL) ||(elem == NULL)){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }

  //reset value
  memset(elem->label,0,CS_CHARBUF*sizeof(char));
  elem->id_start = 0;
  elem->id_end = 0;

  //cs_parser_printresult(result);

  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER);
  switch(ttype){
  case A2L_COMPU_TAB_TYPE_UNKNOWN:
    return CS_FAIL;
    break;
  case A2L_COMPU_TAB_TYPE_VTAB:
    snprintf(regex,CS_REGEX_BUFFER,"(?'all'(?'value'[0-9]+)[\\s]+\"(?'label'(.*?))\")");
    break;
  case A2L_COMPU_TAB_TYPE_VTAB_RANGE:
    snprintf(regex,CS_REGEX_BUFFER,"(?'all'(?'value'[0-9]+)[\\s](?'valueend'[0-9]+)[\\s]+\"(?'label'(.*?))\")");
    break;
  }


  //Precompile regex
  if((re = cs_parser_precompile((const char*)regex)) == NULL){
    _a2l_error("Regex precompilation failed\n");
    return CS_FAIL;
  }
  
  rv = pcre_exec(re, NULL, (const char*)result->result,result->resultlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE*noresults);

  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _a2l_error("No match\n"); 
      pcre_free(re);
      return CS_FAIL;
      break;
    default: 
      _a2l_error("Matching error %d\n", rv);
      pcre_free(re);
      return CS_FAIL;
      break;
    }
  } else {
    if(cs_parser_getsubstring_int(re,result->result,ovector,CS_REGEX_RESULTSIZE*noresults,"value",&elem->id_start,0)){
      _a2l_error("Failed to get value \n");
      pcre_free(re);
      return CS_FAIL;
    }
    if(ttype == A2L_COMPU_TAB_TYPE_VTAB_RANGE){
      if(cs_parser_getsubstring_int(re,result->result,ovector,CS_REGEX_RESULTSIZE*noresults,"valueend",&elem->id_end,0)){
	_a2l_error("Failed to get value end \n");
	pcre_free(re);
	return CS_FAIL;
      }
    }
    if(cs_parser_getsubstring(re,result->result,ovector,CS_REGEX_RESULTSIZE*noresults,"label",namedvalue)){
      _a2l_error("Cannot find label section\n");
      pcre_free(re);
      return CS_FAIL;
    }
    strncpy(elem->label,namedvalue,CS_CHARBUF);
  }
  //Modify search area for next search
  cs_parser_stepnext(result,ovector,CS_REGEX_RESULTSIZE*noresults,1);

  pcre_free(re);   
  return CS_OK;
}

/**************************************************************************/
int cs_a2l_getflag(
		   cs_parser_result_t result,
		   char *name
		   )
/**************************************************************************/
{
  pcre *re = NULL;
  int rv = 0;
  int noresults = 2;
  int ovector[CS_REGEX_RESULTSIZE*noresults];
  char regex[CS_REGEX_BUFFER];
  char namedvalue[CS_PARSER_CHARBUF];

  if(name == NULL){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }

  
  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER);
  snprintf(regex,CS_REGEX_BUFFER,"(?'flag'%s)",name);

  //Precompile regex
  if((re = cs_parser_precompile((const char*)regex)) == NULL){
    _a2l_error("Regex precompilation failed\n");
    return CS_FAIL;
  }

  rv = pcre_exec(re, NULL, (const char*)result.searchin,result.searchlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE*noresults);

  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _a2l_error("No match\n"); 
      pcre_free(re);
      return CS_FAIL;
      break;
    default: 
      _a2l_error("Matching error %d\n", rv);
      pcre_free(re);
      return CS_FAIL;
      break;
    }
  } else {
    if(cs_parser_getsubstring(re,result.searchin,ovector,CS_REGEX_RESULTSIZE*noresults,"flag",namedvalue)){
      pcre_free(re);
      return CS_FAIL;
    } else {
      pcre_free(re);   
      return CS_OK;
    }
  } 
  pcre_free(re);   
  return CS_FAIL;
}


/**************************************************************************/
void cs_a2l_GetByteOrder(
			 cs_parser_result_t lresult,
			 cs_a2l_bo_t defaultBO, 
			 cs_a2l_bo_t *BO
			 )
/**************************************************************************/
{
  char ordervalue[CS_PARSER_CHARBUF];

  if(BO == NULL){
    _a2l_error("Parameter failure\n");
    return;
  } 
  
  if(cs_a2l_getstringvalue(lresult,"BYTE_ORDER",ordervalue,CS_PARSER_CHARBUF)){
    //Cannot find Byte order for signal, try to set to default value
    *BO = defaultBO;
  } else {
    if(strcmp(ordervalue,"MSB_LAST") == 0){
      *BO = A2L_BO_INTEL;
    } else {
      *BO = A2L_BO_MOTOROLLA;
    }
  }
}

/**************************************************************************/
void cs_a2l_GetTsUnit(
		      cs_parser_result_t ltime,
		      cs_a2l_ts_unit_t *value
		      )
/**************************************************************************/
{
  *value = (!cs_a2l_getflag(ltime,"UNIT_1NS")?A2L_TS_UNIT_1NS:*value);
  *value = (!cs_a2l_getflag(ltime,"UNIT_10NS")?A2L_TS_UNIT_10NS:*value);
  *value = (!cs_a2l_getflag(ltime,"UNIT_100NS")?A2L_TS_UNIT_100NS:*value);
  *value = (!cs_a2l_getflag(ltime,"UNIT_1US")?A2L_TS_UNIT_1US:*value);
  *value = (!cs_a2l_getflag(ltime,"UNIT_10US")?A2L_TS_UNIT_10US:*value);
  *value = (!cs_a2l_getflag(ltime,"UNIT_100US")?A2L_TS_UNIT_100US:*value);
  *value = (!cs_a2l_getflag(ltime,"UNIT_1MS")?A2L_TS_UNIT_1MS:*value);
  *value = (!cs_a2l_getflag(ltime,"UNIT_10MS")?A2L_TS_UNIT_10MS:*value);
  *value = (!cs_a2l_getflag(ltime,"UNIT_100MS")?A2L_TS_UNIT_100MS:*value);
  *value = (!cs_a2l_getflag(ltime,"UNIT_1S")?A2L_TS_UNIT_1S:*value);
}

/**************************************************************************/
int cs_a2l_StringToType(
			char* typestring,
			cs_a2l_ag_t AG,
			cs_a2l_data_type_t *dtype,
			int *agsize
			)
/**************************************************************************/
{
  int found = 0;

  if((typestring == NULL)|| (dtype == NULL) || (agsize == NULL)){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  if(strcmp(typestring,"SBYTE") == 0){
    *dtype = A2L_SBYTE;
    *agsize = 1 / cs_a2l_get_ag_bytes(AG);
    found = 1;
  }

  if(strcmp(typestring,"UBYTE") == 0){
    *dtype = A2L_UBYTE;
    *agsize = 1 / cs_a2l_get_ag_bytes(AG);
    found = 1;
  }
  
  if(strcmp(typestring,"SWORD") == 0){
    *dtype = A2L_SWORD;
    *agsize = 2 / cs_a2l_get_ag_bytes(AG);
    found = 1;
  }

  if(strcmp(typestring,"UWORD") == 0){
    *dtype = A2L_UWORD;
    *agsize = 2 / cs_a2l_get_ag_bytes(AG);
    found = 1;
  }

  if(strcmp(typestring,"SLONG") == 0){
    *dtype = A2L_SINT;
    *agsize = 4 / cs_a2l_get_ag_bytes(AG);
    found = 1;
  }

  if(strcmp(typestring,"ULONG") == 0){
    *dtype = A2L_UINT;
    *agsize = 4 / cs_a2l_get_ag_bytes(AG);
    found = 1;
  }

  if(strcmp(typestring,"A_UINT64") == 0){
    *dtype = A2L_UINT64;
    *agsize = 8 / cs_a2l_get_ag_bytes(AG);
    found = 1;
  }

  if(strcmp(typestring,"A_INT64") == 0){
    *dtype = A2L_SINT64;
    *agsize = 8 / cs_a2l_get_ag_bytes(AG);
    found = 1;
  }

  if(strcmp(typestring,"FLOAT32_IEEE") == 0){
    *dtype = A2L_FLOAT;
    *agsize = 4 / cs_a2l_get_ag_bytes(AG);
    found = 1;
  }

  if(strcmp(typestring,"FLOAT64_IEEE") == 0){
    *dtype = A2L_FLOAT64;
    *agsize = 6 / cs_a2l_get_ag_bytes(AG);
    found = 1;
  }

  if(found == 0){
    return CS_FAIL;
  }

  return CS_OK;
}

/**************************************************************************/
char *cs_a2l_TypeToString(
			  cs_a2l_data_type_t dtype
			  )
/**************************************************************************/
{

  switch(dtype){
  case A2L_SBYTE:
    return "SBYTE";
  case A2L_UBYTE:
    return "UBYTE";
  case A2L_SWORD:
    return "SWORD";
  case A2L_UWORD:
    return "UWORD";
  case A2L_SINT:
    return "SLONG";
  case A2L_UINT:
    return "ULONG";
  case A2L_SINT64:
    return "SLONG64";
  case A2L_UINT64:
    return "ULONG64";
  case A2L_FLOAT:
    return "FLOAT32";
  case A2L_FLOAT64:
    return "FLOAT64";
  default:
    return "Unknown";
  }
}

/**************************************************************************/
int cs_a2l_TypeToAG(
		    cs_a2l_data_type_t dtype
		    )
/**************************************************************************/
{

  switch(dtype){
  case A2L_SBYTE:
  case A2L_UBYTE:
    return 1;
  case A2L_SWORD:
  case A2L_UWORD:
    return 2;
  case A2L_SINT:
  case A2L_UINT:
    return 4;
  case A2L_FLOAT:
    return 4;
  default:
    return 0;
  }
}

/** Get number of bytes of value representation depending on AG
    \retval number of bytes
*/
/**************************************************************************/
int cs_a2l_get_ag_bytes( cs_a2l_ag_t AG)
/**************************************************************************/
{
  switch(AG){
  case  A2L_AG_BYTE:
    return 1;
  case  A2L_AG_WORD:
    return 2;
  case  A2L_AG_DWORD:
    return 4;
  case  A2L_AG_RESERVED:
    return 1;
  default:
    return 1;
  }
}

/**************************************************************************/
int cs_a2l_getsectionheader(
			    cs_parser_result_t *result,
			    char *name,
			    char *label
			    )
/**************************************************************************/
{
  pcre *re = NULL;
  int rv = 0;
  
  int noresults = 6;
  int ovector[CS_REGEX_RESULTSIZE*noresults];
  cs_parser_result_t lresult;

  if((result == NULL) || (name == NULL) || (label == NULL)){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  //Precompile regex

  if((re = cs_parser_precompile((const char *)"/begin [A-Z_]+[\\s]+(\")?(?'name'[a-z A-Z0-9_\\.\\[\\]]+[a-zA-Z0-9_\\.\\[\\]]+)(\")?[\\s]?(\"(?'label'(.*?))\")?")) == NULL){
    _a2l_error("Regex precompilation failed\n");
    return CS_FAIL;
  }

  //Define search area
  lresult.searchin = result->result;
  lresult.searchlen = result->resultlen;

  //Debug output
  //cs_parser_printresult(result);

  //Get data, only one fit is allowed
  rv = pcre_exec(re, NULL, (const char*)lresult.searchin,lresult.searchlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE*noresults);

  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _a2l_error("No match\n"); 
      break;
    default: 
      _a2l_error("Matching error %d\n", rv);
      pcre_free(re);
      return CS_FAIL;
      break;
    }
  } else {

    if(cs_parser_getsubstring(re,lresult.searchin,ovector,CS_REGEX_RESULTSIZE*noresults,"name",name)){
      _a2l_verbose("Failed to get substring name\n");
    }

    if(cs_parser_getsubstring(re,lresult.searchin,ovector,CS_REGEX_RESULTSIZE*noresults,"label",label)){
      _a2l_verbose("Failed to get substring label\n");
    }    
  }
 
  pcre_free(re);   
  return CS_OK;
}

/**************************************************************************/
int cs_a2l_search_section(
			  cs_parser_result_t *result,
			  char *sectionname,
			  char *sectionsubname
			  )
/**************************************************************************/
{
  pcre *re = NULL;
  char regex[CS_REGEX_BUFFER];
  char subname_clean[CS_REGEX_BUFFER];
  int rv = 0;
  int ovector[CS_REGEX_RESULTSIZE];

  if((sectionname == NULL) ||
     (result == NULL)){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }

  //Reset result
  result->result = NULL;
  result->resultlen = 0;
  result->resultoffset = 0;

  cs_parser_preparestring(sectionsubname,subname_clean,CS_REGEX_BUFFER);

  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER);
  if(sectionsubname == NULL){
    snprintf(regex,CS_REGEX_BUFFER,"(/begin %s)(.*?)(/end %s)",sectionname,sectionname);
  } else {
    if(strlen(subname_clean) == 0){ //""
      snprintf(regex,CS_REGEX_BUFFER,"(/begin %s[\\s]+[\"]?[\"]?)(.*?)(/end %s)",sectionname,sectionname);
    } else {
      snprintf(regex,CS_REGEX_BUFFER,"(/begin %s[\\s]+[\"]?%s[\"]?)(.*?)(/end %s)",sectionname,subname_clean,sectionname);
    }
  }

  //_a2l_verbose("Search Section: regex:%s\n",regex);
  //Precompile regex
  if((re = cs_parser_precompile((const char *)regex)) == NULL){
    _a2l_error("Regex precompilation failed\n");
    return CS_FAIL;
  }

  //Get data, only one fit is allowed
  rv = pcre_exec(re, NULL, (const char*)result->searchin,result->searchlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE);

  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _a2l_error("No match\n");
      return CS_FAIL;
      break;
    default: 
      _a2l_error("Matching error %d\n", rv);
      pcre_free(re);
      return CS_FAIL;
      break;
    }
  } else {
    result->result = &result->searchin[ovector[0]];
    result->resultlen = ovector[1] - ovector[0];
    result->resultoffset = ovector[0];
  }

  //Calc next values
  cs_parser_calcresult(result);

  pcre_free(re); 

  return CS_OK;
}


/**************************************************************************/
int cs_a2l_list_sections(
			 cs_parser_result_t *result,
			 cs_a2l_list_t *list,
			 char *sectiontype
			 )
/**************************************************************************/
{
  cs_parser_result_t lresult;
  int counter = 0;
  int i;

  if((list == NULL) ||
     (sectiontype == NULL) ||
     (result == NULL)){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  //Reset storage structure
  list->no_result = 0;
  list->resultname = NULL;
  list->resultlabel = NULL;
  list->result = NULL;
  memset(list->sectiontype,0,CS_PARSER_CHARBUF);

  //Store section type
  strncpy(list->sectiontype,sectiontype,CS_PARSER_CHARBUF);

  //Init local result  values
  lresult.searchin = result->searchin;
  lresult.searchlen = result->searchlen;

  //First loop over section to count number of sections
  while(!cs_a2l_search_section(&lresult,sectiontype,NULL)){
    if(lresult.result != NULL)
      list->no_result++;
    if(lresult.next == NULL){ //Final entry, no more entries to search
      break;
    } else {
      lresult.searchin = lresult.next;
      lresult.searchlen = lresult.nextlen; 
    }
  }

  if(list->no_result == 0)
    return CS_OK;

  list->result = (cs_parser_result_t*)malloc(sizeof(cs_parser_result_t) * list->no_result);

  if(list->result == NULL){
    list->no_result = 0;
    return CS_FAIL;
  }

  list->resultlabel = (char **)malloc(sizeof(char*) * list->no_result);
  if(list->resultlabel == NULL){
    _a2l_error("Malloc failure\n");
    cs_a2l_list_free(list);
    return CS_FAIL;
  }

  for(i = 0; i < list->no_result; i++){
    list->resultlabel[i] = (char*)malloc(sizeof(char*) * CS_PARSER_CHARBUF);
    if(list->resultlabel[i] == NULL){
      _a2l_error("Malloc failure\n");
      cs_a2l_list_free(list);
      return CS_FAIL;
    }
    memset(list->resultlabel[i],0,sizeof(char*) * CS_PARSER_CHARBUF);
  }

  
  list->resultname = (char **)malloc(sizeof(char*) * list->no_result);

  if(list->resultname == NULL){
    cs_a2l_list_free(list);
    return CS_FAIL;
  }

  for(i = 0; i < list->no_result; i++){
    list->resultname[i] = (char*)malloc(sizeof(char*) * CS_PARSER_CHARBUF);
    if(list->resultname[i] == NULL){
      _a2l_error("Malloc failure\n");
      cs_a2l_list_free(list);
      return CS_FAIL;
    }
    memset(list->resultname[i],0,sizeof(char*) * CS_PARSER_CHARBUF);
  }

  //Second loop over section
  lresult.searchin = result->searchin;
  lresult.searchlen = result->searchlen;

  //Store section type
  strncpy(list->sectiontype,sectiontype,CS_PARSER_CHARBUF);

  counter = 0;
  while(!cs_a2l_search_section(&lresult,sectiontype,NULL)){
    //Store values
    list->result[counter] = lresult;
    if(cs_a2l_getsectionheader(&lresult, list->resultname[counter], list->resultlabel[counter])){
      _a2l_error("Failed to get Headername and Headerlabel\n");
    } else {
      //_a2l_verbose("Found section header, type: %s, name: %s, label %s \n",sectiontype,list->resultname[counter],list->resultlabel[counter]);
    }
    
    counter++;

    if(counter >= list->no_result)
      break;

    if(lresult.next == NULL){ //Final entry, no more entries to search
      break;
    } else {
      lresult.searchin = lresult.next;
      lresult.searchlen = lresult.nextlen; 
    }
  }

  return CS_OK;
}

/**************************************************************************/
int cs_a2l_list_free(
		     cs_a2l_list_t *list
		     )
/**************************************************************************/
{
  int i;

  if(list == NULL){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  if(list->result){
    free(list->result);
    list->result = NULL;
  }

  if(list->resultlabel){
    for(i = 0; i < list->no_result; i++){
      if(list->resultlabel[i]){
	free(list->resultlabel[i]);
	list->resultlabel[i] = NULL;
      }
    }
    free(list->resultlabel);
    list->resultlabel = NULL;
  }
    
  if(list->resultname){
    for(i = 0; i < list->no_result; i++){
      if(list->resultname[i]){
	free(list->resultname[i]);
	list->resultname[i] = NULL;
      }
    }
    free(list->resultname);
    list->resultname = NULL;
  }

  list->no_result = 0;
  memset(list->sectiontype,0,CS_PARSER_CHARBUF);

  return CS_OK;
}

/**************************************************************************/
void cs_a2l_reset_measurement(
			      cs_a2l_measurement_t *me
			      )
/**************************************************************************/
{
  if(!me){
    _a2l_error("Parameter failure\n");
    return;
  }
  
  memset(me->name,0,sizeof(char)*CS_CHARBUF);
  memset(me->comment,0,sizeof(char)*CS_CHARBUF);
  memset(me->conv,0,sizeof(char)*CS_CHARBUF);
  me->BO = A2L_BO_UNKNOWN;
  me->ev_idx = -1;
  me->prescaler = 1;

  me->bitoffset = 0;
  me->address = 0;
  me->extension = 0;
  me->size = 0;

  memset(&me->cm,0,sizeof(cs_a2l_compu_method_t));
  //Prepare for linear conversion as default
  me->cm.coeff[0] = 1.0; //scale
  me->cm.coeff[1] = 0.0; //offset
  me->cm.algo = A2L_CM_LINEAR; 
  me->dtype = A2L_UINT;
  me->exvalue = NULL;
  me->exts = NULL;
  me->extcallback = NULL;
  me->extcallbackpriv = NULL;
  me->min = 0;
  me->max = 0;
}


/**************************************************************************/
void cs_a2l_init(
		 cs_a2l_t *data
		 )
/**************************************************************************/
{
  int i;
  if(data == NULL){
    _a2l_error("Parameter failure\n");
    return;
  }

  //Reset storage structure
  memset(data,0,sizeof(cs_a2l_t));
  //Reset module storage
  cs_parser_resetresult(&data->module);
  cs_parser_resetresult(&data->iface);
  cs_parser_resetresult(&data->protocol);
  for(i = 0; i < CS_A2L_SIGNALMAX; i++){
    cs_a2l_reset_measurement(&data->signals[i]);
  }
  data->signalcounter = 0;
  data->AG = A2L_AG_BYTE;
  data->BO = A2L_BO_UNKNOWN;
}


/**************************************************************************/
int cs_a2l_open(
		cs_a2l_t *data,
		char *filename
		)
/**************************************************************************/
{

  if((data == NULL) || (filename == NULL)){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }

  if(data->map != NULL){
    _a2l_error("There is already an opened file\n");
    return CS_FAIL;
  }

  cs_a2l_init(data);

  data->filename = filename;

  if(cs_parser_open(data->filename,&data->fs,&data->filelen,&data->map)){
    return CS_FAIL;
  }

  //Cleanup comments
  if(cs_a2l_cleanup((char*)data->map,&data->buffer,data->filelen)){
    return CS_FAIL;
  }
  
  return CS_OK;
}

/**************************************************************************/
int cs_a2l_close(
		 cs_a2l_t *data
		 )
/**************************************************************************/
{
  if(data == NULL){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }


  if(cs_parser_close(data->fs,data->map,data->filelen)){
    return CS_FAIL;
  }

  if(data->buffer!= NULL)
    free(data->buffer);

  //reinit structure
  cs_a2l_init(data);

  return CS_OK;
}

/**************************************************************************/
int cs_a2l_getversion(
		      cs_a2l_t *data,
		      int *major,
		      int *minor
		      )
/**************************************************************************/
{
  pcre *re = NULL;
  char regex[CS_REGEX_BUFFER];
  int rv = 0;
  int noresult = 3;
  int ovector[CS_REGEX_RESULTSIZE*noresult];

  if((data == NULL) ||
     (major == NULL) ||
     (minor == NULL)){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }

  //Reset result
  *major = 0;
  *minor = 0;

  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER);
  snprintf(regex,CS_REGEX_BUFFER,"ASAP2_VERSION (?'major'[0-9]+) (?'minor'[0-9]+)");

  //Precompile regex
  if((re = cs_parser_precompile((const char *)regex)) == NULL){
    _a2l_error("Regex precompilation failed\n");
    return CS_FAIL;
  }

  //Get data, only one fit is allowed
  rv = pcre_exec(re, NULL, (const char*)data->buffer,data->filelen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE * noresult);

  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _a2l_error("No match\n"); 
      break;
    default: 
      _a2l_error("Matching error %d\n", rv);
      pcre_free(re);
      return CS_FAIL;
      break;
    }
  } else {

    if(cs_parser_getsubstring_int(re,data->buffer,ovector,CS_REGEX_RESULTSIZE*noresult,"major",major,0)){
      _a2l_error("Failed to get substring major\n");
      pcre_free(re);
      return CS_FAIL;
    }

    if(cs_parser_getsubstring_int(re,data->buffer,ovector,CS_REGEX_RESULTSIZE*noresult,"minor",minor,0)){
      _a2l_error("Failed to get substring minor\n");
      pcre_free(re);
      return CS_FAIL;
    }
    
  }
  pcre_free(re); 

  return CS_OK;
}

/**************************************************************************/
int cs_a2l_GetProject(
		      cs_a2l_t *data,
		      char * project
		      )
/**************************************************************************/
{
  if((data == NULL) ||
     (project == NULL)){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  data->project.searchin = data->buffer;
  data->project.searchlen = data->filelen;
  if(cs_a2l_search_section(&data->project,"PROJECT",project)){
    _a2l_error("Regex failure\n");
    return CS_FAIL;
  }

  if(data->project.result == NULL){
    _a2l_error("Cannot find PROJECT %s\n",project);
    return CS_FAIL;
  }

  return CS_OK;
}

/**************************************************************************/
int cs_a2l_GetModule(
		     cs_a2l_t *data,
		     char * module
		     )
/**************************************************************************/
{
  if((data == NULL) ||
     (module == NULL)){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  data->module.searchin = data->project.result;
  data->module.searchlen = data->project.resultlen;

  //cs_parser_printresult(&data->project);

  if(cs_a2l_search_section(&data->module,"MODULE",module)){
    _a2l_error("Regex failure\n");
    return CS_FAIL;
  }

  if(data->module.result == NULL){
    _a2l_error("Cannot find MODULE %s\n",module);
    return CS_FAIL;
  }

  return CS_OK;
}


/**************************************************************************/
int cs_a2l_GetAddressMapping(
			     cs_parser_result_t *result,
			     cs_a2l_am_t *entry
			     )
/**************************************************************************/
{
  pcre *re = NULL;
  int rv = 0;
  int noresults = 5;
  int ovector[CS_REGEX_RESULTSIZE*noresults];
  char regex[CS_REGEX_BUFFER];
  int tmp;

  if((result == NULL) ||(entry == NULL)){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }
  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER);
  snprintf(regex,CS_REGEX_BUFFER,"(?'all'ADDRESS_MAPPING[\\s]+(?'start'0x[0-9a-fA-F]+)[\\s]+(?'offset'0x[0-9a-fA-F]+)[\\s]+(?'len'0x[0-9a-fA-F]+))");

  //Precompile regex
  if((re = cs_parser_precompile((const char*)regex)) == NULL){
    _a2l_error("Regex precompilation failed\n");
    return CS_FAIL;
  }
  
  rv = pcre_exec(re, NULL, (const char*)result->result,result->resultlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE*noresults);

  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _a2l_error("No match\n"); 
      pcre_free(re);
      return CS_FAIL;
      break;
    default: 
      _a2l_error("Matching error %d\n", rv);
      pcre_free(re);
      return CS_FAIL;
      break;
    }
  } else {
    if(cs_parser_getsubstring_hex(re,result->result,ovector,CS_REGEX_RESULTSIZE*noresults,"start",&tmp,0)){
      _a2l_error("Failed to get start \n");
      pcre_free(re);
      return CS_FAIL;
    }
    entry->start = (uint32_t)tmp;
    if(cs_parser_getsubstring_hex(re,result->result,ovector,CS_REGEX_RESULTSIZE*noresults,"offset",&tmp,0)){
      _a2l_error("Failed to get offset \n");
      pcre_free(re);
      return CS_FAIL;
    }
    entry->offset = (uint32_t)tmp;
    if(cs_parser_getsubstring_hex(re,result->result,ovector,CS_REGEX_RESULTSIZE*noresults,"len",&tmp,0)){
      _a2l_error("Failed to get len \n");
      pcre_free(re);
      return CS_FAIL;
    }
    entry->len = (uint32_t)tmp;
  }
  //Modify search area for next search
  cs_parser_stepnext(result,ovector,CS_REGEX_RESULTSIZE*noresults,1);

  pcre_free(re);   
  return CS_OK;
}

/**************************************************************************/
int cs_a2l_GetModPar(
		     cs_a2l_t *data,
		     char * name
		     )
/**************************************************************************/
{
  cs_parser_result_t result;
  cs_parser_result_t resultiface;

  if((data == NULL) ||
     (name == NULL)){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }

  result.searchin = data->module.result;
  result.searchlen = data->module.resultlen;

  if(cs_a2l_search_section(&result,"MOD_PAR",name)){
    _a2l_error("Regex failure\n");
    return CS_FAIL;
  }

  if(result.result == NULL){
    _a2l_error("Cannot find MOD_PAR %s\n",name);
    return CS_FAIL;
  }

  //cs_parser_printresult(&result);
  
  //TODO get Memory segments

  //Look for Interface sections
  resultiface.searchin = result.result;
  resultiface.searchlen = result.resultlen;

  if(cs_a2l_search_section(&resultiface,"IF_DATA",data->ifacename)){
    _a2l_error("Regex failure\n");
    return CS_FAIL;
  }

  data->modpar.am_count = 0;

  while(cs_a2l_GetAddressMapping(&resultiface,&data->modpar.am[data->modpar.am_count]) == CS_OK){
    
    data->modpar.am_count++;

    if(data->modpar.am_count >= CS_A2L_AM_SIZEMAX){
      _a2l_error("Adress mapping overrange\n");
      data->modpar.am_count = CS_A2L_AM_SIZEMAX -1;
      break;
    }
  }

  //cs_a2l_print_modpar(&data->modpar);

  return CS_OK;
}


/**************************************************************************/
uint32_t cs_a2l_CalcAddress(
			    cs_a2l_t *data,
			    uint32_t address
			    )
/**************************************************************************/
{
  int i;
  uint32_t rv = address;
  cs_a2l_am_t *map = NULL;
  if(data == NULL){
    _a2l_error("Parameter failure\n");
    return rv;
  }

  for(i = 0; i < data->modpar.am_count; i++){
    map = &data->modpar.am[i];
    if((address >= map->start) && 
       (address < (map->start + map->len))){
      //Found
      rv = address - map->start + map->offset;
      return rv;
    }
  }
  return rv;
}

/**************************************************************************/
int cs_a2l_GetModCommon(
			cs_a2l_t *data
			)
/**************************************************************************/
{
  cs_parser_result_t result;

  if(data == NULL){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  result.searchin = data->module.result;
  result.searchlen = data->module.resultlen;
  if(cs_a2l_search_section(&result,"MOD_COMMON",NULL)){
    _a2l_error("Regex failure\n");
    return CS_FAIL;
  }

  if(result.result == NULL){
    _a2l_error("Cannot find MOD_COMMON\n");
    return CS_FAIL;
  }

  cs_a2l_GetByteOrder(result,A2L_BO_UNKNOWN,&data->BO);

  return CS_OK;
}


/**************************************************************************/
int cs_a2l_GetInterface(
			cs_a2l_t *data,
			char * iface
			)
/**************************************************************************/
{
  if((data == NULL) ||
     (iface == NULL)){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  data->iface.searchin = data->module.result;
  data->iface.searchlen = data->module.resultlen;
  if(cs_a2l_search_section(&data->iface,"IF_DATA",iface)){
    _a2l_error("Regex failure\n");
    return CS_FAIL;
  }

  if(data->iface.result == NULL){
    _a2l_error("Cannot find IF_DATA %s\n",iface);
    return CS_FAIL;
  }
  
  //_a2l_verbose("Found Interface IF_DATA %s\n",iface);

  //Search for fix PROTOCOL_LAYER section
  data->protocol.searchin = data->iface.result;
  data->protocol.searchlen = data->iface.resultlen;
  if(cs_a2l_search_section(&data->protocol,"[PROTOCOL_LAYER|TP_BLOB]",NULL)){
    _a2l_error("Regex failure\n");
    return CS_FAIL;
  }

  if(data->protocol.result == NULL){
    _a2l_error("Cannot find PROTOCOL_LAYER\n");
    return CS_FAIL;
  }

  //Store interface name
  strncpy(data->ifacename,iface,sizeof(char)*CS_CHARBUF);

  return CS_OK;
}

/**************************************************************************/
int cs_a2l_GetMeasurement(
			  cs_a2l_t *data,
			  cs_a2l_measurement_t *me,
			  char *name
			  )
/**************************************************************************/
{
  cs_parser_result_t lmeasure;
  cs_parser_result_t lresult;
  uint8_t rvalue = CS_OK;
  pcre *re = NULL;
  char regex[CS_REGEX_BUFFER];
  int rv = 0;
  int noresult = 14;
  int ovector[CS_REGEX_RESULTSIZE*noresult];
  char namedvalue[CS_PARSER_CHARBUF];
  uint32_t lvalue;


  if((data == NULL) ||
     (me == NULL) ||
     (name == NULL)){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  //reset
  cs_a2l_reset_measurement(me);


  //Search daq section
  lmeasure.searchin = data->module.result;
  lmeasure.searchlen = data->module.resultlen;
  if(cs_a2l_search_section(&lmeasure,"MEASUREMENT",name)){
    _a2l_error("Regex failure searching MEASUREMENT %s\n",name);
    return CS_FAIL;
  }

  if(lmeasure.result == NULL){
    _a2l_error("Cannot find MEASUREMENT %s\n",name);
    return CS_FAIL;
  }
  //cs_parser_printresult(&lmeasure);

  //define serachregion
  lresult.searchin = lmeasure.result;
  lresult.searchlen = lmeasure.resultlen;

  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER);
  snprintf(regex,CS_REGEX_BUFFER,"MEASUREMENT[\\s]+(?'name'(.*?))[\\s]+\"(?'comment'(.*?))\"[\\s]+(/\\*.*?\\*/[\\s]+)?(?'type'[A-Z]+)[\\s]+(?'conv'(.*?))[\\s]+(?'res'[0-9]+)[\\s]+(?'aq'[0-9]+)[\\s]+(?'min'[-0-9\\.]+)[\\s]+(?'max'[-0-9\\.]+)");

  //Precompile regex
  if((re = cs_parser_precompile((const char *)regex)) == NULL){
    _a2l_error("Regex precompilation failed\n");
    return CS_FAIL;
  }
  //Get data, only one fit is allowed
  rv = pcre_exec(re, NULL, (const char*)lmeasure.result,lmeasure.resultlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE * noresult);
  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _a2l_error("No content match for Signal %s\n",name);
      //cs_parser_printresult(&lmeasure);
      pcre_free(re);
      return CS_FAIL;
      break;
    default: 
      _a2l_error("Matching error %d\n", rv);
      pcre_free(re);
      return CS_FAIL;
      break;
    }
  } else {
    
    if(cs_parser_getsubstring(re,lmeasure.result,ovector,CS_REGEX_RESULTSIZE*noresult,"name",namedvalue)){
      _a2l_error("Failed to get substring name\n");
      pcre_free(re);
      return CS_FAIL;
    }
    strncpy(me->name,namedvalue,CS_CHARBUF);

    if(cs_parser_getsubstring(re,lmeasure.result,ovector,CS_REGEX_RESULTSIZE*noresult,"comment",namedvalue)){
      _a2l_error("Failed to get substring comment\n");
      pcre_free(re);
      return CS_FAIL;
    }
    strncpy(me->comment,namedvalue,CS_CHARBUF);

    if(cs_parser_getsubstring(re,lmeasure.result,ovector,CS_REGEX_RESULTSIZE*noresult,"type",namedvalue)){
      _a2l_error("Failed to get substring type\n");
      pcre_free(re);
      return CS_FAIL;
    }
    
    //Set default values
    me->dtype = A2L_SBYTE;
    me->size = 1;

    if(cs_a2l_StringToType(namedvalue,data->AG,&me->dtype,&me->size)){
      _a2l_error("Unknown data type\n");
      pcre_free(re);
      return CS_FAIL;
    }
    if(me->size == 0)
      me->size = 1;

    if(cs_parser_getsubstring(re,lmeasure.result,ovector,CS_REGEX_RESULTSIZE*noresult,"conv",namedvalue)){
      _a2l_error("Failed to get substring conv\n");
      pcre_free(re);
      return CS_FAIL;
    }
    strncpy(me->conv,namedvalue,CS_CHARBUF);

    if(cs_parser_getsubstring_int(re,lmeasure.result,ovector,CS_REGEX_RESULTSIZE*noresult,"res",&me->resolution,0)){
      _a2l_error("Failed to get substring res\n");
      pcre_free(re);
      return CS_FAIL;
    }

    if(cs_parser_getsubstring_int(re,lmeasure.result,ovector,CS_REGEX_RESULTSIZE*noresult,"aq",&me->aqurancy,0)){
      _a2l_error("Failed to get substring aq\n");
      pcre_free(re);
      return CS_FAIL;
    }

    if(cs_parser_getsubstring_double(re,lmeasure.result,ovector,CS_REGEX_RESULTSIZE*noresult,"min",&me->min,0)){
      _a2l_error("Failed to get substring min\n");
      pcre_free(re);
      return CS_FAIL;
    }

    if(cs_parser_getsubstring_double(re,lmeasure.result,ovector,CS_REGEX_RESULTSIZE*noresult,"max",&me->max,0)){
      _a2l_error("Failed to get substring max\n");
      pcre_free(re);
      return CS_FAIL;
    }

    cs_a2l_GetByteOrder(lresult,data->BO,&me->BO);

    me->address = cs_a2l_getuintvalue(lresult,"ECU_ADDRESS",&rvalue);
    if(rvalue){
      _a2l_error("Failed to parse ECU_ADDRESS\n");
      return CS_FAIL;
    }
    lvalue = cs_a2l_getuintvalue(lresult,"ECU_ADDRESS_EXTENSION",&rvalue);
    me->extension= (uint8_t)lvalue;

    lvalue = cs_a2l_getuintvalue(lresult,"BITMASK",&rvalue);
    me->bitoffset= (uint8_t)lvalue;

  }
  pcre_free(re);

  if(!cs_a2l_getflag(lresult,"MATRIX_DIM")){
    _a2l_error("Unsupported datatype MATRIX_DIM\n");
    return CS_FAIL;
  }

  //Get optional Byte order
  cs_a2l_GetByteOrder(lresult,data->BO,&me->BO);  

  //Get compu method
  if(cs_a2l_GetCompuMethod(data,&me->cm,me->conv)){
    _a2l_error("Failed to parse Compu method %s\n",me->conv);
    return CS_FAIL;
  }
  //Set min and max values of computation function
  me->cm.min = me->min;
  me->cm.max = me->max;

  //cs_a2l_print_measurement(me);

  return CS_OK;
}


/**************************************************************************/
void cs_a2l_reset_compuvtab(
			    cs_a2l_compu_vtab_t *tab
			    )
/**************************************************************************/
{
  if(tab == NULL){
    _a2l_error("Parameter failure\n");
    return;
  }

  if((tab->number >= 0) && (tab->elems != NULL))
    free(tab->elems);
  
  //reset
  memset(tab,0,sizeof(cs_a2l_compu_vtab_t));
}

/**************************************************************************/
int cs_a2l_GetCompuVTab(
			cs_a2l_t *data,
			cs_a2l_compu_vtab_t *tab,
			char *name
			)
/**************************************************************************/
{
  cs_parser_result_t lcm;
  cs_parser_result_t lcmdefault;
  pcre *re = NULL;
  char regex[CS_REGEX_BUFFER];
  int rv = 0;
  int noresult = 12;
  int ovector[CS_REGEX_RESULTSIZE*noresult];
  char namedvalue[CS_PARSER_CHARBUF];
  int i;


  if((data == NULL) ||
     (tab == NULL) ||
     (name == NULL)){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  tab->number = 0;
  cs_a2l_reset_compuvtab(tab);

  //Search daq section
  lcm.searchin = data->module.result;
  lcm.searchlen = data->module.resultlen;

  if(!cs_a2l_search_section(&lcm,"COMPU_VTAB_RANGE",name)){
    tab->ttype = A2L_COMPU_TAB_TYPE_VTAB_RANGE;
  } else {
    if(!cs_a2l_search_section(&lcm,"COMPU_VTAB",name)){
      tab->ttype = A2L_COMPU_TAB_TYPE_VTAB;
    }
  }

  if(lcm.result == NULL){
    _a2l_error("Cannot find COMPU_VTAB or COMPU_VTAB_RANGE %s\n",name);
    return CS_FAIL;
  }

  lcmdefault.searchin = lcm.result;
  lcmdefault.searchlen = lcm.resultlen;

  //cs_parser_printresult(&lcm);

  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER);
  if(tab->ttype == A2L_COMPU_TAB_TYPE_VTAB)
    snprintf(regex,CS_REGEX_BUFFER,"COMPU_VTAB[\\s]+(?'name'(.*?))[\\s]+\"(?'comment'(.*?))\"[\\s]+(?'type'([A-Z_]+))[\\s]+(?'elems'([0-9]+))");
  if(tab->ttype == A2L_COMPU_TAB_TYPE_VTAB_RANGE)
    snprintf(regex,CS_REGEX_BUFFER,"COMPU_VTAB_RANGE[\\s]+(?'name'(.*?))[\\s]+\"(?'comment'(.*?))\"[\\s]+(?'elems'([0-9]+))");


  //Precompile regex
  if((re = cs_parser_precompile((const char *)regex)) == NULL){
    _a2l_error("Regex precompilation failed\n");
    return CS_FAIL;
  }
  //Get data, only one fit is allowed
  rv = pcre_exec(re, NULL, (const char*)lcm.result,lcm.resultlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE * noresult);

  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _a2l_error("No match\n"); 
      break;
    default: 
      _a2l_error("Matching error %d\n", rv);
      pcre_free(re);
      return CS_FAIL;
      break;
    }
  } else {
    if(cs_parser_getsubstring(re,lcm.result,ovector,CS_REGEX_RESULTSIZE*noresult,"name",namedvalue)){
      _a2l_error("Failed to get substring name\n");
      pcre_free(re);
      return CS_FAIL;
    }
    strncpy(tab->name,namedvalue,CS_CHARBUF);

    if(cs_parser_getsubstring(re,lcm.result,ovector,CS_REGEX_RESULTSIZE*noresult,"comment",namedvalue)){
      _a2l_error("Failed to get substring comment\n");
      pcre_free(re);
      return CS_FAIL;
    }
    strncpy(tab->comment,namedvalue,CS_CHARBUF);

    if(tab->ttype == A2L_COMPU_TAB_TYPE_VTAB){
      if(cs_parser_getsubstring(re,lcm.result,ovector,CS_REGEX_RESULTSIZE*noresult,"type",namedvalue)){
	_a2l_error("Failed to get substring type\n");
	pcre_free(re);
	return CS_FAIL;
      }
      if(strcmp(namedvalue,"TAB_VERB")!=0){
	_a2l_error("Unknown lookup table\n");
	pcre_free(re);
	return CS_FAIL;
      }
    }
    
    if(cs_parser_getsubstring_int(re,lcm.result,ovector,CS_REGEX_RESULTSIZE*noresult,"elems",&tab->number,0)){
      _a2l_error("Failed to get number of elements \n");
      pcre_free(re);
      return CS_FAIL;
    }    
    if(tab->number <= 0){
      _a2l_error("Invalid number of elements \n");
      pcre_free(re);
      return CS_FAIL;
    }
    //Jump to end of last element
    if(tab->ttype == A2L_COMPU_TAB_TYPE_VTAB){
      if(cs_parser_stepnext(&lcm,ovector,CS_REGEX_RESULTSIZE * noresult,5)){
	_a2l_error("Failed to step to next working point\n");
	pcre_free(re);
	return CS_FAIL;
      }    
    }
    if(tab->ttype == A2L_COMPU_TAB_TYPE_VTAB_RANGE){
      if(cs_parser_stepnext(&lcm,ovector,CS_REGEX_RESULTSIZE * noresult,3)){
	_a2l_error("Failed to step to next working point\n");
	pcre_free(re);
	return CS_FAIL;
      }    
    }
  }

  //Prealloc lookup table
  tab->elems = (cs_a2l_compu_vtab_elem_t*)malloc(sizeof(cs_a2l_compu_vtab_elem_t)*tab->number);
  if(tab->elems == NULL){
    _a2l_error("Malloc failure\n");
    tab->number = 0;
    pcre_free(re);
    return CS_FAIL;
  }
  memset(tab->elems,0,sizeof(cs_a2l_compu_vtab_elem_t)*tab->number);

  for(i = 0 ; i < tab->number; i++){
    //Note lcm is modified
    if(cs_a2l_getlookupvalue(&lcm,tab->ttype,&tab->elems[i])){
      _a2l_error("Failed to parse lookup entries\n");
      tab->number = 0;
      pcre_free(re);
      return CS_FAIL;
    }
  }

  //Get optional default value
  cs_a2l_getstringvaluequoted(lcmdefault,"DEFAULT_VALUE",tab->edefault,CS_CHARBUF);

  //cs_a2l_print_compuvtab(tab);

  pcre_free(re);
  return CS_OK;
}

/**************************************************************************/
void cs_a2l_reset_compumethod(
			      cs_a2l_compu_method_t *cm
			      )
/**************************************************************************/
{
  if(cm == NULL){
    _a2l_error("Parameter failure\n");
    return;
  }
  
  cs_a2l_reset_compuvtab(&cm->tab);

  //reset
  memset(cm,0,sizeof(cs_a2l_compu_method_t));
}

/**************************************************************************/
int cs_a2l_GetCompuMethod(
			  cs_a2l_t *data,
			  cs_a2l_compu_method_t *cm,
			  char *name
			  )
/**************************************************************************/
{
  cs_parser_result_t lcm;
  cs_parser_result_t lresult;
  pcre *re = NULL;
  char regex[CS_REGEX_BUFFER];
  int rv = 0;
  int noresult = 12;
  int ovector[CS_REGEX_RESULTSIZE*noresult];
  char namedvalue[CS_PARSER_CHARBUF];


  if((data == NULL) ||
     (cm == NULL) ||
     (name == NULL)){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  cs_a2l_reset_compumethod(cm);

  if(strcmp(name,"NO_COMPU_METHOD")==0){
    sprintf(cm->name,"No function defined");
    cm->algo = A2L_CM_IDENT;
    return CS_OK;
  }


  //Search daq section
  lcm.searchin = data->module.result;
  lcm.searchlen = data->module.resultlen;
  if(cs_a2l_search_section(&lcm,"COMPU_METHOD",name)){
    _a2l_error("Regex failure searching section COMPU_METHOD %s\n",name);
    return CS_FAIL;
  }

  if(lcm.result == NULL){
    _a2l_error("Cannot find COMPU_METHOD %s\n",name);
    return CS_FAIL;
  }

  //cs_parser_printresult(&lcm);

  //define serachregion
  lresult.searchin = lcm.result;
  lresult.searchlen = lcm.resultlen;

  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER);
  snprintf(regex,CS_REGEX_BUFFER,"COMPU_METHOD[\\s]+(?'name'(.*?))[\\s]+\"(?'comment'(.*?))\"[\\s]+(?'type'([A-Z_]+))");

  //Precompile regex
  if((re = cs_parser_precompile((const char *)regex)) == NULL){
    _a2l_error("Regex precompilation failed\n");
    return CS_FAIL;
  }
  //Get data, only one fit is allowed
  rv = pcre_exec(re, NULL, (const char*)lcm.result,lcm.resultlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE * noresult);

  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _a2l_verbose("No match\n"); 
      break;
    default: 
      _a2l_error("Matching error %d\n", rv);
      pcre_free(re);
      return CS_FAIL;
      break;
    }
  } else {
    if(cs_parser_getsubstring(re,lcm.result,ovector,CS_REGEX_RESULTSIZE*noresult,"name",namedvalue)){
      _a2l_error("Failed to get substring name\n");
      pcre_free(re);
      return CS_FAIL;
    }
    strncpy(cm->name,namedvalue,CS_CHARBUF);

    if(cs_parser_getsubstring(re,lcm.result,ovector,CS_REGEX_RESULTSIZE*noresult,"comment",namedvalue)){
      _a2l_error("Failed to get substring comment\n");
      pcre_free(re);
      return CS_FAIL;
    }
    strncpy(cm->comment,namedvalue,CS_CHARBUF);

    if(cs_parser_getsubstring(re,lcm.result,ovector,CS_REGEX_RESULTSIZE*noresult,"type",namedvalue)){
      _a2l_error("Failed to get substring type\n");
      pcre_free(re);
      return CS_FAIL;
    }

    if(strcmp(namedvalue,"IDENTICAL")==0){
      cm->algo = A2L_CM_IDENT;
    }

    if(strcmp(namedvalue,"FORM")==0){
      cm->algo = A2L_CM_FUNCTION;
      //TODO get formular description
    }

    if(strcmp(namedvalue,"TAB_INTP")==0){
      cm->algo = A2L_CM_TAB_INTP;
      //TODO get table definition
    }

    if(strcmp(namedvalue,"TAB_NOINTP")==0){
      cm->algo = A2L_CM_TAB_NOINTP;
      //TODO get table definition
    }

    if(strcmp(namedvalue,"LINEAR")==0){
      cm->algo = A2L_CM_LINEAR;

      if(cs_a2l_getdoublevaluevector(lresult,"COEFFS_LINEAR",cm->coeff,2)){
	_a2l_error("Failed to parse linear coeffizient\n");
	pcre_free(re);
	return CS_FAIL;
      }
    }

    if(strcmp(namedvalue,"RAT_FUNC")==0){
      cm->algo = A2L_CM_RAT;

      if(cs_a2l_getdoublevaluevector(lresult,"COEFFS",cm->coeff,6)){
	_a2l_error("Failed to parse rational function coeffizient\n");
	pcre_free(re);
	return CS_FAIL;
      }

    }

    if(strcmp(namedvalue,"TAB_VERB")==0){
      cm->algo = A2L_CM_TAB_VERB;
      if(cs_a2l_getstringvalue(lresult,"COMPU_TAB_REF",cm->compuvtab,CS_CHARBUF)){
	_a2l_error("Failed to parse compu tab reference\n");
	pcre_free(re);
	return CS_FAIL;
      }
      if(cs_a2l_GetCompuVTab(data,&cm->tab,cm->compuvtab)){
	_a2l_error("Failed to parse compu tab node\n");
	pcre_free(re);
	return CS_FAIL;
      }

    }

    if(cm->algo == A2L_CM_UNDEF){
      _a2l_error("Unsupported computation method %s\n",namedvalue);
      pcre_free(re);
      return CS_FAIL;
    }
  }

  pcre_free(re);
  return CS_OK;
}

/**************************************************************************/
void cs_a2l_reset_record(
			 cs_a2l_record_t *record
			 )
/**************************************************************************/
{
  if(record == NULL){
    _a2l_error("Parameter failure\n");
    return;
  }

  //reset
  memset(record,0,sizeof(cs_a2l_record_t));
}

/**************************************************************************/
int cs_a2l_GetRecord(
		     cs_a2l_t *data,
		     cs_a2l_record_t *record,
		     char *name
		     )
/**************************************************************************/
{
  cs_parser_result_t lcm;
  cs_parser_result_t lresult;
  cs_parser_result_t lf;
  pcre *re = NULL;
  char regex[CS_REGEX_BUFFER];
  int rv = 0;
  int noresult = 12;
  int ovector[CS_REGEX_RESULTSIZE*noresult];
  char namedvalue[CS_PARSER_CHARBUF];
  int dummy;
  uint8_t loop = 1;

  if((data == NULL) ||
     (record == NULL) ||
     (name == NULL)){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  //reset
  cs_a2l_reset_record(record);

  //Search daq section
  lcm.searchin = data->module.result;
  lcm.searchlen = data->module.resultlen;
  if(cs_a2l_search_section(&lcm,"RECORD_LAYOUT",name)){
    _a2l_error("Regex failure searching section RECORD_LAYOUT %s\n",name);
    return CS_FAIL;
  }

  if(lcm.result == NULL){
    _a2l_error("Cannot find RECORD_LAYOUT %s\n",name);
    return CS_FAIL;
  }

  //cs_parser_printresult(&lcm);

  //store record name
  strncpy(record->name,name,CS_CHARBUF);

  //define serachregion
  lresult.result = lcm.result;
  lresult.resultlen = lcm.resultlen;

  //Loop for arguments
  record->size = 0;

  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER);
  snprintf(regex,CS_REGEX_BUFFER,"(?'all'((?'type'[A-Z_]+)[\\s]+(?'position'[0-9]+)[\\s]+(?'vtype'[A-Z_]+)[\\s]+(?'orientation'[A-Z_]+)[\\s]+(?'access'[A-Z]+)))");
  
  //Precompile regex
  if((re = cs_parser_precompile((const char *)regex)) == NULL){
    _a2l_error("Regex precompilation failed\n");
    return CS_FAIL;
  }
  
  while(loop){
    //Get data, only one fit is allowed
    rv = pcre_exec(re, NULL, (const char*)lresult.result,lresult.resultlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE * noresult);
    
    if(rv < 0) {
      switch(rv) {
      case PCRE_ERROR_NOMATCH:
	loop = 0;
	break;
      default: 
	_a2l_verbose("Matching error %d\n", rv);
	pcre_free(re);
	return CS_FAIL;
	break;
      }
      if(!loop)
	break;
    } else {
      if(cs_parser_getsubstring(re,lresult.result,ovector,CS_REGEX_RESULTSIZE*noresult,"type",namedvalue)){
	_a2l_error("Failed to get substring type\n");
	pcre_free(re);
	return CS_FAIL;
      }
      record->rtype[record->size] = A2L_RECORD_TYPE_UNKNOWN;
      if(strcmp(namedvalue,"FNC_VALUES") == 0){
	record->rtype[record->size] = A2L_RECORD_TYPE_FNC;
      } else {
	_a2l_error("Unknown record type %s\n",namedvalue);
      }
      
      
      if(cs_parser_getsubstring_int(re,lresult.result,ovector,CS_REGEX_RESULTSIZE*noresult,"position",&record->position[record->size],0)){
	_a2l_error("GetRecord: Failed to get position \n");
	pcre_free(re);
	return CS_FAIL;
      }
      
      if(cs_parser_getsubstring(re,lresult.result,ovector,CS_REGEX_RESULTSIZE*noresult,"vtype",namedvalue)){
	_a2l_error("Failed to get substring vtype\n");
	pcre_free(re);
	return CS_FAIL;
      }
      if(cs_a2l_StringToType(namedvalue,data->AG,&record->dtype[record->size],&dummy)){
	_a2l_error("Unknown value type %s\n",namedvalue);
	pcre_free(re);
	return CS_FAIL;
      }
      //Get optional parts
      //define search area
      lf.searchin = lresult.result;
      lf.searchlen = lresult.resultlen;
      //Set default orientation
      record->orientation[record->size] = A2L_RECORD_ORIENT_UNKNOWN;
      record->orientation[record->size] = !cs_a2l_getflag(lf,"COLUMN_DIR")?A2L_RECORD_ORIENT_COLUMN:record->orientation[record->size];
      record->orientation[record->size] = !cs_a2l_getflag(lf,"ROW_DIR")?A2L_RECORD_ORIENT_ROW:record->orientation[record->size];
      record->orientation[record->size] = !cs_a2l_getflag(lf,"INDEC_INCR")?A2L_RECORD_ORIENT_INDEXINCR:record->orientation[record->size];
      //Set default access
      record->access[record->size] = A2L_RECORD_ACCESS_UNKNOWN;
      record->access[record->size] = !cs_a2l_getflag(lf,"DIRECT")?A2L_RECORD_ACCESS_DIRECT:record->access[record->size];
      

      //Validate record line
      switch(record->rtype[record->size]){
      case  A2L_RECORD_TYPE_FNC:
	if((record->access[record->size] == A2L_RECORD_ACCESS_UNKNOWN) || 
	   (record->orientation[record->size] == A2L_RECORD_ORIENT_UNKNOWN)){
	  _a2l_error("Invalid Record format for FNC_VALUES type\n");
	  pcre_free(re);
	  return CS_FAIL;
	}
	break;
      default:
	_a2l_error("Unknown Record format\n,");
      }
      

    }
    record->size++;
    if(record->size >= CS_A2L_RECORD_SIZEMAX){
      _a2l_error("Buffer overflow detected\n");
      pcre_free(re);
      return CS_FAIL;
    }
    //set new search area
    if(loop){
      if(cs_parser_stepnext(&lresult,ovector,CS_REGEX_RESULTSIZE * noresult,1)){
	_a2l_error("Failed to step to next working point\n");
	pcre_free(re);
	return CS_FAIL;
      }
    }
  }

  //cs_a2l_print_record(record);

  pcre_free(re);
  return CS_OK;
}

/**************************************************************************/
void cs_a2l_reset_characteristic(
				 cs_a2l_characteristic_t *ch
				 )
/**************************************************************************/
{

  if(ch == NULL){
    _a2l_error("Parameter failure\n");
    return;
  }

  //reset
  memset(ch,0,sizeof(cs_a2l_characteristic_t));
  
  ch->type = A2L_CHARACTERISTIC_TYPE_UNKNOWN;
  ch->BO = A2L_BO_UNKNOWN;
  cs_a2l_reset_compumethod(&ch->cm);
  cs_a2l_reset_record(&ch->record);
}

/**************************************************************************/
int cs_a2l_GetCharacteristic(
			     cs_a2l_t *data,
			     cs_a2l_characteristic_t *ch,
			     char *name
			     )
/**************************************************************************/
{
  cs_parser_result_t lchar;
  cs_parser_result_t lresult;
  pcre *re = NULL;
  char regex[CS_REGEX_BUFFER];
  int rv = 0;
  int noresult = 20;
  int ovector[CS_REGEX_RESULTSIZE*noresult];
  char namedvalue[CS_PARSER_CHARBUF];
  uint8_t rflag = 0;
  uint32_t rvalue;

  if((data == NULL) ||
     (ch == NULL) ||
     (name == NULL)){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  //reset
  cs_a2l_reset_characteristic(ch);

  //Search daq section
  lchar.searchin = data->module.result;
  lchar.searchlen = data->module.resultlen;
  if(cs_a2l_search_section(&lchar,"CHARACTERISTIC",name)){
    _a2l_error("Regex failure searching CHARACTERISTIC %s\n",name);
    return CS_FAIL;
  }

  if(lchar.result == NULL){
    _a2l_error("Cannot find CHARACTERISTIC %s\n",name);
    return CS_FAIL;
  }
  //cs_parser_printresult(&lchar);

  //define serachregion
  lresult.searchin = lchar.result;
  lresult.searchlen = lchar.resultlen;

  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER);
  snprintf(regex,CS_REGEX_BUFFER,"CHARACTERISTIC[\\s]+(?'name'(.*?))[\\s]+\"(?'comment'(.*?))\"[\\s]+(?'type'([A-Z_]+))[\\s]+0x(?'address'([0-9a-fA-F]+))[\\s]+(?'record'([A-Za-z_0-9\\.]+))[\\s]+(?'maxdiff'([-0-9.]+))[\\s]+(?'compu'([A-Za-z_0-9\\.]+))[\\s]+(?'min'([-0-9E+\\.]+))[\\s]+(?'max'([-0-9E+\\.]+))");
  //_a2l_verbose("Regex: %s\n",regex);
  //Precompile regex
  if((re = cs_parser_precompile((const char *)regex)) == NULL){
    _a2l_error("Regex precompilation failed\n");
    return CS_FAIL;
  }
  //Get data, only one fit is allowed
  rv = pcre_exec(re, NULL, (const char*)lchar.result,lchar.resultlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE * noresult);
  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _a2l_error("No content match for Characteristic %s\n",name);
      //cs_parser_printresult(&lchar);
      pcre_free(re);
      return CS_FAIL;
      break;
    default: 
      _a2l_error("Matching error %d\n", rv);
      pcre_free(re);
      return CS_FAIL;
      break;
    }
  } else {
    
    if(cs_parser_getsubstring(re,lchar.result,ovector,CS_REGEX_RESULTSIZE*noresult,"name",namedvalue)){
      _a2l_error("Failed to get substring name\n");
      pcre_free(re);
      return CS_FAIL;
    }
    strncpy(ch->name,namedvalue,CS_CHARBUF);

    if(cs_parser_getsubstring(re,lchar.result,ovector,CS_REGEX_RESULTSIZE*noresult,"comment",namedvalue)){
      _a2l_error("Failed to get substring comment\n");
      pcre_free(re);
      return CS_FAIL;
    }
    strncpy(ch->comment,namedvalue,CS_CHARBUF);

    if(cs_parser_getsubstring(re,lchar.result,ovector,CS_REGEX_RESULTSIZE*noresult,"type",namedvalue)){
      _a2l_error("Failed to get substring type\n");
      pcre_free(re);
      return CS_FAIL;
    }
    ch->type = A2L_CHARACTERISTIC_TYPE_UNKNOWN;
   
    if(strcmp(namedvalue,"VALUE") == 0){
      ch->type = A2L_CHARACTERISTIC_TYPE_VALUE;
    }
    if(strcmp(namedvalue,"VAL_BLK") == 0){
      ch->type = A2L_CHARACTERISTIC_TYPE_VALBLK;
    }
    if(strcmp(namedvalue,"CURVE") == 0){
      ch->type = A2L_CHARACTERISTIC_TYPE_CURVE;
    }
    if(strcmp(namedvalue,"ASCII") == 0){
      ch->type = A2L_CHARACTERISTIC_TYPE_ASCII;
    }

    if(ch->type == A2L_CHARACTERISTIC_TYPE_UNKNOWN){
      _a2l_error("Unknown characteristic type:%s\n",namedvalue);
      pcre_free(re);
      return CS_FAIL;
    }
    if(cs_parser_getsubstring(re,lchar.result,ovector,CS_REGEX_RESULTSIZE*noresult,"address",namedvalue)){
      _a2l_error("Failed to get substring address\n");
      pcre_free(re);
      return CS_FAIL;
    }
    if(!sscanf(namedvalue,"%X",&ch->address)){
      _a2l_error("Failed to parse substring %s\n",namedvalue);
      pcre_free(re);
      return CS_FAIL;
    }

    if(cs_parser_getsubstring(re,lchar.result,ovector,CS_REGEX_RESULTSIZE*noresult,"record",namedvalue)){
      _a2l_error("Failed to get substring record\n");
      pcre_free(re);
      return CS_FAIL;
    }
    strncpy(ch->recordname,namedvalue,CS_CHARBUF);

    if(cs_parser_getsubstring(re,lchar.result,ovector,CS_REGEX_RESULTSIZE*noresult,"compu",namedvalue)){
      _a2l_error("Failed to get substring compu\n");
      pcre_free(re);
      return CS_FAIL;
    }
    strncpy(ch->conv,namedvalue,CS_CHARBUF);

    if(cs_parser_getsubstring_double(re,lchar.result,ovector,CS_REGEX_RESULTSIZE*noresult,"maxdiff",&ch->maxdiff,0)){
      _a2l_error("Failed to get substring maxdiff\n");
      pcre_free(re);
      return CS_FAIL;
    }

    if(cs_parser_getsubstring_double(re,lchar.result,ovector,CS_REGEX_RESULTSIZE*noresult,"min",&ch->min,0)){
      _a2l_error("Failed to get substring min\n");
      pcre_free(re);
      return CS_FAIL;
    }

    if(cs_parser_getsubstring_double(re,lchar.result,ovector,CS_REGEX_RESULTSIZE*noresult,"max",&ch->max,0)){
      _a2l_error("Failed to get substring max\n");
      pcre_free(re);
      return CS_FAIL;
    }

  }
  pcre_free(re);
  
  cs_a2l_GetByteOrder(lresult,data->BO,&ch->BO);

  cs_a2l_getstringvalue(lresult,"PHYS_UNIT",ch->unit,CS_PARSER_CHARBUF);

  cs_a2l_getstringvalue(lresult,"FORMAT",ch->format,CS_PARSER_CHARBUF);

  rvalue = cs_a2l_getuintvalue(lresult,"ECU_ADDRESS_EXTENSION",&rflag);
  if(!rflag){
    ch->extension= (uint8_t)rvalue;
  }
  
  ch->access = A2L_CHARACTERISTIC_ACCESS_RW;
  if(!cs_a2l_getflag(lresult,"READONLY")){
    ch->access = A2L_CHARACTERISTIC_ACCESS_RO;
  }

  if(cs_a2l_GetRecord(data,&ch->record,ch->recordname)){
    _a2l_error("Failed to parse Record %s\n",ch->recordname);
    return CS_FAIL;
  }

  //Get compu method
  if(cs_a2l_GetCompuMethod(data,&ch->cm,ch->conv)){
    _a2l_error("Failed to parse Compu method %s\n",ch->conv);
    return CS_FAIL;
  }
  //Set min and max for computational function
  ch->cm.min = ch->min;
  ch->cm.max = ch->max;

  //Copy AG, only used in support functions
  ch->AG = data->AG;

  //cs_a2l_print_characteristic(ch);

  return CS_OK;

}

/**************************************************************************/
int cs_a2l_compu_method_linear_create(
				      cs_a2l_compu_method_t* cm,
				      char *name,
				      char *label,
				      char *unit,
				      char *format,
				      double scale,
				      double offset
				      )
/**************************************************************************/
{
  if((cm == NULL) ||
     (name == NULL)){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }

  //Reset structure
  memset(cm,0,sizeof(cs_a2l_compu_method_t));

  strncpy(cm->name,name,CS_CHARBUF);

  if(label != NULL)
    strncpy(cm->comment,label,CS_CHARBUF);

  if(unit != NULL)
    strncpy(cm->unit,unit,CS_CHARBUF);

  if(format != NULL)
    strncpy(cm->format,format,CS_CHARBUF);

  cm->coeff[0] = scale;
  cm->coeff[1] = offset;

  cm->algo = A2L_CM_LINEAR;

  return CS_OK;
}  

/**************************************************************************/
int cs_a2l_compu_method_linear_r2d(
				   cs_a2l_compu_method_t cm,
				   double rawvalue,
				   double *value
				   )
/**************************************************************************/
{
  if(value == NULL){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  *value = (rawvalue * cm.coeff[0]) - cm.coeff[1];
  
  return CS_OK;
}

/**************************************************************************/
int cs_a2l_compu_method_linear_d2r(
				   cs_a2l_compu_method_t cm,
				   double *rawvalue,
				   double value
				   )
/**************************************************************************/
{
  if(rawvalue == NULL){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  if(cm.coeff[0] != 0){
    *rawvalue = (value + cm.coeff[1]) / cm.coeff[0];
  } else {
    *rawvalue = 0;
    _a2l_error("Devision by zero\n");
    return CS_FAIL;
  }
  return CS_OK;
}

/**************************************************************************/
int cs_a2l_compu_method_rat_r2d(
				cs_a2l_compu_method_t cm,
				double rawvalue,
				double *value
				)
/**************************************************************************/
{
  double a = cm.coeff[0];
  double b = cm.coeff[1];
  double c = cm.coeff[2];
  double d = cm.coeff[3];
  double e = cm.coeff[4];
  double f = cm.coeff[5];
  double ldiv = 0;

  if(value == NULL){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  ldiv = d*rawvalue*rawvalue+e*rawvalue+f;

  if(ldiv == 0){
    _a2l_error("Devision by zero\n");
    *value = 0;
    return CS_FAIL;
  }

  *value = (a*rawvalue*rawvalue + b*rawvalue + c)/ ldiv;
  
  
  return CS_OK;
}

/**************************************************************************/
int cs_a2l_compu_method_rat_d2r(
				cs_a2l_compu_method_t cm,
				double *rawvalue,
				double value
				)
/**************************************************************************/
{
  double a = cm.coeff[0];
  double b = cm.coeff[1];
  double c = cm.coeff[2];
  double d = cm.coeff[3];
  double e = cm.coeff[4];
  double f = cm.coeff[5];
  double ldiv = 0;
  double num;
  double y = value;
  double tmp;
  
  if(rawvalue == NULL){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  *rawvalue = 0;

  if((a == 0)&&(d == 0)){
    //optimzed calculation
    //Maxima: solve(y=(b*x+d)/(e*x+f),x);
    //x=-(f*y-c)/(e*y-b)
    ldiv = e*y-b;
    if(ldiv == 0){
      *rawvalue = 0;
      _a2l_error("Devision by zero\n");
      return CS_FAIL;
    }
    num = f*y-c;
    *rawvalue = -1.0*num/ldiv;
    return CS_OK;
  }
  //normal calculation
  //Maxima: solve(y=(a*x*x+b*x+c)/(d*x*x+e*x+f),x);
  //[x=(sqrt((e^2-4*d*f)*y^2+(4*a*f-2*b*e+4*c*d)*y-4*a*c+b^2)-e*y+b)/(2*d*y-2*a),x=-(sqrt((e^2-4*d*f)*y^2+(4*a*f-2*b*e+4*c*d)*y-4*a*c+b^2)+e*y-b)/(2*d*y-2*a)]
  ldiv = 2*d*y-2*a;
  if(ldiv == 0){
    *rawvalue = 0;
    _a2l_error("Devision by zero\n");
    return CS_FAIL;
  }
  num = sqrt((e*e-4*d*f)*y*y+(4*a*f-2*b*e+4*c*d)*y-4*a*c+b*b);
  tmp = (num-e*y+b)/ldiv;
  if((tmp > cm.min)&&(tmp < cm.max)){
    //Value in range
    *rawvalue = tmp;
    return CS_OK;
  }
  tmp = -1.0*(num+e*y-b)/ldiv;
  if((tmp > cm.min)&&(tmp < cm.max)){
    //Value in range
    *rawvalue = tmp;
    return CS_OK;
  }
  _a2l_error("Value out of range\n");
  return CS_FAIL;
   
}

/**************************************************************************/
int cs_a2l_compu_method_r2d(
			    cs_a2l_compu_method_t cm,
			    double rawvalue,
			    double *value
			    )
/**************************************************************************/
{
  if(value == NULL){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  switch(cm.algo){
  case A2L_CM_IDENT:
    *value = rawvalue;
    return CS_OK;
    break;
  case A2L_CM_LINEAR:
    return cs_a2l_compu_method_linear_r2d(cm,rawvalue,value);
    break;
  case A2L_CM_RAT:
    return cs_a2l_compu_method_rat_d2r(cm,value,rawvalue);
    break;
  case A2L_CM_FUNCTION:
    _a2l_error("Method FORM is not supported\n");
    return CS_FAIL;
  case A2L_CM_TAB_INTP:
    _a2l_error("Method TAB_INTP is not supported\n");
    return CS_FAIL;
  case A2L_CM_TAB_NOINTP:
    _a2l_error("Method TAB_NOINTP is not supported\n");
    return CS_FAIL;
  case A2L_CM_TAB_VERB:
    //No value lookup, only forwarding, because it is only human interface conversion
    *value = rawvalue;
    return CS_OK;
    break;
  default:
    _a2l_error("Method is not supported\n");
    return CS_FAIL;
  }
}

/**************************************************************************/
int cs_a2l_compu_method_d2r(
			    cs_a2l_compu_method_t cm,
			    double *rawvalue,
			    double value
			    )
/**************************************************************************/
{
  if(rawvalue == NULL){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  switch(cm.algo){
  case A2L_CM_IDENT:
    *rawvalue = value;
    return CS_OK;
    break;
  case A2L_CM_LINEAR:
    return cs_a2l_compu_method_linear_d2r(cm,rawvalue,value);
    //return cs_a2l_compu_method_linear_d2r(cm,rawvalue,value);
    break;
  case A2L_CM_RAT:
    return cs_a2l_compu_method_rat_r2d(cm,value,rawvalue);
    //return cs_a2l_compu_method_rat_d2r(cm,rawvalue,value);
    break;
  case A2L_CM_FUNCTION:
    _a2l_error("Method FORM is not supported\n");
    return CS_FAIL;
    break;
  case A2L_CM_TAB_INTP:
    _a2l_error("Method TAB_INTP is not supported\n");
    return CS_FAIL;
    break;
  case A2L_CM_TAB_NOINTP:
    _a2l_error("Method TAB_NOINTP is not supported\n");
    return CS_FAIL;
    break;
  case A2L_CM_TAB_VERB:
    //No value lookup, only forwarding, because it is only human interface conversion
    *rawvalue = value;
    return CS_OK;
    break;
  default:
    _a2l_error("Method is not supported\n");
    return CS_FAIL;
  }
  
}


/**************************************************************************/
int cs_a2l_convert_to_double(			
			     cs_a2l_bo_t endianess,
			     uint8_t *udata,
			     int size,
			     cs_a2l_compu_method_t cm,
			     cs_a2l_data_type_t dtype,
			     double *value
						)
/**************************************************************************/
{
  cs_cu_data16_t l16;
  cs_cu_data32_t l32;
  cs_cu_data64_t l64;
  int8_t s8;
  int16_t s16;
  int32_t s32;
  int64_t s64;
  double lvalue = 0;


  if((udata == NULL) ||
     (value == NULL) ||
     (size == 0) ||
     (size > 8) ||
     (endianess == A2L_BO_UNKNOWN)){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }

  _a2l_verbose("Raw value:");
  _a2l_verbose_buffer(udata,size);
  switch(dtype){
  case A2L_UBYTE:
  case A2L_UWORD:
  case A2L_UINT:
  case A2L_UINT64:
    switch(size){
    case 1:
      lvalue = (double)udata[0];
      break;
    case 2:
      l16.u8[0] = udata[0];
      l16.u8[1] = udata[1];
      lvalue = (double)cs_cu_swap16(endianess,l16.u16);
      break;
    case 3:
      l32.u8[0] = udata[0];
      l32.u8[1] = udata[1];
      l32.u8[2] = udata[2];
      l32.u8[3] = 0;
      lvalue = (double)cs_cu_swap32(endianess,l32.u32);
      break;
    case 4:
      l32.u8[0] = udata[0];
      l32.u8[1] = udata[1];
      l32.u8[2] = udata[2];
      l32.u8[3] = udata[3];
      lvalue = (double)cs_cu_swap32(endianess,l32.u32);
      break;
    case 8:
      l64.u8[0] = udata[0];
      l64.u8[1] = udata[1];
      l64.u8[2] = udata[2];
      l64.u8[3] = udata[3];
      l64.u8[4] = udata[4];
      l64.u8[5] = udata[5];
      l64.u8[6] = udata[6];
      l64.u8[7] = udata[7];
      lvalue = (double)cs_cu_swap64(endianess,l64.u64);
      break;
    default:
      //Nothing to do
      break;
    }
    break;
  case A2L_SINT:
  case A2L_SBYTE:
  case A2L_SWORD:
  case A2L_SINT64:
    switch(size){
    case 1:
      s8 = (int8_t)udata[0];
      lvalue = (double)s8;
      break;
    case 2:
      l16.u8[0] = udata[0];
      l16.u8[1] = udata[1];
      s16 = (int16_t)cs_cu_swap16(endianess,l16.u16);
      lvalue = (double)s16;
      break;
    case 3:
      l32.u8[0] = udata[0];
      l32.u8[1] = udata[1];
      l32.u8[2] = udata[2];
      l32.u8[3] = 0;
      s32 = (int32_t)cs_cu_swap32(endianess,l32.u32);
      lvalue = (double)s32;
      break;
    case 4:
      l32.u8[0] = udata[0];
      l32.u8[1] = udata[1];
      l32.u8[2] = udata[2];
      l32.u8[3] = udata[3];
      s32 = (int32_t)cs_cu_swap32(endianess,l32.u32);
      lvalue = (double)s32;;
      break;
    case 8:
      l64.u8[0] = udata[0];
      l64.u8[1] = udata[1];
      l64.u8[2] = udata[2];
      l64.u8[3] = udata[3];
      l64.u8[4] = udata[4];
      l64.u8[5] = udata[5];
      l64.u8[6] = udata[6];
      l64.u8[7] = udata[7];
      s64 = (int64_t)cs_cu_swap64(endianess,l64.u64);
      lvalue = (double)s64;
      break;
    default:
      //Nothing to do
      break;
    }
    break;
  case A2L_FLOAT:
    l32.u32 = 0;
    memcpy(l32.u8,udata,size);
    l32.u32 = cs_cu_swap32(endianess,l32.u32);
    lvalue = (double)l32.flt;
    break;
  case A2L_FLOAT64:
    l64.u64 = 0;
    memcpy(l64.u8,udata,size);
    l64.u64 = cs_cu_swap64(endianess,l64.u64);
    lvalue = (double)l64.dbl;
    break;
  default:
    break;
  }
    

  return cs_a2l_compu_method_r2d(cm,lvalue,value);

}

/**************************************************************************/
int cs_a2l_convert_from_double(			
			       cs_a2l_bo_t endianess,
			       uint8_t *udata,
			       int size,
			       cs_a2l_compu_method_t cm,
			       cs_a2l_data_type_t dtype,
			       double value
						)
/**************************************************************************/
{
  cs_cu_data16_t l16;
  cs_cu_data32_t l32;
  cs_cu_data64_t l64;
  double lvalue = 0;


  if((udata == NULL) ||
     (size == 0) ||
     (endianess == A2L_BO_UNKNOWN)){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }    

  //Reset buffer
  memset(udata,0,sizeof(uint8_t)*size);

  //Convert via compu method
  _a2l_verbose("A2L Convert from Double: Double value: %f \n",value);
  if(cs_a2l_compu_method_d2r(cm,&lvalue,value)){
    _a2l_error("Failed to convert value\n");
    return CS_FAIL;
  }
  _a2l_verbose("Double value after Cumpu method: %f \n",lvalue);
  switch(dtype){
  case A2L_UBYTE:
  case A2L_UWORD:
  case A2L_UINT:
  case A2L_UINT64:
    switch(size){
    case 1:
      udata[0] = (uint8_t)lvalue;
      break;
    case 2:
      l16.u16 = cs_cu_swap16(endianess,(uint16_t)lvalue);
      udata[0] = l16.u8[0];
      udata[1] = l16.u8[1];
      break;
    case 3:
      l32.u32 = cs_cu_swap32(endianess,(uint32_t)lvalue);
      udata[0] = l32.u8[0];
      udata[1] = l32.u8[1]; 
      udata[2] = l32.u8[2];
      udata[3] = 0;
      break;
    case 4:
      l32.u32 = cs_cu_swap32(endianess,(uint32_t)lvalue);
      udata[0] = l32.u8[0];
      udata[1] = l32.u8[1];
      udata[2] = l32.u8[2];
      udata[3] = l32.u8[3];
      break;
    case 8:
      l64.u64 = cs_cu_swap64(endianess,(uint64_t)lvalue);
      udata[0] = l64.u8[0];
      udata[1] = l64.u8[1];
      udata[2] = l64.u8[2];
      udata[3] = l64.u8[3];
      udata[4] = l64.u8[4];
      udata[5] = l64.u8[5];
      udata[6] = l64.u8[6];
      udata[7] = l64.u8[7];
      break;
    default:
      //Nothing to do
      break;
    }
    break;
  case A2L_SINT:
  case A2L_SBYTE:
  case A2L_SWORD:
  case A2L_SINT64:
    switch(size){
    case 1:
      udata[0] = (uint8_t)(int8_t)lvalue;
      break;
    case 2:
      l16.u16 = cs_cu_swap16(endianess,(uint16_t)(int16_t)lvalue);
      udata[0] = l16.u8[0];
      udata[1] = l16.u8[1];
      break;
    case 3:
      l32.u32 = cs_cu_swap32(endianess,(uint32_t)(int32_t)lvalue);
      udata[0] = l32.u8[0];
      udata[1] = l32.u8[1]; 
      udata[2] = l32.u8[2];
      udata[3] = 0;
      break;
    case 4:
      l32.u32 = cs_cu_swap32(endianess,(uint32_t)(int32_t)lvalue);
      udata[0] = l32.u8[0];
      udata[1] = l32.u8[1];
      udata[2] = l32.u8[2];
      udata[3] = l32.u8[3];
      break;
    case 8:
      l64.u64 = cs_cu_swap64(endianess,(uint64_t)(int64_t)lvalue);
      udata[0] = l64.u8[0];
      udata[1] = l64.u8[1];
      udata[2] = l64.u8[2];
      udata[3] = l64.u8[3];
      udata[4] = l64.u8[4];
      udata[5] = l64.u8[5];
      udata[6] = l64.u8[6];
      udata[7] = l64.u8[7];
      break;
    default:
      //Nothing to do
      break;
    }
    break;
  case A2L_FLOAT:
    l32.flt = (float)lvalue;
    l32.u32 = cs_cu_swap32(endianess,l32.u32);
    memcpy(udata,l32.u8,size);
    break;
  case A2L_FLOAT64:
    l64.dbl = lvalue;
    l64.u64 = cs_cu_swap64(endianess,l64.u64);
    memcpy(udata,l64.u8,size);
    break;
  default:
    break;
  }
  _a2l_verbose("Raw buffer after conversion:\n");
  _a2l_debug_buffer(udata,size);
    
  return CS_OK;
}

/**************************************************************************/
void cs_a2l_printlist(
		      cs_a2l_list_t *list
		      )
/**************************************************************************/
{
  int i;
  if(list == NULL){
    _a2l_error("Parameter failure\n");
    return;
  }

  _a2l_info("List of %s sections:\n",list->sectiontype);
  for(i = 0; i < list->no_result; i++){
    _a2l_info("Index %i, Name: %s, Label: %s\n",i,list->resultname[i],list->resultlabel[i]);
  }
  
  return;
}


/**************************************************************************/
void cs_a2l_print_modpar(
			 cs_a2l_modpar_t *mp
			 )
/**************************************************************************/
{
  int i;
  if(mp == NULL){
    _a2l_error("Parameter failure\n");
    return;
  }
  
  _a2l_info("Number of  Memory segments %i\n",mp->ms_count);
  for(i = 0; i< mp->ms_count; i++){
    //TODO
  }

  _a2l_info("Number of Address mappings %i\n",mp->am_count);
  for(i = 0; i< mp->am_count; i++){
    _a2l_info("Index %i Start: 0x%X Map: 0x%X, Length: %X\n",
	      i,
	      mp->am[i].start,
	      mp->am[i].offset,
	      mp->am[i].len);
  }
}

/**************************************************************************/
void cs_a2l_print_record(
			 cs_a2l_record_t *re
			 )
/**************************************************************************/
{
  int i;
  if(re == NULL){
    _a2l_error("Parameter failure\n");
    return;
  }
  
  _a2l_info("Record Name %s\n",re->name);
  _a2l_info("Record Size %i\n",re->size);
  for(i = 0; i< re->size; i++){
    switch(re->rtype[i]){
    case A2L_RECORD_TYPE_UNKNOWN:
      _a2l_info("Record element %i: Unknown Type\n",i);
      break;
    case A2L_RECORD_TYPE_FNC:
      _a2l_info("Record element %i: Type FNC_VALUES\n",i);
      break;
    }
    _a2l_info("Record element %i: Position %i\n",i,re->position[i]);
    _a2l_info("Record element %i: Basictype %s\n",i,cs_a2l_TypeToString(re->dtype[i]));
    switch(re->access[i]){
    case A2L_RECORD_ACCESS_UNKNOWN:
      _a2l_info("Record element %i: Access type Unknown\n",i);
      break;
    case A2L_RECORD_ACCESS_DIRECT:
      _a2l_info("Record element %i: Access type Direct\n",i);
      break;
    }
    switch(re->orientation[i]){
    case A2L_RECORD_ORIENT_UNKNOWN:
      _a2l_info("Record element %i: Orientation Unknown\n",i);
      break;
    case A2L_RECORD_ORIENT_COLUMN:
      _a2l_info("Record element %i: Orientation Column\n",i);
      break;
    case A2L_RECORD_ORIENT_ROW:
      _a2l_info("Record element %i: Orientation Row\n",i);
      break;
    case A2L_RECORD_ORIENT_INDEXINCR:
      _a2l_info("Record element %i: Orientation Index increment\n",i);
      break;
    }
  }
}

/**************************************************************************/
void cs_a2l_print_compuvtab(
			    cs_a2l_compu_vtab_t *tab
			    )
/**************************************************************************/
{
  int i;
  if(tab == NULL){
    _a2l_error("Parameter failure\n");
    return;
  }
  
  _a2l_info("CompuVTab Name %s\n",tab->name);
  _a2l_info("CompuVTab Comment %s\n",tab->comment);
  switch(tab->ttype){
  case A2L_COMPU_TAB_TYPE_UNKNOWN:
    _a2l_info("CompuVTab Type Unknown\n");
    break;
  case A2L_COMPU_TAB_TYPE_VTAB:
    _a2l_info("CompuVTab Type direct lookup\n");
    break;
  case A2L_COMPU_TAB_TYPE_VTAB_RANGE:
    _a2l_info("CompuVTab Type range lookup\n");
    break;
  }
  _a2l_info("CompuVtab Size %i\n",tab->number);
  _a2l_info("CompuVtab Default %s\n",tab->edefault);
  
  for(i = 0; i< tab->number; i++){
    if(tab->ttype == A2L_COMPU_TAB_TYPE_VTAB_RANGE)
      _a2l_info("CompuVTab Index %i ID %i - %i:  %s\n",i,tab->elems[i].id_start,tab->elems[i].id_end,tab->elems[i].label);
    if(tab->ttype == A2L_COMPU_TAB_TYPE_VTAB)
      _a2l_info("CompuVTab Index %i ID %i:  %s\n",i,tab->elems[i].id_start,tab->elems[i].label);

  }
}


/**************************************************************************/
void cs_a2l_print_compumethod(
			      cs_a2l_compu_method_t *cm
			      )
/**************************************************************************/
{
  if(cm == NULL){
    _a2l_error("Parameter failure\n");
    return;
  }
    
  switch(cm->algo){
  case A2L_CM_UNDEF:
    _a2l_info("Compu Method Undefined\n");
    return;
    break;
  case A2L_CM_IDENT:
    _a2l_info("Compu Method No function defined, or take identical value\n");
    return;
    break;
  case A2L_CM_LINEAR:
    _a2l_info("Compu Method Linear: (ax+b) a= %f b= %f\n",cm->coeff[0],cm->coeff[1]);
    break;
  case A2L_CM_FUNCTION:
    _a2l_info("Compu Method Function: TODO\n");
    break;
  case A2L_CM_TAB_INTP:
    _a2l_info("Compu Method Table Lookup with Interpolation: TODO\n");
    break;
  case A2L_CM_TAB_NOINTP:
    _a2l_info("Compu Method Table Lookup without Interpolation: TODO\n");
    break;
  case A2L_CM_RAT:
    _a2l_info("Compu Method Rational function: (a*x*x+b*x+c)/(d*x*x+e*x+f)) a= %f b= %f c= %f d= %f e= %f f= %f\n",cm->coeff[0],cm->coeff[1],cm->coeff[2],cm->coeff[3],cm->coeff[4],cm->coeff[5]);
    break;
  case A2L_CM_TAB_VERB:
    _a2l_info("Compu Method Verbal Table lookup function\n");
    _a2l_info("Compu Method Table definition name: %s\n",cm->compuvtab);
    cs_a2l_print_compuvtab(&cm->tab);
    break;
  default:
    _a2l_info("Compu Method Unknown Algorythm\n");
  }
  _a2l_info("Compu Method Name %s\n",cm->name);
  _a2l_info("Compu Method Comment %s\n",cm->comment);
  _a2l_info("Compu Method Unit %s\n",cm->unit);
  _a2l_info("Compu Method Format %s\n",cm->format);

}

/**************************************************************************/
void cs_a2l_print_measurement(
			      cs_a2l_measurement_t *me
			      )
/**************************************************************************/
{
  if(me == NULL){
    _a2l_error("Parameter failure\n");
    return;
  }

  _a2l_info("Measurement Name %s\n",me->name);
  _a2l_info("Measurement Comment %s\n",me->comment);
  _a2l_info("Measurement Compu method %s\n",me->conv);
  switch(me->BO){
  case A2L_BO_MOTOROLLA:
    _a2l_info("Measurement Byte order Motorola\n");
    break;
  case A2L_BO_INTEL:
    _a2l_info("Measurement Byte order Intel\n");
    break;
  case A2L_BO_UNKNOWN:
    _a2l_info("Measurement Byte order not defined in signal\n");
    break;
  }
  _a2l_info("Measurement Type %s\n",cs_a2l_TypeToString(me->dtype));
  _a2l_info("Measurement Size %i\n",me->size);
  _a2l_info("Measurement Event index %i\n",me->ev_idx);
  _a2l_info("Measurement Prescaler %i\n",me->prescaler);
  _a2l_info("Measurement Minimum %f\n",me->min);
  _a2l_info("Measurement Maximum %f\n",me->max);
  _a2l_info("Measurement Resolution %i\n",me->resolution);
  _a2l_info("Measurement Aqurancy %i\n",me->aqurancy);
  _a2l_info("Measurement Prescaler %i\n",me->prescaler);
  _a2l_info("Measurement ECU address 0x%X\n",me->address);
  _a2l_info("Measurement ECU address extension 0x%X\n",me->extension);
  _a2l_info("Measurement Bitoffset 0x%X\n",me->bitoffset);

  cs_a2l_print_compumethod(&me->cm);
  
}

/**************************************************************************/
void cs_a2l_print_characteristic(
				 cs_a2l_characteristic_t *ch
				 )
/**************************************************************************/
{
  if (ch == NULL){
    _a2l_error("Parameter failure\n");
    return;
  }
    
  _a2l_info("Characteristic Name %s\n",ch->name);
  _a2l_info("Characteristic Comment %s\n",ch->comment);
  _a2l_info("Characteristic Compu method %s\n",ch->conv);
  _a2l_info("Characteristic Unit %s\n",ch->unit);
  _a2l_info("Characteristic Record %s\n",ch->recordname);
  _a2l_info("Characteristic Format %s\n",ch->format);
  _a2l_info("Characteristic MaxDiff %f\n",ch->maxdiff);
  _a2l_info("Characteristic Minimum %f\n",ch->min);
  _a2l_info("Characteristic Maximum %f\n",ch->max);
  _a2l_info("Characteristic ECU address 0x%X\n",ch->address);
  _a2l_info("Characteristic ECU address extension 0x%X\n",ch->extension);
  switch(ch->BO){
  case A2L_BO_MOTOROLLA:
    _a2l_info("Characteristic Byte order Motorola\n");
    break;
  case A2L_BO_INTEL:
    _a2l_info("Characteristic Byte order Intel\n");
    break;
  case A2L_BO_UNKNOWN:
    _a2l_info("Characteristic Byte order not defined in signal\n");
    break;
  }
  switch(ch->type){
  case A2L_CHARACTERISTIC_TYPE_UNKNOWN:
    _a2l_info("Characteristic Unknown characteristic type\n");
    break;
  case A2L_CHARACTERISTIC_TYPE_VALUE:
    _a2l_info("Characteristic Characteristic type value\n");
    break;
  case A2L_CHARACTERISTIC_TYPE_VALBLK:
    _a2l_info("Characteristic Characteristic type value block\n");
    break;
  case A2L_CHARACTERISTIC_TYPE_CURVE:
    _a2l_info("Characteristic Characteristic type curve\n");
    break;
  case A2L_CHARACTERISTIC_TYPE_ASCII:
    _a2l_info("Characteristic Characteristic type ASCII\n");
    break;
  }
  switch(ch->access){
  case A2L_CHARACTERISTIC_ACCESS_RW:
    _a2l_info("Characteristic Access read and write\n");
    break;
  case A2L_CHARACTERISTIC_ACCESS_RO:
    _a2l_info("Characteristic Access read only\n");
    break;
  default:
    _a2l_info("Characteristic Access unknown\n");
    break;
  }
    
  cs_a2l_print_compumethod(&ch->cm);
  cs_a2l_print_record(&ch->record);

}

/**************************************************************************/
void cs_a2l_print_ts(
		     cs_a2l_ts_unit_t ts_unit,
		     char *pretxt
		     )
/**************************************************************************/
{
  if(pretxt == NULL){
    _a2l_error("Parameter failure\n");
    return;
  }
  
  switch(ts_unit){
  case A2L_TS_UNIT_1NS:
    _a2l_info("%s Timestamp unit 1ns\n",pretxt);
    break;
  case A2L_TS_UNIT_10NS:
    _a2l_info("%s Timestamp unit 10ns\n",pretxt);
    break;
  case A2L_TS_UNIT_100NS:
    _a2l_info("%s Timestamp unit 100ns\n",pretxt);
    break;
  case A2L_TS_UNIT_1US:
    _a2l_info("%s Timestamp unit 1 micro s\n",pretxt);
    break;
  case A2L_TS_UNIT_10US:
    _a2l_info("%s Timestamp unit 10 micro s\n",pretxt);
    break;
  case A2L_TS_UNIT_100US:
    _a2l_info("%s Timestamp unit 100 micros\n",pretxt);
    break;
  case A2L_TS_UNIT_1MS:
    _a2l_info("%s Timestamp unit 1ms\n",pretxt);
    break;
  case A2L_TS_UNIT_10MS:
    _a2l_info("%s Timestamp unit 10ms\n",pretxt);
    break;
  case A2L_TS_UNIT_100MS:
    _a2l_info("%s Timestamp unit 100ms\n",pretxt);
    break;
  case A2L_TS_UNIT_1S:
    _a2l_info("%s Timestamp unit 1s\n",pretxt);
    break;
  case A2L_TS_UNIT_1PS:
    _a2l_info("%s Timestamp unit 1picos\n",pretxt);
    break;
  case A2L_TS_UNIT_10PS:
    _a2l_info("%s Timestamp unit 10picos\n",pretxt);
    break;
  case A2L_TS_UNIT_100PS:
    _a2l_info("%s Timestamp unit 10picos\n",pretxt);
    break;
  default:
    _a2l_info("%s Timestamp unit Unknown\n",pretxt);
    break;
  }
}

/**************************************************************************/
int cs_a2l_characteristic_l2d(
			      cs_a2l_characteristic_t *ch,
			      char *label,
			      int len,
			      double *value
			      )
/**************************************************************************/
{
  cs_a2l_compu_vtab_t *vtab = NULL;
  int i;

  if((ch == NULL)||(label == NULL)||(len <= 0)||(value == NULL)){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }
  *value = 0;

  //Check if we can convert
  if(ch->cm.algo != A2L_CM_TAB_VERB){
    _a2l_error("Label to Value: Invalid compumethod type\n");
    return CS_FAIL;
  }

  if(ch->cm.tab.ttype != A2L_COMPU_TAB_TYPE_VTAB){
    _a2l_error("Label to Value: Invalid Lookup type\n");
    return CS_FAIL;
  }

  if(len > CS_CHARBUF){
    _a2l_error("Label to Value: Label length out of range\n");
    return CS_FAIL;
  }

  vtab = &ch->cm.tab;
  
  if(vtab->number <= 0){
    _a2l_error("Label to Value: No Lookup elemets defined\n");
    return CS_FAIL;
  }

  for(i = 0; i < vtab->number;i++){
    if(strcmp(vtab->elems[i].label,label) == 0){
      *value = (double)vtab->elems[i].id_start;
      return CS_OK;
    }
  }

  _a2l_error("Label to Value: Label not found\n"); 
  return CS_FAIL;
}

/**************************************************************************/
int cs_a2l_characteristic_d2l(
			      cs_a2l_characteristic_t *ch,
			      char *label,
			      int *len,
			      double value
			      )
/**************************************************************************/
{
  cs_a2l_compu_vtab_t *vtab = NULL;
  int i;

  if((ch == NULL)||(label == NULL)||(len == NULL)||(*len <= 0)){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }

  if(*len < CS_CHARBUF){
    _a2l_error("Value to Label: Label buffer length out of range\n");
    return CS_FAIL;
  }

  memset(label,0,sizeof(char)**len);
  *len = 0;

  //Check if we can convert
  if(ch->cm.algo != A2L_CM_TAB_VERB){
    _a2l_error("Value to Label: Invalid compumethod type\n");
    return CS_FAIL;
  }

  if(ch->cm.tab.ttype == A2L_COMPU_TAB_TYPE_UNKNOWN){
    _a2l_error("Label to Value: Lookup type not configured\n");
    return CS_FAIL;
  }

  vtab = &ch->cm.tab;
  
  if(vtab->number <= 0){
    _a2l_error("Value to Label: No Lookup elemets defined\n");
    return CS_FAIL;
  }

  switch(vtab->ttype){
  case A2L_COMPU_TAB_TYPE_UNKNOWN:
    break;
  case A2L_COMPU_TAB_TYPE_VTAB:
    for(i = 0; i < vtab->number;i++){
      if(vtab->elems[i].id_start == (int)value){
	strncpy(label,vtab->elems[i].label,CS_CHARBUF);
	*len = CS_CHARBUF;
	return CS_OK;
      }
    }
    break;
  case A2L_COMPU_TAB_TYPE_VTAB_RANGE:
    for(i = 0; i < vtab->number;i++){
      if((vtab->elems[i].id_start <= (int)value) && (vtab->elems[i].id_end > (int)value)){
	strncpy(label,vtab->elems[i].label,CS_CHARBUF);
	*len = CS_CHARBUF;
	return CS_OK;
      }
    }
    break;
  }
  _a2l_error("Value to Label: Value not found\n"); 
  return CS_FAIL;
}

/**************************************************************************/
int cs_a2l_characteristic_limit(
				cs_a2l_characteristic_t *ch,
				double *value
				)
/**************************************************************************/
{
  if((ch == NULL) || (value == NULL)){
    _a2l_error("Parameter failure\n");
    return CS_FAIL;
  }

  //Check for min
  if(*value < ch->min){
    _a2l_verbose("Check Limits: Value < Min\n");
    *value =  ch->min;
  }

  //Check for max
  if(*value > ch->max){
    _a2l_verbose("Check Limits: Value > Max\n");
    *value =  ch->max;
  }

  return CS_OK;
}
