/*  XCP CAN Stack for SocketCAN
    Copyright (C) 2014 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
#include <time.h>
#include <math.h>
#include <byteswap.h>

#include "cs_defines.h"
#include "cs_debug_private.h"
#include "cs_calibration_utils.h"


/*************************************************************************************************/
double cs_cu_ts_unit2double(cs_a2l_ts_unit_t ts_unit, uint8_t period_unit)
/*************************************************************************************************/
{
    double period = 0;
    double scale = 0;
  
    switch(ts_unit){
    case A2L_TS_UNIT_1NS:
        scale = 1.0e-9;
        break;
    case A2L_TS_UNIT_10NS:
        scale = 1.0e-8;
        break;
    case A2L_TS_UNIT_100NS:
        scale = 1.0e-7;
        break;
    case A2L_TS_UNIT_1US:
        scale = 1.0e-6;
        break;
    case A2L_TS_UNIT_10US:
        scale = 1.0e-5;
        break;
    case A2L_TS_UNIT_100US:
        scale = 1.0e-4;
        break;
    case A2L_TS_UNIT_1MS:
        scale = 1.0e-3;
        break;
    case A2L_TS_UNIT_10MS:
        scale = 1.0e-2;
        break;
    case A2L_TS_UNIT_100MS:
        scale = 1.0e-1;
        break;
    case A2L_TS_UNIT_1S:
        scale = 1.0;
        break;
    case A2L_TS_UNIT_1PS:
        scale = 1.0e-12;
        break;
    case A2L_TS_UNIT_10PS:
        scale = 1.0e-11;
        break;
    case A2L_TS_UNIT_100PS:
        scale = 1.0e-12;
        break;
    }
    period = (double)period_unit * scale;
    return period;
}

/*************************************************************************************************/
struct timespec cs_cu_ts_timestamp2ts(uint32_t timestamp, cs_a2l_ts_unit_t resolution)
/*************************************************************************************************/
{
    struct timespec ts;
    uint64_t ts_int = 0;
    double factor = 0;
    int skip = 0;
    switch(resolution){
    case A2L_TS_UNIT_1NS:
    case A2L_TS_UNIT_10NS:
    case A2L_TS_UNIT_100NS:
    case A2L_TS_UNIT_1US:
    case A2L_TS_UNIT_10US:
    case A2L_TS_UNIT_100US:
    case A2L_TS_UNIT_1MS:
    case A2L_TS_UNIT_10MS:
    case A2L_TS_UNIT_100MS:
    case A2L_TS_UNIT_1S:
        factor = pow(10,(double)resolution);
        break;
    case A2L_TS_UNIT_1PS:
    case A2L_TS_UNIT_10PS:
    case A2L_TS_UNIT_100PS:
        //Resolutiom max ns
        factor = pow(10,9);
        skip = (0xC - (uint8_t)resolution);
        break;
    }
    ts_int = (timestamp >> skip) *factor;
    ts.tv_nsec = (int)fmod((double)ts_int,1e9);
    ts.tv_sec = ts_int/1e9;
    return ts;
}


/**************************************************************************/
cs_a2l_bo_t cs_cu_endianess()
/**************************************************************************/
{
    union { long l; char c[sizeof (long)]; } u;
    u.l = 1;
    if(u.c[sizeof (long) - 1] == 1)
        return A2L_BO_MOTOROLLA; //Big Endian, Motorola
    else
        return A2L_BO_INTEL; //Little Endian, Intel

}

/**************************************************************************/
uint16_t cs_cu_swap16(
		      cs_a2l_bo_t endianess,
		      uint16_t value
		      )
/**************************************************************************/
{
    uint16_t rv;
    if(endianess != A2L_BO_UNKNOWN){
      if(cs_cu_endianess() != endianess){
        if(cs_cu_endianess() == A2L_BO_MOTOROLLA)
	  _int_verbose("Swap 16: Bit Motorola to Intel\n");
        else
	  _int_verbose("Swap 16: Bit Intel to Motorola\n");
        //swap bytes
        rv = bswap_16(value);
        _int_verbose("Swap 16: In: 0x%X, OUT 0x%X\n",value,rv);
        return rv;
      }
    } else {
      _int_debug("Swap16: Warning Byte order is not defined\n");
    }
    return value;
}

/**************************************************************************/
uint32_t cs_cu_swap32(
    		      cs_a2l_bo_t endianess,
		      uint32_t value
		      )
/**************************************************************************/
{
    uint32_t rv;
    if(endianess != A2L_BO_UNKNOWN){
      if(cs_cu_endianess() != endianess){
        if(cs_cu_endianess() == A2L_BO_MOTOROLLA)
	  _int_verbose("Swap 32: Bit Motorola to Intel\n");
        else
	  _int_verbose("Swap 32: Bit Intel to Motorola\n");
        //swap bytes
        rv = bswap_32(value);
        _int_verbose("Swap 32: In: 0x%X, OUT 0x%X\n",value,rv);
        return rv;
      }
    } else {
      _int_debug("Swap32: Warning Byte order is not defined\n");
    }
    return value;
}

/**************************************************************************/
uint64_t cs_cu_swap64(
    		      cs_a2l_bo_t endianess,
		      uint64_t value
		      )
/**************************************************************************/
{
    uint64_t rv;
    if(endianess != A2L_BO_UNKNOWN){
      if(cs_cu_endianess() != endianess){
        if(cs_cu_endianess() == A2L_BO_MOTOROLLA)
	  _int_verbose("Swap 64: Bit Motorola to Intel\n");
        else
	  _int_verbose("Swap 64: Bit Intel to Motorola\n");
        //swap bytes
        rv = bswap_64(value);
        _int_verbose("Swap 64: In: 0x%X, OUT 0x%X\n",value,rv);
        return rv;
      }
    } else {
      _int_debug("Swap64: Warning Byte order is not defined\n");
    }
    return value;
}
