/*  CCP CAN Stack for SocketCAN
    Copyright (C) 2015 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <byteswap.h>
#include <unistd.h>

#include "config.h"
#include "cs_defines.h"
#include "cs_time.h"
#include "cs_calibration_utils.h"
#include "cs_socketcan.h"
#include "cs_ccp.h"
#include "cs_ccp_private.h"


/********************************************************************/
cs_debug_level_t cs_ccp_debuglevel(
				   cs_debug_level_t level
				   )
/********************************************************************/
{
  return cs_debuglevel(CS_DEBUG_CCP,level);
}

/********************************************************************/
void cs_ccp_debug_setcallback(
			      void *ctx,
			      void (*callb)(void *ctx,
					    cs_debug_sub_t sub,
					    cs_debug_level_t dl,
					    const char *func,
					    const int line,
					    const char *fmt,
					    va_list ap)
			      )
/********************************************************************/
{
  cs_debug_setcallback(CS_DEBUG_CCP,ctx,callb);
}

/**************************************************************************/
void _ccp_reset_error(
		      cs_ccp_t *data
		      )
/**************************************************************************/
{ 
  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return;
  }

  data->lerror = 0;
  memset(data->lerrorstr,0,CS_CHARBUF);

}

/**************************************************************************/
void _ccp_reset_pkg(
		    cs_ccp_pkg_t *pkg
		    )
/**************************************************************************/
{
  if(pkg == NULL){
    _ccp_error("Parameter failure\n");
    return;
  }
  
  if(pkg == NULL){
    _ccp_error("Parameter failure\n");
    return;
  }
  memset(pkg,0,sizeof(cs_ccp_pkg_t));
}


/**************************************************************************/
int _ccp_odt_isactive(
		      cs_ccp_daq_t *daq,
		      uint8_t odt
		      )
/**************************************************************************/
{
  int i;

  if(daq == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  for(i = 0; i < daq->no_active; i++){
    if(odt == daq->odt_active[i])
      return CS_OK;
  }
  return CS_FAIL;
}

/**************************************************************************/
int _ccp_odt_setactive(
		       cs_ccp_daq_t *daq,
		       uint8_t odt,
		       uint8_t active //0=inactive,1=active
		       )
/**************************************************************************/
{
  int i;
  int counter;

  if(daq == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  uint8_t store[daq->no_odt];

  
  
  if(odt >= daq->no_odt)
    return CS_FAIL;

  if(active){
    //Add to vector
    
    //Check if it is in vector
    if(_ccp_odt_isactive(daq,odt)){
      daq->odt_active[daq->no_active] = odt;
      daq->no_active++;
      if(daq->no_active > daq->no_odt){
	daq->no_active--;
	_ccp_error("Out of range error\n");
	return CS_FAIL;
      }
    }
    return CS_OK;
  } else {
    //Remove from vector

    //Check if it is in vector
    if(_ccp_odt_isactive(daq,odt))
      return CS_OK;
    //Backup
    memcpy(store,daq->odt_active,sizeof(uint8_t)*daq->no_odt);
    //Reset
    memset(daq->odt_active,0xFF,sizeof(uint8_t)*daq->no_odt);
    counter = 0;
    for(i = 0; i < daq->no_active; i++){
      if(odt == store[i])
	continue;
      daq->odt_active[counter] = store[i];
      counter++;
    }
    daq->no_active = counter;
    return CS_OK;
  }
  return CS_FAIL;
}

/**************************************************************************/
int _ccp_daq_isactive(
		      cs_ccp_t *ccp,
		      uint8_t daq
		      )
/**************************************************************************/
{
  int i;

  if(ccp == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  for(i = 0; i < ccp->daqlist.no_active; i++){
    if(daq == ccp->daqlist.daq_active[i])
      return CS_OK;
  }
  return CS_FAIL;
}

/**************************************************************************/
int _ccp_daq_setactive(
		       cs_ccp_t *data,
		       uint8_t daq,
		       uint8_t active //0=inactive,1=active
		       )
/**************************************************************************/
{
  int i;
  int counter;
  uint8_t store[CCP_DAQMAX];

  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  if(daq >= CCP_DAQMAX)
    return CS_FAIL;

  if(active){
    //Add to vector
    
    //Check if it is in vector
    if(_ccp_daq_isactive(data,daq)){
      data->daqlist.daq_active[data->daqlist.no_active] = daq;
      data->daqlist.no_active++;
      if(data->daqlist.no_active > CCP_DAQMAX){
	data->daqlist.no_active--;
	_ccp_error("Out of range error\n");
	return CS_FAIL;
      }
    }
    return CS_OK;
  } else {
    //Remove from vector

    //Check if it is in vector
    if(_ccp_daq_isactive(data,daq))
      return CS_OK;
    //Backup
    memcpy(store,data->daqlist.daq_active,sizeof(uint8_t)*CCP_DAQMAX);
    //Reset
    memset(data->daqlist.daq_active,0xFF,sizeof(uint8_t)*CCP_DAQMAX);
    counter = 0;
    for(i = 0; i < data->daqlist.no_active; i++){
      if(daq == store[i])
	continue;
      data->daqlist.daq_active[counter] = store[i];
      counter++;
    }
    data->daqlist.no_active = counter;
    return CS_OK;
  }
  return CS_FAIL;
}



/*************************************************************************************************/
int _ccp_can_recv_block(cs_ccp_t *ccp, uint32_t *id, uint8_t *data, int *dl, uint32_t timeout, struct timeval *tv_recv)
/*************************************************************************************************/
{

  int rv;
  int loop = 1;

  if((ccp == NULL) || 
     (data == NULL) || 
     (id == NULL) ||
     (tv_recv == NULL) || 
     (dl == NULL)){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }

  _ccp_verbose("CCP CAN Receive blocking\n");
  memset(data,0,sizeof(uint8_t)*CCP_DATALENGTH);

  while(loop){
    rv = cs_socketcan_recv_block(ccp->socket,id,data,dl,tv_recv,timeout);
    switch(rv){
    case CS_SOCKETCAN_OK:
      loop = 0;
      break;
    case CS_SOCKETCAN_FAIL:
      loop = 0;
      return CS_FAIL;
      break;
    case CS_SOCKETCAN_RTR:
    case CS_SOCKETCAN_ERROR:
      //continue recv
      loop++;
      break;
    default:
      return CS_FAIL;
    }
    if(loop > 10)
      return CS_FAIL;
  }

  _ccp_verbose("Received buffer, CAN id hex:%X, Data length  %i\n",*id,*dl);
  _ccp_verbose_buffer(data,*dl);
  return CS_OK;
}


/*************************************************************************************************/
int _ccp_can_recv_nonblock(cs_ccp_t *ccp, uint32_t *id, uint8_t *data, int *dl, struct timeval *tv_recv)
/*************************************************************************************************/
{
  if((ccp == NULL) || 
     (data == NULL) || 
     (id == NULL) ||
     (tv_recv == NULL) || 
     (dl == NULL)){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }

  _ccp_verbose("CCP CAN Receive non blocking\n");

  memset(data,0,sizeof(uint8_t)*CCP_DATALENGTH);

  if(!cs_socketcan_recv_nonblock(ccp->socket,id,data,dl,tv_recv)){    
    _ccp_verbose("Received buffer, CAN id hex:%X\n",*id);
    _ccp_verbose_buffer(data,*dl);
    return CS_OK;
  }
  
  return CS_FAIL;
}

/** Internal send routine
 */
/**************************************************************************/
int _ccp_send(
	      cs_ccp_t *data,
	      cs_ccp_pkg_t *pkg
	      )
/**************************************************************************/
{
  struct timespec ts;
  int rv;

  if((data == NULL) || (pkg == NULL)){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }
    
  //Get send timestamp
  cs_ts_gettime(&ts);
  //Store time
  pkg->tv = cs_ts_timespec2timeval(ts);
  rv = cs_socketcan_send(data->socket,data->ccpa2l.can_send_id,pkg->data,CCP_DATALENGTH,SOCKETCAN_SEND_CAN);
  switch(rv){
  case CS_SOCKETCAN_FAIL:
    _ccp_error("SocketCAN send failure\n");
    data->lerror = CCP_ERROR_SEND;
    return CS_FAIL;
    break;
  case CS_SOCKETCAN_OK:
    return CS_OK;
    break;
  case CS_SOCKETCAN_TXFULL:
    _ccp_error("SocketCAN TX-Queque full\n");
    data->lerror = CCP_ERROR_TXFULL;
    return CS_FAIL;
    break;
  default:
    return CS_FAIL;
  }
  return CS_OK;
}

/** Received DTO, fill into internal structure
    \retval 0 No failure
    \retval 1 Failure
*/
/**************************************************************************/
int _ccp_dto_daq_process(
			 cs_ccp_t *data,
			 uint32_t canid,
			 uint8_t *sdata,
			 int ldata,
			 struct timeval tv
			 )
/**************************************************************************/
{
  int i,j;
  uint8_t odt_pid;
  cs_ccp_daq_t *daq = NULL;
  cs_ccp_odt_t *odt = NULL;
  int flag_first_odt;
  int data_offset = 1;
  double value;
  uint8_t daq_id;
  int did;
  int oid;

  if((data == NULL) ||
     (sdata == NULL)){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }
    
  _ccp_verbose("Start\n");

  if(canid != data->ccpa2l.can_recv_id){
    //DAQ identified by CAN id, actually not supported
    return CS_FAIL;
    for(i = 0; i < data->daqlist.no_active;i++){
      did =  data->daqlist.daq_active[i];
      if(canid == data->daqlist.daq[did].canid){
	//Found daq
	daq = &data->daqlist.daq[did];
	//todo act on data
      }
    }
  }

  odt_pid = sdata[0]; //Get PID
  
  //search odt
  for(i = 0; i < data->daqlist.no_active;i++){
    daq_id = data->daqlist.daq_active[i];
    for(j =0; j < data->daqlist.daq[daq_id].no_active; j++){
      oid = data->daqlist.daq[daq_id].odt_active[j]; 
      if((((data->daqlist.daq[daq_id].first_pid + j) == odt_pid) && data->daqlist.daq[daq_id].running)){
	_ccp_verbose("Found odt: %i, daq %i, odt index %i\n",odt_pid,i,oid);
	if(oid == 0){
	  flag_first_odt = 1;
	} else {
	  flag_first_odt = 0;
	}
	odt = &data->daqlist.daq[daq_id].odt[oid];
	daq = &data->daqlist.daq[daq_id];
	break;
      }
    }
  }
 
  if(odt == NULL){
    _ccp_verbose("Unknown ODT\n");
    return CS_OK;
  }

  //Calculate data offset
  if(flag_first_odt){
    daq->counter++;
    //take receive timestamp of first ODT
    daq->ts_ref = cs_ts_timeval2d(tv);
    _ccp_verbose("Counter %i\n",daq->counter);
  }

  //Set receive timestamp for ODT, to be able to check  for receive timeout
  odt->last = cs_ts_timeval2d(tv);


  //Do data handling
  //note only data copy, no byte swapping
  for(i = 0; i < odt->no_entry; i++){
    _ccp_verbose("Entry %i, copy size: %i\n",i,odt->entry[i].me.size);
    //Memcopy
    memcpy(odt->entry[i].data,&sdata[data_offset],odt->entry[i].me.size);
    
    //Calculate value
    value = 0.0;
    cs_a2l_convert_to_double(
			     odt->entry[i].me.BO,
			     odt->entry[i].data,
			     odt->entry[i].me.size,
			     odt->entry[i].me.cm,
			     odt->entry[i].me.dtype,
			     &value);
    
    //Connect the received value with external references
      
    //Do we have an external value pointer
    if(odt->entry[i].me.exvalue!= NULL){
      *odt->entry[i].me.exvalue = value;
    }
    //Do we have an external timestamp pointer
    if(odt->entry[i].me.exts!= NULL){
      *odt->entry[i].me.exts = daq->ts_ref;
    }
    
    //Do we have an external callback funtion
    if(odt->entry[i].me.extcallback != NULL){
      //Call external function
      (*odt->entry[i].me.extcallback)(value,daq->ts_ref,odt->entry[i].me.extcallbackpriv);
    }

    //Increment data offset
    data_offset += odt->entry[i].me.size;
  }
  return CS_OK;
}

/**************************************************************************/
int _ccp_receive(
		 cs_ccp_t *data,
		 cs_ccp_pkg_t *pkg,
		 uint32_t timeout,
		 int mode_odt
		 )
/**************************************************************************/
{
  uint8_t sdata[CCP_DATALENGTH];
  int ldata = CCP_DATALENGTH;
  uint32_t can_id;
  struct timespec ts_timeout;
  uint8_t do_receive = 1;
  uint8_t received_pkg = 0;


  if((data == NULL) ||
     (pkg == NULL)){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  //Get actual time
  cs_ts_gettime(&ts_timeout);
  //Add timeout
  cs_ts_add_timeout(&ts_timeout,timeout);
  
  pkg->dl = 0;
  
  while(do_receive){
    //Reset raw buffer
    memset(pkg->data,0,sizeof(uint8_t)*CCP_DATALENGTH);
    _ccp_verbose("Receive ... Mode: %s\n",mode_odt?"ODT":"Standard");
    //Check for timeout if we have received one or more ODT package
    if(!mode_odt && cs_ts_checkpast(&ts_timeout)){
      _ccp_verbose("Timeout error, timestamp\n");
      data->lerror = CCP_ERROR_RECVTIMEOUT;
      do_receive = 0;
      break;
    }
    //Received
    received_pkg = 0;
    if(!mode_odt){
      _ccp_verbose("Exec standard mode\n");
      if( _ccp_can_recv_block(data,&can_id,sdata,&ldata,timeout,&pkg->tv)){
	_ccp_verbose("Timeout error\n");
	data->lerror = CCP_ERROR_RECVTIMEOUT;
	do_receive = 0;
      } else {
	received_pkg = 1;
      }
      if(received_pkg && (can_id != data->ccpa2l.can_recv_id)){ //Receive a daq odt with special CAN id
	_ccp_verbose("Unknown CAN id 0x%X, try DAQ processing\n",can_id);
	//_ccp_dto_daq_process(data,can_id,sdata,ldata,pkg->tv);
	received_pkg = 0;	
      }
    } else {
      _ccp_verbose("Exec ODT mode\n");
      if(_ccp_can_recv_nonblock(data,&can_id,sdata,&ldata,&pkg->tv)){
	data->lerror = CCP_ERROR_NOERROR;
	do_receive = 0;
      } else {
	received_pkg = 1;
      }
    } 
    if(received_pkg){
      _ccp_verbose("Received package\n");
      //Chek for invalid data length
      if(!mode_odt){
	if(ldata == 0){
	  data->lerror = CCP_ERROR_RECV;
	  do_receive = 0;
	}
      } else {
	if(ldata == 0){
	  pkg->dl = 0;
	  _ccp_error("ODT mode: Invalid data length\n");
	  return CS_FAIL;
	}
      }

      //Process raw data
      if((ldata > 0) && (data->lerror == CCP_ERROR_NOERROR)){
	switch(sdata[0]){
	case CCP_DTO_CRM:
	case CCP_DTO_EV:
	  _ccp_verbose("Received Response or event paket\n");
	  switch(sdata[1]){
	  case CCP_ERROR_NOERROR:
	  case CCP_RC_DAQOVERLOAD:
	    //Process Response package
	    pkg->dl = ldata;
	    //Check for buffer overflow
	    memcpy(pkg->data,sdata,pkg->dl);
	    do_receive = 0;
	    break;	    
	  case CCP_RC_CMDBUSY:
	  case CCP_RC_DAQBUSY:
	  case CCP_RC_INTERNALTIMEOUT:
	    //Reset timeout
	    //Get actual time
	    cs_ts_gettime(&ts_timeout);
	    //Add timeout
	    cs_ts_add_timeout(&ts_timeout,timeout);
	    _ccp_verbose("Received a Command pending package. Reset timeout.\n");
	    break;
	  case CCP_RC_OVERLOAD:
	  case CCP_RC_CMDUNKNOWN:
	  case CCP_RC_CMDSYNTAX:
	  case CCP_RC_PARAMETER:
	  case CCP_RC_ACCESSDENIED:
	  case CCP_RC_ACCESSLOCKED:
	  case CCP_RC_NOTAVAILABLE:
	    data->lerror = sdata[1];
	    pkg->dl = 0;
	    return CS_FAIL;
	    break;
	  default:
	    //Handle asynchron return codes
	    if(sdata[1] & 0x20){ //Asynchron return code
	      switch(sdata[1]){
	      case CCP_RC_COLDRESTART:
		data->indicator |= CCP_INDICATION_COLDRESTART;
		break;
	      case CCP_RC_CALINIT:
		data->indicator |= CCP_INDICATION_CALINIT;
		break;
	      case CCP_RC_DAQINIT:
		data->indicator |= CCP_INDICATION_DAQINIT;
		break;
	      case CCP_RC_CODEUPDATE:
		data->indicator |= CCP_INDICATION_CODEUPDATE;
		break;
	      default:
		_ccp_verbose("Received unknown asynchron return code\n");
	      }
	    } else {
	      data->lerror = sdata[1];
	      pkg->dl = 0;
	      _ccp_error("Received a Unknown return code.\n");
	      return CS_FAIL;
	    }
	  }
	  if(sdata[0] == CCP_DTO_CRM){ //Handle data, if we received async return codes
	    //Process Response package
	    pkg->dl = ldata;
	    //Check for buffer overflow
	    memcpy(pkg->data,sdata,pkg->dl);
	    do_receive = 0;
	    break;	    
	  }
	default:
	  //Handle DAQ messages
	  _ccp_verbose("Received unknown package id\n");
	  //All other are DT0 packages
	  //Process
	  if(_ccp_dto_daq_process(data,can_id,sdata,ldata,pkg->tv)){
	    _ccp_error("DAQ processing failed\n");
	    pkg->dl = 0;
	    return CS_FAIL;
	  }
	  if(mode_odt)
	    do_receive = 0;
	  break;	  
	}
      }       
    } else {
      _ccp_verbose("No package received\n");
    }
  }
  
  if(!received_pkg)
    return CS_FAIL;
    
  if(data->lerror != CCP_ERROR_NOERROR){
    _ccp_verbose("Error in progress\n");
    return CS_FAIL;
  } else
    return CS_OK;
}



/**************************************************************************/
int _ccp_send_recv_action(
			  cs_ccp_t *data,
			  int cmd,
			  cs_ccp_pkg_t *pkg_send,
			  cs_ccp_pkg_t *pkg_recv,
			  uint8_t ignore_feedback
			  )
/**************************************************************************/
{ 
  int repeat = 1;
  int repeat_counter = 0;
  cs_ccp_pkg_t pkg_dummy;
  int i;

  if((data == NULL) ||
     (pkg_send == NULL) ||
     (pkg_recv == NULL)){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }
    
  if(!ignore_feedback && data->ignore_counter > 0){
    data->ignore_counter = 0;
    for(i = 0; i < data->ignore_counter; i++){
      //Loop for not read ack messages
      _ccp_verbose("Loop for ignored messages");
      _ccp_receive(data,&pkg_dummy,1,0);
      data->lerror = CCP_ERROR_NOERROR;
    }
  }
  _ccp_verbose("Timeout %i\n", data->cmd[cmd].timeout);
  while(repeat_counter < repeat){
    _ccp_verbose("Send Recv Retry Cmd: 0x%X, Retry %i, Last Error %s\n",cmd,repeat_counter,cs_ccp_getlasterrorstring(data));
    _ccp_reset_error(data);
      
    if(_ccp_send(data,pkg_send)){
      _ccp_error("Send failure\n");
      return CS_FAIL;
    }
    if(!ignore_feedback){
      if(_ccp_receive(data,pkg_recv,data->cmd[cmd].timeout,0)){
	_ccp_verbose("Receive negative response: %s\n",cs_ccp_getlasterrorstring(data));
	//error received	
      } else {
	//positive response
	_ccp_verbose("Positive response got DL: %i\n",pkg_recv->dl);
      }
    } else {
      data->ignore_counter++;
    }
    repeat_counter++;
  }
    
  if(data->lerror != CCP_ERROR_NOERROR)
    return CS_FAIL;
  else
    return CS_OK;
}

/********************************************************************/
int _ccp_check_cmd(
		   cs_ccp_t *data,
		   uint8_t cmd
		   )
/********************************************************************/
{
  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  if(!data->cmd[cmd].supported){
    _ccp_error("CMD is not supported by API\n");
    data->lerror = CCP_ERROR_NOTSUPPORTEDAPI;
    return CS_FAIL;
  }

  if(data->cmd[cmd].optional && !data->cmd[cmd].isoperational){
    _ccp_error("CMD is not supported by Slave\n");
    data->lerror = CCP_ERROR_NOTSUPPORTEDSLAVE;
    return CS_FAIL;
  }
  
  return CS_OK;
}

/** Reset ODT enry
 */
/**************************************************************************/
void _ccp_reset_odt_entry(
			  cs_ccp_odt_entry_t *entry
			  )
/**************************************************************************/
{
  if(!entry){
    _ccp_error("Parameter failure\n");
    return;
  }
  
  _ccp_verbose("_CCP Reset ODT entry\n");  
  memset(entry->data,0,sizeof(uint8_t)*CCP_DATALENGTH);
  cs_a2l_reset_measurement(&entry->me);
}

/** Reset ODT
 */
/**************************************************************************/
void _ccp_reset_odt(
		    cs_ccp_odt_t *odt
		    )
/**************************************************************************/
{
  int i;
  if(!odt){
    _ccp_error("Parameter failure\n");
    return;
  }
  
  _ccp_verbose("_CCP Reset ODT\n");
  for(i = 0; i < CCP_ODTENTRYMAX; i++){
    if(&odt->entry[i])
      _ccp_reset_odt_entry(&odt->entry[i]);
  }
  odt->no_entry = 0;
  odt->datasize = 0;
  odt->last = 0;
}


/** Reset DAQ entry
 */
/**************************************************************************/
void _ccp_reset_daq(
		    cs_ccp_daq_t *daq
		    )
/**************************************************************************/
{
  int i;
  
  if(!daq){
    _ccp_error("Parameter failure\n");
    return;
  }
  
  _ccp_verbose("_CCP Reset DAQ\n");
  
  if(daq->odt){
    for(i = 0; i< daq->no_odt; i++){
      _ccp_reset_odt(&daq->odt[i]);
    }
    free(daq->odt);
    daq->odt = NULL;
    free(daq->odt_active);
    daq->odt_active = NULL;
  }
  daq->no_odt = 0;
  daq->no_active = 0;
  daq->resume = 0;            
  daq->running = 0;      
  daq->selected = 0;
  daq->ts_old = 0;
  daq->counter = 0;
  daq->canid = 0;
  daq->period = -1.0; //Negative period to skip timeout check if we have no
                      //period information
}

/** Reset DAQ list
 */
/**************************************************************************/
void _ccp_reset_daqlist(
			cs_ccp_t *data
			)
/**************************************************************************/
{
  int i;

  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return;
  }
  
  _ccp_verbose("_CCP Reset DAQ list\n");
  for(i = 0; i < CCP_DAQMAX; i++){
    _ccp_reset_daq(&data->daqlist.daq[i]);
    data->daqlist.daq_active[i] = 0xFF; //set invalid
  }
  data->daqlist.no_active = 0;
  data->daqlist.clock_shift = 0;
  data->daqlist.clock_base = 0;
  data->daqlist.daq_fac_recvtimeout = 0.0;
}

/** Prepare ODT structure for start
 */
/**************************************************************************/
void _ccp_start_odt(
		    cs_ccp_odt_t *odt
		    )
/**************************************************************************/
{
  int j;

  if(odt == NULL){
    _ccp_error("Parameter failure\n");
    return;
  }
  
  odt->datasize = 0;
  for(j = 0; j < CCP_ODTENTRYMAX; j++){
    odt->datasize += odt->entry[j].me.size;
  }
  //Set last receive time to zero
  odt->last = 0.0;
}

/** Signal start daq in local data structures
 */
/**************************************************************************/
void _ccp_daq_start( 
		    cs_ccp_t *data,
		    int daq
		     )
/**************************************************************************/
{
  int i;
  struct timespec ts_now;
  double now;
  int oid;
  
  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return;
  }
  
  if(_ccp_daq_isactive(data,daq))
    return;

  //Make note for started DAQ list
  data->daqlist.daq[daq].running = 1;
  //Disable selected flag
  data->daqlist.daq[daq].selected = 0;

  for(i = 0; i < data->daqlist.daq[daq].no_active; i++){
    oid = data->daqlist.daq[daq].odt_active[i];
    _ccp_start_odt(&data->daqlist.daq[daq].odt[oid]);
  }
  //Reset counter
  data->daqlist.daq[daq].counter = 0;
  //Set timestamp of next expected package
  if(data->daqlist.daq[daq].period > 0.0){
    cs_ts_gettime(&ts_now);
    now = cs_ts_timespec2d(ts_now);
    data->daqlist.daq[daq].ts_expect = now + data->daqlist.daq[daq].period;
  }
}

/** Signal start selected daq in local data structures
 */
/**************************************************************************/
void _ccp_daq_start_selected( 
			     cs_ccp_t *data
			      )
/**************************************************************************/
{
  int i,k;
  struct timespec ts_now;
  double now;
  int daqid;
  int oid;

  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return;
  }
  
  for(k = 0; k < data->daqlist.no_active; k++){
    daqid = data->daqlist.daq_active[k];
    if(data->daqlist.daq[daqid].selected){
      //Make note for started DAQ list
      data->daqlist.daq[daqid].running = 1;
      //Calculate size of ODT to campare it with received data
      for(i = 0; i < data->daqlist.daq[daqid].no_active; i++){
	oid = data->daqlist.daq[daqid].odt_active[i];
	_ccp_start_odt(&data->daqlist.daq[daqid].odt[oid]);
      }
      //Disable selected flag
      data->daqlist.daq[daqid].selected = 0;
      //Reset counter
      data->daqlist.daq[daqid].counter = 0;
      if(data->daqlist.daq[daqid].period > 0.0){
	//Set timestamp of next expected package
	cs_ts_gettime(&ts_now);
	now = cs_ts_timespec2d(ts_now);
	data->daqlist.daq[daqid].ts_expect = now + data->daqlist.daq[daqid].period;
      }
    }
  }
}

/** Signal start selected daq in local data structures
 */
/**************************************************************************/
void _ccp_daq_stop_selected( 
			    cs_ccp_t *data
			     )
/**************************************************************************/
{
  int i,k;
  int daqid;
  int oid;

  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return;
  }
  
  for(k = 0; k < data->daqlist.no_active; k++){
    daqid = data->daqlist.daq_active[k];
    if(data->daqlist.daq[daqid].selected){
      //Make note for started DAQ list
      data->daqlist.daq[daqid].running = 0;
      for(i = 0; i < data->daqlist.daq[daqid].no_active; i++){
	oid = data->daqlist.daq[daqid].odt_active[i];
	data->daqlist.daq[daqid].odt[oid].datasize = 0;
      }
      //Disable selected flag
      data->daqlist.daq[daqid].selected = 0;
    }
  }
}

/** Signal stop daq in local data structures
 */
/**************************************************************************/
void _ccp_daq_stop( 
		   cs_ccp_t *data,
		   int daq
		    )
/**************************************************************************/
{

  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return;
  }
  
  if(_ccp_daq_isactive(data,daq))
    return;

  data->daqlist.daq[daq].running = 0;
  data->daqlist.daq[daq].selected = 0;
}

/** Signal select daq in local data structures
 */
/**************************************************************************/
void _ccp_daq_select( 
		     cs_ccp_t *data,
		     int daq
		      )
/**************************************************************************/
{

  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return;
  }
  
  if(!_ccp_daq_isactive(data,daq))
    data->daqlist.daq[daq].selected = 1;
}

/********************************************************************/
uint16_t cs_ccp_getversion()
/********************************************************************/
{
  return CS_CCP_VERSION_MAGIC;
}

/********************************************************************/
uint8_t cs_ccp_getlasterror(
			    cs_ccp_t *data         /**< CCP stack structure*/ 
			    )
/********************************************************************/
{
  if(data != NULL)
    return data->lerror;
  
  _ccp_error("Parameter failure\n");
  return 0;
}

/********************************************************************/
char *cs_ccp_getlasterrorstring(
				cs_ccp_t *data         /**< CCP stack structure*/ 
				)
/********************************************************************/
{
  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return "";
  }

  memset(data->lerrorstr,0,CS_CHARBUF);

  if(data->error[data->lerror].source == 0){ // It is an CCP stack failure
    snprintf(data->lerrorstr,CS_CHARBUF,"CCP Failure: %s",data->error[data->lerror].name);
  } else { //It is an API failure
    snprintf(data->lerrorstr,CS_CHARBUF,"CCP API Failure: %s",data->error[data->lerror].name);
  }
  return data->lerrorstr;
}

/********************************************************************/
int cs_ccp_init(
		cs_ccp_t *data         /**< CCP stack structure*/ 
		)
/********************************************************************/
{
  int i;

  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }

  for(i = 0; i < CCP_CMD_BUF; i++){
    memset(data->cmd[i].name,0,CS_CHARBUF);
    strcpy(data->cmd[i].name,"Unknown command");
    data->cmd[i].timeout = 0;
    data->cmd[i].optional = 0;
    data->cmd[i].isoperational = 0;
    data->cmd[i].supported = 0;
  }

  cs_ccp_cmd_t c1 = {.name = "Connect",.timeout = 25,.optional = 0,.supported = 1};
  data->cmd[CCP_CMD_CONNECT] = c1;
  cs_ccp_cmd_t c2 = {.name = "Disconnect",.timeout = 25,.optional = 0,.supported = 1};
  data->cmd[CCP_CMD_DISCONNECT] = c2;
  cs_ccp_cmd_t c3 = {.name = "Test",.timeout = 25,.optional = 1,.supported = 1};
  data->cmd[CCP_CMD_TEST] = c3;
  cs_ccp_cmd_t c4 = {.name = "SetMTA",.timeout = 25,.optional = 0,.supported = 1};
  data->cmd[CCP_CMD_SETMTA] = c4;
  cs_ccp_cmd_t c5 = {.name = "Dnload",.timeout = 25,.optional = 0,.supported = 1};
  data->cmd[CCP_CMD_DNLOAD] = c5;
  cs_ccp_cmd_t c6 = {.name = "Upload",.timeout = 25,.optional = 0,.supported = 1};
  data->cmd[CCP_CMD_UPLOAD] = c6;
  cs_ccp_cmd_t c7 = {.name = "GetDAQSize",.timeout = 25,.optional = 0,.supported = 1};
  data->cmd[CCP_CMD_GETDAQSIZE] = c7;
  cs_ccp_cmd_t c8 = {.name = "SetDAQPtr",.timeout = 25,.optional = 0,.supported = 1};
  data->cmd[CCP_CMD_SETDAQPTR] = c8;
  cs_ccp_cmd_t c9 = {.name = "WriteDAQ",.timeout = 25,.optional = 0,.supported = 1};
  data->cmd[CCP_CMD_WRITEDAQ] = c9;
  cs_ccp_cmd_t c10 = {.name = "StartStop",.timeout = 25,.optional = 0,.supported = 1};
  data->cmd[CCP_CMD_STARTSTOP] = c10;
  cs_ccp_cmd_t c11 = {.name = "GetSessionStatus",.timeout = 25,.optional = 1,.supported = 1};
  data->cmd[CCP_CMD_GETSSTATUS] = c11;
  cs_ccp_cmd_t c12 = {.name = "StartStopAll",.timeout = 25,.optional = 1,.supported = 1};
  data->cmd[CCP_CMD_STARTSTOPALL] = c12;
  cs_ccp_cmd_t c13 = {.name = "CCPVersion",.timeout = 25,.optional = 0,.supported = 1};
  data->cmd[CCP_CMD_CCPVERSION] = c13;
  cs_ccp_cmd_t c14 = {.name = "ExchangeID",.timeout = 25,.optional = 0,.supported = 1};
  data->cmd[CCP_CMD_EXCHANGE_ID] = c14;
  cs_ccp_cmd_t c15 = {.name = "BuildChkSum",.timeout = 30000,.optional = 1,.supported = 1};
  data->cmd[CCP_CMD_BUILDCHKSUM] = c15;
 
  for(i = 0; i < CCP_ERR_BUF; i++){
    memset(data->error[i].name,0,CS_CHARBUF);
    strcpy(data->error[i].name,"Unknown error code");
    data->error[i].source = 1;
    data->error[i].category = CCP_RETURN_CATEGORY_C3;
  }
  cs_ccp_return_t e1 = {.source = 0,.category = CCP_RETURN_CATEGORY_NOERROR,.name = "No error"};
  data->error[CCP_ERROR_NOERROR] = e1;
  cs_ccp_return_t e2 = {.source = 0,.category = CCP_RETURN_CATEGORY_C0,.name = "DAQ processor overload"};
  data->error[CCP_RC_OVERLOAD] = e2;
  cs_ccp_return_t e3 = {.source = 0,.category = CCP_RETURN_CATEGORY_C1,.name = "Command processor busy"};
  data->error[CCP_RC_CMDBUSY] = e3;
  cs_ccp_return_t e4 = {.source = 0,.category = CCP_RETURN_CATEGORY_C1,.name = "DAQ processor busy"};
  data->error[CCP_RC_DAQBUSY] = e4;
  cs_ccp_return_t e5 = {.source = 0,.category = CCP_RETURN_CATEGORY_C1,.name = "Internal timeout"};
  data->error[CCP_RC_INTERNALTIMEOUT] = e5;
  cs_ccp_return_t e6 = {.source = 0,.category = CCP_RETURN_CATEGORY_C1,.name = "Key request"};
  data->error[CCP_RC_KEYREQUEST] = e6;
  cs_ccp_return_t e7 = {.source = 0,.category = CCP_RETURN_CATEGORY_C1,.name = "Session status request"};
  data->error[CCP_RC_SESSIONSTATUS] = e7;
  cs_ccp_return_t e8 = {.source = 0,.category = CCP_RETURN_CATEGORY_C2,.name = "Cold restart request"};
  data->error[CCP_RC_COLDRESTART] = e8;
  cs_ccp_return_t e9 = {.source = 0,.category = CCP_RETURN_CATEGORY_C2,.name = "Calibration data init request"};
  data->error[CCP_RC_CALINIT] = e9;
  cs_ccp_return_t e10 = {.source = 0,.category = CCP_RETURN_CATEGORY_C2,.name = "DAQ list init request"};
  data->error[CCP_RC_DAQINIT] = e10;
  cs_ccp_return_t e11 = {.source = 0,.category = CCP_RETURN_CATEGORY_C2,.name = "Code update request"};
  data->error[CCP_RC_CODEUPDATE] = e11;
  cs_ccp_return_t e12 = {.source = 0,.category = CCP_RETURN_CATEGORY_C3,.name = "Unknown cmd"};
  data->error[CCP_RC_CMDUNKNOWN] = e12;
  cs_ccp_return_t e13 = {.source = 0,.category = CCP_RETURN_CATEGORY_C3,.name = "Command syntax"};
  data->error[CCP_RC_CMDSYNTAX] = e13;
  cs_ccp_return_t e14 = {.source = 0,.category = CCP_RETURN_CATEGORY_C3,.name = "Parameters out of range"};
  data->error[CCP_RC_PARAMETER] = e14;
  cs_ccp_return_t e15 = {.source = 0,.category = CCP_RETURN_CATEGORY_C3,.name = "Access denied"};
  data->error[CCP_RC_ACCESSDENIED] = e15;
  cs_ccp_return_t e16 = {.source = 0,.category = CCP_RETURN_CATEGORY_C3,.name = "Overload"};
  data->error[CCP_RC_OVERLOAD] = e16;
  cs_ccp_return_t e17 = {.source = 0,.category = CCP_RETURN_CATEGORY_C3,.name = "Access blocked"};
  data->error[CCP_RC_ACCESSLOCKED] = e17;
  cs_ccp_return_t e18 = {.source = 0,.category = CCP_RETURN_CATEGORY_C3,.name = "Resource/Function is not available"};
  data->error[CCP_RC_NOTAVAILABLE] = e18;
  cs_ccp_return_t e19 = {.source = 1,.category = CCP_RETURN_CATEGORY_C3,.name = "CAN interface broken"};
  data->error[CCP_ERROR_CANIFACE] = e19;
  cs_ccp_return_t e20 = {.source = 1,.category = CCP_RETURN_CATEGORY_C3,.name = "CAN interface send failure"};
  data->error[CCP_ERROR_SEND] = e20;
  cs_ccp_return_t e21 = {.source = 1,.category = CCP_RETURN_CATEGORY_C3,.name = "CAN interface recv timeout"};
  data->error[CCP_ERROR_RECVTIMEOUT] = e21;
  cs_ccp_return_t e22 = {.source = 1,.category = CCP_RETURN_CATEGORY_C3,.name = "CAN interface recv failure"};
  data->error[CCP_ERROR_RECV] = e22;
  cs_ccp_return_t e23 = {.source = 1,.category = CCP_RETURN_CATEGORY_C3,.name = "Invalid CRM packet"};
  data->error[CCP_ERROR_RECVINVALID] = e23;
  cs_ccp_return_t e24 = {.source = 1,.category = CCP_RETURN_CATEGORY_C3,.name = "CMD is not supported by API"};
  data->error[CCP_ERROR_NOTSUPPORTEDAPI] = e24;
  cs_ccp_return_t e25 = {.source = 1,.category = CCP_RETURN_CATEGORY_C3,.name = "CMD is not supported by Slave"};
  data->error[CCP_ERROR_NOTSUPPORTEDSLAVE] = e25;
  cs_ccp_return_t e26 = {.source = 1,.category = CCP_RETURN_CATEGORY_C3,.name = "Invalid parameter"};
  data->error[CCP_ERROR_PARAMETER] = e26;
  cs_ccp_return_t e27 = {.source = 1,.category = CCP_RETURN_CATEGORY_C3,.name = "Malloc failure"};
  data->error[CCP_ERROR_MALLOC] = e27;
  cs_ccp_return_t e28 = {.source = 1,.category = CCP_RETURN_CATEGORY_C3,.name = "DAQ configuration failure"};
  data->error[CCP_ERROR_DAQCONFIG] = e28;
  cs_ccp_return_t e29 = {.source = 1,.category = CCP_RETURN_CATEGORY_C3,.name = "Cannot find A2L signal"};
  data->error[CCP_ERROR_A2LSIGNAL] = e29;
  cs_ccp_return_t e30 = {.source = 1,.category = CCP_RETURN_CATEGORY_C3,.name = "Incompatible CCP Version"};
  data->error[CCP_ERROR_CCPVERSION] = e30;
  

  memset(&data->daqlist,0,sizeof(cs_ccp_daq_list_t));
  _ccp_reset_daqlist(data);
  data->ignore_feedback = 0; //Set to default, wait for salve response
  data->ignore_counter = 0;
  return CS_OK;
}

/*************************************************************************************************/
int cs_ccp_can_open(cs_ccp_t *data, char *iface)
/*************************************************************************************************/
{
  uint32_t filter[1]; //+ Resp id
  if((iface == NULL) ||
     (data == NULL)){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }

  //Open raw socket
  data->socket = cs_socketcan_init();

  if(data->socket == NULL){
    _ccp_error("Failed init socket: %s\n",iface);
    data->lerror = CCP_ERROR_CANIFACE;
    return CS_FAIL;
  }

  if(cs_socketcan_config_native(data->socket, iface,0)){
    _ccp_error("Failed to configure socket: %s\n",iface);
    data->lerror = CCP_ERROR_CANIFACE;
    return CS_FAIL;
  }

  if(cs_socketcan_open(data->socket)){
    _ccp_error("Failed to open socket: %s\n",iface);
    data->lerror = CCP_ERROR_CANIFACE;
    return CS_FAIL;
  }

  //TODO: Support for DAQ specific CAN id
  //Build up new filter
  filter[0] = data->ccpa2l.can_recv_id;
  _ccp_verbose("Set filter to recv ID: 0x%X\n",data->ccpa2l.can_recv_id);

  if(cs_socketcan_set_filter(data->socket,filter,1)){
    _ccp_error("Failed to create filter\n");
    data->lerror = CCP_ERROR_CANIFACE;
    return CS_FAIL;
  }

  return CS_OK;
}

/*************************************************************************************************/
int cs_ccp_can_close(cs_ccp_t *data)
/*************************************************************************************************/
{
  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  cs_socketcan_close(data->socket);
  cs_socketcan_free(data->socket);
  return CS_OK;
}


/*************************************************************************************************/
void cs_ccp_set_ingnoreresponse(cs_ccp_t *data, uint8_t ignore)
/*************************************************************************************************/
{
  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return;
  }

  data->ignore_feedback = (ignore>0?1:0);
}

/********************************************************************/
int cs_ccp_close(
		 cs_ccp_t *data
		 )
/********************************************************************/
{
  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }

  _ccp_reset_daqlist(data);

  return CS_OK;
}


/********************************************************************/
void cs_ccp_print_status(
			 uint8_t status
			 )
/********************************************************************/
{
  if(status & CCP_SSTATUS_CALINIT)
    _ccp_info("Calibraition is initialized\n");
  if(status & CCP_SSTATUS_DAQINIT)
    _ccp_info("DAQ is initialized\n");
  if(status & CCP_SSTATUS_DAQSAVE)
    _ccp_info("DAQ conf is saved during shutdown\n");
  if(status & CCP_SSTATUS_CALSAVE)
    _ccp_info("CAL conf is saved during shutdown\n");
  if(status & CCP_SSTATUS_RUN)
    _ccp_info("Session in progress\n");
}

/********************************************************************/
void cs_ccp_print_resource(
			   cs_ccp_t *data
			   )
/********************************************************************/
{

  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return;
  }
    
  
  if(data->resource_availability_mask & CCP_RESOURCE_CALIBRATION)
    _ccp_info("Calibration is available\n");
  if(data->resource_availability_mask & CCP_RESOURCE_DAQ)
    _ccp_info("DAQ is available\n");
  if(data->resource_availability_mask & CCP_RESOURCE_PGM)
    _ccp_info("Programming is available\n");
  if(data->resource_protection_mask & CCP_RESOURCE_CALIBRATION)
    _ccp_info("Calibration is protected\n");
  if(data->resource_protection_mask & CCP_RESOURCE_DAQ)
    _ccp_info("DAQ is protected\n");
  if(data->resource_protection_mask & CCP_RESOURCE_PGM)
    _ccp_info("Programming is protected\n");
}

/**************************************************************************/
uint8_t cs_ccp_odt_getmaxactive(
				cs_ccp_t *data,
				uint8_t daq
				)
/**************************************************************************/
{
  int i;
  int maxid = -1;
  uint8_t rv_error = 0xFF;
  cs_ccp_daq_t *daqo = NULL;

  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return rv_error;
  }

  if(daq >= CCP_DAQMAX){
    _ccp_error("DAQ index overrange\n");
    data->lerror = CCP_ERROR_DAQCONFIG;
    return rv_error;
  }

  daqo = &data->daqlist.daq[data->daqlist.daq_active[daq]];

  for(i = 0; i < daqo->no_active; i++){
    if(maxid < (int)daqo->odt_active[i])
      maxid = (int)daqo->odt_active[i];
  }

  if(maxid == -1)
    return 0xFF;

  return (uint8_t)maxid;
}

/**************************************************************************/
int cs_ccp_daq_setperiod(			
			 cs_ccp_t *data,
			 uint8_t daq,
			 double period
						)
/**************************************************************************/
{
  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }

  if(daq >= CCP_DAQMAX){
    _ccp_error("Invalid DAQ id\n");
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL; 
  }

  if(period <= 0){
    _ccp_error("Invalid period\n");
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL; 
  }

  data->daqlist.daq[daq].period = period;

  return CS_OK;
}

/**************************************************************************/
int cs_ccp_daq_setprescaler(			
			    cs_ccp_t *data,
			    uint8_t daq,
			    uint16_t prescaler
						)
/**************************************************************************/
{
  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }

  if(daq >= CCP_DAQMAX){
    _ccp_error("Invalid DAQ id\n");
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL; 
  }

  if(prescaler < 1){
    _ccp_error("Invalid prescaler\n");
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL; 
  }

  data->daqlist.daq[daq].prescaler = prescaler;

  return CS_OK;
}

/**************************************************************************/
int cs_ccp_daq_istimeout(			
			 cs_ccp_t *data,
			 double fac_timeout
						)
/**************************************************************************/
{
  int i, j;
  int rv = CS_OK;
  double now;
  int did;
  struct timespec ts_now;
  int oid;
  
  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }

  //Get actual time
  cs_ts_gettime(&ts_now);
  now = cs_ts_timespec2d(ts_now);
  //Loop over all DAQ, to check ODTs for receive timeout
  for(i = 0; i < data->daqlist.no_active; i++){
    did = data->daqlist.daq_active[i];

    //Is DAQ running
    if(!data->daqlist.daq[did].running)
      continue;

    //Is period configured
    if(data->daqlist.daq[did].period <= 0.0)
      continue;

    for(j = 0; j < data->daqlist.daq[did].no_active; j++){
      oid = data->daqlist.daq[did].odt_active[i];
      //Do we have already received an ODT
      if(data->daqlist.daq[did].odt[oid].last <= 0.0)
	continue;
      
      if((data->daqlist.daq[did].odt[oid].last + 
	  (data->daqlist.daq[did].period * fac_timeout)) < data->daqlist.daq[did].ts_expect){
	//Check if timeout check is deactivated
	if(fac_timeout > 1.0){
	  //Detect timeout
	  _ccp_debug("DAQ timeout detected, DAQ %i, ODT %i, Period %f s Last recv %f s, Expected %f s\n",
		     i,i, data->daqlist.daq[did].period,
		     data->daqlist.daq[did].odt[oid].last,
		     data->daqlist.daq[did].ts_expect);
	  rv = CS_FAIL;
	}
      }
    }
    //Now calc new expected recv timestamps,
    //Notice we do not know when this function is called, so we have to loop
    if(data->daqlist.daq[did].period > 0.0){
      while(data->daqlist.daq[did].ts_expect <= now){
	data->daqlist.daq[did].ts_expect += data->daqlist.daq[i].period;
      }
    } else {
      //if period is zero, set expected time to actual time
      data->daqlist.daq[did].ts_expect = now;
    }
    
  }
  
  return rv;
}

/**************************************************************************/
int cs_ccp_entry_reg_value(			
			   cs_ccp_t *data,
			   uint8_t daq,
			   uint8_t odt,
			   uint8_t entry,
			   int size,
			   double *value,
			   cs_a2l_bo_t BO,
			   cs_a2l_compu_method_t cm,
			   cs_a2l_data_type_t dtype,
			   double *ts,
			   void (*extcallback)(double value, double ts, void *priv),
			   void *priv
						)
/**************************************************************************/
{

  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  if(daq >= CCP_DAQMAX){
    _ccp_error("DAQ index overrange\n");
    data->lerror = CCP_ERROR_DAQCONFIG;
    return CS_FAIL;
  }

  if(odt > data->daqlist.daq[daq].no_odt){
    _ccp_error("ODT index overrange\n");
    data->lerror = CCP_ERROR_DAQCONFIG;
    return CS_FAIL;
  }

  if(entry > CCP_ODTENTRYMAX){
    _ccp_error("ODT entry overrange\n");
    data->lerror = CCP_ERROR_DAQCONFIG;
    return CS_FAIL;
  }

  _ccp_verbose("Register DAQ :0x%X, ODT: 0x%X, Entry: 0x%X, Compu function: %s\n",daq,odt,entry,cm.name);
  data->daqlist.daq[daq].odt[odt].entry[entry].me.exvalue = value;
  data->daqlist.daq[daq].odt[odt].entry[entry].me.size = size;
  data->daqlist.daq[daq].odt[odt].entry[entry].me.BO = BO;
  data->daqlist.daq[daq].odt[odt].entry[entry].me.cm = cm;
  data->daqlist.daq[daq].odt[odt].entry[entry].me.dtype = dtype;
  data->daqlist.daq[daq].odt[odt].entry[entry].me.exts = ts;
  data->daqlist.daq[daq].odt[odt].entry[entry].me.extcallback = extcallback;
  data->daqlist.daq[daq].odt[odt].entry[entry].me.extcallbackpriv = priv;

  if(_ccp_daq_setactive(data,daq,1)){
    _ccp_error("Failed to activate DAQ\n");
    data->lerror = CCP_ERROR_DAQCONFIG;
    return CS_FAIL;
  }

  if(_ccp_odt_setactive(&data->daqlist.daq[daq],odt,1)){
    _ccp_error("Failed to activate ODT\n");
    data->lerror = CCP_ERROR_DAQCONFIG;
    return CS_FAIL;
  }
  
  if(data->daqlist.daq[daq].odt[odt].no_entry < CCP_ODTENTRYMAX)
    data->daqlist.daq[daq].odt[odt].no_entry++;


  return CS_OK;
}

/**************************************************************************/
int cs_ccp_entry_reg_signal(			
			    cs_ccp_t *data,
			    uint8_t daq,
			    uint8_t odt,
			    uint8_t entry,
			    char* signal,
			    double *value,
			    double *ts,
			    void (*extcallback)(double value, double ts, void *priv),
			    void *priv
						)
/**************************************************************************/
{
  cs_a2l_measurement_t me;

  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  //Try to get neasurement signal
  if(cs_a2l_GetMeasurement(&data->ccpa2l.a2l,&me,signal)){
    data->lerror = CCP_ERROR_DAQCONFIG;
    return CS_FAIL;
  }

  //Try to register
  return cs_ccp_entry_reg_value(data,daq,odt,entry,
				me.size,value,me.BO,me.cm,me.dtype,
				ts,extcallback,priv);
}

/**************************************************************************/
int cs_ccp_receive(
		   cs_ccp_t *data
		   )
/**************************************************************************/
{
  cs_ccp_pkg_t pkg_recv;
  int counter = 0;
    
  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }
    
  _ccp_reset_pkg(&pkg_recv);
  //Note only dummy receive pkg, DTO handling in low level routine
  while(!_ccp_receive(data,&pkg_recv,0,1)){
    //process data
    //_ccp_verbose("Looping: %i\n",counter);
    counter++;
  }
  if(data->lerror != CCP_ERROR_NOERROR){
    _ccp_error("Failure: %i, DL:%i\n",counter,pkg_recv.dl);
    return CS_FAIL;
  }else{
    return CS_OK;
  }
}

/**************************************************************************/
void cs_ccp_set_stim(
		     cs_ccp_t *data,
		     uint32_t stim
		     )
/**************************************************************************/
{
  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return;
  }
  
  data->stim = stim;
}

/**************************************************************************/
void cs_ccp_wait_stim(
		      cs_ccp_t *data
		      )
/**************************************************************************/
{
  uint32_t i;
  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return;
  }
  
  for(i = 0; i < data->stim; i++){
    //Wait 1 ms in a loop until stim is reached
    cs_ts_sleep(1);
    //Process async data
    cs_ccp_receive(data);
  }
}

/**************************************************************************/
int cs_ccp_cmd_connect(
		       cs_ccp_t *data
		       )
/**************************************************************************/
{
  cs_ccp_pkg_t pkg_send;
  cs_ccp_pkg_t pkg_recv;
  cs_cu_data16_t l16;
  
  
  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  _ccp_verbose("CCP API Call Connect\n");

  if(_ccp_check_cmd(data,CCP_CMD_CONNECT)){
    //return CS_FAIL;
  }
    
  _ccp_reset_error(data);
  _ccp_reset_pkg(&pkg_send);
  _ccp_reset_pkg(&pkg_recv);
  
  pkg_send.data[0] = CCP_CMD_CONNECT; //
  pkg_send.data[1] = data->wc++; //
  l16.u16 = cs_cu_swap16(A2L_BO_INTEL,data->ccpa2l.station_id);
  pkg_send.data[2] = l16.u8[0];
  pkg_send.data[3] = l16.u8[1];
    
  //Send Recv and retry for special failures
  if(_ccp_send_recv_action(data,CCP_CMD_CONNECT,&pkg_send,&pkg_recv,0)){
    return CS_FAIL;
  }
  
  if(pkg_recv.data[1] != CCP_ERROR_NOERROR){
    _ccp_verbose("CCP Cmd Connect: No Ack received\n");
    return CS_FAIL ;
  }
  
  return CS_OK;
}

/**************************************************************************/
int cs_ccp_cmd_getsessionstatus(
				cs_ccp_t *data,
				uint8_t *status
				)
/**************************************************************************/
{
  cs_ccp_pkg_t pkg_send;
  cs_ccp_pkg_t pkg_recv;
  
  
  if((data == NULL) ||
     (status == NULL)){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  _ccp_verbose("CCP API Call GetSessionStatus\n");

  if(_ccp_check_cmd(data,CCP_CMD_GETSSTATUS)){
    //return CS_FAIL;
  }
    
  _ccp_reset_error(data);
  _ccp_reset_pkg(&pkg_send);
  _ccp_reset_pkg(&pkg_recv);
  
  pkg_send.data[0] = CCP_CMD_GETSSTATUS; //
  pkg_send.data[1] = data->wc++; //
    
  //Send Recv and retry for special failures
  if(_ccp_send_recv_action(data,CCP_CMD_GETSSTATUS,&pkg_send,&pkg_recv,0)){
    return CS_FAIL;
  }
  
  if(pkg_recv.data[1] != CCP_ERROR_NOERROR){
    _ccp_verbose("No Ack received\n");
    return CS_FAIL ;
  }
    
  *status = pkg_recv.data[3];

  return CS_OK;
}

/**************************************************************************/
int cs_ccp_cmd_disconnect(
			  cs_ccp_t *data,
			  uint8_t endofsession
			  )
/**************************************************************************/
{
  cs_ccp_pkg_t pkg_send;
  cs_ccp_pkg_t pkg_recv;
  cs_cu_data16_t l16;
  
  
  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }


  if(!((endofsession == 0)||(endofsession == 1))){
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL;
  }
  
  _ccp_verbose("CCP API Call Disconnect\n");

  if(_ccp_check_cmd(data,CCP_CMD_DISCONNECT)){
    //return CS_FAIL;
  }
    
  _ccp_reset_error(data);
  _ccp_reset_pkg(&pkg_send);
  _ccp_reset_pkg(&pkg_recv);
  
  pkg_send.data[0] = CCP_CMD_DISCONNECT; //
  pkg_send.data[1] = data->wc++; //
  pkg_send.data[2] = endofsession; //
  pkg_send.data[3] = 0; // not defined
  l16.u16 = cs_cu_swap16(A2L_BO_INTEL,data->ccpa2l.station_id);
  pkg_send.data[4] = l16.u8[0];
  pkg_send.data[5] = l16.u8[1];
    
  //Send Recv and retry for special failures
  if(_ccp_send_recv_action(data,CCP_CMD_DISCONNECT,&pkg_send,&pkg_recv,data->ignore_feedback)){
    return CS_FAIL;
  }
  
  if(!data->ignore_feedback){
    if(pkg_recv.data[1] != CCP_ERROR_NOERROR){
      _ccp_verbose("No Ack received\n");
      return CS_FAIL ;
    }
  }
  
  return CS_OK;
}

/**************************************************************************/
int cs_ccp_cmd_ccpversion(
			  cs_ccp_t *data,
			  uint8_t *masterversion,
			  uint8_t *releaseversion
			  )
/**************************************************************************/
{
  cs_ccp_pkg_t pkg_send;
  cs_ccp_pkg_t pkg_recv;
  
  
  if((data == NULL)||(masterversion == NULL)||(releaseversion == NULL)){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  _ccp_verbose("CCP API Call CCPVersion\n");

  if(_ccp_check_cmd(data,CCP_CMD_CCPVERSION)){
    //return CS_FAIL;
  }
    
  _ccp_reset_error(data);
  _ccp_reset_pkg(&pkg_send);
  _ccp_reset_pkg(&pkg_recv);
  
  pkg_send.data[0] = CCP_CMD_CCPVERSION; //
  pkg_send.data[1] = data->wc++; //
  pkg_send.data[2] = CS_CCP_VERSION_MAJOR; //
  pkg_send.data[3] = CS_CCP_VERSION_MINOR;
    
  //Send Recv and retry for special failures
  if(_ccp_send_recv_action(data,CCP_CMD_CCPVERSION,&pkg_send,&pkg_recv,0)){
    return CS_FAIL;
  }
  
  if(pkg_recv.data[1] != CCP_ERROR_NOERROR){
    _ccp_verbose("No Ack received\n");
    return CS_FAIL ;
  }
  
  *masterversion = pkg_recv.data[3];
  *releaseversion = pkg_recv.data[4];

  return CS_OK;
}

/**************************************************************************/
int cs_ccp_cmd_exchangeid(
			  cs_ccp_t *data,
			  uint8_t *length,
			  uint8_t *qualifier,
			  uint8_t *resource_availability_mask,
			  uint8_t *resource_protection_mask
			  )
/**************************************************************************/
{
  cs_ccp_pkg_t pkg_send;
  cs_ccp_pkg_t pkg_recv;
  
  
  if((data == NULL)||
     (length == NULL)){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  _ccp_verbose("CCP API Call Exchange ID\n");

  if(_ccp_check_cmd(data,CCP_CMD_EXCHANGE_ID)){
    //return CS_FAIL;
  }
    
  _ccp_reset_error(data);
  _ccp_reset_pkg(&pkg_send);
  _ccp_reset_pkg(&pkg_recv);
  
  pkg_send.data[0] = CCP_CMD_EXCHANGE_ID; //
  pkg_send.data[1] = data->wc++; //
    
  //Send Recv and retry for special failures
  if(_ccp_send_recv_action(data,CCP_CMD_EXCHANGE_ID,&pkg_send,&pkg_recv,0)){
    return CS_FAIL;
  }
  
  if(pkg_recv.data[1] != CCP_ERROR_NOERROR){
    _ccp_verbose("No Ack received\n");
    return CS_FAIL ;
  }
  
  *length = pkg_recv.data[3];
  if(qualifier!= NULL)
    *qualifier = pkg_recv.data[4];
  if(resource_availability_mask != NULL)
    *resource_availability_mask = pkg_recv.data[5];
  if(resource_protection_mask != NULL)
    *resource_protection_mask = pkg_recv.data[6];

  //Store in data structure
  data->resource_availability_mask = pkg_recv.data[5];
  data->resource_protection_mask = pkg_recv.data[6];
    
  return CS_OK;
}

/**************************************************************************/
int cs_ccp_cmd_setmta(
		      cs_ccp_t *data,
		      uint8_t mtanumber,
		      uint32_t mta,
		      uint8_t adressextension
		      )
/**************************************************************************/
{
  cs_ccp_pkg_t pkg_send;
  cs_ccp_pkg_t pkg_recv;
  cs_cu_data32_t l32;
    
  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }


  if(!((mtanumber == 0)||(mtanumber == 1))){
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL;
  }
  
  _ccp_verbose("CCP API Call set MTA\n");

  if(_ccp_check_cmd(data,CCP_CMD_SETMTA)){
    //return CS_FAIL;
  }
    
  _ccp_reset_error(data);
  _ccp_reset_pkg(&pkg_send);
  _ccp_reset_pkg(&pkg_recv);
  
  pkg_send.data[0] = CCP_CMD_SETMTA; //
  pkg_send.data[1] = data->wc++; //
  pkg_send.data[2] = mtanumber; //
  pkg_send.data[3] = adressextension; //
  l32.u32 = cs_cu_swap32(data->ccpa2l.BO,mta);
  pkg_send.data[4] = l32.u8[0];
  pkg_send.data[5] = l32.u8[1];
  pkg_send.data[6] = l32.u8[2];
  pkg_send.data[7] = l32.u8[3];
    
  //Send Recv and retry for special failures
  if(_ccp_send_recv_action(data,CCP_CMD_SETMTA,&pkg_send,&pkg_recv,data->ignore_feedback)){
    return CS_FAIL;
  }

  if(!data->ignore_feedback){
    if(pkg_recv.data[1] != CCP_ERROR_NOERROR){
      _ccp_verbose("No Ack received\n");
      return CS_FAIL ;
    }
  }

  data->mta.address = mta;
  data->mta.extension = adressextension;
    
  return CS_OK;
}

/**************************************************************************/
int cs_ccp_cmd_dnload(
		      cs_ccp_t *data,
		      uint8_t *tdata,
		      int size
		      )
/**************************************************************************/
{
  cs_ccp_pkg_t pkg_send;
  cs_ccp_pkg_t pkg_recv;
  
  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }
    
  if((tdata == NULL)||(size <= 0)||(size > 5)){
    _ccp_error("Parameter failure\n");
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL;
  }
  
  _ccp_verbose("CCP API Call Dnload\n");

  if(_ccp_check_cmd(data,CCP_CMD_DNLOAD)){
    //return CS_FAIL;
  }
    
  _ccp_reset_error(data);
  _ccp_reset_pkg(&pkg_send);
  _ccp_reset_pkg(&pkg_recv);
  
  pkg_send.data[0] = CCP_CMD_DNLOAD; //
  pkg_send.data[1] = data->wc++; //
  pkg_send.data[2] = (uint8_t)size; //
  memcpy(&pkg_send.data[3],tdata,size);
    
  //Send Recv and retry for special failures
  if(_ccp_send_recv_action(data,CCP_CMD_DNLOAD,&pkg_send,&pkg_recv,data->ignore_feedback)){
    return CS_FAIL;
  }

  if(!data->ignore_feedback){
    if(pkg_recv.data[1] != CCP_ERROR_NOERROR){
      _ccp_verbose("No Ack received\n");
      return CS_FAIL ;
    }
  }
    
  return CS_OK;
}

/**************************************************************************/
int cs_ccp_cmd_upload(
		      cs_ccp_t *data,
		      uint8_t *tdata,
		      int size
		      )
/**************************************************************************/
{
  cs_ccp_pkg_t pkg_send;
  cs_ccp_pkg_t pkg_recv;
  
  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }


  if((tdata == NULL)||(size <= 0)||(size > 5)){
    _ccp_error("Parameter failure\n");
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL;
  }
  
  _ccp_verbose("CCP API Call Upload\n");

  if(_ccp_check_cmd(data,CCP_CMD_UPLOAD)){
    //return CS_FAIL;
  }
    
  _ccp_reset_error(data);
  _ccp_reset_pkg(&pkg_send);
  _ccp_reset_pkg(&pkg_recv);
  
  pkg_send.data[0] = CCP_CMD_UPLOAD; //
  pkg_send.data[1] = data->wc++; //
  pkg_send.data[2] = (uint8_t)size; //
    
  //Send Recv and retry for special failures
  if(_ccp_send_recv_action(data,CCP_CMD_UPLOAD,&pkg_send,&pkg_recv,0)){
    return CS_FAIL;
  }

  if(pkg_recv.data[1] != CCP_ERROR_NOERROR){
    _ccp_verbose("No Ack received\n");
    return CS_FAIL ;
  }
    
  memcpy(tdata,&pkg_recv.data[3],size);

  return CS_OK;
}

/**************************************************************************/
int cs_ccp_cmd_getdaqsize(
			  cs_ccp_t *data,
			  uint8_t daq,
			  uint32_t dto_id,
			  uint8_t *daq_size,
			  uint8_t *pid_first
			  )
/**************************************************************************/
{
  cs_ccp_pkg_t pkg_send;
  cs_ccp_pkg_t pkg_recv;
  cs_cu_data32_t l32;
  int i;

  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }


  if((daq_size == NULL)||(pid_first == NULL)){
    _ccp_error("Parameter failure\n");
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL;
  }

  if(daq >= CCP_DAQMAX){
    _ccp_error("DAQ id out of Range\n");
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL;
  }
  
  _ccp_verbose("CCP API Call GetDAQSize\n");

  if(_ccp_check_cmd(data,CCP_CMD_GETDAQSIZE)){
    //return CS_FAIL;
  }
    
  _ccp_reset_error(data);
  _ccp_reset_pkg(&pkg_send);
  _ccp_reset_pkg(&pkg_recv);
  
  pkg_send.data[0] = CCP_CMD_GETDAQSIZE; //
  pkg_send.data[1] = data->wc++; //
  pkg_send.data[2] = daq; //
  pkg_send.data[3] = 0; // do not care
  l32.u32 = cs_cu_swap32(data->ccpa2l.BO,dto_id);
  pkg_send.data[4] = l32.u8[0];
  pkg_send.data[5] = l32.u8[1];
  pkg_send.data[6] = l32.u8[2];
  pkg_send.data[7] = l32.u8[3];    
  //Send Recv and retry for special failures
  if(_ccp_send_recv_action(data,CCP_CMD_GETDAQSIZE,&pkg_send,&pkg_recv,0)){
    return CS_FAIL;
  }

  if(pkg_recv.data[1] != CCP_ERROR_NOERROR){
    _ccp_verbose("No Ack received\n");
    return CS_FAIL ;
  }
    
  *daq_size = pkg_recv.data[3];
  *pid_first = pkg_recv.data[4];

 
  _ccp_reset_daq(&data->daqlist.daq[daq]);
     
  data->daqlist.daq[daq].no_odt = *daq_size;
  data->daqlist.daq[daq].odt = (cs_ccp_odt_t*)malloc(sizeof(cs_ccp_odt_t)**daq_size);
  if(data->daqlist.daq[daq].odt == NULL){
    _ccp_error("Malloc failure\n");
    data->lerror = CCP_ERROR_MALLOC;
    return CS_FAIL;
  }
  data->daqlist.daq[daq].odt_active = (uint8_t*)malloc(sizeof(uint8_t)**daq_size);
  if(data->daqlist.daq[daq].odt_active == NULL){
    _ccp_error("Malloc failure\n");
    data->lerror = CCP_ERROR_MALLOC;
    return CS_FAIL;
  }
  memset(data->daqlist.daq[daq].odt_active,0xFF,sizeof(uint8_t)**daq_size);
  data->daqlist.daq[daq].first_pid = *pid_first;

  //Reset odts
  for(i = 0; i < *daq_size; i++){
    _ccp_reset_odt(&data->daqlist.daq[daq].odt[i]);
  }

  //Remove DAQ from active list
  if(_ccp_daq_setactive(data,daq,0)){
    _ccp_error("Failed to set DAQ inactive\n");
    data->lerror = CCP_ERROR_DAQCONFIG;
    return CS_FAIL;
  }

  return CS_OK;
}

/**************************************************************************/
int cs_ccp_cmd_setdaqptr(
			 cs_ccp_t *data,
			 uint8_t daq,
			 uint8_t odt,
			 uint8_t element
			 )
/**************************************************************************/
{
  cs_ccp_pkg_t pkg_send;
  cs_ccp_pkg_t pkg_recv;
  cs_ccp_daq_t *daqdata = NULL;
  
  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }

  
  _ccp_verbose("CCP API Call SetDAQPtr\n");

  if(daq >= CCP_DAQMAX){
    _ccp_error("Invalid DAQ id");
    data->lerror = CCP_ERROR_DAQCONFIG;
    return CS_FAIL;
  }

  daqdata = &data->daqlist.daq[daq];

  if(odt >= daqdata->no_odt){
    _ccp_error("Invalid ODT id");
    data->lerror = CCP_ERROR_DAQCONFIG;
    return CS_FAIL;
  }

  if(element >= CCP_ODTENTRYMAX){
    _ccp_error("Invalid Element id");
    data->lerror = CCP_ERROR_DAQCONFIG;
    return CS_FAIL;
  }
    
  if(_ccp_check_cmd(data,CCP_CMD_SETDAQPTR)){
    //return CS_FAIL;
  }
    
  _ccp_reset_error(data);
  _ccp_reset_pkg(&pkg_send);
  _ccp_reset_pkg(&pkg_recv);
  
  pkg_send.data[0] = CCP_CMD_SETDAQPTR; //
  pkg_send.data[1] = data->wc++; //
  pkg_send.data[2] = daq; //
  pkg_send.data[3] = odt; //
  pkg_send.data[4] = element;

  //Send Recv and retry for special failures
  if(_ccp_send_recv_action(data,CCP_CMD_SETDAQPTR,&pkg_send,&pkg_recv,data->ignore_feedback)){
    return CS_FAIL;
  }

  if(!data->ignore_feedback){
    if(pkg_recv.data[1] != CCP_ERROR_NOERROR){
      _ccp_verbose("No Ack received\n");
      return CS_FAIL ;
    }
  }
    
  return CS_OK;
}

/**************************************************************************/
int cs_ccp_cmd_writedaq(
			cs_ccp_t *data,
			int size,
			uint8_t addr_extension,
			uint32_t addr
			)
/**************************************************************************/
{
  cs_ccp_pkg_t pkg_send;
  cs_ccp_pkg_t pkg_recv;
  cs_cu_data32_t l32;
  
  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }

  if(!((size == 1)||(size == 2)||(size==4))){
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL;
  }
  
  _ccp_verbose("CCP API Call WriteDAQ\n");

  if(_ccp_check_cmd(data,CCP_CMD_WRITEDAQ)){
    //return CS_FAIL;
  }
    
  _ccp_reset_error(data);
  _ccp_reset_pkg(&pkg_send);
  _ccp_reset_pkg(&pkg_recv);
  
  pkg_send.data[0] = CCP_CMD_WRITEDAQ; //
  pkg_send.data[1] = data->wc++; //
  pkg_send.data[2] = (uint8_t)size; //
  pkg_send.data[3] = addr_extension; //
  l32.u32 = cs_cu_swap32(data->ccpa2l.BO,addr);
  pkg_send.data[4] = l32.u8[0];
  pkg_send.data[5] = l32.u8[1];
  pkg_send.data[6] = l32.u8[2];
  pkg_send.data[7] = l32.u8[3];    

  //Send Recv and retry for special failures
  if(_ccp_send_recv_action(data,CCP_CMD_WRITEDAQ,&pkg_send,&pkg_recv,data->ignore_feedback)){
    return CS_FAIL;
  }

  if(!data->ignore_feedback){
    if(pkg_recv.data[1] != CCP_ERROR_NOERROR){
      _ccp_verbose("No Ack received\n");
      return CS_FAIL ;
    }
  }
    
  return CS_OK;
}

/**************************************************************************/
int cs_ccp_cmd_startstop(
			 cs_ccp_t *data,
			 uint8_t mode,
			 uint8_t daq,
			 uint8_t odt_last,
			 uint8_t event,
			 uint16_t prescaler
			 )
/**************************************************************************/
{
  cs_ccp_pkg_t pkg_send;
  cs_ccp_pkg_t pkg_recv;
  cs_cu_data16_t l16;

  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }

  if(!((mode == 0)||(mode == 1)||(mode==2)||(prescaler>0))){
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL;
  }
  
    
  switch(mode){
  case 0:
    _ccp_verbose("CCP API Call StartStop Mode Stop\n");
    break;
  case 1:
    _ccp_verbose("CCP API Call StartStop Mode Start\n");
    break;
  case 2:
    _ccp_verbose("CCP API Call StartStop Mode Select\n");
    break;
  }


  if(daq > CCP_DAQMAX){
    _ccp_error("Invalid DAQ id\n");
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL;
  }

  if(_ccp_check_cmd(data,CCP_CMD_STARTSTOP)){
    //return CS_FAIL;
  }


  if(data->daqlist.no_active == 0){
    _ccp_error("No DAQ configured\n");
    data->lerror = CCP_ERROR_DAQCONFIG;
    return CS_FAIL;
  }

  if(_ccp_daq_isactive(data,daq)){
    _ccp_error("This DAQ %i is not configured\n",daq);
    data->lerror = CCP_ERROR_DAQCONFIG;
    return CS_FAIL;
  }
    
  _ccp_reset_error(data);
  _ccp_reset_pkg(&pkg_send);
  _ccp_reset_pkg(&pkg_recv);
  
  pkg_send.data[0] = CCP_CMD_STARTSTOP; //
  pkg_send.data[1] = data->wc++; //
  pkg_send.data[2] = mode; //
  pkg_send.data[3] = daq; //
  pkg_send.data[4] = odt_last; //
  pkg_send.data[5] = event; //
  l16.u16 = cs_cu_swap16(data->ccpa2l.BO,prescaler);
  pkg_send.data[6] = l16.u8[0];
  pkg_send.data[7] = l16.u8[1];

  //Send Recv and retry for special failures
  if(_ccp_send_recv_action(data,CCP_CMD_STARTSTOP,&pkg_send,&pkg_recv,data->ignore_feedback)){
    return CS_FAIL;
  }

  if(!data->ignore_feedback){
    if(pkg_recv.data[1] != CCP_ERROR_NOERROR){
      _ccp_verbose("No Ack received\n");
      return CS_FAIL ;
    }
  }

  switch(mode){
  case 0: //Stop
    _ccp_daq_stop(data,daq);
    break;
  case 1: //Start
    _ccp_daq_start(data,daq);
    break;
  case 2:
    _ccp_daq_select(data,daq);
    break;
  }
      
  return CS_OK;
}


/**************************************************************************/
int cs_ccp_cmd_startstopall(
			    cs_ccp_t *data,
			    uint8_t mode
			    )
/**************************************************************************/
{
  cs_ccp_pkg_t pkg_send;
  cs_ccp_pkg_t pkg_recv;

  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }

  if(!((mode == 0)||(mode == 1))){
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL;
  }
    
  _ccp_verbose("CCP API Call StartStopAll\n");

  if(_ccp_check_cmd(data,CCP_CMD_STARTSTOPALL)){
    //return CS_FAIL;
  }


  if(data->daqlist.no_active == 0){
    _ccp_error("No DAQ configured\n");
    data->lerror = CCP_ERROR_DAQCONFIG;
    return CS_FAIL;
  }
    
  _ccp_reset_error(data);
  _ccp_reset_pkg(&pkg_send);
  _ccp_reset_pkg(&pkg_recv);
  
  pkg_send.data[0] = CCP_CMD_STARTSTOPALL; //
  pkg_send.data[1] = data->wc++; //
  pkg_send.data[2] = mode; //

  //Send Recv and retry for special failures
  if(_ccp_send_recv_action(data,CCP_CMD_STARTSTOPALL,&pkg_send,&pkg_recv,data->ignore_feedback)){
    return CS_FAIL;
  }

  if(!data->ignore_feedback){
    if(pkg_recv.data[1] != CCP_ERROR_NOERROR){
      _ccp_verbose("No Ack received\n");
      return CS_FAIL ;
    }
  }

  switch(mode){
  case 0: //Stop
    _ccp_daq_stop_selected(data);
  case 1: //Start
    _ccp_daq_start_selected(data);
  }
      
  return CS_OK;
}

/**************************************************************************/
int cs_ccp_cmd_test(
		    cs_ccp_t *data
		    )
/**************************************************************************/
{
  cs_ccp_pkg_t pkg_send;
  cs_ccp_pkg_t pkg_recv;
  cs_cu_data16_t l16;
  
  
  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  _ccp_verbose("CCP API Call Test\n");
  
  if(_ccp_check_cmd(data,CCP_CMD_TEST)){
    //return CS_FAIL;
  }
  
  _ccp_reset_error(data);
  _ccp_reset_pkg(&pkg_send);
  _ccp_reset_pkg(&pkg_recv);
  
  pkg_send.dl = 4;
  pkg_send.data[0] = CCP_CMD_TEST; //
  pkg_send.data[1] = data->wc++; //
  l16.u16 = cs_cu_swap16(A2L_BO_INTEL,data->ccpa2l.station_id);
  pkg_send.data[2] = l16.u8[0];
  pkg_send.data[3] = l16.u8[1];
  
  //Send Recv and retry for special failures
  if(_ccp_send_recv_action(data,CCP_CMD_TEST,&pkg_send,&pkg_recv,data->ignore_feedback)){
    return CS_FAIL;
  }
  
  if(data->ignore_feedback){
    if(pkg_recv.data[1] != CCP_ERROR_NOERROR){
      _ccp_verbose("No Ack received\n");
      return CS_FAIL ;
    }
  }
  
  return CS_OK;
}

/**************************************************************************/
int cs_ccp_cmd_buildchksum(
			   cs_ccp_t *data,
			   int blocksize,
			   uint32_t *checksum
			   )
/**************************************************************************/
{
  cs_ccp_pkg_t pkg_send;
  cs_ccp_pkg_t pkg_recv;
  cs_cu_data32_t l32;
  cs_cu_data16_t l16;
  int chksumsize;

  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }

  if((blocksize <= 0)||
     (checksum == NULL)){
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL;
  }
  
  _ccp_verbose("CCP API Call BuildCheckSum\n");
  
  if(_ccp_check_cmd(data,CCP_CMD_BUILDCHKSUM)){
    //return CS_FAIL;
  }
  
  _ccp_reset_error(data);
  _ccp_reset_pkg(&pkg_send);
  _ccp_reset_pkg(&pkg_recv);
  
  pkg_send.dl = 6;
  pkg_send.data[0] = CCP_CMD_BUILDCHKSUM; //
  pkg_send.data[1] = data->wc++; //
  l32.u32 = cs_cu_swap32(data->ccpa2l.BO,(uint32_t)blocksize);
  pkg_send.data[2] = l32.u8[0];
  pkg_send.data[3] = l32.u8[1];
  pkg_send.data[4] = l32.u8[2];
  pkg_send.data[5] = l32.u8[3];
  
  //Send Recv and retry for special failures
  if(_ccp_send_recv_action(data,CCP_CMD_BUILDCHKSUM,&pkg_send,&pkg_recv,0)){
    return CS_FAIL;
  }
  
  if(pkg_recv.data[1] != CCP_ERROR_NOERROR){
    _ccp_verbose("No Ack received\n");
    return CS_FAIL ;
  }
  chksumsize = (int)pkg_recv.data[3];

  if(chksumsize > 4){
    _ccp_debug("Invalid Checksumsize");
    chksumsize = 4;
  }
  switch(chksumsize){
  case 4:
    memset(l32.u8,0,sizeof(uint32_t));
    memcpy(l32.u8,&pkg_recv.data[4],chksumsize);
    *checksum = cs_cu_swap32(data->ccpa2l.BO,l32.u32);
    break;
  case 3:
    memset(l32.u8,0,sizeof(uint32_t));
    memcpy(&l32.u8[1],&pkg_recv.data[4],chksumsize);
    *checksum = cs_cu_swap32(data->ccpa2l.BO,l32.u32);
    break;
  case 2:
    memset(l16.u8,0,sizeof(uint16_t));
    memcpy(l16.u8,&pkg_recv.data[4],chksumsize);
    *checksum = (uint32_t)cs_cu_swap16(data->ccpa2l.BO,l16.u16);
    break;
  case 1:
    *checksum = (uint32_t)pkg_recv.data[4];
    break;
  default:
    _ccp_debug("Invalid Checksumsize");
    *checksum = 0;
  }  

  return CS_OK;
}


