/*  CCP CAN Stack for SocketCAN
    Copyright (C) 2015 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/mman.h>
#include <unistd.h>

#include "cs_ccp_a2l.h"
#include "cs_defines.h"
#include "cs_calibration_utils.h"
#include "cs_ccp_private.h"


/**************************************************************************/
void cs_ccp_a2l_init(
		     cs_ccp_a2l_t *data  /**< Working storage structure*/
		     )
/**************************************************************************/
{
  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return;
  }

  data->ccp_version = 0;
  data->blob_version = 0;
  data->can_send_id = 0;
  data->can_recv_id = 0;
  data->station_id = 0;
  data->BO = A2L_BO_UNKNOWN;
  data->baudrate = 0;
  data->daqmode = CCP_DAQ_MODE_UNKNOWN;
  data->consistency = CCP_CONSISTENCY_UNKNOWN;
  data->addrextension = CCP_ADDREXTENSION_UNKNOWN;
  memset(data->cmd_optional,0,CS_CHARBUF);
  if((data->no_events > 0) && (data->events != NULL)){
    free(data->events);
    data->no_events = 0;
  }
  if((data->no_source > 0) && (data->source != NULL)){
    free(data->source);
    data->no_source = 0;
  }
  cs_a2l_init(&data->a2l);

}

/**************************************************************************/
int cs_ccp_a2l_GetOptionalCmds(
			       cs_parser_result_t *result,
			       int *value
			       )
/**************************************************************************/
{
  pcre *re = NULL;
  int rv = 0;
  int noresults = 3;
  int ovector[CS_REGEX_RESULTSIZE*noresults];
  char regex[CS_REGEX_BUFFER];

  if((result == NULL) || (value == NULL)){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }

  //reset value
  *value = 0;

  //cs_parser_printresult(result);

  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER);
  snprintf(regex,CS_REGEX_BUFFER,"(?'all'OPTIONAL_CMD (?'value'0x[0-9A-Fa-f]+))");

  //Precompile regex
  if((re = cs_parser_precompile((const char*)regex)) == NULL){
    _ccp_error("Regex precompilation failed\n");
    return CS_FAIL;
  }
  
  rv = pcre_exec(re, NULL, (const char*)result->result,result->resultlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE*noresults);

  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      //_ccp_verbose("No match\n"); 
      pcre_free(re);
      return CS_FAIL;
      break;
    default: 
      _ccp_error("Matching error %d\n", rv);
      pcre_free(re);
      return CS_FAIL;
      break;
    }
  } else {
    if(cs_parser_getsubstring_int(re,result->result,ovector,CS_REGEX_RESULTSIZE*noresults,"value",value,0)){
      _ccp_error("Failed to get value \n");
      pcre_free(re);
      return CS_FAIL;
    }    
  }
  //Modify search area for next search
  cs_parser_stepnext(result,ovector,CS_REGEX_RESULTSIZE*noresults,1);

  pcre_free(re);   
  return CS_OK;
}


/**************************************************************************/
int cs_ccp_a2l_GetTransportData(
				cs_ccp_a2l_t *data
				)
/**************************************************************************/
{
  cs_parser_result_t lresult;
  cs_parser_result_t lopt;
  pcre *re = NULL;
  char regex[CS_REGEX_BUFFER];
  int rv = 0;
  int noresult = 20;
  int ovector[CS_REGEX_RESULTSIZE*noresult];
  int value;
  uint8_t rvalue;
  char namedvalue[CS_CHARBUF];
  int counter = 0;

  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  lresult.searchin = data->a2l.iface.result;
  lresult.searchlen = data->a2l.iface.resultlen;
  if(cs_a2l_search_section(&lresult,"TP_BLOB",NULL)){
    _ccp_error("Regex failure\n");
    return CS_FAIL;
  }

  if(lresult.result == NULL){
    _ccp_error("Cannot find Section TP_BLOB\n");
    return CS_FAIL;
  }

  //cs_parser_printresult(&lresult);

  //define serachregion
  lresult.searchin = lresult.result;
  lresult.searchlen = lresult.resultlen;

  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER);
  snprintf(regex,CS_REGEX_BUFFER,"TP_BLOB[\\s]+(?'ccpversion'0x[0-9A-Fa-f]+)[\\s]+(?'blobversion'0x[0-9A-Fa-f]+)[\\s]+(?'sendid'0x[0-9A-Fa-f]+)[\\s]+(?'recvid'0x[0-9A-Fa-f]+)[\\s]+(?'station'0x[0-9A-Fa-f]+)[\\s]+(?'bo'[0-9]+)");
  //_ccp_verbose("Regex: %s\n",regex);
  //Precompile regex
  if((re = cs_parser_precompile((const char *)regex)) == NULL){
    _ccp_error("Regex precompilation failed\n");
    return CS_FAIL;
  }
  //Get data, only one fit is allowed
  rv = pcre_exec(re, NULL, (const char*)lresult.result,lresult.resultlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE * noresult);
  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _ccp_error("No content match %s\n");
      pcre_free(re);
      return CS_FAIL;
      break;
    default: 
      _ccp_error("Matching error %d\n", rv);
      pcre_free(re);
      return CS_FAIL;
      break;
    }
  } else {
    
    if(cs_parser_getsubstring_int(re,lresult.result,ovector,CS_REGEX_RESULTSIZE*noresult,"ccpversion",&value,0)){
      _ccp_error("Failed to get substring ccpversion\n");
      pcre_free(re);
      return CS_FAIL;
    }
    data->ccp_version = (uint16_t)value;

    if(cs_parser_getsubstring_int(re,lresult.result,ovector,CS_REGEX_RESULTSIZE*noresult,"blobversion",&value,0)){
      _ccp_error("Failed to get substring blobversion\n");
      pcre_free(re);
      return CS_FAIL;
    }
    data->blob_version = (uint16_t)value;

    if(cs_parser_getsubstring_int(re,lresult.result,ovector,CS_REGEX_RESULTSIZE*noresult,"sendid",&value,0)){
      _ccp_error("Failed to get substring sendid\n");
      pcre_free(re);
      return CS_FAIL;
    }
    data->can_send_id = (uint32_t)value;

    if(cs_parser_getsubstring_int(re,lresult.result,ovector,CS_REGEX_RESULTSIZE*noresult,"recvid",&value,0)){
      _ccp_error("Failed to get substring recvid\n");
      pcre_free(re);
      return CS_FAIL;
    }
    data->can_recv_id = (uint32_t)value;

    if(cs_parser_getsubstring_int(re,lresult.result,ovector,CS_REGEX_RESULTSIZE*noresult,"station",&value,0)){
      _ccp_error("Failed to get substring station\n");
      pcre_free(re);
      return CS_FAIL;
    }
    data->station_id = (uint16_t)value;

    if(cs_parser_getsubstring_int(re,lresult.result,ovector,CS_REGEX_RESULTSIZE*noresult,"bo",&value,0)){
      _ccp_error("Failed to get substring station\n");
      pcre_free(re);
      return CS_FAIL;
    }
    data->BO = (cs_a2l_bo_t)value;

    lopt = lresult;
    //Jump to end of last element
    if(cs_parser_stepnext(&lopt,ovector,CS_REGEX_RESULTSIZE * noresult,6)){
      _ccp_error("Failed to step to next working point\n");
      pcre_free(re);
      return CS_FAIL;
    }

  }
  pcre_free(re);

  data->baudrate = cs_a2l_getuintvalue(lresult,"BAUDRATE",&rvalue);
  if(rvalue){
    data->baudrate = 0; 
  }

  if(cs_a2l_getstringvalue(lresult,"DAQ_MODE",namedvalue,CS_CHARBUF)){
    _ccp_error("Failed to parse DAQ_MODE\n");
    pcre_free(re);
    return CS_FAIL;
  }
  if(strcmp(namedvalue,"BURST") == 0){
    data->daqmode = CCP_DAQ_MODE_BURST;
  }

  if(cs_a2l_getstringvalue(lresult,"CONSISTENCY",namedvalue,CS_CHARBUF)){
    _ccp_error("Failed to parse CONSISTENCY\n");
    pcre_free(re);
    return CS_FAIL;
  }
  if(strcmp(namedvalue,"DAQ") == 0){
    data->consistency = CCP_CONSISTENCY_DAQ;
  }
  if(strcmp(namedvalue,"ODT") == 0){
    data->consistency = CCP_CONSISTENCY_ODT;
  }

  if(cs_a2l_getstringvalue(lresult,"ADDRESS_EXTENSION",namedvalue,CS_CHARBUF)){
    _ccp_error("Failed to parse ADDRESS_EXTENSION\n");
    pcre_free(re);
    return CS_FAIL;
  }
  if(strcmp(namedvalue,"DAQ") == 0){
    data->addrextension = CCP_ADDREXTENSION_DAQ;
  }
  
  //Get optional cmds
  while(!cs_ccp_a2l_GetOptionalCmds(&lopt,&value)){
    data->cmd_optional[counter] = (uint8_t)value;
    counter++;
  }

  return CS_OK;
}

/**************************************************************************/
int cs_ccp_a2l_GetRaster(
			 cs_ccp_a2l_t *data,
			 cs_ccp_a2l_event_t *ev,
			 char *name
			 )
/**************************************************************************/
{
  cs_parser_result_t lresult;
  pcre *re = NULL;
  char regex[CS_REGEX_BUFFER];
  int rv = 0;
  int noresults = 10;
  int ovector[CS_REGEX_RESULTSIZE*noresults];
  char namedvalue[CS_CHARBUF];
  int value;

  if((data == NULL)||(ev == NULL)||(name == NULL)){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }

  memset(ev,0,sizeof(cs_ccp_a2l_event_t));
  
  lresult.searchin = data->a2l.iface.result;
  lresult.searchlen = data->a2l.iface.resultlen;
  if(cs_a2l_search_section(&lresult,"RASTER",name)){
    _ccp_error("Regex failure\n");
    return CS_FAIL;
  }

  if(lresult.result == NULL){
    _ccp_verbose("Cannot find RASTER %s\n",name);
    return CS_FAIL;
  }

  //cs_parser_printresult(&lresult);

  //define serachregion
  lresult.searchin = lresult.result;
  lresult.searchlen = lresult.resultlen;

  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER);
  snprintf(regex,CS_REGEX_BUFFER,"RASTER[\\s]+\"(?'name'(.*?))\"[\\s]+\"(?'shortname'(.*?))\"[\\s]+(?'id'[0-9]+)[\\s]+(?'cse'[0-9]+)[\\s]+(?'rate'[0-9]+)");
  //_ccp_verbose("Regex: %s\n",regex);
  //Precompile regex
  if((re = cs_parser_precompile((const char *)regex)) == NULL){
    _ccp_error("Regex precompilation failed\n");
    return CS_FAIL;
  }
  //Get data, only one fit is allowed
  rv = pcre_exec(re, NULL, (const char*)lresult.result,lresult.resultlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE * noresults);
  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _ccp_error("No content match %s\n");
      pcre_free(re);
      return CS_FAIL;
      break;
    default: 
      _ccp_error("Matching error %d\n", rv);
      pcre_free(re);
      return CS_FAIL;
      break;
    }
  } else {

    if(cs_parser_getsubstring(re,lresult.result,ovector,CS_REGEX_RESULTSIZE*noresults,"name",namedvalue)){
      _ccp_error("Cannot find name section\n");
      pcre_free(re);
      return CS_FAIL;
    }
    strncpy(ev->name,namedvalue,CS_CHARBUF);

    if(cs_parser_getsubstring(re,lresult.result,ovector,CS_REGEX_RESULTSIZE*noresults,"shortname",namedvalue)){
      _ccp_error("Cannot find shortname section\n");
      pcre_free(re);
      return CS_FAIL;
    }
    strncpy(ev->shortname,namedvalue,CS_CHARBUF);
    
    if(cs_parser_getsubstring_int(re,lresult.result,ovector,CS_REGEX_RESULTSIZE*noresults,"id",&value,0)){
      _ccp_error("Failed to get substring id\n");
      pcre_free(re);
      return CS_FAIL;
    }
    ev->id = (uint8_t)value;

    if(cs_parser_getsubstring_int(re,lresult.result,ovector,CS_REGEX_RESULTSIZE*noresults,"cse",&ev->cse,0)){
      _ccp_error("Failed to get substring cse\n");
      pcre_free(re);
      return CS_FAIL;
    }

    if(cs_parser_getsubstring_int(re,lresult.result,ovector,CS_REGEX_RESULTSIZE*noresults,"rate",&ev->samplerate,0)){
      _ccp_error("Failed to get substring rate\n");
      pcre_free(re);
      return CS_FAIL;
    }
  }
  pcre_free(re);

  return CS_OK;
}

/**************************************************************************/
int cs_ccp_a2l_GetSource(
			 cs_ccp_a2l_t *data,
			 cs_ccp_a2l_source_t *s,
			 char *name
			 )
/**************************************************************************/
{
  cs_parser_result_t lresult;
  cs_parser_result_t lblob;

  pcre *re = NULL;
  char regex[CS_REGEX_BUFFER];
  int rv = 0;
  int noresults = 10;
  int ovector[CS_REGEX_RESULTSIZE*noresults];
  char namedvalue[CS_CHARBUF];
  uint32_t value;
  uint8_t rvalue;
  int i;

  if((data == NULL)||(s == NULL)||(name == NULL)){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }

  memset(s,0,sizeof(cs_ccp_a2l_source_t));
  
  lresult.searchin = data->a2l.iface.result;
  lresult.searchlen = data->a2l.iface.resultlen;
  if(cs_a2l_search_section(&lresult,"SOURCE",name)){
    _ccp_error("Regex failure\n");
    return CS_FAIL;
  }

  if(lresult.result == NULL){
    _ccp_error("Cannot find SOURCE %s\n",name);
    return CS_FAIL;
  }

  //cs_parser_printresult(&lresult);

  //define serachregion
  lresult.searchin = lresult.result;
  lresult.searchlen = lresult.resultlen;

  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER);
  snprintf(regex,CS_REGEX_BUFFER,"SOURCE[\\s]+\"(?'name'(.*?))\"[\\s]+(?'period'[0-9]+)[\\s]+(?'rate'[0-9]+)");
  //_ccp_verbose("Regex: %s\n",regex);
  //Precompile regex
  if((re = cs_parser_precompile((const char *)regex)) == NULL){
    _ccp_error("Regex precompilation failed\n");
    return CS_FAIL;
  }
  //Get data, only one fit is allowed
  rv = pcre_exec(re, NULL, (const char*)lresult.result,lresult.resultlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE * noresults);
  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _ccp_error("No content match %s\n");
      pcre_free(re);
      return CS_FAIL;
      break;
    default: 
      _ccp_error("Matching error %d\n", rv);
      pcre_free(re);
      return CS_FAIL;
      break;
    }
  } else {

    if(cs_parser_getsubstring(re,lresult.result,ovector,CS_REGEX_RESULTSIZE*noresults,"name",namedvalue)){
      _ccp_error("Cannot find name section\n");
      pcre_free(re);
      return CS_FAIL;
    }
    strncpy(s->name,namedvalue,CS_CHARBUF);

    if(cs_parser_getsubstring_int(re,lresult.result,ovector,CS_REGEX_RESULTSIZE*noresults,"period",&s->period,0)){
      _ccp_error("Failed to get substring period\n");
      pcre_free(re);
      return CS_FAIL;
    }

    if(cs_parser_getsubstring_int(re,lresult.result,ovector,CS_REGEX_RESULTSIZE*noresults,"rate",&s->rate,0)){
      _ccp_error("Failed to get substring rate\n");
      pcre_free(re);
      return CS_FAIL;
    }
  }
  pcre_free(re);

  cs_a2l_getstringvalue(lresult,"DISPLAY_IDENTIFIER",s->displayident,CS_CHARBUF);

  lblob = lresult;
  if(cs_a2l_search_section(&lblob,"QP_BLOB",NULL)){
    _ccp_error("Regex failure\n");
    return CS_FAIL;
  }

  if(lblob.result == NULL){
    _ccp_error("Cannot find QP_BLOB section\n");
    return CS_FAIL;
  }

  //cs_parser_printresult(&lblob);

  //define serachregion
  lblob.searchin = lblob.result;
  lblob.searchlen = lblob.resultlen;

  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER);
  snprintf(regex,CS_REGEX_BUFFER,"QP_BLOB[\\s]+(?'id'[0-9]+)");
  //_ccp_verbose("Regex: %s\n",regex);
  //Precompile regex
  if((re = cs_parser_precompile((const char *)regex)) == NULL){
    _ccp_error("Regex precompilation failed\n");
    return CS_FAIL;
  }
  //Get data, only one fit is allowed
  rv = pcre_exec(re, NULL, (const char*)lblob.result,lblob.resultlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE * noresults);
  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _ccp_error("No content match %s\n");
      pcre_free(re);
      return CS_FAIL;
      break;
    default: 
      _ccp_error("Matching error %d\n", rv);
      pcre_free(re);
      return CS_FAIL;
      break;
    }
  } else {

    if(cs_parser_getsubstring_int(re,lblob.result,ovector,CS_REGEX_RESULTSIZE*noresults,"id",&s->id,0)){
      _ccp_error("Failed to get substring id\n");
      pcre_free(re);
      return CS_FAIL;
    }
  }
  pcre_free(re);

  s->length = (int)cs_a2l_getuintvalue(lblob,"LENGTH",&rvalue);
  if(rvalue != CS_OK){
    _ccp_error("Failed to get LENGTH\n");
    return CS_FAIL;
  }
  
  s->can_id = cs_a2l_getuintvalue(lblob,"CAN_ID_FIXED",&rvalue);
  if(rvalue)
    s->can_id = 0;

  value = cs_a2l_getuintvalue(lblob,"RASTER",&rvalue);
  if(!rvalue){
    //There is a Raster ID
    s->event = NULL;
    for(i = 0; i < data->no_events;i++){
      if(data->events[i].id == value){
	s->event = &data->events[i];
	break;
      }
    }
    if(s->event == NULL){
      _ccp_verbose("No existing RASTER with id %i\n",value);
    }
  }

  s->reduction_allowed = (!cs_a2l_getflag(lblob,"REDUCTION_ALLOWED")?1:0);
  
  value = cs_a2l_getuintvalue(lblob,"FIRST_PID",&rvalue);
  s->first_pid = (uint8_t)value;

  return CS_OK;
}


/**************************************************************************/
void cs_ccp_a2l_print_source(
			     cs_ccp_a2l_source_t *data
			     )
/**************************************************************************/
{
  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return;
  }

  _ccp_info("A2L Source Data: Name %s\n",data->name);
  _ccp_info("A2L Source Data: Display identidfier %s\n",data->displayident);
  _ccp_info("A2L Source Data: Id 0x%X\n",data->id);
  _ccp_info("A2L Source Data: Perid %i ms\n",data->period);
  _ccp_info("A2L Source Data: Sample time %i ms\n",data->rate);
  _ccp_info("A2L Source Data: Reduction allowed %s\n",data->reduction_allowed?"Yes":"No");
  _ccp_info("A2L Source Data: First PID 0x%X\n",data->first_pid);
  if(data->event != NULL)
    cs_ccp_a2l_print_raster(data->event);
}

/**************************************************************************/
void cs_ccp_a2l_print_raster(
			     cs_ccp_a2l_event_t *data
			     )
/**************************************************************************/
{
  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return;
  }

  _ccp_info("A2L Event Data: Name %s\n",data->name);
  _ccp_info("A2L Event Data: Short name %s\n",data->shortname);
  _ccp_info("A2L Event Data: Id 0x%X\n",data->id);
  _ccp_info("A2L Event Data: CSE %i\n",data->cse);
  _ccp_info("A2L Event Data: Sample time %i ms\n",data->samplerate);

}

/**************************************************************************/
void cs_ccp_a2l_print(
		      cs_ccp_a2l_t *data
		      )
/**************************************************************************/
{
  int i;
  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return;
  }

  _ccp_info("CCP Version 0x%X\n",data->ccp_version);
  _ccp_info("Blob Version 0x%X\n",data->blob_version);
  _ccp_info("CAN send id 0x%X\n",data->can_send_id);
  _ccp_info("CAN recv id 0x%X\n",data->can_recv_id);
  _ccp_info("Station id 0x%X\n",data->station_id);
  _ccp_info("Baudrate %i\n",data->baudrate);
  switch(data->BO){
  case A2L_BO_MOTOROLLA:
    _ccp_info("Byte order Motorola\n");
    break;
  case A2L_BO_INTEL:
    _ccp_info("Byte order Intel\n");
    break;
  case A2L_BO_UNKNOWN:
    _ccp_info("Byte order not defined in signal\n");
    break;
  }
  switch(data->daqmode){
  case CCP_DAQ_MODE_UNKNOWN:
    _ccp_info("DAQ mode unknown\n");
    break;
  case CCP_DAQ_MODE_BURST:
    _ccp_info("DAQ mode Burst\n");
    break;
  }
  switch(data->consistency){
  case CCP_CONSISTENCY_UNKNOWN:
    _ccp_info("Data consistency unknown\n");
    break;
  case CCP_CONSISTENCY_DAQ:
    _ccp_info("Data consistency DAQ\n");
    break;
  case CCP_CONSISTENCY_ODT:
    _ccp_info("Data consistency ODT\n");
    break;
  }
  switch(data->addrextension){
  case CCP_ADDREXTENSION_UNKNOWN:
    _ccp_info("Address extension unknown\n");
    break;
  case CCP_ADDREXTENSION_DAQ:
    _ccp_info("Address extension DAQ\n");
    break;
  }
  for(i = 0; i < CS_CHARBUF; i++){
    if(data->cmd_optional[i] == 0)
      break;
    _ccp_info("Optional cmd 0x%X\n",data->cmd_optional[i]);
  }
  
  for(i = 0; i < data->no_events; i++){
    cs_ccp_a2l_print_raster(&data->events[i]);
  }

  for(i = 0; i < data->no_source; i++){
    cs_ccp_a2l_print_source(&data->source[i]);
  }

}

/**************************************************************************/
int cs_ccp_a2l_open(
		    cs_ccp_a2l_t *data,
		    char *filename,
		    char *project,
		    char *module,
		    char *interface,
		    char *option
		    )
/**************************************************************************/
{
  int asap2major = 0;
  int asap2minor = 0;
  cs_a2l_list_t list;
  int i;

  if((data == NULL) ||
     (filename == NULL) ||
     (project == NULL) ||
     (module == NULL) ||
     (interface == NULL)){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }

  data->no_events = 0;
  data->no_source = 0;
  cs_ccp_a2l_init(data);
  
  if(cs_a2l_open(&data->a2l,filename)){
    return CS_FAIL;
  }

  //Get ASAP-Version
  if(cs_a2l_getversion(&data->a2l,&asap2major,&asap2minor)){
    _ccp_error("Failed to get ASAP2 version\n");
    return CS_FAIL;
  }  
  _ccp_verbose("%s, A2L Version Major %i Minor %i\n",filename,asap2major,asap2minor);

  //Get Project
  if(cs_a2l_GetProject(&data->a2l,project)){
    _ccp_error("Cannot get Project %s\n",project);
    return CS_FAIL;
  }

  //Get Module
  if(cs_a2l_GetModule(&data->a2l,module)){
    _ccp_error("Cannot get Module %s\n",module);
    return CS_FAIL;
  }

  //Get Module
  if(cs_a2l_GetInterface(&data->a2l,interface)){
    _ccp_error("Cannot get Interface %s\n",interface);
    return CS_FAIL;
  }

  if(option != NULL){
    cs_a2l_GetModPar(&data->a2l,option);
  } else{
    //Try to get standard section
    if(cs_a2l_GetModPar(&data->a2l,"")){
      _ccp_verbose("Failed to get default MOD_PAR\n");
    }
  }

  //Get common parameter if exists
  if(cs_a2l_GetModCommon(&data->a2l)){
    _ccp_verbose("Cannot handle common data\n");
  }

  if(cs_ccp_a2l_GetTransportData(data)){
    _ccp_error("Cannot handle transpport data\n");
    return CS_FAIL;
  }

  //Set default address granulation for CCP
  data->a2l.AG =  A2L_AG_BYTE;

  //Handle event definitions

  //Get list of events section
  if(cs_a2l_list_sections(& data->a2l.iface,&list,"RASTER")){
    _ccp_error("Failed to build up section list\n");
    cs_a2l_list_free(&list);
    return CS_FAIL;
  }
  //printf("Available RASTERs:\n");
  //cs_a2l_printlist(&list);
  data->no_events = list.no_result;

  //Malloc storage
  data->events = (cs_ccp_a2l_event_t*)malloc(sizeof(cs_ccp_a2l_event_t)*data->no_events);

  if(data->events == NULL){
    _ccp_error("Malloc failure\n");
    cs_a2l_list_free(&list);
    return CS_FAIL;
  }
 
  for(i = 0; i < data->no_events; i++){
    //_ccp_verbose("Raster %s\n",list.resultname[i]);

    if(cs_ccp_a2l_GetRaster(data,&data->events[i],list.resultname[i])){
      _ccp_error("Failed to parse Raster %s\n",list.resultname[i]);
      data->no_events = 0;
      free(data->events);
      cs_a2l_list_free(&list);
      return CS_FAIL;
    }
  }

  cs_a2l_list_free(&list);

  //Handle source definitions

  //Get list of events section
  if(cs_a2l_list_sections(& data->a2l.iface,&list,"SOURCE")){
    _ccp_error("Failed to build up section list\n");
    cs_a2l_list_free(&list);
    return CS_FAIL;
  }
  //printf("Available SOURCEs:\n");
  //cs_a2l_printlist(&list);
  data->no_source = list.no_result;

  //Malloc storage
  data->source = (cs_ccp_a2l_source_t*)malloc(sizeof(cs_ccp_a2l_source_t)*data->no_source);

  if(data->events == NULL){
    _ccp_error("Malloc failure\n");
    cs_a2l_list_free(&list);
    return CS_FAIL;
  }

  for(i = 0; i < data->no_source; i++){
    //_ccp_verbose("SOURCE %s\n",list.resultname[i]);

    if(cs_ccp_a2l_GetSource(data,&data->source[i],list.resultname[i])){
      _ccp_error("Failed to parse Source %s\n",list.resultname[i]);
      data->no_source = 0;
      free(data->source);
      cs_a2l_list_free(&list);
      return CS_FAIL;
    }
  }

  cs_a2l_list_free(&list);

  return CS_OK;
}

/**************************************************************************/
int cs_ccp_a2l_close(
		     cs_ccp_a2l_t *data
		     )
/**************************************************************************/
{
  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }

  if((data->no_events > 0) && (data->events != NULL)){
    free(data->events);
    data->no_events = 0;
  }

  if(cs_a2l_close(&data->a2l)){
    return CS_FAIL;
  }
  return CS_OK;
}


/**************************************************************************/
cs_ccp_a2l_source_t *cs_ccp_a2l_GetDAQ(
				       cs_ccp_a2l_t *data,
				       uint8_t daq
				       )
/**************************************************************************/
{
  int i;

  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return NULL;
  }

  for(i = 0; i < data->no_source; i++){
    if(data->source[i].id == daq){
      return &data->source[i];
    }
  }

  return NULL;
}

/**************************************************************************/
double cs_ccp_a2l_GetDAQPeriod(
			       cs_ccp_a2l_t *data,
			       uint8_t daq
			       )
/**************************************************************************/
{
  cs_ccp_a2l_source_t * source = NULL;

  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return 0.0;
  }
  
  source = cs_ccp_a2l_GetDAQ(data,daq);

  if(source == NULL)
    return 0.0;

  if(source->period != 0)
    return ((double)source->period)/1.0e-3;

  if(source->event == NULL)
    return 0.0;
  
  return (double)source->event->samplerate/1.0e-3;
}

/**************************************************************************/
int cs_ccp_a2l_GetDAQEvent(
			   cs_ccp_a2l_t *data,
			   uint8_t daq
			   )
/**************************************************************************/
{
  cs_ccp_a2l_source_t * source = NULL;

  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return -1;
  }
  
  source = cs_ccp_a2l_GetDAQ(data,daq);

  if(source == NULL)
    return -1; //No source found

  if(source->event == NULL)
    return -1; // No event found

  return source->event->id;

}
