/*  CCP CAN Stack for SocketCAN
    Copyright (C) 2015 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/mman.h>
#include <unistd.h>

#include "cs_ccp_api.h"
#include "cs_calibration_utils.h"
#include "cs_defines.h"
#include "cs_ccp_private.h"


/**************************************************************************/
int cs_ccp_api_load(
		    cs_ccp_t *data,
		    char * filename,
		    char* project,
		    char* module,
		    char* option,
		    uint32_t stim
		    )
/**************************************************************************/
{

  if((data == NULL) ||
     (filename == NULL) ||
     (project == NULL) ||
     (module == NULL))
    return CS_FAIL;

  //Try to open file
  if(cs_ccp_a2l_open(&data->ccpa2l,filename,project,module,"ASAP1B_CCP",option)){
    _ccp_error("Failed to open A2L file\n");
    return CS_FAIL;
  }

  //Init protocol stack
  cs_ccp_init(data);

  cs_ccp_set_stim(data,stim);

  return CS_OK;
}


/**************************************************************************/
int cs_ccp_api_open(
		    cs_ccp_t *data,
		    char *iface
		    )
/**************************************************************************/
{
  uint8_t ccpmajor;
  uint8_t ccpminor;
  uint8_t exchangeidlength;
  uint8_t counter;
  uint8_t size;

  if((data == NULL)||(iface == NULL))
    return CS_FAIL;

  //Open CAN interface
  if(cs_ccp_can_open(data,iface)){
    _ccp_error("Failed to open CAN interface: %s\n",cs_ccp_getlasterrorstring(data));
    return CS_FAIL;
  }

  //Connect to slave
  if(cs_ccp_cmd_connect(data)){
    _ccp_error("%s\n",cs_ccp_getlasterrorstring(data));
    return CS_FAIL;
  }
  cs_ccp_wait_stim(data);

  //Get CCP Version
  if(cs_ccp_cmd_ccpversion(data,&ccpmajor,&ccpminor)){
    _ccp_error("%s\n", cs_ccp_getlasterrorstring(data));
    return CS_FAIL;
  }
  //Check Version
  if(ccpmajor != CS_CCP_VERSION_MAJOR){
    data->lerror = CCP_ERROR_CCPVERSION;
    _ccp_verbose("%s\n", cs_ccp_getlasterrorstring(data));
  }
  cs_ccp_wait_stim(data);
  //Get slave resources
  if(cs_ccp_cmd_exchangeid(data,&exchangeidlength,NULL,NULL,NULL)){
    _ccp_error("%s\n", cs_ccp_getlasterrorstring(data));
    return CS_FAIL;
  }
  cs_ccp_wait_stim(data);
  memset(data->slaveid,0,sizeof(char)*CCP_EXCHANGEID_SIZE);
  for(counter = 0; counter < exchangeidlength; counter+=5){
    if((size = (exchangeidlength -counter))> 5){
      size = 5;
    }
    if(cs_ccp_cmd_upload(data,&data->slaveid[counter],size)){
      return CS_FAIL;
    }
    cs_ccp_wait_stim(data);
  }
  return CS_OK;
}

/**************************************************************************/
int cs_ccp_api_close(
		     cs_ccp_t *data
		     )
/**************************************************************************/
{
  if(data == NULL)
    return CS_FAIL;

  if(cs_ccp_cmd_disconnect(data,1)){
    _ccp_error("%s\n",cs_ccp_getlasterrorstring(data));
    return CS_FAIL;
  }
  cs_ccp_wait_stim(data);

  //Close CAN interface
  if(cs_ccp_can_close(data)){
    _ccp_error("Failed to close CAN interface\n");
    return CS_FAIL;
  }

  //Close stack
  if(cs_ccp_close(data)){
    _ccp_error("Failed to close CCP stack: %s\n",cs_ccp_getlasterrorstring(data));
    return CS_FAIL;
  }

  //Close A2L file
  if(cs_ccp_a2l_close(&data->ccpa2l)){
    _ccp_error("Failed to close file\n");
    return CS_FAIL;
  }

  return CS_OK;
}

/**************************************************************************/
char *cs_ccp_api_getslaveid(
			    cs_ccp_t *data
			    )
/**************************************************************************/
{
  if(data == NULL)
    return NULL;

  return (char*)data->slaveid;
}
  
/**************************************************************************/
int cs_ccp_api_addSignalDAQ(
			    cs_ccp_t *data,
			    char* signalname,
			    char* ev_name,
			    int prescaler,
			    double *value,
			    double *ts,
			    void (*extcallback)(double value, double ts, void *priv),
			    void *priv
			    )
/**************************************************************************/
{
  int i;

  if(data == NULL)
    return CS_FAIL;

  if((signalname == NULL) || (ev_name == NULL)){
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL;
  }

  //Check for already configured signal number
  if((data->ccpa2l.a2l.signalcounter + 1) >= CS_A2L_SIGNALMAX){
    _ccp_error("Number of signals overflow\n");
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL;
  }

  //Check for valid prescaler
  if(prescaler < 1){
    _ccp_error("Invalid prescaler\n");
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL;
  }

  //Search for signal
  if(cs_a2l_GetMeasurement(&data->ccpa2l.a2l,&data->ccpa2l.a2l.signals[data->ccpa2l.a2l.signalcounter],signalname)){
    data->lerror = CCP_ERROR_A2LSIGNAL;
    _ccp_error("Cannot get signal from A2L file\n");
    return CS_FAIL;
  }
  
  cs_a2l_print_measurement(&data->ccpa2l.a2l.signals[data->ccpa2l.a2l.signalcounter]);

  //Look if Event exists
  for(i = 0; i < data->ccpa2l.no_events; i++){
    if(strcmp(data->ccpa2l.events[i].name,ev_name) == 0){
      //found Event, store event number
      data->ccpa2l.a2l.signals[data->ccpa2l.a2l.signalcounter].ev_idx = i;
      break;
    }
  }
  if(i >= data->ccpa2l.no_events){
    data->lerror = CCP_ERROR_PARAMETER;
    _ccp_error("Invalid event\n");
    return CS_FAIL;
  }

  //Set prescaler
  data->ccpa2l.a2l.signals[data->ccpa2l.a2l.signalcounter].prescaler = prescaler;

  //Store feedback values
  data->ccpa2l.a2l.signals[data->ccpa2l.a2l.signalcounter].exvalue = value;
  data->ccpa2l.a2l.signals[data->ccpa2l.a2l.signalcounter].exts = ts;
  data->ccpa2l.a2l.signals[data->ccpa2l.a2l.signalcounter].extcallback = extcallback;
  data->ccpa2l.a2l.signals[data->ccpa2l.a2l.signalcounter].extcallbackpriv = priv;
  
  //Increment signalcounter, because we have a new valid system
  data->ccpa2l.a2l.signalcounter++;

  return CS_OK;
}

/**************************************************************************/
int cs_ccp_api_connect(
		       cs_ccp_t *data
		       )
/**************************************************************************/
{

  if(data == NULL)
    return CS_FAIL;

  //Try to connect
  if(cs_ccp_cmd_connect(data)){
    _ccp_error("Failed to connect to slave\n");
    return CS_FAIL;
  }
  cs_ccp_wait_stim(data);
 
  return CS_OK;
}


/**************************************************************************/
int cs_ccp_api_configureDAQ(
			    cs_ccp_t *data
			    )
/**************************************************************************/
{
  int i,j;
  int daq_idx = -1;
  uint8_t daq_size;
  int maxodt = 0xFF; //Maximal number of ODTs per DAQ
  int flag_insert = 0; //Flag storage for insert signal in ODT
  uint8_t did;

  if(data == NULL)
    return CS_FAIL;

  if(data->ccpa2l.no_source == 0){
    _ccp_error("DAQ max is zero\n");
    data->lerror = CCP_ERROR_DAQCONFIG;
    return CS_FAIL;
  }


  //We will build up a structure

  //helperstructure to reorder signals on ODT level
  struct odtdata{
    int num_entry; //Number of signals to register
    int used;      //Used bytes
  };
  //helperstructure to reorder signals on DAQ level
  struct{
    struct odtdata odt[maxodt]; //ODT storage structure
    int max_odt;   //Maximal number of ODT in DAQ
    int num_odt;   //Number of ODTs to register
    int prescaler; //Prescaler for DAQ
    int ev_idx;    //Event index
  }daq[data->ccpa2l.no_source];

  //Init structure
  for(i = 0; i< data->ccpa2l.no_source; i++){
    for(j = 0; j < maxodt; j++){
      daq[i].odt[j].used = 0;
      daq[i].odt[j].num_entry = 0;
      
    }
    daq[i].max_odt = data->ccpa2l.source[i].length;
    daq[i].prescaler = -1;
    daq[i].num_odt = 0;
    daq[i].ev_idx = cs_ccp_a2l_GetDAQEvent(&data->ccpa2l,i);
  }

  //Start working on signals
  for(i = 0; i < data->ccpa2l.a2l.signalcounter; i++){
    daq_idx = -1;
    //First search for existing or predefined daq
    for(j = 0; j < data->ccpa2l.no_source; j++){
      if(daq[j].ev_idx != -1){
	if(daq[j].ev_idx == data->ccpa2l.a2l.signals[i].ev_idx){
	  if((daq[j].prescaler == data->ccpa2l.a2l.signals[i].prescaler) || 
	     (daq[j].prescaler != 0)){
	    daq_idx = j;
	    break;
	  }
	}
      } else {
	//Unspecific DAQ without defined event source, take it
	daq_idx = j;
	break;
      }
    }

    if(daq_idx == -1){
      _ccp_error("Cannot find valid DAQ for signal %s\n",data->ccpa2l.a2l.signals[i].name);
      data->lerror = CCP_ERROR_DAQCONFIG;
      return CS_FAIL;
    }    
    
    

    //Now search for odt
    for(j = 0; j < daq[daq_idx].max_odt; j++){
      flag_insert = 0;
      _ccp_verbose("CCP config DAQ 0x%X, ODT 0x%X used: 0x%X, signal size: 0x%X\n",daq_idx,j,
		   daq[daq_idx].odt[j].used,data->ccpa2l.a2l.signals[i].size);
      if((daq[daq_idx].odt[j].used + data->ccpa2l.a2l.signals[i].size) <= CCP_ODTMAXSIZE){
	flag_insert = 1;
      }

      if(!flag_insert){
	//No space in actual ODT, try next ODT
	continue;
      }

      //Increment used odt, if we use a new one
      if(daq[daq_idx].odt[j].used == 0)
	daq[daq_idx].num_odt++;

      //Register Data
      daq[daq_idx].ev_idx = data->ccpa2l.a2l.signals[i].ev_idx;
      daq[daq_idx].prescaler = data->ccpa2l.a2l.signals[i].prescaler;
      data->ccpa2l.a2l.signals[i].daq_idx = daq_idx;
      data->ccpa2l.a2l.signals[i].odt_idx = j;
      data->ccpa2l.a2l.signals[i].entry_idx = daq[daq_idx].odt[j].num_entry;
      daq[daq_idx].odt[j].num_entry++;
      daq[daq_idx].odt[j].used += data->ccpa2l.a2l.signals[i].size;

      _ccp_verbose("CCP API Configure: \n");
      _ccp_verbose("   Signal index %i\n",i);
      _ccp_verbose("   Register Signal %s:\n",data->ccpa2l.a2l.signals[i].name);
      _ccp_verbose("   Signal size in Bytes %i\n",data->ccpa2l.a2l.signals[i].size);
      _ccp_verbose("   Event %s \n",data->ccpa2l.events[data->ccpa2l.a2l.signals[i].ev_idx].name);
      _ccp_verbose("   Prescaler %i\n",data->ccpa2l.a2l.signals[i].prescaler);
      _ccp_verbose("   DAQ idx %i\n",data->ccpa2l.a2l.signals[i].daq_idx);
      _ccp_verbose("   DAQ max ODT %i\n",daq[data->ccpa2l.a2l.signals[i].daq_idx].max_odt);
      _ccp_verbose("   ODT idx %i\n",data->ccpa2l.a2l.signals[i].odt_idx);
      _ccp_verbose("   Entry idx %i\n",data->ccpa2l.a2l.signals[i].entry_idx);
      _ccp_verbose("   ODT num entries %i\n",daq[daq_idx].odt[j].num_entry);
      _ccp_verbose("   ODT bytes used %i\n",daq[daq_idx].odt[j].used);
      cs_a2l_print_measurement(&data->ccpa2l.a2l.signals[i]);
      break;
    }

    if(!flag_insert){
      _ccp_error("Can not find ODT for signal\n");
      data->lerror = CCP_ERROR_DAQCONFIG;
      return CS_FAIL;
    }

  }

  //Reset DAQ configuration and compare odt size
  for(i = 0; i< data->ccpa2l.no_source; i++){
    did = data->ccpa2l.source[i].id;

    if(did >= CCP_DAQMAX){
      _ccp_error("DAQ id out of range\n");
      data->lerror = CCP_ERROR_DAQCONFIG;
      return CS_FAIL;
    }
    
    //Reset actual configuration
    if(cs_ccp_cmd_getdaqsize(data,did,data->ccpa2l.can_recv_id,&daq_size,&data->daqlist.daq[did].first_pid)){
      _ccp_error("%s\n",cs_ccp_getlasterrorstring(data));
      return CS_FAIL;
    }

    if(daq_size != data->ccpa2l.source[i].length){
      _ccp_verbose("DAQ size, A2L does not fit to Slave\n");
      //TODO howto handle the difference
      //data->lerror = CCP_ERROR_DAQCONFIG;
      //return CS_FAIL;
    }
    cs_ccp_wait_stim(data);
  }
  
  //Register signals
  for(i = 0; i < data->ccpa2l.a2l.signalcounter; i++){
    //Register feedback values
    if(cs_ccp_entry_reg_value(data,
			      data->ccpa2l.a2l.signals[i].daq_idx,
			      data->ccpa2l.a2l.signals[i].odt_idx,
			      data->ccpa2l.a2l.signals[i].entry_idx,
			      data->ccpa2l.a2l.signals[i].size,
			      data->ccpa2l.a2l.signals[i].exvalue,
			      data->ccpa2l.a2l.signals[i].BO,
			      data->ccpa2l.a2l.signals[i].cm,
			      data->ccpa2l.a2l.signals[i].dtype,
			      data->ccpa2l.a2l.signals[i].exts,
			      data->ccpa2l.a2l.signals[i].extcallback,
			      data->ccpa2l.a2l.signals[i].extcallbackpriv
			      )){
      _ccp_error("Failed to register feedback\n");
      return CS_FAIL;
    }

    //Set DAQ period for timeout check
    if(cs_ccp_daq_setperiod(data,
			    data->ccpa2l.a2l.signals[i].daq_idx,
			    cs_ccp_a2l_GetDAQPeriod(&data->ccpa2l,data->ccpa2l.a2l.signals[i].daq_idx) *
			    (double)data->ccpa2l.a2l.signals[i].prescaler)){
      _ccp_error("Failed to set DAQ period: %s\n",cs_ccp_getlasterrorstring(data));
      return CS_FAIL;
    }

    //Set prescaler for startstop cmd
    if(cs_ccp_daq_setprescaler(data,
			       data->ccpa2l.a2l.signals[i].daq_idx,
			       (uint16_t)data->ccpa2l.a2l.signals[i].prescaler)){
      _ccp_error("Failed to set DAQ prescaler: %s\n",cs_ccp_getlasterrorstring(data));
      return CS_FAIL;
    }

    //Set working pointer
    _ccp_error("Set DAQ PTR, DAQ %i, ODT %i, Entry %i\n",
	       data->ccpa2l.a2l.signals[i].daq_idx,
	       data->ccpa2l.a2l.signals[i].odt_idx,
	       data->ccpa2l.a2l.signals[i].entry_idx);
    if(cs_ccp_cmd_setdaqptr(data,
			    data->ccpa2l.a2l.signals[i].daq_idx,
			    data->ccpa2l.a2l.signals[i].odt_idx,
			    data->ccpa2l.a2l.signals[i].entry_idx)){ 
      //always first entry
      _ccp_error("Failed to set DAQ pointer: %s\n",cs_ccp_getlasterrorstring(data));
      return CS_FAIL;
    }
    cs_ccp_wait_stim(data);

    //Store memory addr in ODT Entry
    _ccp_verbose("Write DAQ, Size %i, Extension 0x%X, Address 0x%X\n",
		 data->ccpa2l.a2l.signals[i].size,
		 data->ccpa2l.a2l.signals[i].extension,
		 data->ccpa2l.a2l.signals[i].address);
    if(cs_ccp_cmd_writedaq(data,
			   data->ccpa2l.a2l.signals[i].size,
			   data->ccpa2l.a2l.signals[i].extension,
			   cs_a2l_CalcAddress(&data->ccpa2l.a2l,data->ccpa2l.a2l.signals[i].address))){
      _ccp_error("Failed to write DAQ: %s\n",cs_ccp_getlasterrorstring(data));
      return CS_FAIL;
    }
    cs_ccp_wait_stim(data);

  }
  
  return CS_OK;
}

/**************************************************************************/
int cs_ccp_api_startDAQ(
			cs_ccp_t *data
			)
/**************************************************************************/
{
  int i;
  cs_ccp_daq_t *daq = NULL;

  if(data == NULL)
    return CS_FAIL;

  //Select DAQ list
  for(i = 0; i < data->daqlist.no_active; i++){
    daq = &data->daqlist.daq[data->daqlist.daq_active[i]];
    if(cs_ccp_cmd_startstop(data,
			    1,
			    data->daqlist.daq_active[i],
			    cs_ccp_odt_getmaxactive(data,i),
			    daq->ev,daq->prescaler)){
      _ccp_error("Failed to start DAQ list: %s\n",cs_ccp_getlasterrorstring(data));
      return CS_FAIL;
    }
    cs_ccp_wait_stim(data);
  }
  
  return CS_OK;
}

/**************************************************************************/
int cs_ccp_api_stopDAQ(
		       cs_ccp_t *data
		       )
/**************************************************************************/
{
  int i;
  cs_ccp_daq_t *daq = NULL;

  if(data == NULL)
    return CS_FAIL;


  //Select DAQ list
  for(i = 0; i < data->daqlist.no_active; i++){
    daq = &data->daqlist.daq[data->daqlist.daq_active[i]];
    if(cs_ccp_cmd_startstop(data,0,data->daqlist.daq_active[i],daq->no_odt,daq->ev,daq->prescaler)){
      _ccp_error("Failed to start DAQ list: %s\n",cs_ccp_getlasterrorstring(data));
      return CS_FAIL;
    }
    cs_ccp_wait_stim(data);
  }


  return CS_OK;
}

/**************************************************************************/
int cs_ccp_api_process(
		       cs_ccp_t *data
		       )
/**************************************************************************/
{
  if(data == NULL)
    return CS_FAIL;
  
  if(cs_ccp_receive(data)){
    _ccp_error("Failed to receive data\n");
    return CS_FAIL;
  }
  return CS_OK;
}

/**************************************************************************/
int cs_ccp_api_setDAQTimeout(
			     cs_ccp_t *data,
			     double fac
			     )
/**************************************************************************/
{
  if(data == NULL)
    return CS_FAIL;
    
  data->daqlist.daq_fac_recvtimeout = fac;
    
  return CS_OK;
}

/**************************************************************************/
int cs_ccp_api_hasDAQTimeout(
			     cs_ccp_t *data
			     )
/**************************************************************************/
{
  if(data == NULL)
    return CS_FAIL;   
  //Check for timeouts
  if(cs_ccp_daq_istimeout(data,data->daqlist.daq_fac_recvtimeout)){
    _ccp_verbose("DAQ Timeout detected\n");
    return CS_FAIL;
  }
  return CS_OK;
}

/**************************************************************************/
int cs_ccp_api_dnload_area(
			   cs_ccp_t *data,
			   uint32_t address,
			   uint8_t extension,
			   uint8_t *rdata,
			   int rlen
			   )
/**************************************************************************/
{
  int counter;
  int size;

  if(data == NULL)
    return CS_FAIL;

  if((rdata == NULL)||(rlen <= 0)){
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL;
  }

  //SetMTA
  if(cs_ccp_cmd_setmta(data,0,cs_a2l_CalcAddress(&data->ccpa2l.a2l,address),extension)){
    return CS_FAIL;
  } 
  cs_ccp_wait_stim(data);
  for(counter = 0; counter < rlen; counter+=5){
    //_ccp_verbose("Size: 0x%X Counter: 0x%X\n",rlen,counter);
    if((size = (rlen -counter))> 5){
      size = 5;
    }
    if(size == 0){
      return CS_OK;
    }
    //_ccp_verbose("CCP API Dnload area: Dnload size: 0x%X\n",size);
    if(cs_ccp_cmd_dnload(data,&rdata[counter],size)){
      return CS_FAIL;
    }
    cs_ccp_wait_stim(data);
  }
  return CS_OK;
}

/**************************************************************************/
int cs_ccp_api_upload_area(
			   cs_ccp_t *data,
			   uint32_t address,
			   uint8_t extension,
			   uint8_t *rdata,
			   int rlen
			   )
/**************************************************************************/
{
  int counter;
  int size;

  if(data == NULL)
    return CS_FAIL;

  if((rdata == NULL)||(rlen <= 0)){
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL;
  }

  //SetMTA
  if(cs_ccp_cmd_setmta(data,0,cs_a2l_CalcAddress(&data->ccpa2l.a2l,address),extension)){
    return CS_FAIL;
  } 
  cs_ccp_wait_stim(data);
  for(counter = 0; counter < rlen; counter+=5){
    if((size = (rlen -counter))> 5){
      size = 5;
    }
    if(cs_ccp_cmd_upload(data,&rdata[counter],size)){
      return CS_FAIL;
    }
    cs_ccp_wait_stim(data);
  }
  return CS_OK;
}

/**************************************************************************/
int cs_ccp_api_dnload_value(
			    cs_ccp_t *data,
			    char *name,
			    double *value,
			    int numelem
			    )
/**************************************************************************/
{
  cs_a2l_characteristic_t ch;
  uint8_t rdata[CCP_API_MAXAREA];
  uint8_t tdata[CCP_API_MAXAREA];
  int counter = 0;
  int esize;
  int i;

  if(data == NULL)
    return CS_FAIL;

  if((name == NULL)||(value == NULL)||(numelem <=0)){
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL;
  }

  memset(rdata,0,sizeof(uint8_t)*CCP_API_MAXAREA);

  //Try to get characteristic
  if(cs_a2l_GetCharacteristic(&data->ccpa2l.a2l,&ch,name)){
    _ccp_error("Failed to get characteristic %s data\n",name);
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL;
  }

  cs_a2l_print_characteristic(&ch);

  //Check element size
  if(numelem != ch.record.size){
    _ccp_error("Number of elements does not fit\n");
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL;
  }

  if(ch.record.size > 1){
    _ccp_error("Sorry record size > 1 actually not supported\n");
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL;
  }

  for(i = 0; i < ch.record.size; i++){
    if(ch.record.rtype[i] != A2L_RECORD_TYPE_FNC){
      _ccp_error("Sorry record type actually not supported\n");
      data->lerror = CCP_ERROR_PARAMETER;
      return CS_FAIL;
    }
    
    if(ch.record.access[i] != A2L_RECORD_ACCESS_DIRECT){
      _ccp_error("Sorry record access type actually not supported\n");
      data->lerror = CCP_ERROR_PARAMETER;
      return CS_FAIL;
    }
    _ccp_verbose("Value in process Index: %i Value: %f\n",i,value[i]);
    //Check for Limits
    if(cs_a2l_characteristic_limit(&ch,&value[i])){
      _ccp_error("Limit check failed\n");
      data->lerror = CCP_ERROR_PARAMETER;
      return CS_FAIL;
    }

    memset(tdata,0,sizeof(uint8_t)*CCP_API_MAXAREA);
    esize = cs_a2l_TypeToAG(ch.record.dtype[i]) * cs_a2l_get_ag_bytes(ch.AG);
    if(cs_a2l_convert_from_double(ch.BO,tdata,
				  esize,ch.cm,ch.record.dtype[i],value[i])){
      _ccp_error("Sorry record access type actually not supported\n");
      data->lerror = CCP_ERROR_PARAMETER;
      return CS_FAIL;
    }
    if((counter + esize) > CCP_API_MAXAREA){
      _ccp_error("Invalid size data, buffer overrange\n");
      data->lerror = CCP_ERROR_PARAMETER;
      return CS_FAIL;
    }
    memcpy(&rdata[counter],tdata,esize);
    counter+=esize;
    
  }

  if(cs_ccp_api_dnload_area(data,ch.address,ch.extension,rdata,counter)){
    _ccp_error("Download failed\n");
    return CS_FAIL;
  }

  return CS_OK;
}


/**************************************************************************/
int cs_ccp_api_upload_value(
			    cs_ccp_t *data,
			    char *name,
			    double *value,
			    int *numelem
			    )
/**************************************************************************/
{
  cs_a2l_characteristic_t ch;
  uint8_t rdata[ CCP_API_MAXAREA];
  int counter = 0;
  int rsize = 0;
  int esize = 0;
  int i;

  if(data == NULL)
    return CS_FAIL;

  if((name == NULL)||(value == NULL)||(numelem == NULL)||(*numelem <=0)){
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL;
  }

  memset(rdata,0,sizeof(uint8_t)* CCP_API_MAXAREA);


  //Try to get characteristic
  if(cs_a2l_GetCharacteristic(&data->ccpa2l.a2l,&ch,name)){
    _ccp_error("Failed to get characteristic %s data\n",name);
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL;
  }

  //Check element size
  if(*numelem < ch.record.size){
    _ccp_error("Number of elements does not fit, Buffer is to small\n");
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL;
  }

  if(ch.record.size > 1){
    _ccp_error("Sorry record size > 1 actually not supported\n");
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL;
  }


  for(i = 0; i < ch.record.size; i++){
    if(ch.record.rtype[i] != A2L_RECORD_TYPE_FNC){
      _ccp_error("Sorry record type actually not supported\n");
      data->lerror = CCP_ERROR_PARAMETER;
      return CS_FAIL;
    }
    
    if(ch.record.access[i] != A2L_RECORD_ACCESS_DIRECT){
      _ccp_error("Sorry record access type actually not supported\n");
      data->lerror = CCP_ERROR_PARAMETER;
      return CS_FAIL;
    }

    rsize += cs_a2l_TypeToAG(ch.record.dtype[i]) * cs_a2l_get_ag_bytes(ch.AG);
  }
  _ccp_verbose("Raw data size to upload %i Bytes\n",rsize);
  
  if(cs_ccp_api_upload_area(data,ch.address,ch.extension,rdata,rsize)){
    _ccp_error("CCP API Upload: Upload failed\n");
    return CS_FAIL;
  }
  
  _ccp_verbose("CCP API Upload: Raw buffer received:\n");
  _ccp_verbose_buffer(rdata,rsize);

  for(i = 0; i < ch.record.size; i++){

    esize = cs_a2l_TypeToAG(ch.record.dtype[i]) * cs_a2l_get_ag_bytes(ch.AG);

    if((counter + esize) > CCP_API_MAXAREA){
      _ccp_error("Invalid size data, buffer overrange\n");
      data->lerror = CCP_ERROR_PARAMETER;
      return CS_FAIL;
    }

    if(cs_a2l_convert_to_double(ch.BO,&rdata[counter],
				esize,ch.cm,ch.record.dtype[i],&value[i])){
      _ccp_error("Sorry record access type actually not supported\n");
      data->lerror = CCP_ERROR_PARAMETER;
      return CS_FAIL;
    }
    counter+=esize;
    
  }
  
  *numelem = ch.record.size;

  return CS_OK;
}

/**************************************************************************/
int cs_ccp_api_buildchksum(
			   cs_ccp_t *data,
			   uint32_t address,
			   uint8_t extension,
			   int blocksize,
			   uint32_t *checksum
			   )
/**************************************************************************/
{
  if(data == NULL)
    return CS_FAIL;

  if((blocksize <= 0)||(checksum == NULL)){
    data->lerror = CCP_ERROR_PARAMETER;
    return CS_FAIL;
  }

  if(cs_ccp_cmd_setmta(data,0,cs_a2l_CalcAddress(&data->ccpa2l.a2l,address),extension)){
    return CS_FAIL;
  }

  cs_ccp_wait_stim(data);

  if(cs_ccp_cmd_buildchksum(data,blocksize,checksum)){
    return CS_FAIL;
  }
  return CS_OK;
}
