/*  CCP CAN Stack for SocketCAN
    Copyright (C) 2015 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <byteswap.h>
#include <unistd.h>

#include "config.h"
#include "cs_defines.h"
#include "cs_calibration_utils.h"
#include "cs_socketcan.h"
#include "cs_ccp_private.h"
#include "cs_ccp_decoder.h"

/**************************************************************************/
int _ccp_decoder_append_buffer(
			       char* out,
			       int outlen,
			       uint8_t* buffer,
			       int len
			       )
/**************************************************************************/
{
  char lbuffer[CS_CHARBUF];
  char sbuffer[CS_CHARBUF];
  int i;

  if((out == NULL) ||
     (buffer == NULL)){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  memset(lbuffer,0,sizeof(char)*CS_CHARBUF);

  for(i = 0; i < len; i++){
    memset(sbuffer,0,sizeof(char)*CS_CHARBUF);
    snprintf(sbuffer,CS_CHARBUF," 0x%X",buffer[i]);
    strcat(lbuffer,sbuffer);
  }
  strcat(out,lbuffer);

  return CS_OK;
}

/**************************************************************************/
int cs_ccp_decoder_open(
			cs_ccp_decoder_t *data,
			char* filename,
			char* project,
			char *module
			)
/**************************************************************************/
{
  if((data == NULL) ||
     (filename == NULL) ||
     (project == NULL) ||
     (module == NULL)){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }

  //open A2L file
  if(cs_ccp_a2l_open(&data->ccpa2l,filename,project,module,"ASAP1B_CCP",NULL)){
    _ccp_error("Failed to open A2L\n");
    return CS_FAIL;
  }
  
  data->master_cmd = CCP_CMD_NONE;
  data->data_len = 0;
  
  return CS_OK;
}

/**************************************************************************/
int cs_ccp_decoder_buffer(cs_ccp_decoder_t *data,
			  double ts,
			  uint32_t canid,
			  uint8_t * buffer,
			  int len,
			  char *out,
			  int outlen
			  )
/**************************************************************************/
{
  cs_cu_data16_t l16;
  cs_cu_data32_t l32;
  
  if((data == NULL) ||
     (buffer == NULL) ||
     (out == NULL) ||
     (outlen == 0)){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }

  memset(out,0,sizeof(char)*outlen);

  if(canid == data->ccpa2l.can_send_id){
    if(len < 1){
      _ccp_error("Invalid data length\n");
      return CS_FAIL;
    }
    //_ccp_verbose_buffer(buffer,len);
    switch(buffer[0]){
    case CCP_CMD_CONNECT:
      if(len < 4){
	_ccp_error("CCP Decoder Connect: Invalid data length\n");
	return CS_FAIL;
      }
      memcpy(l16.u8,&buffer[2],2);
      snprintf(out,outlen,"%f: Master Connect WC: %u Station id: 0x%X \n",
	       ts,
	       buffer[1],
	       cs_cu_swap16(A2L_BO_INTEL,l16.u16));
      data->master_cmd = CCP_CMD_CONNECT;
      break;
    case CCP_CMD_EXCHANGE_ID:
      if(len < 2){
	_ccp_error("CCP Decoder Exchange ID: Invalid data length\n");
	return CS_FAIL;
      }
      snprintf(out,outlen,"%f: Master Exchange ID WC: %u \n",
	       ts,
	       buffer[1]);
      data->master_cmd = CCP_CMD_EXCHANGE_ID;
      break;
    case CCP_CMD_GETSSTATUS:
      if(len < 2){
	_ccp_error("CCP Decoder GetSStatus: Invalid data length\n");
	return CS_FAIL;
      }
      snprintf(out,outlen,"%f: Master GetSStatus WC: %u \n",
	       ts,
	       buffer[1]);
      data->master_cmd = CCP_CMD_GETSSTATUS;
      break;
    case CCP_CMD_SETMTA:
      if(len < 8){
	_ccp_error("CCP Decoder SetMTA: Invalid data length\n");
	return CS_FAIL;
      }
      memcpy(l32.u8,&buffer[4],4);
      snprintf(out,outlen,"%f: Master SetMTA WC: %u MTANo: 0x%X Adr-Ext: 0x%X Adr: 0x%X\n",
	       ts,
	       buffer[1],
	       buffer[2],
	       buffer[3],
	       cs_cu_swap32(data->ccpa2l.BO,l32.u32));
      data->master_cmd = CCP_CMD_SETMTA;
      break;
    case CCP_CMD_DNLOAD:
      if(len < 3){
	_ccp_error("CCP Decoder Dnload: Invalid data length\n");
	return CS_FAIL;
      }
      snprintf(out,outlen,"%f: Master Dnload WC: %u Data length %i Data:",
	       ts,
	       buffer[1],
	       buffer[2]);
      if(len < (3 + buffer[2])){
	_ccp_error("CCP Decoder Dnload: Invalid data length\n");
	return CS_FAIL;
      }
      _ccp_decoder_append_buffer(out,outlen,&buffer[3],buffer[2]);
      strcat(out,"\n");
      data->master_cmd = CCP_CMD_DNLOAD;
      break;
    case CCP_CMD_UPLOAD:
      if(len < 3){
	_ccp_error("CCP Decoder Upload: Invalid data length\n");
	return CS_FAIL;
      }
      snprintf(out,outlen,"%f: Master Upload WC: %u Data length %i\n",
	       ts,
	       buffer[1],
	       buffer[2]);
      data->data_len = buffer[2];
      data->master_cmd = CCP_CMD_UPLOAD;
      break;
    case CCP_CMD_GETDAQSIZE:
      if(len < 8){
	_ccp_error("CCP Decoder GetDAQSize: Invalid data length\n");
	return CS_FAIL;
      }
      memcpy(l32.u8,&buffer[4],4);
      snprintf(out,outlen,"%f: Master GetDAQSize WC: %u DAQ: 0x%X CAN-Id: 0x%X\n",
	       ts,
	       buffer[1],
	       buffer[2],
	       cs_cu_swap32(data->ccpa2l.BO,l32.u32));
      data->master_cmd = CCP_CMD_GETDAQSIZE;
      break;
    case CCP_CMD_SETDAQPTR:
      if(len < 5){
	_ccp_error("CCP Decoder SetDAQPtr: Invalid data length\n");
	return CS_FAIL;
      }
      snprintf(out,outlen,"%f: Master SetDAQPtr WC: %u DAQ: 0x%X ODT: 0x%X Entry: 0x%X\n",
	       ts,
	       buffer[1],
	       buffer[2],
	       buffer[3],
	       buffer[4]);
      data->master_cmd = CCP_CMD_SETDAQPTR;
      break;
    case CCP_CMD_WRITEDAQ:
      if(len < 8){
	_ccp_error("CCP Decoder WriteDAQ: Invalid data length\n");
	return CS_FAIL;
      }
      memcpy(l32.u8,&buffer[4],4);
      snprintf(out,outlen,"%f: Master WriteDAQ WC: %u Size: 0x%X Addr-Ext: 0x%X Address: 0x%X\n",
	       ts,
	       buffer[1],
	       buffer[2],
	       buffer[3],
	       cs_cu_swap32(data->ccpa2l.BO,l32.u32));
      data->master_cmd = CCP_CMD_WRITEDAQ;
      break;
    case CCP_CMD_STARTSTOP:
      if(len < 8){
	_ccp_error("CCP Decoder StartStop: Invalid data length\n");
	return CS_FAIL;
      }
      memcpy(l16.u8,&buffer[6],2);
      snprintf(out,outlen,"%f: Master StartStop WC: %u Mode: 0x%X DAQ: 0x%X Last ODT: 0x%X Event: 0x%X Prescaler: 0x%X\n",
	       ts,
	       buffer[1],
	       buffer[2],
	       buffer[3],
	       buffer[4],
	       buffer[5],
	       cs_cu_swap16(data->ccpa2l.BO,l16.u16));
      data->master_cmd = CCP_CMD_SETDAQPTR;
      break;
    case CCP_CMD_DISCONNECT:
      if(len < 6){
	_ccp_error("CCP Decoder Disconnect: Invalid data length\n");
	return CS_FAIL;
      }
      memcpy(l16.u8,&buffer[4],2);
      snprintf(out,outlen,"%f: Master Disconnect WC: %u EndofSession: %s Station id: 0x%X \n",
	       ts,
	       buffer[1],
	       buffer[2]==1?"End":"Temporary",
	       cs_cu_swap16(cs_cu_endianess(),l16.u16));
      data->master_cmd = CCP_CMD_DISCONNECT;
      break;
    case CCP_CMD_BUILDCHKSUM:
      if(len < 6){
	_ccp_error("CCP Decoder BuildChecksum: Invalid data length\n");
	return CS_FAIL;
      }
      memcpy(l32.u8,&buffer[2],4);
      snprintf(out,outlen,"%f: Master BuildChecksum WC: %u Blocksize: 0x%X\n",
	       ts,
	       buffer[1],
	       cs_cu_swap32(data->ccpa2l.BO,l32.u32));
      data->master_cmd = CCP_CMD_BUILDCHKSUM;
      break;
    case CCP_CMD_CCPVERSION:
      if(len < 4){
	_ccp_error("CCP Decoder GetCCPVersion: Invalid data length\n");
	return CS_FAIL;
      }
      snprintf(out,outlen,"%f: Master GetCCPVersion WC: %u Main Protocol Version: 0x%X Release Version: 0x%X\n",
	       ts,
	       buffer[1],
	       buffer[2],
	       buffer[3]);
      data->master_cmd = CCP_CMD_CCPVERSION;
      break;
    case CCP_CMD_TEST:
      if(len < 4){
	_ccp_error("CCP Decoder Test: Invalid data length\n");
	return CS_FAIL;
      }
      memcpy(l16.u8,&buffer[2],2);
      snprintf(out,outlen,"%f: Master Test WC: %u Station Id: 0x%X\n",
	       ts,
	       buffer[1],
	       cs_cu_swap16(A2L_BO_INTEL,l16.u16));
      data->master_cmd = CCP_CMD_TEST;
      break;
    default:
      snprintf(out,outlen,"%f: Master Unknown cmd: 0x%X\n",ts,buffer[0]);
      break;
    }
    return CS_OK;
  }
  if(canid == data->ccpa2l.can_recv_id){ //Slave answer, Events or DAQ frames
    //_ccp_verbose_buffer(buffer,len);
    switch(buffer[0]){
    case CCP_DTO_CRM:
      switch(data->master_cmd){
      case CCP_CMD_NONE:
	return CS_FAIL;
	break;
      case CCP_CMD_CONNECT:
      case CCP_CMD_SETMTA:
      case CCP_CMD_SETDAQPTR:
      case CCP_CMD_DNLOAD:
      case CCP_CMD_WRITEDAQ:
      case CCP_CMD_STARTSTOP:
      case CCP_CMD_DISCONNECT:
      case CCP_CMD_TEST:
	if(len < 3){
	  _ccp_error("CCP Slave Ack: Invalid data length\n");
	  return CS_FAIL;
	}
	snprintf(out,outlen,"%f: Slave Ack RC: 0x%X WC: 0x%X\n",
		 ts,
		 buffer[1],
		 buffer[2]);
	data->master_cmd = CCP_CMD_NONE;
	break;
      case CCP_CMD_UPLOAD:
	if(len < 3){
	  _ccp_error("CCP Slave Upload: Invalid data length\n");
	  return CS_FAIL;
	}
	snprintf(out,outlen,"%f: Slave Upload RC: 0x%X WC: 0x%X Data:",
		 ts,
		 buffer[1],
		 buffer[2]);
	if(len < (3 + data->data_len)){
	  _ccp_error("CCP Slave Upload: Invalid data length\n");
	  return CS_FAIL;
	}
	_ccp_decoder_append_buffer(out,outlen,&buffer[3],data->data_len);
	strcat(out,"\n");
	data->master_cmd = CCP_CMD_NONE;
	data->data_len = 0;
	break;
      case CCP_CMD_GETDAQSIZE:
	if(len < 5){
	  _ccp_error("CCP Slave GetDAQSize: Invalid data length\n");
	  return CS_FAIL;
	}
	snprintf(out,outlen,"%f: Slave GetDAQSize RC: 0x%X WC: %u DAQ-Size: 0x%X PID: 0x%X\n",
		 ts,
		 buffer[1],
		 buffer[2],
		 buffer[3],
		 buffer[4]);
	data->master_cmd = CCP_CMD_NONE;
	break;
      case CCP_CMD_GETSSTATUS:
	if(len < 4){
	  _ccp_error("CCP Slave GetSStatus: Invalid data length\n");
	  return CS_FAIL;
	}
	snprintf(out,outlen,"%f: Slave GetSStatus RC: 0x%X WC: %u Status: 0x%X\n",
		 ts,
		 buffer[1],
		 buffer[2],
		 buffer[3]);
	data->master_cmd = CCP_CMD_NONE;
	break;
      case CCP_CMD_EXCHANGE_ID:
	if(len < 7){
	  _ccp_error("CCP Slave Exchange ID: Invalid data length\n");
	  return CS_FAIL;
	}
	snprintf(out,outlen,"%f: Slave Exchange ID RC: 0x%X WC: %u Length: 0x%X Qualifier: 0x%X ResAM: 0x%X, ResPM: 0x%X\n",
		 ts,
		 buffer[1],
		 buffer[2],
		 buffer[3],
		 buffer[4],
		 buffer[5],
		 buffer[6]);
	data->master_cmd = CCP_CMD_NONE;
	break;
      case CCP_CMD_BUILDCHKSUM:
	if(len < 4){
	  _ccp_error("CCP Slave BuildChecksum: Invalid data length\n");
	  return CS_FAIL;
	}
	memcpy(l32.u8,&buffer[4],4);
	snprintf(out,outlen,"%f: Slave BuildChecksum RC: 0x%X WC: %u Length: 0x%X Checksum:",
		 ts,
		 buffer[1],
		 buffer[2],
		 buffer[3]);
	if(len < (4+ buffer[3])){
	  _ccp_error("CCP Slave BuildChecksum: Invalid data length\n");
	  return CS_FAIL;
	}
	_ccp_decoder_append_buffer(out,outlen,&buffer[4],buffer[3]);
	strcat(out,"\n");
	data->master_cmd = CCP_CMD_NONE;
	break;
      case CCP_CMD_CCPVERSION:
	if(len < 5){
	  _ccp_error("CCP Slave GetCCPVersion: Invalid data length\n");
	  return CS_FAIL;
	}
	snprintf(out,outlen,"%f: Slave GetCCPVersion RC: 0x%X WC: %u Main Protocol Version: 0x%X Release Version: 0x%X\n",
		 ts,
		 buffer[1],
		 buffer[2],
		 buffer[3],
		 buffer[4]);
	data->master_cmd = CCP_CMD_NONE;
	break;
      }
      break;
    case CCP_DTO_EV:
      snprintf(out,outlen,"%f: Slave Receive event frame\n",
	       ts);
      break;
    default:
      //snprintf(out,outlen,"%f: Slave Receive DAQ frame\n",
      //ts);
      break;
    }
    return CS_OK;
  }

  return CS_FAIL;
}

/**************************************************************************/
int cs_ccp_decoder_close(
			 cs_ccp_decoder_t *data
			 )
/**************************************************************************/
{
  if(data == NULL){
    _ccp_error("Parameter failure\n");
    return CS_FAIL;
  }

  cs_ccp_a2l_close(&data->ccpa2l);

  return CS_OK;
}
