/*  cs_dbc.c --  encode and decode CAN messages signals
    Copyright (C) 2013 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <locale.h>


#include "cs_dbc_private.h"
#include "cs_dbc.h"
#include "cs_signal_private.h"
#include "cs_message_private.h"
#include "cs_parser.h"
#include "cs_lists.h"
#include "cs_debug_private.h"

/** \def SCANBUFFER
    Max line length in DBC file
*/
#define SCANBUFFER 1024


#define _dbc_error(...) (cs_error(CS_DEBUG_DBC,__func__,__LINE__,__VA_ARGS__))
#define _dbc_warn(...) (cs_warn(CS_DEBUG_DBC,__func__,__LINE__,__VA_ARGS__))
#define _dbc_info(...) (cs_info(CS_DEBUG_DBC,__func__,__LINE__,__VA_ARGS__))
#define _dbc_debug(...) (cs_debug(CS_DEBUG_DBC,__func__,__LINE__,__VA_ARGS__))
#define _dbc_verbose(...) (cs_verbose(CS_DEBUG_DBC,__func__,__LINE__,__VA_ARGS__))
#define _dbc_error_buffer(DATA,LEN) (cs_error_buffer(CS_DEBUG_DBC,__func__,__LINE__,DATA,LEN))
#define _dbc_warn_buffer(DATA,LEN) (cs_warn_buffer(CS_DEBUG_DBC,__func__,__LINE__,DATA,LEN))
#define _dbc_info_buffer(DATA,LEN) (cs_info_buffer(CS_DEBUG_DBC,__func__,__LINE__,DATA,LEN))
#define _dbc_debug_buffer(DATA,LEN) (cs_debug_buffer(CS_DEBUG_DBC,__func__,__LINE__,DATA,LEN))
#define _dbc_verbose_buffer(DATA,LEN) (cs_verbose_buffer(CS_DEBUG_DBC,__func__,__LINE__,DATA,LEN))


/********************************************************************/
cs_debug_level_t cs_dbc_debuglevel(
				   cs_debug_level_t level
				   )
/********************************************************************/
{
  return cs_debuglevel(CS_DEBUG_DBC,level);
}

/********************************************************************/
void cs_dbc_debug_setcallback(
    void *ctx,
    void (*callb)(void *ctx,
		  cs_debug_sub_t sub,
		  cs_debug_level_t dl,
		  const char *func,
		  const int line,
		  const char *fmt,
		  va_list ap)
    )
/********************************************************************/
{
  cs_debug_setcallback(CS_DEBUG_DBC,ctx,callb);
}

/********************************************************************/
void * _dbc_cpf(void *ext)
/********************************************************************/
{
    return ext;
}

/********************************************************************/
void _dbc_free_generic(void *n)
/********************************************************************/
{
    if(n == NULL)
        return;

    free(n);	     
}

/********************************************************************/
void _dbc_message_free_generic(void *n)
/********************************************************************/
{
    if(n == NULL)
        return;

    cs_message_free((cs_message_t *)n);	     
}

/********************************************************************/
void _dbc_signal_free_generic(void *n)
/********************************************************************/
{
    if(n == NULL)
        return;

    cs_signal_free((cs_signal_t *)n);	     
}


/********************************************************************/
cs_dbc_node_t *_dbc_node_init()
/********************************************************************/
{
    cs_dbc_node_t *data = NULL;
    
    data = (cs_dbc_node_t*)calloc(1,sizeof(cs_dbc_node_t));
    if(data == NULL){
      _dbc_error("Malloc failure\n");
      return NULL;
    }

    if(cs_lists_init(&data->genlist,NULL,NULL) ||
       cs_lists_init(&data->reqlist,NULL,NULL)
       ){
        _dbc_error("Failed to init lists\n");
        cs_lists_destroy(&data->genlist);
        cs_lists_destroy(&data->reqlist);
        free(data);
        return NULL;
    }
  
    return data;
}

/********************************************************************/
void _dbc_node_free(
		    cs_dbc_node_t *sk          /**< Node storage structure*/
		    )
/********************************************************************/
{
  if(sk == NULL){
    _dbc_error("Parameter failure\n");
    return;
  }
  
  cs_lists_destroy(&sk->genlist);
  cs_lists_destroy(&sk->reqlist);
  free(sk);
  sk = NULL;
}

/********************************************************************/
void _dbc_free_list_node(void *n)
/********************************************************************/
{
  if(n == NULL){
    _dbc_error("Parameter failure\n");
    return;
  }	
  
  _dbc_node_free((cs_dbc_node_t *)n);
    
}

/**************************************************************************/
int _dbc_addnode(
		 cs_dbc_t *data,
		 char *name
		 )
/**************************************************************************/
{
  if(data == NULL)
    return CS_FAIL;
  
  if(name == NULL)
    return CS_FAIL;

  if(strlen(name) >= CS_CHARBUF)
    return CS_FAIL;

  cs_dbc_node_t *n = _dbc_node_init();

  if(n == NULL){
    return CS_FAIL;
  }
  
  strncpy(n->name,name,CS_CHARBUF);

  cs_lists_insert(&data->nlist,(void*)n);

  return CS_OK;
}

/**************************************************************************/
cs_dbc_node_t* cs_dbc_getnode_byname(
				     cs_dbc_t *data,    /**< DBC structure*/
				     char *name         /**< Node name*/
				     )
/**************************************************************************/
{
  cs_dbc_node_t *n = NULL;
  int i;

  if((data == NULL) || (name == NULL)){
    _dbc_error("Parameter failure\n");
    return NULL;
  }
  
  for(i = 0; i < cs_dbc_getnonodes(data); i++){
    n = (cs_dbc_node_t*)cs_lists_get_iter(&data->nlist,i);
    if(strcmp(name,n->name)==0){
      return n;
    }
  }
  return NULL;
}

/**************************************************************************/
int _dbc_node_add_requested_message(
				    cs_dbc_t *data,    /**< DBC structure*/
				    char *name,        /**< Node name*/
				    cs_message_t *m    /**< Message pointer*/
				    )
/**************************************************************************/
{
  cs_dbc_node_t *n = NULL;

  if ((n =cs_dbc_getnode_byname(data,name)) == NULL){
    _dbc_warn("Failed to get node\n");
    return CS_FAIL;
  }
  cs_lists_insert(&n->reqlist,(void*)m);
  return CS_OK;
}


/**************************************************************************/
int _dbc_node_add_generated_message(
				    cs_dbc_t *data,    /**< DBC structure*/
				    char *name,        /**< Node name*/
				    cs_message_t *m    /**< Message pointer*/
				    )
/**************************************************************************/
{
  cs_dbc_node_t *n = NULL;

  if ((n =cs_dbc_getnode_byname(data,name)) == NULL){
    _dbc_warn("Failed to get node\n");
    return CS_FAIL;
  }
  cs_lists_insert(&n->genlist,(void*)m);
  return CS_OK;
}


/**************************************************************************/
int cs_dbc_node_no_requested_messages(
				      cs_dbc_node_t *data
				      )
/**************************************************************************/
{
  if(data == NULL){
    _dbc_error("Parameter failure\n");
    return -1;
  }
    
  return cs_lists_size(&data->reqlist);
}

/**************************************************************************/
cs_message_t *cs_dbc_node_get_requested_message_iter(
						     cs_dbc_node_t *data,
						     int index
						     )
/**************************************************************************/
{
  if(data == NULL){
    _dbc_error("Parameter failure\n");
    return NULL;
  }
    
  if((index < 0) || (index >= cs_lists_size(&data->reqlist))){
    _dbc_error("Index out of range\n");
    return NULL;
  }
  
  return (cs_message_t*)cs_lists_get_iter(&data->reqlist,index);
}

/**************************************************************************/
int cs_dbc_node_no_generated_messages(
				      cs_dbc_node_t *data
				      )
/**************************************************************************/
{
  if(data == NULL){
    _dbc_error("Parameter failure\n");
    return -1;
  }
  
  return cs_lists_size(&data->genlist);
}

/**************************************************************************/
cs_message_t *cs_dbc_node_get_generated_message_iter(
						     cs_dbc_node_t *data,
						     int index
						     )
/**************************************************************************/
{ 
  if(data == NULL){
    _dbc_error("Parameter failure\n");
    return NULL;
  }
  
  if((index < 0) || (index >= cs_lists_size(&data->genlist))){
    _dbc_error("Index out of range\n");
    return NULL;
  }
  
  return (cs_message_t*)cs_lists_get_iter(&data->genlist,index);
}


/**************************************************************************/
int cs_dbc_getnonodes(
		      cs_dbc_t *data    /**< DBC structure*/
		      )
/**************************************************************************/
{
  
  if(data == NULL){
    _dbc_error("Parameter failure\n");
    return -1;
  }
  
  return cs_lists_size(&data->nlist);
}

/**************************************************************************/
cs_dbc_node_t *cs_dbc_getnode_iter(
				   cs_dbc_t *data,    /**< DBC structure*/
				   int index         /**< Node index*/
				   )
/**************************************************************************/
{
  cs_dbc_node_t *n = NULL;
  
  if(data == NULL){
    _dbc_error("Parameter failure\n");
    return NULL;
  }

  if((index < 0) || (index >= cs_lists_size(&data->nlist))){
    _dbc_error("Index out of range\n");
    return NULL;
  }
    
  n = (cs_dbc_node_t*)cs_lists_get_iter(&data->nlist,index);
  
  return n;
}

/**************************************************************************/
char *cs_dbc_node_get_name(
			   cs_dbc_node_t *data
			   )
/**************************************************************************/
{
  if(data == NULL){
    _dbc_error("Parameter failure\n");
    return NULL;
  }

  return data->name;
}


/**************************************************************************/
cs_message_t * cs_dbc_getmessage(
		      cs_dbc_t *dbc,
		      uint32_t id
		      )
/**************************************************************************/
{
  struct cs_lists_entry *entry = NULL;
  cs_message_t *dm = NULL;
  
  if(dbc == NULL){
    _dbc_error("Parameter failure\n");
    return NULL;
  }
  
  for(entry = cs_lists_get_first(&dbc->mlist);
      entry != cs_lists_get_last(&dbc->mlist);
      entry = cs_lists_get_next(entry)){
    dm = (cs_message_t*)cs_lists_get_data(entry);
    if(dm->id == id){
      return dm;
    }
  }
  return NULL;
}

/**************************************************************************/
cs_signal_t *cs_dbc_getsignal(
			      cs_dbc_t *dbc,
			      char *slname
			      )
/**************************************************************************/
{
    struct cs_lists_entry *entry = NULL;
    cs_signal_t *s = NULL;
    cs_message_t *dm = NULL;
    

    if(dbc == NULL){
      _dbc_error("Parameter failure\n");
      return NULL;
    }
  
    if(slname == NULL){
        _dbc_error("Invalid signal name pointer\n");
        return NULL;
    }

    for(entry = cs_lists_get_first(&dbc->mlist);
        entry != cs_lists_get_last(&dbc->mlist);
        entry = cs_lists_get_next(entry)){
        dm = (cs_message_t*)cs_lists_get_data(entry);

	if((s = cs_message_get_signal_byname(dm,slname)) != NULL){
	  return s;
	}
    }
    
    return NULL;
}


/**************************************************************************/
int cs_dbc_getnummessages(cs_dbc_t *dbc)
/**************************************************************************/
{
  if(dbc == NULL){
    _dbc_error("Parameter failure\n");
    return -1;
  }
  
  return cs_lists_size(&dbc->mlist);
}


/**************************************************************************/
cs_message_t* cs_dbc_getmessage_iter(cs_dbc_t *dbc, int iter)
/**************************************************************************/
{
  if(dbc == NULL){
    _dbc_error("Parameter failure\n");
    return NULL;
  }
  
  return (cs_message_t*)cs_lists_get_iter(&dbc->mlist,iter);
}

/********************************************************************/
cs_dbc_t *cs_dbc_init()
/********************************************************************/
{
  cs_dbc_t *data = NULL;
  
  data = (cs_dbc_t*)calloc(1,sizeof(cs_dbc_t));
  if(data == NULL){
    _dbc_error("Malloc failure\n");
    return NULL;
  }
    
  if(cs_lists_init(&data->nlist,_dbc_cpf,_dbc_free_list_node) ||
     cs_lists_init(&data->mlist,_dbc_cpf,_dbc_message_free_generic)
     ){
    _dbc_error("Failed to init lists\n");
    cs_lists_destroy(&data->nlist);
    cs_lists_destroy(&data->mlist);
    free(data);
    return NULL;
  }
  
  return data;
}

/********************************************************************/
void cs_dbc_free(
		 cs_dbc_t *sk          /**< DBC storage structure*/
		 )
/********************************************************************/
{
  if(sk == NULL)
    return;
  
  cs_lists_destroy(&sk->nlist);
  cs_lists_destroy(&sk->mlist);
  free(sk);
  sk = NULL;
}

/** Scan for signal labels
      \retval 0 No failure
      \retval 1 Failure
*/
/********************************************************************/
int _dbc_scan_signal_label(
			   cs_signal_t *s, /**< Signal storage structure*/
			   char *line      /**< DBC line to parse*/
			   )
/********************************************************************/
{
  /*
    Parse for something like
    VAL_ 148 AFS_Personality_Desired 7 "unused" 6 "unused" 5 "unused" 4 "unused" 3 "Tow/Haul" 2 "Sport" 1 "Normal" 0 "Comfort"\n
    Values a seperated by space
    A value label pair is defined by format %f "label"
    First we scan for the opening " and closing "
   */
  int i;
  int counter = 0;
  double value;
  int state = 0;
  //State = 0 => Initial whitespace offset search
  //State = 1 => Scanf
  //State = 2 => Search for opening "
  //State = 3 => Search for closing "
  //State = 4 => Search for ' '
  //State = 5 => finished
  char label[SCANBUFFER];
  
  for(i = 0; i < strlen(line); i++) {
    //printf("State: %d: %c\n",state,line[i]);
    switch(state){
    case 0:
      //First ommit number of spaces before start parsing
      if(strncmp((const char *)&line[i], " ", 1) == 0){
	counter++;
      }
      if(counter == 3){
	state = 1;
	break;
      }
      break;
    case 1:
      //Scanf
      if (sscanf(&line[i], "%lf \"%[^\"]s\"", &value, label) == 2 ){
	if(cs_signal_add_label(s,label,value)){
	  _dbc_error("Failed to add label %s for signal %s\n",label,s->name);
	  return CS_FAIL;
	}
	_dbc_verbose("Add label %s:%.3f for signal %s\n",label,value,s->name);
	state = 2;
	break;
      } else {
	//No more to scan
	state = 5;
      }
    case 2:
      //Scan for opening "
      if(strncmp((const char *)&line[i], "\"", 1) == 0){
	state = 3;
      }
      break;
    case 3:
      //Scan for closing "
      if(strncmp((const char *)&line[i], "\"", 1) == 0){
	state = 4;
      }
      break;
     case 4:
      if(strncmp((const char *)&line[i], " ", 1) == 0){
	state = 1;
      }
      break;
    case 5:
      //Scan is finished
      return CS_OK;
    }
  }
  return CS_OK;
}

/** Scan for signal labels
      \retval 0 No failure
      \retval 1 Failure
*/
/********************************************************************/
int _dbc_scan_signal_receiver(
			      cs_signal_t *s, /**< Signal storage structure*/
			      char *line      /**< DBC line to parse*/
			      )
/********************************************************************/
{
  /*
    Parse for something like
    SG_ Supply_Voltage : 15|16@0+ (0.01,0) [0|100] "V"  ECU02,ECU01,ECU03
    A value label pair is defined by format %f "label"
    First we scan for the opening " and closing " and for whitespaces
   */
  int i;
  int state = 0;
  //State = 0 => Search for opening "
  //State = 1 => Search for closing "
  //State = 2 => Search for ' '
  //State = 3 => scanf
  //State = 4 => Search for ,
  //State = 5 => finished
  char name[SCANBUFFER];
  
  for(i = 0; i < strlen(line); i++) {
    //printf("State: %d: %c\n",state,line[i]);
    switch(state){
    case 0:
      //Scan for opening "
      if(strncmp((const char *)&line[i], "\"", 1) == 0){
	state = 1;
      }
      break;
    case 1:
      //Scan for closing "
      if(strncmp((const char *)&line[i], "\"", 1) == 0){
	state = 2;
      }
      break;
    case 2:
      //Scan for whitespaces
      if(strncmp((const char *)&line[i], " ", 1) == 0){
	state = 2;
	break;
      } else {
	state = 3;
	//Note, no break, direct start of scanf
      }
    case 3:
      //Scanf, Receiver are separates by , without whitespace
      if (sscanf(&line[i], "%[^, \n\r]", name) == 1 ){
	if(cs_signal_add_node(s,name)){
	  _dbc_error("Failed to add receiver %s for signal %s\n",name,s->name);
	  //return CS_FAIL;
	}
	_dbc_verbose("Add receiver %s for signal %s\n",name,s->name);
	state = 4;
	//Increment i by string length
	i+=strlen(name)-1;
      } else {
	//No more to scan
	state = 5;
      }
      break;
    case 4:
      //Scan for ","
      if(strncmp((const char *)&line[i], ",", 1) == 0){
	state = 3;
      } else {
	//state = 5;
	//Finished
      }
      break;
    case 5:
      //Scan is finished
      return CS_OK;
    }
  }
  return CS_OK;
}


/** Scan for bus units
      \retval 0 No failure
      \retval 1 Failure
*/
/********************************************************************/
int _dbc_scan_busunits(
		      cs_dbc_t *d, /**< DBC storage structure*/
		      char *line      /**< DBC line to parse*/
		      )
/********************************************************************/
{
  /*
    Parse for something like
    BU_: ECU02 ECU01 ECU03
   */
  int i;
  int state = 0;
  //State = 0 => Search for opening "BU_:"
  //State = 1 => Search for ' '
  //State = 2 => scanf
  //State = 3 => finished
  char name[SCANBUFFER];
  
  for(i = 0; i < strlen(line); i++) {
    //printf("State: %d: %c\n",state,line[i]);
    switch(state){
    case 0:
      //Scan for opening "BU_:"
      if(strncmp((const char *)&line[i], "BU_:", 4) == 0){
	state = 1;
      }
      break;
    case 1:
      //Scan for " "
      if(strncmp((const char *)&line[i], " ", 1) == 0){
	state = 2;
      }
      break;
    case 2:
      //Scanf
      if (sscanf(&line[i], "%[^ \n\r]", name) == 1 ){
	if(_dbc_addnode(d,name)){
	  _dbc_error("Failed to add node %s\n",name);
	  //return CS_FAIL;
	}
	_dbc_verbose("Add node %s\n",name);
	state = 1;
      } else {
	//No more to scan
	state = 3;
      }
      break;
    case 3:
      //Scan is finished
      return CS_OK;
    }
  }
  return CS_OK;
}

/********************************************************************/
int _dbc_build_dependency(
			  cs_dbc_t *data               /**< DBC storage structure*/
			  )
/********************************************************************/
{
  //Loop nodes
  cs_dbc_node_t* n = NULL;
  cs_message_t* m = NULL;
  cs_message_t* rm = NULL;
  cs_signal_t* s = NULL;
  cs_signal_node_t *sn = NULL;
  
  int i,j,k,l,r;
  uint8_t found;

  for(i = 0; i < cs_lists_size(&data->nlist); i++){
    n = (cs_dbc_node_t*)cs_lists_get_iter(&data->nlist,i);
    //printf("Loop node %s\n",n->name);
    //Loop messages
    for(j = 0; j < cs_lists_size(&data->mlist); j++){
      m = (cs_message_t*)cs_lists_get_iter(&data->mlist,j);
      if(strcmp(n->name,m->producer) == 0){
	_dbc_verbose("Node %s produces message %s (0x%X)\n",n->name,m->name,m->id);
	cs_lists_insert(&n->genlist,(void*)m);	
      }
      //loop all signals
      for(k = 0; k < cs_lists_size(&m->listsignal); k++){
	s = (cs_signal_t*)cs_lists_get_iter(&m->listsignal,k);
	//Loop all signal requesting nodes
	for(l = 0; l < cs_lists_size(&s->reqnodes); l++){
	  sn = (cs_signal_node_t*)cs_lists_get_iter(&s->reqnodes,l);
	  //printf("Signal %s Reqnode %s\n",s->name,sn->name);
	  if(strcmp(sn->name,n->name)==0){
	    //This node requests this signal, check if message is already inside list
	    found =0;
	    for(r = 0; r < cs_lists_size(&n->reqlist); r++){
	      rm = (cs_message_t*)cs_lists_get_iter(&n->reqlist,r);
	      if(strcmp(m->name,rm->name)==0){
		found = 1;
		break;
	      }
	    }
	    if(found == 0){
	      cs_lists_insert(&n->reqlist,(void*)m);
	      _dbc_verbose("Node %s requests message %s (0x%X)\n",n->name,m->name,m->id);
	      //We can stop for this message, because it is already requested
	      break;
	    }
	  }
	}
      }
    }
  }
  return CS_OK;
}

/********************************************************************/
int cs_dbc_read(
    cs_dbc_t *data,               /**< DBC storage structure*/
    char *filename                /**< DBC file path and name*/
    )
/********************************************************************/
{
    char line[SCANBUFFER];
    FILE *fp;
    char mname[SCANBUFFER];
    int mid;
    int mdlc;
    char msender[SCANBUFFER];
    char sname[SCANBUFFER];
    char smux[SCANBUFFER];
    char content[SCANBUFFER];
    int sstart;
    int slen;
    int sbt;
    char ssigned;
    int genmsgcycletime;
    int genmsgdelaytime;
    int intvalue;
    double sscale;
    double soffset;
    double smin;
    double smax;
    char sunit[SCANBUFFER];
    int rv;
    char *locale_store = NULL;
  
    cs_message_t *m = NULL;
    cs_signal_t *s = NULL;
   
    if(data == NULL){
      _dbc_error("No data handle defined\n");
      return CS_FAIL;
    }

  
    if(filename == NULL){
        _dbc_error("No filename defined\n");
        return CS_FAIL;
    }


    _dbc_verbose("Try to open DBC file: %s\n",filename);
  
    fp = fopen(filename,"r");
  
    if(fp == NULL){
      _dbc_error("Error opening  %s: %s\n", filename,strerror(errno));
        return CS_FAIL;
    }

    //We have to set the local to 
    locale_store = setlocale(LC_ALL,NULL);
    setlocale(LC_ALL,"C");
    //Note first we only scan for messages and signals
    //all other stuff will be handled in a second run
    while(fgets(line,SCANBUFFER-1,fp)){
        memset(mname,0,sizeof(char)*SCANBUFFER);
        memset(msender,0,sizeof(char)*SCANBUFFER);
        memset(sname,0,sizeof(char)*SCANBUFFER);
        memset(sunit,0,sizeof(char)*SCANBUFFER);
        memset(smux,0,sizeof(char)*SCANBUFFER);
        sscale = 1.0;
        soffset = 0.0;
        smin = 0.0;
        smax = 0.0;
	if(strncmp("BU_:",line,4)==0){
	  if(_dbc_scan_busunits(data,line)){
	    _dbc_error("Failed to scan bus units\n");
	    goto errorhandling;
	  }
        } else if((rv = sscanf(line," BO_ %d %[^:]: %d %s",&mid,mname,&mdlc,msender)) == 4){
	  if(((m = cs_message_init())== NULL)||
	     cs_message_set(m,mid,mdlc,mname,msender,0) ||
	     cs_lists_insert(&data->mlist,(void*)m)){
	    _dbc_error("Failed to init message\n");
	    goto errorhandling;
	  }
	  _dbc_verbose("Add message (0x%X)%s DLC:%d\n",mid,mname,mdlc);
	  continue;
        } else if((rv = sscanf(line," SG_ %s : %d|%d@%d%c (%lf,%lf) [%lf|%lf] \"%[^\"]s\"",sname, &sstart, &slen,&sbt, &ssigned, &sscale, &soffset, &smin, &smax, sunit)) > 5){
	  s = cs_signal_init();
	  if(s == NULL){
	    _dbc_error("Signal init failure\n");
	    goto errorhandling;
	  }
	  if(cs_signal_set_name(s,sname)){
	    _dbc_error("Failed to set signal name %s\n",sname);
	    goto errorhandling;
	  }
	  if(
	     cs_signal_set_signal_type(s,(ssigned=='-')?SIGNAL_TYPE_SIGNED:SIGNAL_TYPE_UNSIGNED)||
	     cs_signal_set_byteorder(s,(sbt==0)?SIGNAL_BT_BIG:SIGNAL_BT_LITTLE)||
	     cs_signal_set_bitoffset(s,sstart)||
	     cs_signal_set_bitlength(s,slen)||
	     cs_signal_set_unit(s,sunit)||
	     cs_signal_set_scale(s,sscale)||
	     cs_signal_set_offset(s,soffset)||
	     cs_signal_set_minmax(s,smin,smax)||
	     _dbc_scan_signal_receiver(s,line)||
	     cs_message_add_signal(m,s)
	     ){
	    _dbc_error("Failed to add signal %s\n",sname);
	  } else {
	    _dbc_verbose("Add signal %s Start:%d length: %d\n",sname,sstart,slen);
	  }
	  continue;
        } else if((rv = sscanf(line," SG_ %s %s : %d|%d@%d%c (%lf,%lf) [%lf|%lf] \"%[^\"]s\"",sname, smux, &sstart, &slen,&sbt, &ssigned, &sscale, &soffset, &smin, &smax, sunit)) > 6){
	  _dbc_error("Sorry muxed signals are not supported until now\n");
	  goto errorhandling;
        } 
    }
    //Back to start of file
    rewind(fp);
    //Second run
    //Parse for additional informations
    while(fgets(line,SCANBUFFER-1,fp)){
        memset(sname,0,sizeof(char)*SCANBUFFER);
        memset(content,0,sizeof(char)*SCANBUFFER);
        if((rv = sscanf(line,"BA_ \"GenMsgCycleTime\" BO_ %d %d",&mid,&genmsgcycletime)) == 2){
            _dbc_verbose("GenMsgCycleTime ID:%d %d us\n",mid,genmsgcycletime);
	    //Note we work with 0.000001 s resolution, DBC only with 0.001s
            if((m = cs_dbc_getmessage(data,mid)) != NULL){
                if(cs_message_set_interval(m,genmsgcycletime*1000)){
                    _dbc_error("Failed to set cycle time for message %s\n",m->name);
                }
            } else {
	      _dbc_error("Failed to get message by id 0x%X\n",m->id);
	    }
            continue;
        } else if((rv = sscanf(line,"BA_ \"GenMsgDelayTime\" BO_ %d %d",&mid,&genmsgdelaytime)) == 2){
            _dbc_verbose("GenMsgDelayTime ID:%d %d us\n",mid,genmsgdelaytime);
	    //Note we work with 0.000001 s resolution, DBC only with 0.001s
	    if((m = cs_dbc_getmessage(data,mid)) != NULL){
                if(cs_message_set_offset(m,genmsgdelaytime*1000)){
                    _dbc_error("Failed to set cycle time offset for message %s\n",m->name);
                }
            } else {
	      _dbc_error("Failed to get message by id 0x%X\n",m->id);
	    }
            continue;
        } else if((rv = sscanf(line,"BA_ \"GenMsgStartDelayTime\" BO_ %d %d",&mid,&intvalue)) == 2){
            _dbc_verbose("GenMsgStartDelayTime ID:%d %d ms\n",mid,intvalue);
	    //Note we work with 0.000001 s resolution, DBC only with 0.001s
	    if((m = cs_dbc_getmessage(data,mid)) != NULL){
                if(cs_message_set_startoffset(m,intvalue*1000)){
                    _dbc_error("Failed to set cycle time start offset for message %s\n",m->name);
                }
            } else {
	      _dbc_error("Failed to get message by id 0x%X\n",m->id);
	    }
            continue;
        } else if((rv = sscanf(line,"BA_ \"GenMsgSendType\" BO_ %d %d",&mid,&intvalue)) == 2){
            _dbc_verbose("GenMsgSendType ID:%d %d\n",mid,intvalue);
	    if((m = cs_dbc_getmessage(data,mid)) != NULL){
	      cs_message_send_type_t st = MESSAGE_STYPE_NOTACTIVE;
	      switch(intvalue){
	      case 0:
		st = MESSAGE_STYPE_CYCLIC;
		break;
	      case 1:
		st = MESSAGE_STYPE_TRIGGERED;
		break;
	      case 2:
		st = MESSAGE_STYPE_CYCLICENABLED;
		break;
	      case 3:
		st = MESSAGE_STYPE_CYCLICTRIGGERED;
		break;
	      case 4:
		st = MESSAGE_STYPE_CYCLICENABLEDTRIGGERED;
		break;
	      default:
		st = MESSAGE_STYPE_NOTACTIVE;
	      }
	      if(cs_message_set_sendtype(m,st)){
		_dbc_error("Failed to set message send type for message %s\n",m->name);
	      }
            } else {
	      _dbc_error("Failed to get message by id 0x%X\n",m->id);
	    }
            continue;
        } else if((rv = sscanf(line,"BA_ \"GenSigStartValue\" SG_ %d %s %d",&mid,sname,&intvalue)) == 3){
            _dbc_verbose("GenSigStartValue  ID:%d Signal: %s %d\n",mid,sname,intvalue);
	    if(((m = cs_dbc_getmessage(data,mid)) != NULL) &&
	       ((s = cs_dbc_getsignal(data,sname)) != NULL)){
	      //Set start value of signal
	      if(cs_signal_set_startrawvalue(s,intvalue)){
		_dbc_error("Failed to set start value for signal %s\n",sname);
	      }
	    } else {
	      _dbc_error("No signal %s found to add startvalue\n",sname);
	    }
            continue;
        } else if((rv = sscanf(line,"VAL_ %d %s",&mid,sname)) == 2){
            _dbc_verbose("VAL_ ID:%d Signal:%s \n",mid,sname);
            if(((m = cs_dbc_getmessage(data,mid)) != NULL) &&
	       ((s = cs_message_get_signal_byname(m,sname)) != NULL)){
              if(_dbc_scan_signal_label(s,line)){
		_dbc_error("Failed to parse labels for signal %s\n",s->name);
	      }
            }
            continue;
        }
    }

    setlocale(LC_ALL,locale_store);
  
    if(fp != NULL){
        fclose(fp);
    }


    //File read is done, build dependency structure
    if(_dbc_build_dependency(data)){
      _dbc_error("Failed to build dependency tree\n");
    }

    
    return CS_OK;

  errorhandling:
  
    setlocale(LC_ALL,locale_store);
  
    //Close parser
    if(fp != NULL){
        fclose(fp);
    }
    return CS_FAIL;
}
