/*
    Copyright (C) 2019 IgH, Andreas Stewering-Bone

    This file is part of the JEOL project
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "cs_debug.h"

struct _debug_states{
  char *subname;
  cs_debug_sub_t subsystem;
  cs_debug_level_t dl;
  void (*_debugcallb)(void *ctx,
		      cs_debug_sub_t sub,
		      cs_debug_level_t dl,
		      const char *func,
		      const int line,
		      const char *fmt,
		      va_list ap);
  void *_debugcallb_data;
};

//Predecleration
static struct _debug_states _debug_global_states[];

#define DEBUG_GET_SUBNAME(SUB) (_debug_global_states[SUB].subname)
#define DEBUG_GET_LEVEL(SUB) (_debug_global_states[SUB].dl)
#define DEBUG_SET_LEVEL(SUB,VALUE) (_debug_global_states[SUB].dl = VALUE)
#define DEBUG_GET_CALLB(SUB) (_debug_global_states[SUB]._debugcallb)
#define DEBUG_GET_CALLB_DATA(SUB) (_debug_global_states[SUB]._debugcallb_data)
#define DEBUG_SET_CALLB(SUB,VALUE) (_debug_global_states[SUB]._debugcallb = VALUE)
#define DEBUG_SET_CALLB_DATA(SUB,VALUE) (_debug_global_states[SUB]._debugcallb_data = VALUE)

#define BYTE_TO_BINARY_PATTERN "%c%c%c%c%c%c%c%c "
#define BYTE_TO_BINARY(byte)			\
    (byte & 0x80 ? '1' : '0'),			\
    (byte & 0x40 ? '1' : '0'),			\
    (byte & 0x20 ? '1' : '0'),			\
    (byte & 0x10 ? '1' : '0'),			\
    (byte & 0x08 ? '1' : '0'),			\
    (byte & 0x04 ? '1' : '0'),			\
    (byte & 0x02 ? '1' : '0'),			\
    (byte & 0x01 ? '1' : '0')

/****************************************************************************/
char *cs_debug_makemessage(const char *fmt, va_list ap)
/****************************************************************************/
{
  //Taken from printf manual page
  int n;
  int size = 255;
  va_list bp;
  char *p = NULL;
  char *np = NULL;
  p = (char*)calloc(size,sizeof(char));

  if(p == NULL)
    return NULL;
  
  while(1){
    va_copy(bp,ap);
    n = vsnprintf(p, size, fmt, bp);
    if(n < 0) {
      free(p);
      return NULL;
    }
    
    if(n < size){
      return p;
    }

    size = n + 1;
    np = realloc(p, size);
    if (np == NULL) {
      free(p);
      return NULL;
    } else {
      memset(np,0,sizeof(char)*size);
      p = np;
    }
  }
  //Should never be reached
  return NULL;
}

/********************************************************************/
void cs_default_debug_out(void *ctx,
			  cs_debug_sub_t sub,
			  cs_debug_level_t dl,
			  const char *func,
			  const int line,
			  const char *fmt,
			  va_list ap)
/********************************************************************/
{
  char *buffer = NULL;

  printf("%s (Func: %s, Line: %d)",DEBUG_GET_SUBNAME(sub),func,line);

  switch(dl){
  case CS_DEBUG_LEVEL_NONE:
    printf("Info: ");
    break;
  case CS_DEBUG_LEVEL_DEBUG:
    printf("Debug: ");
    break;
  case CS_DEBUG_LEVEL_VERBOSE:
    printf("Verbose: ");
    break;
  case CS_DEBUG_LEVEL_ERROR:
    printf("Error: ");
    break;
  case CS_DEBUG_LEVEL_WARN:
    printf("Warning: ");
    break;
  default:
    //Nothing
    break;
  }
  
  buffer = cs_debug_makemessage(fmt,ap);

  if(buffer == NULL){
    printf("Debug output failure\n");
    return;
  }
  printf("%s",buffer);
  free(buffer);
  
}

/** Global Debug callback function pointer
    Default write debug to stdout
    Keep care on order, we jump via index
*/
//Set state table
static struct _debug_states _debug_global_states[] = {
  {"SocketCAN",CS_DEBUG_SOCKETCAN,CS_DEBUG_LEVEL_NONE,cs_default_debug_out,NULL},
  {"IsoTP",CS_DEBUG_ISOTP,CS_DEBUG_LEVEL_NONE,cs_default_debug_out,NULL},
  {"UDS",CS_DEBUG_UDS,CS_DEBUG_LEVEL_NONE,cs_default_debug_out,NULL},
  {"CCP",CS_DEBUG_CCP,CS_DEBUG_LEVEL_NONE,cs_default_debug_out,NULL},
  {"XCP",CS_DEBUG_XCP,CS_DEBUG_LEVEL_NONE,cs_default_debug_out,NULL},
  {"Signal",CS_DEBUG_SIGNAL,CS_DEBUG_LEVEL_NONE,cs_default_debug_out,NULL},
  {"Message",CS_DEBUG_MESSAGE,CS_DEBUG_LEVEL_NONE,cs_default_debug_out,NULL},
  {"Lin",CS_DEBUG_LIN,CS_DEBUG_LEVEL_NONE,cs_default_debug_out,NULL},
  {"Lin Bit",CS_DEBUG_LIN_BIT,CS_DEBUG_LEVEL_NONE,cs_default_debug_out,NULL},
  {"DBC",CS_DEBUG_DBC,CS_DEBUG_LEVEL_NONE,cs_default_debug_out,NULL},
  {"Schedule",CS_DEBUG_SCHEDULE,CS_DEBUG_LEVEL_NONE,cs_default_debug_out,NULL},
  {"A2L",CS_DEBUG_A2L,CS_DEBUG_LEVEL_NONE,cs_default_debug_out,NULL},
  {"Internal",CS_DEBUG_INT,CS_DEBUG_LEVEL_NONE,cs_default_debug_out,NULL},
  {NULL,CS_DEBUG_ENDMARK,CS_DEBUG_LEVEL_NONE,NULL,NULL},
  {NULL}
};

/********************************************************************/
cs_debug_level_t cs_debuglevel(
			       cs_debug_sub_t sub, 
			       cs_debug_level_t level
			       )
/********************************************************************/
{

  if(sub >= CS_DEBUG_ENDMARK){
    //Invalid subsystem
    return CS_DEBUG_LEVEL_NONE;
  }
  
  switch(level){
  case CS_DEBUG_LEVEL_NONE:
  case CS_DEBUG_LEVEL_DEBUG:
  case CS_DEBUG_LEVEL_VERBOSE:
  case CS_DEBUG_LEVEL_ERROR:
  case CS_DEBUG_LEVEL_WARN:
  case CS_DEBUG_LEVEL_INFO:
    //Set value
    DEBUG_SET_LEVEL(sub,level);
    break;
  case CS_DEBUG_LEVEL_READ:
    //Nothing todo
    return DEBUG_GET_LEVEL(sub);
  default:
    return CS_DEBUG_LEVEL_NONE;
  }
  return CS_DEBUG_LEVEL_NONE;
}


/********************************************************************/
void cs_error(cs_debug_sub_t sub,
	      const char *func,
	      const int line,
	      const char *fmt, ...)
/********************************************************************/
{
  va_list ap;
  cs_debug_level_t dl = cs_debuglevel(sub,CS_DEBUG_LEVEL_READ);
  void (* callb)(void *ctx,
		 cs_debug_sub_t sub,
		 cs_debug_level_t dl,
		 const char *func,
		 const int line,
		 const char *fmt,
		 va_list ap) = DEBUG_GET_CALLB(sub);
  void *callb_data = DEBUG_GET_CALLB_DATA(sub);
  
  if(dl >= CS_DEBUG_LEVEL_ERROR){
    if(callb != NULL){
      va_start(ap, fmt);
      (*callb)(callb_data,sub,CS_DEBUG_LEVEL_ERROR,func,line,fmt,ap);
      va_end(ap);
    }
  }
}

/********************************************************************/
void cs_warn(cs_debug_sub_t sub,
	     const char *func,
	     const int line,
	     const char *fmt, ...)
/********************************************************************/
{
  va_list ap;
  cs_debug_level_t dl = cs_debuglevel(sub,CS_DEBUG_LEVEL_READ);
  void (* callb)(void *ctx,
		 cs_debug_sub_t sub,
		 cs_debug_level_t dl,
		 const char *func,
		 const int line,
		 const char *fmt,
		 va_list ap) = DEBUG_GET_CALLB(sub);
  void *callb_data = DEBUG_GET_CALLB_DATA(sub);
  
  if(dl >= CS_DEBUG_LEVEL_WARN){
    if(callb != NULL){
      va_start(ap, fmt);
      (*callb)(callb_data,sub,CS_DEBUG_LEVEL_WARN,func,line,fmt,ap);
      va_end(ap);
    }
  }
}

/********************************************************************/
void cs_info(cs_debug_sub_t sub,
	     const char *func,
	     const int line,
	     const char *fmt, ...)
/********************************************************************/
{
  va_list ap;

  cs_debug_level_t dl = cs_debuglevel(sub,CS_DEBUG_LEVEL_READ);
  void (* callb)(void *ctx,
		 cs_debug_sub_t sub,
		 cs_debug_level_t dl,
		 const char *func,
		 const int line,
		 const char *fmt,
		 va_list ap) = DEBUG_GET_CALLB(sub);
  void *callb_data = DEBUG_GET_CALLB_DATA(sub);
  
  if(dl >= CS_DEBUG_LEVEL_INFO){
    if(callb != NULL){
      va_start(ap, fmt);
      (*callb)(callb_data,sub,CS_DEBUG_LEVEL_INFO,func,line,fmt,ap);
      va_end(ap);
    }
  }
}

/********************************************************************/
void cs_debug(cs_debug_sub_t sub,
	      const char *func,
	      const int line,
	      const char *fmt, ...)
/********************************************************************/
{
  //#ifdef CS_DEBUG
  va_list ap;
  cs_debug_level_t dl = cs_debuglevel(sub,CS_DEBUG_LEVEL_READ);
    void (* callb)(void *ctx,
		   cs_debug_sub_t sub,
		   cs_debug_level_t dl,
		   const char *func,
		   const int line,
		   const char *fmt,
		   va_list ap) = DEBUG_GET_CALLB(sub);
  void *callb_data = DEBUG_GET_CALLB_DATA(sub);
  
  if(dl >= CS_DEBUG_LEVEL_DEBUG){
    if(callb != NULL){
      va_start(ap, fmt);
      (*callb)(callb_data,sub,CS_DEBUG_LEVEL_DEBUG,func,line,fmt,ap);
      va_end(ap);
    }
  }
  //#endif
}

/********************************************************************/
void cs_verbose(cs_debug_sub_t sub,
		const char *func,
		const int line,
		const char *fmt, ...)
/********************************************************************/
{
  #ifdef CS_VERBOSE
  va_list ap;
  cs_debug_level_t dl = cs_debuglevel(sub,CS_DEBUG_LEVEL_READ);
  void (* callb)(void *ctx,
		 cs_debug_sub_t sub,
		 cs_debug_level_t dl,
		 const char *func,
		 const int line,
		 const char *fmt,
		 va_list ap) = DEBUG_GET_CALLB(sub);
  void *callb_data = DEBUG_GET_CALLB_DATA(sub);
  
  if(dl >= CS_DEBUG_LEVEL_VERBOSE){
    if(callb != NULL){
      va_start(ap, fmt);
      (*callb)(callb_data,sub,CS_DEBUG_LEVEL_VERBOSE,func,line,fmt,ap);
      va_end(ap);
    }
  }
  #endif
}

/********************************************************************/
void cs_error_buffer(cs_debug_sub_t sub,
		     const char *func,
		     const int line,
		     uint8_t *data,
		     int len
		     )
/********************************************************************/
{
  int i;
  int bufferlen = len*3+1; // XX_
  char buffer[bufferlen];
  char *ptr = &buffer[0];

  cs_debug_level_t dl = cs_debuglevel(sub,CS_DEBUG_LEVEL_READ);

  if(dl < CS_DEBUG_LEVEL_ERROR){
    return;
  }
  
  memset(buffer,0,sizeof(char)*bufferlen);

  for(i = 0; i < len; i++){
    ptr += sprintf (ptr, "%02X ", data[i]);
  }
  cs_error(sub,func,line,"Buffer len: %i Data; %s\n",len,buffer);
}

/********************************************************************/
void cs_warn_buffer(cs_debug_sub_t sub,
		    const char *func,
		    const int line,
		    uint8_t *data,
		    int len
		    )
/********************************************************************/
{
  int i;
  int bufferlen = len*3+1; // XX_
  char buffer[bufferlen];
  char *ptr = &buffer[0];

  cs_debug_level_t dl = cs_debuglevel(sub,CS_DEBUG_LEVEL_READ);

  if(dl < CS_DEBUG_LEVEL_WARN){
    return;
  }
  
  memset(buffer,0,sizeof(char)*bufferlen);

  for(i = 0; i < len; i++){
    ptr += sprintf (ptr, "%02X ", data[i]);
  }
  cs_warn(sub,func,line,"Buffer len: %i Data; %s\n",len,buffer);
}

/********************************************************************/
void cs_info_buffer(
		    cs_debug_sub_t sub,
		    const char *func,
		    const int line,
		    uint8_t *data,
		    int len
		    )
/********************************************************************/
{
  int i;
  int bufferlen = len*3+1; // XX_
  char buffer[bufferlen];
  char *ptr = &buffer[0];

  cs_debug_level_t dl = cs_debuglevel(sub,CS_DEBUG_LEVEL_READ);

  if(dl < CS_DEBUG_LEVEL_INFO){
    return;
  }
  
  memset(buffer,0,sizeof(char)*bufferlen);

  for(i = 0; i < len; i++){
    ptr += sprintf (ptr, "%02X ", data[i]);
  }
  cs_info(sub,func,line,"Buffer len: %i Data; %s\n",len,buffer);
}

/********************************************************************/
void cs_info_char_buffer(
			 cs_debug_sub_t sub,
			 const char *func,
			 const int line,
			 uint8_t *data,
			 int len
			 )
/********************************************************************/
{
  int i;
  char buffer[len];

  cs_debug_level_t dl = cs_debuglevel(sub,CS_DEBUG_LEVEL_READ);

  if(dl < CS_DEBUG_LEVEL_INFO){
    return;
  }
  
  memset(buffer,0,sizeof(char)*len);

  for(i = 0; i < len; i++){
    buffer[i] = (char)data[i];
  }
  cs_info(sub,func,line,"Buffer len: %i Data; %s\n",len,buffer);
}

/********************************************************************/
void cs_debug_buffer(
		     cs_debug_sub_t sub,
		     const char *func,
		     const int line,
		     uint8_t *data,
		     int len
		     )
/********************************************************************/
{
  //#ifdef CS_DEBUG
  int i;
  int bufferlen = len*3+1; // XX_
  char buffer[bufferlen];
  char *ptr = &buffer[0];

  cs_debug_level_t dl = cs_debuglevel(sub,CS_DEBUG_LEVEL_READ);

  if(dl < CS_DEBUG_LEVEL_DEBUG){
    return;
  }
  
  memset(buffer,0,sizeof(char)*bufferlen);

  for(i = 0; i < len; i++){
    ptr += sprintf (ptr, "%02X ", data[i]);
  }
  cs_debug(sub,func,line,"Buffer len: %i Data; %s\n",len,buffer);
  //#endif
}

/********************************************************************/
void cs_verbose_buffer(
		       cs_debug_sub_t sub,
		       const char *func,
		       const int line,
		       uint8_t *data,
		       int len
		       )
/********************************************************************/
{
  #ifdef CS_VERBOSE
  int i;
  int bufferlen = len*3+1; // XX_
  char buffer[bufferlen];
  char *ptr = &buffer[0];

  cs_debug_level_t dl = cs_debuglevel(sub,CS_DEBUG_LEVEL_READ);

  if(dl < CS_DEBUG_LEVEL_VERBOSE){
    return;
  }
  
  memset(buffer,0,sizeof(char)*bufferlen);

  for(i = 0; i < len; i++){
    ptr += sprintf (ptr, "%02X ", data[i]);
  }
  cs_verbose(sub,func,line,"Buffer len: %i Data; %s\n",len,buffer);
  #endif
}


/********************************************************************/
void cs_verbose_binbuffer(
			  cs_debug_sub_t sub,
			  const char *func,
			  const int line,
			  uint8_t *data,
			  int len
			  )
/********************************************************************/
{
  #ifdef CS_VERBOSE
  int i;
  int bufferlen = len*12+1; // XX_
  char buffer[bufferlen];
  char *ptr = &buffer[0];

  cs_debug_level_t dl = cs_debuglevel(sub,CS_DEBUG_LEVEL_READ);

  if(dl < CS_DEBUG_LEVEL_VERBOSE){
    return;
  }
  
  memset(buffer,0,sizeof(char)*bufferlen);

  for(i = 0; i < len; i++){
    ptr += sprintf (ptr, BYTE_TO_BINARY_PATTERN,BYTE_TO_BINARY(data[i]));
  }
  cs_verbose(sub,func,line,"Buffer len: %i Data; %s\n",len,buffer);
  #endif
}

/********************************************************************/
void cs_debug_setcallback(
			  cs_debug_sub_t sub,
			  void *ctx,
			  void (*callb)(void *ctx,
					cs_debug_sub_t sub,
					cs_debug_level_t dl,
					const char *func,
					const int line,
					const char *fmt,
					va_list ap
					)
			  )
/********************************************************************/
{
  if(sub >= CS_DEBUG_ENDMARK){
    //Invalid subsystem
    return;
  }

  DEBUG_SET_CALLB(sub,callb);
  DEBUG_SET_CALLB_DATA(sub,ctx);
  
}


/********************************************************************/
cs_debug_level_t cs_int_debuglevel(
				   cs_debug_level_t level
				   )
/********************************************************************/
{
  return cs_debuglevel(CS_DEBUG_INT,level);
}

/********************************************************************/
void cs_int_debug_setcallback(
			      void *ctx,
			      void (*callb)(void *ctx,
					    cs_debug_sub_t sub,
					    cs_debug_level_t dl,
					    const char *func,
					    const int line,
					    const char *fmt,
					    va_list ap)
			      )
/********************************************************************/
{
  cs_debug_setcallback(CS_DEBUG_INT,ctx,callb);
}
