/*  
    Copyright (C) 2013 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include "config.h"
#include "cs_isotp.h"
#include "cs_isotp_private.h"
#include "cs_socketcan.h"
#include "cs_time.h"
#include "cs_defines.h"
#include "cs_debug_private.h"

#define _isotp_error(...) (cs_error(CS_DEBUG_ISOTP,__func__,__LINE__,__VA_ARGS__))
#define _isotp_warn(...) (cs_warn(CS_DEBUG_ISOTP,__func__,__LINE__,__VA_ARGS__))
#define _isotp_info(...) (cs_info(CS_DEBUG_ISOTP,__func__,__LINE__,__VA_ARGS__))
#define _isotp_debug(...) (cs_debug(CS_DEBUG_ISOTP,__func__,__LINE__,__VA_ARGS__))
#define _isotp_verbose(...) (cs_verbose(CS_DEBUG_ISOTP,__func__,__LINE__,__VA_ARGS__))
#define _isotp_error_buffer(DATA,LEN) (cs_error_buffer(CS_DEBUG_ISOTP,__func__,__LINE__,DATA,LEN))
#define _isotp_warn_buffer(DATA,LEN) (cs_warn_buffer(CS_DEBUG_ISOTP,__func__,__LINE__,DATA,LEN))
#define _isotp_info_buffer(DATA,LEN) (cs_info_buffer(CS_DEBUG_ISOTP,__func__,__LINE__,DATA,LEN))
#define _isotp_debug_buffer(DATA,LEN) (cs_debug_buffer(CS_DEBUG_ISOTP,__func__,__LINE__,DATA,LEN))
#define _isotp_verbose_buffer(DATA,LEN) (cs_verbose_buffer(CS_DEBUG_ISOTP,__func__,__LINE__,DATA,LEN))




/********************************************************************/
cs_debug_level_t cs_isotp_debuglevel(
				     cs_debug_level_t level
				     )
/********************************************************************/
{
  return cs_debuglevel(CS_DEBUG_ISOTP,level);
}

/********************************************************************/
void cs_isotp_debug_setcallback(
				void *ctx,
				void (*callb)(void *ctx,
					      cs_debug_sub_t sub,
					      cs_debug_level_t dl,
					      const char *func,
					      const int line,
					      const char *fmt,
					      va_list ap)
				)
/********************************************************************/
{
  cs_debug_setcallback(CS_DEBUG_ISOTP,ctx,callb);
}


/********************************************************************/
uint8_t _mues2stim(int value)
/********************************************************************/
{
  //convert from 0.000001s to ISOTP specific code
  uint8_t rv = 0x00; //0ms

  if( value < 0 ){
    _isotp_error("Invalid argument");
    return rv;
  }

  if( value == 0 ){
    return rv;
  }

  if(value >= 1000){
    rv = (uint8_t)(value/1000);
    if(rv > 0x7F)
      rv = 0x7F;
    return rv;
  }

  if((value > 0) && (value < 1000)){
    rv = 0xF0 + (uint8_t)(value/100);
    if(rv < 0xF1)
      rv = 0xF1;
    return rv;
  }
  _isotp_error("Zero convert");
  return rv;
}

/********************************************************************/
struct timespec _stim2ts(uint8_t stim)
/********************************************************************/
{
  //convert from specific ISOTP code to timespec
  struct timespec rv;
  rv.tv_sec = 0;
  rv.tv_nsec = 2e6L; //2ms
  
  if((stim >= 0) && (stim <= 0x7F)){
    rv.tv_nsec = 1e6L * stim;
    return rv;
  }

  if((stim >= 0xF1) && (stim <= 0xF9)){
    rv.tv_nsec = 1e5L * (stim - 0xF0);
    return rv;
  }
  
  //Return fallback value
  _isotp_error("Cannot convert");
  return rv;
}

/********************************************************************/
double _stim2d(uint8_t stim)
/********************************************************************/
{
  //convert from specific ISOTP code double in ms
  double rv;

    if((stim >= 0) && (stim <= 0x7F)){
      rv = (double)stim;
      return rv;
    }

    if((stim >= 0xF1) && (stim <= 0xF9)){
      rv = (double)(stim - 0xF0)*0.1;
      return rv;
    }
    _isotp_error("Cannot convert");
    return rv = 0;
}

/********************************************************************/
uint32_t _stim2mues(uint8_t stim)
/********************************************************************/
{
  //convert from specific ISOTP code double in 0.000001s
  uint32_t rv = 2000;
  
    if((stim >= 0) && (stim <= 0x7F)){
      rv = (uint32_t)stim*1000;
      return rv;
    }

    if((stim >= 0xF1) && (stim <= 0xF9)){
      rv = (uint32_t)(stim - 0xF0)*100;
      return rv;
    }
    _isotp_error("Cannot convert");
    return rv;
}

/* Taken from SocketCAN isotp.c */
/* Copyright (c) 2017 Volkswagen Group Electronic Research*/
/** Get frame len for a given data segment length
    \retval expected frame length
 */
/********************************************************************/
static uint8_t _isotp_padlen(
			     uint8_t datalen /**< Data segment length in bytes*/
			     )
/********************************************************************/
{
  const uint8_t plen[] = {8,  8,  8,  8,  8,  8,  8,  8,  8,	/* 0 - 8 */
			  12, 12, 12, 12,			/* 9 - 12 */
			  16, 16, 16, 16,			/* 13 - 16 */
			  20, 20, 20, 20,			/* 17 - 20 */
			  24, 24, 24, 24,			/* 21 - 24 */
			  32, 32, 32, 32, 32, 32, 32, 32,	/* 25 - 32 */
			  48, 48, 48, 48, 48, 48, 48, 48,	/* 33 - 40 */
			  48, 48, 48, 48, 48, 48, 48, 48};	/* 41 - 48 */
  
  if (datalen > 48)
    return 64;

  return plen[datalen];
}

/** Taken from SocketCAN isotp.c*/
/* Copyright (c) 2017 Volkswagen Group Electronic Research*/
/* check for length optimization and return 1/true when the check fails */
/********************************************************************/
static int _isotp_check_optimized(
				  int len,               /**< Receive buffer length*/
				  int start_index        /**< Start offset to check padding*/
				  )
/********************************************************************/
{
  /*
   * for CAN_DL <= 8 the start_index is equal to the CAN_DL as the
   * padding would start at this point. E.g. if the padding would
   * start at cf.data[7] cf->len has to be 7 to be optimal.
   * Note: The data[] index starts with zero.
   */
  if (len <= ISOTP_LL_TX_DL_8){
    return (len != start_index);
  }

  /*
   * This relation is also valid in the non-linear DLC range, where
   * we need to take care of the minimal next possible CAN_DL.
   * The correct check would be (padlen(cf->len) != padlen(start_index)).
   * But as cf->len can only take discrete values from 12, .., 64 at this
   * point the padlen(cf->len) is always equal to cf->len.
   */
  return (len != _isotp_padlen(start_index));
}

/* Taken from SocketCAN isotp.c*/
/* Copyright (c) 2017 Volkswagen Group Electronic Research*/
/* check padding and return 1/true when the check fails */
/********************************************************************/
static int _isotp_check_pad(
			    cs_isotp_t *data, /**< Storage structure*/
			    uint8_t * buffer, /**< Receive buffer*/
			    int len,          /**< Receive buffer length*/
			    uint32_t start_index   /**< Start index to check padding*/
			    )
/********************************************************************/
{
  int i;

  _isotp_verbose("Entry\n");
  /* no RX_PADDING value => check length of optimized frame length */
  if(!(data->options & ISOTP_OPTION_RXPADDING)){
    if (data->options & ISOTP_OPTION_RXLENGTH){
      if(_isotp_check_optimized(len, start_index)){
	_isotp_error("RX Length check failed\n");
	return 1;
      } else {
	return 0;
      }
    }
    /* no valid test against empty value => ignore frame */
    return 1;
  }

  /* check datalength of correctly padded CAN frame */
  if ((data->options & ISOTP_OPTION_RXLENGTH) &&
      len != _isotp_padlen(len)){
    _isotp_error("RX Length check failed\n");
    return 1;
  }

  /* check padding content */
  if(data->options & ISOTP_OPTION_RXDATA){
    for (i = start_index; i < len; i++){
      if (buffer[i] != data->padbyte){
	_isotp_error("RX padding data invalid\n");
	return 1;
      }
    }
  }
  return 0;
}

/********************************************************************/
int cs_isotp_reset(
		   cs_isotp_t *data
		   )
/********************************************************************/
{
  if(data == NULL)
    return CS_FAIL;


  data->sa = 0;
  data->ta = 0;
  data->rxstim = _mues2stim(ISOTP_RXSTIM);
  data->txstim = 0;
  data->rxbs = ISOTP_RXBS;
  data->txbs = ISOTP_RXBS;
  data->options = ISOTP_OPTION_DEFAULT;
  data->padbyte = ISOTP_PADDING_DEFAULT;
  data->cf_frames = 0;
  data->cf_last = 0;
  data->cf_counter = 0;
  data->ll_mode = ISOTP_LL_MODE_CAN; //CAN 2.0
  data->tx_dl = ISOTP_LL_TX_DL_8; //CAN 2.0 compatible
  data->rx_dl = ISOTP_LL_TX_DL_8; //CAN 2.0 compatible
  data->dl_max =  ISOTP_FRAME_CAN_MAX; //Max Frame length, CAN 2.0, 4095 bytes
  data->ll_send_mode = SOCKETCAN_SEND_CAN; //CAN 2.0
  data->tx_first_FC = 1; //Wait for first FC
  data->seq = 0; //Reset sequence number
  data->segment_position = 0; //Reset data segment position offset
  data->segment_length = 0; //Reset data segment length
  data->transfercb = NULL; //Callback function pointer
  data->transfercb_ctx = NULL; //Callback function context pointer
  return CS_OK;
}

/********************************************************************/
int cs_isotp_getlasterror(
			  cs_isotp_t *data
			  )
/********************************************************************/
{
  if(data == NULL)
    return 0;

  return data->lerror;
}

/********************************************************************/
char* cs_isotp_getlasterrorstring(
				  cs_isotp_t *data
				  )
/********************************************************************/
{
  if(data == NULL)
    return "";

  if((data->lerror == ISOTP_ERROR_SOCKETCAN) &&
     (data->socket != NULL)
     ){
    return cs_socketcan_getlasterrorstring(data->socket);
  }
  return cs_isotp_geterroridstring(data->lerror);
}

/********************************************************************/
char* cs_isotp_geterroridstring(
				uint8_t lerror
				)
/********************************************************************/
{
  switch(lerror){
  case ISOTP_ERROR_NOERROR:
    return "IsoTP: No error";
  case ISOTP_ERROR_SOCKETCAN_INIT:
    return "IsoTP: SocketCAN Init failure";
  case ISOTP_ERROR_SOCKETCAN:
    return "IsoTP: SocketCAN General failure";
  case ISOTP_ERROR_RECV:
    return "IsoTP: Receive failure";
  case ISOTP_ERROR_SEND:
    return "IsoTP: Send failure";
  case ISOTP_ERROR_UNSUPPORTED:
    return "IsoTP: Feature is not supported";
  case ISOTP_ERROR_UNKNOWN:
    return "IsoTP: Unknwon frame";
  case ISOTP_ERROR_SF:
    return "IsoTP: Invalid SF frame";
  case ISOTP_ERROR_FC:
    return "IsoTP: Invalid FC frame";
  case ISOTP_ERROR_FF:
    return "IsoTP: Invalid FF frame";
  case ISOTP_ERROR_CF:
    return "IsoTP: Invalid CF frame";
  case ISOTP_ERROR_UNEXPECTED:
    return "IsoTP: Receive unexpected frame";
  case ISOTP_ERROR_DATA:
    return "IsoTP: Data out of range";
  case ISOTP_ERROR_PARAMETER:
    return "IsoTP: Invalid parameter";
  case ISOTP_ERROR_SEQUENCE:
    return "IsoTP: Invalid sequence number";
  case ISOTP_ERROR_TIMEOUT:
    return "IsoTP: Receive timeout";
  case ISOTP_ERROR_TXFULL:
    return "IsoTP: TX path is full";
  case ISOTP_ERROR_OPTION:
    return "IsoTP: Invalid Option";
  case ISOTP_ERROR_RXPADDING:
    return "IsoTP: RX Padding failure";
  case ISOTP_ERROR_RXLENGTH:
    return "IsoTP: RX Length failure";
  case ISOTP_ERROR_OVERFLOW:
    return "IsoTP: Data buffer overflow on receiver side";
  }

  return "";
}

/********************************************************************/
cs_socketcan_t *cs_isotp_get_socketcan(
				       cs_isotp_t *data
				       )
/********************************************************************/
{
  if( data == NULL)
    return NULL;

  return data->socket;
}

/********************************************************************/
cs_isotp_t *cs_isotp_init(
			  uint32_t sa,
			  uint32_t ta
			  )
/********************************************************************/
{
  cs_isotp_t *data = NULL;
    
  data = (cs_isotp_t*)malloc(sizeof(cs_isotp_t));
  if(data == NULL)
    return NULL;

  data->socket = cs_socketcan_init();

  if(data->socket == NULL){
    _isotp_error("Failed to configure socket\n");
    free(data);
    return NULL;
  }
  
  cs_isotp_reset(data);

  data->sa = sa;
  data->ta = ta;

  return data;
}

/********************************************************************/
int cs_isotp_setoptions(
			cs_isotp_t *data,
			uint8_t option
			)
/********************************************************************/
{
  uint8_t force_txstim = 0;
  if(data == NULL)
    return CS_FAIL;
  
  //Store force txstim
  force_txstim = data->options & ISOTP_OPTION_TXSTIM_FORCE;

  //Set option
  //TODO Check for valid values
  data->options = force_txstim | option;
  return CS_OK;
}

/********************************************************************/
int cs_isotp_setoptions_link_layer(
				   cs_isotp_t *data,
				   cs_isotp_ll_mode_t ll_mode,
				   cs_isotp_ll_tx_dl_t tx_dl
				   )
/********************************************************************/
{
  if(data == NULL)
    return CS_FAIL;

  if(cs_socketcan_isconnected(data->socket) == 1){
    _isotp_error("Link layer configuration can not be changed after open call");
    return CS_FAIL;
  }
  
  if((ll_mode == ISOTP_LL_MODE_CAN) && (tx_dl != ISOTP_LL_TX_DL_8)){
    _isotp_error("Invalid data length for CAN 2.0");
    return CS_FAIL;
  }

  if(ll_mode == ISOTP_LL_MODE_CAN){
    //CAN 2.0
    //Max ISOTP Frame length
    data->dl_max =  ISOTP_FRAME_CAN_MAX;
    data->ll_send_mode = SOCKETCAN_SEND_CAN;
  } else {
    //CAN-FD
    //Max ISOTP Frame length
    data->dl_max = ISOTP_FRAME_CANFD_MAX;
    if(ll_mode ==  ISOTP_LL_MODE_CANFD_BRS){
      data->ll_send_mode = SOCKETCAN_SEND_CANFD_BRS;
    } else {
      data->ll_send_mode = SOCKETCAN_SEND_CANFD;
    }
  }
  
  data->ll_mode = ll_mode;
  data->tx_dl = tx_dl;
  //Always start with the maximum possible
  data->rx_dl = ISOTP_LL_TX_DL_64;
  return CS_OK;
}

/********************************************************************/
int cs_isotp_getmaxbytes(
			 cs_isotp_t *data
			 )
/********************************************************************/
{
  if(data == NULL){
    return -1;
  }
  return data->dl_max;
}

  

/********************************************************************/
int cs_isotp_setpadbyte(
			cs_isotp_t *data,
			uint8_t byte
			)
/********************************************************************/
{
  if(data == NULL)
    return CS_FAIL;

  data->padbyte = byte;

  return CS_OK;
}

/********************************************************************/
int cs_isotp_settiming(
		       cs_isotp_t *data,
		       int txstim,
		       int rxstim
		       )
/********************************************************************/
{
  if(data == NULL)
    return CS_FAIL;

  if(rxstim >= 0){
    data->rxstim = _mues2stim(rxstim);
  }
  if(txstim >= 0){
    data->options |= ISOTP_OPTION_TXSTIM_FORCE;
    data->txstim = _mues2stim(txstim);
  }
  return CS_OK;
}

/********************************************************************/
int cs_isotp_reset_timing(
			  cs_isotp_t *data
			  )
/********************************************************************/
{
  if(data == NULL)
    return CS_FAIL;
  
  data->rxstim = _mues2stim(ISOTP_RXSTIM);
  data->txstim = 0;
  data->options &= ~ISOTP_OPTION_TXSTIM_FORCE;
  return CS_OK;
}
  
/********************************************************************/
int cs_isotp_setrxblocksize(
			    cs_isotp_t *data,
			    uint8_t bs
			    )
/********************************************************************/
{
  if(data == NULL)
    return CS_FAIL;

  if(bs != 0){
    data->rxbs = bs;
  }
  return CS_OK;
}

/********************************************************************/
int cs_isotp_open(
		  cs_isotp_t *data,
		  char *iface,
		  char *server,
		  int port
		  )
/********************************************************************/
{
  //TODO Handling functional adressing in filter context
  cs_socketcan_mode_t smode;
  if(data == NULL)
    return CS_FAIL;
  
  if(iface == NULL)
    return CS_FAIL;
  
  data->lerror = ISOTP_ERROR_NOERROR;

  if(server != NULL){
    if(cs_socketcan_config_socketcand(data->socket,iface,server,port)){
      _isotp_error("Failed to configure SocketCANd socket\n");
      data->lerror = ISOTP_ERROR_SOCKETCAN;
      return CS_FAIL;
    }
  } else {
    switch(data->ll_mode){
    case  ISOTP_LL_MODE_CAN:
    default:
      smode = SOCKETCAN_MODE_CAN;
      break;
    case ISOTP_LL_MODE_CANFD:
    case ISOTP_LL_MODE_CANFD_BRS:
      smode = SOCKETCAN_MODE_CANFD;
      break;
    }
    if(cs_socketcan_config_native(data->socket,iface,smode)){
      _isotp_error("Failed to configure native socket\n");
      data->lerror = ISOTP_ERROR_SOCKETCAN;
      return CS_FAIL;
    }
  }

  if(cs_socketcan_set_filter(data->socket,&data->ta,1)){
    _isotp_error("Failed to configure filter\n");
    data->lerror = ISOTP_ERROR_SOCKETCAN;
    return CS_FAIL;
  }
  
  if(cs_socketcan_open(data->socket)){
    _isotp_error("Failed to open socket\n");
    data->lerror = ISOTP_ERROR_SOCKETCAN;
    return CS_FAIL;
  }
  
  return CS_OK;
}

/********************************************************************/
int cs_isotp_close(
		   cs_isotp_t *data
		   )
/********************************************************************/
{
  if(data == NULL)
    return CS_FAIL;

  data->lerror = ISOTP_ERROR_NOERROR;

  if(cs_socketcan_close(data->socket)){
    _isotp_error("Failed to close socket\n");
    data->lerror = ISOTP_ERROR_SOCKETCAN;
    return CS_FAIL;
  }
  
  return CS_OK;
}

/********************************************************************/
void cs_isotp_free(
		   cs_isotp_t *data
		   )
/********************************************************************/
{
  if(data == NULL){
    return;
  }
  
  cs_socketcan_free(data->socket);
  
  free(data);

  data = NULL;
}

/********************************************************************/
void cs_isotp_transfer_setcallback(
				   cs_isotp_t *data,
				   void *ctx,
				   void (transfercb)(void *ctx,
						     uint8_t mode,
						     uint32_t length,
						     uint32_t counter,
						     uint32_t actbytes
						     )
				   )
/********************************************************************/
{
  if(data == NULL){
    return;
  }
  
  
  data->transfercb = transfercb;
  data->transfercb_ctx = ctx;
  
}


/********************************************************************/
int _isotp_recv(cs_isotp_t *data,
		uint8_t *buffer,
		int *len,
		uint32_t timeout)
/********************************************************************/
{
  struct timeval dummy;
  uint32_t id;
  int rv;
  int loop = 1;
  struct timespec ts_timeout;
  _isotp_verbose("Timeout: %i ms\n",timeout);
  //Get actual timestamp
  cs_ts_gettime(&ts_timeout);
  //Add timeout
  cs_ts_add_timeout(&ts_timeout,timeout);
  //Store original buffer len
  int maxlen = *len;
  
  while(loop){
    //Reset length to the maximum possible
    *len = maxlen;
    //Reset recv buffer
    memset(buffer,0,maxlen*sizeof(uint8_t));
    //Receive
    rv = cs_socketcan_recv_block(data->socket,&id,buffer,len,&dummy,timeout);
    switch(rv){
    case CS_SOCKETCAN_OK:
      //Check for correct id
      if(id != data->ta){
	//Unknown id, skip
	_isotp_warn("Receive unkown CAN-Id: 0x%X\n",id);
      } else{
	//Check minimum length
	if(*len < 1){
	  //Invalid buffer length, skip
	  _isotp_error("Invalid buffer\n");
	} else {
	  _isotp_verbose("Receive success\n");
	  switch(buffer[0] & 0xF0){
	  case ISOTP_SF_ID:
	    _isotp_verbose("SF Frame\n");
	    return ISOTP_SF;
	  case ISOTP_FF_ID:
	    _isotp_verbose("FF frame\n");
	    return ISOTP_FF;
	  case ISOTP_CF_ID:
	    _isotp_verbose("CF frame\n");
	    return ISOTP_CF;
	  case ISOTP_FC_ID:
	    _isotp_verbose("FC frame\n");
	    return ISOTP_FC;
	  default:
	    _isotp_error("Unknown frame 0x%X\n",buffer[0] & 0xF0);
	    //Ignore
	  }
	}
      }
      break;
    case CS_SOCKETCAN_FAIL:
      _isotp_error("Receive failed: %s\n",cs_socketcan_getlasterrorstring(data->socket));
      data->lerror = ISOTP_ERROR_SOCKETCAN;
      return CS_FAIL;
      break;
    case CS_SOCKETCAN_TIMEOUT:
      _isotp_error("Receive timeout\n");
      data->lerror = ISOTP_ERROR_TIMEOUT;
      return CS_FAIL;
      break;
    case CS_SOCKETCAN_TXFULL:
      _isotp_warn("Receive TXFULL\n");
      //Ignore
      /*data->lerror = ISOTP_ERROR_TXFULL;
	return CS_FAIL;*/
      break;
    case CS_SOCKETCAN_RTR:
    case CS_SOCKETCAN_ERROR:
      _isotp_debug("Receive RTR or ERROR\n");
      //Ignore
      break;
    default:
      _isotp_error("Receive unknown ERROR\n");
      data->lerror = ISOTP_ERROR_SOCKETCAN;
      return CS_FAIL;
    }
    if((timeout > 0) && cs_ts_checkpast(&ts_timeout)){
      _isotp_error("Receive timout reached\n");
      data->lerror = ISOTP_ERROR_RECV;
      return CS_FAIL;
    }
  }
  return CS_FAIL;
}

/********************************************************************/
int _isotp_send(cs_isotp_t *data,
		uint8_t *buffer,
		int len)
/********************************************************************/
{
  int rv;
  
  rv = cs_socketcan_send(data->socket,data->sa,buffer,len,data->ll_send_mode);
  switch(rv){
  case CS_SOCKETCAN_OK:
    _isotp_verbose("Send successfully\n");
    data->lerror = ISOTP_ERROR_NOERROR;
    return CS_OK;
    break;
  case CS_SOCKETCAN_TXFULL:
    _isotp_error("TXFULL\n");
    data->lerror = ISOTP_ERROR_TXFULL;
    return CS_FAIL;
    break;
  default:
    _isotp_error("Sending failure\n");
    data->lerror = ISOTP_ERROR_SOCKETCAN;
    return CS_FAIL;
    break;
  }
  return CS_OK;
}

/********************************************************************/
int _isotp_send_SF(cs_isotp_t *data,
		   uint8_t *buffer)
/********************************************************************/
{
  uint8_t sendbuffer[data->tx_dl];
  int datalen = data->tx_dl;
  uint32_t dataoffset;

  /*Prepare message with padbyte*/
  memset(sendbuffer,data->padbyte,datalen);
  
  if(datalen == ISOTP_LL_TX_DL_8){
    //CAN 2.0 and 8 Byte CAN-FD are handled in the traditional way
    dataoffset = 1;
    //Identifier and length in first byte
    sendbuffer[0] = ISOTP_SF_ID | (((uint8_t)data->segment_length) & ISOTP_SF_LMASK);
  } else {
    //CAN-FD frames > 8 Bytes
    dataoffset = 2;
    //Only identifier in first byte
    sendbuffer[0] = ISOTP_SF_ID;
    //length in second byte
    sendbuffer[1] = (uint8_t) data->segment_length;
  }

  //Check len
  if(data->segment_length > (datalen - dataoffset)){
    _isotp_error("Invalid data length\n");
    return CS_FAIL;
  }
  //Copy data into sendbuffer
  memcpy(&sendbuffer[dataoffset],&buffer[data->segment_position],data->segment_length);
  
  if(data->options & ISOTP_OPTION_TXPADDING){
    //Send maximum length with padding bytes
    _isotp_verbose("Send with TXPADDING\n");
    return _isotp_send(data,sendbuffer,datalen);
  } else {
    //Send only data content
    _isotp_verbose("Send\n");
    return _isotp_send(data,sendbuffer,data->segment_length + dataoffset);
  }
  //should never be reached
  return CS_FAIL;
}

/********************************************************************/
int _isotp_recv_SF(
		   cs_isotp_t *data,
		   uint8_t *recvbuffer, 
		   int recvlen,
		   uint8_t *buffer,
		   int len
		   )
/********************************************************************/
{
  uint32_t locallen = 0;
  uint32_t padoffset = 0;
  uint32_t dataoffset = 0;
  
  if(recvlen < 2){
    data->segment_length = 0;
    _isotp_error("Invalid data length\n");
    data->lerror = ISOTP_ERROR_SF;
    return CS_FAIL;
  }

  //Reset segment length to zero
  data->segment_length = 0;

  //Set position, always zero for SF frame
  data->segment_position = 0;
  
  //For CAN 2.0 and CAN-FD length <= 8 Byte the length has to
  //Get length, Note only valid for frames <= 8 Byte
  //For bigger frames it has to be zero
  locallen = (uint32_t)(recvbuffer[0] & ISOTP_SF_LMASK);
  
  if(recvlen <= ISOTP_LL_TX_DL_8){
    //Handle the traditional way
    //First check range
    //maximum allowed is tx_dl - 1, here tx_dl = 8 Byte is assumed
    //for CAN 2.0 7 Bytes
    if((locallen <= 0) ||(locallen > (ISOTP_LL_TX_DL_8-1))){
      _isotp_error("Invalid data range\n");
      data->lerror = ISOTP_ERROR_SF;
      return CS_FAIL;
    }
    //Calculate Padding data startoffstet
    //For 8 Bytes frames this is 1 byte offset + len
    dataoffset = 1;
    padoffset = dataoffset + locallen;
  } else {
    //CAN-FD
    //Note locallen (coded in first byte)
    //has to be zero for frame data > 8 Byte
    if(locallen != 0){
      _isotp_error("Invalid data len for CAN-FD frame\n");
      data->lerror = ISOTP_ERROR_SF;
      return CS_FAIL;
    }
    //Get length length
    //for CAN-FD
    //coded in the first data byte
    locallen =  (uint32_t)recvbuffer[1];
    
    if(locallen >= (recvlen - 2)){
      _isotp_error("Invalid data range for CAN-FD frame\n");
      data->lerror = ISOTP_ERROR_SF;
      return CS_FAIL;
    }
    //Calculate Padding data startoffstet
    //For CAN-FD frames this is 2 byte offset + len
    dataoffset = 2;
    padoffset = dataoffset + locallen;
  }

  //Optional check padding
  if ((data->options & ISOTP_CHECK_PADDING) &&
      _isotp_check_pad(data,recvbuffer ,recvlen,padoffset)){
    data->lerror = ISOTP_ERROR_SF;
    return CS_FAIL;
  }

  //Final check if data fits into storage buffer
  if(locallen > len){
    _isotp_error("Data does not fit into storage buffer\n");
    data->lerror = ISOTP_ERROR_SF;
    return CS_FAIL;
  }
  
  //Frame seems to valid, now copy
  memcpy(&buffer[data->segment_position],&recvbuffer[dataoffset],locallen);

  //Store length
  data->segment_length = locallen;

  return CS_OK;
}

/********************************************************************/
int _isotp_send_FC(
		   cs_isotp_t *data,
		   uint8_t overflow,
		   uint8_t wait
		   )
/********************************************************************/
{
  //Note we take rx_dl not tx_dl, because we sync frame length to the sender
  //rx_dl is defined in the FF frame receive function
  uint8_t sendbuffer[data->rx_dl];
  int len = data->rx_dl;
  
  if(overflow && wait){
    _isotp_error("Invalid parameter\n");
    data->lerror = ISOTP_ERROR_FC;
    return CS_FAIL;
  }

  //Set complete buffer with padbyte
  memset(sendbuffer,data->padbyte,len);

  //Set FC id
  sendbuffer[0] = ISOTP_FC_ID; //This is a FC frame

  //Set additional flags if needed
  if(overflow)
    sendbuffer[0] |= ISOTP_FC_OVERFLOW;
  if(wait)
    sendbuffer[0] |= ISOTP_FC_WAIT;

  //Set block size
  sendbuffer[1] = data->rxbs;
  //Set RX STIM
  sendbuffer[2] = data->rxstim;
  
  _isotp_verbose("Send FC Frame: BS: %i STIM: %f ms\n",data->rxbs,_stim2d(data->rxstim));
  if(data->options & ISOTP_OPTION_TXPADDING){
    _isotp_verbose("Send with TXPADDING\n");
    return _isotp_send(data,sendbuffer,len);
  } else {
    _isotp_verbose("Send\n");
    return _isotp_send(data,sendbuffer,ISOTP_FC_LEN);
  }
}

/********************************************************************/
int _isotp_recv_FC(cs_isotp_t *data)
/********************************************************************/
{
  //We are on the sender side, but always take rx_dl
  uint8_t buffer[data->rx_dl];
  int len = data->rx_dl;

  //Reset buffer
  memset(buffer,0,len*sizeof(uint8_t));
 
  switch(_isotp_recv(data,buffer,&len,ISOTP_TIMEOUT_BS)){
  case ISOTP_FC:
    break;
  case ISOTP_SF:
  case ISOTP_CF:
  case ISOTP_FF:
    _isotp_error("Invalid frame, Expected FC Frame\n");
    return ISOTP_IGNORE;
    break;
  case CS_FAIL:
  default:
    _isotp_error("Receive failure: %s\n",cs_isotp_getlasterrorstring(data));
    return CS_FAIL;
    break;
  }

  /*Check for minimum FC frame length*/
  if(len < ISOTP_FC_LEN){
    _isotp_error("Incompatible FC frame length\n");
    data->lerror = ISOTP_ERROR_FC;
    return CS_FAIL;
  }

  //Check padding
  if((data->options & ISOTP_CHECK_PADDING) && 
     _isotp_check_pad(data,buffer,len,ISOTP_FC_LEN)){
    data->lerror = ISOTP_ERROR_FC;
    return CS_FAIL;
  }
  
  /*Check for Wait*/
  if(buffer[0] & ISOTP_FC_WAIT){
    _isotp_verbose("Wait\n");
    return ISOTP_WAIT;
  }

  /*Check for overflow*/
  if(buffer[0] & ISOTP_FC_OVERFLOW){
    data->lerror = ISOTP_ERROR_OVERFLOW;
    _isotp_error("Overflow on client side\n");
    return CS_FAIL;
  }

  //Check if it is the first FC frame
  if(data->tx_first_FC == 0){
    //No, we are finished
    _isotp_verbose("Receive frame, but do not take data\n");
    return CS_OK;
  }

  //Reset flag
  data->tx_first_FC = 0;
  
  //Take communication parameter
  
  /*get block size*/
  data->txbs = buffer[1];

  if(data->txbs == 0){
    _isotp_verbose("Client will not send any FC frame for the actual transmission\n");
  }
  
  if(data->options & ISOTP_OPTION_TXSTIM_FORCE){
    _isotp_warn("Ignore STIM, set to %f ms\n",_stim2d(data->txstim));
  } else {
    /*get STMIN*/
    data->txstim = buffer[2];
    if(data->txstim  == 0){
     _isotp_warn("Receive zero TXSTIM\n");
     //Set to default //1 ms
     //Reason we can not check if bus is free, we use socketcan
     //and we do not know the bitrate, so be conservative with 1 ms
     data->txstim = _mues2stim(ISOTP_RXSTIM);
    }
    //Check for reserved fields
    if((data->txstim >= 0x80) && (data->txstim <= 0xF0)){
      //Set to maximum, (defined in ISO)
      data->txstim = 0x7F;
    }
    if((data->txstim >= 0xFA) && (data->txstim <= 0xFF)){
      //Set to maximum, (defined in ISO)
      data->txstim = 0x7F;
    }
  }
  _isotp_verbose("BS=%d frames\n",data->txbs);
  _isotp_verbose("STMIN= %f ms\n",_stim2d(data->txstim));

  return CS_OK;
}

/********************************************************************/
int _isotp_send_FF(
		   cs_isotp_t *data,
		   uint8_t *buffer
		   )
/********************************************************************/
{
  uint8_t sendbuffer[data->tx_dl];
  int len = data->tx_dl; 
  uint32_t dataoffset;

  //Reset complete sendbuffer with padbyte
  memset(sendbuffer,data->padbyte,len);

  //Set FF flag
  sendbuffer[0] = ISOTP_FF_ID;  
  
  if(data->segment_length > ISOTP_FRAME_CAN_MAX){
    //Set second byte to zero
    //to singal length encoding
    sendbuffer[1] = 0;
    //Store encoding in the next 4 bytes
    sendbuffer[2] = (uint8_t)(data->segment_length >> 24) & 0xFF;
    sendbuffer[3] = (uint8_t) (data->segment_length >> 16) & 0xFF;
    sendbuffer[4] = (uint8_t) (data->segment_length >> 8) & 0xFF;
    sendbuffer[5] = (uint8_t) data->segment_length & 0xFF;
    dataoffset = 6;
  } else {
    //Encode length in the first and second byte
    sendbuffer[0] |= (data->segment_length >> 8) & 0xf; //Upper four bits into lower bits of first byte
    sendbuffer[1] = data->segment_length & 0xFF; //Lower 8 bits in second byte
    dataoffset = 2;
  }

  //Reset buffer position pointer
  data->segment_position = 0;

  //Copy data into sendbuffer
  memcpy(&sendbuffer[dataoffset],&buffer[data->segment_position],len-dataoffset);

  //Increment segment position
  data->segment_position = len - dataoffset;
  
  //Increment sequence counter
  data->seq++;

  //Init CF counter
  data->cf_counter = 0;
  
  //No padding needed, because we will always send a full frame
  return _isotp_send(data,sendbuffer,len);
}

/********************************************************************/
int _isotp_recv_FF(
		   cs_isotp_t *data,
		   uint8_t *recvbuffer, 
		   int recvlen,
		   uint8_t *buffer,
		   uint32_t len
		   )
/********************************************************************/
{
  uint32_t locallen = 0;
  int dataoffset = 0;
  int datalen = 0;
  
  //Set rx_dl based on the given frame size
  data->rx_dl =_isotp_padlen(recvlen);
  
  _isotp_verbose("Set RX_DL to %d bytes\n",data->rx_dl);

  //Set storage to invalid values
  data->segment_position = 0;
  data->segment_length = 0;

  //Data has to fill the complete frame
  if(recvlen != data->rx_dl){
    _isotp_error("Invalid data length\n");
    data->lerror = ISOTP_ERROR_FF;
    return CS_FAIL;
  }
  
  //Get length the traditional way
  locallen = ((uint32_t)recvbuffer[0] & 0xF) << 8 ;
  locallen += (uint32_t)recvbuffer[1];

  if(locallen > 0){
    //Length is not there, so coded the traditional way
    dataoffset = 2;
  } else {
    //Length is encoded in for bytes for CAN-FD frames > ISOTP_FRAME_CAN_MAX bytes
    locallen = ((uint32_t)recvbuffer[2]) << 24;
    locallen += ((uint32_t)recvbuffer[3]) << 16;
    locallen += ((uint32_t)recvbuffer[4]) << 8;
    locallen += ((uint32_t)recvbuffer[5]);
    dataoffset = 6;
  }
  
  //Check for maximum data length
  if((locallen > data->dl_max)  || (locallen > len)){
    _isotp_error("Overflow data length received\n");
    data->lerror = ISOTP_ERROR_OVERFLOW;
    return CS_FAIL;
  }

  //Get datalen in frame
  datalen = recvlen -dataoffset;
  
  //Copy the data
  memcpy(&buffer[data->segment_position],&recvbuffer[dataoffset],datalen);

  //Set complete segment length
  data->segment_length = locallen;

  //Set already received bytes
  data->segment_position = datalen;

  //precalc number of packets to receive
  data->cf_frames = (data->segment_length - data->segment_position) / (data->rx_dl -1);
  //Number of bytes in last packet
  data->cf_last = (data->segment_length - data->segment_position) % (data->rx_dl -1);

  //If there are bytes in last packet, we have to receive one more packet
  if(data->cf_last > 0){
    data->cf_frames += 1;
  } else {
    //Last frame is full with all  Bytes
    data->cf_last = data->rx_dl - 1;
  }

  //Init CF counter
  data->cf_counter = 0;
  
  _isotp_verbose("Number of CF-Frames to receive: %d\n",data->cf_frames);
  _isotp_verbose("Size of last CF-Frames to receive: %d\n",data->cf_last);
    
  return CS_OK;
}

/********************************************************************/
int _isotp_send_CF(
		   cs_isotp_t *data,
		   uint8_t *buffer
		   )
/********************************************************************/
{
  uint8_t sendbuffer[data->tx_dl];
  int sendlen = data->segment_length - data->segment_position;
  int dataoffset = 1;
  
  //Check for zero length or unsigned int overflow
  //Should not happen, but we check it
  if((sendlen <= 0) || (sendlen > data->dl_max)){
    data->lerror = ISOTP_ERROR_CF;
    _isotp_error("Invalid buffer length\n");
    return CS_FAIL;
  }

  //If bytes to send are more than max frame length
  //we have to reduce the frame length and send another frame
  if(sendlen > (data->tx_dl - dataoffset))
    sendlen = (data->tx_dl - dataoffset);

  //Reset complete sendbuffer with padbyte
  memset(sendbuffer,data->padbyte,data->tx_dl);

  //Set packet ID
  sendbuffer[0] = ISOTP_CF_ID;

  //Add sequence number
  sendbuffer[0] |= (uint8_t)((data->seq) & 0xF); // Sequence counter 0-15

  //Copy data
  memcpy(&sendbuffer[dataoffset],&buffer[data->segment_position],sendlen);

  //Store segment position offset
  data->segment_position += sendlen;

  //Increment sequence
  data->seq += 1;

  //Increment BS counter
  data->cf_counter++;
  
  /*Send frame*/
  if(data->options & ISOTP_OPTION_TXPADDING){
    _isotp_verbose("Send with TXPADDING\n");
    return _isotp_send(data,sendbuffer,data->tx_dl);
  } else {
    _isotp_verbose("Send\n");
    return _isotp_send(data,sendbuffer,sendlen + 1);
  }
}

/********************************************************************/
int _isotp_recv_CF(
		   cs_isotp_t *data,
		   uint8_t *recvbuffer, 
		   int recvlen,
		   uint8_t *buffer,
		   uint32_t len
		   )
/********************************************************************/
{
  int datalen = 0;
  int dataoffset = 1;
  uint8_t localseq = 0;
  
  if((recvlen < 2) || (recvlen > data->rx_dl)){
    _isotp_error("Invalid data length\n");
    data->lerror = ISOTP_ERROR_CF;
    data->segment_length = 0;
    return CS_FAIL;
  }
  
  //Increment packet counter
  data->cf_counter++;

  //Check for last frame
  if(data->cf_counter < data->cf_frames){
    //This is not the last frame
    //We have to use all bytes in the frame
    if(recvlen != data->rx_dl){
      _isotp_error("Invalid packet size\n");
      data->lerror = ISOTP_ERROR_CF;
      data->segment_length = 0;
      return CS_FAIL;
    }
    datalen = recvlen - dataoffset;  
  } else {
    //This is the last frame
    //The number of bytes in this frame was precalculated 
    datalen = data->cf_last;
    if(datalen > (recvlen - dataoffset)){
      _isotp_error("Invalid Receive length %d bytes for last frame, precalced %d bytes\n",recvlen,datalen);
      data->lerror = ISOTP_ERROR_RXLENGTH;
      data->segment_length = 0;
      return CS_FAIL;
    }

    //If received the last frame, we have todo padding check
    //Optional check padding
    if ((data->options & ISOTP_CHECK_PADDING) &&
	_isotp_check_pad(data,recvbuffer ,recvlen,dataoffset + datalen)){
      _isotp_error("Padding check failed\n");
      data->segment_length = 0;
      data->lerror = ISOTP_ERROR_SF;
      return CS_FAIL;
    }    
  }

  //Get sequence number
  localseq = recvbuffer[0] & 0xF;

  //Compare against old sequence number incremented by one and modulo
  if(localseq != ((data->seq +1)&0xF)){
    data->lerror = ISOTP_ERROR_SEQUENCE;
    data->segment_length = 0;
    _isotp_error("Sequence failure detected\n");
    return CS_FAIL;
  }
  
  //Sequence is ok, increment
  data->seq = (data->seq + 1) & 0xF; 

  //Copy data into user buffer
  memcpy(&buffer[data->segment_position],&recvbuffer[dataoffset],datalen);

  //Increment user buffer position
  data->segment_position += datalen;

  return CS_OK;
}

/********************************************************************/
int _isotp_lib_recv(
		    cs_isotp_t *data,
		    uint8_t *buffer, 
		    uint32_t *len,
		    uint32_t timeout
		    )
/********************************************************************/
{
  //CAN receive buffer
  //Note for CAN-FD the buffer size start with the maximum
  //possible size of 64 byte and is reduced to the received FF frame length
  uint8_t recvbuffer[data->rx_dl];
  int recvlen = data->rx_dl;
  //storage for the original buffer length
  int store_recvlen = recvlen;
  int rvalue;
  int rv;
  //This sleep time is only for testing the wait FC frame
  struct timespec sleeptime;
  
  if(data == NULL){
    return CS_FAIL;
  }
  
  if(buffer == NULL){
    data->lerror = ISOTP_ERROR_PARAMETER;
    return CS_FAIL;
  }

  if(len == NULL){
    data->lerror = ISOTP_ERROR_PARAMETER;
    return CS_FAIL;
  }

  //Reset user buffer
  memset(buffer,0,*len*sizeof(uint8_t));
  
  //Reset recvbuffer over the original length
  memset(recvbuffer,0,store_recvlen*sizeof(uint8_t));
  
  switch(_isotp_recv(data,recvbuffer,&recvlen,timeout)){
  case ISOTP_SF:
    rv = _isotp_recv_SF(data,recvbuffer,recvlen,buffer,*len);
    //Set received length
    *len = data->segment_length;
    if(data->transfercb != NULL){
      (*data->transfercb)(data->transfercb_ctx,ISOTP_SF_ID,data->segment_length,data->cf_counter,data->segment_position);
    }
    return rv;
  case ISOTP_FF:
    data->txbs = 0;
    if(_isotp_recv_FF(data,recvbuffer,recvlen,buffer,*len)){
      if(data->lerror == ISOTP_ERROR_OVERFLOW){
	//Received correct but datalen overrange
	//Send overflow FC frame and stop
	_isotp_send_FC(data,1,0);
      }
      *len = 0;
      return CS_FAIL;
    }
    if(data->transfercb != NULL){
      (*data->transfercb)(data->transfercb_ctx,ISOTP_FF_ID,data->segment_length,data->cf_counter,data->segment_position);
    }
    _isotp_verbose("Recv FF Frame Data length: %i, Data index %i, Recv len %i\n",data->segment_length, data->segment_position, recvlen);
    
    if(_isotp_send_FC(data,0,0)){
      *len = 0;
      return CS_FAIL;
    }
    if(data->transfercb != NULL){
      (*data->transfercb)(data->transfercb_ctx,ISOTP_FC_ID,data->segment_length,data->cf_counter,data->segment_position);
    }
    while(1){
      //Reset recvbuffer over the original length
      memset(recvbuffer,0,store_recvlen*sizeof(uint8_t));
      //Reset length to received data length,
      //for CAN-FD defined by FF frame
      //for CAN 2.0 defined fixed to 8 Byte
      recvlen = data->rx_dl;
      rvalue = _isotp_recv(data,recvbuffer,&recvlen,ISOTP_TIMEOUT_CR);
      switch(rvalue){
      case CS_FAIL:
      default:
	_isotp_error("Recieve failure\n");
	*len = 0;
	return CS_FAIL;
	break;
      case ISOTP_SF:
      case ISOTP_FF:
      case ISOTP_FC:
	//Ignore frame
	_isotp_error("Recieve unexpected frame type\n");
	break;
      case ISOTP_CF: 	
	if(_isotp_recv_CF(data,recvbuffer,recvlen,buffer,*len)){
	  *len = 0;
	  return CS_FAIL;
	}
	if(data->transfercb != NULL){
	  (*data->transfercb)(data->transfercb_ctx,ISOTP_CF_ID,data->segment_length,data->cf_counter,data->segment_position);
	}
	_isotp_verbose("Recv CF Frame: Counter: %i Seq: 0x%X, Data index: %i, Recv len: %i\n",
		       data->cf_counter,data->seq,data->segment_position,recvlen-1);
		
	if(data->segment_position >= data->segment_length){
	  //Store length
	  *len = data->segment_length;
	  _isotp_verbose("End of segment detected\n");
	  return CS_OK;
	}
	if((data->rxbs > 0) && ((data->cf_counter % data->rxbs) == 0)){
	  if((data->options & ISOTP_OPTION_RXTESTING) &&
	     ((random()%10) == 0)){
	    //random apply a Wait FC frame
	    //only for testing
	    if(_isotp_send_FC(data,0,1)){
	      *len = 0;
	      return CS_FAIL;
	    }
	    if(data->transfercb != NULL){
	      (*data->transfercb)(data->transfercb_ctx,ISOTP_FC_ID,data->segment_length,data->cf_counter,data->segment_position);
	    }
	    sleeptime.tv_sec = 0;
	    sleeptime.tv_nsec = 500e6; //500ms
	    nanosleep(&sleeptime, NULL); 
	  }
	  //No send the normal FC frame
	  //Send FC frame
	  if(_isotp_send_FC(data,0,0)){
	    *len = 0;
	    return CS_FAIL;
	  }
	  if(data->transfercb != NULL){
	    (*data->transfercb)(data->transfercb_ctx,ISOTP_FC_ID,data->segment_length,data->cf_counter,data->segment_position);
	  }
	}
	break;
      }
    }
    return CS_OK;
  default:
    //Failure
    //Reset buffer
    memset(buffer,0,*len);
    //Reset len
    *len = 0;
    if(data->lerror != ISOTP_ERROR_TIMEOUT){
      _isotp_error("Recieve unexpected frame: %s\n",cs_isotp_getlasterrorstring(data));
    }
    if(data->lerror == ISOTP_ERROR_NOERROR){
      data->lerror = ISOTP_ERROR_UNEXPECTED;
    }
    return CS_FAIL;
  }  
}

/********************************************************************/
int _isotp_lib_send(
		    cs_isotp_t *data,
		    uint8_t *buffer
		    )
/********************************************************************/
{
  int rvalue;
  int loop = 1;
  struct timespec stmintime;

  
  if((data->segment_length <= (ISOTP_LL_TX_DL_8 -1)) || /*CAN 2.0 and short CAN-FD frame*/
      (data->segment_length <= (data->tx_dl -2))        /*CAN-FD > 8 Byte frame*/
      ){
    /*Small frame direct sending*/
    _isotp_verbose("SF-Frame Data len: %i\n", data->segment_length);
    rvalue = _isotp_send_SF(data,buffer);
    if(data->transfercb != NULL){
      (*data->transfercb)(data->transfercb_ctx,ISOTP_SF_ID,data->segment_length,data->cf_counter,data->segment_position);
    }
    return rvalue;
  } else {
    /*Prepare first message*/
    if(_isotp_send_FF(data,buffer)){
      return CS_FAIL;
    }
    if(data->transfercb != NULL){
      (*data->transfercb)(data->transfercb_ctx,ISOTP_FF_ID,data->segment_length,data->cf_counter,data->segment_position);
    }
    _isotp_verbose("FF-Frame Data len: %i, Seq: 0x%X Data index %i\n",
		   data->segment_length, data->seq, data->segment_position);
    /*Wait for FC-Frame*/
    loop = 1;
    while(loop){
      rvalue = _isotp_recv_FC(data);
      if(data->transfercb != NULL){
	(*data->transfercb)(data->transfercb_ctx,ISOTP_FC_ID,data->segment_length,data->cf_counter,data->segment_position);
      }
      switch(rvalue){
      case ISOTP_WAIT:
	//Receive Wait
	//Wait for next FC frame
	_isotp_debug("Wait\n");
	break;
      case ISOTP_IGNORE:
	_isotp_debug("Ignore\n");
	break;
      case CS_FAIL:
	_isotp_error("Failure\n");
	return CS_FAIL;
      case CS_OK:
	loop = 0;
	break;
      default:
	_isotp_error("Recv FC, unhandled return code: 0x%X",rvalue);
	return CS_FAIL;
	break;
      }
    }
    stmintime = _stim2ts(data->txstim);
    while(1){
      if(_isotp_send_CF(data,buffer)){
	return CS_FAIL;
      }
      if(data->transfercb != NULL){
	(*data->transfercb)(data->transfercb_ctx,ISOTP_CF_ID,data->segment_length,data->cf_counter,data->segment_position);
      }
      _isotp_verbose("CF-Frame Seq: 0x%X, Data index %i\n",data->seq, data->segment_position);
      if(data->segment_position >= data->segment_length){
	_isotp_verbose("Finished Paket\n");
	return CS_OK;
      }
      if((data->txbs > 0) && ((data->cf_counter % data->txbs) == 0)){
	/*Wait for FC-Frame*/
	loop = 1;
	while(loop){
	  rvalue = _isotp_recv_FC(data);
	  if(data->transfercb != NULL){
	    (*data->transfercb)(data->transfercb_ctx,ISOTP_FC_ID,data->segment_length,data->cf_counter,data->segment_position);
	  }	  
	  switch(rvalue){
	  case ISOTP_WAIT:
	    _isotp_debug("Wait\n");
	    break;
	  case ISOTP_IGNORE:
	    _isotp_debug("Ignore\n");
	    break;
	  case CS_FAIL:
	    _isotp_error("Failure\n");
	    return CS_FAIL;
	  case CS_OK:
	    loop = 0;
	    break;
	  default:
	    _isotp_error("Recv FC, unhandled return code: 0x%X",rvalue);
	    return CS_FAIL;
	  }
	}
	stmintime = _stim2ts(data->txstim);
      } 
      if(data->txstim != 0){
	nanosleep(&stmintime, NULL); //Sleep STMIN
      }
    }
  }
  return CS_OK;
}


/********************************************************************/
int cs_isotp_send(
		  cs_isotp_t *data,
		  uint8_t *buffer, 
		  uint32_t len
		  )
/********************************************************************/
{

  if(data == NULL)
    return CS_FAIL;

  data->lerror = ISOTP_ERROR_NOERROR;
  
  if((buffer == NULL)||(len == 0)){
    _isotp_error("Parameter failure\n");
    data->lerror = ISOTP_ERROR_PARAMETER;
    return CS_FAIL;
  }
  
  if(len > data->dl_max){
    _isotp_error("Buffer length is out of range\n");
    data->lerror = ISOTP_ERROR_PARAMETER;
    return CS_FAIL;
  }
  
  //rx_dl is set to tx_dl for sender
  data->rx_dl = data->tx_dl;
  //Set flag that we are waiting for first FC frame
  data->tx_first_FC = 1;
  //Reset sequence number
  data->seq = 0;
  //Reset data segment position offset
  data->segment_position = 0;
  //Store length
  data->segment_length = len;
  //Reset BS counter
  data->cf_counter = 0;
  
  return _isotp_lib_send(data,buffer);
  
}

/********************************************************************/
int cs_isotp_recv(
		  cs_isotp_t *data,
		  uint8_t *buffer, 
		  uint32_t *len,
		  uint32_t timeout
		  )
/********************************************************************/
{
  if(data == NULL)
    return CS_FAIL;

  data->lerror = ISOTP_ERROR_NOERROR;

  //Reset rx_dl
  //For CAN 2.0 this is a fix 8 Byte length
  //For CAN-FD is detected by received frame length of the first
  //frame received, so we are starting with 64 Byte length (maximum possible)
  //Note if we received a CAN-FD frame type SF or FF with 8 Bytes
  //it is handled the traditonal why (length coding,...)
  if(data->ll_mode == ISOTP_LL_MODE_CAN){
    data->rx_dl = ISOTP_LL_TX_DL_8;
  } else {
    data->rx_dl = ISOTP_LL_TX_DL_64;
  }

  //Reset sequence number
  data->seq = 0;
  //Reset data segment position offset
  data->segment_position = 0;
  //Reset data segment length
  data->segment_length = 0;
  //Reset BS counter
  data->cf_counter = 0;
  
  if((buffer == NULL)||(len == NULL)||(*len== 0)){
    _isotp_error("Parameter failure\n");
    data->lerror = ISOTP_ERROR_PARAMETER;
    return CS_FAIL;
  }

  return _isotp_lib_recv(data,buffer,len,timeout);
}
