/*  
    Copyright (C) 2016 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "config.h"
#include "cs_isotp_decoder.h"
#include "cs_isotp_private.h"
#include "cs_defines.h"

/* \def STRBUFFER
   Length of string buffer
*/
#define STRBUFFER (ISOTP_FRAME_CANFD_MAX * 4 + 1)

/** \struct cs_isotp_decoder
    IsoTP decoder storage structure
*/
struct cs_isotp_decoder{
  uint32_t sa;                          /**< Source address*/
  uint32_t ta;                          /**< Target address*/
  uint32_t trigger_id;                  /**< Who triggered communication*/
  uint8_t data[ISOTP_FRAME_CANFD_MAX];  /**< Local data storage*/
  uint32_t data_len;                         /**< Local data len*/
  uint32_t data_act;                         /**< Actual data index*/
  void (*fout)(void *priv_data,
	       double ts,
	       uint32_t id,
	       const char *fmt, ...);   /**< Print out function callback*/
  void (*ferror)(void *priv_data,
		 double ts,
		 uint32_t id,
		 const char *fmt, ...); /**< Print out error function callback, May be NULL*/
  void (*fdata)(void *priv_data,
		double ts,
		uint32_t id,
		uint8_t *data,
		int len,
		int error);             /**< Data function callback*/
  uint8_t state;                        /**< Internal state*/
  void *priv_data;                      /**< Private data storage pointer for data callback functions,
					   for high level stack handling*/
  void *priv_data_out;                 /**< Private data storage pointer for output driven callback functions*/
};


//external helper functions, defined in cs_isotp.c
extern uint32_t _stim2mues(uint8_t stim);


/********************************************************************/
void _isotp_decoder_reset_data(
			       cs_isotp_decoder_t * data
			       )
/********************************************************************/
{
  memset(data->data,0,sizeof(uint8_t)*ISOTP_FRAME_CANFD_MAX);
  data->data_act = 0;
}

/********************************************************************/
void _isotp_decoder_reset_state(
				cs_isotp_decoder_t * data
				)
/********************************************************************/
{
  _isotp_decoder_reset_data(data);
  data->state = ISOTP_WAIT;
  data->trigger_id = 0;
}

/********************************************************************/
int _isotp_decoder_check_id(
			    cs_isotp_decoder_t * data,
			    uint32_t id
			    )
/********************************************************************/
{
  if(id == data->sa){
    return CS_OK;
  }
  
  if(id == data->ta){
    return CS_OK;
  }
  return CS_FAIL;
}

/********************************************************************/
char *_isotp_decoder_buffer2str(char *str,
				uint8_t *buffer,
				int len
				)
/********************************************************************/
{
  int i;
  int lcounter= 0;
  //Note len is maximal ISOTP_FRAME_CANFD_MAX
  
  memset(str,0,sizeof(char)*STRBUFFER);
  for(i = 0; i < len; i++){
    lcounter += sprintf(&str[lcounter]," %.2X",buffer[i]);
  }
  //Return string 
  return str;
}

/********************************************************************/
int _isotp_decoder_check_data(
			      cs_isotp_decoder_t * data,
			      double ts,
			      uint32_t id,
			      uint8_t *bdata,
			      int len,
			      uint8_t **udata_start,
			      uint32_t *len_udata
			      )
/********************************************************************/
{
  int fc_bs;
  int fc_stim;
  int cf_seq;
  char lstr[STRBUFFER];
  int dataoffset = 0;
  
  *len_udata = 0;
  *udata_start = bdata;
  //Get type
  switch(bdata[0] & 0xF0){
  case ISOTP_SF_ID:
    //Check length
    if( len < 2){
      if(data->ferror != NULL){
	(*data->ferror)(data->priv_data_out,ts,id,"ISOTP, %s: %s",
			cs_isotp_geterroridstring(ISOTP_ERROR_SF),
			_isotp_decoder_buffer2str(lstr,bdata,len));
      }
      return ISOTP_IGNORE;
    }
    //Get content len
    *len_udata =  (uint32_t)(bdata[0] & ISOTP_SF_LMASK);
    dataoffset = 1;
    //if len is there length is encoded in the next byte
    if(*len_udata == 0){
      dataoffset = 2;
      *len_udata = (uint32_t)bdata[1];
    }
    
    //Get data start
    *udata_start = &bdata[dataoffset];
    //Check length
    if(*len_udata > (len -dataoffset)){
      if(data->ferror != NULL){
	(*data->ferror)(data->priv_data_out,ts,id,"ISOTP, %s: %s",
			cs_isotp_geterroridstring(ISOTP_ERROR_SF),
			_isotp_decoder_buffer2str(lstr,bdata,len));
      }
      return ISOTP_IGNORE;
    }
    if(data->fout != NULL){
      (*data->fout)(data->priv_data_out,ts,id,"ISOTP, SF-Frame: Len %i Bytes: %s",
		    *len_udata,
		    _isotp_decoder_buffer2str(lstr,*udata_start,*len_udata));
    }
    return ISOTP_SF;
  case ISOTP_FF_ID:
    //Check min length
    if(len < ISOTP_LL_TX_DL_8){
      if(data->ferror != NULL){
      	(*data->ferror)(data->priv_data_out,ts,id,"ISOTP, %s: %s",
			cs_isotp_geterroridstring(ISOTP_ERROR_FF),
			_isotp_decoder_buffer2str(lstr,bdata,len));
      }
      return ISOTP_IGNORE;
    }
    //get ISOTP frame length
    //First ols style
    data->data_len = 0;
    data->data_len = (bdata[0] & 0xF) << 8 ;
    data->data_len += bdata[1]; 
    dataoffset = 2;
    
    if(data->data_len == 0){
      //Expect CAN-FD encoding
      data->data_len = ((uint32_t)bdata[2]) << 24;
      data->data_len += ((uint32_t)bdata[3]) << 16;
      data->data_len += ((uint32_t)bdata[4]) << 8;
      data->data_len  += ((uint32_t)bdata[5]);
      dataoffset = 6;
    }
    
    if(data->data_len > ISOTP_FRAME_CANFD_MAX){
      if(data->ferror != NULL){
      	(*data->ferror)(data->priv_data_out,ts,id,"ISOTP, %s: %s",
			cs_isotp_geterroridstring(ISOTP_ERROR_DATA),
			_isotp_decoder_buffer2str(lstr,bdata,len));
      }
      return ISOTP_IGNORE;
    }
    //Get content len
    *len_udata =  len - dataoffset;
    //Get data start
    *udata_start = &bdata[dataoffset];
    if(data->fout != NULL){
      (*data->fout)(data->priv_data_out,ts,id,"ISOTP, FF-Frame: Complete Len %i Bytes,  Len: %i Bytes: %s",
		    data->data_len,
		    *len_udata,
		    _isotp_decoder_buffer2str(lstr,*udata_start,*len_udata));
    }
    return ISOTP_FF;
  case ISOTP_CF_ID:
    //Check min length
    if(len < 2){
      if(data->ferror != NULL){
      	(*data->ferror)(data->priv_data_out,ts,id,"ISOTP, %s: %s",
			cs_isotp_geterroridstring(ISOTP_ERROR_CF),
			_isotp_decoder_buffer2str(lstr,bdata,len));
      }
      return ISOTP_IGNORE;
    }
    cf_seq = bdata [0] & 0xF;
    //Get content len
    *len_udata =  len - 1;
    //Get data start
    *udata_start = &bdata[1];
    if(data->fout != NULL){
      (*data->fout)(data->priv_data_out,ts,id,"ISOTP, CF-Frame: Seq %i Len: %i Bytes: %s",
		    cf_seq,
		    *len_udata,
		    _isotp_decoder_buffer2str(lstr,*udata_start,*len_udata));
    }
    return ISOTP_CF;
  case ISOTP_FC_ID:
    //Check min length
    if(len < ISOTP_FC_LEN){
      if(data->ferror != NULL){
      	(*data->ferror)(data->priv_data_out,ts,id,"ISOTP, %s: %s",
			cs_isotp_geterroridstring(ISOTP_ERROR_FC),
			_isotp_decoder_buffer2str(lstr,bdata,len));
      }
      return ISOTP_IGNORE;
    }
    //Check for Wait
    if(bdata[0] & ISOTP_FC_WAIT){
      if(data->fout != NULL){
	(*data->fout)(data->priv_data_out,ts,id,"ISOTP, FC-Frame: Wait: %s",
		      _isotp_decoder_buffer2str(lstr,bdata,len));
      }
      return ISOTP_IGNORE;
    }
    //Check for overflow
    if(bdata[0] & ISOTP_FC_OVERFLOW){
      if(data->fout != NULL){
	(*data->fout)(data->priv_data_out,ts,id,"ISOTP, FC-Frame: Overflow: %s",
		      _isotp_decoder_buffer2str(lstr,bdata,len));
      }
      return ISOTP_IGNORE;
    }

    //get block size
    fc_bs = bdata[1];
    //get STMIN
    fc_stim = bdata[2];
    if(data->fout != NULL){
      (*data->fout)(data->priv_data_out,ts,id,"ISOTP, FC-Frame: Blocksize %i, STIM %i ms",
		    fc_bs,
		    _stim2mues(fc_stim));
    }
    return ISOTP_FC;
  default:
    if(data->ferror != NULL){
      (*data->ferror)(data->priv_data_out,ts,id,"ISOTP, %s: %s",
		      cs_isotp_geterroridstring(ISOTP_ERROR_UNKNOWN),
		      _isotp_decoder_buffer2str(lstr,bdata,len));
    }
    return ISOTP_IGNORE;
  }
}

/********************************************************************/
int _isotp_decoder_copy_data(
	       cs_isotp_decoder_t * data,
	       double ts,
	       uint32_t id,
	       uint8_t *udata_start,
	       int len_udata
	       )
/********************************************************************/
{
  if((data->data_act + len_udata) >= ISOTP_FRAME_CANFD_MAX){
    if(data->ferror != NULL){
      (*data->ferror)(data->priv_data_out,ts,id,"ISOTP, %s: %s",
		      cs_isotp_geterroridstring(ISOTP_ERROR_DATA));
    }
    return CS_FAIL;
  }
  memcpy(&data->data[data->data_act],udata_start,len_udata);
  data->data_act += len_udata;
  //printf("Data act: %i Bytes\n",data->data_act);
  return CS_OK;
}
  
/********************************************************************/
cs_isotp_decoder_t *cs_isotp_decoder_init(
					  uint32_t sa,
					  uint32_t ta,
					  void (*fout)(void *priv_data_out,
						       double ts,
						       uint32_t id,
						       const char *fmt, ...),
					  void (*ferror)(void *priv_data_out,
							 double ts,
							 uint32_t id,
							 const char *fmt, ...),
					  void (*fdata)(void *priv_data,
							double ts,
							uint32_t id,
							uint8_t *data,
							int len,
							int error),
					  void *priv_data,
					  void *priv_data_out
					  )
/********************************************************************/
{
  cs_isotp_decoder_t *data = NULL;

  if(fdata == NULL){
    return NULL;
  }

  data = (cs_isotp_decoder_t*)malloc(sizeof(cs_isotp_decoder_t));
  if(data == NULL)
    return NULL;

  //Store CAN-ids
  data->sa = sa;
  data->ta = ta;

  //Store handles
  data->fout = fout;
  data->ferror = ferror;
  data->fdata = fdata;

  //Store priv data
  data->priv_data = priv_data;
  data->priv_data_out = priv_data_out;

  //Init state and data
  _isotp_decoder_reset_state(data);

  return data;
}

/********************************************************************/
int cs_isotp_decoder_reset(
			   cs_isotp_decoder_t *data
			   )
/********************************************************************/
{
  if( data == NULL ){
    return CS_FAIL;
  }
  _isotp_decoder_reset_state(data);
  return CS_OK;
}

/********************************************************************/
int cs_isotp_decoder_data(
			  cs_isotp_decoder_t *data,
			  double ts,
			  uint32_t id,
			  uint8_t *bdata,
			  int len
			  )
/********************************************************************/
{
  int new_state;
  uint8_t *udata_start;
  uint32_t len_udata;
  char lstr[STRBUFFER];
  
  if(data == NULL){
    return CS_FAIL;
  }

  if(bdata == NULL){
    return CS_FAIL;
  }

  if(_isotp_decoder_check_id(data,id)){
    //CAN ID does not fit
    return CS_OK;
  }
  new_state = _isotp_decoder_check_data(data,ts,id,bdata,len,&udata_start,&len_udata);

  if(new_state == ISOTP_IGNORE){
    //Ignore frame
    return CS_OK;
  }
  
  switch(data->state){
  case ISOTP_WAIT:
    switch(new_state){
    case ISOTP_SF:
      if(_isotp_decoder_copy_data(data,ts,id,udata_start,len_udata)){
	//Copy failed
	_isotp_decoder_reset_state(data);
	(*data->fdata)(data->priv_data,ts,id,data->data,0,ISOTP_ERROR_DATA);
	break;
      }
      //Emit data
      (*data->fdata)(data->priv_data,ts,id,data->data,data->data_act,ISOTP_ERROR_NOERROR);
      _isotp_decoder_reset_state(data);
      break;
    case ISOTP_FF:
      if(_isotp_decoder_copy_data(data,ts,id,udata_start,len_udata)){
	//Copy failed
	_isotp_decoder_reset_state(data);
	(*data->fdata)(data->priv_data,ts,id,data->data,0,ISOTP_ERROR_DATA);
	break;
      }
      data->trigger_id = id;
      data->state = ISOTP_FF;
      break;
    default:
      if(data->ferror != NULL){
	(*data->ferror)(data->priv_data_out,ts,id,"ISOTP (S00), %s: %s",
			cs_isotp_geterroridstring(ISOTP_ERROR_UNEXPECTED),
			_isotp_decoder_buffer2str(lstr,bdata,len)
			);
      }
      _isotp_decoder_reset_state(data);
      (*data->fdata)(data->priv_data,ts,id,data->data,0,ISOTP_ERROR_UNEXPECTED);
      break;
    }
    break;
  case ISOTP_FF:
    switch(new_state){
    case ISOTP_FC:
      //No data to copy
      data->state = ISOTP_FC;
      break;
    default:
      if(data->ferror != NULL){
	(*data->ferror)(data->priv_data_out,ts,id,"ISOTP (SFF), %s: %s",
			cs_isotp_geterroridstring(ISOTP_ERROR_UNEXPECTED),
			_isotp_decoder_buffer2str(lstr,bdata,len));
      }
      _isotp_decoder_reset_state(data);
      (*data->fdata)(data->priv_data,ts,id,data->data,0,ISOTP_ERROR_UNEXPECTED);
      break;
    }
    break;
  case ISOTP_FC:
  case ISOTP_CF:
    switch(new_state){
    case ISOTP_CF:
      if(_isotp_decoder_copy_data(data,ts,id,udata_start,len_udata)){
	//Copy failed
	_isotp_decoder_reset_state(data);
	(*data->fdata)(data->priv_data,ts,id,data->data,0,ISOTP_ERROR_DATA);
	break;
      }
      if((data->data_act) >= data->data_len){
	//That was the last frame
	//Emit data
	(*data->fdata)(data->priv_data,ts,data->trigger_id,data->data,data->data_act,ISOTP_ERROR_NOERROR);
	_isotp_decoder_reset_state(data);
	break;
      } else {
	data->state = ISOTP_CF;
      }
      break;
    case ISOTP_FC:
      //Nothing todo, there is no data
      break;
    default:
      if(data->ferror != NULL){
	(*data->ferror)(data->priv_data_out,ts,id,"ISOTP (SCFFC), %s: %s",
			cs_isotp_geterroridstring(ISOTP_ERROR_UNEXPECTED),
			_isotp_decoder_buffer2str(lstr,bdata,len));
      }
      _isotp_decoder_reset_state(data);
      (*data->fdata)(data->priv_data,ts,id,data->data,0,ISOTP_ERROR_UNEXPECTED);
      break;
    }
    break;
  default:
    if(data->ferror != NULL){
      (*data->ferror)(data->priv_data_out,ts,id,"ISOTP (Unknown), %s: %s",
		      cs_isotp_geterroridstring(ISOTP_ERROR_UNEXPECTED),
		      _isotp_decoder_buffer2str(lstr,bdata,len));
    }
    _isotp_decoder_reset_state(data);
    (*data->fdata)(data->priv_data,ts,id,data->data,0,ISOTP_ERROR_UNEXPECTED);
    break;
  }
  return CS_OK;
}

/********************************************************************/
int cs_isotp_decoder_close(
			   cs_isotp_decoder_t *data
			   )
/********************************************************************/
{
  if( data == NULL ){
    return CS_FAIL;
  }
  free(data);
  return CS_OK;
}
