/*
    Copyright (C) 2014 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @file cs_isotp_private.h
 * @author Andreas Stewering-Bone
 * @date 9 Oct 2014
 * @brief File containing an ISOTP userspace Layer, private definitions.
 *
 * File containing an ISOTP userspace Layer, private definitions.
 * Limitations:
 * - Only half duplex
 */
#ifndef CS_ISOTP_PRIVATE_H
#define CS_ISOTP_PRIVATE_H

/** \def CS_ISOTP_WAIT
    ISOTP ID Wait
*/
#define ISOTP_WAIT 0x02
/** \def ISOTP_FINISHED
    ISOTP ID Finished transmission
*/
#define ISOTP_FINISHED 0x03
/** \def ISOTP_SF
    ISOTP internal ID Single frame
*/
#define ISOTP_SF 0x04
/** \def ISOTP_FF
    ISOTP internal ID First frame
*/
#define ISOTP_FF 0x05
/** \def ISOTP_CF
    ISOTP internal ID Conductive frame
*/
#define ISOTP_CF 0x06
/** \def ISOTP_FC
    ISOTP internal ID Flow control frame
*/
#define ISOTP_FC 0x07
/** \def ISOTP_IGNORE
    Ignore current frame
*/
#define ISOTP_IGNORE 0x08

/** \def ISOTP_SF_ID
    ISOTP ID Single frame
*/

/** \def ISOTP_SF_LMASK
    ISOTP mask for SF length encoding
*/
#define ISOTP_SF_LMASK 0x0F

/** \def ISOTP_FC_OVERFLOW
    ISOTP FC signal overflow
*/
#define ISOTP_FC_OVERFLOW 0x02

/** \def ISOTP_FC_WAIT
    ISOTP FC signal wait
*/
#define ISOTP_FC_WAIT 0x01

/** \def ISOTP_CF_MAX
    ISOTP max data length of CF frame in bytes
*/
#define ISOTP_CF_MAX 0x07

/** \def ISOTP_FF_MAX
    ISOTP max data length of FF frame in bytes
*/
#define ISOTP_FF_MAX 0x06

/** \def ISOTP_FC_LEN
    Length of a minimal FC frame
*/
#define ISOTP_FC_LEN 3

/** \def ISOTP_TIMEOUT_BS
    Timeout waiting for FC frame in ms
*/
#define ISOTP_TIMEOUT_BS 1000

/** \def ISOTP_TIMEOUT_CR
    Timeout waiting for CF frame in ms
*/
#define ISOTP_TIMEOUT_CR 1000

/** \def ISOTP_FRAME_CAN_MAX
    ISOTP max length of data frame (CAN 2.0) in bytes
*/
#define ISOTP_FRAME_CAN_MAX 4095

/** \def ISOTP_FRAME_CANFD_MAX
    ISOTP max length of data frame (CAN-FD) in bytes
    Actual limited 
    Theoretical maximum (length is coded in 4 Byte value) is
    4294967295 Bytes
*/
#define ISOTP_FRAME_CANFD_MAX 16383

/** \def ISOTP_RXSTIM
    Default RX seperation time in 0.000001s, used for timeout
*/
#define ISOTP_RXSTIM 1000

/** \def ISOTP_RXBS
    Default RX Block size
*/
#define ISOTP_RXBS 20

/** \def ISOTP_CHECK_PADDING
    Helper define for padding check
*/
#define ISOTP_CHECK_PADDING (ISOTP_OPTION_RXLENGTH | ISOTP_OPTION_RXDATA)

/** \struct cs_isotp
    Central stack structure, private definiton
*/
struct cs_isotp{
  cs_socketcan_t *socket;               /**< SocketCAN interface handle*/
  uint8_t txstim;                       /**< Transmit STIM time in 0.000001s*/
  uint8_t rxstim;                       /**< Receive STIM time in 0.000001s*/
  uint32_t sa;                          /**< Source address*/
  uint32_t ta;                          /**< Target address*/
  uint8_t txbs;                         /**< Block size definition*/
  uint8_t rxbs;                         /**< Block size definition*/
  uint8_t lerror;                       /**< Last error code*/
  uint8_t options;                      /**< Option flag*/
  uint8_t padbyte;                      /**< Padding byte content, default 0x00*/
  uint32_t cf_frames;                   /**< Number of CF-Frames to receive*/
  uint32_t cf_last;                     /**< Number of bytes in last CF-Frame*/
  uint32_t cf_counter;                  /**< CF-Frame counter*/
  cs_isotp_ll_mode_t ll_mode;           /**< Link layer running mode*/
  cs_isotp_ll_tx_dl_t tx_dl;            /**< Transmit frame length, default 8 Byte (CAN 2.0 compatible)*/
  cs_isotp_ll_tx_dl_t rx_dl;            /**< Receive frame length, default 8 Byte (CAN 2.0 compatible)*/
  int dl_max;                           /**< Maximum datalength for transfer buffer*/
  cs_socketcan_send_mode_t ll_send_mode;/**< Link layer send mode*/
  uint8_t tx_first_FC;                   /**< Storage flag for first FC on sender side*/
  uint8_t seq;                           /**< Storage for frame sequence number 0-15*/
  uint32_t segment_position;             /**< Buffer offset in data segment to handle in bytes*/
  uint32_t segment_length;               /**< data segment length in bytes*/
  void (*transfercb)(void *ctx,uint8_t mode, uint32_t length, uint32_t counter, uint32_t actbytes); /** <Callback which is called after CF Frame send or receive*/
  void *transfercb_ctx;                  /**< Callback function context pointer*/
};


#endif
