/*  
    Copyright (C) 2017 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "config.h"
#include "cs_defines.h"
#include "cs_lin_private.h"
#include "cs_lin.h"
#include "cs_time.h"
#include "cs_byteorder.h"
#include "cs_debug_private.h"

/** \def PHYS_SET_BIT
    Set bits in bit field
    X is bit field
    Y is offset in bits

*/
#define PHYS_SET_BIT(X,Y) (x[(Y)/8] | (1<<((Y)%8)))

/** Mapping type of Entry types to predefined frame names
    Please care of order, it has to be order of enum cs_lin_schedule_entry_type_t
*/
const char * cs_lin_ldf_schedentrytypemap[] = {
    "MasterReq",
    "SlaveResp",
    "AssignNAD",
    "ConditionalChangeNAD",
    "DataDump",
    "SaveConfiguration",
    "AssignFrameIdRange",
    "FreeFormat",
    "AssignFrameId",
};

#define _lin_error(...) (cs_error(CS_DEBUG_LIN,__func__,__LINE__,__VA_ARGS__))
#define _lin_warn(...) (cs_warn(CS_DEBUG_LIN,__func__,__LINE__,__VA_ARGS__))
#define _lin_info(...) (cs_info(CS_DEBUG_LIN,__func__,__LINE__,__VA_ARGS__))
#define _lin_debug(...) (cs_debug(CS_DEBUG_LIN,__func__,__LINE__,__VA_ARGS__))
#define _lin_verbose(...) (cs_verbose(CS_DEBUG_LIN,__func__,__LINE__,__VA_ARGS__))
#define _lin_error_buffer(DATA,LEN) (cs_error_buffer(CS_DEBUG_LIN,__func__,__LINE__,DATA,LEN))
#define _lin_warn_buffer(DATA,LEN) (cs_warn_buffer(CS_DEBUG_LIN,__func__,__LINE__,DATA,LEN))
#define _lin_info_buffer(DATA,LEN) (cs_info_buffer(CS_DEBUG_LIN,__func__,__LINE__,DATA,LEN))
#define _lin_debug_buffer(DATA,LEN) (cs_debug_buffer(CS_DEBUG_LIN,__func__,__LINE__,DATA,LEN))
#define _lin_verbose_buffer(DATA,LEN) (cs_verbose_buffer(CS_DEBUG_LIN,__func__,__LINE__,DATA,LEN))


/********************************************************************/
cs_debug_level_t cs_lin_debuglevel(
				   cs_debug_level_t level
				   )
/********************************************************************/
{
  return cs_debuglevel(CS_DEBUG_LIN,level);
}

/********************************************************************/
void cs_lin_debug_setcallback(
    void *ctx,
    void (*callb)(void *ctx,
		  cs_debug_sub_t sub,
		  cs_debug_level_t dl,
		  const char *func,
		  const int line,
		  const char *fmt,
		  va_list ap)
    )
/********************************************************************/
{
  cs_debug_setcallback(CS_DEBUG_LIN,ctx,callb);
}


/********************************************************************/
void _lin_debug_conv(
    cs_lin_raw_t *data,
    int len
    )
/********************************************************************/
{
#ifdef CS_DEBUG
    int i;
    int bufferlen = len*10+1; // XX_
    char buffer[bufferlen];
    char *ptr = &buffer[0];
    memset(buffer,0,sizeof(char)*bufferlen);

    for(i = 0; i < len; i++){
        ptr += sprintf (ptr, "%04X ", data[i]);
    }
    _lin_debug("Buffer len: %i Data; %s\n",len,buffer);
#endif
}


/********************************************************************/
void _lin_verbose_conv(
    cs_lin_raw_t *data,
    int len
    )
/********************************************************************/
{
#ifdef CS_VERBOSE
    int i;
    int bufferlen = len*10+1; // XX_
    char buffer[bufferlen];
    char *ptr = &buffer[0];
    memset(buffer,0,sizeof(char)*bufferlen);

    for(i = 0; i < len; i++){
        ptr += sprintf (ptr, "%04X ", data[i]);
    }
    _lin_verbose("Buffer len: %i Data; %s\n",len,buffer);
#endif
}

/********************************************************************/
void * _lin_ldf_cpf(void *ext)
/********************************************************************/
{
    return ext;
}

/********************************************************************/
void _lin_ldf_free_generic(void *n)
/********************************************************************/
{
    if(n == NULL)
        return;

    free(n);	     
}

/********************************************************************/
void _lin_ldf_free_signalencoding(void *n)
/********************************************************************/
{
    cs_lin_enc_t* l = (cs_lin_enc_t*)n;

    if(l == NULL)
        return;
    cs_lists_destroy(&l->l_logical);
    cs_lists_destroy(&l->l_physical);
    free(l);	     
}

/********************************************************************/
void _lin_ldf_free_signalrepresentation(void *n)
/********************************************************************/
{
    cs_lin_enc_representation_t* l = (cs_lin_enc_representation_t*)n;

    if(l == NULL)
        return;
    cs_lists_destroy(&l->l_signalname);
    free(l);	     
}

/********************************************************************/
void _lin_ldf_free_signaldefinition(void *n)
/********************************************************************/
{
    cs_lin_signal_def_t* l = (cs_lin_signal_def_t*)n;

    if(l == NULL)
        return;
    cs_lists_destroy(&l->l_subscribername);
    free(l);	     
}

/********************************************************************/
void _lin_ldf_free_frames(void *n)
/********************************************************************/
{
    cs_lin_frame_t* l = (cs_lin_frame_t*)n;

    if(l == NULL)
        return;
    cs_lists_destroy(&l->l_signal);
    cs_lists_destroy(&l->l_framename);
    cs_lists_destroy(&l->l_frame);
    free(l);	     
}

/********************************************************************/
void _lin_ldf_free_scheduletable(void *n)
/********************************************************************/
{
    cs_lin_schedule_table_t* l = (cs_lin_schedule_table_t*)n;

    if(l == NULL)
        return;
    cs_lists_destroy(&l->l_entry);
    free(l);	     
}

/********************************************************************/
void _lin_ldf_free_node(void *n)
/********************************************************************/
{
    cs_lin_node_t* l = (cs_lin_node_t*)n;

    if(l == NULL)
        return;

    cs_lists_destroy(&l->l_frames_pub);
  
    free(l);	     
}


/** Encode physical signal to converted raw value
    \retval 0 no failure
    \retval 1 failue
*/
/********************************************************************/
int _lin_ldf_encode_physical(
    cs_lin_enc_t* enc,            /**< Encoding structure*/
    double *value,                /**< Physical value to encode*/
    int len,                      /**< Number of values to encode*/
    cs_lin_raw_t *conv,           /**< Converted value array, internal size CS_LIN_SIGNAL_MAX_BYTE*/
    int offset                    /**< Offset in value array*/
    )
/********************************************************************/
{
    int i,j;
    cs_lin_raw_t tmp;
  
    if((enc == NULL) ||
       (value == NULL) ||
       (len <= 0) ||
       (len > CS_LIN_SIGNAL_MAX_BYTE) ||
       (conv == NULL) ||
       (offset < 0) ||
       ((offset + len) >= CS_LIN_SIGNAL_MAX_BYTE)){
        _lin_error("Parameter failure\n");
        return CS_FAIL;
    }
  
    //First check type
    if(!((enc->base == LIN_ENC_DYNAMIC) ||
         (enc->base == LIN_ENC_PHYSICAL))){
        _lin_error("Invalid encoding type for physical encoding\n");
        return CS_FAIL;
    }

    //Loop about signal vector
    for(j = 0; j < len; j++){
        cs_lin_enc_physical_t* pe = NULL;
        uint8_t found = 0;
        //Loop about physical encoding definitions
        for(i = 0; i < cs_lists_size(&enc->l_physical); i++){
            pe =  (cs_lin_enc_physical_t*)cs_lists_get_iter(&enc->l_physical,i);
      
            //Check for working area
            if((value[j] >= pe->min_phys) && (value[j] < pe->max_phys)){
                //Found correct working range
                found = 1;
                //Check for zero scale
                if(DEQUALS_ZERO(pe->scale)){
                    _lin_error("Zero scale detected for encoding: %s\n",enc->name);
                    return CS_FAIL;
                }

                //Calculate
                double dtmp = (value[j]-pe->offset)/pe->scale;
                //compensate double inaccurrancy, tmp has to be an integer
                tmp = (cs_lin_raw_t)round(dtmp);
                //_lin_debug("Physic %f Encoded 0x%X\n",value[j],tmp);
	
                //Check range
                if((tmp < pe->min_raw) || (tmp >= pe->max_raw)){
                    _lin_error("Encode value index %d is out of range\n",j);
                    return CS_FAIL;
                }

                //write output array
                conv[j+offset] = tmp;
                break;
            }
        }
        if(found == 0){
            _lin_error("No physical encoding found for value: %f\n",value[j]);
            return CS_FAIL;
        }
    }
  
    //_lin_verbose("%s Conv buffer:",enc->name);
    //_lin_verbose_conv(conv,CS_LIN_SIGNAL_MAX_BYTE);
    return CS_OK;
}

/** Deccode internal raw value to physical signal
    \retval 0 no failure
    \retval 1 failue
*/
/********************************************************************/
int _lin_ldf_decode_physical(
    cs_lin_enc_t* enc,            /**< Encoding structure*/
    double *value,                /**< Physical value result buffer*/
    int len,                      /**< Number of values to encode*/
    cs_lin_raw_t *conv            /**< Raw value array, internal size CS_LIN_SIGNAL_MAX_BYTE*/
    )
/********************************************************************/
{
    int i,j;
  
    if((enc == NULL) ||
       (value == NULL) ||
       (len <= 0) ||
       (len > CS_LIN_SIGNAL_MAX_BYTE) ||
       (conv == NULL)){
        _lin_error("Parameter failure\n");
        return CS_FAIL;
    }
  
    //First check type
    if(!((enc->base == LIN_ENC_DYNAMIC) ||
         (enc->base == LIN_ENC_PHYSICAL))){
        _lin_error("Invalid encoding type for physical decoding\n");
        return CS_FAIL;
    }

    _lin_verbose("%s Conv buffer:",enc->name);
    _lin_verbose_conv(conv,CS_LIN_SIGNAL_MAX_BYTE);
  
    //Loop about signal vector
    for(j = 0; j < len; j++){
        cs_lin_enc_physical_t* pe = NULL;
        uint8_t found = 0;
        //Loop about physical encoding definitions
        for(i = 0; i < cs_lists_size(&enc->l_physical); i++){
            pe =  (cs_lin_enc_physical_t*)cs_lists_get_iter(&enc->l_physical,i);
      
            //Check for working area
            if((conv[j] >= pe->min_raw) && (conv[j] < pe->max_raw)){
                //Found correct working range
                found = 1;
                //Calculate
                value[j] = (((double)conv[j])*pe->scale)+pe->offset;
                //_lin_debug("Raw 0x%X Decoded %f\n",conv[j],value[j]);
                break;
            }
        }
        if(found == 0){
            _lin_error("No physical encoding found for value: %f\n",value[j]);
            return CS_FAIL;
        }
    }	
    return CS_OK;
}

/** Encode physical signal to converted raw value
    \retval 0 no failure
    \retval 1 failue
*/
/********************************************************************/
int _lin_ldf_encode_logical(
    cs_lin_enc_t* enc,            /**< Encoding structure*/
    const char *value,            /**< Logical value to encode*/
    cs_lin_raw_t *conv,           /**< Converted value array, internal size CS_LIN_SIGNAL_MAX_BYTE*/
    int offset                    /**< Offset in value array*/
    )
/********************************************************************/
{
    int i;
  
    if((enc == NULL) ||
       (value == NULL) ||
       (conv == NULL) ||
       (offset < 0) ||
       (offset >= CS_LIN_SIGNAL_MAX_BYTE)){
        _lin_error("Parameter failure\n");
        return CS_FAIL;
    }
  
    //First check type
    if(!((enc->base == LIN_ENC_DYNAMIC) ||
         (enc->base == LIN_ENC_LOGICAL))){
        _lin_error("Invalid encoding type for logical encoding\n");
        return CS_FAIL;
    }


    cs_lin_enc_logical_t* le = NULL;
    uint8_t found = 0;
    //Loop about logical encoding definitions
    for(i = 0; i < cs_lists_size(&enc->l_logical); i++){
        le =  (cs_lin_enc_logical_t*)cs_lists_get_iter(&enc->l_logical,i);
      
        //Check for working area
        if(strncmp(value,le->text,CS_LIN_CHARBUF) == 0){
            //Found correct encoding definition
            found = 1;
            //write output array
            conv[offset] = le->value;
            break;
        }
    }
    if(found == 0){
        _lin_error("No logiccal encoding found for value: %s\n",value);
        return CS_FAIL;
    }
    _lin_verbose("%s Conv buffer:",enc->name);
    _lin_verbose_conv(conv,CS_LIN_SIGNAL_MAX_BYTE);
  
    return CS_OK;
}

/** Decode raw value to logical text
    \retval 0 no failure
    \retval 1 failue
*/
/********************************************************************/
int _lin_ldf_decode_logical(
    cs_lin_enc_t* enc,            /**< Encoding structure*/
    char *value,                  /**< Logical value to decoding buffer*/
    int buflen,                   /**< Max buffer length*/
    cs_lin_raw_t conv             /**< Valueto convert*/
    )
/********************************************************************/
{
    int i;
  
    if((enc == NULL) ||
       (value == NULL) ||
       (buflen == 0) ||
       (buflen < CS_LIN_CHARBUF)
        ){
        _lin_error("Parameter failure\n");
        return CS_FAIL;
    }
  
    //First check type
    if(!((enc->base == LIN_ENC_DYNAMIC) ||
         (enc->base == LIN_ENC_LOGICAL))){
        _lin_error("Invalid encoding type for logical encoding\n");
        return CS_FAIL;
    }

    //reset result buffer
    memset(value,0,sizeof(char)*buflen);

    _lin_verbose("%s Conv buffer:",enc->name);
    _lin_verbose_conv(&conv,1);
  
    cs_lin_enc_logical_t* le = NULL;
    uint8_t found = 0;
    //Loop about logical encoding definitions
    for(i = 0; i < cs_lists_size(&enc->l_logical); i++){
        le =  (cs_lin_enc_logical_t*)cs_lists_get_iter(&enc->l_logical,i);
      
        if(le->value == conv){
            found = 1;
            strncpy(value,le->text,(buflen <= CS_LIN_CHARBUF)?buflen:CS_LIN_CHARBUF);
            break;
        }
    }
    if(found == 0){
        _lin_error("No logical encoding found for value: %s\n",value);
        return CS_FAIL;
    }	
    return CS_OK;
}

/** Convert internal converted signal to signal raw representation.
    Do optional byte swapping and shifting to correct position
    \retval 0 no failure
    \retval 1 failue
*/
/********************************************************************/
int _lin_signal_conv2raw(
    cs_lin_signal_t *sig          /**< Signal instance pointer*/
    )
/********************************************************************/
{
    int i,j;
    //get signal definition
    cs_lin_signal_def_t *def = sig->signal_def;

    //_lin_verbose("Process signal %s\n",sig->name);
  
    //Reset raw buffer
    memset(sig->raw,0,sizeof(uint8_t)*CS_LIN_MAXDLC);
  
    if(def->initlen > 1){
        //Loop over signal conv length
        for(i = 0; i < def->initlen; i++){
	  //_lin_verbose("Process signal %s vector value [%d]= 0x%X\n",sig->name,i,sig->conv[i]);
            for(j = 0; j < 8; j++){
                //LSB first, note only byte array is defined for signal vector by spec
                //Reorder to LSB
	        sig->raw[i] = (uint8_t)sig->conv[i];
            }
        }
    } else {
        //_lin_verbose("Process signal %s len %d bits, value = 0x%X BitMask 0x%X\n",sig->name,def->len,sig->conv[0],((1<<(def->len))-1));
        //LSB first, note, up to 16 bit possible
        //Mask bit length
        cs_lin_raw_t lraw = sig->conv[0] & ((1<<(def->len))-1);
        //keep in mind, that we are mapping on a byte array
        sig->raw[1] = (uint8_t)(lraw >> 8);
        sig->raw[0] = (uint8_t)(lraw & 0xFF);
    }
    //_lin_verbose("Signal %s Raw vector:",sig->name);
    //_lin_verbose_buffer(sig->raw,CS_LIN_MAXDLC);
    return CS_OK;
}

/** Convert internal raw signal to signal converted representation.
    Do optional byte swapping and shifting to correct position
    \retval 0 no failure
    \retval 1 failue
*/
/********************************************************************/
int _lin_signal_raw2conv(
    cs_lin_signal_t *sig          /**< Signal instance pointer*/
    )
/********************************************************************/
{
    int i;
    //get signal definition
    cs_lin_signal_def_t *def = sig->signal_def;

    _lin_verbose("Process signal %s\n",sig->name);
    
    if(def->initlen > 1){
        //Loop over signal conv length
        for(i = 0; i < def->initlen; i++){
            _lin_verbose("Process vector value [%d]\n",i);
	    sig->conv[i] = 0;
            //LSB first, note only byte array is defined for signal vector by spec
            sig->conv[i] = sig->raw[i];
        }
    } else {
        //LSB first, note, up to 16 bit possible
        //Reset value
        //Note conversion from 8 to 16 bit
        sig->conv[0] = 0;
        sig->conv[0] = (cs_lin_raw_t)sig->raw[0];
        sig->conv[0] |= ((cs_lin_raw_t)sig->raw[1]) << 8; 
    }
    _lin_verbose("Raw vector: ");
    _lin_verbose_buffer(sig->raw,CS_LIN_MAXDLC); 
    return CS_OK;
}

/** Init signal based on init data
    \retval 0 no failure
    \retval 1 failue
*/
/********************************************************************/
int _lin_init_signal(
    cs_lin_signal_t * sig   /**< Signal pointer*/
    )
/********************************************************************/
{
    int i;
    
    //Reset signal conv buffer
    memset(sig->conv,0,sizeof(cs_lin_raw_t)*CS_LIN_SIGNAL_MAX_BYTE);
  
    //Copy init value buffer to conv
    //Do it manually
    for(i = 0; i < sig->signal_def->initlen; i++){
        sig->conv[i] = sig->signal_def->init[i];
    }

    //Now convert to raw buffer
    if(_lin_signal_conv2raw(sig)){
        _lin_error("Failure during converting values\n");
        return CS_FAIL;
    }
  
    //Set signal state to idle
    //_lin_verbose("Signal state: IDLE\n");
    sig->state = LIN_SIGNAL_STATE_IDLE;
  
    return CS_OK;  

}


/** Update frame raw data content based on signal content. Only used for frames to be send
    \retval 0 no failure
    \retval 1 failue
*/
/********************************************************************/
int _lin_update_frame(
    cs_lin_frame_t *frame,             /**< Frame structure to update*/
    int flag                           /**< Flag 0= Update based on signal content, 1=Update based on init data*/
    )
/********************************************************************/
{
    int i,j;

    //_lin_verbose("Update frame %s\n",frame->name);
  
    if((flag == 0) && (frame->state != LIN_STATE_UPDATED)){
        //Nothing to do
        return CS_OK;
    }
  
    //Handle signals
    for(i = 0; i < cs_lists_size(&frame->l_signal); i++){
        cs_lin_signal_t* s =  (cs_lin_signal_t*)cs_lists_get_iter(&frame->l_signal,i);

        //Update in init case
        if(flag == 1){
            if(_lin_init_signal(s)){
                _lin_error("Failure during init signal %s values\n",s->name);
                return CS_FAIL;
            }
            //Signal frame
            //_lin_verbose("Frame %s state: UPDATED\n",frame->name);
            frame->state = LIN_STATE_UPDATED;
        } else {
            //Update only when init, or signal was updated
            if(s->state == LIN_SIGNAL_STATE_UPDATED){
                //Convert conv buffer to raw
                if(_lin_signal_conv2raw(s)){
                    _lin_error("Failure during converting values\n");
                    return CS_FAIL;
                }
	
                //Reset signal state
                //_lin_verbose("Signal %s  state: IDLE\n",s->name);
                s->state = LIN_SIGNAL_STATE_IDLE;

                //Signal frame
                //_lin_verbose("Frame %s state: UPDATED\n",frame->name);
                frame->state = LIN_STATE_UPDATED;
            }
        }
    }

    if(frame->state == LIN_STATE_UPDATED){
        //Build send frame based on signal raw segments
        memset(frame->raw,0,sizeof(uint8_t)*CS_LIN_MAXDLC);
        //Loop about signals and do a bit or with the signals content
        //_lin_verbose("Merge frame: %s \n",frame->name);
        for(i = 0; i < cs_lists_size(&frame->l_signal); i++){
            cs_lin_signal_t* s =  (cs_lin_signal_t*)cs_lists_get_iter(&frame->l_signal,i);
            //_lin_verbose("Signal: %s offset %d bits, signal len %d bits: ",s->name,s->offset,s->signal_def->len);
            //_lin_verbose_buffer(s->raw,CS_LIN_MAXDLC);
            //Loop over bit length
            //Get corresponding signal raw bit
            //Shift significant bit to lowest position
            //mask all other bits
            //Shift to the correct bit position in frame byte
            for(j = 0; j < s->signal_def->len; j++){
	      //_lin_verbose("Frame byte %d, signal byte %d, signal bit shift %d, frame bit %d\n",(s->offset+j)/8,j/8,(j%8),((s->offset + j) % 8));
                frame->raw[(s->offset+j)/8] |= ((s->raw[j/8] >> (j%8)) & 1) << ((s->offset + j) % 8);
                //_lin_verbose("Frame byte %d: 0x%X\n",(s->offset+j)/8,frame->raw[(s->offset+j)/8]);
            }
        }
        //_lin_verbose("Merged frame %s: ",frame->name);
        //_lin_verbose_buffer(frame->raw,CS_LIN_MAXDLC);
    }
    return CS_OK;
}

/** Process frame received raw data. Only used for frames to be received
    \retval 0 no failure
    \retval 1 failue
*/
/********************************************************************/
int _lin_process_frame(
    cs_lin_frame_t *frame,             /**< Frame structure to update*/
    uint8_t *data,                     /**< Raw data segment*/
    int len,                           /**< Data len*/
    double ts                          /**< Actual timestamp*/
    )
/********************************************************************/
{
    int i,j;
  
    _lin_verbose("Process frame %s at time: %f s\n",frame->name,ts);  


    //Copy received data to frame buffer, length was already checked
    memcpy(frame->raw,data,sizeof(uint8_t)*CS_LIN_MAXDLC);
  
  
    _lin_verbose("Processed content: ");
    _lin_verbose_buffer(frame->raw,len);
    //Copy frame content to signal
    for(i = 0; i < cs_lists_size(&frame->l_signal); i++){
        cs_lin_signal_t* s =  (cs_lin_signal_t*)cs_lists_get_iter(&frame->l_signal,i);
        _lin_verbose("Signal: %s offset %d bits, signal len %d bits\n",s->name,s->offset,s->signal_def->len);
        memset(s->raw,0,sizeof(uint8_t)*CS_LIN_MAXDLC);
        for(j = 0; j < (s->signal_def->len); j++){
            //Loop over signal bit length
            //get corresponding frame byte (boffset/8)
            //shift byte so that lowest bit is the interesting one
            //Mask all other bits
            //do bit or with signal raw byte 
            int boffset = s->offset+j;
            _lin_verbose("Frame byte %d, frame byte bit %d, signal byte %d, signal bit shift %d\n",boffset/8,boffset %8,j/8,j%8);
            s->raw[j/8] |= (frame->raw[boffset/8] >> (boffset%8) & 1) << (j%8);
            _lin_verbose("Signal byte %d: 0x%X\n",j/8,s->raw[j/8]);
        }
        _lin_verbose("Signal: %s: ",s->name);
        _lin_verbose_buffer(s->raw,CS_LIN_MAXDLC);
        _lin_signal_raw2conv(s);
        //Mark signal
        _lin_verbose("Signal %s state: UPDATED\n",s->name);
        s->state = LIN_SIGNAL_STATE_UPDATED;
    }
    return CS_OK;
}

/********************************************************************/
void _lin_ldf_free_nodeattributes(void *n)
/********************************************************************/
{
    cs_lin_node_t* l = (cs_lin_node_t*)n;

    if(l == NULL)
        return;

    free(l);	     
}


/********************************************************************/
char* cs_lin_geterroridstring(
    uint8_t lerror
    )
/********************************************************************/
{
    switch(lerror){
    case CS_LIN_ERROR_NOERROR:
        return "No Error";
    case CS_LIN_ERROR_SLLIN:
        return "SLLIN specific Error";
    case CS_LIN_ERROR_SOCKETCAN:
        return "SocketCAN specific Error";
    case CS_LIN_ERROR_INTERNAL:
        return "Internal Error";
    case CS_LIN_ERROR_PARAMETER:
        return "Parameter Error";
    case CS_LIN_ERROR_LDF:
        return "LDF configuration Error";
    case CS_LIN_ERROR_SCHED:
        return "Scheduling Error";
    case CS_LIN_ERROR_CONTENT:
        return "Parity or Checksum Error";
    }

    return "";
}

/********************************************************************/
int cs_lin_getlasterror(
    cs_lin_t *sk
    )
/********************************************************************/
{
    if(sk == NULL)
        return 0;
  
    return sk->lerror;
}

/********************************************************************/
char* cs_lin_getlasterrorstring(
    cs_lin_t *sk
    )
/********************************************************************/
{
    if(sk == NULL)
        return "";

    return cs_lin_geterroridstring(sk->lerror);
}

/********************************************************************/
void _lin_reset(
    cs_lin_t *sk
    )
/********************************************************************/
{
    sk->master = NULL;
    sk->listen_only = 0;
    sk->lerror = CS_LIN_ERROR_NOERROR;
    sk->timeout_ts = 0.0; // in s
}



/********************************************************************/
cs_lin_t *cs_lin_init()
/********************************************************************/
{
    cs_lin_t *data = NULL;
    
    data = (cs_lin_t*)calloc(1,sizeof(cs_lin_t));
    if(data == NULL)
        return NULL;

    if(cs_lists_init(&data->l_nodes,_lin_ldf_cpf,_lin_ldf_free_generic) ||
       cs_lists_init(&data->l_frames,_lin_ldf_cpf,_lin_ldf_free_generic)){
        _lin_error("Failed to init lists\n");
        cs_lists_destroy(&data->l_nodes);
        cs_lists_destroy(&data->l_frames);
        free(data);
        return NULL;
    }
  
    _lin_reset(data);
  
    return data;
}

/********************************************************************/
void cs_lin_free(
    cs_lin_t *sk
    )
/********************************************************************/
{
  
    if(sk == NULL)
        return;

    cs_lists_destroy(&sk->l_nodes);
    cs_lists_destroy(&sk->l_frames);
    free(sk);
    sk = NULL;
}

/********************************************************************/
double cs_lin_get_time()
/********************************************************************/
{
    struct timespec ts;
    double dts;
    cs_ts_gettime(&ts);
    dts = cs_ts_timespec2d(ts);
    //_lin_verbose("Time: %f\n",dts);
    return dts;
}

/********************************************************************/
void cs_lin_set_callback(
    cs_lin_t *sk,
    void *ctx,
    uint8_t send_finished, /**< 0=> No send finished call needed, 1=> send finished call needed*/
    void (*cs_lin_cb_prepfilt)(void *ctx, uint32_t *id, int len),  /**< Prepare recv filter*/
    void (*cs_lin_cb_send)(void *ctx, uint32_t id, uint8_t *data, int len, int enhanced),
    void (*cs_lin_cb_recv)(void *ctx, uint32_t id, int enhanced),
    void (*cs_lin_cb_prepcache)(
        void *ctx,        /**< Context pointer*/
        uint32_t id,      /**< LIN id*/
        uint8_t *data,    /**< Data buffer*/
        int len,          /**< Len data buffer*/
        int enhanced      /**< 0=> Classic checksum, 1=>Enhanced checksum*/)
    )
/********************************************************************/
{
  
    if(sk == NULL)
        return;
    sk->cb_ctx = ctx;
    sk->send_finished_call = send_finished;
    sk->cs_lin_cb_prepfilt = cs_lin_cb_prepfilt;
    sk->cs_lin_cb_send = cs_lin_cb_send;
    sk->cs_lin_cb_recv = cs_lin_cb_recv;
    sk->cs_lin_cb_prepcache = cs_lin_cb_prepcache;

}

/********************************************************************/
int cs_lin_reset_slave(
    cs_lin_t *sk          /**< Socket struct pointer*/
    )
/********************************************************************/
{
    uint32_t ids[CS_LIN_MAX_ID];
    int idlen = 0;
    int j;

    if(sk == NULL){
        _lin_error("Invalid parameter\n");
        return CS_FAIL;
    }
  
    //Prepare frame and signal content
    for(j = 0; j < cs_lists_size(&sk->l_frames); j++){
        cs_lin_frame_t* f =  *((cs_lin_frame_t**)cs_lists_get_iter(&sk->l_frames,j));

        if(f->dir ==  LIN_DIR_RECEIVE){ // Send by slave
            //Init frame content based on init data
            if(_lin_update_frame(f,1)){
                _lin_error("Init frame %s failed\n",f->name);
                sk->lerror = CS_LIN_ERROR_INTERNAL;
                return CS_FAIL;
            }
            //Call prepare cache routine if set
            if(sk->cs_lin_cb_prepcache != NULL){
                (sk->cs_lin_cb_prepcache)(sk->cb_ctx,f->id, f->raw, f->dlc,cs_lin_frame_is_enhanced(f));
            }
        }
        ids[idlen++]=f->id;
        //Reset all frame states
        //_lin_verbose("Frame %s state: IDLE\n",f->name);
        f->state = LIN_STATE_IDLE;
    }

    if(sk->cs_lin_cb_prepfilt != NULL){
        (sk->cs_lin_cb_prepfilt)(sk->cb_ctx,ids,idlen);
    }
  
    //reset error state
    sk->lerror = CS_LIN_ERROR_NOERROR;
    //reset slave state
    //_lin_verbose("Slave state: IDLE\n");
    sk->s_state = S_STATE_IDLE;
    //reset request frame pointer
    sk->s_frame_req = NULL;
  
    return CS_OK;
}

/********************************************************************/
int cs_lin_init_slave(
    cs_lin_t *sk,           /**< Socket struct pointer*/
    cs_lin_ldf_t *ldf,      /**< LDF based configuration*/
    const char **nodes,     /**< Char pointer array, with node names*/
    int no_nodes            /**< Number of node names*/
    )
/********************************************************************/
{
    int i,j,k,l;
    cs_lin_node_t *node = NULL;

    if(sk == NULL){
        _lin_error("Invalid parameter\n");
        return CS_FAIL;
    }
  
    if((ldf == NULL) ||
       (nodes == NULL) ||
       (no_nodes <= 0)){
        _lin_error("Invalid parameter\n");
        sk->lerror = CS_LIN_ERROR_PARAMETER;
        return CS_FAIL;
    }

    //Store LDF structure
    sk->ldf = ldf;
  
    _lin_verbose("Configure slave, Number of Nodes %d\n",no_nodes);
  
    //Search for node names
    for(j = 0; j < no_nodes; j++){
        if(nodes[j] == NULL){
            _lin_error("Invalid node name\n");
            sk->lerror = CS_LIN_ERROR_PARAMETER;
            return CS_FAIL;
        }
        if((node = cs_lin_ldf_get_node(ldf,nodes[j])) == NULL){
            _lin_error("Can not find node config %s\n",nodes[j]);
            sk->lerror = CS_LIN_ERROR_LDF;
            return CS_FAIL;
        }
        cs_lin_node_t** e = NULL;
        e = (cs_lin_node_t**)malloc(sizeof(cs_lin_node_t*));
        if(e == NULL){
            _lin_error("Malloc failue\n");
            sk->lerror = CS_LIN_ERROR_INTERNAL;
            return CS_FAIL;
        }
        *e = node;
        if(cs_lists_insert(&sk->l_nodes,(void*)e)){
            _lin_error("List insert failure\n");
            free(e);
            sk->lerror = CS_LIN_ERROR_INTERNAL;
            return CS_FAIL;
        }
        _lin_verbose("Add node %s\n",node->name);
    }

    //Loop about nodes to get list of frames
    for(j = 0; j < cs_lists_size(&sk->l_nodes); j++){
        node =  *((cs_lin_node_t**)cs_lists_get_iter(&sk->l_nodes,j));
        for(i= 0; i < cs_lists_size(&node->l_frames_pub); i++){
            cs_lin_frame_t* f =  *((cs_lin_frame_t**)cs_lists_get_iter(&node->l_frames_pub,i));
            cs_lin_frame_t** e = NULL;
            e = (cs_lin_frame_t**)malloc(sizeof(cs_lin_frame_t*));
            if(e == NULL){
                _lin_error("Malloc failue\n");
                sk->lerror = CS_LIN_ERROR_INTERNAL;
                return CS_FAIL;
            }
            *e = f;
            if(cs_lists_insert(&sk->l_frames,(void*)e)){
                _lin_error("List insert failure\n");
                free(e);
                sk->lerror = CS_LIN_ERROR_INTERNAL;
                return CS_FAIL;
            }
        }
    }

    //Now add frames with signals we have subscribed
    //Node this is an optional argument
    //Loop all frames
    for(l = 0; l < cs_lists_size(&ldf->l_frames); l++){
        cs_lin_frame_t * lf = (cs_lin_frame_t*)cs_lists_get_iter(&ldf->l_frames,l);
        //Loop signals of frame
        for(j = 0; j < cs_lists_size(&lf->l_signal); j++){
            cs_lin_signal_t* s =  (cs_lin_signal_t*)cs_lists_get_iter(&lf->l_signal,j);
            //Get signal definition pointer 
            cs_lin_signal_def_t *def = s->signal_def;
            //loop subscriber
            for(i= 0; i < cs_lists_size(&def->l_subscribername); i++){
                char *subscribednode =  (char*)cs_lists_get_iter(&def->l_subscribername,i);
                for(k = 0; k < no_nodes; k++){
                    if(strncmp(subscribednode,nodes[k],CS_LIN_CHARBUF) == 0){
                        //We have to handle this signal
                        //Check if frame is already in list
                        if(cs_lin_get_frame_by_id(sk,lf->id) == NULL){
                            //Add frame
                            cs_lin_frame_t** e = NULL;
                            e = (cs_lin_frame_t**)malloc(sizeof(cs_lin_frame_t*));
                            if(e == NULL){
                                _lin_error("Malloc failue\n");
                                sk->lerror = CS_LIN_ERROR_INTERNAL;
                                return CS_FAIL;
                            }
                            *e = lf;
                            if(cs_lists_insert(&sk->l_frames,(void*)e)){
                                _lin_error("List insert failure\n");
                                free(e);
                                sk->lerror = CS_LIN_ERROR_INTERNAL;
                                return CS_FAIL;
                            }
                        }
                    }
                }
            }
        }
    }
    if(cs_lin_debuglevel(CS_DEBUG_LEVEL_READ) >= CS_DEBUG_LEVEL_DEBUG){
        for(i = 0; i < cs_lists_size(&sk->l_frames); i++){
            cs_lin_frame_t* f =  *((cs_lin_frame_t**)cs_lists_get_iter(&sk->l_frames,i));
            if(f->dir == LIN_DIR_SEND){
                //Master send data
                _lin_verbose("Handle Frame: %s ID: 0x%X, send by master\n",f->name,f->id);
            }
            if(f->dir == LIN_DIR_RECEIVE){
                //Master send data
                _lin_verbose("Handle Frame: %s ID: 0x%X, send by slave\n",f->name,f->id);
            }
        }
    }
    return cs_lin_reset_slave(sk);
}

/********************************************************************/
int cs_lin_reset_listener(
    cs_lin_t *sk          /**< Socket struct pointer*/
    )
/********************************************************************/
{
    int j;

    if(sk == NULL){
        _lin_error("Invalid parameter\n");
        return CS_FAIL;
    }
  
    //Prepare frame and signal content
    for(j = 0; j < cs_lists_size(&sk->l_frames); j++){
        cs_lin_frame_t* f =  *((cs_lin_frame_t**)cs_lists_get_iter(&sk->l_frames,j));
        //Reset all frame states
        //_lin_verbose("Frame %s state: IDLE\n",f->name);
        f->state = LIN_STATE_IDLE;
    }
  
    //reset error state
    sk->lerror = CS_LIN_ERROR_NOERROR;
    //reset slave state
    //_lin_verbose("Slave state: IDLE\n");
    sk->s_state = S_STATE_IDLE;
    //reset request frame pointer
    sk->s_frame_req = NULL;
  
    return CS_OK;
}

/********************************************************************/
int cs_lin_init_listener(
    cs_lin_t *sk,           /**< Socket struct pointer*/
    cs_lin_ldf_t *ldf       /**< LDF based configuration*/
    )
/********************************************************************/
{
    int i,j,l;

    if(sk == NULL){
        _lin_error("Invalid parameter\n");
        return CS_FAIL;
    }
  
    if(ldf == NULL){
        _lin_error("Invalid parameter\n");
        sk->lerror = CS_LIN_ERROR_PARAMETER;
        return CS_FAIL;
    }

    //Store LDF structure
    sk->ldf = ldf;
  
    _lin_verbose("Configure Listener\n");
  
    //Now add frames
    //Loop all frames
    for(l = 0; l < cs_lists_size(&ldf->l_frames); l++){
        cs_lin_frame_t * lf = (cs_lin_frame_t*)cs_lists_get_iter(&ldf->l_frames,l);
        //Loop signals of frame
        for(j = 0; j < cs_lists_size(&lf->l_signal); j++){
            //Check if frame is already in list
            if(cs_lin_get_frame_by_id(sk,lf->id) == NULL){
                //Add frame
                cs_lin_frame_t** e = NULL;
                e = (cs_lin_frame_t**)malloc(sizeof(cs_lin_frame_t*));
                if(e == NULL){
                    _lin_error("Malloc failue\n");
                    sk->lerror = CS_LIN_ERROR_INTERNAL;
                    return CS_FAIL;
                }
                *e = lf;
                if(cs_lists_insert(&sk->l_frames,(void*)e)){
                    _lin_error("List insert failure\n");
                    free(e);
                    sk->lerror = CS_LIN_ERROR_INTERNAL;
                    return CS_FAIL;
                }
            }
        }
    }
    if(cs_lin_debuglevel(CS_DEBUG_LEVEL_READ) >= CS_DEBUG_LEVEL_DEBUG){
        for(i = 0; i < cs_lists_size(&sk->l_frames); i++){
            cs_lin_frame_t* f =  *((cs_lin_frame_t**)cs_lists_get_iter(&sk->l_frames,i));
            if(f->dir == LIN_DIR_SEND){
                //Master send data
                _lin_verbose("Handle Frame: %s ID: 0x%X, send by master\n",f->name,f->id);
            }
            if(f->dir == LIN_DIR_RECEIVE){
                //Master send data
                _lin_verbose("Handle Frame: %s ID: 0x%X, send by slave\n",f->name,f->id);
            }
        }
    }
    //Mark that we are a listen only slave
    sk->listen_only = 1;
    return cs_lin_reset_listener(sk);
}


/********************************************************************/
int cs_lin_reset_master(
    cs_lin_t *sk           /**< Socket struct pointer*/
    )
/********************************************************************/
{
    uint32_t ids[CS_LIN_MAX_ID];
    int idlen = 0;
    int j;
  
    if(sk == NULL){
        _lin_error("Invalid parameter\n");
        return CS_FAIL;
    }
  
    //Init frame data content, only for frames which are send by master
    for(j = 0; j < cs_lists_size(&sk->l_frames); j++){
        cs_lin_frame_t* f =  *((cs_lin_frame_t**)cs_lists_get_iter(&sk->l_frames,j));

        if(f->dir ==  LIN_DIR_SEND){ // Send by master
            //Init frame content based on init data
            if(_lin_update_frame(f,1)){
                _lin_error("Init frame %s failed\n",f->name);
                sk->lerror = CS_LIN_ERROR_INTERNAL;
                return CS_FAIL;
            }
        }
        ids[idlen++] = f->id;
        //Reset all frame states
        //_lin_verbose("Frame %s state: IDLE\n",f->name);
        f->state = LIN_STATE_IDLE;
    }

    if(sk->cs_lin_cb_prepfilt != NULL){
        (sk->cs_lin_cb_prepfilt)(sk->cb_ctx,ids,idlen);
    }
  

    //Reset master state
    //_lin_verbose("Master state: IDLE\n");
    sk->m_state = M_STATE_IDLE;
    //reset error state
    sk->lerror = CS_LIN_ERROR_NOERROR;
  
    return CS_OK;
}

/********************************************************************/
int cs_lin_init_master(
    cs_lin_t *sk,           /**< Socket struct pointer*/
    cs_lin_ldf_t *ldf,      /**< LDF based configuration*/
    char *schedtable        /**< Main scheduler table*/
    )
/********************************************************************/
{
    int i,j;
    cs_lin_node_t *node = NULL;
    cs_lin_schedule_table_t* sched = NULL;

    if(sk == NULL){
        _lin_error("Invalid parameter\n");
        return CS_FAIL;
    }
  
    if((ldf == NULL) ||
       (schedtable == NULL)){
        _lin_error("Invalid parameter\n");
        sk->lerror = CS_LIN_ERROR_PARAMETER;
        return CS_FAIL;
    }

    //Store LDF structure
    sk->ldf = ldf;
 
    _lin_verbose("Configure master, Scheduling table %s\n",schedtable);

    //Try to find node
    if((node = cs_lin_ldf_get_master(ldf)) == NULL){
        _lin_error("Can not find master node config\n");
        sk->lerror = CS_LIN_ERROR_LDF;
        return CS_FAIL;
    }

    _lin_verbose("Master node %s\n",node->name);

    //Store master node
    sk->master = node;
 
    //Search for default schedule table
    if((sched = cs_lin_ldf_get_schedulertable(ldf,schedtable,0)) == NULL){
        _lin_error("Can not find scheduler table config %s\n",schedtable);
        sk->lerror = CS_LIN_ERROR_LDF;
        return CS_FAIL;
    }

    //Set default scheduling table
    sk->defaulttable = sched;
    //Set Actual scheduler to NULL
    sk->table = NULL;

    _lin_verbose("Default scheduling table %s\n",schedtable);
  

    //Start preparing frame list
    //Loop about nodes to get list of frames
    for(j = 0; j < cs_lists_size(&ldf->l_nodes); j++){
        node =  (cs_lin_node_t*)cs_lists_get_iter(&ldf->l_nodes,j);
        for(i= 0; i < cs_lists_size(&node->l_frames_pub); i++){
            cs_lin_frame_t* f =  *((cs_lin_frame_t**)cs_lists_get_iter(&node->l_frames_pub,i));
            cs_lin_frame_t** e = NULL;
            e = (cs_lin_frame_t**)malloc(sizeof(cs_lin_frame_t*));
            if(e == NULL){
                _lin_error("Malloc failue\n");
                sk->lerror = CS_LIN_ERROR_INTERNAL;
                return CS_FAIL;
            }
            *e = f;
            if(cs_lists_insert(&sk->l_frames,(void*)e)){
                _lin_error("List insert failure\n");
                free(e);
                sk->lerror = CS_LIN_ERROR_INTERNAL;
                return CS_FAIL;
            }
        }
    }

    //Init scheduler table
    if(cs_lin_master_switch_table(sk,schedtable)){
        _lin_error("Fail to set scheduler table\n");
        sk->lerror = CS_LIN_ERROR_INTERNAL;
        return CS_FAIL;
    }

    if(cs_lin_debuglevel(CS_DEBUG_LEVEL_READ) >= CS_DEBUG_LEVEL_DEBUG){
        for(i = 0; i < cs_lists_size(&sk->l_frames); i++){
            cs_lin_frame_t* f =  *((cs_lin_frame_t**)cs_lists_get_iter(&sk->l_frames,i));
            if(f->dir == LIN_DIR_SEND){
                //Master send data
                _lin_verbose("Handle Frame: %s ID: 0x%X, send by master\n",f->name,f->id);
            }
            if(f->dir == LIN_DIR_RECEIVE){
                //Master send data
                _lin_verbose("Handle Frame: %s ID: 0x%X, send by slave\n",f->name,f->id);
            }
        }
    }  
    return cs_lin_reset_master(sk);
}

/********************************************************************/
int cs_lin_master_switch_table(
    cs_lin_t *sk,         /**< Socket struct pointer*/
    char *schedtable        /**< Scheduler table*/
    )
/********************************************************************/
{
    cs_lin_schedule_table_t* sched = NULL;

    if(sk == NULL){
        _lin_error("Invalid parameter\n");
        return CS_FAIL;
    }
  
    if(schedtable == NULL){
        _lin_error("Invalid parameter\n");
        sk->lerror = CS_LIN_ERROR_PARAMETER;
        return CS_FAIL;
    }
  
    _lin_verbose("Master switch scheduling table %s\n",schedtable);

    //Search for default schedule table
    if((sched = cs_lin_ldf_get_schedulertable(sk->ldf,schedtable,0)) == NULL){
        _lin_error("Can not find scheduler table config %s\n",schedtable);
        sk->lerror = CS_LIN_ERROR_LDF;
        return CS_FAIL;
    }

    //Set active table
    sk->table = sched;

    //Store first and last entry of scheduler table
    sk->sched_entry_first = cs_lists_get_first(&sk->table->l_entry);
    sk->sched_entry_last = cs_lists_get_last(&sk->table->l_entry);

    //Set active entry to first element
    sk->sched_entry_act = sk->sched_entry_first;

    //reset scheduling timestamp
    sk->sched_ts = 0.0;
  
    return CS_OK;
}


/********************************************************************/
int cs_lin_iface_finished(
    cs_lin_t *sk           /**< Socket struct pointer*/
    )
/********************************************************************/
{
    cs_lin_schedule_entry_t * active = NULL;
  
    if(sk == NULL){
        _lin_error("Invalid parameter\n");
        return CS_FAIL;
    }

    if(sk->master == NULL){
        //Slave handling
        if(sk->s_state != S_STATE_SEND){
            //Nothing to do
            return CS_OK;
        }
        //_lin_verbose("Slave state: IDLE\n");
        sk->s_state = S_STATE_IDLE;
        return CS_OK;
    
    } else {
        //Master handling
        if(sk->m_state != M_STATE_SEND){
            //Nothing to do
            return CS_OK;
        }
        active = (cs_lin_schedule_entry_t *)cs_lists_get_data(sk->sched_entry_act);
    
        if(active->frame->dir == LIN_DIR_SEND){
            //Frame with data was send, we expect no feedback
            //_lin_verbose("Master state: IDLE\n");
            sk->m_state = M_STATE_IDLE;
            return CS_OK;
        }

        if(active->frame->dir == LIN_DIR_RECEIVE){
            //Frame with data request was send, we expect feedback
            _lin_verbose("Master state: RECV\n");
            sk->m_state = M_STATE_RECV;
            return CS_OK;
        }

        //Should never be reached
        _lin_error("Statemaschine failure\n");
        sk->lerror = CS_LIN_ERROR_SCHED;
        return CS_FAIL;
    }

    return CS_OK;
}

/********************************************************************/
int cs_lin_process_start(
    cs_lin_t *sk,         /**< Socket struct pointer*/
    double ts             /**< Actual timestamp*/
    )
/********************************************************************/
{
    cs_lin_schedule_entry_t * active = NULL;
  
    if(sk == NULL){
        _lin_error("Invalid parameter\n");
        return CS_FAIL;
    }

    //_lin_verbose("Actual time: %f\n",ts);

    //Store timestamp
    sk->act_ts = ts;

    if(sk->master == NULL){
        //We are a slave
        //Nothing to do
        return CS_OK;
    }

    if((sk->send_finished_call == 0) && (sk->m_state == M_STATE_SEND)){
        //If no send finished call needed, go directly to wait for frame
        sk->m_state = M_STATE_RECV;
        _lin_verbose("Master state: RECV\n");
    }

    if(sk->m_state == M_STATE_SEND){
        _lin_verbose("Wait for send finished\n");
        //We have to wait for send finished call
        return CS_OK;
    }
  
    //Ok we are master, look for recv timeout
    if((sk->m_state == M_STATE_RECV) && (sk->sched_entry_act != NULL)){
        //We are waiting for a frame
        active = (cs_lin_schedule_entry_t *)cs_lists_get_data(sk->sched_entry_act);
        //Check for timeout
        if(sk->act_ts > sk->timeout_ts){
            _lin_warn("Detect timeout for scheduler entry: %s at time: %f s \n",active->name,sk->act_ts);
            //Mark Frame
            _lin_verbose("Frame %s state: TIMEOUT\n",active->frame->name);
            active->frame->state = LIN_STATE_TIMEOUT;
            //Reset master state
            //_lin_verbose("Master state: IDLE\n");
            sk->m_state = M_STATE_IDLE;
            return 2;
        }
    }
    return CS_OK;
}

/********************************************************************/
int cs_lin_process_msg(
    cs_lin_t *sk,       /**< Socket struct pointer*/
    uint32_t id,        /**< Message id received*/
    uint8_t *data,      /**< Data pointer*/
    int len             /**< Data length*/
    )
/********************************************************************/
{
    cs_lin_frame_t *frame = NULL;
    cs_lin_schedule_entry_t * active = NULL;
  
    if((sk == NULL) || (data == NULL)){
        _lin_error("Invalid parameter\n");
        return CS_FAIL;
    }

    if(sk->master != NULL){
        //Master handling

        //TODO handle wake up frames
        _lin_verbose("Raw frame received: Id 0x%X Len %i\n",id,len);
    
        if(sk->m_state != M_STATE_RECV){
	    _lin_warn("System is not in receive state. Received frame: Id 0x%X Len %i: ",id,len);
            _lin_debug_buffer(data,len);
            return CS_OK;
        }
    
        if((frame = cs_lin_get_frame_by_id(sk,id)) == NULL){
            _lin_warn("Unknown frame received. Received frame: Id 0x%X Len %i:",id,len);
            _lin_debug_buffer(data,len);
            return CS_OK;
        }

        active = (cs_lin_schedule_entry_t *)cs_lists_get_data(sk->sched_entry_act);
    
        //Check if frame is the one we expect
        if(frame != active->frame){
            _lin_warn("Unexpected frame received. Received frame: Id 0x%X Len %i:",id,len);
            _lin_debug_buffer(data,len);
            return CS_OK;
        }
    
        if(len != active->frame->dlc){
            _lin_warn("Unexpected frame length received. Ignore. Received frame: Id 0x%X Len %i:",id,len);
            _lin_debug_buffer(data,len);
            return CS_OK;
        }
    
        if(_lin_process_frame(frame,data,len,sk->act_ts)){
            _lin_error("Process frame error. Received frame: Id 0x%X Len %i:",id,len);
            _lin_error_buffer(data,len);
            sk->lerror = CS_LIN_ERROR_INTERNAL;
            return CS_FAIL;
        }
        _lin_verbose("Frame %s state: RECEIVED\n",frame->name);
        frame->state = LIN_STATE_RECEIVED;
        sk->m_state = M_STATE_IDLE;
    } else {
        //Slave handling

        //TODO handle sleep request
        _lin_verbose("Raw frame received: Id 0x%X Len %i\n:",id,len);
    
        if((frame = cs_lin_get_frame_by_id(sk,id)) == NULL){
            _lin_warn("Unknown frame received. Received frame: Id 0x%X Len %i:",id,len);
            _lin_debug_buffer(data,len);
            return CS_OK;
        }

        if((frame->dir == LIN_DIR_RECEIVE) && (sk->listen_only == 0)){
            //Request by master, and we have to handle it
      
            if(len > 0 ){
                _lin_error("Master request frame with non zero length. Received frame: Id 0x%X Len %i:",id,len);
                _lin_error_buffer(data,len);
                sk->lerror = CS_LIN_ERROR_LDF;
                return CS_FAIL;
            }

            _lin_verbose("Frame request received: %s at time: %f s\n",frame->name,sk->act_ts);
      
            //If we have a interface frame cache, response is handled by cache, in other case store id
            if(sk->cs_lin_cb_prepcache != NULL){
                //Handled by interface frame cache
                //Nothing to do
                return CS_OK;
            }

            //we have to send
            _lin_verbose("Slave state: REQ\n");
            sk->s_state = S_STATE_REQ;
            sk->s_frame_req = frame;
            _lin_verbose("Frame %s state: RECEIVED\n",frame->name);
            frame->state = LIN_STATE_RECEIVED;
            return CS_OK;
        } else {
            //Data send by master

            if(len != frame->dlc ){
                _lin_warn("Master request frame with invalid length. Received frame: Id 0x%X Len %i:",id,len);
                _lin_debug_buffer(data,len);
                return CS_OK;
            }

            if(_lin_process_frame(frame,data,len,sk->act_ts)){
                _lin_error("Process frame error. Received frame: Id 0x%X Len %i:",id,len);
                _lin_error_buffer(data,len);
                sk->lerror = CS_LIN_ERROR_INTERNAL;
                return CS_FAIL;
            }
            _lin_verbose("Frame %s state: RECEIVED\n",frame->name);
            frame->state = LIN_STATE_RECEIVED;
        }    
    }
  
    return CS_OK;
}

/********************************************************************/
int cs_lin_process_msg_raw(
    cs_lin_t *sk,       /**< Socket struct pointer*/
    uint8_t id,         /**< Message id received, with parity bits*/
    uint8_t *data,      /**< Data pointer*/
    int len,            /**< Data length*/
    uint8_t checksum    /**< Chesum received*/
    )
/********************************************************************/
{
    cs_lin_frame_t *frame = NULL;
    uint32_t oid = 0; //ID without parity
    uint32_t pid = 0; //ID with calculated parity on our own
    uint8_t lchecksum = 0;
  
    if((sk == NULL) || (data == NULL)){
        _lin_error("Invalid parameter\n");
        return CS_FAIL;
    }

    //Check for parity
    oid = CS_LIN_GET_ID(id);

    cs_lin_id_parity(oid,&pid);

    if(id != (uint8_t)pid){
        _lin_error("Parity error. Received frame: Id 0x%X (with parity) Len %i Checksum 0x%X: ",id,len,checksum);
        _lin_error_buffer(data,len);
        sk->lerror = CS_LIN_ERROR_CONTENT;
        return CS_FAIL;    
    }

    //Check for checksum
    if((frame = cs_lin_get_frame_by_id(sk,oid)) == NULL){
        _lin_warn("Unknown frame received. Received frame: Id 0x%X Len %i: ",oid,len);
        _lin_debug_buffer(data,len);
        return CS_OK;
    }  
  
    lchecksum = cs_lin_checksum(pid,data,len,cs_lin_frame_is_enhanced(frame));

    if(checksum != lchecksum){
      _lin_error("Invalid checksum received. Id 0x%X (with parity) Len %i: Received: 0x%X Calculated 0x%X\n",id,len,checksum,lchecksum);
	_lin_error_buffer(data,len);
        sk->lerror = CS_LIN_ERROR_CONTENT;
        return CS_FAIL;    
    }

    return cs_lin_process_msg(sk,oid,data,len);  
}

/********************************************************************/
int cs_lin_process_end(
    cs_lin_t *sk        /**< Socket struct pointer*/
    )
/********************************************************************/
{
    cs_lin_schedule_entry_t *active = NULL;
    int send = 0;
    int i;
  
    if(sk == NULL){
        _lin_error("Invalid parameter\n");
        return CS_FAIL;
    }

    if(sk->master != NULL){
        //Master handling
        if(sk->m_state == M_STATE_RECV){
            //Master is waiting for a slave response,nothing to do
            return CS_OK;
        }
        if(sk->m_state == M_STATE_SEND){
            //Master is waiting for a send ack, nothing to do
            return CS_OK;
        }
        if(sk->m_state != M_STATE_IDLE){
            //Invalid state,nothing to do
            return CS_OK;
        }

        //Get active scheduling entry
        active = (cs_lin_schedule_entry_t *)cs_lists_get_data(sk->sched_entry_act);
    
        if(sk->sched_ts == 0.0){
            //Table was initialized
            //Set scheduling timestamp
            sk->sched_ts = sk->act_ts + active->delay;
	    sk->timeout_ts = sk->act_ts + active->t_frame_slot;
            send = 1;
        }

        if((send == 0) && (sk->act_ts >= sk->sched_ts)){
            _lin_verbose("Scheduling, Actual time: %f\n",sk->act_ts);
            //Switch to next entry
            sk->sched_entry_act = cs_lists_get_next(sk->sched_entry_act);

            if(sk->sched_entry_act == sk->sched_entry_last){
                //Last entry (no data) in table, restart table
                sk->sched_entry_act = sk->sched_entry_first;
            }

            //Recheck
            if(sk->sched_entry_act == sk->sched_entry_last){
                //Last entry (no data) in table, Invalid table
                _lin_error("Invalid scheduler table: %s ",sk->table->name);
                sk->lerror = CS_LIN_ERROR_LDF;
                return CS_FAIL;	
            }    

            //Get active scheduling entry
            active = (cs_lin_schedule_entry_t *)cs_lists_get_data(sk->sched_entry_act);
            _lin_verbose("Schedule table %s active entry %s frame %s, Master %s\n", sk->table->name,active->name,active->frame->name,
                         ((active->frame->dir == LIN_DIR_SEND)?"send":"receive"));
            //Set scheduling timestamp
            sk->sched_ts = sk->act_ts + active->delay;
	    sk->timeout_ts = sk->act_ts + active->t_frame_slot;
            send = 1;
        }
    
        if(send == 1){
            //We have to handle a entry
            //Get frame
            cs_lin_frame_t *f = active->frame;
            if(f->dir == LIN_DIR_RECEIVE){
                //Send a request

                _lin_verbose("Send Frame request:  %s at time: %f s\n", f->name,sk->act_ts);
                if(sk->cs_lin_cb_recv != NULL){
                    if(sk->send_finished_call == 1){
                        _lin_verbose("Wait for send finished\n");
                        _lin_verbose("Master state: SEND\n");
                        sk->m_state = M_STATE_SEND;
                    } else {
                        _lin_verbose("Wait for receive finished\n");
                        _lin_verbose("Master state: RECV\n");
                        sk->m_state = M_STATE_RECV;
                    }
                    (sk->cs_lin_cb_recv)(sk->cb_ctx,f->id,cs_lin_frame_is_enhanced(f));
                }
                //_lin_verbose("Frame %s state: IDLE\n",f->name);
                f->state = LIN_STATE_IDLE;
            } else {
                //Send data

                //Update signal data
                if(_lin_update_frame(f,0)){
                    _lin_error("Update frame error. Frame: Id 0x%X",f->id);
                    sk->lerror = CS_LIN_ERROR_INTERNAL;
                    return CS_FAIL;
                }

                _lin_verbose("Send Frame data:  %s at time: %f s\n", f->name,sk->act_ts);
	
                _lin_verbose("Send frame: Id 0x%X Len %i: ", f->id, f->dlc);
                _lin_verbose_buffer(f->raw,f->dlc);
                if(sk->cs_lin_cb_send != NULL){
                    if(sk->send_finished_call == 1){
                        _lin_verbose("Wait for send finished\n");
                        _lin_verbose("Master state: SEND\n");
                        sk->m_state = M_STATE_SEND;
                    } else {
                        _lin_verbose("We send data, go back to idle\n");
                        //_lin_verbose("Master state: IDLE\n");
                        sk->m_state = M_STATE_IDLE;
                    }
                    (sk->cs_lin_cb_send)(sk->cb_ctx,f->id, f->raw, f->dlc,cs_lin_frame_is_enhanced(f));
                }
                //_lin_verbose("Frame %s state: IDLE\n",f->name);
                f->state = LIN_STATE_IDLE;
            }
      
        } else {
            //Get frame
            cs_lin_frame_t *f = active->frame;
            //_lin_verbose("Frame %s state: IDLE\n",f->name);
            f->state = LIN_STATE_IDLE;
        }
    } else {
        //Slave handling

        //Loop over all frames to update the frame content
        for(i = 0; i < cs_lists_size(&sk->l_frames); i++){
            cs_lin_frame_t* f =  *((cs_lin_frame_t**)cs_lists_get_iter(&sk->l_frames,i));
            if(f->dir == LIN_DIR_RECEIVE){ // Send by slave
                //_lin_verbose("Update frame id: 0x%X, Name: %s\n",f->id,f->name);
                //Update content
                if(_lin_update_frame(f,0)){
                    _lin_error("Update frame error. Frame: Id 0x%X\n",f->id);
                    sk->lerror = CS_LIN_ERROR_INTERNAL;
                    return CS_FAIL;
                }
                //Call prepare cache routine if set
                if(sk->cs_lin_cb_prepcache != NULL){
                    _lin_verbose("Prepare frame cache update: %s\n", f->name);
                    if(f->state == LIN_STATE_UPDATED){
                        (sk->cs_lin_cb_prepcache)(sk->cb_ctx,f->id, f->raw, f->dlc,cs_lin_frame_is_enhanced(f));
                    }
                }
                //_lin_verbose("Frame %s state: IDLE\n",f->name);
                f->state = LIN_STATE_IDLE;
            } else {
                //Send by slave
                //Reset frame state
                //_lin_verbose("Frame %s state: IDLE\n",f->name);
                f->state = LIN_STATE_IDLE;
            }
        }    

        //Handle sending master request answer if we have no frame cache
        if(sk->cs_lin_cb_prepcache == NULL){
            if(sk->s_state == S_STATE_REQ){
                cs_lin_frame_t *f = sk->s_frame_req;
                _lin_verbose("Send frame (request answer): %s at time: %f s\n", f->name,sk->act_ts);
                _lin_verbose("Send frame: Id 0x%X Len %i: ", f->id, f->dlc);
                _lin_verbose_buffer(f->raw,f->dlc);
                if(sk->cs_lin_cb_send != NULL){
                    if(sk->send_finished_call == 1){
                        _lin_verbose("Wait for send finished\n");
                        _lin_verbose("Slave state: SEND\n");
                        sk->s_state = S_STATE_SEND;
                    } else {
                        _lin_verbose("Send data, go back to idle\n");
                        //_lin_verbose("Slave state: IDLE\n");
                        sk->s_state = S_STATE_IDLE;
                    }
                    (sk->cs_lin_cb_send)(sk->cb_ctx,f->id, f->raw, f->dlc,cs_lin_frame_is_enhanced(f));
                }
            }
        } else {
            _lin_verbose("Request handled by frame cache, go back to idle\n");
            //We have a frame cache, so reset slave state
            //_lin_verbose("Slave state: IDLE\n");
            sk->s_state = S_STATE_IDLE;
        }
    }
    return CS_OK;
}
  
/********************************************************************/
cs_lin_frame_t *cs_lin_get_frame_by_id(
				       cs_lin_t *sk,           /**< Socket struct pointer*/
				       uint32_t id             /**< Frame id*/
				       )
/********************************************************************/
{
    cs_lin_frame_t *rv = NULL;
    int i;
  
    if(sk == NULL){
        _lin_error("Invalid parameter\n");
        return NULL;
    }

    //_lin_verbose("Get frame by id: 0x%X\n",id);
  
    for(i = 0; i < cs_lists_size(&sk->l_frames); i++){
        cs_lin_frame_t* f =  *((cs_lin_frame_t**)cs_lists_get_iter(&sk->l_frames,i));
        if(f->id == id){
            //Found frame
            rv = f;
            _lin_verbose("Got frame by id: 0x%X, Name: %s\n",rv->id,rv->name);
            break;
        }
    }
    if(rv == NULL){
      _lin_debug("Can not find frame by id: 0x%X\n",id);
    }
    return rv;
}

/********************************************************************/
int cs_lin_get_frame_state(
    cs_lin_frame_t * frame  /**< Frame data structure*/
    )
/********************************************************************/
{ 
    if(frame == NULL){
        _lin_error("Invalid parameter\n");
        return -1;
    }
    return frame->state;
}

/********************************************************************/
int cs_lin_reset_frame_state(
    cs_lin_frame_t * frame  /**< Frame data structure*/
    )
/********************************************************************/
{ 
    if(frame == NULL){
        _lin_error("Invalid parameter\n");
        return CS_FAIL;
    }
    _lin_verbose("Frame %s state: IDLE\n",frame->name);
    frame->state = LIN_STATE_IDLE;
    return CS_OK;
}

/********************************************************************/
cs_lin_signal_t *cs_lin_get_frame_signal(
    cs_lin_frame_t * frame,  /**< Frame data structure*/
    const char *name         /**< Signal name*/
    )
/********************************************************************/
{
    cs_lin_signal_t *rv = NULL;
    int i;
  
    if((frame == NULL) ||
       (name == NULL)){
        _lin_error("Invalid parameter\n");
        return NULL;
    }

    //_lin_verbose("Get signal %s from frame 0x%X(%s)\n",name,frame->id,frame->name);
  
    for(i = 0; i < cs_lists_size(&frame->l_signal); i++){
        cs_lin_signal_t* s =  (cs_lin_signal_t*)cs_lists_get_iter(&frame->l_signal,i);
        if(strncmp(s->name,name,CS_LIN_CHARBUF)==0){
            //Found signal
            rv = s;
            _lin_verbose("Got signal by name: %s\n",rv->name);
            break;
        }
    }
    if(rv == NULL){
        _lin_debug("Can not find signal %s in frame 0x%X(%s)\n",name,frame->id,frame->name);
    }
    return rv;
}
 
/********************************************************************/
cs_lin_signal_t *cs_lin_get_signal(
    cs_lin_t *sk,           /**< Socket struct pointer*/
    uint32_t id,            /**< Frame id*/
    const char *name        /**< Signal name*/
    )
/********************************************************************/
{
    cs_lin_signal_t *s = NULL;
    cs_lin_frame_t *f = NULL;

    f = cs_lin_get_frame_by_id(sk,id);

    if(f == NULL){
        return s;
    }
  
    s = cs_lin_get_frame_signal(f,name);

    return s;
}

/********************************************************************/
int cs_lin_update_signal_physical(
    cs_lin_t *sk,           /**< Socket struct pointer*/
    cs_lin_signal_t * sig,  /**< Signal pointer*/
    double *value,          /**< Signal value array*/
    int len                 /**< Value array len*/
    )
/********************************************************************/
{

    if(sk == NULL){
        _lin_error("Parameter failure\n");
        return CS_FAIL;
    }

    if((sig == NULL) ||
       (value == NULL) ||
       (len < 1) ||
       (len > CS_LIN_SIGNAL_MAX_BYTE)){
        _lin_error("Parameter failure\n");
        sk->lerror = CS_LIN_ERROR_PARAMETER;
        return CS_FAIL;
    }

    //Check definition pointer
    if(sig->signal_def == NULL){
        _lin_error("Invalid signal definition for signal %s\n",sig->name);
        sk->lerror = CS_LIN_ERROR_LDF;
        return CS_FAIL;
    }

    //Check frame pointer
    if(sig->frame == NULL){
        _lin_error("Invalid frame definition for signal %s\n",sig->name);
        sk->lerror = CS_LIN_ERROR_LDF;
        return CS_FAIL;
    }

    //Get encoding
    cs_lin_enc_t *enc = sig->signal_def->enc;

    if(enc == NULL){
        _lin_error("Invalid encoding definition for signal %s\n",sig->name);
        sk->lerror = CS_LIN_ERROR_LDF;
        return CS_FAIL;
    }
  
    //Check signal
    if(len !=  sig->signal_def->initlen){
        _lin_error("Invalid value buffer for signal %s\n",sig->name);
        sk->lerror = CS_LIN_ERROR_PARAMETER;
        return CS_FAIL;
    }
    
  
    //Start signal converting
    if(_lin_ldf_encode_physical(enc,value,len,sig->conv,0)){
        _lin_error("Failure during physical encoding for signal %s\n",sig->name);
        //sk->lerror = CS_LIN_ERROR_INTERNAL;
        return CS_FAIL;
    }

    //Set signal state to updated
    //_lin_verbose("Signal %s state: UPDATED\n",sig->name);
    sig->state = LIN_SIGNAL_STATE_UPDATED;

    //Set frame state to updated
    //_lin_verbose("Frame %s state: UPDATED\n",sig->frame->name);
    sig->frame->state = LIN_STATE_UPDATED;
  
    return CS_OK;
  
}

/********************************************************************/
int cs_lin_get_signal_physical(
    cs_lin_t *sk,           /**< Socket struct pointer*/
    cs_lin_signal_t * sig,  /**< Signal pointer*/
    double *value,          /**< Signal value array*/
    int len                 /**< Value array len*/
    )
/********************************************************************/
{
    if(sk == NULL){
        _lin_error("Parameter failure\n");
        return CS_FAIL;
    }

    if((sig == NULL) ||
       (value == NULL) ||
       (len < 1) ||
       (len > CS_LIN_SIGNAL_MAX_BYTE)){
        _lin_error("Parameter failure\n");
        sk->lerror = CS_LIN_ERROR_PARAMETER;
        return CS_FAIL;
    }

    //Check definition pointer
    if(sig->signal_def == NULL){
        _lin_error("Invalid signal definition for signal %s\n",sig->name);
        sk->lerror = CS_LIN_ERROR_LDF;
        return CS_FAIL;
    }

    //Check frame pointer
    if(sig->frame == NULL){
        _lin_error("Invalid frame definition for signal %s\n",sig->name);
        sk->lerror = CS_LIN_ERROR_LDF;
        return CS_FAIL;
    }

    //Get encoding
    cs_lin_enc_t *enc = sig->signal_def->enc;

    if(enc == NULL){
        _lin_error("Invalid encoding definition for signal %s\n",sig->name);
        sk->lerror = CS_LIN_ERROR_LDF;
        return CS_FAIL;
    }
  
    //Check signal
    if(len !=  sig->signal_def->initlen){
        _lin_error("Invalid value buffer for signal %s\n",sig->name);
        sk->lerror = CS_LIN_ERROR_PARAMETER;
        return CS_FAIL;
    }
  
    //Start signal converting
    if(_lin_ldf_decode_physical(enc,value,len,sig->conv)){
        _lin_error("Failure during physical decoding for signal %s\n",sig->name);
        //sk->lerror = CS_LIN_ERROR_INTERNAL;
        return CS_FAIL;
    }
  
    return CS_OK;
}

/********************************************************************/
int cs_lin_update_signal_logical(
    cs_lin_t *sk,           /**< Socket struct pointer*/
    cs_lin_signal_t * sig,  /**< Signal pointer*/
    const char *value,      /**< Signal value*/
    int offset              /**< Offset in value array*/
    )
/********************************************************************/
{
    if(sk == NULL){
        _lin_error("Parameter failure\n");
        return CS_FAIL;
    }

    if((sig == NULL) ||
       (value == NULL)){
        _lin_error("Parameter failure\n");
        sk->lerror = CS_LIN_ERROR_PARAMETER;
        return CS_FAIL;
    }

    //Check definition pointer
    if(sig->signal_def == NULL){
        _lin_error("Invalid signal definition for signal %s \n",sig->name);
        sk->lerror = CS_LIN_ERROR_LDF;
        return CS_FAIL;
    }

    //Check frame pointer
    if(sig->frame == NULL){
        _lin_error("Invalid frame definition for signal %s\n",sig->name);
        sk->lerror = CS_LIN_ERROR_LDF;
        return CS_FAIL;
    }

    //Get encoding
    cs_lin_enc_t *enc = sig->signal_def->enc;

    if(enc == NULL){
        _lin_error("Invalid encoding definition for signal %s \n",sig->name);
        sk->lerror = CS_LIN_ERROR_LDF;
        return CS_FAIL;
    }
  
    //Check signal
    if(offset >=  sig->signal_def->initlen){
        _lin_error("Invalid value offset for signal %s\n",sig->name);
        sk->lerror = CS_LIN_ERROR_PARAMETER;
        return CS_FAIL;
    }
    
  
    //Start signal converting
    if(_lin_ldf_encode_logical(enc,value,sig->conv,offset)){
        _lin_error("Failure during logical encoding for signal %s\n",sig->name);
        sk->lerror = CS_LIN_ERROR_INTERNAL;
        return CS_FAIL;
    }

    //Now convert to raw buffer
    if(_lin_signal_conv2raw(sig)){
        _lin_error("Failure during conevrting values for signal %s\n",sig->name);
        sk->lerror = CS_LIN_ERROR_INTERNAL;
        return CS_FAIL;
    }

    //Set signal state to updated
    _lin_verbose("Signal %s state: UPDATED\n",sig->name);
    sig->state = LIN_SIGNAL_STATE_UPDATED;

    //Signal frame there is work todo
    _lin_verbose("Frame %s state: UPDATED\n",sig->frame->name);
    sig->frame->state = LIN_STATE_UPDATED;

    return CS_OK;
  
}


/********************************************************************/
int cs_lin_get_signal_logical(
    cs_lin_t *sk,           /**< Socket struct pointer*/
    cs_lin_signal_t * sig,  /**< Signal pointer*/
    char *value,            /**< Signal value array*/
    int len,                /**< Value max buffer len*/
    int offset              /**< Offset in signal array*/
    )
/********************************************************************/
{

    if(sk == NULL){
        _lin_error("Parameter failure\n");
        return CS_FAIL;
    }

    if((sig == NULL) ||
       (value == NULL) ||
       (len < 1) ||
       (len < CS_LIN_CHARBUF)){
        _lin_error("Parameter failure\n");
        sk->lerror = CS_LIN_ERROR_PARAMETER;
        return CS_FAIL;
    }

    //Check definition pointer
    if(sig->signal_def == NULL){
        _lin_error("Invalid signal definition for signal %s\n",sig->name);
        sk->lerror = CS_LIN_ERROR_LDF;
        return CS_FAIL;
    }

    //Check frame pointer
    if(sig->frame == NULL){
        _lin_error("Invalid frame definition for signal %s\n",sig->name);
        sk->lerror = CS_LIN_ERROR_LDF;
        return CS_FAIL;
    }

    //Get encoding
    cs_lin_enc_t *enc = sig->signal_def->enc;

    if(enc == NULL){
        _lin_error("Invalid encoding definition for signal %s\n",sig->name);
        sk->lerror = CS_LIN_ERROR_LDF;
        return CS_FAIL;
    }
  
    //Check offset
    if((offset < 0) || (offset > sig->signal_def->initlen)){
        _lin_error("Invalid  offset for signal %s\n",sig->name);
        sk->lerror = CS_LIN_ERROR_PARAMETER;
        return CS_FAIL;
    }
  
    //Start signal converting
    if(_lin_ldf_decode_logical(enc,value,len,sig->conv[offset])){
        _lin_error("Failure during logical decoding for signal %s\n",sig->name);
        sk->lerror = CS_LIN_ERROR_INTERNAL;
        return CS_FAIL;
    }
  
    return CS_OK;
}


/********************************************************************/
int cs_lin_update_signal_raw(
    cs_lin_t *sk,           /**< Socket struct pointer*/
    cs_lin_signal_t * sig,  /**< Signal pointer*/
    cs_lin_raw_t *value,    /**< Signal value*/
    int len                 /**< Value array length*/
    )
/********************************************************************/
{
    int i;
  
    if(sk == NULL){
        _lin_error("Parameter failure\n");
        return CS_FAIL;
    }

    if((sig == NULL) ||
       (value == NULL) ||
       (len <= 0)){
        _lin_error("Parameter failure\n");
        sk->lerror = CS_LIN_ERROR_PARAMETER;
        return CS_FAIL;
    }

    //Check definition pointer
    if(sig->signal_def == NULL){
        _lin_error("Invalid signal definition for signal %s \n",sig->name);
        sk->lerror = CS_LIN_ERROR_LDF;
        return CS_FAIL;
    }

    //Check frame pointer
    if(sig->frame == NULL){
        _lin_error("Invalid frame definition for signal %s\n",sig->name);
        sk->lerror = CS_LIN_ERROR_LDF;
        return CS_FAIL;
    }

    //Get encoding
    cs_lin_enc_t *enc = sig->signal_def->enc;

    if(enc == NULL){
        _lin_error("Invalid encoding definition for signal %s \n",sig->name);
        sk->lerror = CS_LIN_ERROR_LDF;
        return CS_FAIL;
    }
  
    //Check signal
    if(len !=  sig->signal_def->initlen){
        _lin_error("Invalid value length for signal %s\n",sig->name);
        sk->lerror = CS_LIN_ERROR_PARAMETER;
        return CS_FAIL;
    }

    for(i = 0; i < len; i++){
        //Set value
        sig->conv[i] = value[i];
    }

    //Now convert to raw buffer
    if(_lin_signal_conv2raw(sig)){
        _lin_error("Failure during conevrting values for signal %s\n",sig->name);
        sk->lerror = CS_LIN_ERROR_INTERNAL;
        return CS_FAIL;
    }

    //Set signal state to updated
    _lin_verbose("Signal %s state: UPDATED\n",sig->name);
    sig->state = LIN_SIGNAL_STATE_UPDATED;

    //Signal frame there is work todo
    _lin_verbose("Frame %s state: UPDATED\n",sig->frame->name);
    sig->frame->state = LIN_STATE_UPDATED;

    return CS_OK;
  
}

/********************************************************************/
int cs_lin_get_signal_raw(
    cs_lin_t *sk,           /**< Socket struct pointer*/
    cs_lin_signal_t * sig,  /**< Signal pointer*/
    cs_lin_raw_t *value,    /**< Signal value array*/
    int len                 /**< Value array len*/
    )
/********************************************************************/
{
    int i;
  
    if(sk == NULL){
        _lin_error("Parameter failure\n");
        return CS_FAIL;
    }

    if((sig == NULL) ||
       (value == NULL) ||
       (len < 1) ||
       (len > CS_LIN_SIGNAL_MAX_BYTE)){
        _lin_error("Parameter failure\n");
        sk->lerror = CS_LIN_ERROR_PARAMETER;
        return CS_FAIL;
    }

    //Check definition pointer
    if(sig->signal_def == NULL){
        _lin_error("Invalid signal definition for signal %s\n",sig->name);
        sk->lerror = CS_LIN_ERROR_LDF;
        return CS_FAIL;
    }

    //Check frame pointer
    if(sig->frame == NULL){
        _lin_error("Invalid frame definition for signal %s\n",sig->name);
        sk->lerror = CS_LIN_ERROR_LDF;
        return CS_FAIL;
    }
  
    //Check signal
    if(len !=  sig->signal_def->initlen){
        _lin_error("Invalid value buffer for signal %s\n",sig->name);
        sk->lerror = CS_LIN_ERROR_PARAMETER;
        return CS_FAIL;
    }
  
    for(i = 0; i < len; i++){
        //Set value
        value[i] = sig->conv[i];
    }  
  
    return CS_OK;
}


/********************************************************************/
int cs_lin_id_parity(
    uint32_t id,                /**< LIN Id*/
    uint32_t *pid               /**< LIN id with parity bits*/
    )
/********************************************************************/
{

    //Taken from sllin.c
    /* Values of two parity bits in LIN Protected
       Identifier for each particular LIN ID */
    const uint8_t _lin_id_parity_table[] = {
        0x80, 0xc0, 0x40, 0x00, 0xc0, 0x80, 0x00, 0x40,
        0x00, 0x40, 0xc0, 0x80, 0x40, 0x00, 0x80, 0xc0,
        0x40, 0x00, 0x80, 0xc0, 0x00, 0x40, 0xc0, 0x80,
        0xc0, 0x80, 0x00, 0x40, 0x80, 0xc0, 0x40, 0x00,
        0x00, 0x40, 0xc0, 0x80, 0x40, 0x00, 0x80, 0xc0,
        0x80, 0xc0, 0x40, 0x00, 0xc0, 0x80, 0x00, 0x40,
        0xc0, 0x80, 0x00, 0x40, 0x80, 0xc0, 0x40, 0x00,
        0x40, 0x00, 0x80, 0xc0, 0x00, 0x40, 0xc0, 0x80
    };
  
    if(pid == NULL){
        _lin_error("Parameter failure\n");
        return CS_FAIL;
    }

    *pid = id;
  
    if(id > CS_LIN_MAX_ID){
        _lin_error("ID out of range: 0x%X\n",id);
        return CS_FAIL;
    }

    *pid = id | _lin_id_parity_table[(uint8_t)id];

    //_lin_verbose("ID: 0x%X with parity: 0x%X\n",id,*pid);
   
    return CS_OK;
}

/********************************************************************/
int cs_lin_frame_is_enhanced(
    cs_lin_frame_t * frame
    )
/********************************************************************/
{
    if(frame == NULL){
        _lin_error("Parameter failure\n");
        return 0; //Default classic
    }

    if(frame->pubnode->attr->version_protocol >= 2.0){
        return 1; //Enhanced
    }
    return 0; //Classic
}
  
/********************************************************************/
uint8_t cs_lin_checksum(
    uint8_t id,           /**< LIN Id with parity bits*/
    uint8_t *data,        /**< Data buffer*/
    int len,              /**< Data len*/
    int enhanced_fl       /**< 0=> Classic CRC, 1=> Enhaced CRC*/
    )
/********************************************************************/
{
    uint16_t csum = 0;
    uint8_t rv = 0;
    int i;
  
    if (enhanced_fl == 1){
        csum += id;
    }
    for (i = 0; i < len; i++) {
        csum += data[i];
        if (csum > 0xFF){
            csum -= 0xFF;
        }
    }

    rv = (uint8_t)(~csum & 0xff);
  
    if (enhanced_fl == 1){
        _lin_verbose("Calc enhanced checksum: 0x%X\n",rv);
    } else {
        _lin_verbose("Calc classic checksum: 0x%X\n",rv);
    }
  
    return rv;
}

/********************************************************************/
cs_lin_ldf_t *cs_lin_ldf_init()
/********************************************************************/
{
    cs_lin_ldf_t *data = NULL;
    
    data = (cs_lin_ldf_t*)calloc(1,sizeof(cs_lin_ldf_t));
  
    if(data == NULL){
        _lin_error("Malloc failure\n");
        return NULL;
    }

    if(cs_lists_init(&data->l_signalencoding,_lin_ldf_cpf,_lin_ldf_free_signalencoding) ||
       cs_lists_init(&data->l_signalrepresentations,_lin_ldf_cpf,_lin_ldf_free_signalrepresentation) ||
       cs_lists_init(&data->l_signals,_lin_ldf_cpf,_lin_ldf_free_signaldefinition) ||
       cs_lists_init(&data->l_frames,_lin_ldf_cpf,_lin_ldf_free_frames) ||
       cs_lists_init(&data->l_schedulingtables,_lin_ldf_cpf,_lin_ldf_free_scheduletable) ||
       cs_lists_init(&data->l_nodes,_lin_ldf_cpf,_lin_ldf_free_node) ||
       cs_lists_init(&data->l_nodeattributes,_lin_ldf_cpf,_lin_ldf_free_nodeattributes)){
        _lin_error("List init failure\n");
        cs_lists_destroy(&data->l_signalencoding);
        cs_lists_destroy(&data->l_signalrepresentations);
        cs_lists_destroy(&data->l_signals);
        cs_lists_destroy(&data->l_frames);
        cs_lists_destroy(&data->l_schedulingtables);
        cs_lists_destroy(&data->l_nodes);
        cs_lists_destroy(&data->l_nodeattributes);
        free(data);
        return NULL;
    }
  
    return data;
}

/********************************************************************/
void cs_lin_ldf_free(
    cs_lin_ldf_t *ldf
    )
/********************************************************************/
{
  
    if(ldf == NULL)
        return;

    cs_lists_destroy(&ldf->l_signalencoding);
    cs_lists_destroy(&ldf->l_signalrepresentations);
    cs_lists_destroy(&ldf->l_signals);
    cs_lists_destroy(&ldf->l_frames);
    cs_lists_destroy(&ldf->l_schedulingtables);
    cs_lists_destroy(&ldf->l_nodes);
    cs_lists_destroy(&ldf->l_nodeattributes);
    free(ldf);
    ldf = NULL;  
}


/********************************************************************/
int cs_lin_ldf_set_general(
    cs_lin_ldf_t *ldf,         /**< LDF storage structure*/
    double prot_version,       /**< Protocol version*/
    double lang_version,       /**< Language version*/
    uint32_t speed             /**< Baud rate in bits/s*/
    )
/********************************************************************/
{
  
    if(ldf == NULL){
        _lin_error("Parameter failure\n");
        return CS_FAIL;
    }

    if(speed == 0){
        _lin_error("Invalid Baud rate\n");
        return CS_FAIL;
    }
    
    ldf->protocol_version = prot_version;
    ldf->language_version = lang_version;
    ldf->speed = speed;
    ldf->t_bit = 1.0 / speed; //in s
    _lin_verbose("Bit time is %f s\n",ldf->t_bit);
    ldf->t_header_max = CS_LIN_BITS_HEADER * ldf->t_bit * CS_LIN_TMAX_FAKTOR;
    _lin_verbose("Time max header is %f s\n",ldf->t_header_max);
    return CS_OK;
}
  
/********************************************************************/
cs_lin_enc_t* cs_lin_ldf_get_encoding(
    cs_lin_ldf_t *ldf,         /**< LDF storage structure*/
    const char *name,          /**< Enoding name*/
    cs_lin_enc_type_t basetype /**< Encoding base type*/ 
    )
/********************************************************************/
{
    cs_lin_enc_t* e = NULL;
    int i;
  
    if(ldf == NULL){
      _lin_error("Parameter failure\n");
      return NULL;
    }

    if((name == NULL)||(strlen(name) >= CS_LIN_CHARBUF)){
      _lin_error("Invalid name\n");
      return NULL;
    }

    //Search for existing
    for(i = 0; i < cs_lists_size(&ldf->l_signalencoding); i++){
        cs_lin_enc_t* le =  (cs_lin_enc_t*)cs_lists_get_iter(&ldf->l_signalencoding,i);
        if(strncmp(le->name,name,CS_LIN_CHARBUF) == 0){
            //Found node
            e = le;
            break;
        }
    }

    if(e != NULL){
        _lin_verbose("Return existing handle on %s\n",name);
        return e;
    }

    _lin_verbose("Create new encoding definition %s base type: %d\n",name,basetype);
  
    //Create new node
    e = (cs_lin_enc_t*)calloc(1,sizeof(cs_lin_enc_t));
  
    if(e == NULL){
        _lin_error("Malloc failure\n");
        return NULL;
    }

    //Init lists
    if(cs_lists_init(&e->l_logical,_lin_ldf_cpf,_lin_ldf_free_generic) ||
       cs_lists_init(&e->l_physical,_lin_ldf_cpf,_lin_ldf_free_generic)){
        _lin_error("List init failure\n");
        cs_lists_destroy(&e->l_logical);
        cs_lists_destroy(&e->l_physical);
        free(e);
        return NULL;
    }
  
    //Copy name
    strncpy(e->name, name, CS_LIN_CHARBUF);

    //Set basetype
    e->base = basetype;

    //Add entry to list
    if(cs_lists_insert(&ldf->l_signalencoding,(void*)e)){
        _lin_error("List insert failure\n");
        cs_lists_destroy(&e->l_logical);
        cs_lists_destroy(&e->l_physical);
        free(e);
        return NULL;
    }

    return e;
}

/********************************************************************/
int cs_lin_ldf_encoding_add_physical(
    cs_lin_enc_t* enc,      /**< Encoding structure*/
    cs_lin_raw_t min,       /**< Minimal raw value*/
    cs_lin_raw_t max,       /**< Maximal raw value*/
    double scale,           /**< Scale value*/
    double offset,          /**< Offset value*/
    const char *unit        /**< Unit*/
    )
/********************************************************************/
{
    cs_lin_enc_physical_t *e = NULL;
  
    if((enc == NULL) ||
       (min > max) ||
       (min < 0) ||
       (max > 0xFFFF) ||
       (scale == 0.0) ||
       (unit == NULL) ||
       (strlen(unit) >= CS_LIN_CHARBUF)
        ){
        _lin_error("Parameter failure\n");
        return CS_FAIL;
    }

    _lin_verbose("Add physical encoding definition to signal %s: Min %d, Max %d, Scale %f, Offset %f, Unit %s\n",
                 enc->name,min,max,scale,offset,unit);

    //Create new node
    e = (cs_lin_enc_physical_t*)calloc(1,sizeof(cs_lin_enc_physical_t));
  
    if(e == NULL){
        _lin_error("Malloc failure\n");
        return CS_FAIL;
    }
  
    e->min_raw = min;
    e->max_raw = max;
    e->scale = scale;
    e->offset = offset;

    //Calc physical borders
    e->min_phys = (e->min_raw * e->scale) + e->offset;
    e->max_phys = (e->max_raw * e->scale) + e->offset;
  
    strncpy(e->unit,unit,CS_LIN_CHARBUF);

    //Add entry to list
    if(cs_lists_insert(&enc->l_physical,(void*)e)){
        _lin_error("List insert failure\n");
        free(e);
        return CS_FAIL;
    }

    //Set encoding type to DYNAMIC
    enc->base = LIN_ENC_DYNAMIC;

    return CS_OK;
}


 
/********************************************************************/
int cs_lin_ldf_encoding_add_logical(
    cs_lin_enc_t* enc,      /**< Encoding structure*/
    cs_lin_raw_t value,     /**< Raw value*/
    const char *text        /**< Text*/
    )
/********************************************************************/
{
    cs_lin_enc_logical_t *e = NULL;
  
    if((enc == NULL) ||
       (text == NULL) ||
       (strlen(text) >= CS_LIN_CHARBUF)
        ){
        _lin_error("Parameter failure\n");
        return CS_FAIL;
    }

    _lin_verbose("Add logical encoding definition to signal %s: Raw %d, Text %s\n",
                 enc->name,value,text);
  
    //Create new node
    e = (cs_lin_enc_logical_t*)calloc(1,sizeof(cs_lin_enc_logical_t));
  
    if(e == NULL){
        _lin_error("Malloc failure\n");
        return CS_FAIL;
    }
  
    e->value = value;
    strncpy(e->text,text,CS_LIN_CHARBUF);

    //Add entry to list
    if(cs_lists_insert(&enc->l_logical,(void*)e)){
        _lin_error("List insert failure\n");
        free(e);
        return CS_FAIL;
    }

    //Set encoding type to DYNAMIC
    enc->base = LIN_ENC_DYNAMIC;

    return CS_OK;
}

/********************************************************************/
cs_lin_enc_representation_t* cs_lin_ldf_get_signalrepresentation(
    cs_lin_ldf_t *ldf,         /**< LDF storage structure*/
    const char *name          /**< Signal representation name*/
    )
/********************************************************************/
{
    cs_lin_enc_representation_t* e = NULL;
    int i;
  
    if(ldf == NULL){
      _lin_error("Parameter failure\n");
      return NULL;
    }
  
    if((name == NULL)||(strlen(name) >= CS_LIN_CHARBUF)){
        _lin_error("Invalid name\n");
	return NULL;
    }

    //Search for existing
    for(i = 0; i < cs_lists_size(&ldf->l_signalrepresentations); i++){
      cs_lin_enc_representation_t* le =  (cs_lin_enc_representation_t*)cs_lists_get_iter(&ldf->l_signalrepresentations,i);
      if(strncmp(le->name,name,CS_LIN_CHARBUF) == 0){
	//Found node
	e = le;
	break;
      }
    }
    
    if(e != NULL){
        _lin_verbose("Return existing handle on %s\n",name);
        return e;
    }

    _lin_verbose("Add signal represeantation %s\n",name);
  
    //Create new node
    e = (cs_lin_enc_representation_t*)calloc(1,sizeof(cs_lin_enc_representation_t));
  
    if(e == NULL){
        _lin_error("Malloc failure\n");
        return NULL;
    }

    //Init lists
    if(cs_lists_init(&e->l_signalname,_lin_ldf_cpf,_lin_ldf_free_generic)){
        _lin_error("List init failure\n");
        cs_lists_destroy(&e->l_signalname);
        free(e);
        return NULL;
    }
  
    //Copy name
    strncpy(e->name, name, CS_LIN_CHARBUF);

    //Add entry to list
    if(cs_lists_insert(&ldf->l_signalrepresentations,(void*)e)){
        _lin_error("List insert failure\n");
        cs_lists_destroy(&e->l_signalname);
        free(e);
        return NULL;
    }
  
    return e;
}

/********************************************************************/
int cs_lin_ldf_signalrepresentation_add_signal(
    cs_lin_enc_representation_t* rep,      /**< Encoding structure*/
    const char *name                       /**< Signal name*/
    )
/********************************************************************/
{
    char *e = NULL;
  
    if((rep == NULL) ||
       (name == NULL) ||
       (strlen(name) >= CS_LIN_CHARBUF)
        ){
        _lin_error("Parameter failure\n");
        return CS_FAIL;
    }

    _lin_verbose("Add signal %s to represeantation %s\n",name,rep->name);
  
    //Create new node
    e = (char*)calloc(CS_LIN_CHARBUF,sizeof(char));
  
    if(e == NULL){
        _lin_error("Malloc failure\n");
        return CS_FAIL;
    }
  
    strncpy(e,name,CS_LIN_CHARBUF);

    //Add entry to list
    if(cs_lists_insert(&rep->l_signalname,(void*)e)){
        _lin_error("List insert failure\n");
        free(e);
        return CS_FAIL;
    }

    return CS_OK;
}


/********************************************************************/
cs_lin_signal_def_t* cs_lin_ldf_get_signaldefinition(
    cs_lin_ldf_t *ldf,         /**< LDF storage structure*/
    const char *name,          /**< Signal representation name*/
    cs_lin_raw_t *init,        /**< Signal initial values*/
    int initlen,               /**< Init length in byte*/
    int len,                   /**< Signal length in bits*/
    const char *publisher      /**< Signal publisher*/
    )
/********************************************************************/
{
    cs_lin_signal_def_t* e = NULL;
    int i;
  
    if(ldf == NULL){
      _lin_error("Parameter failure\n");
      return NULL;
    }
  
    if((name == NULL)||(strlen(name) >= CS_LIN_CHARBUF)){
        _lin_error("Invalid name\n");
        return NULL;
    }

    if((init == NULL)||(initlen < 1)||(initlen > CS_LIN_SIGNAL_MAX_BYTE)){
        _lin_error("Invalid init value for %s\n",name);
        return NULL;
    }

    if((len < 1) || (len > 64)){
        _lin_error("Invalid bit length for %s\n",name);
        return NULL;
    }
  
    if((initlen > 1) && ((initlen * 8) < len)){
        _lin_error("Invalid bit len compared to init byte len for %s\n",name);
        return NULL;
    }
  
    if((publisher == NULL)||(strlen(publisher) >= CS_LIN_CHARBUF)){
        _lin_error("Invalid publisher\n");
        return NULL;
    }

    //Search for existing
    for(i = 0; i < cs_lists_size(&ldf->l_signals); i++){
        cs_lin_signal_def_t* le =  (cs_lin_signal_def_t*)cs_lists_get_iter(&ldf->l_signals,i);
        if(strncmp(le->name,name,CS_LIN_CHARBUF) == 0){
            //Found node
            e = le;
            break;
        }
    }

    if(e != NULL){
        _lin_verbose("Return existing handle on %s\n",name);
        return e;
    }

    _lin_verbose("Add signal definition %s Bitlen %d, Publisher %s\n",name,len,publisher);
  
    //Create new node
    e = (cs_lin_signal_def_t*)calloc(1,sizeof(cs_lin_signal_def_t));
  
    if(e == NULL){
        _lin_error("Malloc failure\n");
        return NULL;
    }

    //Init lists
    if(cs_lists_init(&e->l_subscribername,_lin_ldf_cpf,_lin_ldf_free_generic)){
        _lin_error("List init failure\n");
        cs_lists_destroy(&e->l_subscribername);
        free(e);
        return NULL;
    }
  
    //Fill structure
    strncpy(e->name, name, CS_LIN_CHARBUF);
    strncpy(e->publisher, publisher, CS_LIN_CHARBUF);
    e->initlen = initlen;
    memcpy(e->init,init,initlen*sizeof(cs_lin_raw_t));
    e->len = len;
  
    //Add entry to list
    if(cs_lists_insert(&ldf->l_signals,(void*)e)){
        _lin_error("List insert failure\n");
        cs_lists_destroy(&e->l_subscribername);
        free(e);
        return NULL;
    }
  
    return e;
}

/********************************************************************/
int cs_lin_ldf_signaldefinition_add_subscriber(
    cs_lin_signal_def_t* sig,      /**< Signal definition structure*/
    const char *subscriber         /**< Subscriber name*/
    )
/********************************************************************/
{
    char *e = NULL;
  
    if((sig == NULL) ||
       (subscriber == NULL) ||
       (strlen(subscriber) >= CS_LIN_CHARBUF)
        ){
        _lin_error("Parameter failure\n");
        return CS_FAIL;
    }

    _lin_verbose("Add subscriber %s to signal definition %s\n",subscriber,sig->name);
  
    //Create new node
    e = (char*)calloc(CS_LIN_CHARBUF,sizeof(char));
  
    if(e == NULL){
        _lin_error("Malloc failure\n");
        return CS_FAIL;
    }
  
    strncpy(e,subscriber,CS_LIN_CHARBUF);

    //Add entry to list
    if(cs_lists_insert(&sig->l_subscribername,(void*)e)){
        _lin_error("List insert failure\n");
        free(e);
        return CS_FAIL;
    }

    return CS_OK;
}


/********************************************************************/
cs_lin_frame_t* cs_lin_ldf_get_frame(
    cs_lin_ldf_t *ldf,         /**< LDF storage structure*/
    const char *name,          /**< Frame name*/
    uint32_t id,               /**< Frame id*/
    int dlc,                   /**< Frame DLC in bytes*/
    const char *publisher      /**< Publisher name*/
    )
/********************************************************************/
{
    cs_lin_frame_t* e = NULL;
    int i;
  
    if(ldf == NULL){
      _lin_error("Parameter failure\n");
      return NULL;
    }
  
    if((name == NULL)||(strlen(name) >= CS_LIN_CHARBUF)){
        _lin_error("Invalid name\n");
        return NULL;
    }

    if((publisher == NULL)||(strlen(publisher) >= CS_LIN_CHARBUF)){
        _lin_error("Invalid publisher\n");
        return NULL;
    }

    if((id <1)||(id > CS_LIN_MAX_ID)){
        _lin_error("Invalid frame id\n");
        return NULL;
    }

    if((dlc < 1) || (dlc > CS_LIN_SIGNAL_MAX_BYTE)){
        _lin_error("Invalid frame dlc\n");
        return NULL;
    }
  
    //Search for existing
    for(i = 0; i < cs_lists_size(&ldf->l_frames); i++){
        cs_lin_frame_t* le =  (cs_lin_frame_t*)cs_lists_get_iter(&ldf->l_frames,i);
        if(strncmp(le->name,name,CS_LIN_CHARBUF) == 0){
            //Found node
            e = le;
            break;
        }
    }

    if(e != NULL){
        _lin_verbose("Return existing handle on %s\n",name);
        return e;
    }

    _lin_verbose("Add frame definition %s, Id 0x%X, DLC %d, publisher %s\n",name,id,dlc,publisher);
  
    //Create new node
    e = (cs_lin_frame_t*)calloc(1,sizeof(cs_lin_frame_t));
  
    if(e == NULL){
        _lin_error("Malloc failure\n");
        return NULL;
    }

    //Init lists
    if(cs_lists_init(&e->l_signal,_lin_ldf_cpf,_lin_ldf_free_generic) ||
       cs_lists_init(&e->l_framename,_lin_ldf_cpf,_lin_ldf_free_generic) ||
       cs_lists_init(&e->l_frame,_lin_ldf_cpf,_lin_ldf_free_generic)
        ){
        _lin_error("List init failure\n");
        cs_lists_destroy(&e->l_signal);
        cs_lists_destroy(&e->l_framename);
        cs_lists_destroy(&e->l_frame);
        free(e);
        return NULL;
    }
  
    //Fill structure
    strncpy(e->name, name, CS_LIN_CHARBUF);
    e->id = id;
    e->dlc = dlc;
    strncpy(e->publisher, publisher, CS_LIN_CHARBUF);

    //Set default checksum type
    e->t_cksum = LIN_CKS_UNKNOWN;
  
    //Add entry to list
    if(cs_lists_insert(&ldf->l_frames,(void*)e)){
        _lin_error("List insert failure\n");
        cs_lists_destroy(&e->l_signal);
        cs_lists_destroy(&e->l_framename);
        cs_lists_destroy(&e->l_frame);
        free(e);
        return NULL;
    }
  
    return e;
}

/********************************************************************/
int cs_lin_ldf_frame_add_signal(
    cs_lin_frame_t* frame,         /**< Frame definition structure*/
    const char *name,              /**< Signal name*/
    int offset                     /**< Signal bit offset in frame*/
    )
/********************************************************************/
{
    cs_lin_signal_t *e = NULL;
  
    if((frame == NULL) ||
       (name == NULL) ||
       (strlen(name) >= CS_LIN_CHARBUF) ||
       (offset < 0) ||
       (offset > (CS_LIN_SIGNAL_MAX_BYTE * 8 -1))
        ){
        _lin_error("Parameter failure\n");
        return CS_FAIL;
    }

    _lin_verbose("Add signal %s to frame definition %s, Offset %d bits\n",name,frame->name,offset);
  
    //Create new node
    e = (cs_lin_signal_t*)calloc(1,sizeof(cs_lin_signal_t));
  
    if(e == NULL){
        _lin_error("Malloc failure\n");
        return CS_FAIL;
    }
  
    strncpy(e->name,name,CS_LIN_CHARBUF);
    e->offset = offset;
    //Store frame pointer in signal
    e->frame = frame;

    //Add entry to list
    if(cs_lists_insert(&frame->l_signal,(void*)e)){
        _lin_error("List insert failure\n");
        free(e);
        return CS_FAIL;
    }

    return CS_OK;
}

/********************************************************************/
cs_lin_schedule_table_t* cs_lin_ldf_get_schedulertable(
    cs_lin_ldf_t *ldf,         /**< LDF storage structure*/
    const char *name,          /**< Scheduler table name*/
    int flag                   /**< 0=> Only serach no creation, 1=> search and creation*/
    )
/********************************************************************/
{
    cs_lin_schedule_table_t* e = NULL;
    int i;
  
    if(ldf == NULL){
      _lin_error("Parameter failure\n");
      return NULL;
    }
  
    if((name == NULL)||(strlen(name) >= CS_LIN_CHARBUF)){
        _lin_error("Invalid name\n");
        return NULL;
    }
  
    //Search for existing
    for(i = 0; i < cs_lists_size(&ldf->l_schedulingtables); i++){
        cs_lin_schedule_table_t* le =  (cs_lin_schedule_table_t*)cs_lists_get_iter(&ldf->l_schedulingtables,i);
        if(strncmp(le->name,name,CS_LIN_CHARBUF) == 0){
            //Found node
            e = le;
            break;
        }
    }

    if(e != NULL){
        _lin_verbose("Return existing handle on %s\n",name);
        return e;
    }

    if(flag == 0){
        _lin_verbose("Scheduler table does not exist %s\n",name);
        return NULL;
    }
  
    _lin_verbose("Add scheduler table definition %s\n",name);
  
    //Create new node
    e = (cs_lin_schedule_table_t*)calloc(1,sizeof(cs_lin_schedule_table_t));
  
    if(e == NULL){
        _lin_error("Malloc failure\n");
        return NULL;
    }

    //Init lists
    if(cs_lists_init(&e->l_entry,_lin_ldf_cpf,_lin_ldf_free_generic)){
        _lin_error("List init failure\n");
        cs_lists_destroy(&e->l_entry);
        free(e);
        return NULL;
    }
  
    //Fill structure
    strncpy(e->name, name, CS_LIN_CHARBUF);
  
    //Add entry to list
    if(cs_lists_insert(&ldf->l_schedulingtables,(void*)e)){
        _lin_error("List insert failure\n");
        cs_lists_destroy(&e->l_entry);
        free(e);
        return NULL;
    }
  
    return e;
}

/********************************************************************/
int cs_lin_ldf_schedulertable_add_entry(
    cs_lin_schedule_table_t *table,/**< Scheduler table definition structure*/
    const char *name,              /**< Frame name*/
    double delay                   /**< Delay in s*/
    )
/********************************************************************/
{
    cs_lin_schedule_entry_t *e = NULL;
    int i;
    cs_lin_schedule_entry_type_t type_e = FRAME; //DEFAULT 
  
    if((table == NULL) ||
       (name == NULL) ||
       (strlen(name) >= CS_LIN_CHARBUF) ||
       (delay < 0)
        ){
        _lin_error("Parameter failure\n");
        return CS_FAIL;
    }

    _lin_verbose("Add scheduler entry %s to scheduler table definition %s with delay %f s\n",name,table->name,delay);
  
    //Check for type
    for(i = 0; i < TYPEMAX; i++){
        if(strncmp(cs_lin_ldf_schedentrytypemap[i],name,CS_LIN_CHARBUF) == 0){
            type_e = i;
            break;
        }
    }

    if(type_e != FRAME){
        _lin_error("Sorry Scheduler entry type %s is not supported until now\n",name);
        return CS_FAIL;
    }
  
    //Create new node
    e = (cs_lin_schedule_entry_t*)calloc(1,sizeof(cs_lin_schedule_entry_t));
  
    if(e == NULL){
        _lin_error("Malloc failure\n");
        return CS_FAIL;
    }
  
    strncpy(e->name,name,CS_LIN_CHARBUF);
    e->entry_type = type_e;
    e->delay = delay;

    //Add entry to list
    if(cs_lists_insert(&table->l_entry,(void*)e)){
        _lin_error("List insert failure\n");
        free(e);
        return CS_FAIL;
    }

    return CS_OK;
}

/********************************************************************/
int cs_lin_ldf_add_node(
    cs_lin_ldf_t *ldf,         /**< LDF storage structure*/
    int ismaster,              /**< 0=Slave, 1=Master*/
    const char *name,          /**< Node name*/
    double timebase,           /**< Time base in s, only for master node*/
    double jitter              /**< Jitter in s, only for master*/
    )
/********************************************************************/
{
    cs_lin_node_t *e = NULL;
  
    if((name == NULL) ||
       (strlen(name) >= CS_LIN_CHARBUF) ||
       (ismaster < 0) ||
       (ismaster > 1)
        ){
        _lin_error("Parameter failure\n");
        return CS_FAIL;
    }
  
    if(ismaster && ((timebase <= 0.0) || (jitter < 0))){
        _lin_error("Invalid timebase or jitter parameter for master node\n");
	return CS_FAIL;
    }

    _lin_verbose("Add node definition %s as %s timebase %f s jitter %f s\n",name,((ismaster==1)?"Master":"Slave"),timebase,jitter);
  
    //Create new node
    e = (cs_lin_node_t*)calloc(1,sizeof(cs_lin_node_t));
  
    if(e == NULL){
        _lin_error("Malloc failure\n");
        return CS_FAIL;
    }
  
    strncpy(e->name,name,CS_LIN_CHARBUF);
    e->timebase = timebase;
    e->jitter = jitter;
    e->ismaster = ismaster;


    //Init lists
    if(cs_lists_init(&e->l_frames_pub,_lin_ldf_cpf,_lin_ldf_free_generic)){
        _lin_error("List init failure\n");
        cs_lists_destroy(&e->l_frames_pub);
        free(e);
        return CS_FAIL;
    }

  
    //Add entry to list
    if(cs_lists_insert(&ldf->l_nodes,(void*)e)){
        _lin_error("List insert failure\n");
        free(e);
        return CS_FAIL;
    }
  
    return CS_OK;
}

/********************************************************************/
cs_lin_node_t* cs_lin_ldf_get_node(
    cs_lin_ldf_t *ldf,         /**< LDF storage structure*/
    const char *name          /**< Node name*/
    )
/********************************************************************/
{
    int i;
    cs_lin_node_t *e = NULL;
    cs_lin_node_t *rv = NULL;
   
    if((ldf == NULL)||
       (name == NULL)){
        _lin_error("Invalid parameter\n");
        return NULL;
    }
   
    for(i = 0; i < cs_lists_size(&ldf->l_nodes); i++){
        e = (cs_lin_node_t*)cs_lists_get_iter(&ldf->l_nodes,i);
        if(strncmp(e->name,name,CS_LIN_CHARBUF) == 0){
            _lin_verbose("Found node %s\n",e->name);
            rv = e;
            break;
        }
    }
    return rv;
}

/********************************************************************/
cs_lin_node_t* cs_lin_ldf_get_master(
    cs_lin_ldf_t *ldf         /**< LDF storage structure*/
    )
/********************************************************************/
{
    int i;
    cs_lin_node_t *e = NULL;
    cs_lin_node_t *rv = NULL;
   
    if((ldf == NULL)){
        _lin_error("Invalid parameter\n");
        return NULL;
    }
   
    for(i = 0; i < cs_lists_size(&ldf->l_nodes); i++){
        e = (cs_lin_node_t*)cs_lists_get_iter(&ldf->l_nodes,i);
        if(e->ismaster == 1){
            _lin_verbose("Found master node %s\n",e->name);
            rv = e;
            break;
        }
    }
    return rv;
}
 
/********************************************************************/
cs_lin_nodeattribute_t* cs_lin_ldf_get_nodeattribute(
    cs_lin_ldf_t *ldf,         /**< LDF storage structure*/
    const char *name,          /**< Node name*/
    double version_protocol,   /**< Protocol version, eg. 2.0*/
    uint32_t configured_nad,   /**< Configured NAD id*/
    uint32_t initial_nad       /**< Initial NAD id*/
    )
/********************************************************************/
{
    cs_lin_nodeattribute_t* e = NULL;
    int i;
  
    if(ldf == NULL){
      _lin_error("Parameter failure\n");
      return NULL;
    }
  
    if((name == NULL)||(strlen(name) >= CS_LIN_CHARBUF)){
        _lin_error("Invalid name\n");
        return NULL;
    }
  
    //Search for existing
    for(i = 0; i < cs_lists_size(&ldf->l_nodeattributes); i++){
        cs_lin_nodeattribute_t* le =  (cs_lin_nodeattribute_t*)cs_lists_get_iter(&ldf->l_nodeattributes,i);
        if(strncmp(le->name,name,CS_LIN_CHARBUF) == 0){
            //Found node
            e = le;
            break;
        }
    }

    if(e != NULL){
        _lin_verbose("Return existing handle on %s\n",name);
        return e;
    }

    _lin_verbose("Add node attribute definition %s, version %f, configured NAD 0x%X, initial NAD 0x%X\n",
                 name,version_protocol,configured_nad,initial_nad);
  
    //Create new node
    e = (cs_lin_nodeattribute_t*)calloc(1,sizeof(cs_lin_nodeattribute_t));
  
    if(e == NULL){
        _lin_error("Malloc failure\n");
        return NULL;
    }
  
    //Fill structure
    strncpy(e->name, name, CS_LIN_CHARBUF);
    e->version_protocol = version_protocol;
    e->configured_nad = configured_nad;
    e->initial_nad = initial_nad;
  
    //Add entry to list
    if(cs_lists_insert(&ldf->l_nodeattributes,(void*)e)){
        _lin_error("List insert failure\n");
        free(e);
        return NULL;
    }
  
    return e;
}

/********************************************************************/
int cs_lin_ldf_apply_config(
    cs_lin_ldf_t *ldf         /**< LDF storage structure*/
    )
/********************************************************************/
{
    int i,j,k;
    cs_lin_node_t* node = NULL;
    cs_lin_nodeattribute_t* nodeattribute = NULL;
    if(ldf == NULL)
        return CS_FAIL;

    //Loop signal representation
    for(i = 0 ; i < cs_lists_size(&ldf->l_signalrepresentations); i++){
        cs_lin_enc_representation_t *e = (cs_lin_enc_representation_t*)cs_lists_get_iter(&ldf->l_signalrepresentations,i);

        //Search for signal encoding definitions
        for(j = 0 ; j < cs_lists_size(&ldf->l_signalencoding); j++){
            cs_lin_enc_t *s = (cs_lin_enc_t*)cs_lists_get_iter(&ldf->l_signalencoding,j);
            if(strncmp(e->name,s->name,CS_LIN_CHARBUF) == 0){
                //Store pointer
                e->enc = s;
                _lin_verbose("Pairing signal representation %s to signal encoding %s\n",e->name,s->name);
            }
        }
        if(e->enc == NULL){
            _lin_error("No encoding representation for %s\n",e->name);
            return CS_FAIL;
        }
    }

    //Sort signals to signal encodings
    for(i = 0 ; i < cs_lists_size(&ldf->l_signalrepresentations); i++){
        cs_lin_enc_representation_t *s = (cs_lin_enc_representation_t*)cs_lists_get_iter(&ldf->l_signalrepresentations,i);
        for(j = 0 ; j < cs_lists_size(&s->l_signalname); j++){
            char *c = (char *)cs_lists_get_iter(&s->l_signalname,j);

            //Search signals
            for(k = 0 ; k < cs_lists_size(&ldf->l_signals); k++){
                cs_lin_signal_def_t *e = (cs_lin_signal_def_t *)cs_lists_get_iter(&ldf->l_signals,k);
                if(strncmp(e->name,c,CS_LIN_CHARBUF) == 0){
                    //Store encoding pointer in signal defintion
                    e->enc = s->enc;
                    _lin_verbose("Pairing signal encoding %s to signal definition %s\n",e->name,c);
                }
            }
        }
    }

    //Loop signal definitions to check for signal encoding
    for(k = 0 ; k < cs_lists_size(&ldf->l_signals); k++){
        cs_lin_signal_def_t *e = (cs_lin_signal_def_t *)cs_lists_get_iter(&ldf->l_signals,k);
        if(e->enc == NULL){
            _lin_error("No encoding definition for signal definition %s\n",e->name);
            return CS_FAIL;
        }
    }
  
    //Loop node list
    for(i = 0 ; i < cs_lists_size(&ldf->l_nodes); i++){
        node = (cs_lin_node_t*)cs_lists_get_iter(&ldf->l_nodes,i);
        //Check and store master node
        if(node->ismaster){
            //Check for double entry
            if(ldf->master != NULL){
                _lin_error("Double master node, check config\n");
                return CS_FAIL;
            }
            ldf->master = node;
        }
    
        //Now search for node attributes for this node
        nodeattribute = NULL;
        for(j = 0; j < cs_lists_size(&ldf->l_nodeattributes); j++){
            cs_lin_nodeattribute_t* e =  (cs_lin_nodeattribute_t*)cs_lists_get_iter(&ldf->l_nodeattributes,j);
            if(strncmp(e->name,node->name,CS_LIN_CHARBUF) == 0){
                //Found node
                nodeattribute = e;
                break;
            }
        }
        if(nodeattribute == NULL){
            _lin_warn("No node attributes for node %s\n",node->name);
        }
        node->attr = nodeattribute;
        _lin_verbose("Pairing node attributes %s to node %s\n",nodeattribute->name,node->name);
        //Now update all frames with node attribute information
        for(j = 0; j < cs_lists_size(&ldf->l_frames); j++){
            cs_lin_frame_t* e =  (cs_lin_frame_t*)cs_lists_get_iter(&ldf->l_frames,j);
            if(strncmp(e->publisher,node->name,CS_LIN_CHARBUF) == 0){
                //Found node

                //Store node pointer in frame
                e->pubnode = node;
	
                //Add to internal frame list
                cs_lin_frame_t** le = NULL;
                le = (cs_lin_frame_t**)malloc(sizeof(cs_lin_frame_t*));
                if(le == NULL){
                    _lin_error("Malloc failue\n");
                    return CS_FAIL;
                }
                *le = e;
                if(cs_lists_insert(&node->l_frames_pub,(void*)le)){
                    _lin_error("List insert failure\n");
                    free(le);
                    return CS_FAIL;
                }
                //Set checksum type
                _lin_verbose("Pairing frame %s to node %s\n",e->name,node->name);
                if(node->attr->version_protocol >= 2.0){
                    e->t_cksum = LIN_CKS_ENHANCED;
                } else {
                    e->t_cksum = LIN_CKS_CLASSIC;
                }
                //Check if direction is already overwriten
                if(e->dir == LIN_DIR_UNKNOWN){
                    //Set working direction
                    if(node->ismaster){
                        _lin_verbose("Store master send for frame %s in node %s\n",e->name,node->name);
                        e->dir = LIN_DIR_SEND;
                    } else {
                        _lin_verbose("Store master receive for frame %s in node %s\n",e->name,node->name);
                        e->dir = LIN_DIR_RECEIVE;
                    }
                }
            }
        }
    }
  
    //Check if there is a existing master node
    if(ldf->master == NULL){
        _lin_error("No master node defined, check config\n");
        return CS_FAIL;
    }
  
    //Next run, loop all frames
    for(i = 0; i < cs_lists_size(&ldf->l_frames); i++){
        cs_lin_frame_t* e =  (cs_lin_frame_t*)cs_lists_get_iter(&ldf->l_frames,i);
    
        //Check for checksum
        if(e->t_cksum == LIN_CKS_UNKNOWN){
            _lin_error("Checksum for frame %s is not configured\n",e->name);
            return CS_FAIL;
        }
    
        //Check for direction
        if(e->dir == LIN_DIR_UNKNOWN){
            _lin_error("Direction for frame %s is not configured\n",e->name);
            return CS_FAIL;
        }

	//Calculate maximum frame time
	e->t_frame_max = ldf->t_header_max + (10 * (e->dlc + 1) * ldf->t_bit *CS_LIN_TMAX_FAKTOR);
	_lin_debug("Frame %s(0x%X) max time is %f s\n",e->name,e->id,	e->t_frame_max);

	
        //Loop over frame signal list
        for(j = 0; j < cs_lists_size(&e->l_signal); j++){
            cs_lin_signal_t* s =  (cs_lin_signal_t*)cs_lists_get_iter(&e->l_signal,j);
      
            //Loop about signal definition list
            for(k = 0; k < cs_lists_size(&ldf->l_signals); k++){
                cs_lin_signal_def_t* d =  (cs_lin_signal_def_t*)cs_lists_get_iter(&ldf->l_signals,k);
                if(strncmp(d->name,s->name,CS_LIN_CHARBUF) == 0){
                    //Found signal definition for frame signal
                    s->signal_def = d;
                    _lin_verbose("Pairing frame %s signal %s to signal definition %s\n",e->name,s->name,d->name);
                    //Check if signal fits into frame
                    if((s->offset + d->len) > (e->dlc * 8)){
                        _lin_error("Signal %s with offset %d Bits and length %d Bits does not fit into frame %s length %d Bits\n",
                                   s->name, s->offset, d->len, e->name, e->dlc*8);
                        return CS_FAIL;
                    }	    
                }
            }
            if(s->signal_def == NULL){
                _lin_error("No signal definition for signal %s\n",s->name);
                return CS_FAIL;
            }
        }
    
    }
  
    //Now loop scheduler tables
    for(i = 0; i < cs_lists_size(&ldf->l_schedulingtables); i++){
        cs_lin_schedule_table_t* e =  (cs_lin_schedule_table_t*)cs_lists_get_iter(&ldf->l_schedulingtables,i);
    
        //Loop all scheduler entries ant search for frames
        for(j = 0; j < cs_lists_size(&e->l_entry); j++){
            cs_lin_schedule_entry_t* s =  (cs_lin_schedule_entry_t*)cs_lists_get_iter(&e->l_entry,j);
      
            //Handle standard frames
            if(s->entry_type == FRAME){
                for(k = 0; k < cs_lists_size(&ldf->l_frames); k++){
                    cs_lin_frame_t* f =  (cs_lin_frame_t*)cs_lists_get_iter(&ldf->l_frames,k);
                    if(strncmp(f->name,s->name,CS_LIN_CHARBUF) == 0){
                        //found frame for scheduler entry
                        s->frame = f;
                        _lin_verbose("Pairing scheduler %s entry to frame definition %s\n",e->name,s->name,f->name);
                    }
                }
		if(s->frame == NULL){
		  _lin_error("No frame defined for scheduler entry %s\n",s->name);
		  return CS_FAIL;
		}
		//Calculate frame slot time
		s->t_frame_slot = ldf->master->jitter + s->frame-> t_frame_max;
		_lin_debug("Scheduler entry %s: Frame slot time is %f s\n",s->name,s->t_frame_slot);
            }
        }
    }
  
    return CS_OK;
}

