/*  
    Copyright (C) 2017 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "config.h"
#include "cs_defines.h"
#include "cs_lin.h"
#include "cs_lin_bit.h"
#include "cs_lin_bit_private.h"
#include "cs_debug_private.h"

#define LIN_BIT_HIGH 0x1
#define LIN_BIT_LOW 0x0


#define _lin_bit_error(...) (cs_error(CS_DEBUG_LIN_BIT,__func__,__LINE__,__VA_ARGS__))
#define _lin_bit_warn(...) (cs_warn(CS_DEBUG_LIN_BIT,__func__,__LINE__,__VA_ARGS__))
#define _lin_bit_info(...) (cs_info(CS_DEBUG_LIN_BIT,__func__,__LINE__,__VA_ARGS__))
#define _lin_bit_debug(...) (cs_debug(CS_DEBUG_LIN_BIT,__func__,__LINE__,__VA_ARGS__))
#define _lin_bit_verbose(...) (cs_verbose(CS_DEBUG_LIN_BIT,__func__,__LINE__,__VA_ARGS__))
#define _lin_bit_error_buffer(DATA,LEN) (cs_error_buffer(CS_DEBUG_LIN_BIT,__func__,__LINE__,DATA,LEN))
#define _lin_bit_warn_buffer(DATA,LEN) (cs_warn_buffer(CS_DEBUG_LIN_BIT,__func__,__LINE__,DATA,LEN))
#define _lin_bit_info_buffer(DATA,LEN) (cs_info_buffer(CS_DEBUG_LIN_BIT,__func__,__LINE__,DATA,LEN))
#define _lin_bit_debug_buffer(DATA,LEN) (cs_debug_buffer(CS_DEBUG_LIN_BIT,__func__,__LINE__,DATA,LEN))
#define _lin_bit_verbose_buffer(DATA,LEN) (cs_verbose_buffer(CS_DEBUG_LIN_BIT,__func__,__LINE__,DATA,LEN))


/********************************************************************/
cs_debug_level_t cs_lin_bit_debuglevel(
				       cs_debug_level_t level
				       )
/********************************************************************/
{
  return cs_debuglevel(CS_DEBUG_LIN_BIT,level);
}

/********************************************************************/
void cs_lin_bit_debug_setcallback(
				  void *ctx,
				  void (*callb)(void *ctx,
						cs_debug_sub_t sub,
						cs_debug_level_t dl,
						const char *func,
						const int line,
						const char *fmt,
						va_list ap)
				  )
/********************************************************************/
{
  cs_debug_setcallback(CS_DEBUG_LIN_BIT,ctx,callb);
}

/** Calculate Lin bit array to interface byte array length
    Uses an internal integrator to compensate if length which is not byte align
    If Integrator > 0.8 add a byte
    \retval interface number of bytes to represent given LIN bits
*/ 
/********************************************************************/
int _lin_phys_bit_calc_len(
    cs_lin_phys_bit_t *data,    /**< Data structure*/
    int bits                    /**< Bit length in LIN side*/
    )
/********************************************************************/
{
    double len;
    int rv = 0;
    len = (data->lin_bits * bits);
    rv  = (int)len;
    data->if_byte_shift = data->if_byte_shift +(len - rv);
    if(data->if_byte_shift > 0.8){
        rv += 1;
        data->if_byte_shift = data->if_byte_shift - 1.0;
        //_lin_bit_verbose("Add Byte LIN bits: %f,shift: %f Len: %f, rv %d, Rest: %f\n",data->lin_bits,data->if_byte_shift,len,rv,len - rv);
    }
    return rv;
}

/** Convert one LIN byte to interface byte array
    Adds Start and Stop bit
    \retval >=0 Length of used buffer in bytes
    \retval -1 Failure
*/
/********************************************************************/
int _lin_phys_bit_set_data(
    cs_lin_phys_bit_t *data,    /**< Data structure*/
    uint8_t byte,               /**< Data byte to write*/
    uint8_t *buffer,            /**< Output buffer*/
    int len                     /**< Output buffer len*/
    )
/********************************************************************/
{
    int offset = 0;
    int iflen;
    int i;

    //Start bit
    iflen = _lin_phys_bit_calc_len(data,CS_LIN_PHYS_TIME_START);
    if(iflen > len){
        _lin_bit_error("Buffer missmatch\n");
        return -1;
    }
    memset(buffer,LIN_BIT_LOW,sizeof(uint8_t)*iflen);
    offset += iflen;
    //Loop about data bits
    for(i = 0; i < CS_LIN_PHYS_BITS; i++){
        //get bit length
        iflen = _lin_phys_bit_calc_len(data,CS_LIN_PHYS_TIME_DATA);
        if((offset + iflen) > len){
            _lin_bit_error("Buffer missmatch\n");
            return -1;
        }
        if(byte & (1<<i)){
            //Bit is not zero
            memset(&buffer[offset],LIN_BIT_HIGH,sizeof(uint8_t)*iflen);
        } else {
            //zero
            memset(&buffer[offset],LIN_BIT_LOW,sizeof(uint8_t)*iflen);
        }
        offset += iflen;
    }
    //stop bit
    iflen = _lin_phys_bit_calc_len(data,CS_LIN_PHYS_TIME_STOP);
    if((offset + iflen) > len){
        _lin_bit_error("Buffer missmatch\n");
        return -1;
    }
    memset(&buffer[offset],LIN_BIT_HIGH,sizeof(uint8_t)*iflen);
    offset += iflen;

    //Interbyte space
    iflen = _lin_phys_bit_calc_len(data,CS_LIN_PHYS_TIME_INTERBYTE);
    if((offset + iflen) > len){
        _lin_bit_error("Buffer missmatch\n");
        return -1;
    }
    memset(&buffer[offset],LIN_BIT_HIGH,sizeof(uint8_t)*iflen);
    offset += iflen;
  
    return offset;
}


/** Init LIN Bit structure
    Precalculate break and sync field
    \retval != NULL Data structure
    \retval NULL Failure
*/
/********************************************************************/
cs_lin_phys_bit_t* cs_lin_phys_bit_init(
    uint32_t lin_baud,     /**< LIN-Baud rate in baud/s*/
    uint32_t if_baud       /**< Interface Baud rate in baud/s*/
    )
/********************************************************************/
{
    cs_lin_phys_bit_t *data = NULL;
    int offset = 0;
    int len;
  
    if((lin_baud == 0) || (if_baud == 0) || (if_baud < lin_baud)){
        _lin_bit_error("Parameter failure\n");
        return NULL;
    }

    /*    if((lin_baud != 19200) || (if_baud != 1000000)){
        _lin_bit_error("Not supported baud rates; Only Lin: 19.2kB and IF: 1MB are actually supported\n");
        return NULL;
	}*/

    _lin_bit_verbose("Lin: %i Baud and Interface: %i Baud\n",lin_baud,if_baud);
  
    data = (cs_lin_phys_bit_t*)calloc(1,sizeof(cs_lin_phys_bit_t));
    if(data == NULL){
        _lin_bit_error("Malloc failure\n");
        return NULL;
    }

    //Actually only 19.2kBaud to 1MBaud is supported
    //means 1 Lin Bit is 52.03 IF Bits
 
    data->if_byte_shift = 0;
    data->lin_bits = ((double)if_baud) / ((double)lin_baud);
    int break_pause_len = _lin_phys_bit_calc_len(data,CS_LIN_PHYS_TIME_BREAK);
    int break_high_len = _lin_phys_bit_calc_len(data,CS_LIN_PHYS_TIME_BREAKDELIMITER + CS_LIN_PHYS_TIME_INTERBYTE);
    data->if_break_len = break_pause_len + break_high_len;
    data->if_break = (uint8_t*)calloc(1,sizeof(uint8_t)*(data->if_break_len + 10)); //10 spare bytes
    if(data->if_break == NULL){
        _lin_bit_error("Malloc failure\n");
        free(data);
        return NULL;
    }
    //Break is LIN CS_LIN_PHYS_TIME_BREAK Bits long, and all zero
    offset = break_pause_len;
    //Set all other bytes to one
    memset(&data->if_break[offset],LIN_BIT_HIGH,(data->if_break_len - offset - 1));
    //Ok break field is done

    data->if_sync_len = (int)(((CS_LIN_PHYS_TIME_BYTE + CS_LIN_PHYS_TIME_INTERBYTE) * data->lin_bits) + 100); //spare bytes
    data->if_sync = (uint8_t*)calloc(1,sizeof(uint8_t)*(data->if_sync_len));
    if(data->if_sync == NULL){
        _lin_bit_error("Malloc failure\n");
        free(data->if_break);
        free(data);
        return NULL;
    }
    if((len = _lin_phys_bit_set_data(data,CS_LIN_SYNC_DATA,data->if_sync,data->if_sync_len)) < 0){
        _lin_bit_error("Failed to init sync field\n");
        free(data->if_sync);
        free(data->if_break);
        free(data);
        return NULL;
    }
    //set sync len to the real used one
    data->if_sync_len = len;
    //Store constant header shift
    data->if_byte_shift_header = data->if_byte_shift;
    //Store minimal size
    data->min_size = data->if_break_len + data->if_sync_len * (3 + CS_LIN_MAXDLC + 10); //Add spare bytes (gap bettween frames)

    //Calc limits
    data->minbreak = data->lin_bits * CS_LIN_PHYS_TIME_BREAK_SCAN_MIN;
    data->maxbreak = data->lin_bits * CS_LIN_PHYS_TIME_BREAK_SCAN_MAX;
    data->minbitlength = data->lin_bits * (1.0 - CS_LIN_PHYS_TIME_BIT_TOL);
    data->maxbitlength = data->lin_bits * (1.0 + CS_LIN_PHYS_TIME_BIT_TOL);

    _lin_bit_verbose("Break length min: %f max: %f\n",data->minbreak,data->maxbreak);
    _lin_bit_verbose("Bit length min: %f max: %f\n",data->minbitlength,data->maxbitlength);

    //TODO measure timing
    //Calc half bit offset
    data->data_start_offset = data->lin_bits / 2.0;

    //reset counter
    data->buffercounter = 0;
    
    return data;
}

/** Free structure
    \retval 1 Failure
    \retval 0 No Failure
*/
/********************************************************************/
int cs_lin_phys_bit_free(
    cs_lin_phys_bit_t* bt         /**< Data structure*/
    )
/********************************************************************/
{
  
    if(bt == NULL){
        _lin_bit_error("Parameter failure\n");
        return CS_FAIL;
    }

    if(bt->if_sync != NULL){
        free(bt->if_sync);
    }

    if(bt->if_break != NULL){
        free(bt->if_break);
    }

    free(bt);
    bt = NULL;

    return CS_OK;
}
  
/** Get the minimum interface data buffer size to handle a LIN frame with maximum length
    \retval >=0 Number of bytes
    \retval -1 Failure
*/ 
/********************************************************************/
int cs_lin_phys_bit_get_buffer_size(
    cs_lin_phys_bit_t* bt         /**< Data structure*/
    )
/********************************************************************/
{
    if(bt == NULL){
        _lin_bit_error("Parameter failure\n");
        return -1;
    }
    return bt->min_size;
}

/** Prepare a send interface send buffer
    \retval 0 No failure
    \retval 1 Failure
*/
/********************************************************************/
int cs_lin_phys_bit_send(
    cs_lin_phys_bit_t* bt,         /**< Data structure*/
    uint8_t *bits,                 /**< Interface bit field*/
    int *bitlen,                   /**< Bitbuffer len, will be overwritten by use of buffer*/
    uint8_t pid,                   /**< Protected identifier*/
    uint8_t *data,                 /**< LIN data*/
    int datalen,                   /**< Lin data length*/
    uint8_t crc                    /**< Check sum*/
    )
/********************************************************************/
{
    int i;
    int bufferlen;
    int len;
  
    if((bt == NULL) ||
       (bits == NULL) ||
       (bitlen == NULL) ||
       (data == NULL)){
        _lin_bit_error("Parameter failure\n");
        return CS_FAIL;
    }

    //Check length, note:
    //Frame has break field, synch data, pid field, data bytes * len and crc
    //We know the break field length and the synch byte length
    //synch byte is a normal data byte, so data bytes, pid field will have the same length (note on integrator)
  
    if(*bitlen < (bt->min_size)){
        _lin_bit_error("Interface bit buffer is to small\n");
        return CS_FAIL;
    }
    //Reset output buffer
    memset(bits,LIN_BIT_HIGH,sizeof(uint8_t)*(*bitlen));

    //Store buffer length
    bufferlen = *bitlen;
  
    //Reset bitlen
    *bitlen = 0;

    //Reset shift
    bt->if_byte_shift = bt->if_byte_shift_header;
  
    //Start with break field
    memcpy(bits,bt->if_break,sizeof(uint8_t)*bt->if_break_len);
    *bitlen += bt->if_break_len;
  
    //Synch field
    memcpy(&bits[*bitlen],bt->if_sync,sizeof(uint8_t)*bt->if_sync_len);
    *bitlen += bt->if_sync_len;
  
    //PID field
    if((len = _lin_phys_bit_set_data(bt,pid,&bits[*bitlen],bufferlen - *bitlen)) < 0){
        _lin_bit_error("Failed to add PID field\n");
        return CS_FAIL;
    }  
    *bitlen +=len;
  
    if(datalen > 0){
        //Data segment
        for(i = 0; i < datalen; i++){
            if((len = _lin_phys_bit_set_data(bt,data[i],&bits[*bitlen],bufferlen - *bitlen)) < 0){
                _lin_bit_error("Failed to add data field %d\n",i);
                return CS_FAIL;
            }
            *bitlen +=len;
        }
    }

    //CRC Field
    if((len = _lin_phys_bit_set_data(bt,crc,&bits[*bitlen],bufferlen - *bitlen)) < 0){
        _lin_bit_error("Failed to add CRC field\n");
        return CS_FAIL;
    }  
    *bitlen +=len;

    //All other bits in frame set to high
    memset(&bits[*bitlen],LIN_BIT_HIGH,bufferlen - *bitlen);
  
    return CS_OK;
}

/** Prepare a send request interface send buffer
    \retval 0 No failure
    \retval 1 Failure
*/
/********************************************************************/
int cs_lin_phys_bit_send_req(
    cs_lin_phys_bit_t* bt,         /**< Data structure*/
    uint8_t *bits,                 /**< Interface bit field*/
    int *bitlen,                   /**< Bitbuffer len, will be overwritten by use of buffer*/
    uint8_t pid                   /**< Protected identifier*/
    )
/********************************************************************/
{
    int bufferlen;
    int len;
  
    if((bt == NULL) ||
       (bits == NULL) ||
       (bitlen == NULL)){
        _lin_bit_error("Parameter failure\n");
        return CS_FAIL;
    }

    //Check length, note:
    //Frame has break field, synch data, pid field, data bytes * len and crc
    //We know the break field length and the synch byte length
    //synch byte is a normal data byte, so data bytes, pid field will have the same length (note on integrator)
  
    if(*bitlen < (bt->min_size)){
        _lin_bit_error("Interface bit buffer is to small\n");
        return CS_FAIL;
    }
    //Reset output buffer
    memset(bits,LIN_BIT_HIGH,sizeof(uint8_t)*(*bitlen));

    //Store buffer length
    bufferlen = *bitlen;
  
    //Reset bitlen
    *bitlen = 0;

    //Reset shift
    bt->if_byte_shift= bt->if_byte_shift_header;
  
    //Start with break field
    memcpy(bits,bt->if_break,sizeof(uint8_t)*bt->if_break_len);
    *bitlen += bt->if_break_len;
  
    //Synch field
    memcpy(&bits[*bitlen],bt->if_sync,sizeof(uint8_t)*bt->if_sync_len);
    *bitlen += bt->if_sync_len;
  
    //PID field
    if((len = _lin_phys_bit_set_data(bt,pid,&bits[*bitlen],bufferlen - *bitlen)) < 0){
        _lin_bit_error("Failed to add PID field\n");
        return CS_FAIL;
    }  
    *bitlen +=len;
  
    //All other bits in frame set to high
    memset(&bits[*bitlen],LIN_BIT_HIGH,bufferlen - *bitlen);
  
    return CS_OK;
}

/********************************************************************/
int cs_lin_phys_lin_scan(
    cs_lin_phys_bit_t* bt,         /**< Data structure*/
    uint8_t *bits,                 /**< Input bitfield*/
    int bitlen,                    /**< Bitfield length in byte*/
    int *offset,                   /**< Startoffset of data segment, if we need a second call*/
    uint8_t *scan_pid,             /**< Scan PID content*/
    uint8_t *scan_data,            /**< Scan data field, assumes length CS_LIN_SIGNAL_MAX_BYTE*/
    int *scan_length,              /**< Scan data length, min CS_LIN_SIGNAL_MAX_BYTE long*/
    uint8_t *scan_crc              /**< Scan CRC field*/		 
    )
/********************************************************************/
{ 
    if((bt == NULL) ||
       (bits == NULL) ||
       (bitlen < 0) ||
       (offset == NULL) ||
       (scan_pid == NULL) ||
       (scan_data == NULL) ||
       (scan_length == NULL) ||
       (*scan_length <  CS_LIN_SIGNAL_MAX_BYTE) ||
       (scan_crc == NULL)
        ){
        _lin_bit_error("Parameter failure\n");
        return -1;
    }


    int maxwait = 0;
    int i,j;
    bt->buffercounter++;
    
    if((*offset >= bitlen) || (*offset < 0)){
        _lin_bit_error("%d:Offset out of buffer failure\n",bt->buffercounter);
        return -1;
    }
    //_lin_bit_verbose("Buffer size: %i Bits\n",bitlen);
    //_lin_bit_verbose("Buffer offset: %i Bits\n",*offset);


    
    //Reset feedback values
    *scan_pid = 0;
    for(j = 2; j < *scan_length; j++){
        scan_data[j] = 0;
    }
    *scan_length = 0;
    *scan_crc = 0;
  
    for(i = *offset; i < bitlen; i++){
      
        switch(bt->scan_state){
        case B_STATE_IDLE:
            //Reset counter
            bt->counter = 0;
            bt->scan_data_state = B_STATE_DATA_IDLE;
            //Scan for break start
            if(bits[i] == LIN_BIT_LOW){
	        _lin_bit_verbose("%d: Detect break start, frame pos: %i\n",bt->buffercounter,i);
                //If low maybe we are inside break field
                bt->counter++;
                bt->scan_state = B_STATE_BREAK;
                break;
            }
            break;
        case B_STATE_BREAK:
            if(bits[i] == LIN_BIT_LOW){
                //We are still getting a low bit; continue
                bt->counter++;
            } else {
                if(bt->counter > bt->minbreak){
		    _lin_bit_verbose("%d: Detect break end, length %d, frame pos: %i \n",bt->buffercounter,bt->counter,i);
                    //We got a high bit, this is the end of the break field
                    bt->counter = 0;
                    bt->scan_state = B_STATE_SYNC;
                    //Reset scan data state
                    bt->scan_data_state = B_STATE_DATA_IDLE;
                    bt->data_idle_counter = 0;
                    break;
                } else {
       		    _lin_bit_warn("%d: Detect break level failure, frame pos: %i\n",bt->buffercounter,i);
                    //We got a high bit, but we do not reach a valid break length, reset
                    bt->counter = 0;
                    bt->scan_state = B_STATE_IDLE;
		    *offset = 0;
                    return 4; //Signal there is inconsistent content
                }
            }
            if(bt->counter > bt->maxbreak){
	        //_lin_bit_verbose("%d: Detect break is too long, frame pos: %i\n",bt->buffercounter,i);
                //We got too much low bits
                //This is not a break maybe cable defect or system is shut off
                bt->counter = 0;
                bt->scan_state = B_STATE_IDLE;
            }
            break;
        case B_STATE_SYNC:
        case B_STATE_DATA:
            switch(bt->scan_data_state){
            case B_STATE_DATA_IDLE:
                if(bits[i] == LIN_BIT_HIGH){
                    //Wait for start bit
                    bt->data_idle_counter++;
                    //We are waiting for interbyte space or Response space
                    //depending on the byte position
                    switch(bt->data_byte_index){
                    case 2:
                        //We got the PID field
                        //We have to wait the response space
                        maxwait = bt->lin_bits * 10; //Max 20x bit times, TODO how to set this value correctly
                        break;
                    default:
                        //Normal data or synch
                        //and wait for interbyte space
                        maxwait = bt->lin_bits * (CS_LIN_PHYS_TIME_INTERBYTE + 0.5 + 1); // 0.5 (Rest stop bit)+ 1 interbyte space + 1 spare
                        break;
                    }
                    if(bt->data_idle_counter > maxwait){
		        //_lin_bit_verbose("%d: Detect Byte start timeout, Timeout %i Bits\n",bt->buffercounter,bt->data_idle_counter);
                        //I think we got a timout
                        //This can be the end of a data frame
                        //This can be that a slave does not answer
                        //This can be a protocol failure
                        //Set offset to the current working point
                        *offset = i;
                        //Reset
                        bt->counter = 0;
                        bt->scan_state = B_STATE_IDLE;
                        if(bt->data_byte_index < 2){
			    _lin_bit_warn("%d: We got on inclomplete header, Received %i Bytes\n",bt->buffercounter,bt->data_byte_index-1);
                            //We did not receive a valid header, Failure
			    *offset = 0;
                            return 4;
                        }
                        if(bt->data_byte_index < 3){
			    _lin_bit_warn("%d: We received a header but no response\n",bt->buffercounter);
                            //We received a valid header, but no response from a slave
			    *offset = 0;
			    *scan_pid = bt->frame_data[1];
                            return 2;
                        }
	      
                        if(bt->data_byte_index == 3){
			    _lin_bit_warn("%d: We received a invalid data segment, Received %i Bytes\n",bt->buffercounter,bt->data_byte_index-1);
                            //We received a valid header, and a data byte,
                            //but we need minimum 2 data bytes
			    *offset = 0;
                            return 4;
                        }
                        if(bt->data_byte_index > 3){
			    _lin_bit_verbose("%d: Received a valid frame, Received %i Bytes\n",bt->buffercounter,bt->data_byte_index-1);
                            //We received a valid header and a valid data segment
                            //Prepare feedback
                            //bt->frame_data[0] is synch field
                            *scan_pid = bt->frame_data[1];
                            for(j = 2; j < (bt->data_byte_index-1); j++){
                                scan_data[j-2] =  bt->frame_data[j];
                            }
                            *scan_length = bt->data_byte_index -3;
                            *scan_crc = bt->frame_data[bt->data_byte_index-1];
                            return 3;
                        }
                    }
                    break;
                } else {
		    _lin_bit_verbose("%d: Detect Start bit start after %d , frame pos: %i\n", bt->buffercounter,bt->data_idle_counter,i);
                    //We got a start bit low
                    bt->scan_data_state = B_STATE_DATA_START;
                    bt->data_idle_counter = 0;
                    bt->counter = 1; //Init with one because this is already content of the start bit
                    //Init integrator
                    //TODO measure timing
                    bt->data_sample_int = bt->data_start_offset;
                    break;
                }
                break;
            case B_STATE_DATA_START:
                //We wait for end of start bit
                if(bits[i] == LIN_BIT_LOW){
                    bt->counter++;
                    if(bt->counter >= bt->data_sample_int){
                        //We reach bit sample position (middle of bit)
                        //Add one bit count to integrator
                        bt->data_sample_int += bt->lin_bits;
                        bt->scan_data_state = B_STATE_DATA_DATA;
                        bt->data_bit_index = 0;
                        if(bt->scan_state == B_STATE_SYNC){
                            //If we prepare for synch byte
                            //Reset storage index on byte level
                            bt->data_byte_index = 0;
                        }
			//Reset internal storage
			bt->frame_data[bt->data_byte_index] = 0;
                        break;
                    }
                } else {
                    //We detect a high bit
                    if( bt->counter < bt->minbitlength){
	  	        _lin_bit_warn("%d: Received invalid Start bit data, too short\n",bt->buffercounter);
                        //Invalid data, reset
                        bt->counter = 0;
                        bt->scan_state = B_STATE_IDLE;
                        break;
                    }
                }
                if( bt->counter > bt->maxbitlength){
                    //Invalid data, bit is too long, reset
		    _lin_bit_verbose("%d: Received invalid bit data, too long\n",bt->buffercounter);
                    bt->counter = 0;
                    bt->scan_state = B_STATE_IDLE;
		    return 4; // Signal for inconsistent data
                }
                break;
            case B_STATE_DATA_DATA:
                //Increment counter
                bt->counter++;
                if( bt->counter > bt->data_sample_int){
                    //We reach the sample point, take data
                    //Add one bit count to integrator
                    bt->data_sample_int += bt->lin_bits;
                    //LSB first
                    if(bt->data_bit_index < CS_LIN_PHYS_BITS){
                        //Store the data
                        bt->frame_data[bt->data_byte_index] |= (bits[i] == LIN_BIT_HIGH) << bt->data_bit_index;
                        _lin_bit_verbose("%d: Set byte %i: Bit index: %i frame pos: %i Bit: %X Byte: %X \n",bt->buffercounter,bt->data_byte_index,bt->data_bit_index,i,bits[i],bt->frame_data[bt->data_byte_index]);
                    } else {	    
                        //We are looking for the Stop bit
                        //Only validate if we got a high bit
                        if(bits[i] != LIN_BIT_HIGH){
                            //Failure
			    _lin_bit_warn("%d: Received invalid stop bit\n",bt->buffercounter);
			    bt->counter = 0;
			    bt->scan_state = B_STATE_IDLE;
			    *offset=0;
			    return 4; // Signal for inconsistent data
                        }
                        //End of byte reached
                        bt->data_bit_index = 0;
                        bt->data_byte_index++;
                        if(bt->scan_state == B_STATE_SYNC){
			  if(bt->frame_data[0] == CS_LIN_SYNC_DATA){
                            _lin_bit_verbose("%d: Received sync byte %X\n",bt->buffercounter,bt->frame_data[0]);
                            //We finished synch byte continue PID and data segment
                            bt->scan_state = B_STATE_DATA;
			  } else {
			    _lin_bit_warn("%d: Received invalid sync byte %X\n",bt->buffercounter,bt->frame_data[0]);
			    //Reset
			    *offset = 0; //back to start
			    bt->counter = 0;
			    bt->scan_state = B_STATE_IDLE;
			    bt->scan_data_state = B_STATE_DATA_IDLE;
			    bt->data_idle_counter = 0;
			    return 4;
			  }
                        } else {
			    _lin_bit_verbose("%d: Received byte %X\n",bt->buffercounter,bt->frame_data[bt->data_byte_index -1]);
                        }
                        //Restart and wait for interbyte space and start bit
                        //We got a high bit, this is the end of the break field
                        bt->counter = 0;
                        //Reset scan data state
                        bt->scan_data_state = B_STATE_DATA_IDLE;
                        bt->data_idle_counter = 0;
                        //Check for byte index
                        if(bt->data_byte_index >= CS_LIN_PHYS_BIT_DATA_STORE){
                            //End of possible data area reached
                            //Continue and wait for timeout
                            //TODO, special handling?
                        }
                        break;
                    }
                    //Increment bit index
                    bt->data_bit_index++;
                }
                break;
            }
            break;
        }
    }
    if(bt->scan_state ==  B_STATE_IDLE){
        //Set offset to the current working point
        *offset = 0;
        //Reset
        bt->counter = 0;
        return 0;
    } else {
        //_lin_bit_verbose("%d: Frame in progress, State: %d, data state: %d\n",bt->buffercounter,bt->scan_state,bt->scan_data_state);
        *offset = i;
        return 1;
    }
}
