/*  
    Copyright (C) 2018 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "config.h"
#include "cs_defines.h"
#include "cs_parser.h"
#include "cs_lin_private.h"
#include "cs_lin.h"
#include "cs_debug_private.h"

#define _lin_info(...) (cs_info(CS_DEBUG_LIN,__func__,__LINE__,__VA_ARGS__))
#define _lin_debug(...) (cs_debug(CS_DEBUG_LIN,__func__,__LINE__,__VA_ARGS__))
#define _lin_verbose(...) (cs_verbose(CS_DEBUG_LIN,__func__,__LINE__,__VA_ARGS__))
#define _lin_debug_buffer(DATA,LEN) (cs_debug_buffer(CS_DEBUG_LIN,__func__,__LINE__,DATA,LEN))
#define _lin_verbose_buffer(DATA,LEN) (cs_verbose_buffer(CS_DEBUG_LIN,__func__,__LINE__,DATA,LEN))

/** Cleanup LDF file
 */
/**************************************************************************/
int _lin_ldf_doc_cleanup(char *in, char **out, int size)
/**************************************************************************/
{
  // remove comments => /*jfhjgfhkgh*/
  // remove comments => // ***\n
  // remove \r
  // replace \t by ' '
  // replace \n by ' '
  char func[] = "_lin_ldf_doc_cleanup";
  int i,j;
  int position;
  int fcp;
  char replace;
  enum {
    NORMAL = 0,
    POSSIBLE_START,
    POSSIBLE_END,
    INCOMMENT_STAR,
    INCOMMENT_LINE
  }stm = NORMAL;

  if((in == NULL)||(out == NULL)||(size <= 0)){
    _lin_debug("%s: Parameter failure\n",func);
    return CS_FAIL;
  }

  //Malloc space
  *out = calloc(size,sizeof(char));
  if(*out == NULL){
    _lin_debug("%s: Malloc failure\n",func);
    return CS_FAIL;
  }
  position = 0;
  for(i = 0; i < size; i++){
    fcp = -1;
    //Remove comments
    switch(stm){
    case NORMAL:
      switch(in[i]){
      case '/':
	stm = POSSIBLE_START;
	//_lin_verbose("%s: Detect possible comment start\n",func);
	break;
      case '\r':
	break;
      case '\n':
      case '\t':
	fcp = -2;
	replace = ' ';
	break;
      default:
	fcp = 0;
      }
      break;
    case POSSIBLE_START:
      if(in[i] == '*'){
	stm = INCOMMENT_STAR;
	//_lin_verbose("%s: Detect multiline comment start\n",func);
      } else {
	if(in[i] == '/'){
	  stm = INCOMMENT_LINE;
	  //_lin_verbose("%s: Detect single line comment start\n",func);
	} else {
	  stm = NORMAL;
	  //_lin_verbose("%s: No comment\n",func);
	  fcp = 1;
	}
      }
      break;
    case POSSIBLE_END:
      if(in[i] == '/'){
	//_lin_verbose("%s: Detect multiline comment end\n",func);
	stm = NORMAL;
      } else {
	stm = INCOMMENT_STAR;
      }
      break;
    case INCOMMENT_STAR:
      if(in[i] == '*'){
	//_lin_verbose("%s: Detect possible multiline comment end\n",func);
	stm = POSSIBLE_END;
      }
      break;
    case INCOMMENT_LINE:
      if(in[i] == '\n'){
	//_lin_verbose("%s: Detect single line  comment end\n",func);
	stm = NORMAL;
      }
      break;
    }
    
    //Copy data
    if(fcp > -1){
      for(j = fcp; j >=0; j--){
	//_lin_verbose("%s: Copy i: %d, position: %d, fcp: %d, content: %c\n",func,i,position,fcp,in[i-j]);
	(*out)[position] = in[i-j];
	position++;
      }
    } else {
      switch(fcp){
      case -2:
	(*out)[position] = replace;
	position++;
	break;
      }
    }
  }
  return CS_OK;
}


/**************************************************************************/
uint32_t _lin_ldf_doc_getuintvalue(
				   cs_parser_result_t result,
				   char *name,
				   uint8_t *error
				   )
/**************************************************************************/
{
  char func[] = "_lin_ldf_doc_getuintvalue";
  pcre *re = NULL;
  int rv = 0;
  int noresults = 4;
  int ovector[CS_REGEX_RESULTSIZE*noresults];
  char regex[CS_REGEX_BUFFER];
  uint32_t value = 0;
  char namedvalue[CS_PARSER_CHARBUF];
  char namedhex[CS_PARSER_CHARBUF];

  if(error == NULL)
    return value;

  *error = CS_OK;
  
  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER);
  if(name != NULL){
    snprintf(regex,CS_REGEX_BUFFER,"%s[\\s]+=[\\s]+(?'hex'(0x)?)(?'value'[0-9a-fA-F]+)",name);
  } else {
    snprintf(regex,CS_REGEX_BUFFER,"[\\s]+(?'hex'(0x)?)(?'value'[0-9a-fA-F]+)");
  }
  //Precompile regex
  if((re = cs_parser_precompile((const char*)regex)) == NULL){
    _lin_debug("%s: Regex precompilation failed\n",func);
    *error = CS_FAIL;
    return value;
  }

  rv = pcre_exec(re, NULL, (const char*)result.searchin,result.searchlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE*noresults);

  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _lin_verbose("%s: No match\n",func); 
      *error = CS_FAIL;
      pcre_free(re);
      return value;
      break;
    default: 
      _lin_debug("%s: Matching error %d\n", func, rv);
      pcre_free(re);
      *error = CS_FAIL;
      return value;
      break;
    }
  } else {
    if(cs_parser_getsubstring(re,result.searchin,ovector,CS_REGEX_RESULTSIZE*noresults,"value",namedvalue)){
      _lin_debug("%s: Cannot find value section\n",func);
      *error = CS_FAIL;
      pcre_free(re);
      return value;
    }
    _lin_verbose("%s: Found value: %s\n",func, namedvalue);
    if(!cs_parser_getsubstring(re,result.searchin,ovector,CS_REGEX_RESULTSIZE*noresults,"hex",namedhex)&&(strcmp(namedhex,"0x")==0)){
      // It is a hex value
      _lin_verbose("%s: Found hex value\n",func);
      if(!sscanf(namedvalue,"%X",&value)){
	_lin_debug("%s: Failed to parse substring value: %s\n",func,namedvalue);
	pcre_free(re);
	*error = CS_FAIL;
	return value;
      }
    } else {
      _lin_verbose("%s: Found dec value\n",func);
      if(!sscanf(namedvalue,"%u",&value)){
	_lin_debug("%s: Failed to parse substring value: %s\n",func,namedvalue);
	pcre_free(re);
	*error = CS_FAIL;
	return value;
      }
    }   
  }
 
  pcre_free(re);   
  return value;
}

/**************************************************************************/
int _lin_ldf_doc_getstringvalue(
				cs_parser_result_t result,
				char *name,
				char* value,
				int maxlen
				)
/**************************************************************************/
{
  char func[] = "_lin_ldf_doc_getstringvalue";
  pcre *re = NULL;
  int rv = 0;
  int noresults = 4;
  int ovector[CS_REGEX_RESULTSIZE*noresults];
  char regex[CS_REGEX_BUFFER];
  char namedvalue[CS_PARSER_CHARBUF];

  if(value == NULL)
    return CS_FAIL;

  //reset value
  memset(value,0,maxlen*sizeof(char));

  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER);
  if(name != NULL){
    snprintf(regex,CS_REGEX_BUFFER,"%s[\\s]?=[\\s]?(?'value'[^\\s]+)",name);
  } else {
    snprintf(regex,CS_REGEX_BUFFER,"[\\s]+(?'value'[^\\s]+)");
  }

  //Precompile regex
  if((re = cs_parser_precompile((const char*)regex)) == NULL){
    _lin_debug("%s: Regex precompilation failed\n",func);
    return CS_FAIL;
  }

  rv = pcre_exec(re, NULL, (const char*)result.searchin,result.searchlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE*noresults);

  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _lin_verbose("%s: No match\n",func); 
      pcre_free(re);
      return CS_FAIL;
      break;
    default: 
      _lin_debug("%s: Matching error %d\n", func, rv);
      pcre_free(re);
      return CS_FAIL;
      break;
    }
  } else {
    if(cs_parser_getsubstring(re,result.searchin,ovector,CS_REGEX_RESULTSIZE*noresults,"value",namedvalue)){
      _lin_debug("%s: Cannot find value section\n",func);
      pcre_free(re);
      return CS_FAIL;
    }
    _lin_verbose("%s: Found value: %s\n",func,namedvalue);
  }
  strncpy(value,namedvalue,maxlen);
  pcre_free(re);   
  return CS_OK;
}

/**************************************************************************/
int _lin_ldf_doc_getstringvaluequoted(
				      cs_parser_result_t result,
				      char *name,
				      char* value,
				      int maxlen
				      )
/**************************************************************************/
{
  char func[] = "_lin_ldf_doc_getstringvaluequoted";  
  pcre *re = NULL;
  int rv = 0;
  int noresults = 4;
  int ovector[CS_REGEX_RESULTSIZE*noresults];
  char regex[CS_REGEX_BUFFER];
  char namedvalue[CS_PARSER_CHARBUF];

  if(value == NULL)
    return CS_FAIL;

  //reset value
  memset(value,0,maxlen*sizeof(char));

  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER);
  if(name != NULL){
    snprintf(regex,CS_REGEX_BUFFER,"%s[\\s]?=[\\s]?\"(?'value'[^\"]+)\"",name);
  } else {
    snprintf(regex,CS_REGEX_BUFFER,"[\\s]+\"(?'value'[^\"]+)\"");
  }

  //Precompile regex
  if((re = cs_parser_precompile((const char*)regex)) == NULL){
    _lin_debug("%s: Regex precompilation failed\n",func);
    return CS_FAIL;
  }

  rv = pcre_exec(re, NULL, (const char*)result.searchin,result.searchlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE*noresults);

  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _lin_verbose("%s: No match\n",func); 
      pcre_free(re);
      return CS_FAIL;
      break;
    default: 
      _lin_debug("%s: Matching error %d\n", func,  rv);
      pcre_free(re);
      return CS_FAIL;
      break;
    }
  } else {
    if(cs_parser_getsubstring(re,result.searchin,ovector,CS_REGEX_RESULTSIZE*noresults,"value",namedvalue)){
      _lin_debug("%s: Cannot find value section\n",func);
      pcre_free(re);
      return CS_FAIL;
    }
    _lin_verbose("%s: Found value: %s\n",func,namedvalue);
  }
  strncpy(value,namedvalue,maxlen);
  pcre_free(re);   
  return CS_OK;
}

/**************************************************************************/
int _lin_ldf_doc_getdoublevaluevector(
				      cs_parser_result_t result,
				      char *name,
				      double *value,
				      int lenmax
				      )
/**************************************************************************/
{
  char func[] = "_lin_ldf_doc_getdoublevaluevector";  
  pcre *re = NULL;
  int rv = 0;
  int noresults = lenmax +2;
  int ovector[CS_REGEX_RESULTSIZE*noresults];
  char regex[CS_REGEX_BUFFER*5];
  char regexcp[CS_REGEX_BUFFER*5];
  char search[CS_PARSER_CHARBUF];
  char namedvalue[CS_PARSER_CHARBUF];
  int i;

  if(value == NULL) 
    return CS_FAIL;
  
  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER*5);
  if(name != NULL){
    snprintf(regex,CS_REGEX_BUFFER*5,"%s[\\s]?=[\\s]?",name);
  }
  for(i = 0; i < lenmax; i++){
    memcpy(regexcp,regex,sizeof(char)*CS_REGEX_BUFFER*5);
    snprintf(regex,CS_REGEX_BUFFER*5,"%s[\\s]+(?'value%d'[0-9\\.-]+)",regexcp,i);
  }

  //Precompile regex
  if((re = cs_parser_precompile((const char*)regex)) == NULL){
    _lin_debug("%s: Regex precompilation failed\n",func);
    return CS_FAIL;
  }

  rv = pcre_exec(re, NULL, (const char*)result.searchin,result.searchlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE*noresults);

  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _lin_verbose("%s: No match\n",func);
      pcre_free(re);
      return CS_FAIL;
      break;
    default: 
      _lin_debug("%s: Matching error %d\n", func,rv);
      pcre_free(re);
      return CS_FAIL;
      break;
    }
  } else {
    for(i = 0; i < lenmax; i++){
      //prepare searchstring
      memset(search,0,sizeof(char)*CS_PARSER_CHARBUF);
      snprintf(search,CS_PARSER_CHARBUF,"value%d",i);
      if(cs_parser_getsubstring(re,result.searchin,ovector,CS_REGEX_RESULTSIZE*noresults,search,namedvalue)){
	_lin_debug("%s: Cannot find value %s\n",func,search);
	pcre_free(re);
	return CS_FAIL;
      }
      if(!sscanf(namedvalue,"%lf",&value[i])){
	_lin_debug("%s: Failed to parse substring value: %s\n",func,namedvalue);
	pcre_free(re);
	return CS_FAIL;
      }
    }
  }
 
  pcre_free(re);   
  return CS_OK;
}


/**************************************************************************/
int _lin_ldf_doc_checkident(char v)
/**************************************************************************/
{
  if((v >= 'A') && (v <= 'Z')){
    return CS_OK;
  }

  if((v >= 'a') && (v <= 'z')){
    return CS_OK;
  }

  if((v >= '0') && (v <= '9')){
    return CS_OK;
  }

  if((v == '_')){
    return CS_OK;
  }
  
  return CS_FAIL;
}

/**************************************************************************/
int _lin_ldf_doc_search_section(
				cs_parser_result_t *start,
				char *sectionname,
				char **label
				)
/**************************************************************************/
{
  char func[] = "_lin_ldf_doc_search_section";  
  char *pos = start->searchin;
  char *posend = &start->searchin[start->searchlen];
  int braket = 0;
  char *labelstart = NULL;
  int labellength = 0;
  char *sectionstart = NULL;
  int sectionlength = 0;
  int flag = 0;
  int flag_label = 0;
  
  if(sectionname != NULL){
    pos = strstr(pos,sectionname);
    if(pos == NULL){
      _lin_debug("%s: Failed to find section %s\n",sectionname);
    }
  }

  for(;pos!=posend;pos++){

    if((sectionname == NULL) && !flag_label && (braket == 0)){
      //_lin_verbose("%s: Process label char %c \n",func,*pos);
      if(!_lin_ldf_doc_checkident(*pos)){
	if(labelstart == NULL){
	  //_lin_verbose("%s: Possible label start detected\n",func);
	  labelstart = pos;
	}
	labellength++;
      } else {
	if(*pos == '{'){
	  //_lin_verbose("%s: Open braket detected; end of label\n",func);
	  flag_label = 1;
	} else {
	  if((labelstart != NULL) && (*pos = ' ')){
	    //Space, do nothing
	    //_lin_verbose("%s: Space after label detected\n",func);
	  } else {
	    //Reset
	    //_lin_verbose("%s: Reset label\n",func);
	    labelstart = NULL;
	    labellength = 0;
	  }
	}
      }
    }
    
    if(*pos == '{'){
      //_lin_verbose("%s: Detect open braket\n",func);
      braket++;
    }

    if(*pos == '}'){
      //_lin_verbose("%s: Detect close braket\n",func);
      if(braket == 0){
	//Ignore
      } else {
	if(braket == 1){
	  //We got an open and closing braket and we are back at level 1
	  //I think we are ready
	  flag = 1;
	  sectionlength--;
	  break;
	}
	braket--;
      }
    }
    if(sectionlength == 1){
      sectionstart = pos;
    }
    if(braket >= 1){
      sectionlength++;
    }
    //_lin_verbose("%s: Process: %c\n",func,*pos);
  }

  if(flag == 1){
    //There was a valid section
    start->result = sectionstart;
    start->resultlen = sectionlength;
    if(sectionname == NULL){
      if(labelstart!= NULL){
	*label = (char*)calloc(labellength+1,sizeof(char));
	if(*label == NULL){
	  _lin_debug("%s: Malloc failure\n",func);
	  return CS_FAIL;
	}
	strncpy(*label,labelstart,labellength*sizeof(char));
      } else {
	_lin_debug("%s: No label found\n",func);
	return CS_FAIL;
      }
    }
    return CS_OK;
  }
  
  _lin_verbose("%s: No section found\n",func);
  return CS_FAIL;
}

/**************************************************************************/
int _lin_ldf_doc_signalencodingtypes(
				     cs_lin_ldf_t *ldf,
				     cs_parser_result_t *section
				     )
/**************************************************************************/
{
  char func[] = "_lin_ldf_doc_signalencodingtypes";

  cs_parser_result_t n;
  cs_parser_result_t definition;
  char *label = NULL;
  cs_parser_regex_t* s = NULL;
  cs_lin_enc_t* enc = NULL;
  cs_parser_resetresult(&n);
  n.searchin = section->result;
  n.searchlen = section->resultlen;

  while(1){
    _lin_verbose("%s, Search for signal encoding definition sections\n",func);
    cs_parser_printsearch(&n);
    if(_lin_ldf_doc_search_section(&n,NULL,&label)){
      _lin_verbose("%s, No section found\n",func);
      break;
    }
    //cs_parser_printresult(&n);


    //Init Encoding
    if((enc = cs_lin_ldf_get_encoding(ldf,label,LIN_ENC_DYNAMIC)) == NULL){
	_lin_debug("%s: Failed to create encoding\n",func);
	return CS_FAIL;      
    }

    
    _lin_verbose("%s: Search logical_values for signal encoding: %s\n",func,label);
    //Search for logical entries
    definition.searchin = n.result;
    definition.searchlen = n.resultlen;
    while ((s = cs_parser_regex(&definition,
				"[\\s]?(?'label'[logical_value]+)[\\s]?,[\\s]?(?'value'[0-9]+)[\\s]?,[\\s]?[\"](?'text'[a-zA-Z0-9]+)[\"][^;]*;",8)) != NULL){
      int value;
      char *text = NULL;
      if(cs_parser_regex_getInt(s,"value",&value,0)){
	_lin_debug("%s: Failed to get value\n",func);
	return CS_FAIL;
      }
      _lin_verbose("%s: Value: %d\n",func, value);
      //Note: Specification says text is optional, but we want to have it
      //TODO
      if(cs_parser_regex_getStr(s,"text",&text)){
	_lin_debug("%s: Failed to get text\n",func);
	return CS_FAIL;
      }
      _lin_verbose("%s: Text: %s\n",func, text);
      
      free(text);
      if(cs_parser_regex_next(s,&definition)){
	//Nothing to search
	_lin_verbose("%s: No next section to search\n",func);
	break;
      } else {
	//_lin_verbose("%s: Next section to search\n",func);
	//cs_parser_printsearch(&definition);
      }
      cs_parser_regex_free(s);      
    }

    _lin_verbose("%s: Search physical_values for signal encoding: %s\n",func,label);
    //Search for physical entries
    definition.searchin = n.result;
    definition.searchlen = n.resultlen;
    cs_parser_printsearch(&definition);
    while ((s = cs_parser_regex(&definition,
				"[\\s]?(?'label'[physical_value]+)[\\s]?,[\\s]?(?'rawstart'[0-9]+)[\\s]?,[\\s]?(?'rawend'[0-9]+)[\\s]?,[\\s]?(?'scale'[0-9\\.\\-]+)[\\s]?,[\\s]?(?'offset'[0-9\\.\\-]+)[\\s]?,[\\s]?[\"](?'unit'[a-zA-Z0-9]+)[\"][^;]*;",8)) != NULL){
      int value;
      double dvalue;
      char *text = NULL;
      if(cs_parser_regex_getInt(s,"rawstart",&value,0)){
	_lin_debug("%s: Failed to get rawstart\n",func);
	return CS_FAIL;
      }
      _lin_verbose("%s: RawStart: %d\n",func, value);
      if(cs_parser_regex_getInt(s,"rawend",&value,0)){
	_lin_debug("%s: Failed to get rawend\n",func);
	return CS_FAIL;
      }
      _lin_verbose("%s: RawEnd: %d\n",func, value);
      if(cs_parser_regex_getDouble(s,"scale",&dvalue,0)){
	_lin_debug("%s: Failed to get scale\n",func);
	return CS_FAIL;
      }
      _lin_verbose("%s: Scale: %f\n",func, dvalue);
      if(cs_parser_regex_getDouble(s,"offset",&dvalue,0)){
	_lin_debug("%s: Failed to get offset\n",func);
	return CS_FAIL;
      }
      _lin_verbose("%s: Offset: %f\n",func, dvalue);
      //Note: Specification says text is optional, but we want to have it
      //TODO
      if(cs_parser_regex_getStr(s,"unit",&text)){
	_lin_debug("%s: Failed to get unit\n",func);
	return CS_FAIL;
      }
      _lin_verbose("%s: Unit: %s\n",func, text);
      free(text);
      if(cs_parser_regex_next(s,&definition)){
	//Nothing to search
	_lin_verbose("%s: No next section to search\n",func);
	break;
      } else {
	//_lin_verbose("%s: Next section to search\n",func);
	//cs_parser_printsearch(&definition);
      }
      cs_parser_regex_free(s);      
    }

    //Prepare for next definition section
    n.searchin = &n.result[n.resultlen];
    n.searchlen = n.searchlen - n.resultlen;
  }
  

  
  return CS_OK;
}


/**************************************************************************/
cs_lin_ldf_doc_t* cs_lin_ldf_doc_init()

/**************************************************************************/
{
  char func[] = "cs_lin_ldf_doc_init";
  cs_lin_ldf_doc_t* data = NULL;
    
  data = (cs_lin_ldf_doc_t*)calloc(1,sizeof(cs_lin_ldf_doc_t));
  if(data == NULL){
    _lin_debug("%s: Malloc failure\n",func);
    return NULL;
  }

  return data;
}


/**************************************************************************/
int cs_lin_ldf_doc_open(
			cs_lin_ldf_doc_t *data,
			char *filename
			)
/**************************************************************************/
{
  char func[] = "cs_lin_ldf_doc_open";
  int buflen = 100;
  char buffer[buflen];
  double protocol_version;
  double language_version;
  double baud;
  cs_parser_result_t rheader;
  cs_parser_result_t rnodes;
  cs_parser_result_t rsignals;
  cs_parser_result_t rframes;
  cs_parser_result_t reventtriggeredframes;
  cs_parser_result_t rnodeattributes;
  cs_parser_result_t rscheduletables;
  cs_parser_result_t rsignalencodingtypes;
  cs_parser_result_t rsignalrepresentation;
  
  if((data == NULL) || (filename == NULL)){
    _lin_debug("%s: Parameter failure\n",func);
    return CS_FAIL;
  }

  if(data->map != NULL){
    _lin_debug("%s: There is already an opened file\n",func);
    return CS_FAIL;
  }

  _lin_verbose("%s: Try to open LDF file: %s\n",func,filename);
  
  data->filename = filename;

  if(cs_parser_open(data->filename,&data->fs,&data->filelen,&data->map)){
    _lin_debug("%s: Failed to open file: %s \n",func,data->filename);
    return CS_FAIL;
  }

  //Cleanup comments
  if(_lin_ldf_doc_cleanup((char*)data->map,&data->buffer,data->filelen)){
    _lin_debug("%s: Failed to cleanup file\n",func);
    return CS_FAIL;
  }

  //Check for file header
  if(strstr(data->buffer,"LIN_description_file") == NULL){
    _lin_debug("%s: File %s is not a valid LDF file\n",func,data->filename);
    return CS_FAIL;
  }

  _lin_verbose("%s: LDF has a valid header\n",func);

  //Section predefinition, these are global sections, so all search sections defined equal
  rsignals.searchin = data->buffer;
  rsignals.searchlen = data->filelen;
  rheader = rsignals;
  rnodes = rsignals;
  rframes = rsignals;
  reventtriggeredframes = rsignals;
  rnodeattributes = rsignals;
  rscheduletables = rsignals;
  rsignalencodingtypes = rsignals;
  rsignalrepresentation = rsignals;

  if(_lin_ldf_doc_search_section(&rnodes,"Nodes",NULL)){
    _lin_debug("%s: Failed to find nodes section\n",func);
    return CS_FAIL;
  }

  cs_parser_printresult(&rnodes);
  
  if(_lin_ldf_doc_search_section(&rsignals,"Signals",NULL)){
    _lin_debug("%s: Failed to find signals section\n",func);
    return CS_FAIL;
  }

  cs_parser_printresult(&rsignals);
  
  if(_lin_ldf_doc_search_section(&rframes,"Frames",NULL)){
    _lin_debug("%s: Failed to find frames section\n",func);
    return CS_FAIL;
  }

  cs_parser_printresult(&rframes);
  
  if(_lin_ldf_doc_search_section(&reventtriggeredframes,"Event_triggered_frames",NULL)){
    _lin_debug("%s: Failed to find event triggered frames section\n",func);
    return CS_FAIL;
  }

  cs_parser_printresult(&reventtriggeredframes);
  
  if(_lin_ldf_doc_search_section(&rnodeattributes,"Node_attributes",NULL)){
    _lin_debug("%s: Failed to find node attributes section\n",func);
    return CS_FAIL;
  }

  cs_parser_printresult(&rnodeattributes);
  
  if(_lin_ldf_doc_search_section(&rscheduletables,"Schedule_tables",NULL)){
    _lin_debug("%s: Failed to find schedule tables section\n",func);
    return CS_FAIL;
  }

  cs_parser_printresult(&rscheduletables);

  if(_lin_ldf_doc_search_section(&rsignalencodingtypes,"Signal_encoding_types",NULL)){
    _lin_debug("%s: Failed to find signal encoding types section\n",func);
    return CS_FAIL;
  }
  cs_parser_printresult(&rsignalencodingtypes);
  
  if(_lin_ldf_doc_search_section(&rsignalrepresentation,"Signal_representation",NULL)){
    _lin_debug("%s: Failed to find signal representation section\n",func);
    return CS_FAIL;
  }
  
  cs_parser_printresult(&rsignalrepresentation);
  
  _lin_verbose("%s: Find global sections, continue\n",func);

  if(_lin_ldf_doc_getstringvaluequoted(rheader,"LIN_protocol_version",buffer,buflen)){
    _lin_debug("%s: Failed to get protocol version\n",func);
    return CS_FAIL;
  }
  if(!sscanf(buffer,"%lf",&protocol_version)){
    _lin_debug("%s: Failed to convert protocol version value: %s\n",func,buffer);
    return CS_FAIL;
  }
  _lin_verbose("%s: Protocol version: %f\n",func,protocol_version);
  
  
  if(_lin_ldf_doc_getstringvaluequoted(rheader,"LIN_language_version",buffer,buflen)){
    _lin_debug("%s: Failed to get language version\n",func);
    return CS_FAIL;
  }
  if(!sscanf(buffer,"%lf",&language_version)){
    _lin_debug("%s: Failed to convert language version value: %s\n",func,buffer);
    return CS_FAIL;
  }

  _lin_verbose("%s: Language version: %f\n",func,protocol_version);
  

  if(_lin_ldf_doc_getdoublevaluevector(rheader,"LIN_speed",&baud,1)){
    _lin_debug("%s: Failed to get LIN speed\n",func);
    return CS_FAIL;
  }

  _lin_verbose("%s: LIN speed: %f\n",func,baud);


  //Prepare storage structure
  if((data->ldf = cs_lin_ldf_init()) == NULL){
    return CS_FAIL;
  }

  if(cs_lin_ldf_set_general(data->ldf,protocol_version,
			    language_version,(uint32_t)(baud*1000.0))){
    return CS_FAIL;
  }

  if(_lin_ldf_doc_signalencodingtypes(data->ldf,&rsignalencodingtypes)){
    _lin_debug("%s: Failed to scan for Encoding type entry\n",func);
    return CS_FAIL;
  }
  

  /*cs_parser_regex_t* s =  cs_parser_regex(rsignalencodingtypes.result,rsignalencodingtypes.resultlen,
					  ".*(?'label'[a-zA-Z_0-9]+)[\\s]?{(?'section'[^}]+)}.*",4);
  if(s == NULL){
    _lin_debug("%s: Failed to scan for Encoding type entry\n",func);
    return CS_FAIL;
  }
  char *label = NULL;
  if(cs_parser_regex_getStr(s,"label",&label)){
    _lin_debug("%s: Failed to get label\n",func);
    return CS_FAIL;
  }
  _lin_verbose("%s: Label: %s\n",func,label);
  */

  
  return CS_OK;
}

/**************************************************************************/
int cs_lin_ldf_doc_close(
			 cs_lin_ldf_doc_t *data
			 )
/**************************************************************************/
{
  char func[] = "cs_lin_ldf_doc_close";
  
  if(data == NULL){
    _lin_debug("%s: Parameter failure\n",func);
    return CS_FAIL;
  }


  if(cs_parser_close(data->fs,data->map,data->filelen)){
    _lin_debug("%s: Parser close failure\n",func);
    return CS_FAIL;
  }

  if(data->buffer!= NULL){
    free(data->buffer);
  }
  
  return CS_OK;
}


/********************************************************************/
void cs_lin_ldf_doc_free(
			 cs_lin_ldf_doc_t *data
			 )
/********************************************************************/
{
  
  if(data == NULL)
    return;
  
  cs_lin_ldf_free(data->ldf);
  
  free(data);

}
