/*  
    Copyright (C) 2017 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

/**
 * @file cs_lin_private.h
 * @author Andreas Stewering-Bone
 * @date 9 Oct 2017
 * @brief File containing lin bus private header.
 *
 * File containing lin bus private header.
*/
#ifndef CS_LIN_PRIVATE_H
#define CS_LIN_PRIVATE_H

#include <inttypes.h>

#include "cs_lin.h"

#include "cs_lists.h"



/** \enum m_state_t
    Master state flag
*/
typedef enum{
  M_STATE_IDLE = 0,              /**< Master idle state*/
  M_STATE_SEND,                  /**< Master is waiting for sending finished*/
  M_STATE_RECV,                  /**< Master is waiting for slave frame*/
  M_STATE_ERROR                  /**< Master general error*/        
}m_state_t;

/** \enum s_state_t
    Slave state flag
*/
typedef enum{
  S_STATE_IDLE = 0,              /**< Slave idle state*/
  S_STATE_SEND,                  /**< Slave is waiting for sending finished*/
  S_STATE_REQ,                   /**< Slave frame was requested*/
  S_STATE_ERROR                  /**< Slave general error*/        
}s_state_t;

/** \struct cs_lin
    Central stack structure, private definiton
*/
struct cs_lin{
  cs_lin_node_t *master;                  /**< Master node pointer*/
  uint8_t listen_only;                    /**< If Slave and flag == 1 only listening to all frames (Listener)*/
  cs_lists_t l_nodes;                     /**< Node pointer list*/
  cs_lists_t l_frames;                    /**< Frame pointer list*/
  cs_lin_ldf_t *ldf;                      /**< LDF storage pointer*/
  cs_lin_schedule_table_t *defaulttable;  /**< Default scheduler table*/
  cs_lin_schedule_table_t *table;         /**< Actual running scheduler table*/
  struct cs_lists_entry *sched_entry_first;/**< First entry of actual scheduler table*/
  struct cs_lists_entry *sched_entry_last; /**< Last entry of actual scheduler table, no valid data*/
  struct cs_lists_entry *sched_entry_act; /**< Actual entry of actual scheduler table*/
  uint8_t lerror;                         /**< Last error flag*/
  void *cb_ctx;                           /**< Callback context pointer*/
  void (*cs_lin_cb_prepfilt)(void *ctx, uint32_t *id, int len);  /**< Prepare recv filter*/
  void (*cs_lin_cb_send)(void *ctx, uint32_t id, uint8_t *data, int len, int enhanced); /**< Master / Slave send callback*/
  void (*cs_lin_cb_recv)(void *ctx, uint32_t id, int enhanced); /**< Master recv request callback*/
  void (*cs_lin_cb_prepcache)(
			      void *ctx,        /**< Context pointer*/
			      uint32_t id,      /**< LIN id*/
			      uint8_t *data,    /**< Data buffer*/
			      int len,          /**< Len data buffer*/
			      int enhanced      /**< 0=> Classic checksum, 1=>Enhanced checksum*/
			      );  /**< Slave prepare callback*/
  m_state_t m_state;                      /**< Actual master state*/
  s_state_t s_state;                      /**< Actual slave state*/
  cs_lin_frame_t *s_frame_req;            /**< Slave frame was requested by master*/
  uint8_t send_finished_call;             /**< 0=> No send finished call needed, 1=> send finished call needed*/
  double act_ts;                          /**< Actual timestamp ins s*/
  double sched_ts;                        /**< Scheduler delay timestamp*/
  double timeout_ts;                      /**< Receive timestamp in s*/
};


/** \struct cs_lin_ldf_doc
    Central LDF parser stack structure, private definiton
*/
struct cs_lin_ldf_doc{
  FILE *fs;                               /**< FILE stream*/
  char *filename;                         /**< Filename to parse*/
  long filelen;                           /**< File length*/
  void *map;                              /**< File map pointer*/
  char *buffer;                           /**< Cleaned up raw file content*/        
  cs_lin_ldf_t *ldf;                      /**< LDF storage pointer*/
};

#ifdef __cplusplus
extern "C" {
#endif
  void _lin_info(const char *fmt, ...);
  void _lin_debug(const char *fmt, ...);
  void _lin_verbose(const char *fmt, ...);
  void _lin_debug_buffer(
			 uint8_t *data,
			 int len
			 );
  void _lin_debug_conv(
		       cs_lin_raw_t *data,
		       int len
		       );
  void _lin_verbose_buffer(
			   uint8_t *data,
			   int len
			   );
  void _lin_verbose_conv(
			 cs_lin_raw_t *data,
			 int len
			 );
  
#ifdef __cplusplus
}
#endif

#endif
