/*  
    Simple linked list implementation
    Copyright (C) 2015 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cs_defines.h"
#include "cs_lists.h"

/**
   Dummy copy function
*/
/**************************************************************************/
void *_cs_lists_dummy_cpy(
			  void *ext
			  )
/**************************************************************************/
{
  return ext;
}

/**
   Dummy del function
*/
/**************************************************************************/
void _cs_lists_dummy_del(
			 void *intern
			 )
/**************************************************************************/
{
  return;
}


/**
   Iterate on list, feedback found entry, care of first entry
*/
/**************************************************************************/
void _cs_lists_iter(
		    cs_lists_t *list,
		    struct cs_lists_entry **search,
		    int iter
		    )
/**************************************************************************/
{
  int i = 0;


  *search = list->first->next;

  while(*search != list->end){
    if(i == iter){
      break;
    }
    i++;
    *search = (*search)->next;
  }

  if((i != iter) || (*search == list->end)){
    *search = NULL;
  }
  
}


/**
   Delete entry
*/
/**************************************************************************/
int _cs_lists_del(
		  cs_lists_t *list,
		  struct cs_lists_entry *search
		  )
/**************************************************************************/
{

  //delete Data
  list->delfc(search->data);
  search->data = NULL;

  //Set next of element before to next element behind
  search->prev->next = search->next;
  //Set prev of element behind to element before
  search->next->prev = search->prev;

  //List is valid again, delete entry
  free(search);

  //Decrement size
  list->list_size--;

  if(list->list_size < 0){
    list->list_size = 0;
  }
  
  return CS_OK;
}

/**************************************************************************/
int cs_lists_init(
		  cs_lists_t *list,
		  void* (*copyf)(void *ext),
		  void (*delf)(void *intern) 
		  )
/**************************************************************************/
{
  if(list == NULL)
    return CS_FAIL;

  //Reset structure
  memset(list,0,sizeof(cs_lists_t));

  //Malloc dummy end
  list->first = (struct cs_lists_entry*)malloc(sizeof(struct cs_lists_entry));

  list->end = (struct cs_lists_entry*)malloc(sizeof(struct cs_lists_entry));

  if((list->first == NULL)||(list->end == NULL))
    return CS_FAIL;

  list->first->prev = list->first;
  list->first->next = list->end;
  list->end->next = list->end;
  list->end->prev = list->first;
  //Set previous of first element ti itself
  list->first->prev = list->first;

  //Check for data functions
  if((copyf == NULL)&&(delf !=NULL)){
    return CS_FAIL;
  }

  if((copyf != NULL)&&(delf ==NULL)){
    return CS_FAIL;
  }

  if(delf != NULL){
    list->delfc = delf;
    list->copyfc = copyf;
  } else {
    list->delfc = _cs_lists_dummy_del;
    list->copyfc = _cs_lists_dummy_cpy;
  }
  list->list_size = 0;
  
  return CS_OK;
}

/**************************************************************************/
int cs_lists_size(
		  cs_lists_t *list
		  )
/**************************************************************************/
{
  if(list == 0)
    return 0;

  return list->list_size;
}

/**************************************************************************/
int cs_lists_destroy(
		     cs_lists_t *list
		     )
/**************************************************************************/
{
  struct cs_lists_entry *entry = NULL;

  if(list == NULL)
    return CS_FAIL;

  if((list->first == NULL) || (list->end == NULL)){
    return CS_FAIL;
  }
  
  entry = list->first->next;
  while(entry != list->end){
    //delete data, do it the fast way
    list->delfc(entry->data);
    entry = entry->next;
    free(entry->prev);
  }
  free(list->first);
  free(list->end);

  //Reset structure
  memset(list,0,sizeof(cs_lists_t));
  return CS_OK;
}

/**************************************************************************/
int cs_lists_reset(
		   cs_lists_t *list
		   )
/**************************************************************************/
{
  struct cs_lists_entry *entry = NULL;
  struct cs_lists_entry *next = NULL;

  if(list == NULL)
    return CS_FAIL;

  if((list->first == NULL) || (list->end == NULL)){
    return CS_FAIL;
  }
  
  entry = list->first->next;
  while(entry != list->end){
    next = entry->next;
    _cs_lists_del(list,entry);
    entry = next;
  }
  list->list_size = 0;
  
  return CS_OK;
}

/**************************************************************************/
int cs_lists_insert(
		    cs_lists_t *list,
		    void *data
		    )
/**************************************************************************/
{
  struct cs_lists_entry *entry = NULL;
  struct cs_lists_entry *search = NULL;

  if(list == NULL)
    return CS_FAIL;

  if(data == NULL)
    return CS_FAIL;

  //Malloc new entry
  entry = (struct cs_lists_entry*)malloc(sizeof(struct cs_lists_entry));

  if(entry == NULL)
    return CS_FAIL;

  memset(entry,0,sizeof(struct cs_lists_entry));

  //set data
  entry->data = (list->copyfc)(data);

  if(entry->data == NULL){
    free(entry);
    return CS_FAIL;
  }

  //search last entry
  search = list->end->prev;

  //set next
  search->next = entry;
  //set previous
  entry->prev = search;

  //set end
  entry->next = list->end;
  list->end->prev = entry;

  //Increment size
  list->list_size++;

  return CS_OK;
}



/**************************************************************************/
int cs_lists_delete(
		    cs_lists_t *list,
		    void *data
		    )
/**************************************************************************/
{
  struct cs_lists_entry *search = NULL;

  if(list == NULL)
    return CS_FAIL;

  if(data == NULL)
    return CS_FAIL;

  //set end data to search adress, sentinell method
  list->end->data = data;

  //set start
  search = list->first->next;

  while(search->data != data){
    search = search->next;
  }

  //Are we sentinell
  if(search == list->end){
    //not found
    return CS_FAIL;
  }
  return _cs_lists_del(list,search);
}

/**************************************************************************/
int cs_lists_delete_iter(
			 cs_lists_t *list,
			 int iter
			 )
/**************************************************************************/
{
  struct cs_lists_entry *search = NULL;

  if(list == NULL)
    return CS_FAIL;

  if(iter < 0)
    return CS_FAIL;

  //get element
  _cs_lists_iter(list,&search,iter);

  if(search == NULL){
    return CS_FAIL;
  }

  if(search == list->end){
    //not found
    return CS_FAIL;
  }

  return _cs_lists_del(list,search);
}

/**************************************************************************/
void * cs_lists_get_iter(
			 cs_lists_t *list,
			 int iter
			 )
/**************************************************************************/
{
  struct cs_lists_entry *search = NULL;

  if(list == NULL)
    return NULL;

  if(iter < 0)
    return NULL;

  //get element
  _cs_lists_iter(list,&search,iter);

  if(search == NULL){
    return NULL;
  }

  return search->data;
}

/**************************************************************************/
struct cs_lists_entry * cs_lists_get_first(
					   cs_lists_t *list
					   )
/**************************************************************************/
{

  if(list == NULL)
    return NULL;

  return list->first->next;
}

/**************************************************************************/
struct cs_lists_entry * cs_lists_get_next(
					  struct cs_lists_entry *entry
					  )
/**************************************************************************/
{

  if(entry == NULL)
    return NULL;

  return entry->next;
}

/**************************************************************************/
struct cs_lists_entry * cs_lists_get_last(
					  cs_lists_t *list
					  )
/**************************************************************************/
{

  if(list == NULL)
    return NULL;

  return list->end;
}

/**************************************************************************/
void* cs_lists_get_data(
			struct cs_lists_entry *entry
			)
/**************************************************************************/
{

  if(entry == NULL)
    return NULL;

  return entry->data;
}
