/*  message.c --  encode and decode CAN messages signals
    Copyright (C) 2013 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "config.h"
#include "cs_message_private.h"
#include "cs_signal_private.h"
#include "cs_message.h"
#include "cs_debug_private.h"


#define _message_error(...) (cs_error(CS_DEBUG_MESSAGE,__func__,__LINE__,__VA_ARGS__))
#define _message_warn(...) (cs_warn(CS_DEBUG_MESSAGE,__func__,__LINE__,__VA_ARGS__))
#define _message_info(...) (cs_info(CS_DEBUG_MESSAGE,__func__,__LINE__,__VA_ARGS__))
#define _message_debug(...) (cs_debug(CS_DEBUG_MESSAGE,__func__,__LINE__,__VA_ARGS__))
#define _message_verbose(...) (cs_verbose(CS_DEBUG_MESSAGE,__func__,__LINE__,__VA_ARGS__))
#define _message_error_buffer(DATA,LEN) (cs_error_buffer(CS_DEBUG_MESSAGE,__func__,__LINE__,DATA,LEN))
#define _message_warn_buffer(DATA,LEN) (cs_warn_buffer(CS_DEBUG_MESSAGE,__func__,__LINE__,DATA,LEN))
#define _message_info_buffer(DATA,LEN) (cs_info_buffer(CS_DEBUG_MESSAGE,__func__,__LINE__,DATA,LEN))
#define _message_debug_buffer(DATA,LEN) (cs_debug_buffer(CS_DEBUG_MESSAGE,__func__,__LINE__,DATA,LEN))
#define _message_verbose_buffer(DATA,LEN) (cs_verbose_buffer(CS_DEBUG_MESSAGE,__func__,__LINE__,DATA,LEN))
#define _message_verbose_binbuffer(DATA,LEN) (cs_verbose_binbuffer(CS_DEBUG_MESSAGE,__func__,__LINE__,DATA,LEN))


/********************************************************************/
cs_debug_level_t cs_message_debuglevel(
				       cs_debug_level_t level
				       )
/********************************************************************/
{
  return cs_debuglevel(CS_DEBUG_MESSAGE,level);
}

/********************************************************************/
void cs_message_debug_setcallback(
    void *ctx,
    void (*callb)(void *ctx,
		  cs_debug_sub_t sub,
		  cs_debug_level_t dl,
		  const char *func,
		  const int line,
		  const char *fmt,
		  va_list ap)
    )
/********************************************************************/
{
  cs_debug_setcallback(CS_DEBUG_MESSAGE,ctx,callb);
}

/**************************************************************************/
void *_cs_message_cp_signal(void *ext)
/**************************************************************************/
{
  cs_signal_t *t = cs_signal_cp((cs_signal_t*)ext);

  return (void*)t;
}

/**************************************************************************/
void _cs_message_del_signal(void *intern)
/**************************************************************************/
{
  cs_signal_free((cs_signal_t*)intern);
}

/**************************************************************************/
cs_message_t *cs_message_init()
/**************************************************************************/
{
  cs_message_t * m = NULL;

  m = (cs_message_t*)calloc(sizeof(cs_message_t),1);
  
  if(m == NULL){
    _message_error("Malloc failure\n");
    return NULL;
  }
  cs_lists_init(&m->listsignal,_cs_message_cp_signal,_cs_message_del_signal);
  cs_message_reset(m);

  return m;
}

/**************************************************************************/
void cs_message_free(cs_message_t *m)
/**************************************************************************/
{

  if(m == NULL){
    return;
  }
  cs_lists_destroy(&m->listsignal);
  free(m);
  m = NULL;
  return;
}

/**************************************************************************/
cs_message_t *cs_message_cp(cs_message_t *ext)
/**************************************************************************/
{
  cs_message_t *t = NULL;
  int numsignals = 0;
  int i;

  if(ext == NULL){
    _message_error("Parameter failure\n");
    return NULL;
  }
  
  t = cs_message_init();
  
  if(t == NULL){
    _message_error("Message init failure\n");
    return NULL;
  }
  
  if(cs_message_reset(t)||
     cs_message_set(t,ext->id,ext->dlc,ext->name,ext->producer,ext->interval)||
     cs_message_set_offset(t,ext->interval_offset)||
     cs_message_set_startoffset(t,ext->interval_startoffset)||
     cs_message_set_sendtype(t,ext->send_type)
     ){
    _message_error("Configure failure\n");
    cs_message_free(t);
    return NULL;
  }
  numsignals = cs_message_get_no_signals(ext);

  //Handle signals
  if(numsignals < 0){
    _message_error("Failed to get number of signals\n");
    cs_message_free(t);
    return NULL;
  }

  for(i = 0; i < numsignals; i++){
     cs_signal_t *ds = NULL;
     ds = cs_message_get_signal_iter(ext,i);
     if(ds == NULL){
       _message_error("Failed to get signal\n");
       cs_message_free(t);
       return NULL;
     }
     if(cs_message_add_signal(t,ds)){
       _message_error("Failed to add signal\n");
       cs_message_free(t);
       return NULL;
     }
  }
  
  return t;
}

/**************************************************************************/
int cs_message_set_interval(
			    cs_message_t *message,
			    uint32_t interval
			    )
/**************************************************************************/
{
  if(message == NULL)
    return CS_FAIL;
  
  message->interval = interval;

  return CS_OK;
}

/**************************************************************************/
uint32_t cs_message_get_interval(
				 cs_message_t *message
				 )
/**************************************************************************/
{
  if(message == NULL)
    return 0;
  
  return message->interval;
}

/**************************************************************************/
int cs_message_set_offset(
			  cs_message_t *message,
			  uint32_t offset
			  )
/**************************************************************************/
{
  if(message == NULL)
    return CS_FAIL;
  
  message->interval_offset = offset;

  return CS_OK;
}

/**************************************************************************/
uint32_t cs_message_get_offset(
			       cs_message_t *message
			       )
/**************************************************************************/
{
  if(message == NULL)
    return 0;
  
  return message->interval_offset;
}

/**************************************************************************/
int cs_message_set_startoffset(
			       cs_message_t *message,
			       uint32_t offset
			       )
/**************************************************************************/
{
    if(message == NULL)
        return CS_FAIL;
    
    message->interval_startoffset = offset;
    
    return CS_OK;
}

/**************************************************************************/
uint32_t cs_message_get_startoffset(
				    cs_message_t *message
				    )
/**************************************************************************/
{
  if(message == NULL)
    return 0;
  
  return message->interval_startoffset;
}

/**************************************************************************/
int cs_message_set_sendtype(
			    cs_message_t *message,
			    cs_message_send_type_t send_type
			    )
/**************************************************************************/
{
  if(message == NULL)
    return CS_FAIL;
  
  message->send_type = send_type;
  
  return CS_OK;
}

/**************************************************************************/
cs_message_send_type_t cs_message_get_sendtype(
					       cs_message_t *message
					       )
/**************************************************************************/
{
  if(message == NULL)
    return MESSAGE_STYPE_NOTACTIVE;

  return message->send_type;
}

/**************************************************************************/
int cs_message_get_no_signals(cs_message_t *message)
/**************************************************************************/
{
  if(message == NULL)
    return -1;
  
  return cs_lists_size(&message->listsignal);
}

/**************************************************************************/
int cs_message_set_id(
		      cs_message_t *message,
		      uint32_t id
		      )
/**************************************************************************/
{
  if(message == NULL){
    return CS_FAIL;
  }

  message->id = id;
  
  return CS_OK;
}

/**************************************************************************/
uint32_t cs_message_get_id(
			   cs_message_t *message
			   )
/**************************************************************************/
{
  if(message == NULL){
    return 0;
  }
  return message->id;
}

/**************************************************************************/
int cs_message_get_dlc(
		       cs_message_t *message
		       )
/**************************************************************************/
{
  if(message == NULL){
    return -1;
  }
  return (int)message->dlc;
}

/**************************************************************************/
int cs_message_set_dlc(
		      cs_message_t *message,
		      uint8_t dlc
		      )
/**************************************************************************/
{
  if(message == NULL){
    return CS_FAIL;
  }

  if(dlc > CS_DATA){
    return CS_FAIL;
  }
  
  message->dlc = dlc;
  
  return CS_OK;
}

/**************************************************************************/
char *cs_message_get_name(
			  cs_message_t *message  /**< Message structure*/
			  )
/**************************************************************************/
{
  if(message == NULL){
    return NULL;
  }
  return message->name;
}

/**************************************************************************/
int cs_message_set_name(
			cs_message_t *message,
			char * name
			)
/**************************************************************************/
{
  if((message == NULL) || (name == NULL)){
    return CS_FAIL;
  }
  
  strncpy(message->name,name,CS_CHARBUF);
  return CS_OK;
}

/**************************************************************************/
char *cs_message_get_producer(
			      cs_message_t *message  /**< Message structure*/
			      )
/**************************************************************************/
{
  if(message == NULL){
    return NULL;
  }
  return message->producer;
}

/**************************************************************************/
int cs_message_set_producer(
			    cs_message_t *message,
			    char * name
			    )
/**************************************************************************/
{
  if((message == NULL) || (name == NULL)){
    return CS_FAIL;
  }
  
  strncpy(message->producer,name,CS_CHARBUF);
  return CS_OK;
}

/**************************************************************************/
cs_signal_t* cs_message_get_signal_iter(cs_message_t *message, int iter)
/**************************************************************************/
{
  if(message == NULL)
    return NULL;
  
  return (cs_signal_t*)cs_lists_get_iter(&message->listsignal,iter);
}

/**************************************************************************/
int cs_message_getmessage(cs_message_t **message,
			  int lt_message,
			  uint32_t id,
			  cs_message_t **rmessage)
/**************************************************************************/
{
  int i;

  if(message == NULL)
    return CS_FAIL;

  if(rmessage == NULL)
    return CS_FAIL;

  *rmessage = NULL;
  for(i = 0; i < lt_message; i++){
    if(id == message[i]->id){
      *rmessage = message[i];
      return CS_OK;
    }
  }  
  return CS_FAIL;
}

/**************************************************************************/
cs_signal_t * cs_message_get_signal_byname(
					   cs_message_t *message,
					   char *slname
					   )
/**************************************************************************/
{
  struct cs_lists_entry *entry = NULL;
  cs_signal_t *s = NULL;

  if(message == NULL)
    return NULL;
  
  if(slname == NULL)
    return NULL;

  for(entry = cs_lists_get_first(&message->listsignal);
      entry != cs_lists_get_last(&message->listsignal);
      entry = cs_lists_get_next(entry)){
      s = (cs_signal_t*)cs_lists_get_data(entry);
    if(strcmp(s->name,slname) == 0){
      return s;
    }
  }
  
  return NULL;
}

/**************************************************************************/
int cs_message_reset(cs_message_t *message)
/**************************************************************************/
{
  if(message == NULL)
    return CS_FAIL;
  
  message->id = 0;
  message->dlc = 0;
  message->direction = MESSAGE_DIR_RECEIVE;
  message->mtype = MESSAGE_TYPE_ASSEMBLED;
  message->send_type = MESSAGE_STYPE_NOTACTIVE;
  message->ptype = MESSAGE_PTYPE_CAN; //Default CAN-2.0
  message->interval = 0;
  message->interval_offset = 0;
  message->interval_startoffset = 0;
  message->timeout = 10e6;
  message->timeoutflag = NULL;
  message->externts = NULL;
  message->externtvh = NULL;
  message->externbuffer = NULL;
  message->msg_counter = 0;
  message->rawbuffer = NULL;
  cs_lists_destroy(&message->listsignal);
  cs_lists_init(&message->listsignal,_cs_message_cp_signal,_cs_message_del_signal);
  return CS_OK;
}

/**************************************************************************/
int cs_message_set(cs_message_t *message,
		   uint32_t id,
		   uint8_t dlc,
		   char* name,
		   char* producer,
		   uint32_t interval)
/**************************************************************************/
{
  if(message == NULL)
    return CS_FAIL;

  if(cs_message_set_id(message,id) ||
     cs_message_set_dlc(message,dlc)||
     cs_message_set_name(message,name)||
     cs_message_set_producer(message,producer)||
     cs_message_set_interval(message,interval)){
    return CS_FAIL;
  }
  return CS_OK;
}

/**************************************************************************/
int cs_message_set_timeout(
			   cs_message_t *message,
			   uint32_t timeout // in nues
			   )
/**************************************************************************/
{
  if(message == NULL)
    return CS_FAIL;

  message->timeout = timeout;

  return CS_OK;  
}

/**************************************************************************/
int cs_message_set_timeoutflag(
			       cs_message_t *message,
			       uint8_t *flag
			       )
/**************************************************************************/
{
  if(message == NULL)
    return CS_FAIL;

  message->timeoutflag = flag;

  return CS_OK;  
}

/**************************************************************************/
uint32_t cs_message_get_timeout(
				cs_message_t *message
				)
/**************************************************************************/
{
  if(message == NULL)
    return 0;

  return message->timeout;
}

/**************************************************************************/
int cs_message_add_signal(
			  cs_message_t *message,
			  cs_signal_t *signal
			  )
/**************************************************************************/
{
  cs_signal_t *lsignal = NULL;
  if(message == NULL)
    return CS_FAIL;

  if(signal == NULL)
    return CS_FAIL;

  lsignal = cs_signal_cp(signal);

  if(lsignal == NULL)
    return CS_FAIL;
  
  cs_lists_insert(&message->listsignal,(void*)lsignal);
  
  return CS_OK;
}

/**************************************************************************/
int cs_message_set_crc_cb(
			  cs_message_t *message,
			  char *slname,
			  int (*crc)(void *ctx,uint8_t *buffer, int len, double *value),
			  void *ctx,
			  uint8_t *crcflag
			  )
/**************************************************************************/
{
  cs_signal_t *signal = NULL;

  if(message == NULL)
    return CS_FAIL;

  if(slname == NULL)
    return CS_FAIL;

  if(crc == NULL)
    return CS_FAIL;

  if((signal = cs_message_get_signal_byname(message,slname)) == NULL){
    return CS_FAIL;
  }

  signal->mtype = SIGNAL_MUX_CRC;
  signal->crc = crc;
  signal->crc_ctx = ctx;
  signal->crc_flag = crcflag;
  return CS_OK;
}

/**************************************************************************/
int cs_message_set_signal_default(
				  cs_message_t *message,
				  char *slname,
				  double default_value
				  )
/**************************************************************************/
{ 
  cs_signal_t *signal = NULL;
  
  if(message == NULL)
    return CS_FAIL;
  
  if(slname == NULL)
    return CS_FAIL;

  if((signal = cs_message_get_signal_byname(message,slname)) == NULL){
    return CS_FAIL;
  }

  signal->mtype = SIGNAL_MUX_STATIC;
  signal->startvalue = default_value;
  
  return CS_OK;
}

/**************************************************************************/
int cs_message_set_signal_defaultlabel(
				       cs_message_t *message,
				       char *slname,
				       char *label
				       )
/**************************************************************************/
{ 
  cs_signal_t *signal = NULL;
  double value;
  
  if(message == NULL)
    return CS_FAIL;
  
  if(slname == NULL)
    return CS_FAIL;

  if(label == NULL)
    return CS_FAIL;

  if((signal = cs_message_get_signal_byname(message,slname)) == NULL){
    return CS_FAIL;
  }

  if(cs_signal_runtime_getlabelvalue(signal,&value,label)){
    return CS_FAIL;
  }

  signal->mtype = SIGNAL_MUX_STATIC;
  signal->startvalue = value;
  
  return CS_OK;
}


/**************************************************************************/
int cs_message_set_signal_mux_type(
				   cs_message_t *message,
				   char *slname,
				   cs_signal_mux_t mtype
				   )
/**************************************************************************/
{
  cs_signal_t *signal = NULL;
  
  if(message == NULL)
    return CS_FAIL;
  
  if(slname == NULL)
    return CS_FAIL;
  
  if((signal = cs_message_get_signal_byname(message,slname)) == NULL){
    return CS_FAIL;
  }

  return cs_signal_set_mux_type(signal,mtype);
}

/**************************************************************************/
int cs_message_set_signal_rvalue(
				 cs_message_t *message,
				 char *slname,
				 double *rvalue,
				 char *label,
				 uint8_t *fault
				 )
/**************************************************************************/
{
  cs_signal_t *signal = NULL;
  
  if(message == NULL)
    return CS_FAIL;
  
  if(slname == NULL)
    return CS_FAIL;

  if(rvalue== NULL)
    return CS_FAIL;

  if((signal = cs_message_get_signal_byname(message,slname)) == NULL){
    return CS_FAIL;
  }

  return cs_signal_set_rvalue(signal,rvalue,label,fault);
}

/**************************************************************************/
int cs_message_set_signalcounter_cb(
				    cs_message_t *message,
				    char *slname,
				    void (*calc)(void *ctx, double *value, int reset),
				    void *ctx
				    )
/**************************************************************************/
{
  cs_signal_t *signal = NULL;

  if(message == NULL)
    return CS_FAIL;

  if(slname == NULL)
    return CS_FAIL;

  if(calc == NULL)
    return CS_FAIL;

  if((signal = cs_message_get_signal_byname(message,slname)) == NULL){
    return CS_FAIL;
  }

  signal->mtype = SIGNAL_MUX_COUNTER;
  signal->calc = calc;
  signal->calc_ctx = ctx;

  return CS_OK;
}

/**************************************************************************/
int cs_message_set_signalcheck_cb(
				  cs_message_t *message,
				  char *slname,
				  uint8_t (*check)(void *ctx, double *value, int reset),
				  void *ctx,
				  uint8_t *flag
				  )
/**************************************************************************/
{
  cs_signal_t *signal = NULL;

  if(message == NULL)
    return CS_FAIL;

  if(slname == NULL)
    return CS_FAIL;

  if(check == NULL)
    return CS_FAIL;

  if((signal = cs_message_get_signal_byname(message,slname)) == NULL){
    return CS_FAIL;
  }
  
  signal->mtype = SIGNAL_MUX_COUNTER;
  signal->check = check;
  signal->check_ctx = ctx;
  signal->check_flag = flag;
  return CS_OK;
}

/**************************************************************************/
int cs_message_set_defaultraw(
			      cs_message_t *message,
			      uint8_t *buffer,
			      int len
			      )
/**************************************************************************/
{
  if(message == NULL)
    return CS_FAIL;

  if(buffer == NULL)
    return CS_FAIL;

  if(len == 0)
    return CS_FAIL;

  if(len > CS_DATA)
    return CS_FAIL;

  message->mtype = MESSAGE_TYPE_DEFAULT;
  memset(message->default_rawvalue,0,CS_DATA);
  memcpy(message->default_rawvalue,buffer,len);

  return CS_OK;
}

/**************************************************************************/
int cs_message_get_defaultraw(
			      cs_message_t *message,
			      uint8_t *buffer
			      )
/**************************************************************************/
{
  if(message == NULL)
    return CS_FAIL;

  if(buffer == NULL)
    return CS_FAIL;

  memcpy(buffer,message->default_rawvalue,message->dlc);

  return CS_OK;
}

/**************************************************************************/
int cs_message_set_rawvalue(
			    cs_message_t *message,
			    uint8_t *buffer,
			    int len
			    )
/**************************************************************************/
{
  if(message == NULL)
    return CS_FAIL;
  
  if(buffer == NULL)
    return CS_FAIL;

  if(len == 0)
    return CS_FAIL;

  if(len > CS_DATA)
    return CS_FAIL;

  message->mtype = MESSAGE_TYPE_RAW;
  message->rawbuffer = buffer;
  
  return CS_OK;
}

/**************************************************************************/
int cs_message_set_direction(
			     cs_message_t *message,
			     cs_message_dir_t direction
			     )
/**************************************************************************/
{
  if(message == NULL)
    return CS_FAIL;

  message->direction = direction;

  return CS_OK;
}

/**************************************************************************/
cs_message_dir_t cs_message_get_direction(
					  cs_message_t *message
					  )
/**************************************************************************/
{
  if(message == NULL)
    return MESSAGE_DIR_NONE;

  return message->direction;
}

/**************************************************************************/
int cs_message_set_type(
			cs_message_t *message,
			cs_message_type_t mtype
			)
/**************************************************************************/
{
  if(message == NULL)
    return CS_FAIL;

  message->mtype = mtype;

  return CS_OK;
}

/**************************************************************************/
cs_message_type_t cs_message_get_type(
				      cs_message_t *message
				      )
/**************************************************************************/
{
  if(message == NULL)
    return MESSAGE_TYPE_ASSEMBLED;

  return message->mtype;
}

/**************************************************************************/
int cs_message_set_ptype(
			 cs_message_t *message,
			 cs_message_ptype_t ptype
			 )
/**************************************************************************/
{
  if(message == NULL)
    return CS_FAIL;

  message->ptype = ptype;
  return CS_OK;
}

/**************************************************************************/
cs_message_ptype_t cs_message_get_ptype(
					cs_message_t *message
					)
/**************************************************************************/
{
  if(message == NULL)
    return MESSAGE_PTYPE_CAN;
  
  return message->ptype;
}


/**************************************************************************/
int cs_message_config_setdirection_byproducer(cs_message_t *messages, 
					      int lt_messages,
					      char *producer,
					      cs_message_dir_t direction)
/**************************************************************************/
{
  int i;
  
  if(messages == NULL)
    return CS_FAIL;

  if(producer == NULL)
    return CS_FAIL;

  for(i = 0; i < lt_messages; i++){
    if(strcmp(producer,messages[i].producer)==0){
      cs_message_set_direction(&messages[i], direction);
    }
  }

  return CS_OK;
}

/**************************************************************************/
int cs_message_set_timestamp(
			     cs_message_t *message,
			     struct timeval *ts
			     )
/**************************************************************************/
{
  if(message == NULL)
    return CS_FAIL;

  if(ts == NULL)
    return CS_FAIL;

  message->externts = ts;

  return CS_OK;
}

/**************************************************************************/
int cs_message_set_timestamp_d(
			       cs_message_t *message,
			       double *tvh
			       )
/**************************************************************************/
{
  if(message == NULL)
    return CS_FAIL;

  if(tvh == NULL)
    return CS_FAIL;

  message->externtvh = tvh;

  return CS_OK;
}

/**************************************************************************/
int cs_message_set_rawbuffer(
			     cs_message_t *message,
			     uint8_t *buffer
			     )
/**************************************************************************/
{
  if(message == NULL)
    return CS_FAIL;

  if(buffer == NULL)
    return CS_FAIL;

  message->externbuffer = buffer;

  return CS_OK;
}

/**************************************************************************/
int cs_message_runtime_get_signalrawvalue(cs_message_t *message,
					 char *slname,
					 int64_t** rawvalue)
/**************************************************************************/
{
  cs_signal_t *signal = NULL;

  if(message == NULL)
    return CS_FAIL;
  
  if(slname == NULL)
    return CS_FAIL;

  if(rawvalue == NULL)
    return CS_FAIL;
  
  *rawvalue = NULL;
  if((signal = cs_message_get_signal_byname(message,slname)) == NULL){
    return CS_FAIL;
  }
  *rawvalue = &signal->rawvalue;
  return CS_OK;
}
	
/**************************************************************************/
int cs_message_runtime_set_signalraw(
				     cs_message_t *message,
				     char *slname,
				     int64_t value
				     )
/**************************************************************************/
{
  cs_signal_t *signal= NULL;
  
  if(message == NULL)
    return CS_FAIL;
  
  if(slname == NULL)
    return CS_FAIL;

  if(message->direction != MESSAGE_DIR_SEND)
    return CS_FAIL;

  if((signal = cs_message_get_signal_byname(message,slname)) == NULL){
    /*Unknown signal*/
    return CS_FAIL;
  }

  if((signal->mtype != SIGNAL_MUX_SIGNAL) && (signal->mtype != SIGNAL_MUX_RAW))
    return CS_FAIL;

  signal->mtype = SIGNAL_MUX_RAW;

  signal->rawvalue = value;
  
  return CS_OK;
}

/**************************************************************************/
int cs_message_runtime_set_signal_std(
				      cs_message_t *message,
				      char *slname
				      )
/**************************************************************************/
{
  cs_signal_t *signal= NULL;
  
  if(message == NULL)
    return CS_FAIL;
  
  if(slname == NULL)
    return CS_FAIL;

  if(message->direction != MESSAGE_DIR_SEND)
    return CS_FAIL;

  if((signal = cs_message_get_signal_byname(message,slname)) == NULL){
    /*Unknown signal*/
    return CS_FAIL;
  }

  if((signal->mtype != SIGNAL_MUX_SIGNAL) && (signal->mtype != SIGNAL_MUX_RAW))
    return CS_FAIL;

  signal->mtype = SIGNAL_MUX_SIGNAL;

  signal->rawvalue = 0;
  
  return CS_OK;
}

/**************************************************************************/
int cs_message_runtime_decode(cs_message_t *message,
			      uint8_t *buffer, int len)
/**************************************************************************/
{
  struct cs_lists_entry *entry = NULL;
  cs_signal_t *signal = NULL;
  double crcvalue;
  double crcbuffervalue;
  double checkvalue;
  uint8_t check_flag;

  if(message == NULL)
    return CS_FAIL;

  if(buffer == NULL)
    return CS_FAIL;

  if(len == 0)
    return CS_FAIL;

  _message_verbose("ID: %x\n",message->id);

  /*Increment message counter*/
  message->msg_counter++;

  switch(message->mtype){
  case MESSAGE_TYPE_RAW:
    if(message->rawbuffer != NULL){
      memcpy(message->rawbuffer,buffer,CS_DATA);
    }
    break;
  case MESSAGE_TYPE_ASSEMBLED:
    for(entry = cs_lists_get_first(&message->listsignal);
	entry != cs_lists_get_last(&message->listsignal);
	entry = cs_lists_get_next(entry)){
      signal = cs_lists_get_data(entry);
      switch(signal->mtype){
      case SIGNAL_MUX_CRC:
	/*Calc at the end*/
	continue;
	break;
      case SIGNAL_MUX_SIGNAL:
	if(signal->value != NULL){
	  if(cs_signal_runtime_decode(signal,
				      signal->value, 
				      signal->fault, 
				      buffer,len)){
	    _message_error("Message %X: Decode value %s failed\n",message->id,signal->name);
	  }
	  if(signal->label != NULL){
	    if(cs_signal_runtime_decodelabel(signal,
					     *signal->value,
					     signal->fault,  
					     signal->label)){
	      _message_error("Message %X: Decode label %s failed\n",message->id,signal->name);
	    }
	  }
	}
	continue;
	break;
      case SIGNAL_MUX_STATIC:
      case SIGNAL_MUX_RAW:
	//Nothing todo
	continue;
	break;
      case SIGNAL_MUX_COUNTER:
	if(signal->check !=NULL){
	  if(cs_signal_runtime_decode(signal,
				      &checkvalue,
				      NULL,
				      buffer,len)){
	    _message_error("Message %X: Decode value %s failed\n",message->id,signal->name);
	  }
	  if(signal->value != NULL)
	    *signal->value = checkvalue;
	  check_flag = (signal->check)(signal->check_ctx,&checkvalue,0);
	  if(signal->check_flag != NULL){
	    *signal->check_flag = check_flag;
	  } else {
	    /*if(check_flag)
	      _message_error("Signal check failed\n");*/
	  }
	}
	continue;
	break;
      case SIGNAL_MUX_MULTIPLEXOR:
      case SIGNAL_MUX_MULTIPLEXED:
	/*TODO*/
	continue;
	break;
      }
    }
    /*Check for CRC signals*/
    for(entry = cs_lists_get_first(&message->listsignal);
	entry != cs_lists_get_last(&message->listsignal);
	entry = cs_lists_get_next(entry)){
      signal = cs_lists_get_data(entry);
      if(signal->mtype == SIGNAL_MUX_CRC){
	if(signal->crc != NULL){
	  if((signal->crc)((void*)signal->crc_ctx,buffer,len, &crcvalue)){
	    /*CRC calc failed*/
	    if(signal->crc_flag != NULL)
	      *signal->crc_flag = CS_FAIL;
	  } else {
	    if(cs_signal_runtime_decode(signal,&crcbuffervalue,NULL, 
					buffer,len)){
	      /*Cannot get signal value*/
	      return CS_FAIL;
	    }
	    if(crcvalue != crcbuffervalue){
	      /*CRC compare failed*/
	      if(signal->crc_flag != NULL)
		*signal->crc_flag = CS_FAIL;
	    } else {
	      if(signal->crc_flag)
		*signal->crc_flag = CS_OK;
	    }
	  }
	  if(signal->value != NULL)
	    *signal->value = crcvalue;
	}
      }
    }
    break;
  case MESSAGE_TYPE_DIAGNOSTIC:
  case MESSAGE_TYPE_DEFAULT:
    /*Nothing todo*/
    break;
  }
  
  return CS_OK;
}

/**************************************************************************/
int cs_message_runtime_encode(
			      cs_message_t *message,
			      uint8_t *buffer,
			      int len,
			      int init
			      )
/**************************************************************************/
{
  struct cs_lists_entry *entry = NULL;
  cs_signal_t *signal = NULL;
  double crcvalue;
  double counter = 0;

  if(message == NULL)
    return CS_FAIL;

  if(buffer == NULL)
    return CS_FAIL;

  if(len == 0)
    return CS_FAIL;
  
  if(len > CS_DATA)
    return CS_FAIL;

  _message_verbose("ID: %x, Init call %s\n",message->id,init==1?"yes":"no");
  
  /*Increment message counter*/
  message->msg_counter++;

  switch(message->mtype){
  case MESSAGE_TYPE_RAW:
    if(message->rawbuffer != NULL){
      memcpy(buffer,message->rawbuffer,len);
    }
    break;
  case MESSAGE_TYPE_DEFAULT:
    memcpy(buffer,message->default_rawvalue,len);
    break;
  case MESSAGE_TYPE_DIAGNOSTIC:
      /*Nothing todo*/
      break;
  case MESSAGE_TYPE_ASSEMBLED:
    for(entry = cs_lists_get_first(&message->listsignal);
	entry != cs_lists_get_last(&message->listsignal);
	entry = cs_lists_get_next(entry)){
      signal = cs_lists_get_data(entry);
      switch(signal->mtype){
      case SIGNAL_MUX_CRC:
	continue;
	break;
      case SIGNAL_MUX_COUNTER:
	if(signal->calc != NULL){
	  (signal->calc)(signal->calc_ctx,&counter,0);
	  //_message_verbose("Calc Counter\n");
	  if(signal->value!=NULL){
	    *(signal->value)= counter;
	  }
	  _message_verbose("Signal to encode: %f\n",counter);
	  if(cs_signal_runtime_encode(signal,
			      counter, 
			      buffer,len,0)){
	    _message_error("Message %X: Encode counter value %s failed\n",message->id,signal->name);
	  }
	}
	continue; 
	break;
      case SIGNAL_MUX_SIGNAL:
	if(signal->value != NULL){
	  _message_verbose("Message %X: Signal to encode: %f\n",message->id,*signal->value);
	  if(cs_signal_runtime_encode(signal,
				      *signal->value, 
				      buffer,len,0)){
	    _message_error("Message %X: Encode value %s failed\n",message->id,signal->name);
	  }
	}
	continue;
	break;
      case SIGNAL_MUX_RAW:
	if(cs_signal_runtime_encode(signal,
				    0, //Dummy value 
				    buffer,len,0)){
	  _message_error("Message %X: Encode raw value %s failed\n",message->id,signal->name);
	}
	continue;
	break;
      case SIGNAL_MUX_STATIC:
	_message_verbose("Message %X: Signal %s value to encode statically: %f\n",message->id,signal->name,signal->startvalue);
	if(init){
	  if(cs_signal_runtime_encode(signal,
				      signal->startvalue, 
				      buffer,
				      len,
				      init
				      )){
	    _message_error("Message %X: Encode value %s failed\n",message->id,signal->name);
	  }
	} else {
	  if(cs_signal_runtime_staticencode(signal, 
					    buffer,len)){
	    _message_error("Message %X: Encode value static %s failed\n",message->id,signal->name);
	  }
	}
	//Store value in feedback value
	if(signal->value!=NULL){
	  *(signal->value)= signal->startvalue;
	}
	continue;
	break;
      case SIGNAL_MUX_MULTIPLEXOR:
      case SIGNAL_MUX_MULTIPLEXED:
	/*TODO*/
	continue;
	break;
      }
    }
    /*Check for CRC signals*/
    for(entry = cs_lists_get_first(&message->listsignal);
	entry != cs_lists_get_last(&message->listsignal);
	entry = cs_lists_get_next(entry)){
      signal = cs_lists_get_data(entry);
      if(signal->mtype == SIGNAL_MUX_CRC){
	if(signal->crc != NULL){
	  if((signal->crc)((void*)signal->crc_ctx,buffer,len, &crcvalue)){
	    /*CRC calc failed*/
	    if(signal->crc_flag)
	      *signal->crc_flag = CS_FAIL;
	  } else {
	    //_message_debug("Signal to encode: %f, Buffer:\n",crcvalue);
	    //_message_debug_buffer(buffer,len);
	    if(cs_signal_runtime_encode(signal,crcvalue, 
					buffer,len,0)){
	      /*Cannot set signal value*/
	      _message_error("Message %X: CRC Signal %s failed to encode:\n",message->id,signal->name);
	      return CS_FAIL;
	    }
	    //_message_debug("Signal valid encoded,Buffer:\n");
	    //_message_debug_buffer(buffer,len);
	    if(signal->crc_flag)
	      *signal->crc_flag = CS_OK;
	  }
	  if(signal->value!=NULL)
	    *signal->value = crcvalue;
	}
      }
    }
    break;
  }
  _message_verbose_binbuffer(buffer,len);
  return CS_OK;
}


/**************************************************************************/
int cs_message_check(cs_message_t *message)
/**************************************************************************/
{
  struct cs_lists_entry *entry = NULL;
  cs_signal_t *signal = NULL;

  if(message == NULL)
    return CS_FAIL;

  switch(message->mtype){
  case MESSAGE_TYPE_RAW:
    if(message->rawbuffer == NULL){
      _message_error("Message check (Id: 0x%X): Empty rawbuffer\n",message->id);
      return CS_FAIL;
    } else {
      return CS_OK;
    }
    break;
  case MESSAGE_TYPE_DEFAULT:
    /*Nothing to check*/
    return CS_OK;
    break;
  case MESSAGE_TYPE_DIAGNOSTIC:
    /*Nothing todo*/
    return CS_OK;
    break;
  case MESSAGE_TYPE_ASSEMBLED:
    for(entry = cs_lists_get_first(&message->listsignal);
	entry != cs_lists_get_last(&message->listsignal);
	entry = cs_lists_get_next(entry)){
      signal = cs_lists_get_data(entry);
      switch(signal->mtype){
      case SIGNAL_MUX_CRC:
	if(signal->crc == NULL){
	  _message_error("Message check (Id: 0x%X): Empty CRC (%s) Routine detected\n",message->id,signal->name);
	  return CS_FAIL;
	}
	if(cs_signal_check(signal,message->dlc)){
	  _message_error("Message check (Id: 0x%X): Signal (%s) check failed\n",message->id,signal->name);
	  return CS_FAIL;
	}
	continue;
	break;
      case SIGNAL_MUX_COUNTER:
	if(signal->calc == NULL){
	  _message_error("Message check (Id: 0x%X): Empty Counter (%s) Routine detected\n",message->id,signal->name);
	  return CS_FAIL;
	}
	if(cs_signal_check(signal,message->dlc)){
	  _message_error("Message check (Id: 0x%X): Signal (%s) check failed\n",message->id,signal->name);
	  return CS_FAIL;
	}
	continue; 
	break;
      case SIGNAL_MUX_SIGNAL:
	if(signal->value == NULL){
	  _message_error("Message check (Id: 0x%X): Empty Signal (%s) value detected\n",message->id,signal->name);
	  return CS_FAIL;
	}
	if(cs_signal_check(signal,message->dlc)){
	  _message_error("Message check (Id: 0x%X): Signal (%s) check failed\n",message->id,signal->name);
	  return CS_FAIL;
	}
	continue;
	break;
      case SIGNAL_MUX_RAW:
	/*Nothing to check*/
	continue;
	break;
      case SIGNAL_MUX_STATIC:
	if(cs_signal_check(signal,message->dlc)){
	  _message_error("Message check (Id: 0x%X): Signal (%s) check failed\n",message->id,signal->name);
	  return CS_FAIL;
	}
	continue;
	break;
      case SIGNAL_MUX_MULTIPLEXOR:
      case SIGNAL_MUX_MULTIPLEXED:
	/*TODO*/
	/*Nothing to check*/
	continue;
	break;
      }
    }
  }
  return CS_OK;
}
