/*  
    Copyright (C) 2014 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <limits.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/mman.h>
#include <unistd.h>

#include "cs_parser.h"
#include "cs_defines.h"
#include "cs_debug_private.h"


/** Open file and map it
 */
/**************************************************************************/
int cs_parser_open(
		   char*filename,
		   FILE **fs,
		   long *filelen,
		   void **map
		   )
/**************************************************************************/
{
  int filedes; 

  if((filename == NULL)||
     (fs == NULL) ||
     (filelen == NULL) ||
     (map == NULL)){
    _int_error("Parameter failure\n");
    return CS_FAIL;
  }

  *fs = fopen(filename,"r");
  if(*fs == NULL){
    _int_error("Cannot open file %s, Error: %s\n",filename,strerror(errno));
    return CS_FAIL;
  }

  //Store filedescriptor
  filedes = fileno(*fs);

  //Get size of file in byte
  //first go to end of file
  if(fseek(*fs, 0, SEEK_END) == -1){
    _int_error("Cannot seek to end of file, Error: %s\n",strerror(errno));
    return CS_FAIL;
  }
  //Get size
  *filelen = ftell(*fs);

  //Back to start
  if(fseek(*fs, 0, SEEK_SET) == -1){
    _int_error("Cannot seek to start of file, Error: %s\n",strerror(errno));
    return CS_FAIL;
  }

  //Try to map
  *map = mmap(0, *filelen, PROT_READ, MAP_PRIVATE, filedes, 0);
  if(*map == MAP_FAILED){
    _int_error("Cannot map file, Error: %s\n",strerror(errno));
    return CS_FAIL;
  }

  return CS_OK;
}

/**************************************************************************/
int cs_parser_close(
		    FILE * fs,
		    void *map,
		    long filelen
		    )
/**************************************************************************/
{

  if(map == NULL){
    _int_error("Parameter failure\n");
    return CS_FAIL;
  }

  //destroy mapping
  if(munmap(map, filelen) == -1){
    _int_error("Cannot unmmap file, Error: %s\n",strerror(errno));
    return CS_FAIL;
  }

  //Close file stream
  if(fclose(fs) != 0){
    _int_error("Cannot close file, Error: %s\n",strerror(errno));
    return CS_FAIL;
  }
  return CS_OK;
}


/** Calculate next values for search loops
 */
/**************************************************************************/
int cs_parser_calcresult(
			 cs_parser_result_t *result
			 )
/**************************************************************************/
{
  if(result == NULL){
    _int_error("Parameter failure\n");
    return CS_FAIL;
  }

  if((result->result == NULL) ||
     (result->resultlen <= 0)){
    result->next = NULL;
    result->nextlen = 0;
  } else {
    if(result->searchlen > (result->resultoffset + result->resultlen)){
      result->next = &result->searchin[result->resultoffset + result->resultlen];
      result->nextlen = result->searchlen - result->resultoffset - result->resultlen;
    } else {
      result->next = NULL;
      result->nextlen = 0;
    }
  }

  return CS_OK;
}

/** Step to after first restult
 */
/**************************************************************************/
int cs_parser_stepnext(
		       cs_parser_result_t *result,
		       int *ovector,
		       int vsize,
		       int index
		       )
/**************************************************************************/
{
  int offset;
  if((result == NULL)||
     (ovector == NULL)||
     (vsize <= 1)||
     ((index*3) >= vsize)
     ){
    _int_error("Parameter failure\n");
    return CS_FAIL;
  }
  offset = index*3;
  result->result = &result->result[ovector[offset+2]];
  result->resultlen = result->resultlen - ovector[offset+2];
  result->resultoffset = ovector[offset+1];
  //cs_parser_printresult(result);
  return CS_OK;
}

/** Reset result structure
 */
/**************************************************************************/
void cs_parser_resetresult(
			   cs_parser_result_t *result
			   )
/**************************************************************************/
{
  result->searchin = NULL;
  result->searchlen = 0;
  result->result = NULL;
  result->resultlen = 0;
  result->next = NULL;
  result->nextlen = 0;
}

/** Precompile regex
    \retval Null if there is an failue
    \retval Pointer to regex structure
*/
/**************************************************************************/
pcre *cs_parser_precompile(
			   const char *pattern
			   )
/**************************************************************************/
{
  pcre *r = NULL;
  const char *error = NULL;
  int erroroffset = 0;
  
  if(pattern == NULL){
    _int_error("Parameter failure\n");
    return r;
  }
  
  r = pcre_compile(pattern, PCRE_DOTALL | PCRE_MULTILINE | PCRE_NEWLINE_ANY,
		   &error, &erroroffset, NULL);
  
  if(r == NULL){
    _int_error("Regex compilation failed, Error: %s at position %i\n",error,erroroffset);
    pcre_free_substring(error);
    return r;
  }
  return r;
}


/**************************************************************************/
int cs_parser_getsubstring(
			   pcre *re,
			   char *searchin,
			   int *ovector,
			   int osize,
			   char *tag,
			   char *substring
			   )
/**************************************************************************/
{
  int namednumber = 0;
  const char *pcrestring = NULL;

  if((re == NULL) ||
     (tag == NULL) ||
     (searchin == NULL) ||
     (ovector == NULL) ||
     (substring == NULL)){
    _int_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  if((namednumber = pcre_get_stringnumber(re,tag)) == PCRE_ERROR_NOSUBSTRING){
    _int_error("Failed to get substring %s\n",tag);
    return CS_FAIL;
  }
  
  memset(substring,0,CS_PARSER_CHARBUF);
  
  if(pcre_get_substring(searchin, ovector,osize,namednumber,(const char**)&pcrestring)< 0){
    _int_error("Failed to get named substring\n");
    return CS_FAIL;
  }
  if(strlen(pcrestring) > CS_PARSER_CHARBUF){
    _int_error("Buffer overrun detected\n");
    return CS_FAIL;
  }

  strncpy(substring,pcrestring,CS_PARSER_CHARBUF);
  pcre_free_substring(pcrestring);
 
  return CS_OK;
}

/**************************************************************************/
int cs_parser_getsubstring_malloc(
				  pcre *re,
				  char *searchin,
				  int *ovector,
				  int osize,
				  char *tag,
				  char **substring
				  )
/**************************************************************************/
{
  int namednumber = 0;
  const char *pcrestring = NULL;
  int len;

  if((re == NULL) ||
     (tag == NULL) ||
     (searchin == NULL) ||
     (ovector == NULL) ||
     (substring == NULL)){
    _int_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  if((namednumber = pcre_get_stringnumber(re,tag)) == PCRE_ERROR_NOSUBSTRING){
    _int_error("Failed to get substring %s\n",tag);
    return CS_FAIL;
  }
    
  if(pcre_get_substring(searchin, ovector,osize,namednumber,(const char**)&pcrestring)< 0){
    _int_error("Failed to get named substring\n");
    return CS_FAIL;
  }

  len = strlen(pcrestring) + 1;
  *substring = NULL;
  *substring = (char*)malloc(len*sizeof(char));

  if(*substring == NULL){
    _int_error("Malloc failure\n");
    return CS_FAIL;
  }

  memset(*substring,0,len*(sizeof(char)));
  
  strncpy(*substring,pcrestring,len);
  pcre_free_substring(pcrestring);
 
  return CS_OK;
}

/**************************************************************************/
int cs_parser_getsubstring_int(
			       pcre *re,
			       char *searchin,
			       int *ovector,
			       int osize,
			       char *tag,
			       int *value,
			       int dvalue
			       )
/**************************************************************************/
{
  char substring[CS_PARSER_CHARBUF];

  if((re == NULL) ||
     (tag == NULL) ||
     (searchin == NULL) ||
     (ovector == NULL) ||
     (value == NULL)){
    _int_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  *value = dvalue;

  if(cs_parser_getsubstring(re,searchin,ovector,osize,tag,substring)){
    return CS_FAIL;
  }

  if(!sscanf(substring,"%i",value)){
    return CS_FAIL;
  }
  return CS_OK;
}

/**************************************************************************/
int cs_parser_getsubstring_hex(
			       pcre *re,
			       char *searchin,
			       int *ovector,
			       int osize,
			       char *tag,
			       int *value,
			       int dvalue
			       )
/**************************************************************************/
{
  char substring[CS_PARSER_CHARBUF];
  long int l;

  if((re == NULL) ||
     (tag == NULL) ||
     (searchin == NULL) ||
     (ovector == NULL) ||
     (value == NULL)){
    _int_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  *value = dvalue;

  if(cs_parser_getsubstring(re,searchin,ovector,osize,tag,substring)){
    return CS_FAIL;
  }
  l = strtol(substring, NULL, 16);
  if((l == LONG_MIN)|| (l == LONG_MAX)){
    return CS_FAIL;
  }
  *value = (int)l;
  return CS_OK;
}

/**************************************************************************/
int cs_parser_getsubstring_double(
				  pcre *re,
				  char *searchin,
				  int *ovector,
				  int osize,
				  char *tag,
				  double *value,
				  double dvalue
				  )
/**************************************************************************/
{
  char substring[CS_PARSER_CHARBUF];  

  if((re == NULL) ||
     (tag == NULL) ||
     (searchin == NULL) ||
     (ovector == NULL) ||
     (value == NULL)){
    _int_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  *value = dvalue;

  if(cs_parser_getsubstring(re,searchin,ovector,osize,tag,substring)){
    return CS_FAIL;
  }

  if(!sscanf(substring,"%lf",value)){
    return CS_FAIL;
  }
  return CS_OK;
}



/**************************************************************************/
void cs_parser_printresult(
			   cs_parser_result_t *result
			   )
/**************************************************************************/
{
  char *buffer = NULL;
  int slen = 0;
  if(result == NULL){
    _int_error("Parameter failure\n");
    return;
  }

  if(result->resultlen <= 0)
    return;

  slen = result->resultlen + 1;

  buffer = (char*)malloc(sizeof(char)*slen);

  if(buffer == NULL)
    return;

  memset(buffer,0,slen);

  memcpy(buffer,result->result,result->resultlen); //Skip last 0 byte

  _int_info("Regex result Len: %i, String: %s\n",result->resultlen,buffer);

  free(buffer);
}

/**************************************************************************/
void cs_parser_printsearch(
			   cs_parser_result_t *result
			   )
/**************************************************************************/
{
  char *buffer = NULL;
  int slen = 0;
  if(result == NULL){
    _int_error("Parameter failure\n");
    return;
  }

  if(result->searchlen <= 0)
    return;

  slen = result->searchlen + 1;

  buffer = (char*)malloc(sizeof(char)*slen);

  if(buffer == NULL)
    return;

  memset(buffer,0,slen);

  memcpy(buffer,result->searchin,result->searchlen); //Skip last 0 byte

  _int_info("Regex result Len: %i, String: %s\n",result->searchlen,buffer);

  free(buffer);
}

/**************************************************************************/
int cs_parser_search_stringline(
				cs_parser_result_t *result,
				char *name
				)
/**************************************************************************/
{
  pcre *re = NULL;
  char regex[CS_REGEX_BUFFER];
  int rv = 0;
  int ovector[CS_REGEX_RESULTSIZE];

  if((name == NULL) ||
     (result == NULL)){
    _int_error("Parameter failure\n");
    return CS_FAIL;
  }

  //Reset result
  result->result = NULL;
  result->resultlen = 0;
  result->resultoffset = 0;

  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER);
  snprintf(regex,CS_REGEX_BUFFER,"%s",name);

  //Precompile regex
  if((re = cs_parser_precompile((const char *)regex)) == NULL){
    _int_error("Regex precompilation failed\n");
    return CS_FAIL;
  }

  //Get data, only one fit is allowed
  rv = pcre_exec(re, NULL, (const char*)result->searchin,result->searchlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE);

  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _int_verbose("No match\n"); 
      pcre_free(re);
      return CS_FAIL;
      break;
    default: 
      _int_error("Matching error %d\n", rv);
      pcre_free(re);
      return CS_FAIL;
      break;
    }
  } else {
    result->result = &result->searchin[ovector[0]];
    result->resultlen = ovector[1] - ovector[0];
    result->resultoffset = ovector[0];
  }

  //Calc next values
  cs_parser_calcresult(result);
  pcre_free(re); 

  return CS_OK;
}

/**************************************************************************/
int cs_parser_preparestring(
			    char *raw,
			    char *clean,
			    int buflen
			    )
/**************************************************************************/
{
  //int print = 0;
  int in,out;
  if((raw == NULL)||(clean == NULL)){
    _int_error("Parameter failure\n");
    return CS_FAIL;
  }

  in = 0;
  out = 0;

  memset(clean,0,sizeof(char)*buflen);

  while(raw[in]!= 0){
    switch (raw[in]){
    case '[':
    case ']':
      clean[out]=92;//'\'
      out++;
      //print = 1;
    default:
      clean[out] = raw[in];
    }
    in++;
    out++;
    if(out >= buflen){
      //Buffer length failure
      return CS_FAIL;
    }
  }
  /*if(print)
    _int_info("Cleaned up string: old: %s, new: %s\n",raw,clean);*/
  return CS_OK;
}



/**************************************************************************/
cs_parser_regex_t* cs_parser_regex(
				   cs_parser_result_t *search,
				   char *regex,
				   int noresult
				   )
/**************************************************************************/
{
  cs_parser_regex_t* rv = NULL;
  int rv_exec;

  if((search == NULL)||
     (search->searchin == NULL) ||
     (search->searchlen < 1) ||
     (regex == NULL) ||
     (strlen(regex) < 1) ||
     (noresult < 1)){
    _int_error("Parameter failure\n");
    return NULL;
  }
  
  //Malloc structure
  if((rv = (cs_parser_regex_t*)calloc(1,sizeof(cs_parser_regex_t))) == NULL){
    _int_error("Malloc failure\n");
    return NULL;
  }
  
  if((rv->ovector = (int*)calloc(noresult*CS_REGEX_RESULTSIZE,sizeof(int))) == NULL){
    _int_error("Malloc failure\n");
    free(rv);
    return NULL;
  }

  //Store arguments
  rv->search = search;
  rv->noresult = noresult;
  
  //Precompile regex
  if((rv->re = cs_parser_precompile((const char *)regex)) == NULL){
    _int_error("Regex precompilation failed\n");
    free(rv->ovector);
    free(rv);
    return NULL;
  }
  //Get data, only one fit is allowed
  rv_exec = pcre_exec(rv->re, NULL, (const char*)rv->search->searchin,rv->search->searchlen, 0,PCRE_NOTEMPTY, rv->ovector, CS_REGEX_RESULTSIZE * rv->noresult);
  if(rv_exec <= 0) {
    switch(rv_exec) {
    case PCRE_ERROR_NOMATCH:
      _int_verbose("No content match\n");
      break;
    case 0:
      _int_error("The result vector was not big enough\n");
      break;
    default: 
      _int_error("Matching error %d\n", rv_exec);
      break;
    }
    pcre_free(rv->re);
    free(rv->ovector);
    free(rv);
    return NULL;
  } 
  _int_verbose("Found %d results\n",rv_exec);
  if(rv_exec < rv->noresult){
    //Reduce to found size
    rv->noresult = rv_exec;
  }
  return rv;
}

/**************************************************************************/
void cs_parser_regex_free(
			  cs_parser_regex_t* st
			  )
/**************************************************************************/
{
  if(st == NULL){
    _int_error("Parameter failure\n");
    return;
  }

  pcre_free(st->re);
  free(st->ovector);
  free(st);
  st= NULL;
}

/**************************************************************************/
int cs_parser_regex_next(
			 cs_parser_regex_t* st,
			 cs_parser_result_t *next
			 )
/**************************************************************************/
{
  if((st == NULL) || (next == NULL)){
    _int_error("Parameter failure\n");
    return CS_FAIL;
  }

  //st->ovector[1] is the end index of the complete matched string
  
  if((st->search->searchlen - st->ovector[1]) > 2){
    //We are not finished
    next->searchlen = st->search->searchlen - st->ovector[1];
    next->searchin = &st->search->searchin[st->ovector[1]];

    return CS_OK;
  }
  _int_verbose("Nothing to search\n");
  return CS_FAIL;
}

 
/**************************************************************************/
int cs_parser_regex_getStr(
			   cs_parser_regex_t* st,
			   char *label,
			   char **value
			   )
/**************************************************************************/
{ 
  if((st == NULL) ||
     (label == NULL) ||
     (value == NULL)){
    _int_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  if(st->re == NULL){
    _int_error("%s: Regex is not prepared\n");
  }

  if(cs_parser_getsubstring_malloc(st->re,st->search->searchin,st->ovector,CS_REGEX_RESULTSIZE * st->noresult,
				   label, value)){
    return CS_FAIL;
  }
  
  return CS_OK;
}

/**************************************************************************/
int cs_parser_regex_getInt(
			   cs_parser_regex_t* st,
			   char *label,
			   int *value,
			   int defaultvalue
			   )
/**************************************************************************/
{
  char *substring = NULL;

  *value = defaultvalue;
  
  if((st == NULL) ||
     (label == NULL) ||
     (value == NULL)){
    _int_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  if(st->re == NULL){
    _int_error("Regex is not prepared\n");
  }

  if(cs_parser_getsubstring_malloc(st->re,st->search->searchin,st->ovector,CS_REGEX_RESULTSIZE * st->noresult,
				   label, &substring)){
    return CS_FAIL;
  }
  
  if(!sscanf(substring,"%i",value)){
    free(substring);
    return CS_FAIL;
  }
  free(substring);
  return CS_OK;
}

/**************************************************************************/
int cs_parser_regex_getHex(
			   cs_parser_regex_t* st,
			   char *label,
			   int *value,
			   int defaultvalue
			   )
/**************************************************************************/
{;
  char *substring = NULL;
  long int l;
  
  *value = defaultvalue;
  
  if((st == NULL) ||
     (label == NULL) ||
     (value == NULL)){
    _int_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  if(st->re == NULL){
    _int_error("Regex is not prepared\n");
  }

  if(cs_parser_getsubstring_malloc(st->re,st->search->searchin,st->ovector,CS_REGEX_RESULTSIZE * st->noresult,
				   label, &substring)){
    return CS_FAIL;
  }
  

  l = strtol(substring, NULL, 16);
  if((l == LONG_MIN)|| (l == LONG_MAX)){
    free(substring);
    return CS_FAIL;
  }
  *value = (int)l;

  free(substring);
  return CS_OK;
}

/**************************************************************************/
int cs_parser_regex_getDouble(
			      cs_parser_regex_t* st,
			      char *label,
			      double *value,
			      double defaultvalue
			      )
/**************************************************************************/
{
  char *substring = NULL;

  *value = defaultvalue;
  
  if((st == NULL) ||
     (label == NULL) ||
     (value == NULL)){
    _int_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  if(st->re == NULL){
    _int_error("Regex is not prepared\n");
  }

  if(cs_parser_getsubstring_malloc(st->re,st->search->searchin,st->ovector,CS_REGEX_RESULTSIZE * st->noresult,
				   label, &substring)){
    return CS_FAIL;
  }

  if(!sscanf(substring,"%lf",value)){
    free(substring);
    return CS_FAIL;
  }
  
  free(substring);
  return CS_OK;
}
