/*
  Copyright (C) 2019 IgH, Andreas Stewering-Bone

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <limits.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <sched.h>
#include <signal.h>
#include <unistd.h>
#include <sys/mman.h>

#include "config.h"
#include "cs_defines.h"
#include "cs_process.h"
#include "cs_time.h"
#include "cs_debug_private.h"

#define _process_error(...) (cs_error(CS_DEBUG_PROCESS,__func__,__LINE__,__VA_ARGS__))
#define _process_warn(...) (cs_warn(CS_DEBUG_PROCESS,__func__,__LINE__,__VA_ARGS__))
#define _process_info(...) (cs_info(CS_DEBUG_PROCESS,__func__,__LINE__,__VA_ARGS__))
#define _process_debug(...) (cs_debug(CS_DEBUG_PROCESS,__func__,__LINE__,__VA_ARGS__))
#define _process_verbose(...) (cs_verbose(CS_DEBUG_PROCESS,__func__,__LINE__,__VA_ARGS__))
#define _process_error_buffer(DATA,LEN) (cs_error_buffer(CS_DEBUG_PROCESS,__func__,__LINE__,DATA,LEN))
#define _process_warn_buffer(DATA,LEN) (cs_warn_buffer(CS_DEBUG_PROCESS,__func__,__LINE__,DATA,LEN))
#define _process_info_buffer(DATA,LEN) (cs_info_buffer(CS_DEBUG_PROCESS,__func__,__LINE__,DATA,LEN))
#define _process_debug_buffer(DATA,LEN) (cs_debug_buffer(CS_DEBUG_PROCESS,__func__,__LINE__,DATA,LEN))
#define _process_verbose_buffer(DATA,LEN) (cs_verbose_buffer(CS_DEBUG_PROCESS,__func__,__LINE__,DATA,LEN))


/********************************************************************/
cs_debug_level_t cs_process_debuglevel(
				       cs_debug_level_t level
				       )
/********************************************************************/
{
  return cs_debuglevel(CS_DEBUG_PROCESS,level);
}

/********************************************************************/
void cs_process_debug_setcallback(
				  void *ctx,
				  void (*callb)(void *ctx,
						cs_debug_sub_t sub,
						cs_debug_level_t dl,
						const char *func,
						const int line,
						const char *fmt,
						va_list ap)
				  )
/********************************************************************/
{
  cs_debug_setcallback(CS_DEBUG_PROCESS,ctx,callb);
}

/************************************************************/
void cs_process_sleep(
		      uint32_t sleeptime // in ms
		      )
/************************************************************/
{
  struct timespec wtime;

  if(sleeptime == 0){
    return;
  }

  wtime = cs_ts_timeout2ts(sleeptime);
  
  clock_nanosleep(CLOCK_MONOTONIC,0,&wtime,NULL);
  
}

/**************************************************************************/
cs_process_t *cs_process_init()
/**************************************************************************/
{
  cs_process_t * p = NULL;

  p = (cs_process_t*)calloc(sizeof(cs_process_t),1);

  if(p == NULL){
    _process_error("Malloc failure\n");
    return NULL;
  }

  cs_process_reset(p);

  return p;
}

/**************************************************************************/
void cs_process_free(
		     cs_process_t *process
		     )
/**************************************************************************/
{
  if(process == NULL)
    return;

  free(process);
  process = NULL;
  
}

/**************************************************************************/
int cs_process_reset(
		     cs_process_t *process
		     )
/**************************************************************************/
{
  if(process == NULL){
    _process_error("Parameter failure\n");
    return CS_FAIL;
  }


  return CS_OK;
}

/************************************************************/
int cs_process_prepare(
		       cs_process_t * process,
		       int priority,
		       double sampletime,
		       int overruns_allowed
		       )
/************************************************************/
{
  struct sched_param param;
  
  if(process == NULL){
    _process_error("Parameter failure\n");
    return CS_FAIL;
  }

  _process_verbose("Priority: %i\n",priority);
  _process_verbose("Sampletime: %f\n",sampletime);
  _process_verbose("Allowed overruns: %i\n",overruns_allowed);

  if(priority >= 0){
    /*Set realtime stuff*/
    mlockall(MCL_CURRENT|MCL_FUTURE);
    /*Set process priority*/
    param.sched_priority = priority;
    if(sched_setscheduler(0, SCHED_FIFO, &param) == -1) {
      _process_warn("Failed to set process priority\n");
    }
  }

  /*Init sleep time*/
  process->sleep_time.tv_sec = 0;
  process->sleep_time.tv_nsec = (int)(sampletime * 1000000000.0);
  process->overruns = 0;
  process->sampletime = sampletime;
  process->overruns_allowed = overruns_allowed;

  /*Set time values*/
  clock_gettime(CLOCK_MONOTONIC, &process->start_time);
  process->wakeup_time = cs_ts_addtimespec(process->start_time,process->sleep_time);

  return CS_OK;
}

/************************************************************/
void cs_process_cycle_start(
			    cs_process_t *process
			    )
/************************************************************/
{
  if(process == NULL){
    _process_error("Parameter failure\n");
    return;
  }

  /*Prepare Sleeping*/
  process->wakeup_time = cs_ts_addtimespec(process->wakeup_time,process->sleep_time);
  clock_nanosleep(CLOCK_MONOTONIC,TIMER_ABSTIME,&process->wakeup_time,NULL);
  
  /*Store old start time for cycle calculation*/
  process->start_time_old = process->start_time;
  /*Get proc start time*/
  clock_gettime(CLOCK_MONOTONIC, &process->start_time);
  /*Get proc external time*/
  clock_gettime(CLOCK_REALTIME, &process->external_time);
}

/************************************************************/
int cs_process_cycle_end(
			 cs_process_t *process
			 )
/************************************************************/
{
  if(process == NULL){
    _process_error("Parameter failure\n");
    return CS_FAIL;
  }

  /*Statistics*/
  /*Update time value*/
  process->dbltime = cs_ts_timespec2d(process->start_time);
  /*Get proc final time*/
  clock_gettime(CLOCK_MONOTONIC, &process->end_time);
  /*Calculate execution time*/
  process->exec_time = cs_ts_difftimespec(process->start_time, process->end_time);
  process->exec_time_d = cs_ts_timespec2d(process->exec_time);
  /*Calculate peridical calling time*/
  process->cycle_time = cs_ts_difftimespec(process->start_time_old, process->start_time);
  process->cycle_time_d = cs_ts_timespec2d(process->cycle_time);
  /*Check for overruns*/
  if(cs_ts_timespec2d(process->cycle_time) > (1.5* process->sampletime)){
    /*Check for overrun intmax limit*/
    if(process->overruns != INT_MAX)
      process->overruns++;
    if(process->overruns_allowed >= 0){
      if(process->overruns > process->overruns_allowed){
	return 3;
      }
      return 2;
    }
  }
  return CS_OK;
}
