/*  scheduler.h --  scheduling CAN messages
    Copyright (C) 2013 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
#include <string.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/uio.h>
#include <net/if.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <limits.h>

#include "config.h"
#include "cs_schedule.h"
#include "cs_schedule_private.h"
#include "cs_socketcan.h"
#include "cs_defines.h"
#include "cs_time.h"
#include "cs_message_private.h"
#include "cs_signal_private.h"
#include "cs_debug_private.h"



#define _schedule_error(...) (cs_error(CS_DEBUG_SCHEDULE,__func__,__LINE__,__VA_ARGS__))
#define _schedule_warn(...) (cs_warn(CS_DEBUG_SCHEDULE,__func__,__LINE__,__VA_ARGS__))
#define _schedule_info(...) (cs_info(CS_DEBUG_SCHEDULE,__func__,__LINE__,__VA_ARGS__))
#define _schedule_debug(...) (cs_debug(CS_DEBUG_SCHEDULE,__func__,__LINE__,__VA_ARGS__))
#define _schedule_verbose(...) (cs_verbose(CS_DEBUG_SCHEDULE,__func__,__LINE__,__VA_ARGS__))
#define _schedule_error_buffer(DATA,LEN) (cs_error_buffer(CS_DEBUG_SCHEDULE,__func__,__LINE__,DATA,LEN))
#define _schedule_warn_buffer(DATA,LEN) (cs_warn_buffer(CS_DEBUG_SCHEDULE,__func__,__LINE__,DATA,LEN))
#define _schedule_info_buffer(DATA,LEN) (cs_info_buffer(CS_DEBUG_SCHEDULE,__func__,__LINE__,DATA,LEN))
#define _schedule_debug_buffer(DATA,LEN) (cs_debug_buffer(CS_DEBUG_SCHEDULE,__func__,__LINE__,DATA,LEN))
#define _schedule_verbose_buffer(DATA,LEN) (cs_verbose_buffer(CS_DEBUG_SCHEDULE,__func__,__LINE__,DATA,LEN))


/********************************************************************/
cs_debug_level_t cs_schedule_debuglevel(
					cs_debug_level_t level
					)
/********************************************************************/
{
  return cs_debuglevel(CS_DEBUG_SCHEDULE,level);
}

/********************************************************************/
void cs_schedule_debug_setcallback(
    void *ctx,
    void (*callb)(void *ctx,
		  cs_debug_sub_t sub,
		  cs_debug_level_t dl,
		  const char *func,
		  const int line,
		  const char *fmt,
		  va_list ap)
    )
/********************************************************************/
{
  cs_debug_setcallback(CS_DEBUG_SCHEDULE,ctx,callb);
}

/**************************************************************/
cs_schedule_message_t *_schedule_get_message(
					     cs_schedule_t *scheduler,
					     uint32_t id,
					     cs_message_dir_t dir
					     )
/**************************************************************/
{
  struct cs_lists_entry *entry = NULL;
  cs_lists_t *l = NULL;
  cs_schedule_message_t* scm = NULL;

  switch(dir){
  case MESSAGE_DIR_SEND:
    l = &scheduler->l_send;
    break;
  case MESSAGE_DIR_RECEIVE:
    l = &scheduler->l_recv;
    break;
  case MESSAGE_DIR_NONE:
  case MESSAGE_DIR_SKIP:
    return NULL;
  }

  for(entry = cs_lists_get_first(l);
      entry != cs_lists_get_last(l);
      entry = cs_lists_get_next(entry)){
    scm = (cs_schedule_message_t*)cs_lists_get_data(entry);
    if(scm->message->id == id){
      return scm;
    }
  }
  
  return NULL;
}

/** Add message for list handling
 */
/**************************************************************************/
void *_schedule_add_message( void *ext)
/**************************************************************************/
{
  return ext;
}

/** Delete message for list handling
 */
/**************************************************************************/
void _schedule_del_message( void *intern)
/**************************************************************************/
{
  if(intern != NULL)
    free(intern);
}

/**************************************************************************/
cs_schedule_t *cs_schedule_init()
/**************************************************************************/
{
  cs_schedule_t * sc = NULL;

  sc = (cs_schedule_t*)calloc(sizeof(cs_schedule_t),1);
  
  if(sc == NULL){
    _schedule_error("Malloc failure\n");
    return NULL;
  }

  cs_lists_init(&sc->l_send,_schedule_add_message,_schedule_del_message);
  cs_lists_init(&sc->l_recv,_schedule_add_message,_schedule_del_message);

  cs_schedule_reset(sc);

  return sc;
}

/**************************************************************************/
void cs_schedule_free(
		      cs_schedule_t *scheduler
		      )
/**************************************************************************/
{
  if(scheduler == NULL){
    return;
  }
  cs_lists_destroy(&scheduler->l_send);
  cs_lists_destroy(&scheduler->l_recv);
  free(scheduler);
  scheduler = NULL;
}
  
/**************************************************************************/
int cs_schedule_reset(
		      cs_schedule_t *sc
		      )
/**************************************************************************/
{
  struct cs_lists_entry *entry = NULL;
  cs_schedule_message_t *scm = NULL;
  
  if(sc == NULL)
    return CS_FAIL;
  sc->cycle = 0;
  sc->max_messpcycle = 1;
  sc->send_shifted_messages = 0;
  sc->send_no_messages = 0;
  sc->send_tx_full_counter = 0;
  sc->send_error_counter = 0;
  
  //Reset messages
  for(entry = cs_lists_get_first(&sc->l_send);
      entry != cs_lists_get_last(&sc->l_send);
      entry = cs_lists_get_next(entry)){
    scm = (cs_schedule_message_t*)cs_lists_get_data(entry);
    scm->counter = 0;
    scm->state = SCHEDULE_STATE_IDLE;
    scm->action = SCHEDULE_MSGACTION_PAUSED;
  }

    for(entry = cs_lists_get_first(&sc->l_recv);
      entry != cs_lists_get_last(&sc->l_recv);
      entry = cs_lists_get_next(entry)){
    scm = (cs_schedule_message_t*)cs_lists_get_data(entry);
    scm->counter = 0;
    scm->state = SCHEDULE_STATE_IDLE;
    scm->action = SCHEDULE_MSGACTION_PAUSED;
  }

  return CS_OK;
}

/**************************************************************/
int cs_schedule_exist_message(
			      cs_schedule_t *scheduler,
			      uint32_t id
			      )
/**************************************************************/
{ 
  if(scheduler == NULL){
    _schedule_error("Parameter failure\n");
    return CS_FAIL;
  }

  //check if message already exist
  if(_schedule_get_message(scheduler,id,MESSAGE_DIR_SEND)!= NULL){
    _schedule_warn("Message 0x%X already exists in send list\n",id);
    return 2;
  }

  if(_schedule_get_message(scheduler,id,MESSAGE_DIR_RECEIVE)!= NULL){
    _schedule_warn("Message 0x%X already exists in receive list\n",id);
    return 2;
  }

  return CS_OK;
}
/**************************************************************/
int cs_schedule_add_message(
			    cs_schedule_t *scheduler,
			    cs_message_t* message
			    )
/**************************************************************/
{
  cs_schedule_message_t *scm = NULL;
  uint8_t dummy[CS_DATA];

  
    
  if((scheduler == NULL) || (message == NULL)){
    _schedule_error("Parameter failure\n");
    return CS_FAIL;
  }

  //check if message already exist
  if(_schedule_get_message(scheduler,message->id,MESSAGE_DIR_SEND)!= NULL){
    _schedule_warn("Message %s already exists in send list\n",cs_message_get_name(message));
    return 2;
  }

  if(_schedule_get_message(scheduler,message->id,MESSAGE_DIR_RECEIVE)!= NULL){
    _schedule_warn("Message %s already exists in receive list\n",cs_message_get_name(message));
    return 2;
  }

  if(cs_message_check(message)){
    _schedule_error("Message %s parameter failure\n",cs_message_get_name(message));
    return CS_FAIL;
  }

  
  scm = (cs_schedule_message_t *)calloc(sizeof(cs_schedule_message_t),1);

  if(scm == NULL){
    _schedule_error("Malloc failure\n");
    return CS_FAIL;
  }

  //Note not copy message, only store pointer
  //So keep care on external free
  scm->message = message;

  //Store basic data for faster access
  scm->id = message->id;
  scm->cycle = message->interval;
  /*Init counter with start and interval offset*/
  scm->counter = message->interval_offset;// + message->interval_startoffset;
  scm->timeout = message->timeout;
  scm->action = SCHEDULE_MSGACTION_RUN;

  //set send type
  switch(message->ptype){
  case MESSAGE_PTYPE_CANFD:
    scm->smode = SOCKETCAN_SEND_CANFD;
    break;
  case MESSAGE_PTYPE_CAN:
  default:
    scm->smode = SOCKETCAN_SEND_CAN;
    break;
  }
  
  if(message->direction == MESSAGE_DIR_RECEIVE){
    //Add ro receive list
    cs_lists_insert(&scheduler->l_recv,(void*)scm);
  }

  if(message->direction == MESSAGE_DIR_SEND){
    /*Precalc values for send direction*/
    memset(dummy,0,sizeof(uint8_t)*CS_DATA);
    if(cs_message_runtime_encode(message,dummy,message->dlc,1)){
      /*Encoding failed*/
      _schedule_error("Encoding failed: ID: %X\n",message->id);
      return CS_FAIL;
    }
    //Add to send list
    cs_lists_insert(&scheduler->l_send,(void*)scm);
  }

  return CS_OK;
}

/**************************************************************/
int cs_schedule_set_cycletime(
			      cs_schedule_t *scheduler,
			      uint32_t cycle
			      )
/**************************************************************/
{
  if(scheduler == NULL){
    _schedule_error("Parameter failure\n");
    return CS_FAIL;
  }

  if(cycle == 0){
    _schedule_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  scheduler->cycle = cycle;
  return CS_OK;
}

/**************************************************************/
int cs_schedule_set_maxpcycle(
			      cs_schedule_t *scheduler,
			      int  max_messpcycle
			      )
/**************************************************************/
{
  if(scheduler == NULL){
    _schedule_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  if(max_messpcycle < 1){
    _schedule_error("Parameter failure\n");
    return CS_FAIL;
  }

  scheduler->max_messpcycle = max_messpcycle;
  return CS_OK;
}

/**************************************************************/
int cs_schedule_set_msgaction(
			      cs_schedule_t *scheduler,
			      uint32_t id,
			      cs_schedule_msgaction_t action
			      )
/**************************************************************/
{
  cs_schedule_message_t* dummy = NULL;

  if(scheduler == NULL){
    _schedule_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  if((dummy = _schedule_get_message(scheduler,id,MESSAGE_DIR_SEND)) == NULL){
    _schedule_error("Failed to get message ID 0x%X\n",id);
    return CS_FAIL;
  }
  
  dummy->action = action;

  return CS_OK;
}

/**************************************************************/
int cs_schedule_set_msgaction_all(
				  cs_schedule_t *scheduler,
				  cs_schedule_msgaction_t action
				  )
/**************************************************************/
{
  struct cs_lists_entry *entry = NULL;
  cs_schedule_message_t* scm = NULL;
  
  if(scheduler == NULL){
    _schedule_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  for(entry = cs_lists_get_first(&scheduler->l_send);
      entry != cs_lists_get_last(&scheduler->l_send);
      entry = cs_lists_get_next(entry)){
    scm = (cs_schedule_message_t*)cs_lists_get_data(entry); 
    scm->action = action;
  }
 
  return CS_OK;
}

/**************************************************************/
cs_schedule_msgstate_t cs_schedule_get_msgstate(
						cs_schedule_t *scheduler,
						uint32_t id
						)
/**************************************************************/
{
  cs_schedule_message_t* dummy = NULL;

  if(scheduler == NULL){
    _schedule_error("Parameter failure\n");
    return SCHEDULE_STATE_IDLE;
  }

  if((dummy = _schedule_get_message(scheduler,id,MESSAGE_DIR_RECEIVE)) == NULL){
    _schedule_error("Failed to get Message ID 0x%X\n",id);
    return SCHEDULE_STATE_IDLE;
  }

  return dummy->state;
}

/**************************************************************/
int cs_schedule_open(
		     cs_schedule_t *scheduler,
		     char *iface, 
		     cs_socketcan_mode_t mode
		     )
/**************************************************************/
{
  uint32_t *canid = NULL;
  int num_id = 0;
  int i;
  
  if((scheduler == NULL) || (iface == NULL)){
    _schedule_error("Parameter failure\n");
    return CS_FAIL;
  }

  scheduler->socket = cs_socketcan_init();

  if(scheduler->socket == NULL){
    _schedule_error("Failed to init SocketCAN\n");
    return CS_FAIL;
  }

  //Configure SocketCAN
  if(cs_socketcan_config_native(scheduler->socket,iface,mode)){
    _schedule_error("Failed to config socket: %s\n",cs_socketcan_getlasterrorstring(scheduler->socket));
    return CS_FAIL;
  }
  
  //Open socket
  if(cs_socketcan_open(scheduler->socket)){
    _schedule_error("Failed to open socket: %s\n",cs_socketcan_getlasterrorstring(scheduler->socket));
    return CS_FAIL;
  }

  //Set filter for receive messages
  num_id = cs_lists_size(&scheduler->l_recv);
  canid = (uint32_t*)calloc(sizeof(uint32_t),num_id);
  if(canid == NULL){
    _schedule_error("Failed to malloc Id array\n");
    return CS_FAIL;
  }
  
  for(i = 0; i < num_id; i++){
    cs_schedule_message_t *m = (cs_schedule_message_t *)cs_lists_get_iter(&scheduler->l_recv,i);
    canid[i] = m->id;
  }
  //Skip filter to get real busload
  /*
  if(cs_socketcan_set_filter(scheduler->socket,canid,num_id)){
    _schedule_error("Failed to set filter on receive messages: %s\n",cs_socketcan_getlasterrorstring(scheduler->socket));
    free(canid);
    return CS_FAIL;
  }
  */
  free(canid);

  return CS_OK;
}


/**************************************************************/
int cs_schedule_close(cs_schedule_t *scheduler)
/**************************************************************/
{
  if(scheduler == NULL){
    _schedule_error("Parameter failure\n");
    return CS_FAIL;
  }

  cs_socketcan_close(scheduler->socket);

  cs_socketcan_free(scheduler->socket);
  
  return CS_OK;
}


/**************************************************************/
int cs_schedule_checkmessages(cs_schedule_t *scheduler)
/**************************************************************/
{
  int i;
  int rv;

  if(scheduler == NULL){
    _schedule_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  /*Only check for send messages,
   check if every signal in message is configured*/
  for(i = 0; i < cs_lists_size(&scheduler->l_send); i++){
    cs_schedule_message_t *m = (cs_schedule_message_t *)cs_lists_get_iter(&scheduler->l_send,i);
    if((rv = cs_message_check(m->message))){
      return rv;
    }
  }
  return CS_OK;
}


/**************************************************************/
void _cs_schedule_resetstates(cs_schedule_t *scheduler)
/**************************************************************/
{
  int i;
  
  if(scheduler == NULL){
    _schedule_error("Parameter failure\n");
    return;
  }
  
  /*Reset state flags*/
  for(i = 0; i < cs_lists_size(&scheduler->l_recv); i++){
    cs_schedule_message_t *m = (cs_schedule_message_t *)cs_lists_get_iter(&scheduler->l_recv,i);
    m->state = SCHEDULE_STATE_IDLE;  
  }
  
}

/**************************************************************/
void _cs_schedule_idle(cs_schedule_t *scheduler)
/**************************************************************/
{
  int i;
  
  if(scheduler == NULL){
    _schedule_error("Parameter failure\n");
    return;
  }
  
  /*Reset counter*/
  for(i = 0; i < cs_lists_size(&scheduler->l_recv); i++){
    cs_schedule_message_t *m = (cs_schedule_message_t *)cs_lists_get_iter(&scheduler->l_recv,i);
    m->timeout = m->message->timeout;
    if(m->message->timeoutflag != NULL){
      *m->message->timeoutflag = CS_FAIL;
    }
  }
  for(i = 0; i < cs_lists_size(&scheduler->l_send); i++){
    cs_schedule_message_t *m = (cs_schedule_message_t *)cs_lists_get_iter(&scheduler->l_send,i);
    m->counter = m->message->interval;
    /*External buffer*/
    if(m->message->externbuffer != NULL)
      memset(m->message->externbuffer,0,CS_DATA);
  }
}

/** Send CAN frame
    \retval 0 Success
    \retval 1 Failure
    \retval 2 TX-Full, resend in next cycle
*/
/**************************************************************/
int _cs_schedule_sendframe(
			   cs_schedule_t *scheduler,
			   cs_schedule_message_t *m
			   )
/**************************************************************/
{
  uint8_t sendbuffer[CS_DATA];
  int rv;
  cs_message_t *message = m->message;
  
  memset(sendbuffer,0,CS_DATA);
  if(cs_message_runtime_encode(message,sendbuffer,message->dlc,0)){
    /*Encoding failed, TODO Handling*/
    _schedule_error("Encoding failed: ID: %X\n",message->id);
    return CS_FAIL;
  }
  /*External buffer*/
  if(message->externbuffer != NULL)
    memcpy(message->externbuffer,sendbuffer,CS_DATA);
	 
  rv =  cs_socketcan_send(scheduler->socket,message->id,sendbuffer,message->dlc,m->smode);
  switch(rv){
  case CS_SOCKETCAN_FAIL:
    scheduler->send_error_counter++;
    return CS_FAIL;
    break;
  case CS_SOCKETCAN_OK:
    return CS_OK;
    break;
  case CS_SOCKETCAN_TXFULL:
    scheduler->send_tx_full_counter++;
    //resend in next cycle
    return 2;
    break;
  default:
    scheduler->send_error_counter++;
    return CS_FAIL;
  }
  return CS_OK;
}

/**************************************************************/
int _cs_schedule_recieve(cs_schedule_t *scheduler)
/**************************************************************/
{
  uint32_t id;
  uint8_t recvbuffer[CS_DATA];
  int recvlen = CS_DATA;
  cs_message_t *message = NULL;
  cs_schedule_message_t* sched_message = NULL;
  struct timeval tv;

  //Reset msg counter
  scheduler->recv_no_messages = 0;
  
  while(cs_socketcan_recv_nonblock(scheduler->socket,&id,recvbuffer,&recvlen,&tv) == CS_OK) {
    scheduler->recv_no_messages++;
    /*get message pointer*/
    if((sched_message = _schedule_get_message(scheduler,id,MESSAGE_DIR_RECEIVE)) == NULL){
      /*Message unknown, skip*/
      continue;
    }
    message = sched_message->message;
    if(message ==NULL){
      /*Message pointer failure, skip*/
      continue;
    }
    
    /*Reset Timeout*/
    sched_message->timeout = message->timeout;
    /*Set Timestamp*/
    message->ts = tv;
    
    /*Set external buffer*/
    if(message->externbuffer != NULL)
      memcpy(message->externbuffer,recvbuffer,recvlen);
    
    /*Set external ts*/
    if(message->externts != NULL){
      *message->externts = tv;
    }

    /*Set external tvh*/
    if(message->externtvh != NULL){
      *message->externtvh = cs_ts_timeval2d(tv);
    }
    
    sched_message->state = SCHEDULE_STATE_RECIEVED;
    _schedule_verbose("Scheduler Recieve message\n");
    /*Ok message is prepared to recieve*/
    if(cs_message_runtime_decode(message,recvbuffer,recvlen))
      /*Message decoding failed, TODO handling*/
      continue;

    //Set length to original buffer size
    recvlen = CS_DATA;
  }
  return CS_OK;
}

/**************************************************************/
int cs_schedule_schedule(cs_schedule_t *scheduler, cs_scheduler_mode_t mode)
/**************************************************************/
{
  int sendm_counter = 0;
  struct cs_lists_entry *entry = NULL;
  cs_schedule_message_t *m = NULL;
  int rv;
  uint8_t skip = 0;

  if(scheduler == NULL){
    _schedule_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  _cs_schedule_resetstates(scheduler);

  if(mode == SCHEDULE_MODE_IDLE){
    //_schedule_verbose("Mode Idle\n");
    _cs_schedule_idle(scheduler);
    return CS_OK;
  }

  /*First read raw socket*/
  if(mode==SCHEDULE_MODE_RECIEVE){
    //_schedule_verbose("Mode Recieve\n");
    if(_cs_schedule_recieve(scheduler))
      return CS_FAIL;

    for(entry = cs_lists_get_first(&scheduler->l_recv);
	entry != cs_lists_get_last(&scheduler->l_recv);
	entry = cs_lists_get_next(entry)){
      m = (cs_schedule_message_t*)cs_lists_get_data(entry);
      if(m->action == SCHEDULE_MSGACTION_PAUSED){
	_schedule_verbose("Recieve  message is paused\n");
	m->timeout = m->message->timeout;
	if(m->message->timeoutflag != NULL){
	  *m->message->timeoutflag = CS_FAIL;
	}
	continue;
      }
      if(m->timeout <= scheduler->cycle){
	/*Handle timeout*/
	m->timeout = 0;
	if(m->message->timeoutflag != NULL){
	  *m->message->timeoutflag = CS_FAIL;
	}
      } else {
	/*Reduce timeout counter*/
	m->timeout -= scheduler->cycle;
	/*Reset timeout flag*/
	if(m->message->timeoutflag != NULL){
	  *m->message->timeoutflag = CS_OK;
	}
      }
      //_schedule_debug("Timeout: %d\n",scheduler->m_recv[i].timeout);
    }
    return CS_OK;
  }
  
  if(mode == SCHEDULE_MODE_SEND){
    //Reset send counter
    scheduler->send_no_messages = 0;
    //_schedule_verbose("Mode Send\n");
    for(entry = cs_lists_get_first(&scheduler->l_send);
	entry != cs_lists_get_last(&scheduler->l_send);
	entry = cs_lists_get_next(entry)){
      m = (cs_schedule_message_t*)cs_lists_get_data(entry);
      if(m->action == SCHEDULE_MSGACTION_PAUSED){
	_schedule_verbose("Send  message is paused\n");
	m->counter = m->message->interval_offset;// + m->message->interval_startoffset;
	/*External buffer*/
	if(m->message->externbuffer != NULL)
	  memset(m->message->externbuffer,0,CS_DATA);
	continue;
      }
      if(m->counter <= scheduler->cycle){
	/*Prepare to send*/      
	/*Reset counter to interval*/
	m->counter = m->message->interval;
	/*Check if previous messages was send*/
	if(m->state == SCHEDULE_STATE_SEND){
	  _schedule_verbose("Send buffer overflow detected\n");
	} else {
	  m->state = SCHEDULE_STATE_SEND;
	  scheduler->send_no_messages++;
	}
      } else {
	/*reduce counter*/
	m->counter -= scheduler->cycle;
      }
    }
    //Increment send counter by actual number of new messages
    scheduler->send_shifted_messages += scheduler->send_no_messages;
    
    /*Send messages and send max max_messpcycle messages per cycle*/
    for(entry = cs_lists_get_first(&scheduler->l_send);
	entry != cs_lists_get_last(&scheduler->l_send);
	entry = cs_lists_get_next(entry)){
      m = (cs_schedule_message_t*)cs_lists_get_data(entry);
      if(m->state == SCHEDULE_STATE_SEND){
	if(sendm_counter < scheduler->max_messpcycle){
	  rv = _cs_schedule_sendframe(scheduler, m);
	  switch(rv){
	  case CS_FAIL:
	    //Failure, direct signal to user
	    return CS_FAIL;
	    break;
	  case CS_OK:
	    //Well done, reset states
	    m->state = SCHEDULE_STATE_IDLE;
	    sendm_counter++;
	    if(scheduler->send_shifted_messages > 0){
	      scheduler->send_shifted_messages--;
	    }
	    break;
	  case 2:
	    //Receive a TX-FULL
	    //Skip all other messages
	    //Try next cycle again
	    skip = 1;
	    break;
	  }
	  if(skip == 1){
	    //There is a problem with sending messages
	    //break loop
	    break;
	  }
	}
      }
    }
  }

  return CS_OK;
}

/**************************************************************/
int cs_schedule_get_send_errorcounter(
				      cs_schedule_t *scheduler
				      )
/**************************************************************/
{
  
  if(scheduler == NULL){
    _schedule_error("Parameter failure\n");
    return -1;
  }
  return (int)scheduler->send_error_counter;
}

/**************************************************************/
  int cs_schedule_get_send_txfullerrorcounter(
					      cs_schedule_t *scheduler
					      )
/**************************************************************/
{
  if(scheduler == NULL){
    _schedule_error("Parameter failure\n");
    return -1;
  }
  return (int)scheduler->send_tx_full_counter;
}

/**************************************************************/
int cs_schedule_get_send_msgcounter(
				    cs_schedule_t *scheduler
				    )
/**************************************************************/
{
  if(scheduler == NULL){
    _schedule_error("Parameter failure\n");
    return -1;
  }
  return (int)scheduler->send_no_messages;
}

/**************************************************************/
int cs_schedule_get_recv_msgcounter(
				    cs_schedule_t *scheduler /**< Scheduler structure*/
				    )
/**************************************************************/
{
  if(scheduler == NULL){
    _schedule_error("Parameter failure\n");
    return -1;
  }
  return (int)scheduler->recv_no_messages;
}
  
/**************************************************************/
int cs_schedule_get_send_msgshiftedcounter(
					   cs_schedule_t *scheduler
					   )
/**************************************************************/
{
  if(scheduler == NULL){
    _schedule_error("Parameter failure\n");
    return -1;
  }
  return (int)scheduler->send_shifted_messages;
}
