/*  signal.c --  encode and decode CAN messages signals
    Copyright (C) 2013 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */


#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "config.h"
#include "cs_signal_private.h"
#include "cs_signal.h"
#include "cs_debug.h"
#include "cs_debug_private.h"

/** \def BITSINBYTE
Number of Bits in Byte
*/
#define BITSINBYTE 8

/** \def MAXSIGNALBITS
    Maximal allowed bits in one signal
*/
#define MAXSIGNALBITS 63

#define _signal_error(...) (cs_error(CS_DEBUG_SIGNAL,__func__,__LINE__,__VA_ARGS__))
#define _signal_warn(...) (cs_warn(CS_DEBUG_SIGNAL,__func__,__LINE__,__VA_ARGS__))
#define _signal_info(...) (cs_info(CS_DEBUG_SIGNAL,__func__,__LINE__,__VA_ARGS__))
#define _signal_debug(...) (cs_debug(CS_DEBUG_SIGNAL,__func__,__LINE__,__VA_ARGS__))
#define _signal_verbose(...) (cs_verbose(CS_DEBUG_SIGNAL,__func__,__LINE__,__VA_ARGS__))
#define _signal_error_buffer(DATA,LEN) (cs_error_buffer(CS_DEBUG_SIGNAL,__func__,__LINE__,DATA,LEN))
#define _signal_warn_buffer(DATA,LEN) (cs_warn_buffer(CS_DEBUG_SIGNAL,__func__,__LINE__,DATA,LEN))
#define _signal_info_buffer(DATA,LEN) (cs_info_buffer(CS_DEBUG_SIGNAL,__func__,__LINE__,DATA,LEN))
#define _signal_debug_buffer(DATA,LEN) (cs_debug_buffer(CS_DEBUG_SIGNAL,__func__,__LINE__,DATA,LEN))
#define _signal_verbose_buffer(DATA,LEN) (cs_verbose_buffer(CS_DEBUG_SIGNAL,__func__,__LINE__,DATA,LEN))
#define _signal_verbose_binbuffer(DATA,LEN) (cs_verbose_binbuffer(CS_DEBUG_SIGNAL,__func__,__LINE__,DATA,LEN))




/********************************************************************/
cs_debug_level_t cs_signal_debuglevel(
				      cs_debug_level_t level
				      )
/********************************************************************/
{
  return cs_debuglevel(CS_DEBUG_SIGNAL,level);
}

/********************************************************************/
void cs_signal_debug_setcallback(
    void *ctx,
    void (*callb)(void *ctx,
		  cs_debug_sub_t sub,
		  cs_debug_level_t dl,
		  const char *func,
		  const int line,
		  const char *fmt,
		  va_list ap)
    )
/********************************************************************/
{
  cs_debug_setcallback(CS_DEBUG_SIGNAL,ctx,callb);
}

/*
    A CAN signal with position, size, unit and other information. A
    signal is part of a message.

    Signal bit numbering in a message:

    .. code:: text

       Byte:       0        1        2        3
              +--------+--------+--------+--- - -
              |        |        |        |
              +--------+--------+--------+--- - -
       Bit:    7      0 15     8 23    16 31

    Big endian signal with start bit 2 and length 5 (0=LSB, 4=MSB):

    .. code:: text

       Byte:       0        1        2        3
              +--------+--------+--------+--- - -
              |    |432|10|     |        |
              +--------+--------+--------+--- - -
       Bit:    7      0 15     8 23    16 31

    Little endian signal with start bit 2 and length 9 (0=LSB, 8=MSB):

    .. code:: text

       Byte:       0        1        2        3
              +--------+--------+--------+--- - -
              |543210| |    |876|        |
              +--------+--------+--------+--- - -
       Bit:    7      0 15     8 23    16 31

*/


/**************************************************************************/
int cs_signal_check(
		    cs_signal_t *signal, /**< Signal structure*/
		    uint8_t mlength      /**< Message length*/
		    )
/**************************************************************************/
{
  if(signal == NULL){
    return CS_FAIL;
  }

  //We should do further checks in the future
  
  signal->start_offset = signal->offset & 0x07;
  signal->start_byte = signal->offset / BITSINBYTE;
  if(signal->bt == SIGNAL_BT_BIG) { /* big endian */
    signal->end_byte     = signal->start_byte + (7 + signal->length - signal->start_offset - 1)/BITSINBYTE;
    signal->end_offset   = (signal->start_offset - signal->length + 1) & 7;
    /*
      Example
      for offset = 2, bit_len = 5
      => start_offset = (2 & 7) = 2
      => start_byte = ( 2 / 8) = 0
      end_byte:
      0 + (7 + 5 -2 -1) / 8 = 0 + 9 /8 = 1
      end_byte = 1
      end_offset:
      !!Unsigned operation!! 2 - 5  + 1= 65534 (b1111111111111110)
      (b1111111111111110) & 0x7 (b111) = 110 => 6
      end_offset = 6

      for offset = 48, bit_len = 16
      => start_offset = (48 & 7) = 0
      => start_byte = ( 48 / 8) = 6
      end_byte:
      0 + (7 + 5 -2 -1) / 8 = 0 + 9 /8 = 1
      end_byte = 1
      end_offset:
      !!Unsigned operation!! 2 - 5  + 1= 65534 (b1111111111111110)
      (b1111111111111110) & 0x7 (b111) = 110 => 6
      end_offset = 6

     */
  } else {
    /* little endian - similar algorithm with reverse bit significance  */
    signal->end_byte     = signal->start_byte + (signal->length + signal->start_offset - 1)/BITSINBYTE;
    signal->end_offset   = (signal->start_offset + signal->length - 1) & 7;
    /*
      Example
      for offset = 2, bit_len = 9
      => start_offset = (2 & 7) = 2
      => start_byte = ( 2 / 8) = 0
      end_byte:
      0 + (9 +2 -1) / 8 = 0 + 10 /8 = 1
      end_byte = 1
      end_offet:
      end_offset = (2 + 9 -1) & 7 = 10 (b1010) & 7 (b111) = b010 = 2 
     */
  }

  /*
    Precalculation of signed bit for SIGNED signals 
   */
  signal->signbit= (int64_t) (1L << ((int64_t) signal->length - 1L));

    
  _signal_verbose("Startbyte: %i, Startoffset: %i, Endbyte: %i, Endoffset: %i\n",
		  signal->start_byte,
		  signal->start_offset,
		  signal->end_byte,
		  signal->end_offset);

  if(mlength > 0){
    //We can do a message length based check
    if((signal->start_byte >= mlength) || (signal->end_byte >= mlength)){
      _signal_error("Signal %s data out of message length\n",signal->name);
      return CS_FAIL;
    }
  }

  
  return CS_OK;
}
  
/**************************************************************************/
int _extract_data(int64_t *rvalue,
		  uint16_t start_byte,
		  uint16_t start_offset,
		  uint16_t end_byte,
		  uint16_t end_offset,
		  cs_signal_byteorder_t bt,
		  uint8_t *buffer,
		  uint8_t len)
/**************************************************************************/
{
  /*
    Note this function is prepared to handle signals based on DBC based
    offset and byteorder
    That means Startbit is LSB0 for little and  MSB0 for big endianess
    Bit length is oriented from start bit to the right higher byte
    This function is based on http://cantools.sf.net/
   */
  int  work_byte;
  uint8_t  data;
  int  shift = 0;
  *rvalue = 0;

  if((start_byte < 0) ||
     (start_byte >= len) ||
     (end_byte < 0) ||
     (end_byte >= len)){
    _signal_error("Byte index out of range Start: %i End: %i Len: %i\n",
		  start_byte,
		  end_byte,
		  len);
    return CS_FAIL;
  }
  
  /* 0 = Big Endian, 1 = Little Endian */    
  if(bt == SIGNAL_BT_BIG) { /* big endian */
    _signal_verbose("Extract Big: Startbyte: %i, Startoffset: %i, Endbyte: %i, Endoffset: %i\n",start_byte,start_offset,end_byte,end_offset);
    /*
      Startbit|            | diff = bit_len     
            MSB|   |    |  |LSB
    start_byte|    |    |   |end_byte   
  start_offset||           ||end_offset
     */

    /* 
       loop over all source bytes from start_byte to end_byte 
       so MSB is taken first
    */    
    for(work_byte = start_byte; work_byte <= end_byte; work_byte++) {
      /* fetch source byte */
      data = buffer[work_byte];
      _signal_verbose("Extract Big: work_byte: %i, Raw Data: %X\n",work_byte,data);
      /* process source byte */
      if(work_byte == start_byte && start_offset != 7) {
	/* less that 8 bits in start byte? mask out unused bits */
	/* 
	   example start_offset = 2
	   mask = b11111111 >> 5 => b111
	*/
	data &= (uint8_t)~0 >> (7 - start_offset);
	shift = start_offset + 1;
      } else {
	shift = BITSINBYTE; /* use all eight bits */
      }
      if(work_byte == end_byte && end_offset != 0) {
	/* less that 8 bits in end byte? shift out unused bits */
	/*
	  example end_offset = 6
	*/
	data >>= end_offset;
	shift -= end_offset;
      }
      _signal_verbose("Extract Big: Byte: %i, Data shifted/masked: %X,Shift: %i\n",work_byte,data,shift);
      /* store processed byte */
      *rvalue <<= shift; /* make room for shift bits */
      *rvalue |= data;   /* insert new bits at low position */
      _signal_verbose("Extract Big: Value: %lli\n",*rvalue);
    }
  } else {
    _signal_verbose("Extract Little: Startbyte: %i, Startoffset: %i, Endbyte: %i, Endoffset: %i\n",start_byte,start_offset,end_byte,end_offset);
    /*
       Startbit|           | diff = bit_len     
            LSB|   |    |  |MSB
    start_byte|    |    |   |end_byte   
    Bit number7   07   07  0
  start_offset |   |       ||end_offset
     */
    
    /* 
       loop over all source bytes from end_byte to start_byte 
       so MSB is taken first
    */    
    for(work_byte = end_byte; work_byte >= start_byte; work_byte--) {
      data = buffer[work_byte];
      if(work_byte == end_byte && end_offset != 7) {
	/*
	  example end_offset = 2
	  mask = b11111111 >> (7-2 = 5) => b111
	*/
	data &= (uint8_t)~0 >> (7 - end_offset);
	shift = end_offset + 1;
      } else {
	shift = BITSINBYTE;
      }
      if(work_byte == start_byte && start_offset != 0) {
	data >>= start_offset;
	shift -= start_offset;
      }
      _signal_verbose("Extract Little: Byte: %i, Data: %X,Shift: %i\n",work_byte,data,shift);
      *rvalue <<= shift;
      *rvalue |= data;
    }
  }
  _signal_verbose("Extract: Final Value %lli\n",*rvalue);
  return CS_OK;
}

/**************************************************************************/
int _pack_data(int64_t value,
	       uint16_t start_byte,
	       uint16_t start_offset,
	       uint16_t end_byte,
	       uint16_t end_offset,
	       cs_signal_byteorder_t bt,
	       uint8_t *buffer,
	       uint8_t len)
/**************************************************************************/
{
  /*
    Note this function is prepared to handle signals based on DBC based
    offset and byteorder
    That means Startbit is LSB0 for little and MSB0 for big endianess
    Bit length is oriented from start bit to the right higher byte
  */
  uint8_t data = 0;
  int work_byte;
  int shift = 0;
  uint8_t byte_shift;
  uint8_t mask;


  if((start_byte < 0) ||
     (start_byte >= len) ||
     (end_byte < 0) ||
     (end_byte >= len)){
    _signal_error("Byte index out of range Start: %i End: %i Len: %i\n",
		  start_byte,
		  end_byte,
		  len);
    return CS_FAIL;
  }

  _signal_verbose("Signal pack buffer before pack\n");
  _signal_verbose_binbuffer(buffer,len);
  
  /* 0 = Big Endian, 1 = Little Endian */    
  if(bt == SIGNAL_BT_BIG) { /* big endian */
    _signal_verbose("Pack Big: Startbyte: %i, Startoffset: %i, Endbyte: %i, Endoffset: %i\n",start_byte,start_offset,end_byte,end_offset);
    _signal_verbose("Pack Big: Value start: %lli\n",value);
    
    /*
      Startbit|            | diff = bit_len     
            MSB|   |    |  |LSB
    start_byte|    |    |   |end_byte   
  start_offset||           ||end_offset
     */
    
    /* loop over all source bytes from end_byte to start_byte
       so start with LSB bit
     */
    for(work_byte = end_byte; work_byte >= start_byte; work_byte--) {
      _signal_verbose("Pack Big: work_byte: %i, Start Raw Data: %X\n",work_byte,buffer[work_byte]);
      /*Shift raw value*/
      /*Note first time shift is 0*/
      value >>= shift;
      _signal_verbose("Pack Big: Value shifted: %lli,Shift: %i\n",value,shift);
      if(work_byte == start_byte && start_offset != 7) {
	shift = start_offset +1;
	byte_shift = 0;
	mask = 0xFF >> (BITSINBYTE - start_offset  -1);
	/*
	  example start_offset = 2
	  shift = 2+1 = 3
	  byte_shift = 0
	  mask = b11111111 
	*/
      } else {
	shift = BITSINBYTE;  
	byte_shift = 0;
	mask = 0xFF;
      }
      if(work_byte == end_byte && end_offset != 0) {
	/* less that 8 bits in end byte? shift out unused bits */
	shift = BITSINBYTE - end_offset;
	byte_shift = end_offset;
	mask = mask  & (0xFF << end_offset);
	/*
	  example end_offset = 6
	  shift = 8 -6 = 2
	  byte_shift = 6
	  Note keep in mind , the mask is for short signals
	  already reduces by start offset,
	  the example shown is for long signals
	  mask = b11111111 >> 6 => b11000000
	*/
      }
      _signal_verbose("Pack Big: Byte: %i, Data: %X, Mask: %X, Shift: %i, Byte Shift: %i\n",work_byte,data,mask,shift,byte_shift);
      _signal_verbose("Pack Big: Buffer Byte before mask bits: 0x%X\n",buffer[work_byte]);
      _signal_verbose("Pack Big: Bit AND argument: %X\n",(uint8_t)(~mask));
      /*Delete bits in buffer based on mask*/
      buffer[work_byte] &= (uint8_t)(~mask);
      _signal_verbose("Pack Big: Buffer Byte after mask bits: 0x%X\n",buffer[work_byte]);
      
      /*Get raw data*/
      data = ((uint8_t)(value & 0xFF)) << byte_shift;

      buffer[work_byte] |= data;
      _signal_verbose("Pack Big: work_byte: %i, End Raw Data: %X\n",work_byte,buffer[work_byte]);
    }
  } else {
    /* little endian - similar algorithm with reverse bit significance  */
    _signal_verbose("Pack Little: Startbyte: %i, Startoffset: %i, Endbyte: %i, Endoffset: %i\n",start_byte,start_offset,end_byte,end_offset);
    _signal_verbose("Pack Little: Input value %lli\n",value);    
    if((start_byte < 0) || (start_byte >= len) || (end_byte < 0) || (end_byte >= len)){
      _signal_error("Little Endian: Byte index out of range Start: %i End: %i Len: %i\n",start_byte,end_byte,len);
      return CS_FAIL;
    }

    /*
      Startbit|            | diff = bit_len     
            LSB|   |    |  |MSB
    start_byte|    |    |   |end_byte   
  start_offset||           ||end_offset
    */

    /* loop over all source bytes from start_byte to end_byte
       so start with LSB bit
     */
    for(work_byte = start_byte; work_byte <= end_byte; work_byte++) {
      /*Shift raw value*/
      /*Note shift starts with 0*/
      value >>= shift;
      _signal_verbose("Pack Little: Value shifted: %lli,Shift: %i\n",value,shift);
      if(work_byte == end_byte && end_offset != 7) {
	_signal_verbose("Pack Little: Work byte == End byte offset != 7\n");
	shift = end_offset +1;
	byte_shift = 0;
	mask = 0xFF >> (BITSINBYTE - (end_offset + 1));
	/*
	  example end_offset = 2
	  shift = 2+1 = 3
	  byte_shift = 0
	  mask = 0xFF >> ((8-(2+1)) = 5) => b111 
	*/
      } else {
       	shift = BITSINBYTE;  
	byte_shift = 0;
	mask = 0xFF;
      }
      if(work_byte == start_byte && start_offset != 0) {
	_signal_verbose("Pack Little: Work byte == Start byte offset != 0\n");
	shift = BITSINBYTE - start_offset;
	byte_shift = start_offset;
	mask = mask & (0xFF << start_offset);
	/*
	  example start_offset = 2
	  shift = 8 -2 = 6
	  byte_shift = 2
	  Note keep in mind , the mask is for short signals
	  already reduces by end offset,
	  the example shown is for long signals
	  mask = 0xFF >> 2 => b111111
	*/
      }

      
      _signal_verbose("Pack Little: Byte: %i, Data: %X, Mask: %X, Shift: %i, Byte Shift: %i\n",work_byte,data,mask,shift,byte_shift);
      _signal_verbose("Pack Little: Buffer Byte before mask bits: 0x%X\n",buffer[work_byte]);
      /*Delete bits in buffer based on mask*/
      _signal_verbose("Pack Little: Bit AND argument: %X\n",(uint8_t)(~mask));
      buffer[work_byte] &=(uint8_t)(~mask);
      _signal_verbose("Pack Little: Buffer Byte after mask bits: 0x%X\n",buffer[work_byte]);
      
      /*Get raw data*/
      data = ((uint8_t)(value & 0xFF)) << byte_shift;
      _signal_verbose("Pack Little: Data before OR: %X\n",data);
      
      buffer[work_byte] |= data;
      _signal_verbose("Pack Little: work_byte: %i, End Raw Data: %X\n",work_byte,buffer[work_byte]);
    }
  }
  _signal_verbose("Signal pack buffer after pack\n");
  _signal_verbose_binbuffer(buffer,len);
  return CS_OK;
}

/** Copy label content for list handling
 */
/**************************************************************************/
void *_cs_signal_cp_label( void *ext)
/**************************************************************************/
{
  void *t = NULL;

  t = malloc(sizeof(cs_signal_label_t));

  if(t == NULL)
    return NULL;

  memcpy(t,ext,sizeof(cs_signal_label_t));

  return t;
}

/** Delete label for list handling
 */
/**************************************************************************/
void _cs_signal_del_label( void *intern)
/**************************************************************************/
{
  if(intern != NULL)
    free(intern);
}

/** Copy node content for list handling
 */
/**************************************************************************/
void *_cs_signal_cp_node( void *ext)
/**************************************************************************/
{
  void *t = NULL;

  t = malloc(sizeof(cs_signal_node_t));

  if(t == NULL)
    return NULL;

  memcpy(t,ext,sizeof(cs_signal_node_t));

  return t;
}

/** Delete node for list handling
 */
/**************************************************************************/
void _cs_signal_del_node( void *intern)
/**************************************************************************/
{
  if(intern != NULL)
    free(intern);
}

/**************************************************************************/
cs_signal_t *cs_signal_cp(cs_signal_t *ext)
/**************************************************************************/
{
  cs_signal_t *t = NULL;
  struct cs_lists_entry *entry = NULL;
  cs_signal_label_t *l = NULL;
  cs_signal_node_t *n = NULL;

  if(ext == NULL){
    _signal_error("Parameter failure\n");
    return NULL;
  }
  t = cs_signal_init();
  
  if(t == NULL){
    _signal_error("Signal init failure\n");
    return NULL;
  }
  
  if(cs_signal_reset(t)||
     cs_signal_set_signal_type(t,ext->type)||
     cs_signal_set_byteorder(t,ext->bt)||
     cs_signal_set_bitoffset(t,ext->offset)||
     cs_signal_set_bitlength(t,ext->length)||
     cs_signal_set_name(t,ext->name)||
     cs_signal_set_unit(t,ext->unit)||
     cs_signal_set_note(t,ext->notes)||
     cs_signal_set_scale(t,ext->dscale)||
     cs_signal_set_offset(t,ext->doffset)||
     cs_signal_set_minmax(t,ext->min,ext->max)
     ){
    _signal_error("Configure failure\n");
    free(t);
    return NULL;
  }
  //Label list
  for(entry = cs_lists_get_first(&ext->labels);
      entry != cs_lists_get_last(&ext->labels);
      entry = cs_lists_get_next(entry)){
    l = (cs_signal_label_t*)cs_lists_get_data(entry);
    cs_signal_add_label(t,l->label,l->value);
  }
  //Node list
  for(entry = cs_lists_get_first(&ext->reqnodes);
      entry != cs_lists_get_last(&ext->reqnodes);
      entry = cs_lists_get_next(entry)){
    n = (cs_signal_node_t*)cs_lists_get_data(entry);
    cs_signal_add_node(t,n->name);
  }
  return t;
}

/**************************************************************************/
cs_signal_t *cs_signal_init()
/**************************************************************************/
{
  cs_signal_t * s = NULL;

  s = (cs_signal_t*)calloc(sizeof(cs_signal_t),1);
  
  if(s == NULL){
    _signal_error("Malloc failure\n");
    return NULL;
  }
  cs_lists_init(&s->labels,_cs_signal_cp_label,_cs_signal_del_label);
  cs_lists_init(&s->reqnodes,_cs_signal_cp_node,_cs_signal_del_node);
  cs_signal_reset(s);

  return s;
}

/**************************************************************************/
void cs_signal_free(cs_signal_t *signal)
/**************************************************************************/
{
  if(signal == NULL)
    return;

  cs_lists_destroy(&signal->labels);
  cs_lists_destroy(&signal->reqnodes);

  free(signal);
  signal = NULL;
  
}

/**************************************************************************/
char* cs_signal_get_name(cs_signal_t *signal)
/**************************************************************************/
{
  if(signal == NULL)
    return NULL;

  return signal->name;
}

/**************************************************************************/
int cs_signal_set_name(cs_signal_t *signal, char * name)
/**************************************************************************/
{
  if((signal == NULL) || (name == NULL)){
    return CS_FAIL;
  }
  
  strncpy(signal->name,name,CS_CHARBUF);
  return CS_OK;
}

/**************************************************************************/
char* cs_signal_get_unit(cs_signal_t *signal)
/**************************************************************************/
{
  if(signal == NULL)
    return NULL;

  return signal->unit;
}

/**************************************************************************/
int cs_signal_set_unit(cs_signal_t *signal, char * unit)
/**************************************************************************/
{
  if((signal == NULL) || (unit == NULL)){
    return CS_FAIL;
  }
  
  strncpy(signal->unit,unit,CS_CHARBUF);
  return CS_OK;
}

/**************************************************************************/
char* cs_signal_get_note(cs_signal_t *signal)
/**************************************************************************/
{
  if(signal == NULL)
    return NULL;

  return signal->notes;
}

/**************************************************************************/
int cs_signal_set_note(cs_signal_t *signal, char * note)
/**************************************************************************/
{
  if((signal == NULL) || (note == NULL)){
    return CS_FAIL;
  }
  
  strncpy(signal->notes,note,CS_CHARBUF);
  return CS_OK;
}

/**************************************************************************/
uint16_t cs_signal_get_bitlength(cs_signal_t *signal)
/**************************************************************************/
{
  if(signal == NULL)
    return 0;

  return signal->length;
}

/**************************************************************************/
int cs_signal_set_bitlength(cs_signal_t *signal, uint16_t bitlength)
/**************************************************************************/
{ 
  if((signal == NULL) || (bitlength == 0)){
    return CS_FAIL;
  }
  signal->length = bitlength;

  //Check for signal length > MAXSIGNALBITS
  if(signal->length >  MAXSIGNALBITS){
    _signal_error("Signal %s:  length > %d Bits is not supported \n",signal->name,MAXSIGNALBITS);
  }
  
  return CS_OK;
}

/**************************************************************************/
uint16_t cs_signal_get_bitoffset(cs_signal_t *signal)
/**************************************************************************/
{
  if(signal == NULL)
    return 0;

  return signal->offset;
}

/**************************************************************************/
int cs_signal_set_bitoffset(cs_signal_t *signal, uint16_t bitoffset)
/**************************************************************************/
{
  if(signal == NULL){
    return CS_FAIL;
  }
  signal->offset = bitoffset;
  return CS_OK;
}

/**************************************************************************/
double cs_signal_get_scale(cs_signal_t *signal)
/**************************************************************************/
{
  if(signal == NULL)
    return 0.0;

  return signal->dscale;
}

/**************************************************************************/
int cs_signal_set_scale(cs_signal_t *signal, double scale)
/**************************************************************************/
{
  if(signal == NULL){
    return CS_FAIL;
  }
  signal->dscale = scale;
  return CS_OK;
}

/**************************************************************************/
double cs_signal_get_offset(cs_signal_t *signal)
/**************************************************************************/
{
  if(signal == NULL)
    return 0.0;

  return signal->doffset;
}

/**************************************************************************/
int cs_signal_set_offset(cs_signal_t *signal, double offset)
/**************************************************************************/
{
  if(signal == NULL){
    return CS_FAIL;
  }
  signal->doffset = offset;
  return CS_OK;
}

/**************************************************************************/
double cs_signal_get_min(cs_signal_t *signal)
/**************************************************************************/
{
  if(signal == NULL)
    return 0.0;

  return signal->min;
}

/**************************************************************************/
double cs_signal_get_max(cs_signal_t *signal)
/**************************************************************************/
{
  if(signal == NULL)
    return 0.0;

  return signal->max;
}

/**************************************************************************/
int cs_signal_set_minmax(cs_signal_t *signal, double min, double max)
/**************************************************************************/
{
  if((signal == NULL)||(min > max)){
    return CS_FAIL;
  }
  signal->min = min;
  signal->max = max;
  return CS_OK;
}

/**************************************************************************/
cs_signal_type_t cs_signal_get_signal_type(cs_signal_t *signal)
/**************************************************************************/
{
  if(signal == NULL)
    return SIGNAL_TYPE_SIGNED;

  return signal->type;
}

/**************************************************************************/
int cs_signal_set_signal_type(cs_signal_t *signal, cs_signal_type_t stype)
/**************************************************************************/
{
  if(signal == NULL){
    return CS_FAIL;
  }
  signal->type = stype;
  return CS_OK;
}

/**************************************************************************/
cs_signal_byteorder_t cs_signal_get_byteorder(cs_signal_t *signal)
/**************************************************************************/
{
  if(signal == NULL)
    return SIGNAL_BT_BIG;

  return signal->bt;
}

/**************************************************************************/
int cs_signal_set_byteorder(cs_signal_t *signal, cs_signal_byteorder_t bt)
/**************************************************************************/
{
  if(signal == NULL){
    return CS_FAIL;
  }
  signal->bt = bt;
  return CS_OK;
}

/**************************************************************************/
cs_signal_mux_t cs_signal_get_mux_type(cs_signal_t *signal)
/**************************************************************************/
{
  if(signal == NULL)
    return SIGNAL_MUX_STATIC;

  return signal->mtype;
}

/**************************************************************************/
int cs_signal_set_mux_type(cs_signal_t *signal, cs_signal_mux_t mux)
/**************************************************************************/
{
  if(signal == NULL){
    return CS_FAIL;
  }
  signal->mtype = mux;
  return CS_OK;
}

/**************************************************************************/
int cs_signal_set_startrawvalue(
				cs_signal_t *signal,
				int64_t rawstartvalue
				)
/**************************************************************************/
{
  if(signal == NULL){
    return CS_FAIL;
  }
  signal->startvalue = rawstartvalue * signal->dscale + signal->doffset;
  return CS_OK;
}

/**************************************************************************/
int cs_signal_set_startvalue(
			     cs_signal_t *signal,
			     double startvalue
			     )
/**************************************************************************/
{
  if(signal == NULL){
    return CS_FAIL;
  }
  signal->startvalue = startvalue;
  return CS_OK;
}

/**************************************************************************/
double cs_signal_get_startvalue(
				   cs_signal_t *signal
				   )
/**************************************************************************/
{
  if(signal == NULL){
    return 0;
  }
  return signal->startvalue;
}
  
/**************************************************************************/
int cs_signal_set(
		  cs_signal_t *signal,   /**< Signal structure*/
		  char *name,            /**< Signal name*/
		  char *unit,            /**< Signal unit*/
		  uint16_t bitoffset,    /**< Bit offset in message*/
		  uint16_t bitlength,    /**< Bit length*/
		  cs_signal_byteorder_t bt, /**< Signal byte order*/
		  cs_signal_mux_t mux,  /**< Signal max type*/
		  cs_signal_type_t stype, /**< Signal signedness type*/
		  double scale,          /**< Signal scale*/
		  double offset,         /**< Signal offset*/
		  double min,            /**< Signal min value*/
		  double max             /**< Signal max value*/
		  )
/**************************************************************************/
{
  if(signal == NULL)
    return CS_FAIL;

  if(cs_signal_set_name(signal,name)){
    return CS_FAIL;
  }
  
  if(
     cs_signal_set_unit(signal,unit) ||
     cs_signal_set_bitoffset(signal,bitoffset) ||
     cs_signal_set_bitlength(signal,bitlength) ||
     cs_signal_set_byteorder(signal,bt) ||
     cs_signal_set_mux_type(signal,mux) ||
     cs_signal_set_signal_type(signal,stype) ||
     cs_signal_set_scale(signal,scale) ||
     cs_signal_set_offset(signal,offset) ||
     cs_signal_set_minmax(signal,min,max)
     ){
    return CS_FAIL;
  }
  
  return CS_OK;
}

/**************************************************************************/
int cs_signal_reset(cs_signal_t *signal)
/**************************************************************************/
{
  if(signal == NULL)
    return CS_FAIL;

  signal->type = SIGNAL_TYPE_SIGNED;
  signal->bt = SIGNAL_BT_LITTLE;
  signal->offset = 0;
  signal->length = 1;
  signal->mtype = SIGNAL_MUX_STATIC;
  signal->startvalue = 0.0; // Default
  signal->dscale = 1.0;
  signal->doffset = 0.0;
  signal->min = 0.0;
  signal->max = 1.0;
  signal->calc = NULL;
  signal->check = NULL;
  signal->check_flag = NULL;
  signal->crc = NULL;
  signal->crc_flag = NULL;
  signal->value = NULL;
  signal->fault = NULL;
  signal->label = NULL;
  memset(signal->buffer_rawvalue,0,CS_SIGNAL_MAXDLC);
  memset(signal->name,0,CS_CHARBUF);
  memset(signal->unit,0,CS_CHARBUF);
  memset(signal->notes,0,CS_CHARBUF);
  //destroy  and init again 
  cs_lists_destroy(&signal->labels);
  cs_lists_destroy(&signal->reqnodes);
  cs_lists_init(&signal->labels,_cs_signal_cp_label,_cs_signal_del_label);
  cs_lists_init(&signal->reqnodes,_cs_signal_cp_node,_cs_signal_del_node);
  return CS_OK;
}

/**************************************************************************/
int cs_signal_set_rvalue(cs_signal_t *signal,
			 double *value,
			 char *label,
			 uint8_t *fault)
/**************************************************************************/
{
  if(signal == NULL)
    return CS_FAIL;

  signal->value = value;
  signal->label = label;
  signal->fault = fault;

  return CS_OK;
}


/**************************************************************************/
int cs_signal_add_label(cs_signal_t *signal,
		       char *label,
		       double value)
/**************************************************************************/
{
  
  cs_signal_label_t l;

  if(signal == NULL)
    return CS_FAIL;
  
  if(label == NULL)
    return CS_FAIL;

  if(strlen(label) >= CS_CHARBUF)
    return CS_FAIL;

  strncpy(l.label,label,CS_CHARBUF);
  l.value = value;

  cs_lists_insert(&signal->labels,(void*)&l);

  return CS_OK;
}

/**************************************************************************/
int cs_signal_add_node(
		       cs_signal_t *signal,
		       char *name
		       )
/**************************************************************************/
{
  
  cs_signal_node_t l;

  if(signal == NULL)
    return CS_FAIL;
  
  if(name == NULL)
    return CS_FAIL;

  if(strlen(name) >= CS_CHARBUF)
    return CS_FAIL;

  strncpy(l.name,name,CS_CHARBUF);

  cs_lists_insert(&signal->reqnodes,(void*)&l);

  return CS_OK;
}

/**************************************************************************/
int cs_signal_runtime_decode(cs_signal_t *signal,
			     double *rvalue, 
			     uint8_t *fault,
			     uint8_t *buffer, 
			     int len)
/**************************************************************************/
{
  int64_t signedRawValue = 0;

  if(signal == NULL)
    return CS_FAIL;

  if(rvalue == NULL)
    return CS_FAIL;

  if(buffer == NULL)
    return CS_FAIL;

  if(_extract_data(&signal->rawvalue,
		   signal->start_byte,signal->start_offset,
		   signal->end_byte,signal->end_offset,
		   signal->bt,
		   buffer, (uint8_t)len))
    return CS_FAIL;

  switch(signal->type) {
  case SIGNAL_TYPE_SIGNED:
    if(signal->length > MAXSIGNALBITS){
      //We can not handle this, internal 64 bit data width
      *rvalue = 0;
      break;
    }
    /*
      Note highest bit for signed representation must not be used and
      the maximal signal length must not be higher than 63 bit, this will 
      collide with the raw value signal variable type (sint64)
      Calculation:
      Set a variable (signal->signbit, precalculated) with the highest bit set in bitfield.
      Remove the highest bit in bitfield via XOR, this will remove the highest bit for negative values
      For positive values (without highest bit set) the value of the signbit is added. 
      Reduce the result by signbit value, for positive values get back the original value,
      for negative values the result is a negative one.
      Examples:

      Positive value
      Signal(16 bit, scale 0.1, offset 0.0)
      signal->rawvalue = 1556 (positive)
      signal->signbit = 32768
      XOR = 34324 (bigger value)
      XOR - signbit = 1556
      Scaled = 155.6

      Negative value
      Signal(16 bit, scale 0.1, offset 0.0)
      signal->rawvalue = 64595 (negative, highest bit set)
      signal->signbit = 32768
      XOR = 31827 (smaller value)
      XOR - signbit = -941
      Scaled = -94.1
     */
    _signal_verbose("Signed bit value: %lli \n",signal->signbit);
    _signal_verbose("Signed  XOR: %lli \n",(signal->rawvalue ^ signal->signbit));
    signedRawValue = (signal->rawvalue ^ signal->signbit) - signal->signbit;
    _signal_verbose("Signed Raw Value: %lli \n",signedRawValue);
    *rvalue = (double) signedRawValue * signal->dscale + signal->doffset;
    break;
  case SIGNAL_TYPE_UNSIGNED:
    if(signal->length > MAXSIGNALBITS){
      //We can not handle this, internal 64 bit data width
      *rvalue = 0;
      break;
    }
    *rvalue = (double)signal->rawvalue * signal->dscale +  signal->doffset;
    break;
  case SIGNAL_TYPE_SINGLE:
    _signal_error("Type SINGLE is not implemented\n");
    *rvalue = 0.0;
    return CS_FAIL;
    /*rvalue = (double)((float)signal->rawvalue * (float)signal->dscale + (float)signal->doffset);*/
    break;
  case SIGNAL_TYPE_DOUBLE:
    _signal_error("Type DOUBLE is not implemented\n");
    *rvalue = 0.0;
    return CS_FAIL;
    /**rvalue = (double)signal->rawvalue * signal->dscale + signal->doffset;*/
    break;
  }

  /*Check for overrange*/
  if(fault != NULL ){ 
    if((*rvalue > signal->max)||(*rvalue < signal->min)){
      *fault = 1;
    } else {
      *fault = 0;
    }
  }

  _signal_verbose("Final Value float: %f\n",*rvalue);
  return CS_OK;
}

/**************************************************************************/
int cs_signal_runtime_staticencode(cs_signal_t *signal,
				   uint8_t *buffer, 
				   int len)
/**************************************************************************/
{
  int i;

  if(signal == NULL)
    return CS_FAIL;
  
  if(buffer == NULL)
    return CS_FAIL;

  if(len > CS_SIGNAL_MAXDLC){
    _signal_error("Signal buffer length overrange detected\n");
    return CS_FAIL;
  }
  _signal_verbose("Static encode for signal %s, signal buffer:\n",signal->name);
  _signal_verbose_binbuffer(signal->buffer_rawvalue,len);
  /*Bit or with precalced value*/
  for(i = 0;i < len;i++){
    buffer[i] |= signal->buffer_rawvalue[i];
  }
  _signal_verbose("After merge:\n");
  _signal_verbose_binbuffer(buffer,len);
  
  return CS_OK;
}

/**************************************************************************/
int cs_signal_runtime_encode(cs_signal_t *signal,
			     double value, 
			     uint8_t *buffer, 
			     int len,
			     int init)
/**************************************************************************/
{
  int64_t signedRawValue = 0;
  double rawvalue = 0;

  if(signal == NULL)
    return CS_FAIL;

  if(buffer == NULL)
    return CS_FAIL;

  if((init == 1 ) && (len > CS_SIGNAL_MAXDLC)){
    _signal_error("Signal buffer length overrange detected\n");
    return CS_FAIL;
  }
  
  if((init == 1) && (signal->mtype != SIGNAL_MUX_STATIC)){
    _signal_error("Signal encode init call on non static signal\n");
    return CS_OK;
  }
  
  if(signal->mtype != SIGNAL_MUX_RAW){

    rawvalue = value;
    
    /*Limit value*/
    if(rawvalue > signal->max)
      rawvalue = signal->max;
    
    if(rawvalue < signal->min)
      rawvalue = signal->min;
    
    signal->rawvalue = 0; 
    
    if(!DEQUALS_ZERO(signal->dscale)){
      switch(signal->type) {
      case SIGNAL_TYPE_SIGNED:
	if(signal->length > MAXSIGNALBITS){
	  //We can not handle this, internal 64 bit data width
	  signal->rawvalue = 0;
	  break;
	}
	/*
	  Note highest bit for signed representation must not be used and
	  the maximal signal length must not be higher than 63 bit, this will 
	  collide with the raw value signal variable type (sint64)
	  Calculation:
	  First scaling and convert to int64.
	  After that add signbit value and a bit-xor with the signbit
	  
	  Examples:
	  
	  Positive value
	  Signal(16 bit, scale 0.1, offset 0.0)
	  Value = 155.6
	  Uncaled = 1556
	  Unscaled + signbit = 34324
	  XOR = 1556
	  signal->rawvalue = 1556 (positive)
	  
	  Negative value
	  Signal(16 bit, scale 0.1, offset 0.0)
	  Value = -42.4
	  Unscaled -424
	  Unscaled + signbit = 32343
	  XOR = 65111
	  signal->rawvalue = 65111 (negative)
	*/
	signedRawValue = (int64_t) floor((rawvalue - signal->doffset) / signal->dscale);
	_signal_verbose("Signed bit value: %lli \n",signal->signbit);
	_signal_verbose("Raw value: %lli \n",signedRawValue);
	signal->rawvalue = (signedRawValue + signal->signbit) ^ signal->signbit;
	break;
      case SIGNAL_TYPE_UNSIGNED:
	if(signal->length > MAXSIGNALBITS){
	  //We can not handle this, internal 64 bit data width
	  signal->rawvalue = 0;
	  break;
	}
	signal->rawvalue = (int64_t) floor((rawvalue - signal->doffset) / signal->dscale);
	break;
      case SIGNAL_TYPE_SINGLE:
	/*Not tested*/
	/*signal->rawvalue = (int64_t) floorf(((float)(rawvalue) - (float)signal->doffset) / (float)signal->dscale);*/
	signal->rawvalue = 0;
	_signal_error("Type FLOAT is not implemented\n");
	return CS_FAIL;
	break;
      case SIGNAL_TYPE_DOUBLE:
	/*Not tested*/
	/*signal->rawvalue = (int64_t) floor((rawvalue - signal->doffset) / signal->dscale);*/
	signal->rawvalue = 0;
	_signal_error("Type DOUBLE is not implemented\n");
	return CS_FAIL;
	break;
      } 
    } else {
      signal->rawvalue = 0;
    }
  }

  _signal_verbose("Signal %s: %f, Raw value before pack %lli\n",signal->name,rawvalue,signal->rawvalue);
  
  if(_pack_data(signal->rawvalue,
		signal->start_byte,signal->start_offset,
		signal->end_byte,signal->end_offset,
		signal->bt,
		buffer,
		(uint8_t)len)){
    _signal_error("Signal encode: Packing failed\n");
    return CS_FAIL;
  }
  
  if(init == 1){
    _signal_verbose("Signal encode: Init Packing call\n");
    memset(signal->buffer_rawvalue,0,CS_SIGNAL_MAXDLC);
    if(_pack_data(signal->rawvalue,
		  signal->start_byte,signal->start_offset,
		  signal->end_byte,signal->end_offset,
		  signal->bt,
		  signal->buffer_rawvalue,
		  (uint8_t)len)){
      _signal_error("Signal encode: Init Packing failed\n");
      return CS_FAIL;
    }
    _signal_verbose_binbuffer(signal->buffer_rawvalue,len);
  }

  return CS_OK;
}


/**************************************************************************/
int cs_signal_runtime_decodelabel(cs_signal_t *signal,
				  double value,
				  uint8_t *fault,
				  char *label)
/**************************************************************************/
{  
  struct cs_lists_entry *entry = NULL;
  cs_signal_label_t *l = NULL;
  double lvalue;

  if(signal == NULL)
    return CS_FAIL;


  if(label == NULL)
    return CS_FAIL;

  memset(label,0,CS_CHARBUF);

  if((fault!=NULL)&&(*fault > 0)){
    lvalue = (double) signal->rawvalue;
  } else {
    lvalue = value;
  }
  
  for(entry = cs_lists_get_first(&signal->labels);
      entry != cs_lists_get_last(&signal->labels);
      entry = cs_lists_get_next(entry)){
    l = (cs_signal_label_t*)cs_lists_get_data(entry);
    if(DEQUALS_ZERO(l->value - lvalue)){
      strncpy(label,l->label,CS_CHARBUF);
      return CS_OK;
    }
  }

  return CS_FAIL;
}

/**************************************************************************/
int cs_signal_runtime_getlabelvalue(cs_signal_t *signal,
				    double *value, 
				    char *label)
/**************************************************************************/
{
  struct cs_lists_entry *entry = NULL;
  cs_signal_label_t *l = NULL;

  if(signal == NULL)
    return CS_FAIL;

  if(label == NULL)
    return CS_FAIL;

  if(value == NULL)
    return CS_FAIL;
  
  *value = 0;
  for(entry = cs_lists_get_first(&signal->labels);
      entry != cs_lists_get_last(&signal->labels);
      entry = cs_lists_get_next(entry)){
    l = (cs_signal_label_t*)cs_lists_get_data(entry);
    if(strcmp(label,l->label) == 0){
      *value = l->value;
      return CS_OK;
    }
  }

  return CS_FAIL;
}


/**************************************************************************/
int cs_signal_getnolabel(
			 cs_signal_t *signal    /**< Signal structure*/
			 )
/**************************************************************************/
{
  
  if(signal == NULL)
    return CS_FAIL;

  return cs_lists_size(&signal->labels);
}

/**************************************************************************/
int cs_signal_getlabel_iter(
			    cs_signal_t *signal,    /**< Signal structure*/
			    int index,              /**< Label index*/
			    double *value,          /**< Value feedback pointer*/
			    char *label,            /**< Label buffer*/
			    int bufsize             /**< Label buffer size*/
			    )
/**************************************************************************/
{
  cs_signal_label_t *l = NULL;
  
  if(signal == NULL){
    _signal_error("Signal pointer failure\n");
    return CS_FAIL;
  }

  if(index >= cs_lists_size(&signal->labels)){
    _signal_error("Index out of range\n");
    return CS_FAIL;
  }

  if((value == NULL) || (label == NULL)){
    _signal_error("Feedback value pointer failure\n");
    return CS_FAIL;
  }

  if(bufsize < CS_CHARBUF){
    _signal_error("Label buffer is too small\n");
    return CS_FAIL;
  }

  l = (cs_signal_label_t*)cs_lists_get_iter(&signal->labels,index);
  *value = l->value;
  strncpy(label, (const char *)l->label, CS_CHARBUF);

  return CS_OK;
}

/**************************************************************************/
int cs_signal_getnonodes(
			 cs_signal_t *signal    /**< Signal structure*/
			 )
/**************************************************************************/
{
  
  if(signal == NULL)
    return CS_FAIL;

  return cs_lists_size(&signal->reqnodes);
}

/**************************************************************************/
int cs_signal_getnode_iter(
			   cs_signal_t *signal,    /**< Signal structure*/
			   int index,              /**< Node index*/
			   char *node,             /**< Node name buffer*/
			   int bufsize             /**< Label buffer size*/
			   )
/**************************************************************************/
{
  cs_signal_node_t *l = NULL;
  
  if(signal == NULL){
    _signal_error("Signal pointer failure\n");
    return CS_FAIL;
  }

  if(index >= cs_lists_size(&signal->reqnodes)){
    _signal_error("Index out of range\n");
    return CS_FAIL;
  }

  if(node == NULL){
    _signal_error("Feedback value pointer failure\n");
    return CS_FAIL;
  }

  if(bufsize < CS_CHARBUF){
    _signal_error("Label buffer is too small\n");
    return CS_FAIL;
  }

  l = (cs_signal_node_t*)cs_lists_get_iter(&signal->reqnodes,index);
  strncpy(node, (const char *)l->name, CS_CHARBUF);

  return CS_OK;
}
