/* 
    Copyright (C) 2018 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @file cs_signal_private.h
 * @author Andreas Stewering-Bone
 * @date 20 Dec 2018
 * @brief File containing restbus simulation signal private header.
 *
 * File containing restbus simulation signal handling private header.
 */
#ifndef CS_CAN_SIGNAL_PRIVATE_H
#define CS_CAN_SIGNAL_PRIVATE_H

#include <inttypes.h>
#include <stdarg.h>
#include "cs_defines.h"
#include "cs_lists.h"
#include "cs_signal.h"

/** \def CS_SIGNAL_MAXLABEL
    Maximum number of signal labels
*/
/*#define CS_SIGNAL_MAXLABEL 64*/

/** \def CS_SIGNAL_MAXDLC
    Maximal number of bytes in a message
    TODO Keep in synch with CAN interface, 
    to reduce places of definition
    Note prepared for CANFD
*/
#define CS_SIGNAL_MAXDLC 64


/** \struct cs_signal_label_t
    Signal label definition
*/
typedef struct{
  double value;          /**< Numeric Value*/
  char label[CS_CHARBUF];/**< Label string*/
}cs_signal_label_t;

/** \struct cs_signal_node_t
    Signal requesting node definition
*/
typedef struct{
  char name[CS_CHARBUF];/**< Label string*/
}cs_signal_node_t;

/** \struct cs_signal
    Signal definition structure
*/
struct cs_signal{
  char name[CS_CHARBUF];     /**< Signal name*/
  char unit[CS_CHARBUF];     /**< Signal unit*/
  char notes[CS_CHARBUF];    /**< Signal notes*/
  cs_signal_type_t type;     /**< Signal type*/
  cs_signal_byteorder_t bt;  /**< Byteorder*/
  uint16_t offset;           /**< Bitoffset*/
  uint16_t length;           /**< Signal length in bits*/
  double dscale;             /**< Scale factor*/
  double doffset;            /**< Offset factor*/
  double min;                /**< Minimal value*/
  double max;                /**< Maximum value*/
  int64_t multiplexcount;    /**< Multiplex counter*/
  int64_t rawvalue;          /**< Converted signal raw value*/
  double startvalue;         /**< Signal raw start value (rg.defined by DBC file)*/
  uint8_t buffer_rawvalue[CS_SIGNAL_MAXDLC];/**< Shifted raw signal, for signal static encoding*/
  cs_signal_mux_t mtype;     /**< Mux type*/
  void (*calc)(void *ctx, double *value, int reset); /**< Alive counter callback, for sending*/
  uint8_t (*check)(void *ctx,double *value, int reset); /**< Alive counter callback, for receiving to check value*/
  void *calc_ctx;           /**< Alive counter callback contect pointer*/
  void *check_ctx;           /**< Alive counter callback contect pointer*/
  uint8_t *check_flag;       /**< Feedback flag for checking received alive counter*/
  int (*crc)(void *ctx, uint8_t *buffer,int len, double *value); /**< CRC callback function*/
  void *crc_ctx;             /**< CRC callback context pointer*/
  uint8_t *crc_flag;         /**< Feedback flag for checking received CRC value*/
  cs_lists_t labels;         /**< Label list*/
  double *value;             /**< Pointer for converted signals*/
  uint8_t *fault;            /**< Pointer for signal failures*/
  char *label;               /**< Pointer for converted signal labels*/
  uint16_t start_byte;       /**< Start byte for packing, used internally*/
  uint16_t start_offset;     /**< Start bit offset for packing, used internally*/
  uint16_t end_byte;         /**< End byte for packing, used internally*/
  uint16_t end_offset;       /**< End bit offset for packing, used internally*/
  int64_t signbit;           /**< Signed bit representation, used internally*/
  cs_lists_t reqnodes;       /**< List of CAN nodes which need this signal*/
};


#endif
