/*  
    Copyright (C) 2013 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <net/if.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <linux/sockios.h>
#include <linux/can.h>
#include <linux/can/error.h>
#include <linux/can/raw.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/utsname.h>

#include "config.h"
#include "cs_socketcan.h"
#include "cs_socketcan_private.h"
#include "cs_time.h"
#include "cs_debug_private.h"

/** \def SOCKETCAN_CAN_MTU_OLD
    Sizeof struct can_frame on old system, eq. Debian 8
    Struct is not binary size compatible to struct canfd_frame
*/
#define SOCKETCAN_CAN_MTU_OLD 13
/** \def SOCKETCAN_CAN_MTU_NEW
    Sizeof struct can_frame on actual systems
    Struct is binary size compatible to struct canfd_frame
*/
#define SOCKETCAN_CAN_MTU_NEW 16

/* Taken from can-utils/lib.c */
/* CAN DLC to real data length conversion helpers */
static const uint8_t _socketcan_dlc2len[] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 12, 16, 20, 24, 32, 48, 64};

/* get data length from can_dlc with sanitized can_dlc */
uint8_t _socketcan_can_dlc2len(uint8_t can_dlc)
{
  return _socketcan_dlc2len[can_dlc & 0x0F];
}

static const uint8_t _socketcan_len2dlc[] = {0, 1, 2, 3, 4, 5, 6, 7, 8,		/* 0 - 8 */
					     9, 9, 9, 9,			/* 9 - 12 */
					     10, 10, 10, 10,			/* 13 - 16 */
					     11, 11, 11, 11,			/* 17 - 20 */
					     12, 12, 12, 12,			/* 21 - 24 */
					     13, 13, 13, 13, 13, 13, 13, 13,	/* 25 - 32 */
					     14, 14, 14, 14, 14, 14, 14, 14,	/* 33 - 40 */
					     14, 14, 14, 14, 14, 14, 14, 14,	/* 41 - 48 */
					     15, 15, 15, 15, 15, 15, 15, 15,	/* 49 - 56 */
					     15, 15, 15, 15, 15, 15, 15, 15};	/* 57 - 64 */

/* map the sanitized data length to an appropriate data length code */
uint8_t _socketcan_can_len2dlc(uint8_t len)
{
  if (len > 64)
    return 0xF;

  return _socketcan_len2dlc[len];
}

#define _socketcan_error(...) (cs_error(CS_DEBUG_SOCKETCAN,__func__,__LINE__,__VA_ARGS__))
#define _socketcan_warn(...) (cs_warn(CS_DEBUG_SOCKETCAN,__func__,__LINE__,__VA_ARGS__))
#define _socketcan_info(...) (cs_info(CS_DEBUG_SOCKETCAN,__func__,__LINE__,__VA_ARGS__))
#define _socketcan_debug(...) (cs_debug(CS_DEBUG_SOCKETCAN,__func__,__LINE__,__VA_ARGS__))
#define _socketcan_verbose(...) (cs_verbose(CS_DEBUG_SOCKETCAN,__func__,__LINE__,__VA_ARGS__))
#define _socketcan_error_buffer(DATA,LEN) (cs_error_buffer(CS_DEBUG_SOCKETCAN,__func__,__LINE__,DATA,LEN))
#define _socketcan_warn_buffer(DATA,LEN) (cs_warn_buffer(CS_DEBUG_SOCKETCAN,__func__,__LINE__,DATA,LEN))
#define _socketcan_info_buffer(DATA,LEN) (cs_info_buffer(CS_DEBUG_SOCKETCAN,__func__,__LINE__,DATA,LEN))
#define _socketcan_debug_buffer(DATA,LEN) (cs_debug_buffer(CS_DEBUG_SOCKETCAN,__func__,__LINE__,DATA,LEN))
#define _socketcan_verbose_buffer(DATA,LEN) (cs_verbose_buffer(CS_DEBUG_SOCKETCAN,__func__,__LINE__,DATA,LEN))




/********************************************************************/
cs_debug_level_t cs_socketcan_debuglevel(
					 cs_debug_level_t level
					 )
/********************************************************************/
{
  return cs_debuglevel(CS_DEBUG_SOCKETCAN,level);
}

/********************************************************************/
void cs_socketcan_debug_setcallback(
				    void *ctx,
				    void (*callb)(void *ctx,
						  cs_debug_sub_t sub,
						  cs_debug_level_t dl,
						  const char *func,
						  const int line,
						  const char *fmt,
						  va_list ap)
    )
/********************************************************************/
{
  cs_debug_setcallback(CS_DEBUG_SOCKETCAN,ctx,callb);
}

/********************************************************************/
int cs_socketcan_decode_errorclass(
				   cs_socketcan_t *sk,
				   uint32_t class,
				   uint8_t *data,
				   int datalen
				   )
/********************************************************************/
{ 
  if((data != NULL) && (datalen < 5))
    return CS_SOCKETCAN_FAIL;

  if(data && (data[0]  ==  CAN_ERR_LOSTARB)){
    /*lost arbitration*/
    //No error
    sk->error_cnt_lostarb++;
    _socketcan_verbose("Error: Lost arbitation\n");
  }

  if(class & CAN_ERR_TX_TIMEOUT){
    sk->lerror = CS_SOCKETCAN_ERROR_TXTIMEOUT;
    sk->error_cnt_txtimeout++;
    _socketcan_debug("Error: TX timeout\n");
  }
  if(class & CAN_ERR_CRTL){
    /*Controller failure*/
    if(data && (data[1] == 0x00)){
      sk->lerror = CS_SOCKETCAN_ERROR_CTRL_UNKNOWN;
      sk->error_cnt_ctrl++;
      _socketcan_debug("Error: Controller unspecific\n");
    }
    if(data && (data[1] & 0x01)){
      sk->lerror = CS_SOCKETCAN_ERROR_CTRL_RXOVERFLOW;
      sk->error_cnt_ctrl++;
      _socketcan_debug("Error: Controller RX overflow\n");
    }
    if(data && (data[1] & 0x02)){
      sk->lerror = CS_SOCKETCAN_ERROR_CTRL_TXOVERFLOW;
      sk->error_cnt_ctrl++;
      _socketcan_debug("Error: Controller TX overflow\n");
    }
    if(data && (data[1] &  0x04)){
      //No error
      sk->warning_cnt_ctrl_rxerror++;
      _socketcan_verbose("Warning: Controller RX Error level\n");
    }
    if(data && (data[1] &  0x08)){
      //No error
      sk->warning_cnt_ctrl_txerror++;
      _socketcan_verbose("Warning: Controller TX Error level\n");
    }
    if(data && (data[1] &  0x10)){
      sk->lerror = CS_SOCKETCAN_ERROR_CTRL_RXERROR;
      sk->error_cnt_ctrl_rxerror++;
      _socketcan_debug("Error: Controller passive RX Error\n");
    }
    if(data && (data[1] &  0x20)){
      sk->lerror = CS_SOCKETCAN_ERROR_CTRL_TXERROR;
      sk->error_cnt_ctrl_txerror++;
      _socketcan_debug("Error: Controller passive TX Error\n");
    }
  }
  if(class & CAN_ERR_TRX){
    /*Transmitter Failure*/
    sk->error_cnt_trx++;
    sk->lerror = CS_SOCKETCAN_ERROR_TRX;
    if(data){
      _socketcan_debug("Error: Transmitter: 0x%X\n",data[4]);
    } else {
      _socketcan_debug("Error: Transmitter\n");
    }
  }
  if(class & CAN_ERR_ACK){
    /*Transmitter Recieved no ack*/
    sk->error_cnt_ack++;
    sk->lerror = CS_SOCKETCAN_ERROR_ACK;
    _socketcan_debug("Error Ack\n");
  }
  if(class & CAN_ERR_PROT){
    /*Protokoll failure*/
    sk->error_cnt_prot++;
    sk->lerror = CS_SOCKETCAN_ERROR_PROT;
    if(data){
      _socketcan_debug("Error: Protocol: 0x%X 0x%X\n",data[2],data[3]);
    } else {
      _socketcan_debug("Error: Protocol\n");
    }
  }
  if(class & CAN_ERR_BUSOFF){
    /*Bus is off*/
    sk->error_cnt_busoff++;
    sk->lerror = CS_SOCKETCAN_ERROR_BUSOFF;
    _socketcan_debug("Error: Bus off\n");
  }
  if(class & CAN_ERR_BUSERROR){
    /*Buserror*/
    sk->error_cnt_buserror++;
    sk->lerror = CS_SOCKETCAN_ERROR_BUSERROR;
    _socketcan_debug("Error: Bus error\n");
  }
  if(class & CAN_ERR_RESTARTED){
    /*CAN-Bus restarted*/
    sk->error_cnt_busrestarted++;
    //No error flag
    _socketcan_verbose("Error: Bus restarted\n");
  }
  return CS_SOCKETCAN_OK;
}

/********************************************************************/
char* cs_socketcan_geterroridstring(
				    uint8_t lerror
				    )
/********************************************************************/
{  
  switch(lerror){
  case CS_SOCKETCAN_ERROR_NOERROR:
    return "SocketCAN: No Error";
  case CS_SOCKETCAN_ERROR_SOCKET:
    return "SocketCAN: Socket error";
  case CS_SOCKETCAN_ERROR_LOSTARB:
    return "SocketCAN: Lost arbitration";
  case CS_SOCKETCAN_ERROR_TXTIMEOUT:
    return "SocketCAN: TX timeout";
  case CS_SOCKETCAN_ERROR_CTRL_UNKNOWN:
    return "SocketCAN: Controller Unspecific Error";
  case CS_SOCKETCAN_ERROR_CTRL_RXOVERFLOW:
    return "SocketCAN: Controller RX Overflow";
  case CS_SOCKETCAN_ERROR_CTRL_TXOVERFLOW:
    return "SocketCAN: Controller TX Overflow";
  case CS_SOCKETCAN_ERROR_CTRL_RXERROR:
    return "SocketCAN: Controller RX Error";
  case CS_SOCKETCAN_ERROR_CTRL_TXERROR:
    return "SocketCAN: Controller TX Error";
  case CS_SOCKETCAN_ERROR_TRX:
    return "SocketCAN: Transmitter failure";
  case CS_SOCKETCAN_ERROR_ACK:
    return "SocketCAN: Acknoledge failure";
  case CS_SOCKETCAN_ERROR_PROT:
    return "SocketCAN: Protocol failure";
  case CS_SOCKETCAN_ERROR_BUSERROR:
    return "SocketCAN: Bus error";
  case CS_SOCKETCAN_ERROR_BUSOFF:
    return "SocketCAN: Bus off";
  case CS_SOCKETCAN_ERROR_PARAMETER:
    return "SocketCAN: API parameter failure";
  case CS_SOCKETCAN_ERROR_MALLOC:
    return "SocketCAN: Internal malloc failure";
  case CS_SOCKETCAN_ERROR_FRAME:
    return "SocketCAN: Invalid CAN frame";
  case CS_SOCKETCAN_ERROR_MODE:
    return "SocketCAN: Invalid CAN mode requested";
  case CS_SOCKETCAN_ERROR_BUFFER:
    return "SocketCAN: Invalid user buffer size";
  case CS_SOCKETCAN_ERROR_ERRNO:
    return "SocketCAN: Errno defined failure";
  }

  return "";
}

/********************************************************************/
int cs_socketcan_getlasterror(
			      cs_socketcan_t *sk
			      )
/********************************************************************/
{
  if(sk == NULL)
    return 0;
  
  return sk->lerror;
}

/********************************************************************/
char* cs_socketcan_getlasterrorstring(
				      cs_socketcan_t *sk
				      )
/********************************************************************/
{
  if(sk == NULL)
    return "";

  if(sk->lerror == CS_SOCKETCAN_ERROR_ERRNO){
    return strerror(sk->lerrno);
  }
  
  return cs_socketcan_geterroridstring(sk->lerror);
}

/** SocketCANd low level receive
    \retval -2 socket read failure
    \retval -1 nothing found
    \retval 0 all ok, check cmd buffer
*/
/********************************************************************/
int _socketcand_receive(
			cs_socketcan_t *sk,
			uint32_t timeout
			)
/********************************************************************/
{
  int rv;
  int i, start, stop;
  fd_set rdfs;
  struct timeval ts;
  div_t mydiv;
  int srv;
  
  /*timeout in ms*/
  mydiv = div(timeout, 1000);
  ts.tv_sec = mydiv.quot;
  ts.tv_usec = (int32_t)mydiv.rem * 1000L;
  
  /* if there are no more elements in the buffer read more data from the
   * socket.
   */
  if(!sk->socketcand_more_elements) {
    _socketcan_verbose("Recv Timeout: %d s, %ld us\n",ts.tv_sec,ts.tv_usec);
    FD_ZERO(&rdfs);
    FD_SET(sk->socket, &rdfs);
    srv = select(sk->socket +1, &rdfs, NULL, NULL, &ts);
    _socketcan_verbose("Select return %i\n",srv);
    if ( srv < 0) {
      _socketcan_error("Select failed\n");
      return -2;
    }
    if(srv > 0) {
      if (FD_ISSET(sk->socket, &rdfs)) {
	rv = read(sk->socket, sk->socketcand_cmd_buffer+sk->socketcand_cmd_index, CS_SOCKETCAN_SOCKETCAND_MAXLEN-sk->socketcand_cmd_index);
	if(rv < 0 ){
	  //Read failure
	  _socketcan_error("Read failure\n");
	  return -2;
	}
	sk->socketcand_cmd_index += rv;
      }
    }
  }
  
  sk->socketcand_more_elements = 0;
  
  /* find first '<' in string */
  start = -1;
  for(i=0; i<sk->socketcand_cmd_index; i++) {
    if(sk->socketcand_cmd_buffer[i] == '<') {
      start = i;
      _socketcan_verbose("Detect start: %i\n",start);
      break;
    }
  }
  
  /*
   * if there is no '<' in string it makes no sense to keep data because
   * we will never be able to construct a command of it
   */
  if(start == -1) {
    sk->socketcand_cmd_index = 0;
    _socketcan_error("No start of frame detected\n");
    return -1;
  }

  /* check whether the command is completely in the buffer */
  stop = -1;
  for(i=1; i<sk->socketcand_cmd_index; i++) {
    if(sk->socketcand_cmd_buffer[i] == '>') {
      stop = i;
      _socketcan_verbose("Detect stop: %i\n",stop);
      break;
    }
  }

  /* if no '>' is in the string we have to wait for more data */
  if(stop == -1) {
    _socketcan_error("No end of frame detected\n");
    return -1;
  }

  /* copy string to new destination and correct cmd_buffer */
  for(i=start; i<=stop; i++) {
    sk->socketcand_detectbuf[i-start] = sk->socketcand_cmd_buffer[i];
  }
  sk->socketcand_detectbuf[i-start] = '\0';


  /* if only this message was in the buffer we're done */
  if(stop == sk->socketcand_cmd_index-1) {
    sk->socketcand_cmd_index = 0;
  } else {
    /* check if there is a '<' after the stop */
    start = -1;
    for(i=stop; i<sk->socketcand_cmd_index; i++) {
      if(sk->socketcand_cmd_buffer[i] == '<') {
	start = i;
	_socketcan_verbose("Detect next start: %i\n",start);
	break;
      }
    }

    /* if there is none it is only garbage we can remove */
    if(start == -1) {
      sk->socketcand_cmd_index = 0;
      return 0;
      /* otherwise we copy the valid data to the beginning of the buffer */
    } else {
      for(i=start; i<sk->socketcand_cmd_index; i++) {
	sk->socketcand_cmd_buffer[i-start] = sk->socketcand_cmd_buffer[i];
      }
      sk->socketcand_cmd_index -= start;
      /* check if there is at least one full element in the buffer */
      stop = -1;
      for(i=1; i<sk->socketcand_cmd_index; i++) {
	if(sk->socketcand_cmd_buffer[i] == '>') {
	  stop = i;
	  _socketcan_verbose("Detect next stop: %i\n",stop);
	  break;
	}
      }
      
      if(stop != -1) {
	sk->socketcand_more_elements = 1;
	_socketcan_verbose("Second frame available\n");
      }
    }
  }
  return 0;
}

/** _socketcand_wait_receive
    Low level receive routine for SocketCANd
    \param sk Structure pointer
    \param tv Receive timestamp pointer 
    \param id CAN id pointer
    \param len Received bytes
    \param data Received buffer
    \param timeout Timeout in ms, if value is zero function is non blocking
    \retval CS_SOCKETCAND_SOCKET Socket failure
    \retval CS_SOCKETCAND_TIMEOUT Receive timeout
    \retval CS_SOCKETCAND_NONE Receive nothing
    \retval CS_SOCKETCAND_CONNECT Receive server connect
    \retval CS_SOCKETCAND_ERROR Receive error frame
    \retval CS_SOCKETCAND_FRAME Receive data frame
    \retval CS_SOCKETCAND_INVALID Receive invalid frame content
*/
/********************************************************************/
cs_socketcan_socketcand_frame_t _socketcand_wait_receive(
							 cs_socketcan_t *sk,
							 struct timeval *tv,
							 uint32_t *id,
							 int *len,
							 uint8_t *data,
							 uint32_t timeout
							 )
/********************************************************************/
{
  int rv;
  int lts_second;
  int lts_usec;
  char data_str[CS_SOCKETCAN_CHARBUF+1];
  //char buffer[CS_SOCKETCAN_CHARBUF];
  int rv_scan;
  int llen;
  struct timespec ts;
  //struct timespec ats;
  int i;
  uint8_t found;
  cs_ts_gettime(&ts);
  cs_ts_add_timeout(&ts,timeout);
  while(1){
    rv = _socketcand_receive(sk,timeout);
    if(rv == -2) {
      _socketcan_warn("Connection terminated while waiting for command\n");
      return CS_SOCKETCAND_SOCKET;
    }
    if(rv == -1){
      //Nothing was received,
      //wait for next interval
      if(timeout == 0){
	//if there is no timeout, return fastly
	return CS_SOCKETCAND_NONE;
      } else {
	//Nothing todo
      }
    }
    if(rv == 0){
      //Received valid frame
      //_socketcan_verbose("SocketCANd: Receive raw frame: %s\n",sk->socketcand_detectbuf);
      //Ckeck for frame type
      if(strncmp(sk->socketcand_detectbuf,"< hi >",6) == 0){
	_socketcan_debug("Receive server connect\n");
	return CS_SOCKETCAND_CONNECT;
      }
      if(strncmp(sk->socketcand_detectbuf,"< ok >",6) == 0){
	_socketcan_debug("Receive server ok\n");
	return CS_SOCKETCAND_OK;
      }
      if(strncmp(sk->socketcand_detectbuf,"< echo >",8) == 0){
	_socketcan_debug("Receive server echo\n");
	return CS_SOCKETCAND_ECHO;
      }
      if(strncmp(sk->socketcand_detectbuf,"< error",7) == 0){
	_socketcan_debug("Receive error frame: %s\n",sk->socketcand_detectbuf);
	rv_scan = sscanf(sk->socketcand_detectbuf, "< error %d >",id);
	if(rv_scan < 1){
	  _socketcan_error("Invalid frame detected: %s\n",sk->socketcand_detectbuf);
	  return CS_SOCKETCAND_INVALID;
	}
	return CS_SOCKETCAND_ERROR;
      }
      
      if(strncmp(sk->socketcand_detectbuf,"< frame",7) == 0){
	//Check for frame content
	_socketcan_verbose("Receive data frame: %s\n",sk->socketcand_detectbuf);
	//reset buffer
	memset(data,0,sizeof(uint8_t)*sk->max_dlc);
	memset(data_str,0,sizeof(data_str));

	//Check for max string len
	if(strlen(sk->socketcand_detectbuf) > CS_SOCKETCAN_CHARBUF){
	  _socketcan_error("Invalid frame length detected: %s\n",sk->socketcand_detectbuf);
	  return CS_SOCKETCAND_INVALID;
	}
	
	//Disassemble
	rv_scan = sscanf(sk->socketcand_detectbuf, "< frame %x %d.%d %s >",
			 id,
			 &lts_second,
			 &lts_usec,
			 data_str);
	
	if(rv_scan < 3){
	  _socketcan_error("Invalid frame detected: %s\n",sk->socketcand_detectbuf);
	  return CS_SOCKETCAND_INVALID;
	}
	
	//Check if frame is on receive filter list
	found = 0;
	if(sk->socketcand_num_ids > 0){
	  for( i = 0; i < sk->socketcand_num_ids; i++){
	    if( sk->socketcand_filter_ids[i] == *id){
	      found = 1;
	    }
	  }
	} else {
	  //No filter defined, take everything
	  found = 1;
	}
	
	
	if(found == 0){
	  //Skip frame
	  _socketcan_verbose("Skip frame: ID 0x%X \n",*id);
	  *tv = cs_ts_assembletimeval(0,0);
	  *id = 0;
	  *len = 0;
	} else {
	  *tv = cs_ts_assembletimeval(lts_second,lts_usec);
	  
	  llen = sscanf(data_str, "%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx",
			&data[0], &data[1],
			&data[2], &data[3],
			&data[4], &data[5],
			&data[6], &data[7]);
	  
	  if(llen < 0){
	    //Failure during scan
	    _socketcan_error("Invalid frame data: %s\n",data_str);
	    *tv = cs_ts_assembletimeval(0,0);
	    *id = 0;
	    return CS_SOCKETCAND_INVALID;
	  }
	  *len = llen;
	  _socketcan_verbose("Receive frame: ID 0x%X TS %f LEN %i %s\n",*id,cs_ts_timeval2d(*tv),*len,data_str);
	  return CS_SOCKETCAND_FRAME;
	}
      }

      //there was a unknown frame,
      //do not know how to handle
      //ignore
    }
    if(timeout == 0){
      //if there is no timeout, return fastly
      return CS_SOCKETCAND_NONE;
    } else {
      //cs_ts_gettime(&ats);
      //cs_ts_timespec2c(ats,buffer,CS_SOCKETCAN_CHARBUF);
      //_socketcan_verbose("Receive check timeout act time: %s \n",buffer);
      if(cs_ts_checkpast(&ts) == 1){
	_socketcan_error("Receive Timout\n");
	//Timeout
	return CS_SOCKETCAND_TIMEOUT;
      }
    }
  }
  //Should never be reached
  _socketcan_verbose("Should never be reached\n");
  return CS_SOCKETCAND_NONE;
}  

/********************************************************************/
int _socketcand_send(
		     cs_socketcan_t * sk,
		     char *buffer,
		     int len
		     )
/********************************************************************/
{
  _socketcan_verbose("Low level send: %s\n",buffer);
  if(send(sk->socket, (const void*)buffer, len, 0) < 0){
    if(errno == ENOBUFS){
      _socketcan_error("Socket: Write failed because of full TX Queque: %s\n",strerror(errno));
      sk->lerror = CS_SOCKETCAN_ERROR_SOCKET;
      return CS_SOCKETCAN_TXFULL;
    }
    _socketcan_error("Failed to send: %s\n",strerror(errno));
    sk->lerror = CS_SOCKETCAN_ERROR_SOCKET;
    return CS_SOCKETCAN_FAIL;
  }
  return CS_SOCKETCAN_OK;
}


/********************************************************************/
void cs_socketcan_reset(
			cs_socketcan_t *sk
			)
/********************************************************************/
{
  sk->socket = -1;
  sk->con_type = SOCKETCAN_CON_NATIVE;
  sk->mode = SOCKETCAN_MODE_CAN; //Traditional CAN 2.0
  sk->max_dlc = CAN_MAX_DLEN; // 8 Bytes, CAN 2.0
  sk->mtu = CAN_MTU; // sizeof(struct can_frame) for CAN 2.0
  memset(sk->interface,0,sizeof(char)*CS_SOCKETCAN_CHARBUF);
  memset(sk->hostname,0,sizeof(char)*CS_SOCKETCAN_CHARBUF);
  sk->port = -1;
  sk->rxbuffersize = CS_SOCKETCAN_RXBUFFSIZE;
  sk->socketcand_more_elements = 0;
  sk->socketcand_cmd_index = 0;
  memset(sk->socketcand_detectbuf,0,sizeof(char)*CS_SOCKETCAN_SOCKETCAND_MAXLEN);
  memset(sk->socketcand_cmd_buffer,0,sizeof(char)*CS_SOCKETCAN_SOCKETCAND_MAXLEN);
  memset(sk->socketcand_filter_ids,0,sizeof(uint16_t)*CS_SOCKETCAN_SOCKETCAND_MAX_FILTER);
  sk->socketcand_num_ids = 0;
  cs_socketcan_errorcnt_reset(sk);
  sk->lerror = CS_SOCKETCAN_ERROR_NOERROR;
  sk->lerrno = 0;
}

/********************************************************************/
cs_socketcan_t *cs_socketcan_init()
/********************************************************************/
{
  cs_socketcan_t *data = NULL;
    
  data = (cs_socketcan_t*)malloc(sizeof(cs_socketcan_t));
  if(data == NULL)
    return NULL;
  
  cs_socketcan_reset(data);

  return data;
}

/********************************************************************/
void cs_socketcan_free(
		       cs_socketcan_t *sk
		       )
/********************************************************************/
{
  if(sk == NULL)
    return;

  if(sk->socket != -1){
    close(sk->socket);
  }
  free(sk);
  sk = NULL;
}

/********************************************************************/
int cs_socketcan_isconnected(
			      cs_socketcan_t *sk
			      )
/********************************************************************/
{
  //Failure
  if(sk == NULL)
    return -1;

  //not connected
  if(sk->socket < 0)
    return 0;

  //connected
  return 1;
}

/********************************************************************/
int cs_socketcan_checkcompat()
/********************************************************************/
{
  struct utsname kdata;
  int kernel_major = 0;
  int kernel_minor = 0;
  
  if(uname(&kdata) < 0){
    _socketcan_error("Failed to get kernel data");
    return -1;
  }
  // kdata.release has format 3.18.46-rt50-etherlab
  if(sscanf(kdata.release,"%d.%d",&kernel_major,&kernel_minor) != 2){
    _socketcan_error("Failed to scan kernel data: %s",kdata.release);
    return -1;
  }

  _socketcan_debug("Size of CAN_MTU: %d Bytes",CAN_MTU);
  
  if(CAN_MTU != SOCKETCAN_CAN_MTU_NEW){
    _socketcan_warn("You are using a old header libc can.h format, Please update");
    _socketcan_warn("This libc API is only useable up to kernel 4.2");
    if((kernel_major < 4) || ( (kernel_major == 4) && (kernel_minor <= 2))){
      _socketcan_warn("Old kernel API detected, try it");
      return 0;
    } else {
      _socketcan_warn("New kernel API detected, try it with big risk");
      return 1;
    }
  } else {
    _socketcan_warn("You are using the new header libc can.h format.");
    _socketcan_warn("This libc API is only useable with kernel higher than 4.2");
    if((kernel_major < 4) || ( (kernel_major == 4) && (kernel_minor <= 2))){
      _socketcan_warn("Old kernel API detected, try it");
      return 2;
    } else {
      _socketcan_debug("New kernel API detected. Well done");
      return 3;
    }
  }
  return -1;
}
  
/********************************************************************/
void cs_socketcan_errorcnt_reset(
				 cs_socketcan_t *sk
				 )
/********************************************************************/
{
  if(sk == NULL)
    return;

  sk->warning_cnt_ctrl_rxerror = 0;
  sk->warning_cnt_ctrl_txerror = 0;
  sk->error_cnt_ctrl_rxerror = 0;
  sk->error_cnt_ctrl_txerror = 0;
  sk->error_cnt_ctrl = 0;
  sk->error_cnt_trx = 0;
  sk->error_cnt_ack = 0;
  sk->error_cnt_prot = 0;
  sk->error_cnt_busoff = 0;
  sk->error_cnt_buserror = 0;
  sk->error_cnt_busrestarted = 0;
  sk->error_cnt_lostarb = 0;
  sk->error_cnt_txtimeout = 0;
}

/** Local routine to check for CAN FD capabilities
    \retval -1 General Failure
    \retval 0 No CAN FD capabilities
    \retval 1 CAN FD capabilities
*/
/********************************************************************/
int _socketcan_check_canfd(
			   int socket,         /**< CAN socket handle*/
			   struct ifreq *ifr  /**< Prepared interface structure*/
			   )
/********************************************************************/
{
  /** Taken from can-utils/cangen.c */
  /* check if the frame fits into the CAN netdevice */
  if (ioctl(socket, SIOCGIFMTU, ifr) < 0) {
    _socketcan_error("Failed to get interface information");
    return -1;
  }
  
  if (ifr->ifr_mtu != CANFD_MTU) {
    _socketcan_warn("CAN interface %s is not CAN FD capable - sorry.",ifr->ifr_name);
    return 0;
  }

  //interface seems to be CAN FD capable
  return 1;
  
}

/********************************************************************/
int cs_socketcan_check_canfd(
			     char *ifcan
			     )
/********************************************************************/
{
  int s = 0;
  int rv;
  struct ifreq ifr;
  
  if(ifcan == NULL){
    _socketcan_error("No interface name defined \n");
    return -1;
  }

  if ((s = socket(PF_CAN, SOCK_RAW, CAN_RAW)) < 0) {
    _socketcan_error("Failed to init Socket\n");
    return -1;
  }
  
  strcpy(ifr.ifr_name,ifcan);
  if(ioctl(s, SIOCGIFINDEX, &ifr) < 0) {
    _socketcan_error("Failed to get interface index Socket\n");
    close(s);
    return -1;
  }

  rv = _socketcan_check_canfd(s,&ifr);

  close(s);
  return rv;
}

/********************************************************************/
int cs_socketcan_canfd_isenabled(
				 cs_socketcan_t *sk
				 )
/********************************************************************/
{
  if(sk == NULL){
    return -1;
  }

  if(sk->max_dlc == CAN_MAX_DLEN){
    //Traditional CAN 2.0
    return 0;
  }

  //CAN-FD
  return 1;
}

/********************************************************************/
int cs_socketcan_maxdlc(
			cs_socketcan_t *sk
			)
/********************************************************************/
{
  if(sk == NULL){
    return 0;
  }
  return sk->max_dlc;
}

/** Try to prepare CAN FD mode 
    \retval 0 success
    \retval 1 Failure
*/
/********************************************************************/
int _socketcan_canfd_prepare(
			     cs_socketcan_t *sk
			     )
/********************************************************************/
{
  int canfd_on = 1;

  //Extend maximum allowed DLC
  sk->max_dlc = CANFD_MAX_DLEN;
  sk->mtu = CANFD_MTU;	// sizeof(struct canfd_frame)

  if(setsockopt(sk->socket, SOL_CAN_RAW, CAN_RAW_FD_FRAMES, &canfd_on, sizeof(canfd_on)) == -1){
    _socketcan_error("Failed to enable CAN-FD mode for interface: %s, %s\n",sk->interface,strerror(errno));
    sk->lerror = CS_SOCKETCAN_ERROR_ERRNO;
    sk->lerrno = errno;
    return CS_SOCKETCAN_FAIL;    
  }
  
  return CS_SOCKETCAN_OK;
}

/********************************************************************/
int cs_socketcan_config_native(
			       cs_socketcan_t *sk,
			       char *ifcan,
			       cs_socketcan_mode_t mode
			       )
/********************************************************************/
{
  if(sk == NULL)
    return CS_SOCKETCAN_FAIL;

  _socketcan_verbose("Config native\n");
  if(ifcan == NULL){
    _socketcan_error("No interface name defined \n");
    sk->lerror = CS_SOCKETCAN_ERROR_PARAMETER;
    return CS_SOCKETCAN_FAIL;
  }
  sk->con_type = SOCKETCAN_CON_NATIVE;
  strncpy(sk->interface,ifcan,CS_SOCKETCAN_CHARBUF);
  sk->mode = mode;


  return CS_SOCKETCAN_OK;
}

/********************************************************************/
int cs_socketcan_config_socketcand(
				   cs_socketcan_t *sk,
				   char *ifcan,
				   char *hostname,
				   int port
				   )
/********************************************************************/
{
  struct sockaddr_in serveraddr;
  struct hostent *server_ent;
  struct timeval tv;
  uint32_t id;
  int len;
  cs_socketcan_socketcand_frame_t rv;
  uint32_t timeout = 1000; // Timeout for server connect in ms
  char buffer[CS_SOCKETCAN_CHARBUF];
  
  if(sk == NULL)
    return CS_SOCKETCAN_FAIL;

  uint8_t data[sk->max_dlc];
  
  _socketcan_verbose("Config SocketCANd\n");
  if(ifcan == NULL){
    _socketcan_error("No interface name defined \n");
    sk->lerror = CS_SOCKETCAN_ERROR_PARAMETER;
    return CS_SOCKETCAN_FAIL;
  }
  
  if(hostname == NULL){
    _socketcan_error("No hostname defined \n");
    sk->lerror = CS_SOCKETCAN_ERROR_PARAMETER;
    return CS_SOCKETCAN_FAIL;
  }

  sk->con_type = SOCKETCAN_CON_SOCKETCAND;
  strncpy(sk->interface,ifcan,CS_SOCKETCAN_CHARBUF);
  strncpy(sk->hostname,hostname,CS_SOCKETCAN_CHARBUF);
  sk->port = port;

  sk->socket = socket(AF_INET, SOCK_STREAM, 0);
  if(sk->socket < 0) {
    _socketcan_error("Open socket failed: %s\n",strerror(errno));
    sk->lerror = CS_SOCKETCAN_ERROR_ERRNO;
    sk->lerrno = errno;
    return CS_SOCKETCAN_FAIL;
  }
  
  memset(&serveraddr,0, sizeof(serveraddr));
  serveraddr.sin_family = AF_INET;
  serveraddr.sin_port = htons(sk->port);
  
  server_ent = gethostbyname(sk->hostname);
  if(server_ent == 0) {
    _socketcan_error("Invalid hostname: %s\n",sk->hostname);
    sk->lerror = CS_SOCKETCAN_ERROR_SOCKET;
    return CS_SOCKETCAN_FAIL;
  }

  bcopy(server_ent->h_addr, &(serveraddr.sin_addr.s_addr),
	server_ent->h_length);


  if(connect(sk->socket, (struct sockaddr*)&serveraddr,
	     sizeof(serveraddr)) != 0) {
    _socketcan_error("Failed to connect hostname: %s, %s\n",sk->hostname,strerror(errno));
    sk->lerror = CS_SOCKETCAN_ERROR_ERRNO;
    sk->lerrno = errno;
    return CS_SOCKETCAN_FAIL;
  }

  //Wait for connect flag < hi >
  rv = _socketcand_wait_receive(sk,&tv,&id,&len,data,timeout);
  if(rv != CS_SOCKETCAND_CONNECT){
    _socketcan_error("Failed to wait for connect from %s\n",sk->hostname);
    return CS_SOCKETCAN_FAIL;
  }

  //Open interface
  memset(buffer,0,sizeof(char)*CS_SOCKETCAN_CHARBUF);
  snprintf(buffer,CS_SOCKETCAN_CHARBUF,"< open %s >",sk->interface);
  if(_socketcand_send(sk,buffer,strlen(buffer))){
    //Failed to send
    return CS_SOCKETCAN_FAIL;
  }

  //Wait for ok
  rv = _socketcand_wait_receive(sk,&tv,&id,&len,data,timeout);
  if(rv != CS_SOCKETCAND_OK){
    _socketcan_error("Failed to wait for open interfac ack from %s\n",sk->hostname);
    return CS_SOCKETCAN_FAIL;
  }
  
  return CS_SOCKETCAN_OK;
}

/********************************************************************/
int _socketcan_set_filter_native(cs_socketcan_t *sk, 
				 uint32_t *ids,
				 int num_ids) 
/********************************************************************/
{
  int i;
  struct can_filter *filter = NULL;
 
  //Reset filter
  /*disable receiving on raw socket*/
  setsockopt(sk->socket, SOL_CAN_RAW, CAN_RAW_FILTER, NULL, 0);

  filter = (struct can_filter*)malloc(sizeof(struct can_filter)*num_ids);
  
  if(filter == NULL){
    _socketcan_error("Malloc failure\n");
    sk->lerror = CS_SOCKETCAN_ERROR_MALLOC;
    return CS_SOCKETCAN_FAIL;
  }
    
  memset(filter,0,sizeof(struct can_filter)*num_ids);

  for(i = 0; i < num_ids; i++){
    _socketcan_verbose("Set filter CAN id: hex %X\n",ids[i]);
    filter[i].can_id = ids[i]; 
    filter[i].can_mask = CAN_EFF_MASK; //0x7FF;
    filter[i].can_mask &= ~CAN_ERR_FLAG;
  }

  if(setsockopt(sk->socket, SOL_CAN_RAW, CAN_RAW_FILTER,filter, sizeof(struct can_filter)*num_ids) != 0){
    _socketcan_error("setsockopt failed: %s\n",strerror(errno));
    sk->lerror = CS_SOCKETCAN_ERROR_ERRNO;
    sk->lerrno = errno;
    return CS_SOCKETCAN_FAIL;
  }

  free(filter);
  
  return CS_SOCKETCAN_OK; 
}

/********************************************************************/
int _socketcan_open_native(cs_socketcan_t *sk) 
/********************************************************************/
{
  struct sockaddr_can addr;
  struct ifreq ifr;
  can_err_mask_t err_mask;
  int rv;
  
  _socketcan_verbose("Socket open: %s\n",sk->interface);

  if ((sk->socket = socket(PF_CAN, SOCK_RAW, CAN_RAW)) < 0) {
    _socketcan_error("Failed to init Socket: %s\n",strerror(errno));
    sk->lerror = CS_SOCKETCAN_ERROR_ERRNO;
    sk->lerrno = errno;
    return CS_SOCKETCAN_FAIL;
  }
  memset(&addr, 0, sizeof(addr));
  addr.can_family = PF_CAN;
  
  if (strcmp(sk->interface, "any") == 0)
    addr.can_ifindex = 0;
  else {
    strcpy(ifr.ifr_name,sk->interface);
    if(ioctl(sk->socket, SIOCGIFINDEX, &ifr) < 0) {
      _socketcan_error("Failed to get interface index Socket: %s\n",strerror(errno));
      sk->lerror = CS_SOCKETCAN_ERROR_PARAMETER;
      return CS_SOCKETCAN_FAIL;
    }
    addr.can_ifindex = ifr.ifr_ifindex;
  }  

  if((sk->mode == SOCKETCAN_MODE_FLEX) || (sk->mode == SOCKETCAN_MODE_CANFD)){
    //Test for CAN-FD
    rv = _socketcan_check_canfd(sk->socket,&ifr);
    if(rv < 0){
      sk->lerror = CS_SOCKETCAN_ERROR_SOCKET;
      return CS_SOCKETCAN_FAIL;
    }
    if((sk->mode == SOCKETCAN_MODE_CANFD) && (rv == 0)){
      //Force CAN-FD but it is not available
      _socketcan_error("CAN-FD is forced but not available\n");
      sk->lerror = CS_SOCKETCAN_ERROR_SOCKET;
      return CS_SOCKETCAN_FAIL;
    }
    if(rv == 1){
      //CAN-FD is available, prepare
      if(_socketcan_canfd_prepare(sk)){
	return CS_SOCKETCAN_FAIL;
      }
    }
  }
    
  //Activate Error msg recieve
  //Define classes of errors
  err_mask = ( CAN_ERR_TX_TIMEOUT |
	       CAN_ERR_BUSERROR |   
	       CAN_ERR_TRX |
	       CAN_ERR_BUSOFF |
	       CAN_ERR_CRTL |
	       CAN_ERR_PROT |
	       CAN_ERR_RESTARTED);
  //Note no CAN_ERR_LOSTARB lost arbitration
  setsockopt(sk->socket, SOL_CAN_RAW, CAN_RAW_ERR_FILTER,
	     &err_mask, sizeof(err_mask));
  
  //Resize RX Buffer
  setsockopt(sk->socket, SOL_SOCKET, SO_RCVBUFFORCE,&sk->rxbuffersize,sizeof(sk->rxbuffersize));
  
  fcntl(sk->socket, F_SETFL , O_NONBLOCK);
  if (bind(sk->socket, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
    _socketcan_error("Failed to bind Socket: %s\n",strerror(errno));
    sk->lerror = CS_SOCKETCAN_ERROR_ERRNO;
    sk->lerrno = errno;
    return CS_SOCKETCAN_FAIL;
  }

  //If there are already configured filter set filter
  if(sk->socketcand_num_ids > 0){
    if(_socketcan_set_filter_native(sk,sk->socketcand_filter_ids,sk->socketcand_num_ids)){
      _socketcan_error("Failed to set filter\n");
      return CS_SOCKETCAN_FAIL;
    }
  }
  
  return CS_SOCKETCAN_OK;
}

/********************************************************************/
int _socketcan_open_socketcand(cs_socketcan_t *sk) 
/********************************************************************/
{
  struct timeval tv;
  uint32_t id;
  int len;
  uint8_t data[sk->max_dlc];
  cs_socketcan_socketcand_frame_t rv;
  uint32_t timeout = 1000; // Timeout for server connect in ms
  char buffer[CS_SOCKETCAN_CHARBUF];
  struct timespec ts_start;
  struct timespec ts_end;
  //First measure latency
  memset(buffer,0,sizeof(char)*CS_SOCKETCAN_CHARBUF);
  snprintf(buffer,CS_SOCKETCAN_CHARBUF,"< echo >");
  if(_socketcand_send(sk,buffer,strlen(buffer))){
    //Failed to send
    return CS_SOCKETCAN_FAIL;
  }
  //Get actual time
  cs_ts_gettime(&ts_start);
  //Wait for response
  rv = _socketcand_wait_receive(sk,&tv,&id,&len,data,timeout);
  if(rv != CS_SOCKETCAND_ECHO){
    _socketcan_error("Failed to wait for echo from %s\n",sk->hostname);
    return CS_SOCKETCAN_FAIL;
  }
  cs_ts_gettime(&ts_end);
  //Calc diff
  _socketcan_verbose("Measured latency: %fs\n",cs_ts_timespec2d(ts_end)-cs_ts_timespec2d(ts_start));

  //Switch to Raw Mode
  memset(buffer,0,sizeof(char)*CS_SOCKETCAN_CHARBUF);
  snprintf(buffer,CS_SOCKETCAN_CHARBUF,"< rawmode >");
  if(_socketcand_send(sk,buffer,strlen(buffer))){
    //Failed to send
    return CS_SOCKETCAN_FAIL;
  }
  //Wait for response
  rv = _socketcand_wait_receive(sk,&tv,&id,&len,data,timeout);
  if(rv != CS_SOCKETCAND_OK){
    _socketcan_error("Failed to wait for ok from %s\n",sk->hostname);
    return CS_SOCKETCAN_FAIL;
  }

  return CS_SOCKETCAN_OK;
}

/********************************************************************/
int cs_socketcan_open(cs_socketcan_t *sk) 
/********************************************************************/
{
  if(sk == NULL)
    return CS_SOCKETCAN_FAIL;
  _socketcan_verbose("Socket Open\n");
  switch(sk->con_type){
  case SOCKETCAN_CON_NATIVE:
    return _socketcan_open_native(sk);
    break;
  case SOCKETCAN_CON_SOCKETCAND:
    return _socketcan_open_socketcand(sk);
    break;
  }
  return CS_SOCKETCAN_OK;
}
/********************************************************************/
int cs_socketcan_close(cs_socketcan_t *sk)
/********************************************************************/
{
  _socketcan_verbose("Socket close\n");
  if(sk == NULL)
    return CS_SOCKETCAN_FAIL;

  switch(sk->con_type){
  case SOCKETCAN_CON_NATIVE:
    if(sk->socket >= 0)
      close(sk->socket);
    sk->socket = -1;
    break;
  case SOCKETCAN_CON_SOCKETCAND:
    if(sk->socket >= 0)
      close(sk->socket);
    sk->socket = -1;
    break;
  }
  
  return CS_SOCKETCAN_OK;
}

/********************************************************************/
int cs_socketcan_getsocket(cs_socketcan_t *sk)
/********************************************************************/
{
  if(sk == NULL)
    return -1;

  return sk->socket;
}


/********************************************************************/
 int cs_socketcan_set_filter(cs_socketcan_t *sk, 
			     uint32_t *ids,
			     int num_ids) 
 /********************************************************************/
{
  int i;
  
  if(sk == NULL)
    return CS_SOCKETCAN_FAIL;

  _socketcan_verbose("Socket set filter\n");
  
  if(ids == NULL){
    sk->lerror = CS_SOCKETCAN_ERROR_PARAMETER;
    return CS_SOCKETCAN_FAIL;
  }

  if(num_ids > CS_SOCKETCAN_SOCKETCAND_MAX_FILTER){
    _socketcan_error("Number if filter id > max number");
    sk->lerror = CS_SOCKETCAN_ERROR_PARAMETER;
    return CS_SOCKETCAN_FAIL;
  }

  //Reset filter
  memset(sk->socketcand_filter_ids,0,sizeof(uint32_t)*CS_SOCKETCAN_SOCKETCAND_MAX_FILTER);
  sk->socketcand_num_ids = 0;

  //Store ids
  for(i = 0; i < num_ids; i++){
    _socketcan_verbose("Set filter for ID: 0x%X\n",ids[i]);
    sk->socketcand_filter_ids[i] = ids[i];
  }
  sk->socketcand_num_ids = num_ids;

  if((sk->con_type == SOCKETCAN_CON_NATIVE) && (sk->socket >= 0 )){
    //if socket is already open, and is native type, apply filter settings
    if(_socketcan_set_filter_native(sk,sk->socketcand_filter_ids,sk->socketcand_num_ids)){
      _socketcan_error("Failed to set filter\n");
      return CS_SOCKETCAN_FAIL;
    }
  }
  
  return CS_SOCKETCAN_OK; 
}

/********************************************************************/
int _socketcan_send_socketcand(
			       cs_socketcan_t *sk,
			       uint32_t id, 
			       uint8_t *buffer,
			       int len,
			       uint32_t flags
			       )
/********************************************************************/
{
  char data[CS_SOCKETCAN_CHARBUF];
  char vbuffer[CS_SOCKETCAN_CHARBUF];
  int i;

  //Check for unsupported flags
  if(flags & CAN_RTR_FLAG){
    _socketcan_error("Socket: RTR Frames are not supported\n");
      sk->lerror = CS_SOCKETCAN_ERROR_SOCKET;
      return CS_SOCKETCAN_FAIL;
  }

  memset(data,0,sizeof(char)*CS_SOCKETCAN_CHARBUF);
  //Add header
  snprintf(data,CS_SOCKETCAN_CHARBUF,"< send %X %i ",id,len);
  //Add data
  for(i = 0; i < len; i++){
    memset(vbuffer,0,sizeof(char)*CS_SOCKETCAN_CHARBUF);
    snprintf(vbuffer,CS_SOCKETCAN_CHARBUF,"%.2X ",buffer[i]);
    strcat(data,vbuffer);
  }
  //Add tail
  strcat(data," >");
  return _socketcand_send(sk,data,strlen(data));
}

/********************************************************************/
int _socketcan_send_native(cs_socketcan_t *sk,
			   uint32_t id, 
			   uint8_t *buffer,
			   int len,
			   uint32_t flags,
			   cs_socketcan_send_mode_t mode
			   )
/********************************************************************/
{
  int rv;
  struct canfd_frame frame;
    
  memset(&frame,0,sizeof(frame));

  frame.can_id = id;
  //Check for extended ID
  if((frame.can_id & CAN_EFF_MASK) > CAN_SFF_MASK){
    //Set EFF Flag
    frame.can_id |=  CAN_EFF_FLAG;
  }
  //Set all other flags
  frame.can_id |=  flags;

  if((mode == SOCKETCAN_SEND_CANFD) && (sk->mtu == CAN_MTU)){
    _socketcan_error("Write request for CAN-FD, but only CAN 2.0 available\n");
    sk->lerror = CS_SOCKETCAN_ERROR_MODE;
    return CS_SOCKETCAN_FAIL;
  }

  if(len > sk->max_dlc){
    _socketcan_error("Write request length to long for interface mode\n");
    sk->lerror = CS_SOCKETCAN_ERROR_MODE;
    return CS_SOCKETCAN_FAIL;
  }
  
  if((mode == SOCKETCAN_SEND_CANFD)||(mode == SOCKETCAN_SEND_CANFD_BRS)){
    //CAN-FD
    frame.len = _socketcan_can_dlc2len(_socketcan_can_len2dlc(len));
    if(mode == SOCKETCAN_SEND_CANFD_BRS){
      //Bit-Rate switch, enable second bitrate
      frame.flags |= CANFD_BRS;
    }
  } else {
    //CAN 2.0
    if(len > CAN_MAX_DLEN){
      _socketcan_error("Write request length to long for selected mode CAN 2.0\n");
      sk->lerror = CS_SOCKETCAN_ERROR_MODE;
      return CS_SOCKETCAN_FAIL;
    }
    frame.len = len;
  }
  
  memcpy(frame.data,buffer,len);

  
  rv = write(sk->socket, &frame, sk->mtu); 

  if(rv == 0){
    _socketcan_error("Write 0 Bytes\n");
    sk->lerror = CS_SOCKETCAN_ERROR_SOCKET;
    return CS_SOCKETCAN_FAIL;
  }

  if( rv < 0) {
    if(errno == ENOBUFS){
      sk->lerror = CS_SOCKETCAN_ERROR_ERRNO;
      sk->lerrno = errno;
      _socketcan_error("Write failed because of full TX Queque: %s\n",strerror(errno));
      return CS_SOCKETCAN_TXFULL;
    } else {
      _socketcan_error("Write failed: %s\n",strerror(errno));
      return CS_SOCKETCAN_FAIL;
    }
  }

  if( rv < sk->mtu) {
    _socketcan_error("Write incomplete\n");
    sk->lerror = CS_SOCKETCAN_ERROR_SOCKET;
    return CS_SOCKETCAN_FAIL;
  }
  
  return CS_SOCKETCAN_OK;

}

/********************************************************************/
int cs_socketcan_send(cs_socketcan_t *sk,
		      uint32_t id, 
		      uint8_t *buffer,
		      int len,
		      cs_socketcan_send_mode_t mode
		      )
/********************************************************************/
{
  if(sk == NULL)
    return CS_SOCKETCAN_FAIL;

  if(buffer == NULL){
    sk->lerror = CS_SOCKETCAN_ERROR_PARAMETER;
    return CS_SOCKETCAN_FAIL;
  }
  
  if(sk->socket < 0){
    _socketcan_error("Socket is not open\n");
    sk->lerror = CS_SOCKETCAN_ERROR_SOCKET;
    return CS_SOCKETCAN_FAIL;
  }

   if(len > sk->max_dlc){
     _socketcan_error("Invalid buffer length\n");
    sk->lerror = CS_SOCKETCAN_ERROR_PARAMETER;
    return CS_SOCKETCAN_FAIL;
  }
  
  switch(sk->con_type){
  case SOCKETCAN_CON_NATIVE:
    _socketcan_verbose("Via native Socket Id: 0x%X Len: %i\n",id,len);
    return _socketcan_send_native(sk,id,buffer,len,0,mode); 
    break;
  case SOCKETCAN_CON_SOCKETCAND:
    _socketcan_verbose("Via SocketCANd Id: 0x%X Len: %i\n",id,len);
    return _socketcan_send_socketcand(sk,id,buffer,len,0); 
    break;
  }
    
  return CS_SOCKETCAN_OK;

}

/********************************************************************/
int cs_socketcan_send_rtr(
			  cs_socketcan_t *sk,           /**< Socket struct pointer*/
			  uint32_t id                   /**< CAN id*/
			  )
/********************************************************************/
{ 
  if(sk == NULL)
    return CS_SOCKETCAN_FAIL;

  uint8_t data[sk->max_dlc]; //Only dummy
  
  if(sk->socket < 0){
    _socketcan_error("Socket is not open\n");
    sk->lerror = CS_SOCKETCAN_ERROR_SOCKET;
    return CS_SOCKETCAN_FAIL;
  }

  switch(sk->con_type){
  case SOCKETCAN_CON_NATIVE:
    _socketcan_verbose("Via native Socket Id: 0x%X\n",id);
    //only send in CAN2.0 mode because RTR information is only in the header
    return _socketcan_send_native(sk,id,data,0,CAN_RTR_FLAG,SOCKETCAN_SEND_CAN); 
    break;
  case SOCKETCAN_CON_SOCKETCAND:
    _socketcan_verbose("Via SocketCANd Id: 0x%X\n",id);
    return _socketcan_send_socketcand(sk,id,data,0,CAN_RTR_FLAG); 
    break;
  }
    
  return CS_SOCKETCAN_OK;
  
}
/********************************************************************/
int _socketcan_recv_non_block_native(cs_socketcan_t *sk,
				     uint32_t *id,
				     uint8_t *buffer,
				     int *len,
				     struct timeval *tv
				     )
/********************************************************************/
{
  int nbytes;
  canid_t class;
  struct canfd_frame frame;
  
  memset(&frame,0,CANFD_MTU);

  nbytes = read(sk->socket, &frame, CANFD_MTU);
  
  _socketcan_verbose("Received Bytes %i\n",nbytes);

  /*Please note
    We are working with the CAN-FD frame structure.
    Also the socket can not handle CAN-FD we can take this storage structure.

    Based on the number of bytes recieved we can decide if it is
    - a CAN 2.0 frame
    - a CAN-FD frame

   */
    
  if(nbytes < 0){
    //Nothing to read
    //_socketcan_verbose("Read failure: %s\n",strerror(errno));
    sk->lerror = CS_SOCKETCAN_ERROR_ERRNO;
    sk->lerrno = errno;
    *len = 0;
    return CS_SOCKETCAN_FAIL;
  }

  if(nbytes > 0) {
    gettimeofday(tv, NULL);
    /*Proceed the last message*/
    /* Check for header incompatible sizes 
    */
    if((nbytes == CANFD_MTU) || (nbytes == CAN_MTU)){
      if (ioctl(sk->socket, SIOCGSTAMP, tv) < 0) {
	_socketcan_error("Get Timestamp failed: %s\n",strerror(errno));
	sk->lerror = CS_SOCKETCAN_ERROR_ERRNO;
	sk->lerrno = errno;
	*len = 0;
	return CS_SOCKETCAN_FAIL;
      }
      /*Handle message frame*/
      if(frame.can_id & CAN_ERR_FLAG) {
	/*Get class of failure*/
	class = frame.can_id & CAN_EFF_MASK;
	/*Check Data length for error frames*/
	if(frame.len != CAN_ERR_DLC){
	  /*Invalid frame length*/
	  *len = 0;
	  sk->lerror = CS_SOCKETCAN_ERROR_FRAME;
	  return CS_SOCKETCAN_FAIL;
	}
	cs_socketcan_decode_errorclass(sk,class,frame.data,frame.len);
	*len = 0;
	return CS_SOCKETCAN_ERROR;
      }
      
      if(frame.can_id & CAN_EFF_FLAG){
	*id = frame.can_id & CAN_EFF_MASK;
      } else {
	*id = frame.can_id & CAN_SFF_MASK;
      }
      
      if(frame.can_id & CAN_RTR_FLAG){
	/*Ignore RTR frames*/
	_socketcan_verbose("Receive RTR frame: Id 0x%X\n",*id);
	*len = 0;
	return CS_SOCKETCAN_RTR;
      }
      
      if(frame.len > sk->max_dlc){
	_socketcan_error("Received buffer bigger is bigger than max frame size, received bytes: %i\n",frame.len);
	*len = 0;
	sk->lerror = CS_SOCKETCAN_ERROR_PARAMETER;
	return CS_SOCKETCAN_FAIL;
      }
      
      //check for maximum allowed length
      if(*len < frame.len){
	_socketcan_error("Received buffer bigger than storage buffer, CAN-ID 0x%X, Buffer %i, Recv: %i\n",
			 *id,
			 *len,
			 frame.len);
	*len = 0;
	sk->lerror = CS_SOCKETCAN_ERROR_BUFFER;
	return CS_SOCKETCAN_FAIL;
      }
      *len = frame.len;
      memcpy(buffer,frame.data,*len);
      return CS_SOCKETCAN_OK;	
    } else {
      _socketcan_error("Unknown receive size\n"); 
      *len = 0;
      sk->lerror = CS_SOCKETCAN_ERROR_FRAME;
      return CS_SOCKETCAN_FAIL;
    }
  } else {
    _socketcan_error("Read 0 Bytes\n"); 
    *len = 0;
    //No error
    return CS_SOCKETCAN_FAIL;
  }
  _socketcan_verbose("No packet received\n"); 
  *len = 0;
  //No error
  return CS_SOCKETCAN_FAIL;
}

/********************************************************************/
int _socketcan_recv_nonblock_socketcand(cs_socketcan_t *sk,
					uint32_t *id,
					uint8_t *buffer,
					int *len,
					struct timeval *tv)
/********************************************************************/
{
  int rv;
  //Zero timeout
  rv = _socketcand_wait_receive(sk,tv,id,len,buffer,0);
      
  if(rv == CS_SOCKETCAND_SOCKET){
    sk->lerror = CS_SOCKETCAN_ERROR_SOCKET;
    return CS_SOCKETCAN_FAIL;
  }
  if(rv == CS_SOCKETCAND_CONNECT){
    //No failure
    *len = 0;
    return CS_SOCKETCAN_FAIL;
  }
  if(rv == CS_SOCKETCAND_OK){
    *len = 0;
    //No failure
    return CS_SOCKETCAN_FAIL;
  }
  if(rv == CS_SOCKETCAND_ECHO){
    *len = 0;
    //No failure
    return CS_SOCKETCAN_FAIL;
  }
  if(rv == CS_SOCKETCAND_INVALID){
    *len = 0;
    //No failure
    return CS_SOCKETCAN_FAIL;
  }
  if(rv == CS_SOCKETCAND_ERROR){
    *len = 0;
    cs_socketcan_decode_errorclass(sk,*id,NULL,0);
    return CS_SOCKETCAN_FAIL;
  }
  if(rv == CS_SOCKETCAND_TIMEOUT){
    *len = 0;
    //No failure
    return CS_SOCKETCAN_FAIL;
  }
  if(rv == CS_SOCKETCAND_NONE){
    *len = 0;
    //No failure
    return CS_SOCKETCAN_ERROR;
  }
  if(rv != CS_SOCKETCAND_FRAME){
    *len = 0;
    //Should never be reached
    return CS_SOCKETCAN_FAIL;
  }

  //In all other cases, success
  return CS_SOCKETCAN_OK;	
 }

/********************************************************************/
int cs_socketcan_recv_nonblock(cs_socketcan_t *sk,
			       uint32_t *id,
			       uint8_t *buffer,
			       int *len,
			       struct timeval *tv)
/********************************************************************/
{
  if(sk == NULL)
    return CS_SOCKETCAN_FAIL;

  if(buffer == NULL){
    sk->lerror = CS_SOCKETCAN_ERROR_PARAMETER;
    return CS_SOCKETCAN_FAIL;
  }

  if(len == NULL){
    sk->lerror = CS_SOCKETCAN_ERROR_PARAMETER;
    return CS_SOCKETCAN_FAIL;
  }

  if(id == NULL){
    sk->lerror = CS_SOCKETCAN_ERROR_PARAMETER;
    return CS_SOCKETCAN_FAIL;
  }

  if(tv == NULL){
    sk->lerror = CS_SOCKETCAN_ERROR_PARAMETER;
    return CS_SOCKETCAN_FAIL;
  }

  if(sk->socket < 0){
    _socketcan_error("Socket is not open\n");
    sk->lerror = CS_SOCKETCAN_ERROR_SOCKET;
    return CS_SOCKETCAN_FAIL;
  }


  *id = 0;

  switch(sk->con_type){
  case SOCKETCAN_CON_NATIVE:
    return _socketcan_recv_non_block_native(sk,id,buffer,len,tv); 
    break;
  case SOCKETCAN_CON_SOCKETCAND:
    return _socketcan_recv_non_block_native(sk,id,buffer,len,tv);
    break;
  }
  return CS_SOCKETCAN_FAIL;
}


/********************************************************************/
int _socketcan_recv_block_native(cs_socketcan_t *sk,
				 uint32_t *id,
				 uint8_t *buffer,
				 int *len,
				 struct timeval *tv,
				 uint32_t timeout)
/********************************************************************/
{
  fd_set rdfs;
  struct timeval ts;
  div_t mydiv;
  int nbytes;
  struct canfd_frame frame;
  int rv;
  canid_t class;

  /*Please note
    We are working with the CAN-FD frame structure.
    Also the socket can not handle CAN-FD we can take this storage structure.

    Based on the number of bytes recieved we can decide if it is
    - a CAN 2.0 frame
    - a CAN-FD frame

   */
  
  memset(&frame,0,CANFD_MTU);

  /*timeout in ms*/
  mydiv = div(timeout, 1000);
  ts.tv_sec = mydiv.quot;
  ts.tv_usec = (int32_t)mydiv.rem * 1000L;

  _socketcan_verbose("Recv Timeout: %d s, %ld us\n",ts.tv_sec,ts.tv_usec);
  FD_ZERO(&rdfs);
  FD_SET(sk->socket, &rdfs);

  rv = select(sk->socket +1, &rdfs, NULL, NULL, &ts);
  _socketcan_verbose("Select return %i\n",rv);
  if ( rv < 0) {
    _socketcan_error("Select failed: %s\n",strerror(errno));
    *len = 0;
    sk->lerror = CS_SOCKETCAN_ERROR_ERRNO;
    sk->lerrno = errno;
    return CS_SOCKETCAN_FAIL;
  }
  if(rv > 0) {
    if (FD_ISSET(sk->socket, &rdfs)) {
      nbytes = read(sk->socket, &frame, CANFD_MTU);
      _socketcan_verbose("Read return %i\n",nbytes);
      if(nbytes > 0) {
	gettimeofday(tv, NULL);
	/*Proceed the last message*/
	if((nbytes == CAN_MTU) || (nbytes == CANFD_MTU)) {
	  if (ioctl(sk->socket, SIOCGSTAMP, tv) < 0) {
	    _socketcan_error("Get timestamp failed: %s\n",strerror(errno));
	    *len = 0;
	    sk->lerror = CS_SOCKETCAN_ERROR_ERRNO;
	    sk->lerrno = errno;
	    return CS_SOCKETCAN_FAIL;
	  }
	  /*Handle message frame*/
	  if(frame.can_id & CAN_ERR_FLAG) {
	    /*Get class of failure*/
	    class = frame.can_id & CAN_EFF_MASK;
	    /*Check Data length for error frames*/
	    if(frame.len != CAN_ERR_DLC){
	      /*Invalid frame length*/
	      *len = 0;
	      sk->lerror = CS_SOCKETCAN_ERROR_FRAME;
	      return CS_SOCKETCAN_FAIL;
	    }
	    cs_socketcan_decode_errorclass(sk,class,frame.data,frame.len);
	    *len = 0;
	    return CS_SOCKETCAN_ERROR;
	  }

	  if(frame.can_id & CAN_EFF_FLAG){
	    *id = frame.can_id & CAN_EFF_MASK;
	  } else {
	    *id = frame.can_id & CAN_SFF_MASK;
	  }
	  
	  if(frame.can_id & CAN_RTR_FLAG){
	    _socketcan_verbose("Receive RTR frame: Id 0x%X\n",*id);
	    *len = 0;
	    return CS_SOCKETCAN_RTR;
	  }
	  
	  if(frame.len > sk->max_dlc){
	    _socketcan_error("Received buffer bigger is bigger than max frame size %i, received bytes: %i\n",sk->max_dlc,frame.len);
	    *len = 0;
	    sk->lerror = CS_SOCKETCAN_ERROR_PARAMETER;
	    return CS_SOCKETCAN_FAIL;
	  }
	  
	  //check for maximum allowed length
	  if(*len < frame.len){
	    _socketcan_error("Received buffer %i bytes, bigger than storage buffer %i bytes\n",frame.len,*len);
	    *len = 0;
	    sk->lerror = CS_SOCKETCAN_ERROR_BUFFER;
	    return CS_SOCKETCAN_FAIL;
	  }
	  *len = frame.len;
	  memcpy(buffer,frame.data,*len);
	  _socketcan_verbose("Received Frame CAN id 0x%X, DLC %i\n",*id,*len);
	  return CS_SOCKETCAN_OK;	
	} else {
	  _socketcan_error("Unknown receive size\n");
	  *len = 0;
	  sk->lerror = CS_SOCKETCAN_ERROR_FRAME;
	  return CS_SOCKETCAN_FAIL;
	}
      } else {
	*len = 0;
	if(rv < 0 ){
	  _socketcan_error("Read failed: %s\n",strerror(errno));
	  sk->lerror = CS_SOCKETCAN_ERROR_ERRNO;
	  sk->lerrno = errno;
	} else {
	  _socketcan_error("Read failed\n");
	  sk->lerror = CS_SOCKETCAN_ERROR_SOCKET;
	}
	return CS_SOCKETCAN_FAIL;
      }
    }
  }
  _socketcan_verbose("Timeout\n");
  *len = 0;
  return CS_SOCKETCAN_TIMEOUT;
}

/********************************************************************/
int _socketcan_recv_block_socketcand(cs_socketcan_t *sk,
				     uint32_t *id,
				     uint8_t *buffer,
				     int *len,
				     struct timeval *tv,
				     uint32_t timeout)
/********************************************************************/
{
  int rv;
  rv = _socketcand_wait_receive(sk,tv,id,len,buffer,timeout);

  _socketcan_verbose("Receive block rv=%i\n",rv);
  if(rv == CS_SOCKETCAND_SOCKET){
    sk->lerror = CS_SOCKETCAN_ERROR_SOCKET;
    return CS_SOCKETCAN_FAIL;
  }
  if(rv == CS_SOCKETCAND_CONNECT){
    //No failure
    *len = 0;
    return CS_SOCKETCAN_ERROR;
  }
  if(rv == CS_SOCKETCAND_OK){
    *len = 0;
    //No failure
    return CS_SOCKETCAN_ERROR;
  }
  if(rv == CS_SOCKETCAND_ECHO){
    *len = 0;
    //No failure
    return CS_SOCKETCAN_ERROR;
  }
  if(rv == CS_SOCKETCAND_INVALID){
    *len = 0;
    sk->lerror = CS_SOCKETCAN_ERROR_FRAME;
    return CS_SOCKETCAN_FAIL;
  }
  if(rv == CS_SOCKETCAND_ERROR){
    *len = 0;
    cs_socketcan_decode_errorclass(sk,*id,NULL,0);
    return CS_SOCKETCAN_ERROR;
  }
  if(rv == CS_SOCKETCAND_TIMEOUT){
    *len = 0;
    //No failure
    return CS_SOCKETCAN_TIMEOUT;
  }
  if(rv != CS_SOCKETCAND_FRAME){
    *len = 0;
    //Should never be reached
    return CS_SOCKETCAN_FAIL;
  }

  //In all other cases, success
  return CS_SOCKETCAN_OK;	
 }


/********************************************************************/
int cs_socketcan_recv_block(cs_socketcan_t *sk,
			    uint32_t *id,
			    uint8_t *buffer,
			    int *len,
			    struct timeval *tv,
			    uint32_t timeout)
/********************************************************************/
{
  if(sk == NULL)
    return CS_SOCKETCAN_FAIL;

  if(buffer == NULL){
    sk->lerror = CS_SOCKETCAN_ERROR_PARAMETER;
    return CS_SOCKETCAN_FAIL;
  }

  if(id == NULL){
    sk->lerror = CS_SOCKETCAN_ERROR_PARAMETER;
    return CS_SOCKETCAN_FAIL;
  }

  if(tv == NULL){
    sk->lerror = CS_SOCKETCAN_ERROR_PARAMETER;
    return CS_SOCKETCAN_FAIL;
  }

  if(sk->socket < 0){
    _socketcan_error("Socket is not open\n");
    sk->lerror = CS_SOCKETCAN_ERROR_SOCKET;
    return CS_SOCKETCAN_FAIL;
  }

  *id = 0;
  
  switch(sk->con_type){
  case SOCKETCAN_CON_NATIVE:
    _socketcan_verbose("Via native Socket\n");
    return _socketcan_recv_block_native(sk,id,buffer,len,tv,timeout); 
    break;
  case SOCKETCAN_CON_SOCKETCAND:
    _socketcan_verbose("Via SocketCANd\n");
    return _socketcan_recv_block_socketcand(sk,id,buffer,len,tv,timeout);
    break;
  }

  return CS_SOCKETCAN_FAIL;
}

