/*  
    Copyright (C) 2017 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
#ifndef CS_SOCKETCAN_PRIVATE_H
#define CS_SOCKETCAN_PRIVATE_H

/**
 * @file cs_socketcan_private.h
 * @author Andreas Stewering-Bone
 * @date 9 Mar 2017
 * @brief File containing SOCKETCAN abstraction layer private header.
 *
 * File containing SOCKETCAN abstraction layer private header.
 */

#include <inttypes.h>

/* \def CS_SOCKETCAN_CHARBUF
   SocketCAN charbuf size
*/
#define CS_SOCKETCAN_CHARBUF 255

/** \def CS_SOCKETCAN_RXBUFFSIZE (2*1024*1024)
    SocketCAN RX Socket Buffer size (2MB)
*/
#define CS_SOCKETCAN_RXBUFFSIZE (2*1024*1024)

/** \def CS_SOCKETCAN_SOCKETCAND_MAXLEN
    SocketCANd receive buffer length 100kByte
*/
#define CS_SOCKETCAN_SOCKETCAND_MAXLEN (100*1024)

/** \def CS_SOCKETCAN_SOCKETCAND_MAX_FILTER
    SocketCANd maximum number of filter ids
*/
#define CS_SOCKETCAN_SOCKETCAND_MAX_FILTER 0xff

/** \enum cs_socketcan_con_t
    SocketCAN connection type
*/
typedef enum{
  SOCKETCAN_CON_NATIVE = 0x00,            /**< Local SocketCAN socket*/
  SOCKETCAN_CON_SOCKETCAND = 0x01,               /**< SocketCANd connection*/
}cs_socketcan_con_t;

/** \enum cs_socketcan_socketcand_frame_t
    SocketCANd frame types
*/
typedef enum{
  CS_SOCKETCAND_SOCKET = 0x00,            /**< Socket failure*/
  CS_SOCKETCAND_NONE,                     /**< Nothing received*/
  CS_SOCKETCAND_TIMEOUT,                  /**< Nothing received timeout*/
  CS_SOCKETCAND_CONNECT,                  /**< Server connect < hi >*/
  CS_SOCKETCAND_OK,                       /**< Server connect < ok >*/
  CS_SOCKETCAND_ECHO,                     /**< Server echo < echo >*/
  CS_SOCKETCAND_ERROR,                    /**< Error frame < error ...>*/
  CS_SOCKETCAND_FRAME,                    /**< data frame < frame ...>*/
  CS_SOCKETCAND_INVALID,                  /**< Invalid content received*/
}cs_socketcan_socketcand_frame_t;

/** \struct cs_socketcan
    Central stack structure, private definiton
*/
struct cs_socketcan{
  int socket;                             /**< SocketCAN socket*/
  char interface[CS_SOCKETCAN_CHARBUF];   /**< SocketCAN interface name*/
  cs_socketcan_mode_t mode;               /**< Running mode*/
  int max_dlc;                            /**< Max DLC*/
  int mtu;                                /**< Interface MTU*/
  char hostname[CS_SOCKETCAN_CHARBUF];    /**< SocketCANs host name*/
  int port;                               /**< SocketCANd server port*/ 
  cs_socketcan_con_t con_type;            /**< Interface type*/
  uint64_t warning_cnt_ctrl_rxerror;      /**< Warning counter Controller RX error*/
  uint64_t warning_cnt_ctrl_txerror;      /**< Warning counter Controller TX error*/
  uint64_t error_cnt_ctrl_rxerror;      /**< Error counter Controller RX error*/
  uint64_t error_cnt_ctrl_txerror;      /**< Error counter Controller TX error*/
  uint64_t error_cnt_ctrl;                /**< Error counter Controller*/
  uint64_t error_cnt_trx;                 /**< Error counter Transmitter*/
  uint64_t error_cnt_ack;                 /**< Error counter no ACK received*/
  uint64_t error_cnt_prot;                /**< Error counter protocol*/
  uint64_t error_cnt_busoff;              /**< Error counter bus off*/
  uint64_t error_cnt_buserror;            /**< Error counter bus error*/
  uint64_t error_cnt_busrestarted;           /**< Error counter bus restarted*/
  uint64_t error_cnt_lostarb;             /**< Error Lost arbitation*/
  uint64_t error_cnt_txtimeout;           /**< Error TX Timeout*/
  int rxbuffersize;                       /**< SocketCAN native RX buf size*/
  uint8_t socketcand_more_elements;       /**< SocketCANd flag for more commands*/
  int socketcand_cmd_index;               /**< SocketCANd cmd index*/
  char socketcand_detectbuf[CS_SOCKETCAN_SOCKETCAND_MAXLEN]; /**< SocketCANd detected < XXX > buffer*/
  char socketcand_cmd_buffer[CS_SOCKETCAN_SOCKETCAND_MAXLEN]; /**< SocketCANd raw cmd buffer*/
  uint32_t socketcand_filter_ids[CS_SOCKETCAN_SOCKETCAND_MAX_FILTER];         /**< Array of filter id CAN ids, usage SocketCANd*/
  int socketcand_num_ids;                 /**< Length of array of active CAN filter ids, usage SocketCANd*/
  uint8_t lerror;                         /**< Last error flag*/
  int lerrno;                             /**< Stored errno number*/  
 
};


#endif
