/*  CAN Stack for SocketCAN
    Copyright (C) 2017 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <inttypes.h>
#include <stdio.h>
#include <string.h>


#include <time.h>
#include <sys/time.h>

#include "cs_defines.h"
#include "cs_time.h"

/** Timeout in ms to timespec
    \retval timespec
*/
/*************************************************************************************************/
struct timespec cs_ts_timeout2ts(uint32_t timeout)
/*************************************************************************************************/
{
    struct timespec ts;
    uint32_t mydiv = 1000;
    //timeout in 0.001s
    ts.tv_sec = (timeout/ mydiv);
    ts.tv_nsec = ((long)(timeout % mydiv))*1e6;
    return ts;
}

/****************************************************************************/
struct timespec cs_ts_addtimespec(struct timespec time1, struct timespec time2)
/****************************************************************************/
{
    struct timespec result;
    if((time1.tv_nsec + time2.tv_nsec) >= 1e9) {
      result.tv_sec = time1.tv_sec + time2.tv_sec + 1;
      result.tv_nsec = time1.tv_nsec +time2.tv_nsec -1e9;
    } else {
      result.tv_sec = time1.tv_sec + time2.tv_sec ;
      result.tv_nsec = time1.tv_nsec +time2.tv_nsec;
    }
    return result;
}

/****************************************************************************/
struct timespec cs_ts_difftimespec(struct timespec start, struct timespec end)
/****************************************************************************/
{
  struct timespec temp;
  if ((end.tv_nsec-start.tv_nsec)<0) {
    temp.tv_sec = end.tv_sec-start.tv_sec-1;
    temp.tv_nsec = 1000000000+end.tv_nsec-start.tv_nsec;
  } else {
    temp.tv_sec = end.tv_sec-start.tv_sec;
    temp.tv_nsec = end.tv_nsec-start.tv_nsec;
  }
  return temp;
}

/*************************************************************************************************/
void cs_ts_add_timeout(struct timespec *ts, uint32_t timeout)
/*************************************************************************************************/
{
    // timeout in ms
    *ts =cs_ts_addtimespec(*ts,cs_ts_timeout2ts(timeout));
}

/*************************************************************************************************/
void cs_ts_next(struct timespec *ts, uint32_t timeout)
/*************************************************************************************************/
{
  //Timeout in ms
  //Get actual time
  cs_ts_gettime(ts);
  //Add timeout
  cs_ts_add_timeout(ts,timeout);
}


/** Compare timestamps
    \retval 0 if ta <= tb
    \retval 1 if ta > tb
*/
/*************************************************************************************************/
int cs_ts_cmp(struct timespec *ta,struct timespec *tb)
/*************************************************************************************************/
{
  /*char buffer[CS_CHARBUF];
  cs_ts_timespec2c(*ta,buffer,CS_CHARBUF);
  printf("TA: %s ",buffer);
  cs_ts_timespec2c(*tb,buffer,CS_CHARBUF);
  printf("TB: %s \n",buffer);*/
  if((ta->tv_sec < tb->tv_sec) ||
     ((ta->tv_sec == tb->tv_sec) && (ta->tv_nsec < tb->tv_nsec))){
    //printf("TA < TB\n");
    return 0;
  } else{
    //printf("TA >= TB\n");
    return 1;
  }
}

/*************************************************************************************************/
void cs_ts_gettime(struct timespec *ts)
/*************************************************************************************************/
{
    clock_gettime(CLOCK_MONOTONIC, ts);
}

/*************************************************************************************************/
void cs_ts_gettimeval(struct timeval *tv)
/*************************************************************************************************/
{
  struct timespec ts;
  cs_ts_gettime(&ts);
  *tv = cs_ts_timespec2timeval(ts);
}

/** Check if timestamp is in the past
    \retval 0 if ts is not in the past
    \retval 1 if ts is in the past
*/
/*************************************************************************************************/
int cs_ts_checkpast(struct timespec *ts)
/*************************************************************************************************/
{
    struct timespec now;
    cs_ts_gettime(&now);
    /*char buffer[CS_CHARBUF];
    cs_ts_timespec2c(now,buffer,CS_CHARBUF);
    printf("Now: %s\n",buffer);*/
    return cs_ts_cmp(&now,ts);
}

/****************************************************************************/
struct timeval cs_ts_timespec2timeval(struct timespec time)
/****************************************************************************/
{
    struct timeval tv;
    tv.tv_sec = time.tv_sec;
    tv.tv_usec = time.tv_nsec / 1000;
    return tv;
}

/****************************************************************************/
double cs_ts_timespec2d(struct timespec time)
/****************************************************************************/
{
    return (double)time.tv_sec + (double)time.tv_nsec * 1.0e-9;
}

/****************************************************************************/
void cs_ts_timespec2c(struct timespec time, char *buffer, int len)
/****************************************************************************/
{
  memset(buffer,0,sizeof(char)*len);
  snprintf(buffer,len,"%li.%09li",(long int)time.tv_sec,(long int)time.tv_nsec);
}

/****************************************************************************/
double cs_ts_timeval2d(struct timeval time)
/****************************************************************************/
{
    return (double)time.tv_sec + (double)time.tv_usec * 1.0e-6;
}

/****************************************************************************/
struct timeval cs_ts_assembletimeval(int sec, int usec)
/****************************************************************************/
{
  struct timeval rv;
  rv.tv_sec = sec;
  rv.tv_usec = usec;
  return rv;
}

/** Sleep relative timeout in ms
 */
/*************************************************************************************************/
void cs_ts_sleep(uint32_t timeout)
/*************************************************************************************************/
{
    struct timespec ts = cs_ts_timeout2ts(timeout);
    nanosleep(&ts, NULL);
}

/** Sleep absolute timeout in ms
 */
/*************************************************************************************************/
void cs_ts_sleep_next(struct timespec *ts)
/*************************************************************************************************/
{
  clock_nanosleep(CLOCK_MONOTONIC,TIMER_ABSTIME,ts,NULL);
}
