/*
    Copyright (C) 2013 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <string.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/fcntl.h>
#include <unistd.h>
#include <stdarg.h>
#include <stdio.h>
#include <errno.h>

#include "config.h"
#include "cs_uds.h"
#include "cs_uds_private.h"
#include "cs_debug_private.h"

//UDS-API Return codes
/** \def CS_UDS_TIMEOUT
    Return code Socket timeout
*/
#define CS_UDS_TIMEOUT 2
/** \def CS_UDS_FUNC_FAIL
    Return code Function call failed
*/
#define CS_UDS_FUNC_FAIL 3
/** \def CS_UDS_CYCLIC
    Return code Timeout by Cyclic data overload
*/
#define CS_UDS_CYCLIC 4
/** \def CS_UDS_NEGRESP
    Return code Recv negative response
*/
#define CS_UDS_NEGRESP 5

/** \def CS_UDS_TXFULL
    Return code CAN Transmit queque is full, retry
*/
#define CS_UDS_TXFULL 6

/** \def CS_UDS_RESPSUP
    Return cde for Response supressed request
*/
#define CS_UDS_RESPSUP 7

/** \def CS_UDS_MAXRETRY
    Max number of retry sending
*/
#define CS_UDS_MAXRETRY 10

#define _uds_error(...) (cs_error(CS_DEBUG_UDS,__func__,__LINE__,__VA_ARGS__))
#define _uds_warn(...) (cs_warn(CS_DEBUG_UDS,__func__,__LINE__,__VA_ARGS__))
#define _uds_info(...) (cs_info(CS_DEBUG_UDS,__func__,__LINE__,__VA_ARGS__))
#define _uds_debug(...) (cs_debug(CS_DEBUG_UDS,__func__,__LINE__,__VA_ARGS__))
#define _uds_verbose(...) (cs_verbose(CS_DEBUG_UDS,__func__,__LINE__,__VA_ARGS__))
#define _uds_error_buffer(DATA,LEN) (cs_error_buffer(CS_DEBUG_UDS,__func__,__LINE__,DATA,LEN))
#define _uds_warn_buffer(DATA,LEN) (cs_warn_buffer(CS_DEBUG_UDS,__func__,__LINE__,DATA,LEN))
#define _uds_info_buffer(DATA,LEN) (cs_info_buffer(CS_DEBUG_UDS,__func__,__LINE__,DATA,LEN))
#define _uds_debug_buffer(DATA,LEN) (cs_debug_buffer(CS_DEBUG_UDS,__func__,__LINE__,DATA,LEN))
#define _uds_verbose_buffer(DATA,LEN) (cs_verbose_buffer(CS_DEBUG_UDS,__func__,__LINE__,DATA,LEN))


/********************************************************************/
cs_debug_level_t cs_uds_debuglevel(
				   cs_debug_level_t level
				   )
/********************************************************************/
{
  return cs_debuglevel(CS_DEBUG_UDS,level);
}

/********************************************************************/
void cs_uds_debug_setcallback(
			      void *ctx,
			      void (*callb)(void *ctx,
					    cs_debug_sub_t sub,
					    cs_debug_level_t dl,
					    const char *func,
					    const int line,
					    const char *fmt,
					    va_list ap)
			      )
/********************************************************************/
{
  cs_debug_setcallback(CS_DEBUG_UDS,ctx,callb);
}

/********************************************************************/
void _uds_debug_packet(cs_uds_data_t *packet)
/********************************************************************/
{
#ifdef CS_DEBUG
  cs_debug_level_t dl = cs_uds_debuglevel(CS_DEBUG_LEVEL_READ);
  if(packet == NULL)
    return;

  if(dl >= CS_DEBUG_LEVEL_VERBOSE){
    _uds_verbose("=========================================\n");
    _uds_verbose("Mode           : %s\n",(packet->mt == DIAG?"Diagnostic":"Remote Diagnostic"));
    _uds_verbose("Source Id      : 0x%x\n",packet->sa);
    _uds_verbose("Destination Id : 0x%x\n",packet->ta);
    _uds_verbose("Adressing Mode : %s\n",(packet->tat == PHYS?"Physical":"Functional"));
    _uds_verbose("Adress extender: 0x%x\n",packet->ae);
    _uds_verbose("RA             : 0x%x\n",packet->ra);
    _uds_verbose("Service ID     : 0x%x\n",packet->pci.si);
    _uds_verbose("Neg Resp ID    : 0x%x\n",packet->pci.nrsi);
    _uds_verbose("Resp requested : %s\n",(packet->resp == RESPREQ?"Yes":"No"));
    _uds_verbose("Time recv.     : %li . %li\n",packet->tv_response.tv_sec,packet->tv_response.tv_usec);
    _uds_verbose("Data length    : %i\n",packet->lt);
    _uds_verbose_buffer(packet->a_data,packet->lt);
  }
#endif
}

/********************************************************************/
void _reset_error(cs_uds_service_t *data)
/********************************************************************/
{
  if(data)
    data->lerror = CS_UDS_ERROR_PR;
}


/********************************************************************/
int _reset_timing(cs_uds_service_t *data)
/********************************************************************/
{
  if(data==NULL)
    return CS_FAIL;

  /*Set default timing values*/
  /*Server*/
  data->tserver.P2CANServer = CS_UDS_P2CANSERVER;
  data->tserver.P2CANServerE = CS_UDS_P2CANSERVERE;

  /*Client*/
  data->tclient.pcandelta = CS_UDS_PCANDELTA;
  data->tclient.P2CANClient = CS_UDS_P2CANCLIENT + CS_UDS_PCANDELTA;
  data->tclient.P2CANClientE = CS_UDS_P2CANCLIENTE + CS_UDS_PCANDELTA;
  data->tclient.P3CANClientPhys = CS_UDS_P3CANCLIENTPHYS + CS_UDS_PCANDELTA;
  data->tclient.P3CANClientFunc = CS_UDS_P3CANCLIENTFUNC + CS_UDS_PCANDELTA;

  return CS_OK;
}

/********************************************************************/
int _reset_sockets(cs_uds_service_t *data)
/********************************************************************/
{
  if(data==NULL)
    return CS_FAIL;


  data->ta = 0;
  data->sa = 0;
  data->lerror = CS_UDS_ERROR_PR;
  _reset_timing(data);
  cs_ts_gettime(&data->ts_nextrequest);

  return CS_OK;
}

/*************************************************************************************************/
int _cs_uds_recv(cs_uds_service_t *data, 
		 cs_uds_data_t *packet,
		 uint8_t *buffer,
		 int lenmax,
		 uint32_t *lenrcv,
		 uint32_t timeout)
/*************************************************************************************************/
{
  int rv;
  *lenrcv = lenmax;
  
  rv = cs_isotp_recv(data->isotp,
		     buffer,
		     lenrcv, 
		     timeout);
  if(rv != CS_OK){
    _uds_error("ISOTP error %s\n",cs_isotp_getlasterrorstring(data->isotp));
    return CS_FAIL;
  }

  return CS_OK;
  
}

/********************************************************************/
int _cs_uds_send(cs_uds_service_t *data,
		 cs_uds_data_t *packet)
/********************************************************************/
{
  uint8_t buffer[CS_UDS_MAXBYTE];
  int len = 0;
  
  if(data == NULL)
    return CS_FAIL;

  if(packet == NULL)
    return CS_FAIL;
  
  memset(buffer,0,CS_UDS_MAXBYTE);
  if(packet->pci.nrsi != CS_UDS_ERROR_PR){ //negative message
    buffer[0] = packet->pci.nrsi;
    buffer[1] = packet->pci.si;
    memcpy(&buffer[2],packet->a_data,packet->lt);
    len = 2 + packet->lt;
    /*No response bit needed*/
  } else { //positive response
    buffer[0] = packet->pci.si;
    memcpy(&buffer[1],packet->a_data,packet->lt);
    buffer[1] |= packet->resp == RESPSUP?0x80:0;
    len = 1 + packet->lt;
  }

  if(cs_isotp_send(data->isotp,
		   buffer,
		   len)){
    switch(cs_isotp_getlasterror(data->isotp)){
    case ISOTP_ERROR_NOERROR:
      _uds_verbose("Sending successfully\n");
      return CS_OK;
      break;
    case ISOTP_ERROR_TXFULL:
      _uds_warn("TXFULL Error");
      return CS_UDS_TXFULL;
      break;
    default:
      _uds_error("Unspecific Error: %i\n",cs_isotp_getlasterror(data->isotp));
      return CS_FAIL;
    }
  }
  return CS_OK;
}

/*************************************************************************************************/
int _client_recv_and_prepare(cs_uds_data_t *packet,cs_uds_service_t *data, uint32_t timeout)
/*************************************************************************************************/
{
  uint32_t nbytes = CS_UDS_MAXBYTE;
  uint8_t buffer[CS_UDS_MAXBYTE];
   
  if(data == NULL)
    return CS_FAIL;

  _reset_error(data);
  
  if(packet == NULL)
    return CS_FAIL;  
  
  memset(packet->a_data,0,CS_UDS_MAXBYTE);
  memset(buffer,0,CS_UDS_MAXBYTE);
  
  //Set Ta to inform ISOTP Layer
  packet->ta = data->ta;
  packet->sa = data->sa;
  if(_cs_uds_recv(data, packet,buffer, 
		  CS_UDS_MAXBYTE, &nbytes,timeout) == CS_OK){
    _uds_verbose("Received successfully\n");
    if(nbytes < 1){
      _uds_error("Received zero size package\n");
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      return CS_FAIL;
    }
    /*Fill data structure*/
    packet->pci.nrsi = buffer[0] == CS_UDS_ERROR_SNSIAS ? CS_UDS_ERROR_SNSIAS : CS_UDS_ERROR_PR;
    packet->ta = data->ta;
    packet->sa = data->sa;
    if(packet->pci.nrsi){
      if(nbytes < 3){
	_uds_error("Received invalid failure package\n");
	data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
	return CS_FAIL;
      }
      packet->pci.si = buffer[1];
      packet->lt = nbytes -2;
      data->lerror = buffer[2];
      memcpy(packet->a_data,&buffer[2], packet->lt);
    } else {
      packet->pci.si = buffer[0];
      packet->lt = nbytes -1;
      data->lerror = CS_UDS_ERROR_PR;
      memcpy(packet->a_data,&buffer[1],packet->lt);
    }
    _uds_debug_packet(packet);

    //Calc next timestamp
    cs_ts_next(&data->ts_nextrequest,data->tserver.P2CANServer);
    return CS_OK;
  } else {
    if(cs_isotp_getlasterror(data->isotp) == ISOTP_ERROR_TIMEOUT){
      data->lerror = CS_UDS_ERROR_CLIENT_STIMEOUT;
    } else {
      _uds_error("ISOTP receive failure %s\n",cs_isotp_getlasterrorstring(data->isotp));
      data->lerror = CS_UDS_ERROR_CLIENT_ISOTP;
      return CS_FAIL;
    }

    //Calc next timestamp
    cs_ts_next(&data->ts_nextrequest,data->tserver.P2CANServer);

    _uds_debug("Timeout P2CANServer ISOTP RX socket\n");
    return CS_UDS_TIMEOUT;
  }
  return CS_OK;
}

/*************************************************************************************************/
int _client_check_recvmessage(cs_uds_service_t *data, cs_uds_data_t *rd, int explen, 
			      uint8_t sessionid)
/*************************************************************************************************/
{
  if(data == NULL)
    return CS_FAIL;

  _reset_error(data);

  if(rd == NULL)
    return CS_FAIL;

  if(rd->pci.nrsi == CS_UDS_ERROR_SNSIAS){
    _uds_debug("Recieved Neg Response\n");
      if(rd->lt >= 1){
	_uds_debug("Recieved Error code %x \n", rd->a_data[0]);
	data->lerror = rd->a_data[0];
      }
      return CS_UDS_NEGRESP;
  }
  if(rd->pci.si != (sessionid + CS_UDS_SI_RESPOFFSET)){
    _uds_debug("Recv unexpected SI: %x\n", rd->pci.si);
    //TODO: Callback framework for cyclic data
    return CS_UDS_CYCLIC;
  }

  if(explen > 0){
    if(rd->lt != explen){
      _uds_error("Expected %i Bytes, got %i Bytes \n", explen, rd->lt);
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      return CS_UDS_FUNC_FAIL;
    }
  }
  return CS_OK;
}

/*************************************************************************************************/
int _uds_pack_and_send(cs_uds_data_t *packet,cs_uds_service_t *data)
/*************************************************************************************************/
{
  int rv;
  int loop = 1;

  if(data == NULL)
    return CS_FAIL;

  _reset_error(data);

  if(packet == NULL){
    _uds_error("Invalid arguments\n");
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;
    return CS_FAIL; 
  }
  //get send timestamp
  cs_ts_gettimeval(&packet->tv_response);
  _uds_debug_packet(packet);

  if(packet->lt > (CS_UDS_MAXBYTE -1)){ //MAX - 1Byte SID
    data->lerror = CS_UDS_ERROR_CLIENT_REQDATA;
    _uds_error("Overrange max packet size\n");
    return CS_FAIL;
  }

  //Loop sending, if TX-Queque is full
  while(loop){
    //Sleep for next try
    cs_ts_sleep_next(&data->ts_nextrequest);
    rv = _cs_uds_send(data, packet);
    //Calc next timestamp
    cs_ts_next(&data->ts_nextrequest,data->tserver.P2CANServer);
    switch(rv){
    case CS_OK:
      loop = 0;
      _uds_verbose("Sending successfully\n");
      return CS_OK;
      break;
    case CS_FAIL:
      loop = 0;
      _uds_error("Socket failure\n");
      data->lerror = CS_UDS_ERROR_CLIENT_ISOTP;
      return CS_FAIL;
      break;
    case CS_UDS_TXFULL:
      _uds_warn("TX-Queque full, retry\n");
      loop++;
      break;
    default:
      loop = 0;
      _uds_error("Sending parameter invalid\n");
      data->lerror = CS_UDS_ERROR_CLIENT_REQDATA;
      return CS_FAIL;
      break;
    }

    if(loop > CS_UDS_MAXRETRY){
      //Sending was retried Xtimes, we should inform user
      loop = 0;
      _uds_error("TXFULL error\n");
      data->lerror = CS_UDS_ERROR_CLIENT_TXFULL;
      return CS_FAIL;
    }
  }
  return CS_OK;
}

/*************************************************************************************************/
int _client_wait_send_recv(cs_uds_service_t *data, cs_uds_data_t *sd, cs_uds_data_t *rd, 
                           int explen, uint8_t sessionid)
/*************************************************************************************************/
{
    int i = 0;
    int ret;
    //Set default timeout
    uint32_t timeout = data->tserver.P2CANServer;
    
    if(_uds_pack_and_send(sd,data)){
      _uds_error("Sending failed\n");
      return CS_FAIL;
    }

    //try to read
    memset(rd,0,sizeof(cs_uds_data_t));

    i=0; // Reset Loop counter
    while(!_client_recv_and_prepare(rd,data,timeout)){
        ret = _client_check_recvmessage(data, rd, explen, sessionid); 
        _uds_verbose("Ret: %i, Error: %s\n",ret,cs_uds_strlerror(data));
        switch(ret){
        case CS_OK:
	  _uds_verbose("Paket was accepted\n");
            return CS_OK;
            break;
        case CS_UDS_NEGRESP:
            if(data->lerror == CS_UDS_ERROR_RCRRP){
                //Try to recieve next package
                _uds_debug("Receive Pending response, set extended timeout, wait for answer\n");
                timeout = data->tserver.P2CANServerE;
                memset(rd,0,sizeof(cs_uds_data_t));
            } else {
                _uds_error("Function failed: %s\n",cs_uds_strlerror(data));
                return CS_FAIL;
            }
            break;
        case CS_FAIL:
        case CS_UDS_FUNC_FAIL:
            _uds_error("Function failed: %s\n",cs_uds_strlerror(data));
            return CS_FAIL;
            break;
        case CS_UDS_CYCLIC:
            timeout = data->tserver.P2CANServer;
            if(i > CS_UDS_MAXCYCLIC){
                _uds_error("Timeout by cylic data\n");
                return CS_FAIL;
            }
            i++;
            break;
        default:
            _uds_error("Unknown return code\n");
            return CS_FAIL;
        }
    }

    if(data->lerror != CS_UDS_ERROR_CLIENT_STIMEOUT){
      _uds_error("Failure\n");
      return CS_FAIL;
    }
    
    //Check for supressed response
    if(sd->resp == RESPSUP){
        _uds_verbose("Response is suppressed\n");
        if(data->lerror == CS_UDS_ERROR_CLIENT_STIMEOUT){
            _reset_error(data);
	    _uds_debug("Reset receive timeout failure, because Response is suppressed\n");
        }
        return CS_UDS_RESPSUP;
    } else {
      _uds_debug("Receive timeout failure, because Response is not suppressed\n");
      return CS_FAIL;
    }
    

    return CS_OK;
}

/*************************************************************************************************/
/*************************************************************************************************/
/*Interface functions */
/*************************************************************************************************/
/*************************************************************************************************/

/********************************************************************/
uint16_t cs_uds_getversion()
/********************************************************************/
{
  return CS_UDS_VERSION_MAGIC;
}

/*************************************************************************************************/
void cs_uds_reseterror(cs_uds_service_t *data)
/*************************************************************************************************/
{
  if(data == NULL)
    return;

  _reset_error(data);
}

/*************************************************************************************************/
uint16_t cs_uds_getlerror(cs_uds_service_t *data)
/*************************************************************************************************/
{
  if(data == NULL)
    return 0;

  return data->lerror;
}

/*************************************************************************************************/
char * cs_uds_strlerror(cs_uds_service_t *data)
/*************************************************************************************************/
{
  if(data == NULL)
    return 0;

  if((data->lerror == CS_UDS_ERROR_CLIENT_ISOTP) &&
     (data->isotp != NULL)
     ){
    return cs_isotp_getlasterrorstring(data->isotp);
  }
  
  return cs_uds_strlerrorid(data->lerror);
}

/*************************************************************************************************/
char * cs_uds_strlerrorid(uint16_t lerror)
/*************************************************************************************************/
{  
  switch(lerror){
  case CS_UDS_ERROR_PR:
    return "UDS: Positive response";
  case CS_UDS_ERROR_GR:
    return "UDS: General reject";
  case CS_UDS_ERROR_SNS:
    return "UDS: Service not supported";
  case CS_UDS_ERROR_SFNS:
    return "UDS: Subfunction not supported";
  case CS_UDS_ERROR_IMLOIF:
    return "UDS: Invalid message length or invalid format";
  case CS_UDS_ERROR_RTL:
    return "UDS: Response too long";
  case CS_UDS_ERROR_BRR:
    return "UDS: Busy repeat request";
  case CS_UDS_ERROR_CNC:
    return "UDS: Condition not correct";
  case CS_UDS_ERROR_RSE:
    return "UDS: Request sequence error";
  case CS_UDS_ERROR_NRFSC:
    return "UDS: No response form subnet component";
  case CS_UDS_ERROR_FPEORA:
    return "UDS: Failure prevents execution of requested Action" ;
  case CS_UDS_ERROR_ROOR:
    return "UDS: Request out of range";
  case CS_UDS_ERROR_SAD:
    return "UDS: Security access denied";
  case CS_UDS_ERROR_IK:
    return "UDS: Invalid key";
  case CS_UDS_ERROR_ENOA:
    return "UDS: Exceed number of attempts";
  case CS_UDS_ERROR_RTDNE:
    return "UDS: Required time delay not expired";
  case CS_UDS_ERROR_UDNA:
    return "UDS: Upload download not accepted";
  case CS_UDS_ERROR_TDS:
    return "UDS: Transfer data suspended";
  case CS_UDS_ERROR_GPF:
    return "UDS: General programming failure";
  case CS_UDS_ERROR_WBSC:
    return "UDS: Wrong block sequence counter"; 
  case CS_UDS_ERROR_RCRRP:
    return "UDS: Request correctly received-response pending";
  case CS_UDS_ERROR_SFNSIAS:
    return "UDS: Subfunction not supported in active session";
  case CS_UDS_ERROR_SNSIAS:
    return "UDS: Service not supported in active session";
  case CS_UDS_ERROR_RPMTH:
    return "UDS: RPM too high";
  case CS_UDS_ERROR_RPMTL:
    return "UDS: RPM too low";
  case CS_UDS_ERROR_EIR:
    return "UDS: Engine is running";
  case CS_UDS_ERROR_EINR:
    return "UDS: Engine is not running";
  case CS_UDS_ERROR_ERTTL:
    return "UDS: Runtime too low";
  case CS_UDS_ERROR_TEMPTH:
    return "UDS: Temperature too high";
  case CS_UDS_ERROR_TEMPTL:
    return "UDS: Temperature too low";
  case CS_UDS_ERROR_VSTH:
    return "UDS: Vehicle speed too high";
  case CS_UDS_ERROR_VSTL:
    return "UDS: Vehicle speed too low";
  case CS_UDS_ERROR_TPTH:
    return "UDS: Throttle pedal too high";
  case CS_UDS_ERROR_TPTL:
    return "UDS: Throttle pedal too low";
  case CS_UDS_ERROR_TRNIN:
    return "UDS: Transmisssion range not in neutral";
  case CS_UDS_ERROR_TRNIG:
    return "UDS: Transmission range not in gear";
  case CS_UDS_ERROR_BSNC:
    return "UDS: Brake switches not closed";
  case CS_UDS_ERROR_SLNIP:
    return "UDS: Shifter lever not in park";
  case CS_UDS_ERROR_TCCL:
    return "UDS: Torque converter clutch locked";
  case CS_UDS_ERROR_VTH:
    return "UDS: Voltage too high";
  case CS_UDS_ERROR_VTL:
    return "UDS: Voltage too low";
  case CS_UDS_ERROR_CLIENT_REQDATA:
    return "UDS: Client API: Requested data out of range";
  case CS_UDS_ERROR_CLIENT_RESPDATA:
    return "UDS: Client API: Response data out of range";
  case CS_UDS_ERROR_CLIENT_APICALL:
    return "UDS: Client API: API call failure";
  case CS_UDS_ERROR_CLIENT_SOCKET:
    return "Client API: Socket read/write failure";
  case CS_UDS_ERROR_CLIENT_SITABLE:
    return "Client API: SI table entry missing";
  case CS_UDS_ERROR_CLIENT_STIMEOUT:
    return "Client API: Server timeout";
  case CS_UDS_ERROR_CLIENT_MALLOC:
    return "Client API: Malloc failure";
  case CS_UDS_ERROR_CLIENT_IFACE:
    return "Client API: Hardware interface failure";
  case CS_UDS_ERROR_CLIENT_BUFFER:
    return "Client API: Internal Buffer size failure";
  case CS_UDS_ERROR_CLIENT_TXFULL:
    return "Client API: CAN TX-Queque full";
  case CS_UDS_ERROR_CLIENT_ISOTP:
    return "Client API: ISOTP layer";
  default:
    return "UDS: Unknown error";
  }
}

/*************************************************************************************************/
cs_uds_service_t *cs_uds_init()
/*************************************************************************************************/
{
  cs_uds_service_t *data = NULL;
  
  data = (cs_uds_service_t*)calloc(1,sizeof(cs_uds_service_t));
  
  if(data == NULL){
    _uds_error("Malloc failure\n");
    return NULL;
  }
  
  _reset_error(data);
  _reset_sockets(data);

  return data;
}

/*************************************************************************************************/
void cs_uds_free(
		 cs_uds_service_t *data
		 )
/*************************************************************************************************/
{
  if(data == NULL){
    return;
  }

  if(data->isotp != NULL){
    cs_isotp_close(data->isotp);
    cs_isotp_free(data->isotp);
  }

  free(data);
  data = NULL;
}


/*************************************************************************************************/
int cs_uds_can_init(
		    cs_uds_service_t *data,
		    char *iface,
		    uint32_t sa,
		    uint32_t ta
		    )
/*************************************************************************************************/
{
  if(data == NULL){
    _uds_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  if((iface == NULL) || (sa == 0) || (ta == 0)){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;
    _uds_error("Parameter failure\n");
    return CS_FAIL;
  }

  if(strlen(iface) >= CS_CHARBUF){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;
    _uds_error("Check inteface name length %s\n",iface);
    return CS_FAIL;
  } 

  memcpy(data->ifcan, iface, strlen(iface)+1);
  data->sa = sa;
  data->ta = ta;

  _uds_verbose("Channel SA:%i TA:%i on Interface\n",data->sa, data->ta,data->ifcan);

  _reset_timing(data);
  //TODO Handling functional adressing
  if((data->isotp = cs_isotp_init(data->sa,data->ta)) == NULL){
    data->lerror = CS_UDS_ERROR_CLIENT_IFACE;
    _uds_error("Failed to init ISOTP\n");
    return CS_FAIL;
  }
  
  //Enable TX_Padding for default, TODO is this really default?
  if(cs_isotp_setoptions(data->isotp,ISOTP_OPTION_TXPADDING)){
    _uds_error("Failed to set ISOTP TX padding: %s\n",cs_isotp_getlasterrorstring(data->isotp));
    data->lerror = CS_UDS_ERROR_CLIENT_IFACE;
    return CS_FAIL;
  }

  return CS_OK;
}

/*************************************************************************************************/
cs_isotp_t *cs_uds_get_isotp_handle(
				     cs_uds_service_t *data
				     )
/*************************************************************************************************/
{
  if(data == NULL){
    _uds_error("Parameter failure\n");
    return NULL;
  }
  
  if(data->isotp == NULL){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;
    _uds_error("CAN interface is not prepared\n");
    return NULL;
  }

  return data->isotp;
}

/*************************************************************************************************/
int cs_uds_can_open(
		    cs_uds_service_t *data
		    )
/*************************************************************************************************/
{
  if(data == NULL){
    _uds_error("Parameter failure\n");
    return CS_FAIL;
  }

  if(data->isotp == NULL){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;
    _uds_error("CAN interface is not prepared\n");
    return CS_FAIL;
  }
  
  if(cs_isotp_open(data->isotp,data->ifcan,NULL,0)){
    data->lerror = CS_UDS_ERROR_CLIENT_IFACE;
    _uds_error("Failed to open ISOTP: %s\n",cs_isotp_getlasterrorstring(data->isotp));
    return CS_FAIL;
  }

  return CS_OK;
}

/*************************************************************************************************/
int cs_uds_can_close(
		      cs_uds_service_t *data
		      )
/*************************************************************************************************/
{ 
  if(data == NULL){
    _uds_error("Parameter failure\n");
    return CS_FAIL;
  }

  if(data->isotp == NULL){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;
    _uds_error("CAN interface is not prepared\n");
    return CS_FAIL;
  }

  if(cs_isotp_close(data->isotp)){
    data->lerror = CS_UDS_ERROR_CLIENT_IFACE;
    _uds_error("Failed to close ISOTP: %s\n",cs_isotp_getlasterrorstring(data->isotp));
    return CS_FAIL;
  }
  
  return CS_OK;
}
  
/*************************************************************************************************/
int cs_uds_channel_reset_timing(
				cs_uds_service_t *data           /**< UDS data service structure*/
				)
/*************************************************************************************************/
{
  if(data == NULL)
    return CS_FAIL;

  _reset_error(data);
  _reset_timing(data);
  return CS_OK;
}
  
/*************************************************************************************************/
int cs_uds_channel_settiming(
			     cs_uds_service_t *data, 
			     struct cs_uds_time_server tserver,
			     uint16_t pcandelta
			     )
/*************************************************************************************************/
{
  if(data == NULL){
    _uds_error("Parameter failure\n");
    return CS_FAIL;
  }

  _reset_error(data);

  if((data->sa == 0) || (data->ta == 0)){
    _uds_error("Parameter failure\n");
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;
    return CS_FAIL;
  }

  data->tserver = tserver;

  //Warning if low value
  if(data->tserver.P2CANServer < 10){
    _uds_warn("P2CANServer low value %i ms detected\n",data->tserver.P2CANServer);
  }

  if(data->tserver.P2CANServerE < 50){
    _uds_warn("P2CANServerE low value %i ms detected\n",data->tserver.P2CANServerE);
  }
  
  //Check for zero values
  if(data->tserver.P2CANServer == 0){
    _uds_warn("Zero P2CANServer detected, set to default value\n");
    data->tserver.P2CANServer = CS_UDS_P2CANSERVER;
  }

  if(data->tserver.P2CANServerE == 0){
    _uds_warn("Zero P2CANServerE detected, set to default value\n");
    data->tserver.P2CANServerE = CS_UDS_P2CANSERVERE;
  }

  data->tclient.pcandelta = pcandelta;
  data->tclient.P2CANClient = data->tserver.P2CANServer + pcandelta;
  data->tclient.P2CANClientE = data->tserver.P2CANServerE + pcandelta;
  data->tclient.P3CANClientPhys = data->tserver.P2CANServer + pcandelta;
  data->tclient.P3CANClientFunc = data->tserver.P2CANServer + pcandelta;

  _uds_verbose("Set Client Timing\n");
  _uds_verbose("P2CANClient: %i ms\n",data->tclient.P2CANClient);
  _uds_verbose("P2CANClientE: %i ms\n",data->tclient.P2CANClientE);
  _uds_verbose("P3CANClientPhys: %i ms\n",data->tclient.P3CANClientPhys);
  _uds_verbose("P3CANClientFunc: %i ms\n",data->tclient.P3CANClientFunc);

  return CS_OK;
}

/*************************************************************************************************/
int cs_uds_channel_gettiming(cs_uds_service_t *data, struct cs_uds_time_server *tserver)
/*************************************************************************************************/
{
  if(data == NULL)
    return CS_FAIL;

  _reset_error(data);

  if(tserver == NULL){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;
    return CS_FAIL;
  }

  if((data->sa == 0) || (data->ta == 0)){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;
    return CS_FAIL;
  }
  *tserver = data->tserver;

  return CS_OK;
}

/*************************************************************************************************/
int cs_uds_receive_nonblock(cs_uds_service_t *data, uint32_t timeout)
/*************************************************************************************************/
{
  cs_uds_data_t packet;
  uint32_t nbytes = CS_UDS_MAXBYTE;
  uint8_t buffer[CS_UDS_MAXBYTE];
   
  if(data == NULL)
    return CS_FAIL;

  _reset_error(data);
    
  memset(packet.a_data,0,CS_UDS_MAXBYTE);
  memset(buffer,0,CS_UDS_MAXBYTE);
  
  //Set Ta to inform ISOTP Layer
  packet.ta = data->ta;
  packet.sa = data->sa;
  if(_cs_uds_recv(data,
		  &packet,buffer, 
		  CS_UDS_MAXBYTE,
		  &nbytes,
		  timeout
		  ) == CS_OK){
    /*Fill data structure*/
    packet.pci.nrsi = buffer[0] == CS_UDS_ERROR_SNSIAS ? CS_UDS_ERROR_SNSIAS : CS_UDS_ERROR_PR;
    packet.ta = data->ta;
    packet.sa = data->sa;
    if(packet.pci.nrsi){
      packet.pci.si = buffer[1];
      packet.lt = nbytes -2;
      data->lerror = buffer[2];
      memcpy(packet.a_data,&buffer[2], packet.lt);
    } else {
      packet.pci.si = buffer[0];
      packet.lt = nbytes -1;
      data->lerror = CS_UDS_ERROR_PR;
      memcpy(packet.a_data,&buffer[1],packet.lt);
    }

    _uds_debug_packet(&packet);

    return CS_OK;
  }
  return CS_OK;
}
 
/*************************************************************************************************/
int cs_uds_client_diagsession(cs_uds_service_t *data, uint8_t subfunc, 
			      resp_t resp, cs_uds_diagsession_t *rvalue)
/*************************************************************************************************/
{
  cs_uds_data_t sd;
  cs_uds_data_t rd;
  int ret;

  if(data == NULL)
    return CS_FAIL;

  _reset_error(data);  

  if(rvalue == NULL)
    return CS_FAIL;
  
  if((data->sa == 0) || (data->ta == 0)){
    _uds_error("Channel not configured\n");
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;
    return CS_FAIL;
  }

  if((data->ta == CS_UDS_FUNC_ID) && (resp != RESPSUP)){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Functional adressing with response request is not supported\n");
    return CS_FAIL;
  }
  
  /*Check subfuction*/
  switch(subfunc){
  case CS_UDS_SI_DSC_DS:
  case CS_UDS_SI_DSC_PRGS:
  case CS_UDS_SI_DSC_EXTDS:
  case CS_UDS_SI_DSC_SSDS:
  case CS_UDS_SI_DSC_EOL:
  case CS_UDS_SI_DSC_SUP:
    //supported
    _uds_verbose("Subfunction %x \n",subfunc);
    break;
  default:
    _uds_error("Session in not supported\n");
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;
    return CS_FAIL;
  }
  
  /*Prepare paket to send*/
  memset(&sd,0,sizeof(sd));
  sd.mt = DIAG;
  sd.sa = data->sa;
  sd.ta = data->ta;
  sd.tat = (data->ta == CS_UDS_FUNC_ID)?FUNC:PHYS;
  sd.ae = 0;
  sd.pci.si = CS_UDS_SI_DSC;
  sd.resp = resp;
  sd.a_data[0] = subfunc;
  sd.lt = sizeof(subfunc);

  rvalue->no = 0;

  ret = _client_wait_send_recv(data, &sd, &rd, 5, CS_UDS_SI_DSC);  
  if(ret){
      if(ret == CS_UDS_RESPSUP){
          return CS_OK;
      } else {
          return ret;
      }
  }

  
  //Build up return structure, only one is actually supported
  rvalue->sa[0] = rd.sa;
  // Data at buffer pos 1
  rvalue->tserver[0].P2CANServer = (((uint16_t)rd.a_data[1]) << 8) |  (uint16_t)rd.a_data[2];
  rvalue->tserver[0].P2CANServerE = (((uint16_t)rd.a_data[3]) << 8) |  (uint16_t)rd.a_data[4];
  _uds_verbose("Recv Times Id: %x ServerMax: %i ms  ServerMax*: %i ms\n",
	     rvalue->sa[0] = rd.sa,
	     rvalue->tserver[0].P2CANServer,
	     rvalue->tserver[0].P2CANServerE);
  rvalue->no = 1;
  
  return CS_OK;
}


/*************************************************************************************************/
int cs_uds_client_ecureset(cs_uds_service_t *data, uint8_t subfunc, 
			   resp_t resp, uint8_t *powerdown)
/*************************************************************************************************/
{
  cs_uds_data_t sd;
  cs_uds_data_t rd;
  int ret;

  if(data == NULL)
    return CS_FAIL;
  
  _reset_error(data);

  if((data->sa == 0) || (data->ta == 0)){
    _uds_error("Channel not configured\n");
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    return CS_FAIL;
  }
  
  if(data->ta == CS_UDS_FUNC_ID){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Functional adressing is not supported\n");
    return CS_FAIL;
  }

  /*Check subfuction*/
  switch(subfunc){
  case CS_UDS_SI_ER_HR:
  case CS_UDS_SI_ER_KOFFONR:
  case CS_UDS_SI_ER_SR:
  case CS_UDS_SI_ER_DRPSD:
    //supported
    _uds_verbose("Subfunction %x \n",subfunc);
    break;
  case CS_UDS_SI_ER_ERPSD:
    //Check return value pointer
    if(powerdown == NULL){
      data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
      return CS_FAIL;
    }
    //supported
    _uds_verbose("Subfunction %x \n",subfunc);
    break;
  default:
    _uds_error("Session in not supported\n");
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    return CS_FAIL;
  }
  
  /*Prepare paket to send*/
  memset(&sd,0,sizeof(sd));
  sd.mt = DIAG;
  sd.sa = data->sa;
  sd.ta = data->ta;
  sd.tat = (data->ta == CS_UDS_FUNC_ID)?FUNC:PHYS;
  sd.ae = 0;
  sd.pci.si = CS_UDS_SI_ER;
  sd.resp = resp;
  sd.a_data[0] = subfunc;
  sd.lt = sizeof(subfunc);

  ret = _client_wait_send_recv(data, &sd, &rd, -1,  CS_UDS_SI_ER);  
  if(ret){
      if(ret == CS_UDS_RESPSUP){
          return CS_OK;
      } else {
          return ret;
      }
  }
  
  if(subfunc == CS_UDS_SI_ER_ERPSD){
    //Check recieve data length
    if(rd.lt != 1){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      return CS_FAIL;
    }
    *powerdown = rd.a_data[0];
    _uds_verbose("Powerdowntime: %i \n",
		 *powerdown);
  }

  return CS_OK;
}

/*************************************************************************************************/
int cs_uds_client_securityaccess(cs_uds_service_t *data, uint8_t subfunc,
				 uint8_t *sdbuffer, int sdlen,
				 uint8_t *rcvbuffer, int *rcvlen)
/*************************************************************************************************/
{
  cs_uds_data_t sd;
  cs_uds_data_t rd;
  int ret;

  if(data == NULL)
    return CS_FAIL;
  
  _reset_error(data);

  if((data->sa == 0) || (data->ta == 0)){
    _uds_error("Channel not configured\n");
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    return CS_FAIL;
  }
  
  if(data->ta == CS_UDS_FUNC_ID){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Functional adressing is not supported\n");
    return CS_FAIL;
  }

  /*Check subfuction, and different levels*/
  if((subfunc == 0) || (subfunc > 0x43)){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Subfunction is not supported\n");
    return CS_FAIL;
  }

  /*Modulo 2*/
  /*RSD Level 1,3,5,...*/
  /*SK Level 2,4,6,...*/
  switch((subfunc & 0x1)){
  case CS_UDS_SI_SA_RSD: // 1, 3, 5, 7,...
    //supported
    _uds_verbose("Subfunction %x \n",subfunc);
    //Check for recv buffer
    if((rcvlen == NULL)||(rcvbuffer == NULL)){
      data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
      _uds_error("Check recieve parameter\n");
      return CS_FAIL;
    }
    break;
  case 0: //CS_UDS_SI_SA_SK // 2,4,6,8,...
    //supported
    _uds_verbose("Subfunction %x \n",subfunc);
    if((sdlen > 0)&&(sdbuffer == NULL)){
      data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
      _uds_error("Check send parameter\n");
      return CS_FAIL;
    }
    break;
  }


  /*Prepare paket to send*/
  memset(&sd,0,sizeof(sd));
  sd.mt = DIAG;
  sd.sa = data->sa;
  sd.ta = data->ta;
  sd.tat = PHYS;
  sd.ae = 0;
  sd.pci.si = CS_UDS_SI_SA;
  sd.resp = RESPREQ;
  sd.a_data[0] = subfunc;
  sd.lt = 1;

  //Check data length and pointer
  if(sdlen > 0){
    if((sdbuffer==NULL) && ((sdlen+1) > CS_UDS_MAXBYTE)){
      data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
      _uds_error("Invalid send buffer\n");
      return CS_FAIL;
    }
    /*Add optional sendbuffer*/
    sd.lt += sdlen;
    memcpy(&sd.a_data[1],sdbuffer,sdlen);
  } else {

  }

  ret = _client_wait_send_recv(data, &sd, &rd, -1, CS_UDS_SI_SA);  
  if(ret)
      return ret;

  if(rd.lt < 1){
    _uds_error("Expected minimal 1 Byte \n");
    data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
    return CS_UDS_FUNC_FAIL;
  }

  /*Handle return values*/
  if(subfunc & CS_UDS_SI_SA_RSD){

    //Check recv buffer len
    if(*rcvlen < (rd.lt -1)){
      _uds_error("Recv buffer overflow\n");
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      return CS_UDS_FUNC_FAIL; 
    }

    *rcvlen = rd.lt -1;
    memcpy(rcvbuffer,&rd.a_data[1],*rcvlen);
  }
 
  return CS_OK;
}


/*************************************************************************************************/
int cs_uds_client_controldtcsettings(cs_uds_service_t *data, uint8_t subfunc, resp_t resp)
/*************************************************************************************************/
{
  cs_uds_data_t sd;
  cs_uds_data_t rd;
  int ret;

  if(data == NULL)
    return CS_FAIL;

  _reset_error(data);
  
  if((data->sa == 0) || (data->ta == 0)){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Channel not configured\n");
    return CS_FAIL;
  }

  if(data->ta == CS_UDS_FUNC_ID){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Functional adressing is not supported\n");
    return CS_FAIL;
  }

  /*Check subfuction*/
  switch(subfunc){
  case CS_UDS_SI_CDTCS_ON:
  case CS_UDS_SI_CDTCS_OFF:
    //supported
    _uds_verbose("Subfunction %x \n",subfunc);
    break;
  default:
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Session in not supported\n");
    return CS_FAIL;
  }

  
  /*Prepare paket to send*/
  memset(&sd,0,sizeof(sd));
  sd.mt = DIAG;
  sd.sa = data->sa;
  sd.ta = data->ta;
  sd.tat = PHYS;
  sd.ae = 0;
  sd.pci.si = CS_UDS_SI_CDTCS;
  sd.resp = resp;
  sd.a_data[0] = subfunc;
  sd.lt = 1;

  ret = _client_wait_send_recv(data, &sd, &rd, 1, CS_UDS_SI_CDTCS);  
  if(ret){
      if(ret == CS_UDS_RESPSUP){
          return CS_OK;
      } else {
          return ret;
      }
  }
  return CS_OK;
}

/*************************************************************************************************/
int cs_uds_client_linkcontrol(cs_uds_service_t *data, uint8_t subfunc, resp_t resp, uint32_t sidata)
/*************************************************************************************************/
{
  cs_uds_data_t sd;
  cs_uds_data_t rd;
  int ret;
 
  if(data == NULL)
    return CS_FAIL;

  _reset_error(data);
  
  if((data->sa == 0) || (data->ta == 0)){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Channel not configured\n");
    return CS_FAIL;
  }

  if(data->ta == CS_UDS_FUNC_ID){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Functional adressing is not supported\n");
    return CS_FAIL;
  }

  /*Prepare paket to send*/
  memset(&sd,0,sizeof(sd));

  /*Check subfuction*/
  switch(subfunc){
  case CS_UDS_SI_LC_VMTWFP:
    switch(sidata){
    case CS_UDS_SI_LC_PC9600:
    case CS_UDS_SI_LC_PC19200:
    case CS_UDS_SI_LC_PC38400:
    case CS_UDS_SI_LC_PC57600:
    case CS_UDS_SI_LC_CAN125000:
    case CS_UDS_SI_LC_CAN250000:
    case CS_UDS_SI_LC_CAN500000:
    case CS_UDS_SI_LC_CAN1000000:
    case CS_UDS_SI_LC_PROGSU:
      //supported
      break;
    default:
      //not supported
      data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
      _uds_error("Data is not supported\n");
    }
    sd.a_data[0] = subfunc;
    sd.a_data[1] = (uint8_t) sidata;
    sd.lt = 2;
    break;
  case CS_UDS_SI_LC_VMTWSP:
    sd.a_data[0] = subfunc;
    /*24 Bit datarange*/
    sd.a_data[1] = (uint8_t)((sidata >> 16) & 0xFF);
    sd.a_data[2] = (uint8_t)((sidata >> 8) & 0xFF);
    sd.a_data[3] = (uint8_t)(sidata & 0xFF);
    sd.lt = 4;
    break;
  case CS_UDS_SI_LC_TM:
    sd.a_data[0] = subfunc;
    sd.lt = 1;
    break;
  default:
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Session is not supported\n");
    return CS_FAIL;
  }

  sd.mt = DIAG;
  sd.sa = data->sa;
  sd.ta = data->ta;
  sd.tat = PHYS;
  sd.ae = 0;
  sd.pci.si = CS_UDS_SI_LC;
  sd.resp = resp;

  ret = _client_wait_send_recv(data, &sd, &rd, 1, CS_UDS_SI_LC);  
  if(ret){
      if(ret == CS_UDS_RESPSUP){
          return CS_OK;
      } else {
          return ret;
      }
  }
  return CS_OK;
}

/*************************************************************************************************/
int cs_uds_client_testerpresent(cs_uds_service_t *data, resp_t resp)
/*************************************************************************************************/
{
  cs_uds_data_t sd;
  cs_uds_data_t rd;
  int rv = CS_OK;
 
  if(data == NULL)
    return CS_FAIL;

  _reset_error(data);
  
  if((data->sa == 0) || (data->ta == 0)){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Channel not configured\n");
    return CS_FAIL;
  }

  if((data->ta == CS_UDS_FUNC_ID) && (resp != RESPSUP)){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Functional adressing with response request is not supported\n");
    return CS_FAIL;
  }

  /*Prepare paket to send*/
  memset(&sd,0,sizeof(sd));

  sd.a_data[0] = 0x00;
  sd.lt = 1;
  sd.mt = DIAG;
  sd.sa = data->sa;
  sd.ta = data->ta;
  sd.tat = resp;
  sd.ae = 0;
  sd.pci.si = CS_UDS_SI_TP;
  sd.resp = resp;


  rv = _client_wait_send_recv(data, &sd, &rd, 1, CS_UDS_SI_TP); 
  if(rv){
      if(rv == CS_UDS_RESPSUP){
          return CS_OK;
      } else {
          return rv;
      }
  }

  return CS_OK;
}

/*************************************************************************************************/
int cs_uds_client_sleep(cs_uds_service_t *data, uint32_t time)
/*************************************************************************************************/
{
  uint32_t i = 0;

  if(data == NULL)
    return CS_FAIL;

  if(time == 0)
    return CS_OK;

  while(i < time){
    if(cs_uds_client_testerpresent(data, RESPSUP))
      return CS_FAIL;
    sleep(1);
    i++;
  }
  return CS_OK;
}

/*************************************************************************************************/
int cs_uds_client_readDatabyIdentifier(cs_uds_service_t *data, uint16_t did, 
				       int *len, void *rvdata)
/*************************************************************************************************/
{
  cs_uds_data_t sd;
  cs_uds_data_t rd;
  int rv;

  if(data == NULL)
    return CS_FAIL;

  _reset_error(data);

  if(len == NULL){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    return CS_FAIL;
  }

  if(rvdata == NULL){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    return CS_FAIL;
  }

  /*For dynamic data length*/
  /*if(*len == 0){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    return CS_FAIL;
    }*/
  
  if(did == 0){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    return CS_FAIL;
  }

  if(((*len) + 2) > CS_UDS_MAXBYTE){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Max Recievebuffer size reached\n");
    return CS_FAIL;
  }
  
  if((data->sa == 0) || (data->ta == 0)){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Channel not configured\n");
    return CS_FAIL;
  }

  if(data->ta == CS_UDS_FUNC_ID){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Functional adressing is not supported\n");
    return CS_FAIL;
  }

  
  /*Prepare paket to send*/
  memset(&sd,0,sizeof(sd));
  sd.mt = DIAG;
  sd.sa = data->sa;
  sd.ta = data->ta;
  sd.tat = PHYS;
  sd.ae = 0;
  sd.pci.si = CS_UDS_SI_RDB;
  sd.resp = RESPREQ;
  sd.a_data[0] = (did >>8) & 0xFF;
  sd.a_data[1] = did & 0xFF;
  sd.lt = 2;

  if((*len) >= 0){
    *len += 2;
  }

  rv = _client_wait_send_recv(data, &sd, &rd, *len, CS_UDS_SI_RDB); 
  if(rv)
      return rv;


  /*Get data out of rcv message*/
  *len = rd.lt-2;
  if(*len < 0)
    return CS_FAIL;

  memcpy(rvdata,&rd.a_data[2],*len);

  return CS_OK;
}

/*************************************************************************************************/
int cs_uds_client_writeDatabyIdentifier(cs_uds_service_t *data, uint16_t did, 
					uint16_t len, void *senddata)
/*************************************************************************************************/
{
  cs_uds_data_t sd;
  cs_uds_data_t rd;

  if(data == NULL)
    return CS_FAIL;

  _reset_error(data);

  if(senddata == NULL){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    return CS_FAIL;
  }

  if(len == 0){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    return CS_FAIL;
  }
  
  if(did == 0){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    return CS_FAIL;
  }
  
  if((len + 2) > CS_UDS_MAXBYTE){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Max Sendbuffer size reached\n");
    return CS_FAIL;
  }

  if((data->sa == 0) || (data->ta == 0)){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Channel not configured\n");
    return CS_FAIL;
  }

  if(data->ta == CS_UDS_FUNC_ID){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Functional adressing is not supported\n");
    return CS_FAIL;
  }

  
  /*Prepare paket to send*/
  memset(&sd,0,sizeof(sd));
  sd.mt = DIAG;
  sd.sa = data->sa;
  sd.ta = data->ta;
  sd.tat = PHYS;
  sd.ae = 0;
  sd.pci.si = CS_UDS_SI_WDB;
  sd.resp = RESPREQ;
  sd.a_data[0] = (did >>8) & 0xFF;
  sd.a_data[1] = did & 0xFF;
  memcpy(&sd.a_data[2],senddata,len);
  sd.lt = 2 +len;

  return _client_wait_send_recv(data, &sd, &rd, 2, CS_UDS_SI_WDB); 
}

/*************************************************************************************************/
int cs_uds_client_routinecontrol(cs_uds_service_t *data, uint8_t subfunc,
				 resp_t resp, uint16_t routine,
				 uint8_t *param, int rqlen,
				 uint8_t *routineinfo,
				 uint8_t *rvalue, int *resplen)
/*************************************************************************************************/
{
  cs_uds_data_t sd;
  cs_uds_data_t rd;
  int rv;
  int bufsize;
  int len;
 
  if(data == NULL)
    return CS_FAIL;

  _reset_error(data);
  
  if((rqlen + 3) > CS_UDS_MAXBYTE){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Max Sendbuffer size reached\n");
    return CS_FAIL;
  }

  if((param == NULL) && rqlen > 0){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Invalid parameter, check request parameter\n");
    return CS_FAIL;
  }

  /*Check response buffer parameter*/
  if((rvalue == NULL) != (resplen == NULL)){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Invalid parameter, check response parameter\n");
    return CS_FAIL;
  }

  if(routineinfo != NULL)
    bufsize = 4;
  else
    bufsize = 3;
  
  if((rvalue != NULL) && ((*resplen + bufsize) > CS_UDS_MAXBYTE)){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Max Recievebuffer size reached\n");
    return CS_FAIL;
  }

  if((data->sa == 0) || (data->ta == 0)){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Channel not configured\n");
    return CS_FAIL;
  }

  if(data->ta == CS_UDS_FUNC_ID){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Functional adressing is not supported\n");
    return CS_FAIL;
  }

  /*Check subfuction*/
  switch(subfunc){
  case CS_UDS_SI_RC_STR:
  case CS_UDS_SI_RC_STPR:
  case CS_UDS_SI_RC_RRR:
    break;
  default:
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Session is not supported\n");
    return CS_FAIL;
  }


  /*Prepare paket to send*/
  memset(&sd,0,sizeof(sd));

  sd.mt = DIAG;
  sd.sa = data->sa;
  sd.ta = data->ta;
  sd.tat = PHYS;
  sd.ae = 0;
  sd.pci.si = CS_UDS_SI_RC;
  sd.resp = resp;
  sd.a_data[0] = subfunc;
  sd.a_data[1] = (routine >>8) & 0xFF;
  sd.a_data[2] = routine & 0xFF;

  if((param!=NULL) && (rqlen >0)){
    memcpy(&sd.a_data[3],param,rqlen);
    sd.lt = rqlen + 3;
  } else {
    sd.lt = 3;
  }
  
  if(rvalue == NULL){
    /*Call Routine without return parameter*/
    if(routineinfo != NULL){
      rv = _client_wait_send_recv(data, &sd, &rd, 4, CS_UDS_SI_RC);  
      if(rv){
          if(rv == CS_UDS_RESPSUP){
              return CS_OK;
          } else {
              return rv;
          }
      }
      *routineinfo = rd.a_data[3];
      return CS_OK;
    } else {
      rv = _client_wait_send_recv(data, &sd, &rd, 3, CS_UDS_SI_RC);  
      if(rv){
          if(rv == CS_UDS_RESPSUP){
              return CS_OK;
          } else {
              return rv;
          }
      }
      return CS_OK;
    }
  } else {
    /*Call Routine with return parameter*/
    if(routineinfo !=  NULL){
      if(*resplen > 0){
	len = 4 + *resplen;
      } else {
	/*dynamic length*/
	len = -1;
      }
      rv = _client_wait_send_recv(data, &sd, &rd, len, CS_UDS_SI_RC);
      if(rv){
          if(rv == CS_UDS_RESPSUP){
              return CS_OK;
          } else {
              return rv;
          }
      }
      *resplen = rd.lt - 4;
      if(*resplen < 0){
	_uds_error("Expected minimal %i Bytes, got %i Bytes \n", 4, rd.lt);
	data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
	return CS_UDS_FUNC_FAIL;	
      }
      *routineinfo = rd.a_data[3];
      memcpy(rvalue,&rd.a_data[4],*resplen);
      _uds_verbose_buffer(rvalue, rd.lt);
      return CS_OK;
    } else {
      if(*resplen > 0){
	len = 3 + *resplen;
      } else {
	/*dynamic length*/
	len = -1;
      }
      rv = _client_wait_send_recv(data, &sd, &rd, len, CS_UDS_SI_RC);
      if(rv){
          if(rv == CS_UDS_RESPSUP){
              return CS_OK;
          } else {
              return rv;
          }
      }
      *resplen = rd.lt - 3;
      if(*resplen < 0){
	_uds_error("Expected minimal %i Bytes, got %i Bytes \n", 3, rd.lt);
	data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
	return CS_UDS_FUNC_FAIL;	
      }
      memcpy(rvalue,&rd.a_data[3],*resplen);
      _uds_verbose_buffer(rvalue, *resplen);
      return CS_OK;
    }
  }
}


/*************************************************************************************************/
int cs_uds_client_ClearDiagnosticInformation(cs_uds_service_t *data, resp_t resp, uint32_t dtcmask)
/*************************************************************************************************/
{
  cs_uds_data_t sd;
  cs_uds_data_t rd;
  int rv;

  if(data == NULL)
    return CS_FAIL;

  _reset_error(data);
  
  if((data->sa == 0) || (data->ta == 0)){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Channel not configured\n");
    return CS_FAIL;
  }

  if(data->ta == CS_UDS_FUNC_ID){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Functional adressing is not supported\n");
    return CS_FAIL;
  }

  /*Prepare paket to send*/
  memset(&sd,0,sizeof(sd));
  
  sd.a_data[0] = (uint8_t)((dtcmask & 0xFF0000) >> 16);
  sd.a_data[1] = (uint8_t)((dtcmask & 0x00FF00) >> 8);
  sd.a_data[2] = (uint8_t)(dtcmask & 0x0000FF);
  sd.lt = 3;
  sd.mt = DIAG;
  sd.sa = data->sa;
  sd.ta = data->ta;
  sd.tat = PHYS;
  sd.ae = 0;
  sd.pci.si = CS_UDS_SI_CDTCI;
  sd.resp = resp;

  rv = _client_wait_send_recv(data, &sd, &rd, 0, CS_UDS_SI_CDTCI);  
  if(rv){
      if(rv == CS_UDS_RESPSUP){
          return CS_OK;
      } else {
          return rv;
      }
  }
  return CS_OK;
}

/*************************************************************************************************/
int cs_uds_client_ReadDTCInformation(cs_uds_service_t *data, resp_t resp,
				     uint8_t subfunc, cs_uds_dtc_reqdata_t *reqdata,
				     cs_uds_dtc_respdata_t *respdata)
/*************************************************************************************************/
{
  cs_uds_data_t sd;
  cs_uds_data_t rd;
  int rv;
  div_t q;
  int i;
 
  if(data == NULL)
    return CS_FAIL;

  _reset_error(data);
  
  if(reqdata == NULL){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;
    return CS_FAIL;
  }

  if(respdata == NULL){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;
    return CS_FAIL;
  }
  
  if((data->sa == 0) || (data->ta == 0)){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Channel not configured\n");
    return CS_FAIL;
  }

  if(data->ta == CS_UDS_FUNC_ID){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Functional adressing is not supported\n");
    return CS_FAIL;
  }

  switch(subfunc){
  case CS_UDS_SI_RDTCI_RFTFDTC: //not tested
  case CS_UDS_SI_RDTCI_RFCDTC:  //not tested
  case CS_UDS_SI_RDTCI_RMRTFDTC: //not tested
  case CS_UDS_SI_RDTCI_RMRCDTC: //not tested
  case CS_UDS_SI_RDTCI_RMMDTCBSM: //not tested
  case CS_UDS_SI_RDTCI_RMMDEDRBDN: //not implemented
  case CS_UDS_SI_RDTCI_RNOMMDTCBSM: //not tested
  case CS_UDS_SI_RDTCI_RNOOEOBDDTCBSM: //not tested
  case CS_UDS_SI_RDTCI_ROBDDTCBSM: //not tested
  case CS_UDS_SI_RDTCI_RDTCFDC: //not fully implemented
  case CS_UDS_SI_RDTCI_RDTCWPS: //not tested
  case CS_UDS_SI_RDTCI_RDTCEDBR: //not implemented
  case CS_UDS_SI_RDTCI_RUDMDTCBSM: //not tested
  case CS_UDS_SI_RDTCI_RUDMDTCSSBDTC: //not implemented
  case CS_UDS_SI_RDTCI_RUDMDTCEDRBDN: //not implemented
  case CS_UDS_SI_RDTCI_RWWHOBDDTCBMR: //not tested
  case CS_UDS_SI_RDTCI_RWWHOBDDTCWPS: //not tested
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Subfunction is not supported\n");
    return CS_FAIL;
    break;
  default:
    //nothing todo
    i=0; //dummy
  }

  /*Prepare paket to send*/
  memset(&sd,0,sizeof(sd));

  sd.a_data[0] = subfunc;
  /*Check subfuction*/
  switch(subfunc){
  case CS_UDS_SI_RDTCI_RNODTCBSM:
  case CS_UDS_SI_RDTCI_RNOMMDTCBSM:
  case CS_UDS_SI_RDTCI_RNOOEOBDDTCBSM:
  case CS_UDS_SI_RDTCI_RDTCBSM:
  case CS_UDS_SI_RDTCI_RMMDTCBSM:
  case CS_UDS_SI_RDTCI_ROBDDTCBSM:
    sd.a_data[1] = reqdata->status_mask;
    sd.lt = 2;
    break;
  case CS_UDS_SI_RDTCI_RDTCSSI:
    sd.lt = 1;
    break;
  case CS_UDS_SI_RDTCI_RDTCSSBDTC:
    sd.a_data[1] = (uint8_t)((reqdata->dtc_mask & 0xFF0000)>> 16);
    sd.a_data[2] = (uint8_t)((reqdata->dtc_mask & 0xFF00)>> 8);
    sd.a_data[3] = (uint8_t)(reqdata->dtc_mask & 0xFF);
    sd.a_data[4] = reqdata->snapdata_recno;
    sd.lt = 5;
    if(reqdata->snapdata_recno == 0xFF){
      data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
      _uds_error("Sorry, only explicit snaprecord adressing is supported\n");
      return CS_FAIL;
    }
    break;
  case CS_UDS_SI_RDTCI_RDTCSDBRN:
    if(reqdata->snapdata_recno == 0xFF){
      data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
      _uds_error("Sorry, only explicit snaprecord adressing is supported\n");
      return CS_FAIL;
    }
    sd.a_data[1] = reqdata->snapdata_recno;
    sd.lt = 2;
    break;
  case CS_UDS_SI_RDTCI_RNODTCBSMR:
  case CS_UDS_SI_RDTCI_RDTCBSMR:
    sd.a_data[1] = reqdata->severeity_mask;
    sd.a_data[2] = reqdata->status_mask;
    sd.lt = 3;
    break;
  case CS_UDS_SI_RDTCI_RSIODTC:
    sd.a_data[1] = (uint8_t)((reqdata->dtc_mask & 0xFF0000)>> 16);
    sd.a_data[2] = (uint8_t)((reqdata->dtc_mask & 0xFF00)>> 8);
    sd.a_data[3] = (uint8_t)(reqdata->dtc_mask & 0xFF);
    sd.lt = 4;
    break;
  case CS_UDS_SI_RDTCI_RDTCEDRBDN:
    if(reqdata->extdata_recno == 0xFF){
      data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
      _uds_error("Sorry, only explicit extrecord adressing is supported\n");
      return CS_FAIL;
    }
    sd.a_data[1] = (uint8_t)((reqdata->dtc_mask & 0xFF0000)>> 16);
    sd.a_data[2] = (uint8_t)((reqdata->dtc_mask & 0xFF00)>> 8);
    sd.a_data[3] = (uint8_t)(reqdata->dtc_mask & 0xFF);
    sd.a_data[4] = reqdata->extdata_recno;
    sd.lt = 5;
    break;
  case CS_UDS_SI_RDTCI_RSUPDTC:
  case CS_UDS_SI_RDTCI_RFTFDTC:
  case CS_UDS_SI_RDTCI_RFCDTC:
  case CS_UDS_SI_RDTCI_RMRTFDTC:
  case CS_UDS_SI_RDTCI_RMRCDTC:
  case CS_UDS_SI_RDTCI_RDTCWPS:
    sd.lt = 1;
    break;
  case CS_UDS_SI_RDTCI_RUDMDTCBSM:
    sd.a_data[1] = reqdata->status_mask;
    sd.a_data[2] = reqdata->memory_select;
    sd.lt = 3;
    break;
  case CS_UDS_SI_RDTCI_RWWHOBDDTCBMR:
    sd.a_data[1] = reqdata->group_id;
    sd.a_data[2] = reqdata->severeity_mask;
    sd.a_data[3] = reqdata->status_mask;
    sd.lt = 4;    
    break;
  case CS_UDS_SI_RDTCI_RWWHOBDDTCWPS:
    sd.a_data[1] = reqdata->group_id;
    sd.lt = 2;    
  default:
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Session is not supported\n");
    return CS_FAIL;
  }

  sd.mt = DIAG;
  sd.sa = data->sa;
  sd.ta = data->ta;
  sd.tat = PHYS;
  sd.ae = 0;
  sd.pci.si = CS_UDS_SI_RDTCI;
  sd.resp = resp;


  rv = _client_wait_send_recv(data, &sd, &rd,-1, CS_UDS_SI_RDTCI);
  if(rv){
      if(rv == CS_UDS_RESPSUP){
          return CS_OK;
      } else {
          return rv;
      }
  }

  /*Reset response data*/
  memset(respdata,0,sizeof(cs_uds_dtc_respdata_t));

  /*Check response*/
  switch(subfunc){
  case CS_UDS_SI_RDTCI_RNODTCBSM:
  case CS_UDS_SI_RDTCI_RNODTCBSMR:
  case CS_UDS_SI_RDTCI_RNOMMDTCBSM:
  case CS_UDS_SI_RDTCI_RNOOEOBDDTCBSM:
    if(rd.lt != 5){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      return CS_FAIL;
    }
    respdata->avail_mask = rd.a_data[1];
    respdata->format = rd.a_data[2];
    respdata->count = ((uint16_t)rd.a_data[3])<<8;
    respdata->count |= ((uint16_t)rd.a_data[4]);
    break;
  case CS_UDS_SI_RDTCI_RFTFDTC:
  case CS_UDS_SI_RDTCI_RFCDTC:
  case CS_UDS_SI_RDTCI_RMRTFDTC:
  case CS_UDS_SI_RDTCI_RMRCDTC:
    if(rd.lt < 2){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length\n");
      return CS_FAIL;
    }
    respdata->avail_mask = rd.a_data[1];
    /*malloc data array*/
    q = div(rd.lt -2,3);
    if(q.rem != 0){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length\n");
      return CS_FAIL;
    }
    respdata->lt_statusdtcrec = q.quot;
    if(respdata->lt_statusdtcrec > 0){
      if(respdata->lt_statusdtcrec >  CS_UDS_MAXDTC){
	data->lerror = CS_UDS_ERROR_CLIENT_BUFFER;
	_uds_error("Buffer overrange\n");
	return CS_FAIL;
      }
      for(i = 0; i < respdata->lt_statusdtcrec; i++){
	respdata->statusdtcrec[i].dtc =  ((uint32_t) rd.a_data[2 + i *3 + 0]) << 16;
	respdata->statusdtcrec[i].dtc |= ((uint32_t) rd.a_data[2 + i *3 + 1]) << 8;
	respdata->statusdtcrec[i].dtc |= (uint32_t) rd.a_data[2 + i *3 + 2];
      }
    }
    break;
  case CS_UDS_SI_RDTCI_RDTCBSM:
  case CS_UDS_SI_RDTCI_RSUPDTC:
  case CS_UDS_SI_RDTCI_RMMDTCBSM:
  case CS_UDS_SI_RDTCI_ROBDDTCBSM:
  case CS_UDS_SI_RDTCI_RDTCWPS:
    if(rd.lt < 2){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length\n");
      return CS_FAIL;
    }
    respdata->avail_mask = rd.a_data[1];
    /*malloc data array*/
    q = div(rd.lt -2,4);
    if(q.rem != 0){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length, div= %i, rem = %i\n",q.quot,q.rem);
      return CS_FAIL;
    }
    respdata->lt_statusdtcrec = q.quot;
    if(respdata->lt_statusdtcrec > 0){
      if(respdata->lt_statusdtcrec >  CS_UDS_MAXDTC){
	data->lerror = CS_UDS_ERROR_CLIENT_BUFFER;
	_uds_error("Buffer overrange\n");
	return CS_FAIL;
      }
      for(i = 0; i < respdata->lt_statusdtcrec; i++){
	respdata->statusdtcrec[i].dtc =  ((uint32_t) rd.a_data[2 + i *4 + 0]) << 16;
	respdata->statusdtcrec[i].dtc |= ((uint32_t) rd.a_data[2 + i *4 + 1]) << 8;
	respdata->statusdtcrec[i].dtc |= (uint32_t) rd.a_data[2 + i *4 + 2];
	respdata->statusdtcrec[i].status = rd.a_data[2 + i *4 + 3];
      }
    }
    break;
  case CS_UDS_SI_RDTCI_RDTCSSI:
    if(rd.lt < 1){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length\n");
      return CS_FAIL;
    }
    /*malloc data array*/
    q = div(rd.lt -1,4);
    if(q.rem != 0){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length\n");
      return CS_FAIL;
    }
    respdata->lt_statusdtcrec = q.quot;
    if(respdata->lt_statusdtcrec > 0){
      if(respdata->lt_statusdtcrec >  CS_UDS_MAXDTC){
	data->lerror = CS_UDS_ERROR_CLIENT_BUFFER;
	_uds_error("Buffer overrange\n");
	return CS_FAIL;
      }
      for(i = 0; i < respdata->lt_statusdtcrec; i++){
	respdata->statusdtcrec[i].dtc =  ((uint32_t) rd.a_data[1 + i *4 + 0]) << 16;
	respdata->statusdtcrec[i].dtc |= ((uint32_t) rd.a_data[1 + i *4 + 1]) << 8;
	respdata->statusdtcrec[i].dtc |= (uint32_t) rd.a_data[1 + i *4 + 2];
	respdata->statusdtcrec[i].snaprecno = rd.a_data[1 + i *4 + 3];
      }
    }
    break;
  case CS_UDS_SI_RDTCI_RDTCSSBDTC:
    //Note only explicit snapshotrecord adressing is supported
    if(rd.lt < 1){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length\n");
      return CS_FAIL;
    }
    if((rd.lt < 6)&&(rd.lt != 5)){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length\n");
      return CS_FAIL;
    }
    respdata->lt_statusdtcrec = 1;
    //Get DTC and Status record
    respdata->statusdtcrec[0].dtc =  ((uint32_t) rd.a_data[1 + 0]) << 16;
    respdata->statusdtcrec[0].dtc |= ((uint32_t) rd.a_data[1 + 1]) << 8;
    respdata->statusdtcrec[0].dtc |= (uint32_t) rd.a_data[1 +  2];
    respdata->statusdtcrec[0].status = rd.a_data[1 + 3];

    if((rd.lt > 6)&&(rd.lt < 7)){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length\n");
      return CS_FAIL;
    }
    //Get Snapshot Id
    respdata->statusdtcrec[0].snaprecno = rd.a_data[1 + 4];
    //Get number of DID entries
    respdata->no_didrec = rd.a_data[1 + 5];

    if((respdata->no_didrec > 0)&&(rd.lt < 9)){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length\n");
      return CS_FAIL;
    }
    //Reset data storage
    memset(respdata->didrec,0,CS_UDS_MAXBYTE);
    //Get data
    respdata->lt_didrec = rd.lt - 7;
    memcpy(respdata->didrec,&rd.a_data[7],respdata->lt_didrec);

    break;
  case CS_UDS_SI_RDTCI_RDTCSDBRN:
    //Note only explicit snapshotrecord adressing is supported
    if(rd.lt < 2){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length\n");
      return CS_FAIL;
    }
    respdata->lt_statusdtcrec = 1;
    //Get Snapshot Id
    respdata->statusdtcrec[0].snaprecno = rd.a_data[1 + 0];
    if(rd.lt == 2)
      break;

    if((rd.lt < 7)&&(rd.lt != 6)){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length\n");
      return CS_FAIL;
    }

    //Get DTC and Status record
    respdata->statusdtcrec[0].dtc =  ((uint32_t) rd.a_data[1 + 1]) << 16;
    respdata->statusdtcrec[0].dtc |= ((uint32_t) rd.a_data[1 + 2]) << 8;
    respdata->statusdtcrec[0].dtc |= (uint32_t) rd.a_data[1 +  3];
    respdata->statusdtcrec[0].status = rd.a_data[1 + 4];

    if((rd.lt > 7)&&(rd.lt < 8)){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length\n");
      return CS_FAIL;
    }
    //Get number of DID entries
    respdata->no_didrec = rd.a_data[1 + 5];

    if((respdata->no_didrec > 0)&&(rd.lt < 10)){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length\n");
      return CS_FAIL;
    }
    //Reset data storage
    memset(respdata->didrec,0,CS_UDS_MAXBYTE);
    //Get data
    respdata->lt_didrec = rd.lt - 7;
    memcpy(respdata->didrec,&rd.a_data[7],respdata->lt_didrec);
    break;
 case CS_UDS_SI_RDTCI_RDTCEDRBDN:
   /*Note only explicit extrecord adressing is supported*/
   if(rd.lt < 1){
     data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
     _uds_error("Invalid response data length\n");
     return CS_FAIL;
   }
   if((rd.lt < 6)&&(rd.lt != 5)){
     data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
     _uds_error("Invalid response data length\n");
     return CS_FAIL;
   }
   respdata->lt_statusdtcrec = 1;

   //Get DTC and Status record
   respdata->statusdtcrec[0].dtc =  ((uint32_t) rd.a_data[1 + 0]) << 16;
   respdata->statusdtcrec[0].dtc |= ((uint32_t) rd.a_data[1 + 1]) << 8;
   respdata->statusdtcrec[0].dtc |= (uint32_t) rd.a_data[1 +  2];
   respdata->statusdtcrec[0].status = rd.a_data[1 + 3];    
   
   //Finish if there is no data segment
   if(rd.lt == 5){
     respdata->extdata.lt = 0;
     respdata->extdata.id = 0;
     return CS_OK;
   }

   //Check for recordno and number of objects
   if((rd.lt > 5) && (rd.lt  < 7)){
     data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
     _uds_error("Invalid response data length\n");
     return CS_FAIL;
   }
    
   //Is the recno correct
   if(rd.a_data[5] != reqdata->extdata_recno){
     data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
     _uds_error("Invalid Ext data record data id recieved\n");
     return CS_FAIL;
   }
   
   respdata->extdata.id = rd.a_data[5];
   respdata->extdata.lt = rd.lt - 6;

   if(respdata->extdata.lt > CS_UDS_MAXBYTE){
     data->lerror = CS_UDS_ERROR_CLIENT_BUFFER;
     _uds_error("Buffer overrange\n");
     return CS_FAIL;
   }
    
   //Now the data segment
   memcpy(respdata->extdata.data,&rd.a_data[6],respdata->extdata.lt);
   break;
  case CS_UDS_SI_RDTCI_RDTCBSMR:
  case CS_UDS_SI_RDTCI_RSIODTC:
    if(rd.lt < 2){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length\n");
      return CS_FAIL;
    }
    respdata->avail_mask = rd.a_data[1];
    /*malloc data array*/
    q = div(rd.lt -2,6);
    if(q.rem != 0){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length\n");
      return CS_FAIL;
    }
    respdata->lt_statusdtcrec = q.quot;
    if(respdata->lt_statusdtcrec > 0){
      if(respdata->lt_statusdtcrec >  CS_UDS_MAXDTC){
	data->lerror = CS_UDS_ERROR_CLIENT_BUFFER;
	_uds_error("Buffer overrange\n");
	return CS_FAIL;
      }
      for(i = 0; i < respdata->lt_statusdtcrec; i++){
	respdata->statusdtcrec[i].severity = rd.a_data[2 + i *6 + 0];
	respdata->statusdtcrec[i].functionalunit = rd.a_data[2 + i *6 + 1];
	respdata->statusdtcrec[i].dtc =  ((uint32_t) rd.a_data[2 + i *6 + 2]) << 16;
	respdata->statusdtcrec[i].dtc |= ((uint32_t) rd.a_data[2 + i *6 + 3]) << 8;
	respdata->statusdtcrec[i].dtc |= (uint32_t) rd.a_data[2 + i *6 + 4];
	respdata->statusdtcrec[i].status = rd.a_data[2 + i *6 + 5];
      }
    }    
    break;
  case CS_UDS_SI_RDTCI_RDTCFDC:
    if(rd.lt < 1){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length\n");
      return CS_FAIL;
    }
    /*malloc data array*/
    q = div(rd.lt -1,4);
    if(q.rem != 0){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length\n");
      return CS_FAIL;
    }
    respdata->lt_statusdtcrec = q.quot;
    if(respdata->lt_statusdtcrec > 0){
      if(respdata->lt_statusdtcrec >  CS_UDS_MAXDTC){
	data->lerror = CS_UDS_ERROR_CLIENT_BUFFER;
	_uds_error("Buffer overrange\n");
	return CS_FAIL;
      }
      for(i = 0; i < respdata->lt_statusdtcrec; i++){
	respdata->statusdtcrec[i].dtc =  ((uint32_t) rd.a_data[1 + i *4 + 0]) << 16;
	respdata->statusdtcrec[i].dtc |= ((uint32_t) rd.a_data[1 + i *4 + 1]) << 8;
	respdata->statusdtcrec[i].dtc |= (uint32_t) rd.a_data[1 + i *4 + 2];
	respdata->statusdtcrec[i].faultcounter = rd.a_data[1 + i *4 + 3];
      }
    }    
    break;
  case CS_UDS_SI_RDTCI_RUDMDTCBSM:
    if(rd.lt < 3){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length\n");
      return CS_FAIL;
    }
    respdata->avail_mask = rd.a_data[2];
    /*malloc data array*/
    q = div(rd.lt -3,4);
    if(q.rem != 0){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length\n");
      return CS_FAIL;
    }
    respdata->lt_statusdtcrec = q.quot;
    if(respdata->lt_statusdtcrec > 0){
      if(respdata->lt_statusdtcrec >  CS_UDS_MAXDTC){
	data->lerror = CS_UDS_ERROR_CLIENT_BUFFER;
	_uds_error("Buffer overrange\n");
	return CS_FAIL;
      }
      for(i = 0; i < respdata->lt_statusdtcrec; i++){
	respdata->statusdtcrec[i].dtc =  ((uint32_t) rd.a_data[3 + i *4 + 0]) << 16;
	respdata->statusdtcrec[i].dtc |= ((uint32_t) rd.a_data[3 + i *4 + 1]) << 8;
	respdata->statusdtcrec[i].dtc |= (uint32_t) rd.a_data[3 + i *4 + 2];
	respdata->statusdtcrec[i].status = rd.a_data[3 + i *4 + 3];
      }
    }    
    break;
  case CS_UDS_SI_RDTCI_RWWHOBDDTCBMR:
    if(rd.lt < 5){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length\n");
      return CS_FAIL;
    }
    respdata->functionalgroup = rd.a_data[1];
    respdata->avail_mask = rd.a_data[2];
    respdata->severity_mask = rd.a_data[3];
    respdata->format = rd.a_data[4];
    /*malloc data array*/
    q = div(rd.lt -5,5);
    if(q.rem != 0){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length\n");
      return CS_FAIL;
    }
    respdata->lt_statusdtcrec = q.quot;
    if(respdata->lt_statusdtcrec > 0){
      if(respdata->lt_statusdtcrec >  CS_UDS_MAXDTC){
	data->lerror = CS_UDS_ERROR_CLIENT_BUFFER;
	_uds_error("Buffer overrange\n");
	return CS_FAIL;
      }
      for(i = 0; i < respdata->lt_statusdtcrec; i++){
	respdata->statusdtcrec[i].severity =  rd.a_data[5 + i *5 + 0];
	respdata->statusdtcrec[i].dtc =  ((uint32_t) rd.a_data[5 + i *5 + 1]) << 16;
	respdata->statusdtcrec[i].dtc |= ((uint32_t) rd.a_data[5 + i *5 + 2]) << 8;
	respdata->statusdtcrec[i].dtc |= (uint32_t) rd.a_data[5 + i *5 + 3];
	respdata->statusdtcrec[i].status = rd.a_data[5 + i *5 + 4];
      }
    }    
    break;
  case CS_UDS_SI_RDTCI_RWWHOBDDTCWPS:
    if(rd.lt < 4){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length\n");
      return CS_FAIL;
    }
    respdata->functionalgroup = rd.a_data[1];
    respdata->avail_mask = rd.a_data[2];
    respdata->format = rd.a_data[3];
    /*malloc data array*/
    q = div(rd.lt -4,4);
    if(q.rem != 0){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length\n");
      return CS_FAIL;
    }
    respdata->lt_statusdtcrec = q.quot;
    if(respdata->lt_statusdtcrec > 0){
      if(respdata->lt_statusdtcrec >  CS_UDS_MAXDTC){
	data->lerror = CS_UDS_ERROR_CLIENT_BUFFER;
	_uds_error("Buffer overrange\n");
	return CS_FAIL;
      }
      for(i = 0; i < respdata->lt_statusdtcrec; i++){
	respdata->statusdtcrec[i].dtc =  ((uint32_t) rd.a_data[4 + i *4 + 0]) << 16;
	respdata->statusdtcrec[i].dtc |= ((uint32_t) rd.a_data[4 + i *4 + 1]) << 8;
	respdata->statusdtcrec[i].dtc |= (uint32_t) rd.a_data[4 + i *4 + 2];
	respdata->statusdtcrec[i].status = rd.a_data[4 + i *4 + 3];
      }
    }    
    break;
  default:
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Session is not supported\n");
    return CS_FAIL;
  }

  return CS_OK;
}

/*************************************************************************************************/
int _uds_client_request(
			  cs_uds_service_t *data,
			  uint8_t SI,
			  uint8_t compression,
			  uint8_t encryption,
			  uint64_t adress,
			  uint64_t len,
			  uint64_t *maxblocklen
			  )
/*************************************************************************************************/
{
  cs_uds_data_t sd;
  cs_uds_data_t rd;
  int rv;
  int i;
  int recvlen;


  if(data == NULL){
    _uds_error("Parameter failure\n");
    return CS_FAIL;
  }

  _reset_error(data);

  switch(SI){
  case CS_UDS_SI_RD:
    //Nothing
    break;
  case CS_UDS_SI_RU:
    //Nothing
    break;
  default:
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;
    _uds_error("Invalid subfunction\n");
    return CS_FAIL;
    break;
  }

  if(maxblocklen == NULL){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;
    _uds_error("Parameter maxblocklen failure\n");
    return CS_FAIL;
  }
  
  if((data->sa == 0) || (data->ta == 0)){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Channel not configured\n");
    return CS_FAIL;
  }

  if(data->ta == CS_UDS_FUNC_ID){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Functional adressing is not supported\n");
    return CS_FAIL;
  }

  
  /*Prepare paket to send*/
  memset(&sd,0,sizeof(sd));
  sd.mt = DIAG;
  sd.sa = data->sa;
  sd.ta = data->ta;
  sd.tat = PHYS;
  sd.ae = 0;
  sd.pci.si = SI;
  sd.resp = RESPREQ;
  sd.a_data[0] = ((compression & 0xF)<<4) | (encryption & 0xF);
  sd.a_data[1] = (0x4 << 4) | 0x4; //4Byte adress, 4 byte length
  sd.a_data[2] = (uint8_t)((adress >> 24) & 0xFF); 
  sd.a_data[3] = (uint8_t)((adress >> 16) & 0xFF); 
  sd.a_data[4] = (uint8_t)((adress >> 8) & 0xFF); 
  sd.a_data[5] = (uint8_t)(adress & 0xFF); 
  sd.a_data[6] = (uint8_t)((len >> 24) & 0xFF); 
  sd.a_data[7] = (uint8_t)((len >> 16) & 0xFF); 
  sd.a_data[8] = (uint8_t)((len >> 8) & 0xFF); 
  sd.a_data[9] = (uint8_t)(len & 0xFF); 
  sd.lt = 10;

  *maxblocklen = 0;
  //We do not know the recv data length
  rv = _client_wait_send_recv(data, &sd, &rd, -1, SI); 
  if(rv)
    return rv;


  /*Get data out of rcv message*/
  if(rd.lt < 2){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length\n");
      return CS_FAIL;
  }

  //Calculate the following data segment
  recvlen = (rd.a_data[0] >> 4) & 0xF;
  _uds_verbose("Block data length coded in %i bytes\n",recvlen);

  /*Get data out of rcv message*/
  if(rd.lt < (1 + recvlen)){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length\n");
      return CS_FAIL;
  }
  
  for(i = 0; i < recvlen; i++){
    *maxblocklen |= ((uint64_t)rd.a_data[i + 1]) << ((recvlen - (i +1)) * 8);
  }
  _uds_verbose("Block data length %i bytes\n",*maxblocklen);

  return CS_OK;
}

/*************************************************************************************************/
int cs_uds_client_requestDownload(
				  cs_uds_service_t *data,
				  uint8_t compression,
				  uint8_t encryption,
				  uint64_t adress,
				  uint64_t len,
				  uint64_t *maxblocklen
				  )
/*************************************************************************************************/
{
  return _uds_client_request(data,CS_UDS_SI_RD,compression,encryption,adress,len,maxblocklen);
}

/*************************************************************************************************/
int cs_uds_client_requestUpload(
				cs_uds_service_t *data,
				uint8_t compression,
				uint8_t encryption,
				uint64_t adress,
				uint64_t len,
				uint64_t *maxblocklen
				)
/*************************************************************************************************/
{
  return _uds_client_request(data,CS_UDS_SI_RU,compression,encryption,adress,len,maxblocklen);
}


/*************************************************************************************************/
int cs_uds_client_transferData(
			       cs_uds_service_t *data,
			       uint8_t SI,
			       uint8_t seqcounter,
			       uint64_t maxblocklen,
			       uint8_t *transferdata,
			       uint64_t *transferlen
			       )
/*************************************************************************************************/
{
  cs_uds_data_t sd;
  cs_uds_data_t rd;
  int rv;
  int i;
  uint64_t sendlen = 0;

  if(data == NULL)
    return CS_FAIL;

  if((transferdata == NULL) ||
     (transferlen == NULL)){
    _uds_error("Parameter failure\n");
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    return CS_FAIL;
  } 

  _reset_error(data);

  switch(SI){
  case CS_UDS_SI_RD:
    //Nothing
    break;
  case CS_UDS_SI_RU:
    //Nothing
    break;
  default:
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;
    _uds_error("Parameter failure\n");
    return CS_FAIL;
    break;
  }
  
  if((data->sa == 0) || (data->ta == 0)){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Channel not configured\n");
    return CS_FAIL;
  }

  if(data->ta == CS_UDS_FUNC_ID){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Functional adressing is not supported\n");
    return CS_FAIL;
  }

  if(maxblocklen > CS_UDS_MAXBYTE){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Block size buffer overrange\n");
    return CS_FAIL;
  }
  
  /*Prepare paket to send*/
  memset(&sd,0,sizeof(sd));
  sd.mt = DIAG;
  sd.sa = data->sa;
  sd.ta = data->ta;
  sd.tat = PHYS;
  sd.ae = 0;
  sd.pci.si = CS_UDS_SI_TD;
  sd.resp = RESPREQ;
  sd.a_data[0] = seqcounter;
  switch(SI){
  case CS_UDS_SI_RD:
    //Reduce block length by SI and seqcounter
    if((maxblocklen - 2) >= *transferlen){
      sendlen = *transferlen;
    }
    if((maxblocklen -2) < *transferlen){
      sendlen = maxblocklen;
    }
    _uds_verbose("Sendlen = %lli\n",sendlen);
    for(i = 0; i < sendlen; i++){
      sd.a_data[1 +i] = transferdata[i];
    }
    sd.lt = 1 + sendlen;
    break;
  case CS_UDS_SI_RU:
    sd.lt = 1;    
    break;
  }

  //We do not know the recv data length
  rv = _client_wait_send_recv(data, &sd, &rd, -1, CS_UDS_SI_TD); 
  if(rv)
    return rv;

  switch(SI){
  case CS_UDS_SI_RD:
    *transferlen = sendlen;
    break;
  case CS_UDS_SI_RU:
    if(rd.lt < 1){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length\n");
      return CS_FAIL;
    }
    if(rd.lt < 2){
      *transferlen = 0;
    } else{
      if(*transferlen < (rd.lt - 1)){
	data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
	_uds_error("Upload buffer overrange\n");
	return CS_FAIL;
      }
      *transferlen = rd.lt - 1;
      //copy data
      for(i = 0; i < *transferlen; i++){
	transferdata[i] = rd.a_data[1 + i];
      }
    }
    break;
  }
  return CS_OK;
}

/*************************************************************************************************/
int cs_uds_client_transferExit(
			       cs_uds_service_t *data,
			       uint8_t *senddata,
			       int sendlen,
			       uint8_t *recvdata,
			       int *recvlen
			       )
/*************************************************************************************************/
{
  cs_uds_data_t sd;
  cs_uds_data_t rd;
  int rv;

  if(data == NULL){
    _uds_error("Parameter failure\n");
    return CS_FAIL;
  }

  _reset_error(data);

  if((sendlen > 0) && (senddata == NULL)){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;
    _uds_error("Parameter failure\n");
    return CS_FAIL;
  }

  if((recvlen != 0) && (recvdata == NULL)){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;
    _uds_error("Parameter failure\n");
    return CS_FAIL;
  }

  if((recvlen == 0) && (recvdata != NULL)){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;
    _uds_error("Parameter failure\n");
    return CS_FAIL;
  }

  if(sendlen > CS_UDS_MAXBYTE){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;
    _uds_error("Parameter failure\n");
    return CS_FAIL;
  }

  if((data->sa == 0) || (data->ta == 0)){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Channel not configured\n");
    return CS_FAIL;
  }

  if(data->ta == CS_UDS_FUNC_ID){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    _uds_error("Functional adressing is not supported\n");
    return CS_FAIL;
  }
  
  /*Prepare paket to send*/
  memset(&sd,0,sizeof(sd));
  sd.mt = DIAG;
  sd.sa = data->sa;
  sd.ta = data->ta;
  sd.tat = PHYS;
  sd.ae = 0;
  sd.pci.si = CS_UDS_SI_RTE;
  sd.resp = RESPREQ;
  sd.lt = 0;
  if(sendlen > 0){
    memcpy(sd.a_data,senddata,sendlen);
    sd.lt += sendlen;
  }

  //We do not know the recv data length
  rv = _client_wait_send_recv(data, &sd, &rd, -1, CS_UDS_SI_RTE); 
  if(rv)
    return rv;


  /*Get data out of rcv message*/
  if(rd.lt > 0){
    if((!recvlen) || (*recvlen == 0)){
      data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
      _uds_error("Invalid response data length, expected no data\n");
      return CS_FAIL;
    }
    if(rd.lt > *recvlen){
      _uds_error("Invalid response data length, buffer to small\n");
      return CS_FAIL;
    }
    *recvlen = rd.lt;
    memcpy(recvdata,rd.a_data,*recvlen);
  }
  return CS_OK;
}

/*************************************************************************************************/
int cs_uds_client_api_transfer(
			       cs_uds_service_t *data,
			       uint8_t SI,
			       uint8_t compression,
			       uint8_t encryption,
			       uint64_t address,
			       uint8_t *transferdata,
			       uint64_t transferlen,
			       uint8_t *senddata,
			       int sendlen,
			       uint8_t *recvdata,
			       int *recvlen
			       )
/*************************************************************************************************/
{
  uint8_t seqcounter = 1;
  uint64_t maxblocklen;
  uint64_t transfered = 0;
  uint64_t blocktransfered = 0;

  if(data == NULL){
    _uds_error("Parameter failure\n");
    return CS_FAIL;
  }

  if((transferdata) && 
     (transferlen == 0)){
    _uds_error("Parameter failure\n");
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    return CS_FAIL;
  }

  switch(SI){
  case CS_UDS_SI_RD:
    if(cs_uds_client_requestDownload(data,compression,encryption,address,transferlen,&maxblocklen))
      return CS_FAIL;
    break;
  case CS_UDS_SI_RU:
    if(cs_uds_client_requestUpload(data,compression,encryption,address,transferlen,&maxblocklen))
      return CS_FAIL;
    break;
  default:
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    return CS_FAIL;
    break;
  }  

  //Reduce maxblocklen - SI Id and blockcounter
  blocktransfered = (maxblocklen - 2);
  while(transfered < transferlen){
    _uds_verbose("Transfer Segment size %lli, already transfered %lli\n",
		 transferlen,transfered);
    if(transfered > transferlen){
      _uds_error("Buffer failure %lli\n",transfered);
      data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
      return CS_FAIL;
    }
    if(transferlen < blocktransfered){
      blocktransfered = transferlen;
    }
    cs_uds_client_testerpresent(data, RESPSUP);
    if(cs_uds_client_transferData(data,SI,seqcounter,maxblocklen,&transferdata[transfered],&blocktransfered))
      return CS_FAIL;
    seqcounter++;
    transfered += blocktransfered;
    if((transferlen - transfered) < (maxblocklen -2)){
      //Block is not a full frame, take rest
      blocktransfered = transferlen - transfered;
    } else {
      //Block is a full frame
      blocktransfered = (maxblocklen - 2);
    }
  }
  _uds_verbose("Transfer finished Segment size %lli, already transfered %lli\n",
	       transferlen,transfered);
  if(cs_uds_client_transferExit(data,senddata,sendlen,recvdata,recvlen))
    return CS_FAIL;
  return CS_OK;
}
 

/*************************************************************************************************/
int cs_uds_client_api_flash_init(
				 cs_uds_client_flash_t *flash,
				 cs_uds_service_t *data,
				 uint8_t SI,
				 uint8_t compression,
				 uint8_t encryption,
				 uint64_t address,
				 uint8_t *transferdata,
				 uint64_t transferlen
				 )
/*************************************************************************************************/ 
{
  if(data == NULL){
    _uds_error("Parameter failure\n");
    return CS_FAIL;
  }

  if((flash == NULL) ||
     (transferdata == NULL) ||
     (transferlen == 0)
     ){
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;
    _uds_error("Parameter failure\n");
    return CS_FAIL;
  }

  flash->udsdata = data;
  flash->transferdata = transferdata;
  flash->transferlen = transferlen;
  flash->SI = SI;

  switch(flash->SI){
  case CS_UDS_SI_RD:
    if(cs_uds_client_requestDownload(flash->udsdata,compression,encryption,address,flash->transferlen,&flash->maxblocklen))
      return CS_FAIL;
    break;
  case CS_UDS_SI_RU:
    if(cs_uds_client_requestUpload(flash->udsdata,compression,encryption,address,flash->transferlen,&flash->maxblocklen))
      return CS_FAIL;
    break;
  default:
    data->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    return CS_FAIL;
    break;
  }  

  if(flash->maxblocklen <= 2){
    data->lerror = CS_UDS_ERROR_CLIENT_RESPDATA;
    _uds_error("Invalid block length\n");
    return CS_FAIL;
  }

  _uds_verbose("Block max size %lli\n",flash->maxblocklen);
  
  flash->seqcounter = 1;
  flash->transfered = 0;
 
  return CS_OK;
}

/*************************************************************************************************/
int cs_uds_client_api_flash_run(
				cs_uds_client_flash_t *flash
				)
/*************************************************************************************************/ 
{
  uint64_t blocktransfered = 0;

  if((flash == NULL) ||
     (flash->udsdata == NULL)){
    _uds_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  if((flash->transferlen - flash->transfered) < (flash->maxblocklen -2)){
    //Block is not a full frame, take rest
    blocktransfered = flash->transferlen - flash->transfered;
  } else {
    //Block is a full frame
    blocktransfered = (flash->maxblocklen - 2);
  }

  if(flash->transfered > flash->transferlen){
    _uds_error("Buffer failure %lli\n",flash->transfered);
    flash->udsdata->lerror = CS_UDS_ERROR_CLIENT_APICALL;    
    return CS_FAIL;
  }
  if(cs_uds_client_transferData(flash->udsdata,
				flash->SI,
				flash->seqcounter,
				flash->maxblocklen,
				&flash->transferdata[flash->transfered],
				&blocktransfered))
    return CS_FAIL;
  
  flash->seqcounter++;
  //Add the transfered byte to the general send bytes
  flash->transfered += blocktransfered;

  //Check if we are finished
  if(flash->transfered >=flash->transferlen)
    return 2;

  return CS_OK;
}

/*****************************************************************************/
/*****************************************************************************/
/** Server functions                                                         */
/*****************************************************************************/
/*****************************************************************************/

/*****************************************************************************/
void _uds_server_prepare_neg(
			     uint8_t failure,
			     const char* failurestr,
			     cs_uds_data_t *sendpacket,
			     uint8_t *buffer,
			     uint16_t *bufferlen
			     )
/*****************************************************************************/
{
  _uds_debug(failurestr);
  sendpacket->result = CS_FAIL; // Function not supported
  buffer[0] = failure;
  *bufferlen = 1;
}

/*************************************************************************************************/
int _uds_server_recv(cs_uds_service_t *data, 
		     uint8_t *buffer,
		     int lenmax,
		     uint32_t *lenrcv,
		     uint16_t timeout)
/*************************************************************************************************/
{
  int rv;
  *lenrcv = CS_UDS_MAXBYTE;
 
  rv = cs_isotp_recv(data->isotp, 
		     buffer,
		     lenrcv,
		     timeout);
  if(rv == CS_FAIL){
    if(cs_isotp_getlasterror(data->isotp) != ISOTP_ERROR_TIMEOUT){
      _uds_verbose("ISOTP error %s\n",cs_isotp_getlasterrorstring(data->isotp));
    }
    return CS_FAIL;
  }
  return CS_OK;
}

/*************************************************************************************************/
int _uds_server_recv_and_prepare(
				 cs_uds_service_t *data,
				 cs_uds_data_t *packet
				 )
/*************************************************************************************************/
{
  uint32_t nbytes;
  uint8_t buffer[CS_UDS_MAXBYTE];

  if(data == NULL)
    return CS_FAIL;

  data->lerror = CS_UDS_ERROR_PR; 
  
  if(packet == NULL)
    return CS_FAIL;  
  
  memset(packet->a_data,0,CS_UDS_MAXBYTE);
  memset(buffer,0,CS_UDS_MAXBYTE);
  
  if(!_uds_server_recv(data, 
		       buffer,
		       CS_UDS_MAXBYTE,
		       &nbytes,
		       0)){
    //Fill data structure
    /*switch(packet->tat){
    case FUNC:
      packet->ta = CS_UDS_FUNC_ID;
      packet->sa = data->ta; 
    case PHYS:
      packet->sa = data->ta;
      packet->ta = data->sa; 
      }*/
    //Actuallly only physical adressing is supported
    packet->sa = data->ta;
    packet->ta = data->sa; 
    packet->mt = DIAG;
    packet->lt = nbytes -1 ; //Skip SI
    packet->pci.si = buffer[0]; //& 0x3F; //extract si, skipping first bit and response offset
    switch(packet->pci.si){
    case CS_UDS_SI_DSC:
    case CS_UDS_SI_ER:
    case CS_UDS_SI_SA:
    case CS_UDS_SI_TP:
    case CS_UDS_SI_CDTCS:
    case CS_UDS_SI_LC:
      /*Subfunction and response bit are supported*/
      packet->resp = (buffer[1] & 0x80)? RESPSUP : RESPREQ;
      /*Remove response bit from buffer*/
      buffer[1] &= 0x7F;
      break;
    case CS_UDS_SI_RDB:
    default:
      packet->resp = RESPREQ;
      /*Subfunction and response bit are NOT supported*/
      break;
    }
    memcpy(packet->a_data,&buffer[1],packet->lt);
    
    _uds_debug_packet(packet);
    
    return CS_OK;
  } else {
    //Nothing received
    return CS_FAIL;
  }
    return CS_OK;
}

/*****************************************************************************/
void _server_handle_ds(cs_uds_service_t *data,
		       cs_uds_data_t *rcvpacket,
		       cs_uds_data_t *sendpacket,
		       uint8_t *buffer,
		       uint16_t *bufferlen)
/*****************************************************************************/
{
  int rv = CS_UDS_ERROR_PR; //No failure
  if(data->srv_cb.server_cb_diagsession == NULL){
    _uds_server_prepare_neg(CS_UDS_ERROR_SNS,"Server Diagsession: No callback registered\n",sendpacket,buffer,bufferlen);
    return;
  }
  
  if(rcvpacket->lt != 1){
    _uds_server_prepare_neg(CS_UDS_ERROR_IMLOIF,"Server Diagsession: Invalid format\n",sendpacket,buffer,bufferlen);
    return;
  }
  switch(rcvpacket->a_data[0]){ /*Subfunction*/
  case CS_UDS_SI_DSC_DS:
  case CS_UDS_SI_DSC_PRGS:
  case CS_UDS_SI_DSC_EXTDS:
  case CS_UDS_SI_DSC_SSDS:
  case CS_UDS_SI_DSC_EOL:
  case CS_UDS_SI_DSC_SUP:
    rv = (data->srv_cb.server_cb_diagsession)(data->srv_ctx,rcvpacket->a_data[0]);

    if(rv != CS_UDS_ERROR_PR){
      _uds_server_prepare_neg(rv,"Server Diagsession: Negative return value\n",sendpacket,buffer,bufferlen);
      return;
    }
    break;
  default:
    _uds_server_prepare_neg(CS_UDS_ERROR_SFNS,"Server Diagsession: Subfunction is not supported\n",sendpacket,buffer,bufferlen);
    return;
  }
  sendpacket->result = CS_OK;
  buffer[0] = rcvpacket->a_data[0];
  buffer[2] = (uint8_t)(data->tserver.P2CANServer & 0xFF);
  buffer[1] = (uint8_t)((data->tserver.P2CANServer >> 8) & 0xFF);
  buffer[4] = (uint8_t)(data->tserver.P2CANServerE & 0xFF);
  buffer[3] = (uint8_t)((data->tserver.P2CANServerE >> 8) & 0xFF);
  *bufferlen = 5;
}

/*****************************************************************************/
void _server_handle_er(cs_uds_service_t *data,
		       cs_uds_data_t *rcvpacket,
		       cs_uds_data_t *sendpacket,
		       uint8_t *buffer,
		       uint16_t *bufferlen)
/*****************************************************************************/
{
  int rv = CS_UDS_ERROR_PR; //No failure
  uint8_t powerDownTime = 0xFF; //not available
  
  if(data->srv_cb.server_cb_ecureset == NULL){
    _uds_server_prepare_neg(CS_UDS_ERROR_SNS,"Server ECU Reset: No callback registered\n",sendpacket,buffer,bufferlen);
    return;
  }
 
  if(rcvpacket->lt != 1){
    _uds_server_prepare_neg(CS_UDS_ERROR_IMLOIF,"Server ECU Reset: Invalid format\n",sendpacket,buffer,bufferlen);
    return;
  }
  switch(rcvpacket->a_data[0]){
  case CS_UDS_SI_ER_HR:
  case CS_UDS_SI_ER_KOFFONR:
  case CS_UDS_SI_ER_SR:
  case CS_UDS_SI_ER_ERPSD:
  case CS_UDS_SI_ER_DRPSD:
     rv = data->srv_cb.server_cb_ecureset(data->srv_ctx,rcvpacket->a_data[0], &powerDownTime);

     if(rv != CS_UDS_ERROR_PR){
       _uds_server_prepare_neg(rv,"Server ECU Reset: Negative return value\n",sendpacket,buffer,bufferlen);
       return;
     }
     break;
  default:
    _uds_server_prepare_neg(CS_UDS_ERROR_SFNS,"Server ECU Reset: Subfunction is not supported\n",sendpacket,buffer,bufferlen);
    return;
  }
  sendpacket->result = CS_OK;
  buffer[0] = rcvpacket->a_data[0];
  *bufferlen = 1;
  if(rcvpacket->a_data[0] == CS_UDS_SI_ER_ERPSD){
    buffer[1] = powerDownTime;
    *bufferlen = 2;
  }
}

/*****************************************************************************/
void _server_handle_sa(cs_uds_service_t *data,
		       cs_uds_data_t *rcvpacket,
		       cs_uds_data_t *sendpacket,
		       uint8_t *buffer,
		       uint16_t *bufferlen)
/*****************************************************************************/
{
  int rv = CS_UDS_ERROR_PR; //No failure
  int seedlen = CS_CHARBUF;
  
  if((data->srv_cb.server_cb_securityaccess_seed == NULL) || (data->srv_cb.server_cb_securityaccess_key == NULL)){
    _uds_server_prepare_neg(CS_UDS_ERROR_SNS,"Server Security Access: No callback registered\n",sendpacket,buffer,bufferlen);
    return;
  }
 
  if(rcvpacket->lt < 1){
    _uds_server_prepare_neg(CS_UDS_ERROR_IMLOIF,"Server Security Access: Invalid format\n",sendpacket,buffer,bufferlen);
    return;
  }

  if((rcvpacket->a_data[0] == 0) || (rcvpacket->a_data[0] >= 0x43)){
    _uds_server_prepare_neg(CS_UDS_ERROR_SFNS,"Server Security Access: Subfunction is not supported\n",sendpacket,buffer,bufferlen);
    return;
  }

  switch(rcvpacket->a_data[0] & 0x1){
  case CS_UDS_SI_SA_RSD: //subfunx 1,3,5,7,....
    rv = data->srv_cb.server_cb_securityaccess_seed(data->srv_ctx,rcvpacket->a_data[0], &buffer[1], &seedlen);
    if(rv != CS_UDS_ERROR_PR){
      _uds_server_prepare_neg(rv,"Server Security Access: Negative return value\n",sendpacket,buffer,bufferlen);
      return;
    }
    _uds_verbose("Request seed\n");
    sendpacket->result = CS_OK;
    buffer[0] = rcvpacket->a_data[0];
    *bufferlen = 1 + seedlen;
    return;
  case 0: //CS_UDS_SI_SA_SK subfunc 2,4,6,8,...
    rv = data->srv_cb.server_cb_securityaccess_key(data->srv_ctx,rcvpacket->a_data[0], &rcvpacket->a_data[1], rcvpacket->lt -1);
    if(rv != CS_UDS_ERROR_PR){
      _uds_server_prepare_neg(rv,"Server Security Access: Negative return value\n",sendpacket,buffer,bufferlen);
      return;
    }

    _uds_verbose("Authentication successfully\n");
    sendpacket->result = CS_OK;
    buffer[0] = rcvpacket->a_data[0];
    *bufferlen = 1;
    return;
  }
}

/*****************************************************************************/
void _server_handle_cdts(cs_uds_service_t *data,
			 cs_uds_data_t *rcvpacket,
			 cs_uds_data_t *sendpacket,
			 uint8_t *buffer,
			 uint16_t *bufferlen
			 )
/*****************************************************************************/
{
  int rv = CS_UDS_ERROR_PR; //No failure
  
  if(data->srv_cb.server_cb_controldtcsetting == NULL){
    _uds_server_prepare_neg(CS_UDS_ERROR_SNS,"Server Control DTC setting: No callback registered\n",sendpacket,buffer,bufferlen);
    return;
  }
 
  if(rcvpacket->lt != 1){
    _uds_server_prepare_neg(CS_UDS_ERROR_IMLOIF,"Server Conrol DTC setting: Invalid format\n",sendpacket,buffer,bufferlen);
    return;
  }

  switch(rcvpacket->a_data[0]){ /*Subfunction*/
  case CS_UDS_SI_CDTCS_ON:
  case CS_UDS_SI_CDTCS_OFF:
    rv = data->srv_cb.server_cb_controldtcsetting(data->srv_ctx,rcvpacket->a_data[0]);
    if(rv != CS_UDS_ERROR_PR){
      _uds_server_prepare_neg(rv,"Server Control DTC setting: Negative return value\n",sendpacket,buffer,bufferlen);
      return;
    }
    break;
  default:
    _uds_server_prepare_neg(CS_UDS_ERROR_SFNS,"Server Control DTC setting: Subfunction is not supported\n",sendpacket,buffer,bufferlen);
    return;
  }
  sendpacket->result = CS_OK;
  buffer[0] = rcvpacket->a_data[0];
  *bufferlen = 1;
}

/*****************************************************************************/
void _server_handle_lc(cs_uds_service_t *data,
		       cs_uds_data_t *rcvpacket,
		       cs_uds_data_t *sendpacket,
		       uint8_t *buffer,
		       uint16_t *bufferlen)
/*****************************************************************************/
{
  int rv = CS_UDS_ERROR_PR; //No failure
  
  if(data->srv_cb.server_cb_linkcontrol == NULL){
    _uds_server_prepare_neg(CS_UDS_ERROR_SNS,"Server Link control: No callback registered\n",sendpacket,buffer,bufferlen);
    return;
  }

  if(rcvpacket->lt < 1){
    _uds_server_prepare_neg(CS_UDS_ERROR_IMLOIF,"Server Link control: Invalid format\n",sendpacket,buffer,bufferlen);
    return;
  }
  
  switch(rcvpacket->a_data[0]){ /*Subfunction*/
  case CS_UDS_SI_LC_VMTWFP:
    if(rcvpacket->lt != 2){
      _uds_server_prepare_neg(CS_UDS_ERROR_IMLOIF,"Server Link control: Invalid format\n",sendpacket,buffer,bufferlen);
      return;
    }
    switch(rcvpacket->a_data[1]){
    case CS_UDS_SI_LC_PC9600:
    case CS_UDS_SI_LC_PC19200:
    case CS_UDS_SI_LC_PC38400:
    case CS_UDS_SI_LC_PC57600:
    case CS_UDS_SI_LC_CAN125000:
    case CS_UDS_SI_LC_CAN250000:
    case CS_UDS_SI_LC_CAN500000:
    case CS_UDS_SI_LC_CAN1000000:
    case CS_UDS_SI_LC_PROGSU:
      rv = data->srv_cb.server_cb_linkcontrol(data->srv_ctx,rcvpacket->a_data[0],0);
      if(rv != CS_UDS_ERROR_PR){
	_uds_server_prepare_neg(rv,"Server Link control: Negative return value\n",sendpacket,buffer,bufferlen);
	return;
      }
      break;
    default:
      _uds_server_prepare_neg(CS_UDS_ERROR_ROOR,"Server Link control: Data is not supported\n",sendpacket,buffer,bufferlen);
      return;
    }
    break;
  case CS_UDS_SI_LC_VMTWSP:
    if(rcvpacket->lt != 4){
      _uds_server_prepare_neg(CS_UDS_ERROR_IMLOIF,"Server Link control: Invalid format\n",sendpacket,buffer,bufferlen);
      return;
    }
    //Assemble
    uint32_t ldata = 0;
    ldata = (uint32_t)(rcvpacket->a_data[1]<<16);
    ldata |= (uint32_t)(rcvpacket->a_data[2]<<8);
    ldata |= (uint32_t)rcvpacket->a_data[3];
    rv = data->srv_cb.server_cb_linkcontrol(data->srv_ctx,rcvpacket->a_data[0],ldata);
    if(rv != CS_UDS_ERROR_PR){
      _uds_server_prepare_neg(rv,"Server Link control: Negative return value\n",sendpacket,buffer,bufferlen);
      return;
    }
    break;
  case CS_UDS_SI_LC_TM:
    if(rcvpacket->lt != 1){
      _uds_server_prepare_neg(CS_UDS_ERROR_IMLOIF,"Server Link control: Invalid format\n",sendpacket,buffer,bufferlen);
      return;
    }
    rv = data->srv_cb.server_cb_linkcontrol(data->srv_ctx,rcvpacket->a_data[0],0);
    if(rv != CS_UDS_ERROR_PR){
      _uds_server_prepare_neg(rv,"Server Link control: Negative return value\n",sendpacket,buffer,bufferlen);
      return;
    }
    break;
  default:
    _uds_server_prepare_neg(CS_UDS_ERROR_SFNS,"Server Control DTC setting: Subfunction is not supported\n",sendpacket,buffer,bufferlen);
    return;
  }
  sendpacket->result = CS_OK;
  buffer[0] = rcvpacket->a_data[0];
  *bufferlen = 1;
}

/*****************************************************************************/
void _server_handle_tp(cs_uds_service_t *data,
		       cs_uds_data_t *rcvpacket,
		       cs_uds_data_t *sendpacket,
		       uint8_t *buffer,
		       uint16_t *bufferlen)
/*****************************************************************************/
{
  int rv = CS_UDS_ERROR_PR; //No failure
  
  if(data->srv_cb.server_cb_testerpresent == NULL){
    _uds_server_prepare_neg(CS_UDS_ERROR_SNS,"Server Tester present: No callback registered\n",sendpacket,buffer,bufferlen);
    return;
  }

  if(rcvpacket->lt != 1){
    _uds_server_prepare_neg(CS_UDS_ERROR_IMLOIF,"Server Tester present: Invalid format\n",sendpacket,buffer,bufferlen);
    return;
  }

  rv = data->srv_cb.server_cb_testerpresent(data->srv_ctx);
  if(rv != CS_UDS_ERROR_PR){
    _uds_server_prepare_neg(rv,"Server Tester present: Negative return value\n",sendpacket,buffer,bufferlen);
    return;
  }
  
  sendpacket->result = CS_OK;
  buffer[0] = rcvpacket->a_data[0];
  *bufferlen = 1;
}

/*****************************************************************************/
void _server_handle_rdb(cs_uds_service_t *data,
			cs_uds_data_t *rcvpacket,
			cs_uds_data_t *sendpacket,
			uint8_t *buffer,
			uint16_t *bufferlen)
/*****************************************************************************/
{
  int rv = CS_UDS_ERROR_PR; //No failure
  uint16_t did;
  uint8_t didbuffer[CS_UDS_MAXBYTE];
  int didlen = CS_UDS_MAXBYTE;

  memset(didbuffer,0, sizeof(didbuffer));
  
  if(data->srv_cb.server_cb_readdatabyidentifier == NULL){
    _uds_server_prepare_neg(CS_UDS_ERROR_SNS,"Server Read Data by Identifier: No callback registered\n",sendpacket,buffer,bufferlen);
    return;
  }

  if(rcvpacket->lt != 2){ //TODO multi DID request
    _uds_server_prepare_neg(CS_UDS_ERROR_IMLOIF,"Server Read data by identifier: Invalid format\n",sendpacket,buffer,bufferlen);
    return;
  }
  did = (((uint16_t)rcvpacket->a_data[0]) << 8) & 0xFF00;
  did |= ((uint16_t)rcvpacket->a_data[1]) & 0xFF;

  rv = data->srv_cb.server_cb_readdatabyidentifier(data->srv_ctx,did,didbuffer,&didlen);
  if(rv != CS_UDS_ERROR_PR){
    _uds_server_prepare_neg(rv,"Server Read Data by identifier: Negative return value\n",sendpacket,buffer,bufferlen);
    return;
  }

  sendpacket->result = CS_OK;
  //Copy DID
  buffer[0] = rcvpacket->a_data[0];
  buffer[1] = rcvpacket->a_data[1];
  memcpy(&buffer[2],didbuffer,didlen);
  *bufferlen = 2 + didlen;
}

/*****************************************************************************/
void _server_handle_wdb(cs_uds_service_t *data,
			cs_uds_data_t *rcvpacket,
			cs_uds_data_t *sendpacket,
			uint8_t *buffer,
			uint16_t *bufferlen)
/*****************************************************************************/
{
  int rv = CS_UDS_ERROR_PR; //No failure

  uint16_t did;
  uint8_t didbuffer[CS_UDS_MAXBYTE];
  int didlen = CS_UDS_MAXBYTE;

  memset(didbuffer,0, sizeof(didbuffer));
  
  if(data->srv_cb.server_cb_writedatabyidentifier == NULL){
    _uds_server_prepare_neg(CS_UDS_ERROR_SNS,"Server Write Data by Identifier: No callback registered\n",sendpacket,buffer,bufferlen);
    return;
  }

  if(rcvpacket->lt <= 2){ //TODO multi DID request
    _uds_server_prepare_neg(CS_UDS_ERROR_IMLOIF,"Server Write Data by identifier: Invalid format\n",sendpacket,buffer,bufferlen);
    return;
  }

  did = (((uint16_t)rcvpacket->a_data[0]) << 8) & 0xFF00;
  did |= ((uint16_t)rcvpacket->a_data[1]) & 0xFF;

  didlen = rcvpacket->lt - 2;
  memcpy(didbuffer,&rcvpacket->a_data[2],sizeof(uint8_t)*didlen);

  rv = data->srv_cb.server_cb_writedatabyidentifier(data->srv_ctx,did,didbuffer,didlen);
  if(rv != CS_UDS_ERROR_PR){
    _uds_server_prepare_neg(rv,"Server Write Data by identifier: Negative return value\n",sendpacket,buffer,bufferlen);
    return;
  }

  sendpacket->result = CS_OK;
  //Copy DID
  buffer[0] = rcvpacket->a_data[0];
  buffer[1] = rcvpacket->a_data[1];
  *bufferlen = 2;
}

/*****************************************************************************/
void _server_handle_rc(cs_uds_service_t *data,
		       cs_uds_data_t *rcvpacket,
		       cs_uds_data_t *sendpacket,
		       uint8_t *buffer,
		       uint16_t *bufferlen)
/*****************************************************************************/
{
  uint16_t routine;
  uint8_t i;
  int rv = CS_UDS_ERROR_PR; //No failure
  uint8_t reqbuffer[CS_UDS_MAXBYTE];
  int reqlen = CS_UDS_MAXBYTE;
  uint8_t respbuffer[CS_UDS_MAXBYTE];
  int resplen = CS_UDS_MAXBYTE;
  uint8_t routineinfo = 0;
  uint8_t send_routineinfo = 0;

  memset(reqbuffer,0,sizeof(reqbuffer));
  memset(reqbuffer,0,sizeof(respbuffer));
  
  if(data->srv_cb.server_cb_routinecontrol == NULL){
    _uds_server_prepare_neg(CS_UDS_ERROR_SNS,"Server Routine control: No callback registered\n",sendpacket,buffer,bufferlen);
    return;
  }

  if(rcvpacket->lt < 3){
    _uds_server_prepare_neg(CS_UDS_ERROR_IMLOIF,"Server Routine control: Invalid format\n",sendpacket,buffer,bufferlen);
    return;
  }

  switch(rcvpacket->a_data[0]){
  case CS_UDS_SI_RC_STR:
  case CS_UDS_SI_RC_STPR:
  case CS_UDS_SI_RC_RRR:
    routine = (((uint16_t)rcvpacket->a_data[1]) << 8) & 0xFF00;
    routine |= ((uint16_t)rcvpacket->a_data[2]) & 0xFF;

    reqlen = rcvpacket->lt - 3;
    memcpy(reqbuffer,&rcvpacket->a_data[3],reqlen);

    rv = (data->srv_cb.server_cb_routinecontrol)(data->srv_ctx,rcvpacket->a_data[0],routine,reqbuffer,reqlen,respbuffer,&resplen,&routineinfo,&send_routineinfo);
    if(rv != CS_UDS_ERROR_PR){
      _uds_server_prepare_neg(rv,"Server Write Data by identifier: Negative return value\n",sendpacket,buffer,bufferlen);
      return;
    }
    sendpacket->result = CS_OK;
    i = 0;
    buffer[i++] = rcvpacket->a_data[0]; //Subfunction
    buffer[i++] = rcvpacket->a_data[1]; //HB Routine
    buffer[i++] = rcvpacket->a_data[2]; //LB Routine
    if(send_routineinfo){
      buffer[i++] = routineinfo;
    }

    if((CS_UDS_MAXBYTE - i -resplen) < 0){
      //Invalid resplen
      _uds_server_prepare_neg(CS_UDS_ERROR_GPF,"Server Routine control: Internal buffer failure\n",sendpacket,buffer,bufferlen);
      return;
    }
    memcpy(&buffer[i],respbuffer,sizeof(uint8_t)*resplen);
    *bufferlen = i + resplen;
    break;
  default:
    _uds_server_prepare_neg(CS_UDS_ERROR_SFNS,"Server Routine control: Subfunction is not supported\n",sendpacket,buffer,bufferlen);
    return;
  }

}

/*****************************************************************************/
void _server_handle_cdtci(cs_uds_service_t *data,
			  cs_uds_data_t *rcvpacket,
			  cs_uds_data_t *sendpacket,
			  uint8_t *buffer,
			  uint16_t *bufferlen)
/*****************************************************************************/
{
  int rv = CS_UDS_ERROR_PR; //No failure
  
  if(data->srv_cb.server_cb_cleardtcinformation == NULL){
    _uds_server_prepare_neg(CS_UDS_ERROR_SNS,"Server Clear DTC information: No callback registered\n",sendpacket,buffer,bufferlen);
    return;
  }

  if(rcvpacket->lt != 3){
    _uds_server_prepare_neg(CS_UDS_ERROR_IMLOIF,"Server Clear DTC information: Invalid format\n",sendpacket,buffer,bufferlen);
    return;
  }
  
  uint32_t dtcmask = 0;
  dtcmask = (uint32_t)(rcvpacket->a_data[0]<<16);
  dtcmask |= (uint32_t)(rcvpacket->a_data[1]<<8);
  dtcmask |= (uint32_t)rcvpacket->a_data[2];

  rv = data->srv_cb.server_cb_cleardtcinformation(data->srv_ctx,dtcmask);
  if(rv != CS_UDS_ERROR_PR){
    _uds_server_prepare_neg(rv,"Server Clear DTC information: Negative return value\n",sendpacket,buffer,bufferlen);
    return;
  }  
  sendpacket->result = CS_OK;  
  *bufferlen = 0;
}

/*****************************************************************************/
void _server_handle_rdtcipr(cs_uds_service_t *data,
			    cs_uds_data_t *rcvpacket,
			    cs_uds_data_t *sendpacket,
			    uint8_t *buffer,
			    uint16_t *bufferlen)
/*****************************************************************************/
{
  uint32_t dtc = 0;
  int i;
  int rv = CS_UDS_ERROR_PR; //No failure
  uint8_t subfunc;
  uint8_t dtcstatusmask = 0;
  uint8_t dtcseveritymask = 0;
  uint8_t dtcavailmask = 0;
  uint8_t format = 0;
  uint16_t count = 0;
  uint8_t severitybuffer[CS_UDS_MAXBYTE];
  uint8_t funcunitbuffer[CS_UDS_MAXBYTE];
  uint32_t dtcbuffer[CS_UDS_MAXBYTE];
  uint8_t statusbuffer[CS_UDS_MAXBYTE];
  int buflen = 0;
  
  if(rcvpacket->lt < 1){
    _uds_server_prepare_neg(CS_UDS_ERROR_IMLOIF,"Server Read DTC information: Invalid format\n",sendpacket,buffer,bufferlen);
    return;
  }

  subfunc = rcvpacket->a_data[0]; 
  /*Check subfuction*/
  switch(subfunc){
  case CS_UDS_SI_RDTCI_RFTFDTC:
  case CS_UDS_SI_RDTCI_RFCDTC:
  case CS_UDS_SI_RDTCI_RMRTFDTC:
  case CS_UDS_SI_RDTCI_RMRCDTC:
  case CS_UDS_SI_RDTCI_RMMDTCBSM:
  case CS_UDS_SI_RDTCI_RMMDEDRBDN:
  case CS_UDS_SI_RDTCI_RNOMMDTCBSM:
  case CS_UDS_SI_RDTCI_RNOOEOBDDTCBSM:
  case CS_UDS_SI_RDTCI_ROBDDTCBSM:
  case CS_UDS_SI_RDTCI_RDTCFDC:
  case CS_UDS_SI_RDTCI_RDTCWPS:
  case CS_UDS_SI_RDTCI_RDTCEDBR:
  case CS_UDS_SI_RDTCI_RUDMDTCBSM:
  case CS_UDS_SI_RDTCI_RUDMDTCSSBDTC:
  case CS_UDS_SI_RDTCI_RUDMDTCEDRBDN:
  case CS_UDS_SI_RDTCI_RWWHOBDDTCBMR:
  case CS_UDS_SI_RDTCI_RWWHOBDDTCWPS:
  case CS_UDS_SI_RDTCI_RDTCSSBDTC:
  case CS_UDS_SI_RDTCI_RDTCEDRBDN:
  case CS_UDS_SI_RDTCI_RDTCSDBRN:
    _uds_server_prepare_neg(CS_UDS_ERROR_SFNS,"Server Read DTC information: Subfunction is not supported\n",sendpacket,buffer,bufferlen);
    return;
  case CS_UDS_SI_RDTCI_RNODTCBSM:
  case CS_UDS_SI_RDTCI_RDTCBSM:
    if(rcvpacket->lt != 2){
      _uds_server_prepare_neg(CS_UDS_ERROR_IMLOIF,"Server Read DTC information: Invalid format\n",sendpacket,buffer,bufferlen);
      return;
    }
    dtcstatusmask = rcvpacket->a_data[1];
    break;
  case CS_UDS_SI_RDTCI_RNODTCBSMR:
  case CS_UDS_SI_RDTCI_RDTCBSMR:
    if(rcvpacket->lt != 3){
      _uds_server_prepare_neg(CS_UDS_ERROR_IMLOIF,"Server Read DTC information: Invalid format\n",sendpacket,buffer,bufferlen);
      return;
    }
    dtcseveritymask = rcvpacket->a_data[1];
    dtcstatusmask = rcvpacket->a_data[2];
    break;
  case CS_UDS_SI_RDTCI_RDTCSSI:
    if(rcvpacket->lt != 1){
      _uds_server_prepare_neg(CS_UDS_ERROR_IMLOIF,"Server Read DTC information: Invalid format\n",sendpacket,buffer,bufferlen);
      return;
    }
    break;
  case CS_UDS_SI_RDTCI_RSIODTC:
    if(rcvpacket->lt != 2){
      _uds_server_prepare_neg(CS_UDS_ERROR_IMLOIF,"Server Read DTC information: Invalid format\n",sendpacket,buffer,bufferlen);
      return;
    }
    dtc = (((uint32_t)rcvpacket->a_data[1])<<16) & 0xFF0000;
    dtc |= (((uint32_t)rcvpacket->a_data[2])<<8) & 0xFF00;
    dtc |= ((uint32_t)rcvpacket->a_data[3]) & 0xFF;
    break;
  case CS_UDS_SI_RDTCI_RSUPDTC:
    if(rcvpacket->lt != 1){
      _uds_server_prepare_neg(CS_UDS_ERROR_IMLOIF,"Server Read DTC information: Invalid format\n",sendpacket,buffer,bufferlen);
      return;
    }
    break;
  default:
    _uds_server_prepare_neg(CS_UDS_ERROR_SFNS,"Server Read DTC information: Subfunction is not supported\n",sendpacket,buffer,bufferlen);
    return;
  }

  /*Now handle return values*/
  switch(subfunc){
  case CS_UDS_SI_RDTCI_RNODTCBSM:
  case CS_UDS_SI_RDTCI_RNODTCBSMR:
    if(data->srv_cb.server_cb_readdtcinformation_getnumberofdtcbymask == NULL){
      _uds_server_prepare_neg(CS_UDS_ERROR_SNS,"Server Read DTC information: No callback registered\n",sendpacket,buffer,bufferlen);
      return;
    }
    rv = (data->srv_cb.server_cb_readdtcinformation_getnumberofdtcbymask)(data->srv_ctx,subfunc,dtcstatusmask,dtcseveritymask,&dtcavailmask,&format,&count);
    if(rv != CS_UDS_ERROR_PR){
      _uds_server_prepare_neg(rv,"Server Read DTC information: Negative return value\n",sendpacket,buffer,bufferlen);
      return;
    }
    sendpacket->result = CS_OK;  
    buffer[0] = subfunc; 
    buffer[1] = dtcavailmask;
    buffer[2] = format;
    buffer[3] = (uint8_t)((count & 0xFF00)>>8);
    buffer[4] = (uint8_t)(count & 0xFF);
    *bufferlen = 5;
    break;
  case CS_UDS_SI_RDTCI_RSUPDTC:
  case CS_UDS_SI_RDTCI_RDTCBSM:
    if(data->srv_cb.server_cb_readdtcinformation_getdtcbymask == NULL){
      _uds_server_prepare_neg(CS_UDS_ERROR_SNS,"Server Read DTC information: No callback registered\n",sendpacket,buffer,bufferlen);
      return;
    }
    rv = (data->srv_cb.server_cb_readdtcinformation_getdtcbymask)(data->srv_ctx,subfunc,dtcstatusmask,&dtcavailmask,dtcbuffer,statusbuffer,&buflen);
    if(rv != CS_UDS_ERROR_PR){
      _uds_server_prepare_neg(rv,"Server Read DTC information: Negative return value\n",sendpacket,buffer,bufferlen);
      return;
    }
    sendpacket->result = CS_OK;  
    buffer[0] = subfunc; 
    buffer[1] = dtcavailmask;
    //Check length
    if((2+ buflen *4) > CS_UDS_MAXBYTE){
      //Invalid resplen
      _uds_server_prepare_neg(CS_UDS_ERROR_GPF,"Server Read DTC information: Internal buffer failure\n",sendpacket,buffer,bufferlen);
      return;
    }
    for(i = 0; i < buflen; i++){
      buffer[2 + i*4 + 0] = (uint8_t)((dtcbuffer[i] & 0xFF0000)>>16);
      buffer[2 + i*4 + 1] = (uint8_t)((dtcbuffer[i] & 0xFF00)>>8);
      buffer[2 + i*4 + 2] = (uint8_t)(dtcbuffer[i] & 0xFF);
      buffer[2 + i*4 + 3] = statusbuffer[i];
    }
    *bufferlen = 2 +i*4;
    break;
  case CS_UDS_SI_RDTCI_RDTCSSI:
    if(data->srv_cb.server_cb_readdtcinformation_reportsnapshotidentification == NULL){
      _uds_server_prepare_neg(CS_UDS_ERROR_SNS,"Server Read DTC information: No callback registered\n",sendpacket,buffer,bufferlen);
      return;
    }
    rv = (data->srv_cb.server_cb_readdtcinformation_reportsnapshotidentification)(data->srv_ctx,dtcbuffer,statusbuffer,&buflen);
    if(rv != CS_UDS_ERROR_PR){
      _uds_server_prepare_neg(rv,"Server Read DTC information: Negative return value\n",sendpacket,buffer,bufferlen);
      return;
    }
    sendpacket->result = CS_OK;  
    buffer[0] = subfunc;
    //Check length
    if((1+ buflen *4) > CS_UDS_MAXBYTE){
      //Invalid resplen
      _uds_server_prepare_neg(CS_UDS_ERROR_GPF,"Server Read DTC information: Internal buffer failure\n",sendpacket,buffer,bufferlen);
      return;
    }
    for(i = 0; i < buflen; i++){
      buffer[1 + i*4 + 0] = (uint8_t)((dtcbuffer[i] & 0xFF0000)>>16);
      buffer[1 + i*4 + 1] = (uint8_t)((dtcbuffer[i] & 0xFF00)>>8);
      buffer[1 + i*4 + 2] = (uint8_t)(dtcbuffer[i] & 0xFF);
      buffer[1 + i*4 + 3] = statusbuffer[i];
    }
    *bufferlen = 1 +i*4;
    break;
  case CS_UDS_SI_RDTCI_RDTCBSMR:
  case CS_UDS_SI_RDTCI_RSIODTC:
    if(data->srv_cb.server_cb_readdtcinformation_reportdtcmaskrecord == NULL){
      _uds_server_prepare_neg(CS_UDS_ERROR_SNS,"Server Read DTC information: No callback registered\n",sendpacket,buffer,bufferlen);
      return;
    }
    rv = (data->srv_cb.server_cb_readdtcinformation_reportdtcmaskrecord)(data->srv_ctx,subfunc,dtcstatusmask,dtcseveritymask,
									 dtc,&dtcavailmask,severitybuffer,funcunitbuffer,
									 dtcbuffer,statusbuffer,&buflen);
    if(rv != CS_UDS_ERROR_PR){
      _uds_server_prepare_neg(rv,"Server Read DTC information: Negative return value\n",sendpacket,buffer,bufferlen);
      return;
    }

    //Check length
    if((2+ buflen *6) > CS_UDS_MAXBYTE){
      //Invalid resplen
      _uds_server_prepare_neg(CS_UDS_ERROR_GPF,"Server Read DTC information: Internal buffer failure\n",sendpacket,buffer,bufferlen);
      return;
    }
    
    sendpacket->result = CS_OK;  
    //Mirror subfunction
    buffer[0] = subfunc; 
    buffer[1] = dtcavailmask;
    
    for(i = 0; i < buflen; i++){
      buffer[2 + i * 6 + 0] = severitybuffer[i];
      buffer[2 + i * 6 + 1] = funcunitbuffer[i];
      buffer[2 + i * 6 + 2] = (uint8_t)((dtcbuffer[i] & 0xFF0000)>>16);
      buffer[2 + i * 6 + 3] = (uint8_t)((dtcbuffer[i] & 0xFF00)>>8);
      buffer[2 + i * 6 + 4] = (uint8_t)(dtcbuffer[i] & 0xFF);
      buffer[2 + i * 6 + 5] = statusbuffer[i];
    }
    *bufferlen = 2 + i * 6;
    break;
  default:
    _uds_server_prepare_neg(CS_UDS_ERROR_SFNS,"Server Routine control: Subfunction is not supported\n",sendpacket,buffer,bufferlen);
    return;
  }
}


/*****************************************************************************/
void _server_handle_uploaddownload(cs_uds_service_t *data,
				   cs_uds_data_t *rcvpacket,
				   cs_uds_data_t *sendpacket,
				   uint8_t *buffer,
				   uint16_t *bufferlen)
/*****************************************************************************/
{
  int rv = CS_UDS_ERROR_PR; //No failure
  uint8_t compression = 0;
  uint8_t encryption = 0;
  uint8_t address_len = 0;
  uint8_t segmentlength_len = 0;
  uint64_t address = 0;
  uint64_t segmentlength = 0;
  uint64_t blocksize = 0;
  uint64_t blocksize_cp = 0;
  
  uint8_t blocklength_len = 0;
  int i;

  if(data->srv_cb.server_cb_uploaddownload == NULL){
    _uds_server_prepare_neg(CS_UDS_ERROR_SNS,"Server Upload/Download request: No callback registered\n",sendpacket,buffer,bufferlen);
    return;
  }
  
  if(rcvpacket->lt < 2){
    _uds_server_prepare_neg(CS_UDS_ERROR_IMLOIF,"Server Upload/Download request: Invalid format\n",sendpacket,buffer,bufferlen);
    return;
  }
  
  compression = (rcvpacket->a_data[0] >> 4) & 0xF;
  encryption = rcvpacket->a_data[0] & 0xF;
  address_len = (rcvpacket->a_data[1] >> 4) & 0xF;
  segmentlength_len = rcvpacket->a_data[1] & 0xF;

  if(rcvpacket->lt < (2 + address_len + segmentlength_len)){
    _uds_server_prepare_neg(CS_UDS_ERROR_IMLOIF,"Server Upload/Download request: Invalid format\n",sendpacket,buffer,bufferlen);
    return;
  }
  
  for(i = 0; i < address_len; i++){
    address |= ((uint64_t)rcvpacket->a_data[2 + i]) << ((address_len - (i + 1))*8); 
  }
  for(i = 0; i < segmentlength_len; i++){
    segmentlength |= ((uint64_t)rcvpacket->a_data[2 + address_len + i]) << ((segmentlength_len - (i + 1))*8); 
  }
  
  rv = (data->srv_cb.server_cb_uploaddownload)(data->srv_ctx,rcvpacket->pci.si,compression, encryption,
					       address,segmentlength,&blocksize);
  if(rv != CS_UDS_ERROR_PR){
    _uds_server_prepare_neg(rv,"Server Read DTC information: Negative return value\n",sendpacket,buffer,bufferlen);
    return;
  }
  
  //Prepare response
  blocklength_len = 1;
  blocksize_cp = blocksize;
  for(i = 0; i < 8; i++){
    if((blocksize_cp = (blocksize_cp >> 8)) > 0){
      blocklength_len++;
    } else{
      break;
    }
  }
  sendpacket->result = CS_OK;
  buffer[0] = 0;
  buffer[0] = blocklength_len << 4;
  for(i = 0; i <blocklength_len; i++){
    buffer[1 + i] = (uint8_t)((blocksize >> (blocklength_len - (i + 1)) * 8) & 0xFF);
  }
  *bufferlen = 1 + blocklength_len;
}


/*****************************************************************************/
void _server_handle_transferData(cs_uds_service_t *data,
				 cs_uds_data_t *rcvpacket,
				 cs_uds_data_t *sendpacket,
				 uint8_t *buffer,
				 uint16_t *bufferlen)
/*****************************************************************************/
{
  int rv = CS_UDS_ERROR_PR; //No failure
  uint8_t transferdirection;
  uint8_t blockcounter = 0;
  int datalen = CS_UDS_MAXBYTE -1;
  uint8_t databuffer[CS_UDS_MAXBYTE];

  if(data->srv_cb.server_cb_transfer == NULL){
    _uds_server_prepare_neg(CS_UDS_ERROR_SNS,"Server Transfer: No callback registered\n",sendpacket,buffer,bufferlen);
    return;
  }

  if(rcvpacket->lt < 1){
    _uds_server_prepare_neg(CS_UDS_ERROR_IMLOIF,"Server Transfer: Invalid format\n",sendpacket,buffer,bufferlen);
    return;
  }
  
  if(rcvpacket->lt != 1){
    transferdirection = CS_UDS_SI_RD;
    datalen = rcvpacket->lt -1;
    memcpy(databuffer,&rcvpacket->a_data[1],sizeof(uint8_t)*datalen);
  } else {
    transferdirection = CS_UDS_SI_RU;
  }	  

  blockcounter = rcvpacket->a_data[0];

  rv = (data->srv_cb.server_cb_transfer)(data->srv_ctx,transferdirection,blockcounter,databuffer,&datalen);
  
  if(rv != CS_UDS_ERROR_PR){
    _uds_server_prepare_neg(rv,"Server Transfer: Negative return value\n",sendpacket,buffer,bufferlen);
    return;
  }
  
  sendpacket->result = CS_OK;
  buffer[0] = blockcounter;
  *bufferlen = 1;
  if(transferdirection == CS_UDS_SI_RU){
    memcpy(&buffer[1],databuffer,sizeof(char)*datalen);
    *bufferlen = 1 + datalen;
  }
}

/*****************************************************************************/
void _server_handle_transferExit(cs_uds_service_t *data,
				 cs_uds_data_t *rcvpacket,
				 cs_uds_data_t *sendpacket,
				 uint8_t *buffer,
				 uint16_t *bufferlen)
/*****************************************************************************/
{
  int rv = CS_UDS_ERROR_PR; //No failure
  int len = 0;
  if(data->srv_cb.server_cb_transferexit == NULL){
    _uds_server_prepare_neg(CS_UDS_ERROR_SNS,"Server Transfer exit: No callback registered\n",sendpacket,buffer,bufferlen);
    return;
  }

  rv = (data->srv_cb.server_cb_transferexit)(data->srv_ctx,rcvpacket->a_data,rcvpacket->lt,buffer,&len);
  
  if(rv != CS_UDS_ERROR_PR){
    _uds_server_prepare_neg(rv,"Server Transfer: Negative return value\n",sendpacket,buffer,bufferlen);
    return;
  }
  *bufferlen = len;
}

/*****************************************************************************/
void cs_uds_server_set_context(
			       cs_uds_service_t *data,
			       void *srv_ctx,
			       cs_uds_server_cb_t *cb
			       )
/*****************************************************************************/
{
  if(data == NULL)
    return;

  if(cb == NULL)
    return;

  data->srv_ctx = srv_ctx;
  memcpy(&data-> srv_cb,cb,sizeof(cs_uds_server_cb_t));
  
}

/*****************************************************************************/
void cs_uds_server_process(
			   cs_uds_service_t *data
			   )
/*****************************************************************************/
{
  cs_uds_data_t rcvpacket;
  cs_uds_data_t sendpacket;
  uint8_t buffer[CS_UDS_MAXBYTE];
  uint16_t bufferlen = 0;
  
  /*Reset send data*/
  memset(&sendpacket,0,sizeof(sendpacket));
  memset(&rcvpacket,0,sizeof(sendpacket));
  memset(buffer,0,sizeof(buffer));
  bufferlen = 0;
      
  if(!_uds_server_recv_and_prepare(data,&rcvpacket) && (rcvpacket.mt == DIAG)){      
    switch(rcvpacket.pci.si){
    case CS_UDS_SI_DSC: //Default session
      _server_handle_ds(data,&rcvpacket,&sendpacket,
			buffer,&bufferlen);
      break;
    case CS_UDS_SI_ER:
      _server_handle_er(data,&rcvpacket,&sendpacket,
			buffer,&bufferlen);
      break;
    case CS_UDS_SI_SA:
      _server_handle_sa(data,&rcvpacket,&sendpacket,
			buffer,&bufferlen);
      break;
    case CS_UDS_SI_CDTCS:
      _server_handle_cdts(data,&rcvpacket,&sendpacket,
			  buffer,&bufferlen);
      break;
    case CS_UDS_SI_LC:
      _server_handle_lc(data,&rcvpacket,&sendpacket,
			buffer,&bufferlen);
      break;
    case CS_UDS_SI_TP:
      _server_handle_tp(data,&rcvpacket,&sendpacket,
			buffer,&bufferlen);
      break;
    case CS_UDS_SI_RDB:
      _server_handle_rdb(data,&rcvpacket,&sendpacket,
			 buffer,&bufferlen);
      break;
    case CS_UDS_SI_WDB:
      _server_handle_wdb(data,&rcvpacket,&sendpacket,
			 buffer,&bufferlen);
      break;
    case CS_UDS_SI_RC:
      _server_handle_rc(data,&rcvpacket,&sendpacket,
			buffer,&bufferlen);
      break;
    case CS_UDS_SI_CDTCI:
      _server_handle_cdtci(data,&rcvpacket,&sendpacket,
			   buffer,&bufferlen);
      break;
    case CS_UDS_SI_RDTCI:
      _server_handle_rdtcipr(data,&rcvpacket,&sendpacket,
			     buffer,&bufferlen);
      break;
    case CS_UDS_SI_RD:
    case CS_UDS_SI_RU:
      _server_handle_uploaddownload(data,&rcvpacket,&sendpacket,
				    buffer,&bufferlen);
	break;
    case CS_UDS_SI_TD:
      _server_handle_transferData(data,&rcvpacket,&sendpacket,
				  buffer,&bufferlen);
      break;
    case CS_UDS_SI_RTE:
      _server_handle_transferExit(data,&rcvpacket,&sendpacket,
				  buffer,&bufferlen);
      break;
    default:
      _uds_server_prepare_neg(CS_UDS_ERROR_SNS,"Server: Service is not supported\n",&sendpacket,buffer,&bufferlen);
    }
       
    /*Send packet*/
    if(rcvpacket.resp == RESPREQ){
      sendpacket.sa = rcvpacket.sa;
      sendpacket.ta = rcvpacket.ta;
      sendpacket.pci.si = rcvpacket.pci.si;
      sendpacket.mt = RDIAG;
      sendpacket.tat = PHYS;
      sendpacket.ae = 0;
      sendpacket.resp = RESPREQ;
	
      if(sendpacket.result == CS_FAIL){
	sendpacket.pci.nrsi = CS_UDS_ERROR_SNSIAS;
	memcpy(sendpacket.a_data,buffer,bufferlen);
	sendpacket.lt = bufferlen;
      } else {
	sendpacket.pci.nrsi = CS_UDS_ERROR_PR;
	sendpacket.pci.si |= CS_UDS_SI_RESPOFFSET;
	memcpy(sendpacket.a_data,buffer,bufferlen);
	sendpacket.lt = bufferlen;
      }
      if(_uds_pack_and_send(&sendpacket,data)){
	_uds_debug("Pack and send failed\n");
      }
    } else {
      _uds_verbose("Response suppressed\n");
    }
  }
}
