/*  
    Copyright (C) 2016 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "config.h"
#include "cs_uds_decoder.h"
#include "cs_uds.h"
#include "cs_uds_private.h"
#include "cs_defines.h"

/* \def STRBUFFER
   Length of string buffer
*/
#define STRBUFFER (CS_UDS_MAXBYTE * 4 + 1)

/** \enum cs_uds_decoder_state
    State flag to split more than 1 UDS Channel
*/
typedef enum{
  STATE_WAIT = 0,       /**< Wait for request*/
  STATE_RESPONSE        /**< Wait for response*/
} cs_uds_decoder_state_t;

/** \struct cs_uds_decoder
    UDS decoder storage structure
*/
struct cs_uds_decoder{
  uint32_t sa;                          /**< Source address*/
  uint32_t ta;                          /**< Target address*/
  void (*fout)(void *priv_data,
	       double ts,
	       uint32_t id,
	       const char *fmt, ...);   /**< Print out function callback*/
  void (*ferror)(void *priv_data,
		 double ts,
		 uint32_t id,
		 const char *fmt, ...); /**< Print out error function callback, May be NULL*/
  void (*fdata)(void *priv_data,
		double ts,
		uint32_t id,
		uint8_t *data,
		int len,
		int error);             /**< Data function callback*/
  void *priv_data;                      /**< Private data storage pointer for callback functions*/
  cs_isotp_decoder_t *isotp;            /**< ISOTP decoding layer*/
  cs_uds_decoder_state_t state;         /**< Request / Response state*/
};


/********************************************************************/
void _uds_decoder_reset(
			cs_uds_decoder_t * data
			)
/********************************************************************/
{
  data->sa = 0;
  data->ta = 0;
  data->fout = NULL;
  data->ferror = NULL;
  data->fdata = NULL;
  data->priv_data = NULL;
  data->isotp = NULL;
  data->state = STATE_WAIT;
}


/********************************************************************/
int _uds_decoder_check_id(
			  cs_uds_decoder_t * data,
			  uint32_t id
			  )
/********************************************************************/
{
  /*  if((id == data->sa) && (data->state == STATE_WAIT)){
    return CS_OK;
  }
  
  if((id == data->ta) && (data->state == STATE_RESPONSE)){
    return CS_OK;
  }
  */
  if(id == data->sa){
    return CS_OK;
    }
  
  if(id == data->ta){
    return CS_OK;
  }
    
  return CS_FAIL;
}

/********************************************************************/
char *_uds_decoder_buffer2str(char *str,
			      uint8_t *buffer,
			      int len
			      )
/********************************************************************/
{
  const int llen = 255;
  char local[llen];
  int i;
  int lcounter= 0;
  //Note len is maximal CS_UDS_MAXBYTE
  
  memset(str,0,sizeof(char)*STRBUFFER);
  for(i = 0; i < len; i++){
    memset(local,0,sizeof(char)*llen);
    lcounter += sprintf(&str[lcounter]," %.2X",buffer[i]);
  }
  //Return string 
  return str;
}

/********************************************************************/
int _uds_decoder_check_data(
			    cs_uds_decoder_t * data,
			    uint32_t id,
			    uint8_t *bdata,
			    int len,
			    cs_uds_data_t *packet
			    )
/********************************************************************/
{
  
  if(len < 2){
    return CS_FAIL;
  }
  
  memset(packet->a_data,0,CS_UDS_MAXBYTE);

  //Check for functional id
  if(id == CS_UDS_FUNC_ID){
    packet->tat = FUNC;
  } else {
    packet->tat = PHYS;
  }
  
  //Fill data structure
  packet->pci.nrsi = bdata[0] == CS_UDS_ERROR_SNSIAS ? CS_UDS_ERROR_SNSIAS : CS_UDS_ERROR_PR;
  if(packet->pci.nrsi){
    //negative response
    //get service id
    packet->pci.si = bdata[1];
    //get len
    packet->lt = len -2;
    memcpy(packet->a_data,&bdata[2], packet->lt);
  } else {
    //positive reponse
    if(bdata[0] & CS_UDS_SI_RESPOFFSET){
      packet->mt = RDIAG;
    } else {
      packet->mt = DIAG;
    }
    //get service id
    packet->pci.si = bdata[0] & 0x3F ; //extract si, skipping first bit and response offset
    switch(packet->pci.si){
    case CS_UDS_SI_DSC:
    case CS_UDS_SI_ER:
    case CS_UDS_SI_SA:
    case CS_UDS_SI_TP:
    case CS_UDS_SI_CDTCS:
    case CS_UDS_SI_LC:
      //Subfunction and response bit are supported
      packet->resp = (bdata[1] & 0x80)? RESPSUP : RESPREQ;
      //Remove response bit from buffer
      bdata[1] &= 0x7F;
      break;
    default:
      //Always has to respond
      packet->resp = RESPREQ;
      break;
    }
    packet->lt = len -1;
    memcpy(packet->a_data,&bdata[1],packet->lt);
  }
  return CS_OK;
}

/********************************************************************/
char* _uds_decoder_mt2str(a_mt_t mt)
/********************************************************************/
{
  switch(mt){
  case DIAG:
    return "Response";
  case RDIAG:
    return "Request";
  }
  return "Unknown"; //We should not reach this
}

/********************************************************************/
char* _uds_decoder_resp2str(resp_t resp)
/********************************************************************/
{
  switch(resp){
  case RESPREQ:
    return "requested";
  case RESPSUP:
    return "suppressed";
  }
  return "Unknown"; //We should not reach this
}

/********************************************************************/
char* _uds_decoder_tat2str(a_tat_t tat)
/********************************************************************/
{
  switch(tat){
  case FUNC:
    return "Functional adressing";
  case PHYS:
    return "Physical adressing";
  }
  return "Unknown"; //We should not reach this
}

/********************************************************************/
char* _uds_decoder_si2str(uint8_t si)
/********************************************************************/
{
  switch(si){
  case CS_UDS_SI_DSC:
    return "Session control";
  case CS_UDS_SI_ER:
    return "ECU reset";
  case CS_UDS_SI_SA:
    return "Security access";
  case CS_UDS_SI_TP:
    return "Tester present";
  case CS_UDS_SI_CDTCS:
    return "Control DTC settings";
  case CS_UDS_SI_LC:
    return "Link control";
  case CS_UDS_SI_RDB:
    return "Read Byte by identifier";
  case CS_UDS_SI_WDB:
    return "Write Byte by identifier";
  case CS_UDS_SI_RC:
    return "Routine control";
  case CS_UDS_SI_RDTCI:
    return "Read DTC information";
  case CS_UDS_SI_CDTCI:
    return "Clear DTC information";
  case CS_UDS_SI_RD:
    return "Request download";
  case CS_UDS_SI_RU:
    return "Request upload";
  case CS_UDS_SI_TD:
    return "Transfer data";
  case CS_UDS_SI_RTE:
    return "Request Transfer exit";
  case CS_UDS_SI_RFT:
    return "Request File transfer";
  default:
    return "Unknown SI";
  }
  return "Unknown"; //We should not reach this
}

/*************************************************************************************************/
char * _uds_decoder_si_dsc2str(uint8_t dsc)
/*************************************************************************************************/
{
  switch(dsc){
  case CS_UDS_SI_DSC_DS:
    return "Default session";
  case CS_UDS_SI_DSC_PRGS:
    return "Programming session";
  case CS_UDS_SI_DSC_EXTDS:
    return "Extended Diagnostic session";
  case CS_UDS_SI_DSC_SSDS:
    return "Safety System Diagnostic session";
  case CS_UDS_SI_DSC_EOL:
    return "EOL Extended Diagnostic session";
  case CS_UDS_SI_DSC_SUP:
    return "Supplier Diagnostic session";
  default:
    return "Unknown session";
  }
  return "Unknown"; //We should not reach this
}

/*************************************************************************************************/
char * _uds_decoder_si_er2str(uint8_t er)
/*************************************************************************************************/
{
  switch(er){
  case CS_UDS_SI_ER_HR:
    return "Hard reset";
  case CS_UDS_SI_ER_KOFFONR:
    return "Key Off On reset";
  case CS_UDS_SI_ER_SR:
    return "Soft reset";
  case CS_UDS_SI_ER_ERPSD:
    return "Enable Rapid Power shutdown";
  case CS_UDS_SI_ER_DRPSD:
    return "Disable Rapid Power shutdown";
  default:
    return "Unknown request";
  }
  return "Unknown"; //We should not reach this
}

/*************************************************************************************************/
char * _uds_decoder_si_sa2str(uint8_t sa)
/*************************************************************************************************/
{
  switch(sa){
  case CS_UDS_SI_SA_RSD:
    return "Request seed";
  case CS_UDS_SI_SA_SK:
    return "Send key";
  default:
    return "Unknown request";
  }
  return "Unknown"; //We should not reach this
}

/*************************************************************************************************/
char * _uds_decoder_si_cdtc2str(uint8_t cdtc)
/*************************************************************************************************/
{
  switch(cdtc){
  case  CS_UDS_SI_CDTCS_ON:
    return "On";
  case CS_UDS_SI_CDTCS_OFF:
    return "Off";
  default:
    return "Unknown request";
  }
  return "Unknown"; //We should not reach this
}

/*************************************************************************************************/
char * _uds_decoder_si_lc2str(uint8_t lc)
/*************************************************************************************************/
{
  switch(lc){
  default:
    return "Unknown request";
  }
  return "Unknown"; //We should not reach this
}

/*************************************************************************************************/
char * _uds_decoder_si_rc2str(uint8_t rc)
/*************************************************************************************************/
{
  switch(rc){
  case CS_UDS_SI_RC_STR:
    return "Start request";
  case CS_UDS_SI_RC_STPR:
    return "Stop request";
  case CS_UDS_SI_RC_RRR:
    return "Request result";
  default:
    return "Unknown request";
  }
  return "Unknown"; //We should not reach this
}

/*************************************************************************************************/
char * _uds_decoder_si_rdtci2str(uint8_t rdtci)
/*************************************************************************************************/
{
  switch(rdtci){
  case CS_UDS_SI_RDTCI_RNODTCBSM:
    return "Report Number of DTC by Status Mask";
  case CS_UDS_SI_RDTCI_RDTCBSM:
    return " Report DTC by Status Mask";
  default:
    return "Unknown request";
  }
  return "Unknown"; //We should not reach this
}

/*************************************************************************************************/
char * _uds_decoder_si_cdtcs2str(uint8_t cdtcs)
/*************************************************************************************************/
{
  switch(cdtcs){
  case  CS_UDS_SI_CDTCS_ON:
    return "On";
  case  CS_UDS_SI_CDTCS_OFF:
    return " Off";
  default:
    return "Unknown request";
  }
  return "Unknown"; //We should not reach this
}

/********************************************************************/
void _uds_decoder_cb_fdata(void *priv_data,
			   double ts,
			   uint32_t id,
			   uint8_t *data,
			   int len,
			   int error
			   )
/********************************************************************/
{
  char lstr[STRBUFFER];
  cs_uds_data_t packet;
  cs_uds_decoder_t *pdata = (cs_uds_decoder_t*)priv_data;
  uint16_t tserver;
  uint16_t tservere;
  uint8_t powerdown;
  uint8_t subfunction;
  uint16_t did;
  uint16_t routine;
  uint32_t dtc;
  uint8_t compression;
  uint8_t encryption;
  uint32_t adress;
  uint32_t tlen;
  //uint8_t adress_len;
  //uint8_t len_len;
  uint8_t seqcounter;
  
  if(error != ISOTP_ERROR_NOERROR){
    //Error, ISOTP failure
    return;
  }
  
  if(_uds_decoder_check_data(pdata,id,data,len,&packet)){
    //Error, Invalid packet
    return;
  }

  if(pdata->sa == id){
    if(packet.resp == RESPREQ){
      pdata->state = STATE_RESPONSE;
    } else {
      //Response is suppressed, but may be there is a negative response
      //TODO
      pdata->state = STATE_WAIT;
    }
  } else {
    pdata->state = STATE_WAIT;
  }
  
  //printf("UDS, New data: %f\n",ts);
  memset(lstr,0,sizeof(char)*STRBUFFER);
  
  //Handle neg response
  if(packet.pci.nrsi == CS_UDS_ERROR_SNSIAS){
    if(pdata->ferror != NULL){
      (*pdata->ferror)(pdata->priv_data,ts,id,"UDS, Neg Response, %s Error: %s",
		     _uds_decoder_si2str(packet.pci.si),
		     cs_uds_strlerrorid(packet.a_data[0]));
    }
    return;
  }

  switch(packet.pci.si){
  case CS_UDS_SI_DSC:
    if(packet.mt == DIAG){
      //Request
      //Check length
      if(packet.lt < 1){
	//Error
	return;
      }
      if(pdata->fout != NULL){
	(*pdata->fout)(pdata->priv_data,ts,id,"UDS, Request (Response %s) %s: Session: %s",
		       _uds_decoder_resp2str(packet.resp),
		       _uds_decoder_si2str(packet.pci.si),
		       _uds_decoder_si_dsc2str(packet.a_data[0]));
      } 
    } else {
      //Response
      //Check length
      if(packet.lt < 5){
	//Error
	return;
      }
      tserver = (((uint16_t)packet.a_data[1]) << 8) |  (uint16_t)packet.a_data[2];
      tservere = (((uint16_t)packet.a_data[3]) << 8) |  (uint16_t)packet.a_data[4];
      if(pdata->fout != NULL){
	(*pdata->fout)(pdata->priv_data,ts,id,"UDS, Response %s: Session: %s, Timing: %i ms, Extended %i ms",
		       _uds_decoder_si2str(packet.pci.si),
		       _uds_decoder_si_dsc2str(packet.a_data[0]),
		       tserver,
		       tservere);
      } 
    }
    return;
  case CS_UDS_SI_ER:
    if(packet.mt == DIAG){
      //Request
      //Check length
      if(packet.lt < 1){
	//Error
	return;
      }
      if(pdata->fout != NULL){
	(*pdata->fout)(pdata->priv_data,ts,id,"UDS, Request (Response %s) %s: Type: %s",
		       _uds_decoder_resp2str(packet.resp),
		       _uds_decoder_si2str(packet.pci.si),
		       _uds_decoder_si_er2str(packet.a_data[0]));
      } 
    } else {
      //Check length
      if(packet.lt < 1){
	if(pdata->fout != NULL){
	  (*pdata->fout)(pdata->priv_data,ts,id,"UDS, Response %s",
			 _uds_decoder_si2str(packet.pci.si));
	}
      } else {
	powerdown = packet.a_data[0];
	if(pdata->fout != NULL){
	  (*pdata->fout)(pdata->priv_data,ts,id,"UDS, Response %s: Powerdown: %i ms",
			 _uds_decoder_si2str(packet.pci.si),
			 powerdown);
	}
      }
    }
    return;
  case CS_UDS_SI_SA:
    //Check length
    if(packet.lt < 1){
      //Error
      return;
    }
    subfunction = packet.a_data[0];
    if(packet.mt == DIAG){
      //Request
      switch((subfunction & 0x1)){
      case CS_UDS_SI_SA_RSD: // 1, 3, 5, 7,...
	if(pdata->fout != NULL){
	  (*pdata->fout)(pdata->priv_data,ts,id,"UDS, Request (Response %s) %s: Type: %s, Level: 0x%X",
			 _uds_decoder_resp2str(packet.resp),
			 _uds_decoder_si2str(packet.pci.si),
			 _uds_decoder_si_sa2str(CS_UDS_SI_SA_RSD),
			 subfunction);
	}
	break;
      case 0: //CS_UDS_SI_SA_SK // 2,4,6,8,...
	if(packet.lt < 2){
	  //Error
	  return;
	}
	if(pdata->fout != NULL){
	  (*pdata->fout)(pdata->priv_data,ts,id,"UDS, Request (Response %s) %s: Type: %s, Level: 0x%X, Key: %s",
			 _uds_decoder_resp2str(packet.resp),
			 _uds_decoder_si2str(packet.pci.si),
			 _uds_decoder_si_sa2str(CS_UDS_SI_SA_SK),
			 subfunction,
			 _uds_decoder_buffer2str(lstr,&packet.a_data[1],packet.lt-1));
	}
	break;
      }  
    } else {
      //Response
      switch((subfunction & 0x1)){
      case CS_UDS_SI_SA_RSD: // 1, 3, 5, 7,...
	if(packet.lt < 2){
	  //Error
	  return;
	}
	if(pdata->fout != NULL){
	  (*pdata->fout)(pdata->priv_data,ts,id,"UDS, Response %s: Type: %s, Level: 0x%X, Seed: %s",
			 _uds_decoder_si2str(packet.pci.si),
			 _uds_decoder_si_sa2str(CS_UDS_SI_SA_RSD),
			 subfunction,
			 _uds_decoder_buffer2str(lstr,&packet.a_data[1],packet.lt-1));
	}	
	break;
      case 0: //CS_UDS_SI_SA_SK // 2,4,6,8,...
	if(pdata->fout != NULL){
	  (*pdata->fout)(pdata->priv_data,ts,id,"UDS, Response %s: Type: %s, Level: 0x%X",
			 _uds_decoder_si2str(packet.pci.si),
			 _uds_decoder_si_sa2str(CS_UDS_SI_SA_SK),
			 subfunction
			 );
	}	
	break;
      }  
    }
    return;
  case CS_UDS_SI_TP:
    if(packet.mt == DIAG){
      //Request
      if(pdata->fout != NULL){
	(*pdata->fout)(pdata->priv_data,ts,id,"UDS, Request (Response %s)  %s",
		       _uds_decoder_resp2str(packet.resp),
		       _uds_decoder_si2str(packet.pci.si));
      }
    } else {
      //Response
      if(pdata->fout != NULL){
	(*pdata->fout)(pdata->priv_data,ts,id,"UDS, Response %s",
		       _uds_decoder_si2str(packet.pci.si)
		       );
      }
    }
    return;
  case CS_UDS_SI_CDTCS:
    if(packet.mt == DIAG){
      //Request
      if(packet.lt < 1){
	//Error
	return;
      }
      subfunction = packet.a_data[0];
      if(pdata->fout != NULL){
	(*pdata->fout)(pdata->priv_data,ts,id,"UDS, Request (Response %s)  %s, Subfunction: %s",
		       _uds_decoder_resp2str(packet.resp),
		       _uds_decoder_si2str(packet.pci.si),
		       _uds_decoder_si_cdtcs2str(subfunction));
      }
    } else {
      //Response
      if(pdata->fout != NULL){
	(*pdata->fout)(pdata->priv_data,ts,id,"UDS, Response %s",
		       _uds_decoder_si2str(packet.pci.si)
		       );
      }
    }
    break;
  case CS_UDS_SI_LC:
    if(packet.mt == DIAG){
      //Request
      if(pdata->fout != NULL){
	//TODO not complete
	(*pdata->fout)(pdata->priv_data,ts,id,"UDS, Request (Response %s)  %s",
		       _uds_decoder_resp2str(packet.resp),
		       _uds_decoder_si2str(packet.pci.si));
      }
    } else {
      //Response
      if(pdata->fout != NULL){
	(*pdata->fout)(pdata->priv_data,ts,id,"UDS, Response %s",
		       _uds_decoder_si2str(packet.pci.si)
		       );
      }
    }
    break;
  case CS_UDS_SI_RDB:
    if(packet.lt < 2){
      //Error
      return;
    }
    //Note only one DID is handled
    did = (((uint16_t)packet.a_data[0]) << 8) & 0xFF00;
    did |= ((uint16_t)packet.a_data[1]) & 0xFF;
    if(packet.mt == DIAG){
      //Request
      if(pdata->fout != NULL){
	(*pdata->fout)(pdata->priv_data,ts,id,"UDS, Request (Response %s)  %s, DID: 0x%.4X",
		       _uds_decoder_resp2str(packet.resp),
		       _uds_decoder_si2str(packet.pci.si),
		       did);
      }
    } else {
      //Response
      if(pdata->fout != NULL){
	(*pdata->fout)(pdata->priv_data,ts,id,"UDS, Response %s: DID: 0x%.4X, Data: %s",
		       _uds_decoder_si2str(packet.pci.si),
		       did,
		       _uds_decoder_buffer2str(lstr,&packet.a_data[2],packet.lt-2)
		       );
      }	
    }
    return;
  case CS_UDS_SI_WDB:
    if(packet.mt == DIAG){
      //Request
      if(packet.lt < 2){
	//Error
	return;
      }
      //Note only one DID is handled
      did = (((uint16_t)packet.a_data[0]) << 8) & 0xFF00;
      did |= ((uint16_t)packet.a_data[1]) & 0xFF;
      if(pdata->fout != NULL){
	(*pdata->fout)(pdata->priv_data,ts,id,"UDS, Request (Response %s)  %s, DID: 0x%.4X, Data: %s",
		       _uds_decoder_resp2str(packet.resp),
		       _uds_decoder_si2str(packet.pci.si),
		       did,
		       _uds_decoder_buffer2str(lstr,&packet.a_data[2],packet.lt-2)
		       );
      }
    } else {
      //Response
      if(pdata->fout != NULL){
	(*pdata->fout)(pdata->priv_data,ts,id,"UDS, Response %s",
		       _uds_decoder_si2str(packet.pci.si)
		       );
      }
    }
    break;
  case CS_UDS_SI_RC:
    if(packet.lt < 3){
      //Error
      return;
    }
    subfunction = packet.a_data[0];
    routine = ((uint16_t ) packet.a_data[1]) << 8;
    routine |= packet.a_data[2];
    
    if(packet.mt == DIAG){
      //Request
      if(pdata->fout != NULL){
	(*pdata->fout)(pdata->priv_data,ts,id,"UDS, Request (Response %s)  %s, Subfunction: %s, Routine: 0x%.4X, Data: %s",
		       _uds_decoder_resp2str(packet.resp),
		       _uds_decoder_si2str(packet.pci.si),
		       _uds_decoder_si_rc2str(subfunction),
		       routine,
		       _uds_decoder_buffer2str(lstr,&packet.a_data[3],packet.lt-3)
		       );
      }
    } else {
      //Response
      if(pdata->fout != NULL){
	(*pdata->fout)(pdata->priv_data,ts,id,"UDS, Response %s, Subfunction %s, Routine 0x%.4X, Data: %s",
		       _uds_decoder_si2str(packet.pci.si),
		       _uds_decoder_si_rc2str(subfunction),
		       routine,
		       _uds_decoder_buffer2str(lstr,&packet.a_data[3],packet.lt-3)
		       );
      }
    }
    break;
  case CS_UDS_SI_RDTCI:
    if(packet.mt == DIAG){
      //Request
      //TODO not complete
      if(pdata->fout != NULL){
	(*pdata->fout)(pdata->priv_data,ts,id,"UDS, Request (Response %s)  %s",
		       _uds_decoder_resp2str(packet.resp),
		       _uds_decoder_si2str(packet.pci.si));
      }
    } else {
      //Response
      if(pdata->fout != NULL){
	(*pdata->fout)(pdata->priv_data,ts,id,"UDS, Response %s",
		       _uds_decoder_si2str(packet.pci.si)
		       );
      }
    }
    break;
  case CS_UDS_SI_CDTCI:
    if(packet.mt == DIAG){
      //Request
      if(packet.lt < 3){
	//Error
	return;
      }
      dtc = (uint32_t)(packet.a_data[0]<<16);
      dtc |= (uint32_t)(packet.a_data[1]<<8);
      dtc |= (uint32_t)packet.a_data[2];
      if(pdata->fout != NULL){
	(*pdata->fout)(pdata->priv_data,ts,id,"UDS, Request (Response %s)  %s, DTC-Mask: 0x%.4X",
		       _uds_decoder_resp2str(packet.resp),
		       _uds_decoder_si2str(packet.pci.si),
		       dtc);
      }
    } else {
      //Response
      if(pdata->fout != NULL){
	(*pdata->fout)(pdata->priv_data,ts,id,"UDS, Response %s",
		       _uds_decoder_si2str(packet.pci.si)
		       );
      }
    }
    break;
  case CS_UDS_SI_RD:
  case CS_UDS_SI_RU:
    if(packet.mt == DIAG){
      //Request
      if(packet.lt < 10){
	//Error
	return;
      }
      compression = ((packet.a_data[0] &0xF0) >> 4);
      encryption = (packet.a_data[0] & 0xF);
      //adress_len = ((packet.a_data[1] &0xF0) >> 4);
      //len_len = (packet.a_data[1] & 0xF);
      adress = (uint32_t)packet.a_data[5];
      adress |= ((uint32_t)packet.a_data[4] << 8);
      adress |= ((uint32_t)packet.a_data[3] << 16);
      adress |= ((uint32_t)packet.a_data[2] << 24);
      tlen = (uint32_t)packet.a_data[9];
      tlen |= ((uint32_t)packet.a_data[8] << 8);
      tlen |= ((uint32_t)packet.a_data[7] << 16);
      tlen |= ((uint32_t)packet.a_data[6] << 24);
      if(pdata->fout != NULL){
	(*pdata->fout)(pdata->priv_data,ts,id,"UDS, Request (Response %s)  %s, Compression: 0x%X, Encryption: 0x%X, Adress: 0x%.4X, Length: %i Bytes",
		       _uds_decoder_resp2str(packet.resp),
		       _uds_decoder_si2str(packet.pci.si),
		       compression,
		       encryption,
		       adress,
		       tlen);
      }      
    } else {
      //Response
      if(pdata->fout != NULL){
	(*pdata->fout)(pdata->priv_data,ts,id,"UDS, Response %s",
		       _uds_decoder_si2str(packet.pci.si)
		       );
      }
    }
    break;
  case CS_UDS_SI_TD:
    if(packet.lt < 1){
      //Error
      return;
    }
    seqcounter = packet.a_data[0];
    if(packet.mt == DIAG){
      //Request
      if(packet.lt > 1){
	//send data
	if(pdata->fout != NULL){
	  (*pdata->fout)(pdata->priv_data,ts,id,"UDS, Request (Response %s)  %s, Counter: %i, Data: %s",
			 _uds_decoder_resp2str(packet.resp),
			 _uds_decoder_si2str(packet.pci.si),
			 seqcounter,
			 _uds_decoder_buffer2str(lstr,&packet.a_data[1],packet.lt-1)
		       );
      } else {
	//request data
	  if(pdata->fout != NULL){
	    (*pdata->fout)(pdata->priv_data,ts,id,"UDS, Request (Response %s)  %s, Counter: %i",
			   _uds_decoder_resp2str(packet.resp),
			   _uds_decoder_si2str(packet.pci.si),
			   seqcounter
			   );
	  }
	}
      }
    } else {
      //Response
      if(packet.lt > 1){
	//receive data
	if(pdata->fout != NULL){
	  (*pdata->fout)(pdata->priv_data,ts,id,"UDS, Response %s, Counter %i, Data: %s",
			 _uds_decoder_si2str(packet.pci.si),
			 seqcounter,
			 _uds_decoder_buffer2str(lstr,&packet.a_data[1],packet.lt-1)
			 );
	}
      } else {
	//ack send data
	if(pdata->fout != NULL){
	  (*pdata->fout)(pdata->priv_data,ts,id,"UDS, Response %s, Counter %i",
			 _uds_decoder_si2str(packet.pci.si),
			 seqcounter
			 );
	}
      }
    }
    break;
  case CS_UDS_SI_RTE:
    if(packet.mt == DIAG){
      //Request
      if(pdata->fout != NULL){
	(*pdata->fout)(pdata->priv_data,ts,id,"UDS, Request (Response %s)  %s, Data: %s",
		       _uds_decoder_resp2str(packet.resp),
		       _uds_decoder_si2str(packet.pci.si),
		       _uds_decoder_buffer2str(lstr,&packet.a_data[0],packet.lt)
		       );
      }
    } else {
      //Response
      if(pdata->fout != NULL){
	(*pdata->fout)(pdata->priv_data,ts,id,"UDS, Response %s, Data: %s",
		       _uds_decoder_si2str(packet.pci.si),
		       _uds_decoder_buffer2str(lstr,&packet.a_data[0],packet.lt)
		       );
      }
    }
    break;
  case CS_UDS_SI_RFT:
    //Todo not completed
    if(packet.mt == DIAG){
      //Request
      if(pdata->fout != NULL){
	(*pdata->fout)(pdata->priv_data,ts,id,"UDS, Request (Response %s)  %s",
		       _uds_decoder_resp2str(packet.resp),
		       _uds_decoder_si2str(packet.pci.si)
		       );
      }
    } else {
      //Response
      if(pdata->fout != NULL){
	(*pdata->fout)(pdata->priv_data,ts,id,"UDS, Response %s",
		       _uds_decoder_si2str(packet.pci.si)
		       );
	}
    }
    break;
  default:
    break;
  }
  
  return;
}

/********************************************************************/
cs_uds_decoder_t *cs_uds_decoder_init(
			uint32_t sa,
			uint32_t ta,
			void (*fout)(void *priv_data,
				     double ts,
				     uint32_t id,
				     const char *fmt, ...),
			void (*ferror)(void *priv_data,
				       double ts,
				       uint32_t id,
				       const char *fmt, ...),
			void (*fdata)(void *priv_data,
				      double ts,
				      uint32_t id,
				      uint8_t *data,
				      int len,
				      int error),
			void *priv_data,
			int isotpflags
			)
/********************************************************************/
{
  void *isotp_error = NULL;
  void *isotp_all = NULL;
  cs_uds_decoder_t *data = NULL;
    
  if(fdata == NULL){
    return NULL;
  }

  data = (cs_uds_decoder_t *)malloc(sizeof(cs_uds_decoder_t));
  if(data == NULL)
    return NULL;
  
  _uds_decoder_reset(data);

  //Store CAN-ids
  data->sa = sa;
  data->ta = ta;
  
  //Store handles
  data->fout = fout;
  data->ferror = ferror;
  data->fdata = fdata;

  //Store priv data
  data->priv_data = priv_data;

  isotp_error = NULL;
  isotp_all = NULL;
  //Init ISOTP decoder
  if((isotpflags & UDS_DECODER_ISOTP_ERROR) > 0)
    isotp_error = ferror;

  if((isotpflags & UDS_DECODER_ISOTP_ALL) > 0)
    isotp_all = fout;
  
  if((data->isotp = cs_isotp_decoder_init(
					  sa,
					  ta,
					  isotp_all,
					  isotp_error,
					  _uds_decoder_cb_fdata,
					  data, //Stack private data
					  data->priv_data //Output private data
					  )
      )==NULL){
    //Failure
    free(data);
    return NULL;
  }
  
  return data;
}

/********************************************************************/
int cs_uds_decoder_reset(
			 cs_uds_decoder_t *data
			 )
/********************************************************************/
{
  if( data == NULL ){
    return CS_FAIL;
  }
 
  return CS_OK;
}
 
/********************************************************************/
int cs_uds_decoder_data(
			cs_uds_decoder_t *data,
			double ts,
			uint32_t id,
			uint8_t *bdata,
			int len
			)
/********************************************************************/
{
  if(data == NULL){
    return CS_FAIL;
  }
   
  if(bdata == NULL){
    return CS_FAIL;
  }
   
  if(_uds_decoder_check_id(data,id)){
    //CAN ID does not fit
    return CS_OK;
  }
   
  //Call ISOTP stack
  if(cs_isotp_decoder_data(data->isotp,ts,id,bdata,len)){
    return CS_FAIL;
  }
   
  return CS_OK;
}
 
/********************************************************************/
int cs_uds_decoder_close(
			 cs_uds_decoder_t *data
			 )
/********************************************************************/
{
  if( data == NULL ){
    return CS_FAIL;
  }
  cs_isotp_decoder_close(data->isotp);
  free(data);
  return CS_OK;
}
