/*  
    Copyright (C) 2014 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <string.h>
#include <byteswap.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include "cs_byteorder.h"
#include "cs_vbf.h"
#include "cs_parser.h"
#include "cs_defines.h"
#include "cs_vbf_crc_ccitt.h"


/********************************************************************/
void _vbf_debug(const char *fmt, ...)
/********************************************************************/
{
#ifdef CS_DEBUG
  char buffer[CS_CHARBUF];
  va_list ap;
  cs_vbf_debug_level_t dl = cs_vbf_debuglevel(VBF_DEBUG_LEVEL_READ);

  if(dl >= VBF_DEBUG_LEVEL_DEBUG){
    printf("VBF Debug: ");
    memset(buffer,0,sizeof(buffer));
    va_start(ap, fmt);
    vsnprintf(buffer,CS_CHARBUF, fmt, ap);
    va_end(ap);
    printf(buffer);
  }
#endif
}

/********************************************************************/
void _vbf_verbose(const char *fmt, ...)
/********************************************************************/
{
#ifdef CS_DEBUG
  char buffer[CS_CHARBUF];
  va_list ap;
  cs_vbf_debug_level_t dl = cs_vbf_debuglevel(VBF_DEBUG_LEVEL_READ);

  if(dl >= VBF_DEBUG_LEVEL_VERBOSE){
    printf("VBF Verbose: ");
    memset(buffer,0,sizeof(buffer));
    va_start(ap, fmt);
    vsnprintf(buffer,CS_CHARBUF, fmt, ap);
    va_end(ap);
    printf(buffer);
  }
#endif
}

/********************************************************************/
int cs_vbf_getlasterror(
			cs_vbf_t *data
			)
/********************************************************************/
{
  if(data == NULL)
    return 0;

  return data->lerror;
}

/********************************************************************/
char* cs_vbf_getlasterrorstring(
				cs_vbf_t *data
				)
/********************************************************************/
{
  if(data == NULL)
    return "";
  
  return cs_vbf_geterroridstring(data->lerror);
}

/********************************************************************/
char* cs_vbf_geterroridstring(
			      uint8_t lerror
			      )
/********************************************************************/
{
  switch(lerror){
  case VBF_ERROR_NOERROR:
    return "No error";
  case VBF_ERROR_VERSION:
    return "Incompatible version";
  case VBF_ERROR_FORMAT:
    return "Invalid format";
  case VBF_ERROR_FILE:
    return "Invalid file";
  case VBF_ERROR_UNSUPPORTED:
    return "Unsupported feature";
  }
  

  return "";
}

/********************************************************************/
cs_vbf_debug_level_t cs_vbf_debuglevel(
				       cs_vbf_debug_level_t level
				       )
/********************************************************************/
{
  //Declare a static variable to get a global debug flag in application context
  static cs_vbf_debug_level_t vbf_debug_level = VBF_DEBUG_LEVEL_NONE;

  switch(level){
  case VBF_DEBUG_LEVEL_NONE:
  case VBF_DEBUG_LEVEL_DEBUG:
  case VBF_DEBUG_LEVEL_VERBOSE:
    vbf_debug_level = level;
    break;
  case VBF_DEBUG_LEVEL_READ:
    //Nothing todo
    return vbf_debug_level;
  }
  return vbf_debug_level;
}
/********************************************************************/


/** Function swap raw data to processor specific data, type uint16_t
    \retval swapped uint16_t
*/
/**************************************************************************/
uint16_t cs_vbf_swap16(
		     uint16_t value
		     )
/**************************************************************************/
{
  uint16_t rv;
  if(cs_bo_endianess() != BO_MOTOROLLA){ //!= Big Endian
    if(cs_bo_endianess() == BO_INTEL)
      _vbf_verbose(" VBF Swap 16: Bit Motorola to Intel\n");
    else
      _vbf_verbose("VBF Swap 16: Bit Intel to Motorola\n");
    //swap bytes
    rv = bswap_16(value);
    _vbf_verbose("VBF Swap 16: In: 0x%X, OUT 0x%X\n",value,rv);
    return rv;
  }
  return value;
}

/** Function swap raw data to processor specific data, type uint32_t
    \retval swapped uint32_t
*/
/**************************************************************************/
uint32_t cs_vbf_swap32(
		     uint32_t value
		     )
/**************************************************************************/
{
  uint32_t rv;
  if(cs_bo_endianess() != BO_MOTOROLLA){ //!= Big Endian
    if(cs_bo_endianess() == BO_INTEL)
      _vbf_verbose("VBF Swap 32: Bit Motorola to Intel\n");
    else
      _vbf_verbose("VBF Swap 32: Bit Intel to Motorola\n");
    //swap bytes
    rv = bswap_32(value);
    _vbf_verbose("VBF Swap 32: In: 0x%X, OUT 0x%X\n",value,rv);
    return rv;
  }
  return value;
}


/** Reset structure */
/**************************************************************************/
void _vbf_init(cs_vbf_t * vbf)
/**************************************************************************/
{
  memset(vbf,0,sizeof(cs_vbf_t));
}

/** Find value pairs in list
    { 0x123, 0x 234}
*/
/**************************************************************************/
int _vbf_getvaluepair(
		      cs_parser_result_t *result,
		      uint32_t *v1,
		      uint32_t *v2
		      )
/**************************************************************************/
{
  pcre *re = NULL;
  char regex[CS_REGEX_BUFFER];
  int rv = 0;
  int noresult = 5;
  int ovector[CS_REGEX_RESULTSIZE*noresult];
  char rbuffer[CS_PARSER_CHARBUF];

  *v1 = 0;
  *v2 = 0;
  
  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER);
  snprintf(regex,CS_REGEX_BUFFER,"{[\\s]?0x(?'value1'[0-9a-fA-F]+),[\\s?]0x(?'value2'[0-9a-fA-F]+)}");

  //_vbf_debug("VBF GetValuePair: Regex Buffer: %s\n",regex);

  //Precompile regex
  if((re = cs_parser_precompile((const char *)regex)) == NULL){
    _vbf_debug("VBF GetValuePair: Regex precompilation failed\n");
    return CS_FAIL;
  }

  //Get data, only one fit is allowed
  rv = pcre_exec(re, NULL, (const char*)result->searchin,result->searchlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE * noresult);

  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _vbf_verbose("VBF GetValuePair: No match\n"); 
      pcre_free(re);
      return CS_FAIL;
      break;
    default: 
      _vbf_debug("VBF GetValuePair: Matching error %d\n", rv);
      pcre_free(re);
      return CS_FAIL;
      break;
    }
  } else {
    //Set result to end of region, to set startpoint for next search
    result->result = &result->searchin[ovector[0]];
    result->resultlen = ovector[1] - ovector[0];
    result->resultoffset = ovector[0];
    cs_parser_calcresult(result);
    //cs_parser_printresult(result);

    if(cs_parser_getsubstring(re,result->searchin,ovector,CS_REGEX_RESULTSIZE*noresult,"value1",rbuffer)){
      _vbf_debug("VBF GetValuePair: Failed to get substring value1\n");
      pcre_free(re);
      return CS_FAIL;
    }
    if(!sscanf(rbuffer,"%X",v1)){
      _vbf_debug("VBF GetValuePair: Failed to parse substring %s\n",rbuffer);
      pcre_free(re);
      return CS_FAIL;
    }

    if(cs_parser_getsubstring(re,result->searchin,ovector,CS_REGEX_RESULTSIZE*noresult,"value2",rbuffer)){
      _vbf_debug("VBF GetValuePair: Failed to get substring value2\n");
      pcre_free(re);
      return CS_FAIL;
    }
    if(!sscanf(rbuffer,"%X",v2)){
      _vbf_debug("VBF GetValuePair: Failed to parse substring %s\n",rbuffer);
      pcre_free(re);
      return CS_FAIL;
    }
  }
  pcre_free(re); 

  return CS_OK;
}

/** Find section from file
    name = {
    {0x123,0xabd},
    {0x123,0xabd},
    ...
    };
    \retval 0 success
*/
/**************************************************************************/
int _vbf_getvaluesection(
			 cs_parser_result_t *result,
			 char *name
			 )
/**************************************************************************/
{
  pcre *re = NULL;
  char regex[CS_REGEX_BUFFER];
  int rv = 0;
  int noresult = 5;
  int ovector[CS_REGEX_RESULTSIZE*noresult];

  
  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER);
  snprintf(regex,CS_REGEX_BUFFER,"%s = {(.*?)};",name);

  //_vbf_debug("VBF GetValueSection: Regex Buffer: %s\n",regex);

  //Precompile regex
  if((re = cs_parser_precompile((const char *)regex)) == NULL){
    _vbf_debug("VBF GetValueSection: Regex precompilation failed\n");
    return CS_FAIL;
  }

  //Get data, only one fit is allowed
  rv = pcre_exec(re, NULL, (const char*)result->searchin,result->searchlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE * noresult);

  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _vbf_debug("VBF GetValueSection: No match\n"); 
      pcre_free(re);
      return CS_FAIL;
      break;
    default: 
      _vbf_debug("VBF GetValueSection: Matching error %d\n", rv);
      pcre_free(re);
      return CS_FAIL;
      break;
    }
  } else {
    result->result = &result->searchin[ovector[2]];
    result->resultlen = ovector[3] - ovector[2];
    result->resultoffset = ovector[3];
    cs_parser_calcresult(result);
    //cs_parser_printresult(result);
  }
  pcre_free(re); 

  return CS_OK;
}

/** Find section from file
    {
    ...
    ...
    }
    \retval 0 success
*/
/**************************************************************************/
int _vbf_getsection(
		    cs_parser_result_t *result,
		    char * name
		   )
/**************************************************************************/
{
  pcre *re = NULL;
  char regex[CS_REGEX_BUFFER];
  int rv = 0;
  int noresult = 5;
  int ovector[CS_REGEX_RESULTSIZE*noresult];

  
  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER);
  if(name == NULL){
    snprintf(regex,CS_REGEX_BUFFER,"\\{(?:(?>[^{}])|(?R))*\\}");
  } else {
    snprintf(regex,CS_REGEX_BUFFER,"%s\\{(?:(?>[^{}])|(?R))*\\}",name);
  }

  //_vbf_debug("VBF GetSection: Regex Buffer: %s\n",regex);

  //Precompile regex
  if((re = cs_parser_precompile((const char *)regex)) == NULL){
    _vbf_debug("VBF GetSection: Regex precompilation failed\n");
    return CS_FAIL;
  }

  //Get data, only one fit is allowed
  rv = pcre_exec(re, NULL, (const char*)result->searchin,result->searchlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE * noresult);

  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _vbf_debug("VBF GetSection: No match\n"); 
      pcre_free(re);
      return CS_FAIL;
      break;
    default: 
      _vbf_debug("VBF GetSection: Matching error %d\n", rv);
      pcre_free(re);
      return CS_FAIL;
      break;
    }
  } else {
    result->result = &result->searchin[ovector[0]];
    result->resultlen = ovector[1] - ovector[0];
    result->resultoffset = ovector[0];
    cs_parser_calcresult(result);
    //cs_parser_printresult(result);

  }
  pcre_free(re); 

  return CS_OK;
}

/** Get Value from file
    name = <string>;
    \retval 0 success
*/
/**************************************************************************/
int _vbf_getvalue(
		  cs_parser_result_t *result,
		  char *name,
		  char *rbuffer
		  )
/**************************************************************************/
{
  pcre *re = NULL;
  char regex[CS_REGEX_BUFFER];
  int rv = 0;
  int noresult = 5;
  int ovector[CS_REGEX_RESULTSIZE*noresult];

  
  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER);
  snprintf(regex,CS_REGEX_BUFFER,"%s = (?'value'.*?);",name);

  //_vbf_debug("VBF GetValue: Regex Buffer: %s\n",regex);

  //Precompile regex
  if((re = cs_parser_precompile((const char *)regex)) == NULL){
    _vbf_debug("VBF GetValue: Regex precompilation failed\n");
    return CS_FAIL;
  }

  //Get data, only one fit is allowed
  rv = pcre_exec(re, NULL, (const char*)result->searchin,result->searchlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE * noresult);

  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _vbf_debug("VBF GetValue: No match\n"); 
      pcre_free(re);
      return CS_FAIL;
      break;
    default: 
      _vbf_debug("VBF GetValue: Matching error %d\n", rv);
      pcre_free(re);
      return CS_FAIL;
      break;
    }
  } else {

    if(cs_parser_getsubstring(re,result->searchin,ovector,CS_REGEX_RESULTSIZE*noresult,"value",rbuffer)){
      _vbf_debug("VBF GetValue: Failed to get substring\n");
      pcre_free(re);
      return CS_FAIL;
    }
    result->result = &result->searchin[ovector[0]];
    result->resultlen = ovector[1] - ovector[0];
    result->resultoffset = ovector[0];
    cs_parser_calcresult(result);
    //cs_parser_printresult(result);
  }
  pcre_free(re); 

  return CS_OK;
}

/** Get string Value from file
    name = "<string>";
    \retval 0 success
*/
/**************************************************************************/
int _vbf_getstringvalue(
			cs_parser_result_t result,
			char *name,
			char **rbuffer
			)
/**************************************************************************/
{
  pcre *re = NULL;
  char regex[CS_REGEX_BUFFER];
  int rv = 0;
  int noresult = 5;
  int ovector[CS_REGEX_RESULTSIZE*noresult];

  
  //Build up regex
  memset(regex,0,sizeof(char)*CS_REGEX_BUFFER);
  snprintf(regex,CS_REGEX_BUFFER,"%s = \"(?'value'.*?)\";",name);

  //_vbf_debug("VBF GetStringValue: Regex Buffer: %s\n",regex);

  //Precompile regex
  if((re = cs_parser_precompile((const char *)regex)) == NULL){
    _vbf_debug("VBF GetStringValue: Regex precompilation failed\n");
    return CS_FAIL;
  }

  //Get data, only one fit is allowed
  rv = pcre_exec(re, NULL, (const char*)result.searchin,result.searchlen, 0,PCRE_NOTEMPTY, ovector, CS_REGEX_RESULTSIZE * noresult);

  if(rv < 0) {
    switch(rv) {
    case PCRE_ERROR_NOMATCH:
      _vbf_debug("VBF GetStringValue: No match\n"); 
      pcre_free(re);
      return CS_FAIL;
      break;
    default: 
      _vbf_debug("VBF GetStringValue: Matching error %d\n", rv);
      pcre_free(re);
      return CS_FAIL;
      break;
    }
  } else {

    if(cs_parser_getsubstring_malloc(re,result.searchin,ovector,CS_REGEX_RESULTSIZE*noresult,"value",rbuffer)){
      _vbf_debug("VBF GetStringValue: Failed to get substring\n");
      pcre_free(re);
      return CS_FAIL;
    }
  }
  pcre_free(re); 

  return CS_OK;
}

/** Get Value, interpretation for float
    \retval 0 success
*/
/**************************************************************************/
int _vbf_getfloatvalue(
			cs_parser_result_t result,
			char *name,
			float *value
			)
/**************************************************************************/
{
  char buffer[CS_PARSER_CHARBUF];
  if(_vbf_getvalue(&result,name,buffer)){
    return CS_FAIL;
  }

  //_vbf_debug("VBF GetFloatValue: Buffer: %s\n",buffer);

  if(!sscanf(buffer,"%f",value)){
    _vbf_debug("VBF GetFloatValue: Failed to parse substring %s\n",buffer);
    return CS_FAIL;
  }
  return CS_OK;
}

/** Get Value, interpretation for hexadecimal uint32 value
    \retval 0 success
*/
/**************************************************************************/
int _vbf_gethexvalue(
			cs_parser_result_t result,
			char *name,
			uint32_t *value
			)
/**************************************************************************/
{
  char buffer[CS_PARSER_CHARBUF];
  if(_vbf_getvalue(&result,name,buffer)){
    return CS_FAIL;
  }
  
  if(!sscanf(buffer,"%X",value)){
    _vbf_debug("VBF GetHexValue: Failed to parse substring %s\n",buffer);
    return CS_FAIL;
  }
  return CS_OK;
}


/** Get software part type
    \retval 0 success
*/
/**************************************************************************/
int _vbf_softwareparttype(
			  cs_parser_result_t result,
			  cs_vbf_part_t *type
			  )
/**************************************************************************/
{
  char buffer[CS_PARSER_CHARBUF];
  if(_vbf_getvalue(&result,"sw_part_type",buffer)){
    return CS_FAIL;
  }

  *type = VBF_UNKNOWN;

  if(strncmp(buffer,"CARCFG",CS_PARSER_CHARBUF) == 0)
    *type = VBF_CARCFG;

  if(strncmp(buffer,"CUSTOM",CS_PARSER_CHARBUF) == 0)
    *type = VBF_CUSTOM;

  if(strncmp(buffer,"DATA",CS_PARSER_CHARBUF) == 0)
    *type = VBF_DATA;

  if(strncmp(buffer,"EXE",CS_PARSER_CHARBUF) == 0)
    *type = VBF_EXE;

  if(strncmp(buffer,"SBL",CS_PARSER_CHARBUF) == 0)
    *type = VBF_SBL;

  if(strncmp(buffer,"SIGCFG",CS_PARSER_CHARBUF) == 0)
    *type = VBF_SIGCFG;

  if(strncmp(buffer,"TEST",CS_PARSER_CHARBUF) == 0)
    *type = VBF_TEST;

  return CS_OK;
}

/** Get network type
    \retval 0 success
*/
/**************************************************************************/
int _vbf_networktype(
			  cs_parser_result_t result,
			  cs_vbf_network_t *type
			  )
/**************************************************************************/
{
  char buffer[CS_PARSER_CHARBUF];
  if(_vbf_getvalue(&result,"network",buffer)){
    return CS_FAIL;
  }

  *type = VBF_CAN_UNKNOWN;

  if(strncmp(buffer,"CAN_HS",CS_PARSER_CHARBUF) == 0)
    *type = VBF_CAN_HS;

  if(strncmp(buffer,"CAN_MS",CS_PARSER_CHARBUF) == 0)
    *type = VBF_CAN_MS;

  return CS_OK;
}

/** Get frame format type
    \retval 0 success
*/
/**************************************************************************/
int _vbf_frametype(
		   cs_parser_result_t result,
		   cs_vbf_frame_t *type
			  )
/**************************************************************************/
{
  char buffer[CS_PARSER_CHARBUF];
  if(_vbf_getvalue(&result,"frame_format",buffer)){
    return CS_FAIL;
  }

  *type = VBF_FRAME_UNKNOWN;

  if(strncmp(buffer,"CAN_STANDARD",CS_PARSER_CHARBUF) == 0)
    *type = VBF_FRAME_STANDARD;

  return CS_OK;
}

/**************************************************************************/
int cs_vbf_open(
		cs_vbf_t *vbf,
		char *filename
		)
/**************************************************************************/
{
  cs_parser_result_t result;
  cs_parser_result_t header;
  cs_parser_result_t erase;
  uint32_t format_id = 0;
  uint8_t * work = NULL;
  uint32_t u32;
  uint16_t u16;
  uint16_t lchecksum;

  if((vbf == NULL) ||
     (filename == NULL))
    return CS_FAIL;

  _vbf_init(vbf);

  vbf->filename = filename;
  if(cs_parser_open(vbf->filename,&vbf->fs,&vbf->filelen,&vbf->map)){
    _vbf_debug("VBF: Faild to open file %s\n",vbf->filename);
    vbf->lerror = VBF_ERROR_FILE;
    return CS_FAIL;
  }

  //define search region for global flags
  result.searchin = vbf->map;
  result.searchlen = vbf->filelen;
  
  //Get version information
  if(_vbf_getfloatvalue(result,"vbf_version",&vbf->vbf_version)){
    _vbf_debug("VBF Open: Failed to get version information\n");
    vbf->lerror = VBF_ERROR_FORMAT;
    return CS_FAIL;
  }

  _vbf_debug("VBF Open: Get file version: %3.2f\n",vbf->vbf_version);

  if(vbf->vbf_version > (float)VBF_MAX_VERSION){
    _vbf_debug("VBF Open: Sorry Max version %3.2f is supported, File is version %3.2f\n", 
	     (float)VBF_MAX_VERSION, 
	     vbf->vbf_version);
    vbf->lerror = VBF_ERROR_VERSION;
    return CS_FAIL;
  }

  //Get header segment information
  if(_vbf_getsection(&result,NULL)){
    _vbf_debug("VBF Open: Can not find header segment\n");
    vbf->lerror = VBF_ERROR_FORMAT;
    return CS_FAIL;
  }

  //Store binary section start
  vbf->binarystart = (uint8_t*)result.result + result.resultlen;
  vbf->binaryend = &((uint8_t*)vbf->map)[vbf->filelen - 1];
  vbf->binarylength = vbf->binaryend - vbf->binarystart +1;

  _vbf_debug("VBF Open: Binary section start 0x%X end 0x%X Length %i Bytes\n",
	     vbf->binarystart,vbf->binaryend,vbf->binarylength);

  //Define new search reagion
  header.searchin = result.result;
  header.searchlen = result.resultlen;
  
  //Get software part number
  if(_vbf_getstringvalue(header,"sw_part_number",&vbf->vbf_swpartnumber)){
    _vbf_debug("VBF Open: Failed to get software part information\n");
    vbf->lerror = VBF_ERROR_FORMAT;
    return CS_FAIL;
  }
  _vbf_debug("VBF Open: Get software part number: %s\n",vbf->vbf_swpartnumber);

  //Get software part type
  if(_vbf_softwareparttype(header,&vbf->vbf_parttype)){
    _vbf_debug("VBF Open: Failed to get software part type\n");
    vbf->lerror = VBF_ERROR_FORMAT;
    return CS_FAIL;
  }

  switch(vbf->vbf_parttype){
  case VBF_UNKNOWN:
  case VBF_TEST:
  case VBF_CARCFG:
  case VBF_SIGCFG:
    _vbf_debug("VBF Open: Sorry Software part type is actually not supported\n");
    vbf->lerror = VBF_ERROR_UNSUPPORTED;
    return CS_FAIL;
    break;
  case VBF_SBL:
    _vbf_debug("VBF Open: Software part type: Secondary Bootloader\n");
    break;
  case VBF_EXE:
  case VBF_DATA:
    _vbf_debug("VBF Open: Software part type: Execution\n");
    break;
  default:
    _vbf_debug("VBF Open: Sorry Software part type is actually not supported\n");
    vbf->lerror = VBF_ERROR_UNSUPPORTED;
    return CS_FAIL;
    break;
  }
  
  //Get network type
  if(_vbf_networktype(header,&vbf->vbf_networtype)){
    _vbf_debug("VBF Open: Failed to get network type\n");
    vbf->lerror = VBF_ERROR_FORMAT;
    return CS_FAIL;
  }

  switch(vbf->vbf_networtype){
  case VBF_CAN_MS:
    _vbf_debug("VBF Open: Network type: Medium speed CAN\n");
    break;
  case VBF_CAN_HS:
    _vbf_debug("VBF Open: Network type: High speed CAN\n");
    break;
  default:
    _vbf_debug("VBF Open: Sorry network type is actually not supported\n");
    vbf->lerror = VBF_ERROR_UNSUPPORTED;
    return CS_FAIL;
    break;
  }

  //Get ECU CAN address
  if(_vbf_gethexvalue(header,"ecu_address",&vbf->vbf_ecuaddr)){
    _vbf_debug("VBF Open: Failed to get ECU CAN Id\n");
    vbf->lerror = VBF_ERROR_FORMAT;
    return CS_FAIL;
  }
  _vbf_debug("VBF Open: Get ECU CAN id: 0x%X\n",vbf->vbf_ecuaddr);

  //Get frame type
  if(_vbf_frametype(header,&vbf->vbf_frame)){
    _vbf_debug("VBF Open: Failed to get frame type\n");
    vbf->lerror = VBF_ERROR_FORMAT;
    return CS_FAIL;
  }

  switch(vbf->vbf_frame){
  case VBF_FRAME_STANDARD:
    _vbf_debug("VBF Open: Frame type: Standard CAN frame\n");
    break;
  default:
    _vbf_debug("VBF Open: Sorry frame type is actually not supported\n");
    vbf->lerror = VBF_ERROR_UNSUPPORTED;
    return CS_FAIL;
    break;
  }

  if(vbf->vbf_version >= VBF_MAX_VERSION){
    //Get format id
    if(_vbf_gethexvalue(header,"data_format_identifier",&format_id)){
      _vbf_debug("VBF Open: Failed to get Format Id\n");
      vbf->lerror = VBF_ERROR_FORMAT;
      return CS_FAIL;
    }
    vbf->vbf_format_id = (uint8_t)format_id;
    _vbf_debug("VBF Open: Format id: 0x%X\n",vbf->vbf_format_id);
  } 
  //Get file checksum
  if(_vbf_gethexvalue(header,"file_checksum",&vbf->vbf_filechecksum)){
    _vbf_debug("VBF Open: Failed to get File checksum\n");
    vbf->lerror = VBF_ERROR_FORMAT;
    return CS_FAIL;
  }
  _vbf_debug("VBF Open: File checksum: 0x%X\n",vbf->vbf_filechecksum);

  //Get specific content
  switch(vbf->vbf_parttype){
  case VBF_SBL:
    //No erase blocks
    vbf->vbf_no_eraseblocks = 0;
    //Only one block
    vbf->vbf_no_blocks = 1;
    //get call address
    if(_vbf_gethexvalue(header,"call",&vbf->vbf_calladdress)){
      _vbf_debug("VBF Open: Failed to get Call address\n");
      vbf->lerror = VBF_ERROR_FORMAT;
      return CS_FAIL;
    }
    _vbf_debug("VBF Open: Call address: 0x%X\n",vbf->vbf_calladdress);    
    break;
  case VBF_EXE:
  case VBF_DATA:
    erase = header;
    if(_vbf_getvaluesection(&erase,"erase")){
      _vbf_debug("VBF Open: Can not parse Erase segment\n");
      vbf->lerror = VBF_ERROR_FORMAT;
      return CS_FAIL;
    }
    //Update search region
    erase.searchin = erase.result;
    erase.searchlen = erase.resultlen;
    vbf->vbf_no_eraseblocks = 0;
    while(_vbf_getvaluepair(&erase,
			    &vbf->vbf_eraseblocks[vbf->vbf_no_eraseblocks].address,
			    &vbf->vbf_eraseblocks[vbf->vbf_no_eraseblocks].length) == 0){
      //Set search behind actual result
      erase.searchin = erase.next;
      erase.searchlen = erase.nextlen;
      _vbf_debug("VBF Open: Erase Block %i Address: 0x%X Length 0x%X\n",
	       vbf->vbf_no_eraseblocks,
	       vbf->vbf_eraseblocks[vbf->vbf_no_eraseblocks].address,
	       vbf->vbf_eraseblocks[vbf->vbf_no_eraseblocks].length
	       );
      if(vbf->vbf_no_eraseblocks < (VBF_MAX_BLOCKS -1)){
	vbf->vbf_no_eraseblocks++;
      } else {
	_vbf_debug("VBF Open: Erase block overrange detected\n");
	vbf->lerror = VBF_ERROR_FORMAT;
	return CS_FAIL;
	break;
      }
    }    
    _vbf_debug("VBF Open: Number of erase blocks %i\n",vbf->vbf_no_eraseblocks);
    break;
  default:
    //Nothing todo
    break;
  }

  //Handle binary part
  if(vbf->binarylength < 7){
    _vbf_debug("VBF Open: Invalid binary section length %i Bytes\n",vbf->binarylength);
    vbf->lerror = VBF_ERROR_FORMAT;
    return CS_FAIL;
  }

  work = vbf->binarystart;
  vbf->vbf_no_blocks = 0;
  while(work < vbf->binaryend){
    //TODO Byteorder if running on no Intel plattform
    _vbf_debug("VBF Open: Binary section %i:\n",vbf->vbf_no_blocks);
    if((work + VBF_ADDRESS_SIZE) > vbf->binaryend){
      _vbf_debug("VBF Open: Binary memory failure: address\n");
      vbf->lerror = VBF_ERROR_FORMAT;
      return CS_FAIL;
    }
    memcpy((void*)&u32,work,VBF_ADDRESS_SIZE);
    vbf->vbf_blocks[vbf->vbf_no_blocks].address = cs_vbf_swap32(u32);
    _vbf_debug("VBF Open:         Address : 0x%X\n",vbf->vbf_blocks[vbf->vbf_no_blocks].address);
    work+=VBF_ADDRESS_SIZE;
    if((work + VBF_LENGTH_SIZE) > vbf->binaryend){
      _vbf_debug("VBF Open: Binary memory failure, length\n");
      vbf->lerror = VBF_ERROR_FORMAT;
      return CS_FAIL;
    }
    memcpy((void*)&u32,work,VBF_LENGTH_SIZE);
    vbf->vbf_blocks[vbf->vbf_no_blocks].length = cs_vbf_swap32(u32);
    _vbf_debug("VBF Open:          Length : 0x%X\n",vbf->vbf_blocks[vbf->vbf_no_blocks].length);
    work+=VBF_LENGTH_SIZE;
    vbf->vbf_blocks[vbf->vbf_no_blocks].data = work;
    _vbf_debug("VBF Open:      Data start : 0x%X\n",vbf->vbf_blocks[vbf->vbf_no_blocks].data);
    _vbf_debug("VBF Open:      Data  end  : 0x%X\n",vbf->vbf_blocks[vbf->vbf_no_blocks].data + vbf->vbf_blocks[vbf->vbf_no_blocks].length -1);
    work+=vbf->vbf_blocks[vbf->vbf_no_blocks].length;
    if((work + (VBF_CHECKSUM_SIZE -1)) > vbf->binaryend){
      _vbf_debug("VBF Open: Binary memory failure, checksum\n");
      vbf->lerror = VBF_ERROR_FORMAT;
      return CS_FAIL;
    }
    memcpy((void*)&u16,work,VBF_CHECKSUM_SIZE);
    vbf->vbf_blocks[vbf->vbf_no_blocks].checksum = cs_vbf_swap16(u16);
    _vbf_debug("VBF Open:        Checksum : 0x%X\n",vbf->vbf_blocks[vbf->vbf_no_blocks].checksum);
    work+=VBF_CHECKSUM_SIZE;

    //Calc Checksum
    lchecksum = cs_vbf_crc_ccitt_init();
    lchecksum = cs_vbf_crc_ccitt_update(lchecksum,
					vbf->vbf_blocks[vbf->vbf_no_blocks].data,
					vbf->vbf_blocks[vbf->vbf_no_blocks].length);
    lchecksum = cs_vbf_crc_ccitt_finalize(lchecksum);
    _vbf_debug("VBF Open: Recalced Checksum: 0x%X\n",lchecksum);
    /*if(lchecksum != vbf->vbf_blocks[vbf->vbf_no_blocks].checksum){
      _vbf_debug("VBF Open: Check sum falure\n");
      return CS_FAIL;
      }*/

    if(vbf->vbf_no_blocks < (VBF_MAX_BLOCKS -1)){
      vbf->vbf_no_blocks++;
    } else {
      _vbf_debug("VBF Open: Block overrange detected\n");
      vbf->lerror = VBF_ERROR_FORMAT;
      return CS_FAIL;
      break;
    }
    
  }
  _vbf_debug("VBF Open: Number of data blocks: %i\n",vbf->vbf_no_blocks);

  return CS_OK;
}


/**************************************************************************/
int cs_vbf_close(
		 cs_vbf_t *vbf
		 )
/**************************************************************************/
{
  if(vbf == NULL)
    return CS_FAIL;

  if(cs_parser_close(vbf->fs,vbf->map,vbf->filelen))
    return CS_FAIL;

  if(vbf->vbf_description != NULL){
    free(vbf->vbf_description);
  }

  if(vbf->vbf_swpartnumber != NULL){
    free(vbf->vbf_swpartnumber);
  }
  
  _vbf_init(vbf);

  return CS_OK;
}


/**************************************************************************/
void cs_vbf_print(
		  cs_vbf_t *vbf
		  )
/**************************************************************************/
{
  int i;
  if(vbf == NULL)
    return;

  printf("VBF File information\n");
  printf("              Filename: %s\n",vbf->filename);
  printf("              Filesize: %li Bytes\n",vbf->filelen);
  printf("           Description: %s\n",vbf->vbf_description);
  printf("           VBF Version: %f\n",vbf->vbf_version);
  printf("         VBF Part type: ");
  switch(vbf->vbf_parttype){
  case VBF_UNKNOWN:
    printf("Undefined\n");
    break;
  case VBF_CUSTOM:
    printf("CUSTOM\n");
    break;
  case VBF_TEST:
    printf("TEST\n");
    break;
  case VBF_CARCFG:
    printf("CARCFG\n");
    break;
  case VBF_SIGCFG:
    printf("SIGCFG\n");
    break;
  case VBF_SBL:
    printf("SBL\n");
    printf("           Call adress: 0x%X\n",vbf->vbf_calladdress);
    break;
  case VBF_EXE:
    printf("EXE\n");
    break;
  case VBF_DATA:
    printf("DATA\n");
    break;
  }
  printf("      VBF network type: ");
  switch(vbf->vbf_networtype){
  case VBF_CAN_MS:
    printf("Medium speed CAN\n");
    break;
  case VBF_CAN_HS:
    printf("High speed CAN\n");
    break;
  case VBF_CAN_UNKNOWN:
    printf("Not defined\n");
    break;
  }
  printf("        VBF frame type: ");
  switch(vbf->vbf_frame){
  case VBF_FRAME_STANDARD:
    printf("Standard CAN frame\n");
    break;
  case VBF_FRAME_UNKNOWN:
    printf("Not defined\n");
    break;
  }
  printf("            ECU Adress: 0x%X\n",vbf->vbf_ecuaddr);
  if(vbf->vbf_version >= VBF_MAX_VERSION){
    printf("             Format id: 0x%X\n",vbf->vbf_format_id);
  }
  printf("         File checksum: 0x%X\n",vbf->vbf_filechecksum);
  printf("Number of erase blocks: %i\n",vbf->vbf_no_eraseblocks);
  for(i = 0; i < vbf->vbf_no_eraseblocks; i++){
    printf("           Erase block %i adress: 0x%X\n",i,vbf->vbf_eraseblocks[i].address);
    printf("           Erase block %i length: %i Bytes\n",i,vbf->vbf_eraseblocks[i].length);
  }
  printf("      Number of blocks: %i\n",vbf->vbf_no_blocks);
  for(i = 0; i < vbf->vbf_no_blocks; i++){
    printf("           Block %i adress: 0x%X\n",i,vbf->vbf_blocks[i].address);
    printf("           Block %i length: %i Bytes\n",i,vbf->vbf_blocks[i].length);
    printf("           Block %i checksum: 0x%X\n",i,vbf->vbf_blocks[i].checksum);
  }

}
