/*  XCP CAN Stack for SocketCAN
    Copyright (C) 2014 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <byteswap.h>
#include <unistd.h>

#include "config.h"
#include "cs_defines.h"
#include "cs_time.h"
#include "cs_calibration_utils.h"
#include "cs_xcp.h"
#include "cs_xcp_private.h"



/********************************************************************/
cs_debug_level_t cs_xcp_debuglevel(
				   cs_debug_level_t level
				   )
/********************************************************************/
{
  return cs_debuglevel(CS_DEBUG_XCP,level);
}

/********************************************************************/
void cs_xcp_debug_setcallback(
			      void *ctx,
			      void (*callb)(void *ctx,
					    cs_debug_sub_t sub,
					    cs_debug_level_t dl,
					    const char *func,
					    const int line,
					    const char *fmt,
					    va_list ap)
			      )
/********************************************************************/
{
  cs_debug_setcallback(CS_DEBUG_XCP,ctx,callb);
}


/** Init error handling table
 */
/**************************************************************************/
void _xcp_init_error_handling(
			      cs_xcp_t *data
			      )
/**************************************************************************/
{
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return;
  }
  
  memset(data->error_handling,0,sizeof(cs_xcp_error_handling_t)*CS_XCP_ERROR_HANDLINGLENGTH);

  //Connect normal
  cs_xcp_error_handling_t e0 = {.cmd = XCP_CMD_CONNECT, .subcmd = 0, .error= XCP_ERROR_TIMEOUTT1, .preaction = XCP_PA_NONE, .action = XCP_AC_REPEAT};
  data->error_handling[0] = e0;
  //Connect user defined
  cs_xcp_error_handling_t e1 = {.cmd = XCP_CMD_CONNECT, .subcmd = 1, .error= XCP_ERROR_TIMEOUTT6, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[1] = e1;

  //Disconnect
  cs_xcp_error_handling_t e2 = {.cmd = XCP_CMD_DISCONNECT, .subcmd = 0, .error= XCP_ERROR_TIMEOUTT1, .preaction = XCP_PA_SYNCH, .action = XCP_AC_REPEAT};
  data->error_handling[2] = e2;
  cs_xcp_error_handling_t e3 = {.cmd = XCP_CMD_DISCONNECT, .subcmd = 0, .error= XCP_ERROR_CMD_BUSY, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[3] = e3;
  cs_xcp_error_handling_t e4 = {.cmd = XCP_CMD_DISCONNECT, .subcmd = 0, .error= XCP_ERROR_PGM_ACTIVE, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[4] = e4;

  //Synch
  cs_xcp_error_handling_t e5 = {.cmd = XCP_CMD_SYNCH, .subcmd = 0, .error= XCP_ERROR_TIMEOUTT1, .preaction = XCP_PA_SYNCH, .action = XCP_AC_REPEAT2};
  data->error_handling[5] = e5;
  cs_xcp_error_handling_t e6 = {.cmd = XCP_CMD_SYNCH, .subcmd = 0, .error= XCP_ERROR_CMD_UNKNOWN, .preaction = XCP_PA_NONE, .action = XCP_AC_NONE};
  data->error_handling[6] = e6;
  cs_xcp_error_handling_t e7 = {.cmd = XCP_CMD_SYNCH, .subcmd = 0, .error= XCP_ERROR_CMD_SYNCH, .preaction = XCP_PA_NONE, .action = XCP_AC_OK};
  data->error_handling[7] = e7;

  //Get status
  cs_xcp_error_handling_t e8 = {.cmd = XCP_CMD_GET_STATUS, .subcmd = 0, .error= XCP_ERROR_TIMEOUTT1, .preaction = XCP_PA_SYNCH, .action = XCP_AC_REPEAT2};
  data->error_handling[8] = e8;

  //Get comm mode info
  cs_xcp_error_handling_t e9 = {.cmd = XCP_CMD_GET_COMM_MODE_INFO, .subcmd = 0, .error= XCP_ERROR_TIMEOUTT1, .preaction = XCP_PA_SYNCH, .action = XCP_AC_REPEAT2};
  data->error_handling[9] = e9;
  cs_xcp_error_handling_t e10 = {.cmd = XCP_CMD_GET_COMM_MODE_INFO, .subcmd = 0, .error= XCP_ERROR_CMD_BUSY, .preaction = XCP_PA_NONE, .action = XCP_AC_REPEAT};
  data->error_handling[10] = e10;

  //Get ID
  cs_xcp_error_handling_t e11 = {.cmd = XCP_CMD_GET_ID, .subcmd = 0, .error= XCP_ERROR_TIMEOUTT1, .preaction = XCP_PA_SYNCH, .action = XCP_AC_REPEAT2};
  data->error_handling[11] = e11;
  cs_xcp_error_handling_t e12 = {.cmd = XCP_CMD_GET_ID, .subcmd = 0, .error= XCP_ERROR_CMD_BUSY, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[12] = e12;

  //Set MTA
  cs_xcp_error_handling_t e13 = {.cmd = XCP_CMD_SET_MTA, .subcmd = 0, .error= XCP_ERROR_TIMEOUTT1, .preaction = XCP_PA_SYNCH, .action = XCP_AC_REPEAT2};
  data->error_handling[13] = e13;
  cs_xcp_error_handling_t e14 = {.cmd = XCP_CMD_SET_MTA, .subcmd = 0, .error= XCP_ERROR_CMD_BUSY, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[14] = e14;
  cs_xcp_error_handling_t e15 = {.cmd = XCP_CMD_SET_MTA, .subcmd = 0, .error= XCP_ERROR_PGM_ACTIVE, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[15] = e15;

  //Upload
  cs_xcp_error_handling_t e16 = {.cmd = XCP_CMD_UPLOAD, .subcmd = 0, .error= XCP_ERROR_TIMEOUTT1, .preaction = XCP_PA_SYNCH | XCP_PA_SETMTA, .action = XCP_AC_REPEAT2};
  data->error_handling[16] = e16;
  cs_xcp_error_handling_t e17 = {.cmd = XCP_CMD_UPLOAD, .subcmd = 0, .error= XCP_ERROR_CMD_BUSY, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[17] = e17;
  cs_xcp_error_handling_t e18 = {.cmd = XCP_CMD_UPLOAD, .subcmd = 0, .error= XCP_ERROR_PGM_ACTIVE, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[18] = e18;
  cs_xcp_error_handling_t e19 = {.cmd = XCP_CMD_GET_DAQ_PROCESSOR_INFO, .subcmd = 0, .error= XCP_ERROR_TIMEOUTT1, .preaction = XCP_PA_SYNCH, .action = XCP_AC_REPEAT2};
  data->error_handling[19] = e19;
  cs_xcp_error_handling_t e20 = {.cmd = XCP_CMD_GET_DAQ_PROCESSOR_INFO, .subcmd = 0, .error= XCP_ERROR_CMD_BUSY, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[20] = e20;
  cs_xcp_error_handling_t e21 = {.cmd = XCP_CMD_GET_DAQ_PROCESSOR_INFO, .subcmd = 0, .error= XCP_ERROR_PGM_ACTIVE, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[21] = e21;
  cs_xcp_error_handling_t e22 = {.cmd = XCP_CMD_GET_DAQ_RESOLUTION_INFO, .subcmd = 0, .error= XCP_ERROR_TIMEOUTT1, .preaction = XCP_PA_SYNCH, .action = XCP_AC_REPEAT2};
  data->error_handling[22] = e22;
  cs_xcp_error_handling_t e23 = {.cmd = XCP_CMD_GET_DAQ_RESOLUTION_INFO, .subcmd = 0, .error= XCP_ERROR_CMD_BUSY, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[23] = e23;
  cs_xcp_error_handling_t e24 = {.cmd = XCP_CMD_GET_DAQ_RESOLUTION_INFO, .subcmd = 0, .error= XCP_ERROR_PGM_ACTIVE, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[24] = e24;
  cs_xcp_error_handling_t e25 = {.cmd = XCP_CMD_FREE_DAQ, .subcmd = 0, .error= XCP_ERROR_TIMEOUTT1, .preaction = XCP_PA_SYNCH, .action = XCP_AC_REPEAT2};
  data->error_handling[25] = e25;
  cs_xcp_error_handling_t e26 = {.cmd = XCP_CMD_FREE_DAQ, .subcmd = 0, .error= XCP_ERROR_CMD_BUSY, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[26] = e26;
  cs_xcp_error_handling_t e27 = {.cmd = XCP_CMD_FREE_DAQ, .subcmd = 0, .error= XCP_ERROR_PGM_ACTIVE, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[27] = e27;
  cs_xcp_error_handling_t e28 = {.cmd = XCP_CMD_ALLOC_DAQ, .subcmd = 0, .error= XCP_ERROR_TIMEOUTT1, .preaction = XCP_PA_SYNCH, .action = XCP_AC_REPEAT2};
  data->error_handling[28] = e28;
  cs_xcp_error_handling_t e29 = {.cmd = XCP_CMD_ALLOC_DAQ, .subcmd = 0, .error= XCP_ERROR_CMD_BUSY, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[29] = e29;
  cs_xcp_error_handling_t e30 = {.cmd = XCP_CMD_ALLOC_DAQ, .subcmd = 0, .error= XCP_ERROR_PGM_ACTIVE, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[30] = e30;
  cs_xcp_error_handling_t e31 = {.cmd = XCP_CMD_ALLOC_ODT, .subcmd = 0, .error= XCP_ERROR_TIMEOUTT1, .preaction = XCP_PA_SYNCH, .action = XCP_AC_REPEAT2};
  data->error_handling[31] = e31;
  cs_xcp_error_handling_t e32 = {.cmd = XCP_CMD_ALLOC_ODT, .subcmd = 0, .error= XCP_ERROR_CMD_BUSY, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[32] = e32;
  cs_xcp_error_handling_t e33 = {.cmd = XCP_CMD_ALLOC_ODT, .subcmd = 0, .error= XCP_ERROR_PGM_ACTIVE, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[33] = e33;
  cs_xcp_error_handling_t e34 = {.cmd = XCP_CMD_CLEAR_DAQ_LIST, .subcmd = 0, .error= XCP_ERROR_TIMEOUTT1, .preaction = XCP_PA_SYNCH, .action = XCP_AC_REPEAT2};
  data->error_handling[34] = e34;
  cs_xcp_error_handling_t e35 = {.cmd = XCP_CMD_CLEAR_DAQ_LIST, .subcmd = 0, .error= XCP_ERROR_CMD_BUSY, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[35] = e35;
  cs_xcp_error_handling_t e36 = {.cmd = XCP_CMD_CLEAR_DAQ_LIST, .subcmd = 0, .error= XCP_ERROR_PGM_ACTIVE, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[36] = e36;

  cs_xcp_error_handling_t e37 = {.cmd = XCP_CMD_SET_DAQ_PTR, .subcmd = 0, .error= XCP_ERROR_TIMEOUTT1, .preaction = XCP_PA_SYNCH, .action = XCP_AC_REPEAT2};
  data->error_handling[37] = e37;
  cs_xcp_error_handling_t e38 = {.cmd = XCP_CMD_SET_DAQ_PTR, .subcmd = 0, .error= XCP_ERROR_CMD_BUSY, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[38] = e38;
  cs_xcp_error_handling_t e39 = {.cmd = XCP_CMD_SET_DAQ_PTR, .subcmd = 0, .error= XCP_ERROR_PGM_ACTIVE, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[39] = e39;

  cs_xcp_error_handling_t e40 = {.cmd = XCP_CMD_WRITE_DAQ, .subcmd = 0, .error= XCP_ERROR_TIMEOUTT1, .preaction = XCP_PA_SYNCH, .action = XCP_AC_REPEAT2};
  data->error_handling[40] = e40;
  cs_xcp_error_handling_t e41 = {.cmd = XCP_CMD_WRITE_DAQ, .subcmd = 0, .error= XCP_ERROR_CMD_BUSY, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[41] = e41;
  cs_xcp_error_handling_t e42 = {.cmd = XCP_CMD_WRITE_DAQ, .subcmd = 0, .error= XCP_ERROR_PGM_ACTIVE, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[42] = e42;

  cs_xcp_error_handling_t e43 = {.cmd = XCP_CMD_SET_DAQ_LIST_MODE, .subcmd = 0, .error= XCP_ERROR_TIMEOUTT1, .preaction = XCP_PA_SYNCH, .action = XCP_AC_REPEAT2};
  data->error_handling[43] = e43;
  cs_xcp_error_handling_t e44 = {.cmd = XCP_CMD_SET_DAQ_LIST_MODE, .subcmd = 0, .error= XCP_ERROR_CMD_BUSY, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[44] = e44;
  cs_xcp_error_handling_t e45 = {.cmd = XCP_CMD_SET_DAQ_LIST_MODE, .subcmd = 0, .error= XCP_ERROR_PGM_ACTIVE, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[45] = e45;

  cs_xcp_error_handling_t e46 = {.cmd = XCP_CMD_GET_DAQ_LIST_MODE, .subcmd = 0, .error= XCP_ERROR_TIMEOUTT1, .preaction = XCP_PA_SYNCH, .action = XCP_AC_REPEAT2};
  data->error_handling[46] = e46;
  cs_xcp_error_handling_t e47 = {.cmd = XCP_CMD_GET_DAQ_LIST_MODE, .subcmd = 0, .error= XCP_ERROR_CMD_BUSY, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[47] = e47;
  cs_xcp_error_handling_t e48 = {.cmd = XCP_CMD_GET_DAQ_LIST_MODE, .subcmd = 0, .error= XCP_ERROR_PGM_ACTIVE, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[48] = e48;

  cs_xcp_error_handling_t e49 = {.cmd = XCP_CMD_ALLOC_ODT_ENTRY, .subcmd = 0, .error= XCP_ERROR_TIMEOUTT1, .preaction = XCP_PA_SYNCH, .action = XCP_AC_REPEAT2};
  data->error_handling[49] = e49;
  cs_xcp_error_handling_t e50 = {.cmd = XCP_CMD_ALLOC_ODT_ENTRY, .subcmd = 0, .error= XCP_ERROR_CMD_BUSY, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[50] = e50;
  cs_xcp_error_handling_t e51 = {.cmd = XCP_CMD_ALLOC_ODT_ENTRY, .subcmd = 0, .error= XCP_ERROR_PGM_ACTIVE, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[51] = e51;

  cs_xcp_error_handling_t e52 = {.cmd = XCP_CMD_ALLOC_ODT_ENTRY, .subcmd = 0, .error= XCP_ERROR_TIMEOUTT1, .preaction = XCP_PA_SYNCH, .action = XCP_AC_REPEAT2};
  data->error_handling[52] = e52;
  cs_xcp_error_handling_t e53 = {.cmd = XCP_CMD_ALLOC_ODT_ENTRY, .subcmd = 0, .error= XCP_ERROR_CMD_BUSY, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[53] = e53;
  cs_xcp_error_handling_t e54 = {.cmd = XCP_CMD_ALLOC_ODT_ENTRY, .subcmd = 0, .error= XCP_ERROR_PGM_ACTIVE, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[54] = e54;

  cs_xcp_error_handling_t e55 = {.cmd = XCP_CMD_START_STOP_SYNCH, .subcmd = 0, .error= XCP_ERROR_TIMEOUTT1, .preaction = XCP_PA_SYNCH, .action = XCP_AC_REPEAT2};
  data->error_handling[55] = e55;
  cs_xcp_error_handling_t e56 = {.cmd = XCP_CMD_START_STOP_SYNCH, .subcmd = 0, .error= XCP_ERROR_CMD_BUSY, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[56] = e56;
  cs_xcp_error_handling_t e57 = {.cmd = XCP_CMD_START_STOP_SYNCH, .subcmd = 0, .error= XCP_ERROR_PGM_ACTIVE, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[57] = e57;

  cs_xcp_error_handling_t e58 = {.cmd = XCP_CMD_GET_DAQ_CLOCK, .subcmd = 0, .error= XCP_ERROR_TIMEOUTT1, .preaction = XCP_PA_SYNCH, .action = XCP_AC_REPEAT2};
  data->error_handling[58] = e58;
  cs_xcp_error_handling_t e59 = {.cmd = XCP_CMD_GET_DAQ_CLOCK, .subcmd = 0, .error= XCP_ERROR_CMD_BUSY, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[59] = e59;
  cs_xcp_error_handling_t e60 = {.cmd = XCP_CMD_GET_DAQ_CLOCK, .subcmd = 0, .error= XCP_ERROR_PGM_ACTIVE, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[60] = e60;

  cs_xcp_error_handling_t e61 = {.cmd = XCP_CMD_GET_DAQ_EVENT_INFO, .subcmd = 0, .error= XCP_ERROR_TIMEOUTT1, .preaction = XCP_PA_SYNCH, .action = XCP_AC_REPEAT2};
  data->error_handling[61] = e61;
  cs_xcp_error_handling_t e62 = {.cmd = XCP_CMD_GET_DAQ_EVENT_INFO, .subcmd = 0, .error= XCP_ERROR_CMD_BUSY, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[62] = e62;
  cs_xcp_error_handling_t e63 = {.cmd = XCP_CMD_GET_DAQ_EVENT_INFO, .subcmd = 0, .error= XCP_ERROR_PGM_ACTIVE, .preaction = XCP_PA_WAIT_T7, .action = XCP_AC_REPEAT};
  data->error_handling[63] = e63;
}

/** Get Failure Preaction command
    \retval Preaction command
*/
/**************************************************************************/
cs_xcp_pa_t _xcp_get_preaction(
			       cs_xcp_t *data,        /**< XCP stack data pointer*/
			       uint8_t cmd,           /**< Actual command*/
			       uint8_t subcmd,        /**< Optional sub command, default 0*/
			       uint16_t error         /**< Received error*/
			       )
/**************************************************************************/
{
  int i;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return XCP_PA_NONE;
  }
    
  for(i = 0; i < CS_XCP_ERROR_HANDLINGLENGTH; i++){
    if((data->error_handling[i].cmd == cmd) &&
       (data->error_handling[i].subcmd == subcmd) &&
       (data->error_handling[i].error == error)){
      _xcp_verbose("Cmd %i: Get Preaction Cmd: %i\n",cmd,data->error_handling[i].preaction);
      return data->error_handling[i].preaction;
    }
  }
  return XCP_PA_NONE;
}

/** Get Failure Action command
    \retval Action command
*/
/**************************************************************************/
cs_xcp_ac_t _xcp_get_action(
			    cs_xcp_t *data,        /**< XCP stack data pointer*/
			    uint8_t cmd,           /**< Actual command*/
			    uint8_t subcmd,        /**< Optional sub command, default 0*/
			    uint16_t error         /**< Received error*/
			    )
/**************************************************************************/
{
  int i;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return XCP_AC_NONE;
  }
    
  for(i = 0; i < CS_XCP_ERROR_HANDLINGLENGTH; i++){
    if((data->error_handling[i].cmd == cmd) &&
       (data->error_handling[i].subcmd == subcmd) &&
       (data->error_handling[i].error == error)){
      _xcp_verbose("Cmd %i: Get Action Cmd: %i\n",cmd,data->error_handling[i].action);
      return data->error_handling[i].action;
    }
  }
  return XCP_AC_NONE;
}

/** Reset resources structure
 */
/**************************************************************************/
void _xcp_reset_resources(
			  cs_xcp_t *data
			  )
/**************************************************************************/
{
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return;
  }
  
  data->resources.CALPAG = 0;
  data->resources.DAQ = 0;
  data->resources.STIM = 0;
  data->resources.PGM = 0;
}

/** Reset internal error flag
 */
/**************************************************************************/
void _xcp_reset_error(
		      cs_xcp_t *data
		      )
/**************************************************************************/
{
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return;
  }
  
  data->lerror = XCP_ERROR_NOERROR;
}




/** Reset ODT enry
 */
/**************************************************************************/
void _xcp_reset_odt_entry(
			  cs_xcp_odt_entry_t *entry
			  )
/**************************************************************************/
{
  if(!entry){
    _xcp_error("Parameter failure\n");
    return;
  }

  _xcp_verbose("_XCP Reset ODT entry\n");  
  memset(entry->data,0,sizeof(uint8_t)*CS_XCP_DATALENGTH);
  cs_a2l_reset_measurement(&entry->me);

}

/** Reset ODT
 */
/**************************************************************************/
void _xcp_reset_odt(
		    cs_xcp_odt_t *odt
		    )
/**************************************************************************/
{
  int i;
  if(!odt){
    _xcp_error("Parameter failure\n");
    return;
  }

  _xcp_verbose("_XCP Reset ODT\n");
  for(i = 0; i < odt->no_entry; i++){
    if(&odt->entry[i])
      _xcp_reset_odt_entry(&odt->entry[i]);
  }
  odt->entry = NULL;
  odt->no_entry = 0;
}

/** Reset DAQ entry
 */
/**************************************************************************/
void _xcp_reset_daq(
		    cs_xcp_daq_t *daq
		    )
/**************************************************************************/
{
  int i;

  if(!daq){
    _xcp_error("Parameter failure\n");
    return;
  }
  
  _xcp_verbose("_XCP Reset DAQ\n");

  if(daq->odt){
    for(i = 0; i< daq->no_odt; i++){
      _xcp_reset_odt(&daq->odt[i]);
    }
    free(daq->odt);
    daq->odt = NULL;
    daq->no_odt = 0;
    daq->resume = 0;            
    daq->running = 0;      
    daq->selected = 0;
    daq->ts_old = 0;
    daq->counter = 0;
    daq->canid = 0;
  }
    
}

/** Reset DAQ list
 */
/**************************************************************************/
void _xcp_reset_daqlist(
			cs_xcp_t *data
			)
/**************************************************************************/
{
  int i;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return;
  }
    
  _xcp_verbose("_XCP Reset DAQ list\n");
  if((data->daqlist.no_daq > 0) && (data->daqlist.daq != NULL)){
    for(i = 0; i < data->daqlist.no_daq; i++){
      _xcp_reset_daq(&data->daqlist.daq[i]);
    }
    free(data->daqlist.daq);
    data->daqlist.daq = NULL;
    data->daqlist.no_daq = 0;
  }
  data->work_entry = NULL;
  data->daqlist.clock_shift = 0;
  data->daqlist.clock_base = 0;
}

/** Handle failure preaction
 */
/**************************************************************************/
int _xcp_handler_preaction(
			   cs_xcp_t *data,
			   uint8_t cmd,
			   uint8_t subcmd,
			   cs_xcp_pa_t preaction
			   )
/**************************************************************************/
{

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  if(preaction == XCP_PA_NONE){
    //Nothing to do
  } 
  if(preaction & XCP_PA_WAIT_T7){
    //Sleep
    cs_ts_sleep(data->xcpa2l.commode.timeout[XCP_A2L_TO_T7]);
  }
  if(preaction & XCP_PA_SYNCH){
    //Prefend recursive call
    if(cmd !=  XCP_CMD_SYNCH)
      if(cs_xcp_cmd_synch(data))
	return CS_FAIL;
  }
  if(preaction & XCP_PA_SETMTA){
    if(cs_xcp_cmd_set_mta(data,data->mta.address,data->mta.extension))
      return CS_FAIL;
  }
  return CS_OK;
}

/** Handle failure action
 */
/**************************************************************************/
int _xcp_handler_action(
			cs_xcp_t *data,
			uint8_t cmd,
			uint8_t subcmd,
			cs_xcp_ac_t action,
			int *repeat
			)
/**************************************************************************/
{

  if((data == NULL) ||
     (repeat == NULL)){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }

  *repeat = 0;
    
  switch(action){
  case XCP_AC_NONE:
    //Nothing to do, Failure 
    return CS_FAIL;
  case XCP_AC_REPEAT2:
    *repeat = 2;
    break;
  case XCP_AC_REPEAT:
    *repeat = CS_XCP_REPEATMAX;
    break;
  case XCP_AC_OK:
    //reset failure;
    _xcp_reset_error(data);
    return CS_OK;
    break;
  }
  return CS_OK;
}

/** Preinit package data
 */
/**************************************************************************/
void _xcp_reset_pkg(
		    cs_xcp_pkg_t *pkg
		    )
/**************************************************************************/
{
  if(pkg == NULL){
    _xcp_error("Parameter failure\n");
    return;
  }
  
  memset(pkg,0,sizeof(cs_xcp_pkg_t));
}

/** Search odt by pid
    \retval Pointer to odt structure
    \retval NULL in failure case
*/
/**************************************************************************/
cs_xcp_odt_t *_xcp_search_odt(
			      cs_xcp_t *data,
			      uint16_t pid
			      )
/**************************************************************************/
{
  int i;
  
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return NULL;
  }
  

  for(i = 0; i < data->daqlist.no_daq; i++){
    //TODO
  }
  _xcp_error("Sorry not implemented\n");
  
  return NULL;
}

/**************************************************************************/
cs_xcp_indicator_t cs_xcp_get_indicator(
					cs_xcp_t *data
					)
/**************************************************************************/
{
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return XCP_INDICATOR_INVALID;
  }
  return data->indicator;
}

/**************************************************************************/
void cs_xcp_reset_indicator(
			    cs_xcp_t *data,
			    cs_xcp_indicator_t indicator
			    )
/**************************************************************************/
{
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return;
  }
  data->indicator &= (~indicator);
}

/** Received DTO, fill into internal structure
    \retval 0 No failure
    \retval 1 Failure
*/
/**************************************************************************/
int _xcp_dto_daq_process(
			 cs_xcp_t *data,
			 uint32_t canid,
			 uint8_t *sdata,
			 int ldata,
			 struct timeval tv
			 )
/**************************************************************************/
{
  int i,j;
  int data_offset = 0;
  uint16_t odt_pid;
  int daq_no;
  cs_xcp_daq_t *daq = NULL;
  cs_xcp_odt_t *odt = NULL;
  cs_cu_data16_t l16;
  cs_cu_data32_t l32;
  int flag_first_odt;
  uint32_t ts_diff = 0;
  double value;

  if((data == NULL) ||
     (sdata == NULL)){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
    
  _xcp_verbose("Start\n");

  if(data->iftype == XCP_CAN){
    if(canid != cs_xcp_can_getRespId(data->ifcan)){
      //DAQ identified by CAN id, actually not supported
      return CS_FAIL;
      for(i = 0; i < data->daqlist.no_daq;i++){
	if((canid == data->daqlist.daq[i].canid) && (data->daqlist.daq[i].dir == XCP_DAQ_DIR_DAQ)){
	  //Found daq
	  daq = &data->daqlist.daq[i];
	  //todo act on data
	}
      }
    }
  }

  //Do we have to check Overload via MSB in ODT?
  if((data->xcpa2l.daq_proc.overload_mode == XCP_A2L_ODT_OVERLOAD_MSB) && (sdata[0] > XCP_OVERFLOW_IDENT_MAXODT)){
    //Overload indication detected
    _xcp_info("Overload indication detected.\n");
    odt_pid = (uint16_t)(sdata[0] & XCP_OVERFLOW_IDENT_MAXODT); //Mask out MSB
    data->indicator |= XCP_INDICATOR_OVERLOAD;
  } else {
    odt_pid = (uint16_t)sdata[0]; //Get PID
  }

  switch(data->xcpa2l.daq_proc.odt_ident){
  case XCP_A2L_ODT_IDENT_ABS:
    _xcp_verbose("Absolute ODT PID\n");
    data_offset = 1; //Skip PID
    //search odt
    for(i = 0; i < data->daqlist.no_daq;i++){
      for(j =0; j < data->daqlist.daq[i].no_odt; j++){
	if((data->daqlist.daq[i].dir == XCP_DAQ_DIR_DAQ) &&
	   (((data->daqlist.daq[i].first_pid + j) == odt_pid) && data->daqlist.daq[i].running)){
	  _xcp_verbose("Found odt: %i, daq %i, odt index %i\n",odt_pid,i,j);
	  if(j == 0){
	    flag_first_odt = 1;
	  } else {
	    flag_first_odt = 0;
	  }
	  odt = &data->daqlist.daq[i].odt[j];
	  daq = &data->daqlist.daq[i];
	  break;
	}
      }
    }
    break;
  case XCP_A2L_ODT_IDENT_RAB:
    _xcp_verbose("Relative ODT PID, absolute DAQ (Byte)\n");
    data_offset = 2; //Skip PID, and DAQ byte
    l16.u8[0] = sdata[1];
    l16.u8[1] = 0;
    daq_no = (int)cs_cu_swap16(data->xcpa2l.commode.BO,l16.u16);
    if(daq_no >= data->daqlist.no_daq){
      data->lerror = XCP_ERROR_DAQ_PROCESS;
      return CS_FAIL;
    }
    daq = &data->daqlist.daq[daq_no];
    if(daq->dir != XCP_DAQ_DIR_DAQ){
      data->lerror = XCP_ERROR_DAQ_PROCESS;
      return CS_FAIL;
    }
    for(j =0; j < daq->no_odt; j++){
      if(((daq->first_pid + j) == odt_pid) && daq->running){
	_xcp_verbose("Found odt: %i, daq %i, odt index %i\n",odt_pid,daq_no,j);
	if(j == 0){
	  flag_first_odt = 1;
	} else {
	  flag_first_odt = 0;
	}
	odt = &daq->odt[j];
	break;
      }
    }
    break;
  case XCP_A2L_ODT_IDENT_RAW:
    _xcp_verbose("Relative ODT PID, absolute DAQ (Word)\n");
    data_offset = 3; //Skip PID, and DAQ word
    l16.u8[0] = sdata[1];
    l16.u8[1] = sdata[2];
    daq_no = (int)cs_cu_swap16(data->xcpa2l.commode.BO,l16.u16);
    if(daq_no >= data->daqlist.no_daq){
      data->lerror = XCP_ERROR_DAQ_PROCESS;
      return CS_FAIL;
    }
    daq = &data->daqlist.daq[daq_no];
    if(daq->dir != XCP_DAQ_DIR_DAQ){
      data->lerror = XCP_ERROR_DAQ_PROCESS;
      return CS_FAIL;
    }
    for(j =0; j < daq->no_odt; j++){
      if(((daq->first_pid + j) == odt_pid) && daq->running){
	_xcp_verbose("Found odt: %i, daq %i, odt index %i\n",odt_pid,daq_no,j);
	if(j == 0){
	  flag_first_odt = 1;
	} else {
	  flag_first_odt = 0;
	}
	odt = &daq->odt[j];
	break;
      }
    }
    break;
  case XCP_A2L_ODT_IDENT_RAWA:
    _xcp_verbose("Relative ODT PID, absolute DAQ (Word,aligned)\n");
    data_offset = 4; //Skip PID, and DAQ word and fill byte
    l16.u8[0] = sdata[2];
    l16.u8[1] = sdata[3];
    daq_no = cs_cu_swap16(data->xcpa2l.commode.BO,l16.u16);
    if(daq_no >= data->daqlist.no_daq){
      data->lerror = XCP_ERROR_DAQ_PROCESS;
      return CS_FAIL;
    }
    daq = &data->daqlist.daq[daq_no];
    if(daq->dir != XCP_DAQ_DIR_DAQ){
      data->lerror = XCP_ERROR_DAQ_PROCESS;
      return CS_FAIL;
    }
    for(j =0; j < daq->no_odt; j++){
      if(((daq->first_pid + j) == odt_pid) && daq->running){
	_xcp_verbose("Found odt: %i, daq %i, odt index %i\n",odt_pid,daq_no,j);
	if(j == 0){
	  flag_first_odt = 1;
	} else {
	  flag_first_odt = 0;
	}
	odt = &daq->odt[j];
	break;
      }
    }
    break;
  }
  
  if(odt == NULL){
    //Skip ODT, is not known by system
    //data->lerror = XCP_ERROR_ODT_UNKNOWN;
    _xcp_debug("Unknown ODT\n");
    return CS_OK;
  }

  //Calculate data offset
  if(flag_first_odt){
    if(daq->timestamp){
      if((ldata - data_offset) < data->xcpa2l.daq_proc.ts_size){
	data->lerror = XCP_ERROR_ODT_SIZE;
	_xcp_error("Invalid ODT size\n");
	return CS_FAIL;
      }
      l32.u32 = 0;
      for(i = 0; i < data->xcpa2l.daq_proc.ts_size; i++){
	l32.u8[i] = sdata[data_offset + i];
      }
      l32.u32 = cs_cu_swap32(data->xcpa2l.commode.BO,l32.u32);
      //Calc diff with old timestamp
      switch(data->xcpa2l.daq_proc.ts_size){
      case 1:
	ts_diff = (uint8_t)((uint8_t)l32.u32 -(uint8_t)daq->ts_old );
	break;
      case 2:
	ts_diff = (uint16_t)((uint16_t)l32.u32 - (uint16_t)daq->ts_old);
	break;
      case 4:
	ts_diff = (uint32_t)l32.u32 - (uint32_t)daq->ts_old;
	break;
      default:
	ts_diff = 0;
	break;
      }
      //if it is the first received ODT set clock to receive timestamp
      if(daq->ts_ref == 0.0)
	daq->ts_ref = cs_ts_timeval2d(tv);
      // Add diff to reference
      daq->ts_ref += cs_ts_timespec2d(cs_cu_ts_timestamp2ts(ts_diff,(uint8_t)data->xcpa2l.daq_proc.ts_unit));
      //Store old timestamp
      daq->ts_old = l32.u32;
      data_offset += data->xcpa2l.daq_proc.ts_size;
    } else {
      //No timestamp available, take receive timestamp of first ODT
      daq->ts_ref = cs_ts_timeval2d(tv);
    }
    daq->counter++;
    _xcp_verbose("Counter %i\n",daq->counter);
  }

  //TODO Timestamp, and relative PID and DAQ
  if(odt->datasize > (ldata - data_offset)){
    _xcp_error("ODT expected length %i, is %i, Raw length %i:\n",odt->datasize,ldata-data_offset,ldata);
    _xcp_error_buffer(sdata,ldata);
    data->lerror = XCP_ERROR_ODT_SIZE;
    return CS_FAIL;
  }

  //Set receive timestamp for ODT, to be able to check  for receive timeout
  odt->last = cs_ts_timeval2d(tv);

  _xcp_verbose("Data offset: %i\n",data_offset);

  //Do data handling
  //note only data copy, no byte swapping
  for(i = 0; i < odt->no_entry; i++){
    _xcp_verbose("Entry %i, copy size: %i\n",i,odt->entry[i].me.size * cs_a2l_get_ag_bytes(data->xcpa2l.commode.AG));
    //Memcopy
    memcpy(odt->entry[i].data,&sdata[data_offset],odt->entry[i].me.size * cs_a2l_get_ag_bytes(data->xcpa2l.commode.AG));
    
    //Calculate value
    value = 0.0;
    cs_xcp_convert_to_double(data,
			     odt->entry[i].data,
			     odt->entry[i].me.size * cs_a2l_get_ag_bytes(data->xcpa2l.commode.AG),
			     odt->entry[i].me.cm,
			     odt->entry[i].me.dtype,
			     &value);
    
    //Connect the received value with external references

    //Do we have an external value pointer
    if(odt->entry[i].me.exvalue!= NULL){
      *odt->entry[i].me.exvalue = value;
    }
    //Do we have an external timestamp pointer
    if(odt->entry[i].me.exts!= NULL){
      *odt->entry[i].me.exts = daq->ts_ref;
    }
    
    //Do we have an external callback funtion
    if(odt->entry[i].me.extcallback != NULL){
      //Call external function
      (*odt->entry[i].me.extcallback)(value,daq->ts_ref,odt->entry[i].me.extcallbackpriv);
    }

    //Increment data offset
    data_offset += odt->entry[i].me.size * cs_a2l_get_ag_bytes(data->xcpa2l.commode.AG);

  }

  return CS_OK;
}

/** Internal send routine
 */
/**************************************************************************/
int _xcp_send(
	      cs_xcp_t *data,
	      cs_xcp_pkg_t *pkg
	      )
/**************************************************************************/
{
  uint8_t sdata[CS_XCP_FRAMELENGTH];
  int ldata = 0;
  uint32_t can_id = 0;
  struct timespec ts;

  if((data == NULL) ||
     (pkg == NULL)){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }

    
  if(pkg->dl > CS_XCP_DATALENGTH){
    _xcp_error("Invalid Data segement length\n");
    return CS_FAIL;
  }
  
  memset(sdata,0,sizeof(uint8_t)*CS_XCP_FRAMELENGTH);
  
  ldata = pkg->dl;
  
  switch(pkg->pkgtype){
  case XCP_PKG_NONE:
    data->lerror = XCP_ERROR_API;
    break;
  case XCP_PKG_CMD:
    if(ldata > (data->ifmax -1)){
      data->lerror = XCP_ERROR_API;
      break;
    }
    sdata[0] = pkg->PID;
    memcpy(&sdata[1],pkg->data,ldata);
    ldata +=1;
    can_id = cs_xcp_can_getCmdId(data->ifcan);
    break;
  default:
    data->lerror = XCP_ERROR_UNSUPPORTED;
  }
  
  //Get send timestamp
  cs_ts_gettime(&ts);
  //Store time
  pkg->tv = cs_ts_timespec2timeval(ts);

  if(data->lerror == XCP_ERROR_NOERROR){
    switch(data->iftype){
    case XCP_CAN:
      if(cs_xcp_can_send(data->ifcan,can_id,sdata,ldata)){
	data->lerror = XCP_ERROR_SEND;
      }
      break;
    case XCP_USB:
      data->lerror = XCP_ERROR_UNSUPPORTED;
      break;
    case XCP_ETHERNET:
      data->lerror = XCP_ERROR_UNSUPPORTED;
      break;
    default:
      data->lerror = XCP_ERROR_INTERFACE;    
    }    
  }

  if(data->lerror != XCP_ERROR_NOERROR)
    return CS_FAIL;
  else
    return CS_OK;
}

/** Internal receive routine
 */
/**************************************************************************/
int _xcp_receive(
		 cs_xcp_t *data,
		 cs_xcp_pkg_t *pkg,
		 uint32_t timeout,
		 int mode_odt
		 )
/**************************************************************************/
{
  uint8_t sdata[CS_XCP_FRAMELENGTH];
  int ldata;
  uint32_t can_id;
  struct timespec ts_timeout;
  uint8_t do_receive = 1;
  uint8_t received_pkg = 0;

  if((data == NULL) ||
     (pkg == NULL)){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
    
  //Get actual time
  cs_ts_gettime(&ts_timeout);
  //Add timeout
  cs_ts_add_timeout(&ts_timeout,timeout);

  pkg->dl = 0;

  //_xcp_debug("Receive buffer:\n");
  //_xcp_debug_buffer(sdata,data->ifmax);

  while(do_receive){
    //Reset raw buffer
    ldata = data->ifmax;
    memset(sdata,0,sizeof(uint8_t)*CS_XCP_FRAMELENGTH);
    _xcp_verbose("Receive ... Mode: %s\n",mode_odt?"ODT":"Standard");
    //#if 0
    //Check for timeout if we have received one or more ODT package
    if(!mode_odt && cs_ts_checkpast(&ts_timeout)){
      _xcp_error("Timeout error, timestamp\n");
      data->lerror = XCP_ERROR_TIMEOUT;
      do_receive = 0;
      break;
    }
    //#endif
    //Receive
    switch(data->iftype){
    case XCP_CAN:
      _xcp_verbose("Receive via CAN\n");
      received_pkg = 0;
      if(!mode_odt){
	_xcp_verbose("Exec standard mode\n");
	if(cs_xcp_can_recv(data->ifcan,&can_id,sdata,&ldata,timeout,&pkg->tv)){
	  _xcp_error("Timeout error\n");
	  data->lerror = XCP_ERROR_TIMEOUT;
	  do_receive = 0;
	} else {
	  received_pkg = 1;
	}
	if(received_pkg && (can_id != cs_xcp_can_getRespId(data->ifcan))){ //Receive a daq odt with special CAN id
	  _xcp_error("Unknown CAN id 0x%X, try DAQ processing\n",can_id);
	  _xcp_dto_daq_process(data,can_id,sdata,ldata,pkg->tv);
	  received_pkg = 0;	
	}
      } else {
	_xcp_verbose("Exec ODT mode\n");
	if(cs_xcp_can_recv_nonblock(data->ifcan,&can_id,sdata,&ldata,&pkg->tv)){
	  data->lerror = XCP_ERROR_NOERROR;
	  do_receive = 0;
	} else {
	  received_pkg = 1;
	}
      }
      break;
    case XCP_USB:
      _xcp_error("USB is not supported\n");
      data->lerror = XCP_ERROR_UNSUPPORTED;
      do_receive = 0;
      break;
    case XCP_ETHERNET:
      _xcp_error("Ethernet is not supported\n");
      data->lerror = XCP_ERROR_UNSUPPORTED;
      do_receive = 0;
      break;       
    default:
      _xcp_error("Unknown interface is not supported\n");
      data->lerror = XCP_ERROR_INTERFACE;
      do_receive = 0;
      break;
    }
    
    if(received_pkg){
      _xcp_verbose("Received package\n");
      //Chek for invalid data length
      if(!mode_odt){
	if(ldata == 0){
	  data->lerror = XCP_ERROR_RECV;
	  do_receive = 0;
	}
      } else {
	if(ldata == 0){
	  pkg->dl = 0;
	  _xcp_error("ODT mode: Invalid data length\n");
	  return CS_FAIL;
	}
      }

      //Process raw data
      if((ldata > 0) && (data->lerror == XCP_ERROR_NOERROR)){
	switch(sdata[0]){
	case XCP_PKG_RESP:
	  _xcp_verbose("Received Response paket\n");
	  //Process Response package
	  pkg->PID = XCP_PKG_RESP;
	  pkg->dl = ldata -1;
	  //Check for buffer overflow
	  if((ldata -1) > CS_XCP_DATALENGTH){
	    data->lerror = XCP_ERROR_BUFFER;
	  } else {
	    memcpy(pkg->data,&sdata[1],pkg->dl);
	    do_receive = 0;
	  }
	  break;
	case XCP_PKG_EV:
	  //Process Event package
	  switch(sdata[1]){
	  case XCP_EV_CMD_PENDING:
	    //Reset timeout
	    //Get actual time
	    cs_ts_gettime(&ts_timeout);
	    //Add timeout
	    cs_ts_add_timeout(&ts_timeout,timeout);
	    _xcp_verbose("Received a Command pending package. Reset timeout.\n");
	    break;
	  case XCP_EV_DAQ_OVERLOAD:
	    //Overload indication
	    if(data->xcpa2l.daq_proc.overload_mode != XCP_A2L_ODT_OVERLOAD_EV){
	      data->lerror = XCP_ERROR_UNEXPECTED;
	      _xcp_error("Received unexpected Overload indication paket.\n");
	      return CS_FAIL;
	    } else {
	      _xcp_debug("Received Overload indication paket.\n");
	      data->indicator |= XCP_INDICATOR_OVERLOAD;
	    }
	    break;
	  case XCP_EV_SESSION_TERMINATED:
	    //Session was terminated
	    data->indicator |= XCP_INDICATOR_EVENT_SESSION_TERMINATED;
	    break;
	  case XCP_EV_TRANSPORT:
	    //Transport layer event
	    data->indicator |= XCP_INDICATOR_EVENT_TRANSPORT;
	    break;
	  case XCP_EV_RESUME_MODE:
	    //Slave started in resume mode
	    data->indicator |= XCP_INDICATOR_EVENT_RESUME;
	    //TODO Data handling
	    break;
	  case XCP_EV_CLEAR_DAQ:
	    //DAQ configuration was cleared
	    data->indicator |= XCP_INDICATOR_EVENT_CLEAR_DAQ;
	    break;
	  case XCP_EV_STORE_DAQ:
	    //DAQ configuration was stored
	    data->indicator |= XCP_INDICATOR_EVENT_STORE_DAQ;
	  case XCP_EV_STORE_CAL:
	    //Calibration was stored
	    data->indicator |= XCP_INDICATOR_EVENT_STORE_CAL;
	    break;
	  default:
	    return CS_FAIL;
	  }
	  break;
	case XCP_PKG_ERR:
	  _xcp_verbose("Received Error paket\n");
	  //Process Error package
	  pkg->PID = XCP_PKG_ERR;
	  pkg->dl = ldata -1;
	  //Check for minimum length
	  if(pkg->dl < 1){
	    data->lerror = XCP_ERROR_UNEXPECTED;
	    do_receive = 0;
	  } else {
	    pkg->data[0] = sdata[1];
	    data->lerror = sdata[1];
	    do_receive = 0;
	  }
	  break;
	case XCP_PKG_SERV:
	  _xcp_verbose("Received Service paket\n");
	  //Process Service package
	  break;
	default:
	  _xcp_error("Received unknown package\n");
	  //All other are DT0 packages
	  //Process
	  if(_xcp_dto_daq_process(data,can_id,sdata,ldata,pkg->tv)){
	    _xcp_error("DAQ processing failed\n");
	    return CS_FAIL;
	  }
	  if(mode_odt)
	    do_receive = 0;
	  
	  break;
	}
      } else {
	_xcp_error("Data length zero or error registered\n");
      }
    } else {
      _xcp_verbose("No package received\n");
    }
  }
  
  if(!received_pkg)
    return CS_FAIL;

  if(data->lerror != XCP_ERROR_NOERROR){
    _xcp_error("Error in progress\n");
    return CS_FAIL;
  } else
    return CS_OK;
}


/** Internal routine, send recv, if an error is recieved do preaction 
    and action cmd
*/
/**************************************************************************/
int _xcp_send_recv_action(
			  cs_xcp_t *data,
			  cs_xcp_pkg_t *pkg_send,
			  cs_xcp_pkg_t *pkg_recv,
			  uint8_t cmd,
			  uint8_t subcmd,
			  uint32_t timeout,
			  uint16_t timeout_error
			  )
/**************************************************************************/
{ 
  cs_xcp_pa_t preaction = XCP_PA_NONE;
  cs_xcp_ac_t action = XCP_AC_NONE;
  int repeat = 1;
  int repeat_counter = 0;

  if((data == NULL) ||
     (pkg_send == NULL) ||
     (pkg_recv == NULL)){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
    
  _xcp_verbose("Timeout %i Timeout error %i\n", timeout, timeout_error);
  while(repeat_counter < repeat){
    _xcp_verbose("Send Recv Retry Cmd: 0x%X, Retry %i, Last Error %s\n",cmd,repeat_counter,cs_xcp_get_lerrorstr(data));
    _xcp_reset_error(data);

    if(_xcp_send(data,pkg_send)){
      _xcp_error("Send failure\n");
      return CS_FAIL;
    }
    if(_xcp_receive(data,pkg_recv,timeout,0)){
      _xcp_verbose("Receive negative response: %s\n",cs_xcp_get_lerrorstr(data));
      //error received
      if(data->lerror == XCP_ERROR_TIMEOUT){
	//Change to specific error
	data->lerror = timeout_error;
      }
      if(data->lerror & XCP_SEVERITY_S3){
	//Abort connection
	return CS_FAIL;
      }

      //Lookup failure preaction and action cmd
      preaction = _xcp_get_preaction(data,cmd,subcmd,data->lerror);
      action = _xcp_get_action(data,cmd,subcmd,data->lerror);
	
      if(_xcp_handler_preaction(data,cmd,subcmd,preaction))
	return CS_FAIL;

      if(_xcp_handler_action(data,cmd,subcmd,action,&repeat))
	return CS_FAIL;
	
	
    } else {
      //positive response
      _xcp_verbose("Positive response: Exp DL: %i, Got DL: %i\n",pkg_recv->exp_dl,pkg_recv->dl);
    }
    repeat_counter++;
  }

  if(data->lerror != XCP_ERROR_NOERROR)
    return CS_FAIL;
  else
    return CS_OK;
}


/** Signal start daq in local data structures
 */
/**************************************************************************/
void _xcp_daq_start( 
		    cs_xcp_t *data,
		    int daq,
		    uint8_t pid
		     )
/**************************************************************************/
{
  int i,j;
  struct timespec ts_now;
  double now;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return;
  }

  //Make note for started DAQ list
  data->daqlist.daq[daq].running = 1;
  //Disable selected flag
  data->daqlist.daq[daq].selected = 0;
  //Store the pid
  if(data->xcpa2l.daq_proc. odt_ident == XCP_A2L_ODT_IDENT_ABS){
    _xcp_verbose("Received first PID %i Event %i for DAQ %i\n",pid,data->daqlist.daq[daq].ev,daq);
    data->daqlist.daq[daq].first_pid = (uint16_t)pid;
  }

  //Calculate size of ODT to campare it with received data
  for(i = 0; i < data->daqlist.daq[daq].no_odt; i++){
    data->daqlist.daq[daq].odt[i].datasize = 0;
    for(j = 0; j < data->daqlist.daq[daq].odt[i].no_entry; j++){
      data->daqlist.daq[daq].odt[i].datasize += data->daqlist.daq[daq].odt[i].entry[j].me.size *  cs_a2l_get_ag_bytes(data->xcpa2l.commode.AG);
    }
    //Set last receive time to zero
    data->daqlist.daq[daq].odt[i].last = 0.0;
  }
  //Reset counter
  data->daqlist.daq[daq].counter = 0;
  //Set timestamp of next expected package
  cs_ts_gettime(&ts_now);
  now = cs_ts_timespec2d(ts_now);
  data->daqlist.daq[daq].ts_expect = now + data->daqlist.daq[daq].period;
}

/** Signal start selected daq in local data structures
 */
/**************************************************************************/
void _xcp_daq_start_selected( 
			     cs_xcp_t *data
			      )
/**************************************************************************/
{
  int i,j,k;
  struct timespec ts_now;
  double now;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return;
  }

  for(k = 0; k < data->daqlist.no_daq; k++){
    if(data->daqlist.daq[k].selected){
      //Make note for started DAQ list
      data->daqlist.daq[k].running = 1;
      //Calculate size of ODT to campare it with received data
      for(i = 0; i < data->daqlist.daq[k].no_odt; i++){
	data->daqlist.daq[k].odt[i].datasize = 0;
	for(j = 0; j < data->daqlist.daq[k].odt[i].no_entry; j++){
	  data->daqlist.daq[k].odt[i].datasize += data->daqlist.daq[k].odt[i].entry[j].me.size *  cs_a2l_get_ag_bytes(data->xcpa2l.commode.AG);
	}
	//Set last receive time to zero
	data->daqlist.daq[k].odt[i].last = 0.0;
      }
      //Disable selected flag
      data->daqlist.daq[k].selected = 0;
      //Reset counter
      data->daqlist.daq[k].counter = 0;
      //Set timestamp of next expected package
      cs_ts_gettime(&ts_now);
      now = cs_ts_timespec2d(ts_now);
      data->daqlist.daq[k].ts_expect = now + data->daqlist.daq[k].period;
    }
  }
}

/** Signal start selected daq in local data structures
 */
/**************************************************************************/
void _xcp_daq_stop_selected( 
			    cs_xcp_t *data
			     )
/**************************************************************************/
{
  int i,k;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return;
  }
    
  for(k = 0; k < data->daqlist.no_daq; k++){
    if(data->daqlist.daq[k].selected){
      //Make note for started DAQ list
      data->daqlist.daq[k].running = 0;
      for(i = 0; i < data->daqlist.daq[k].no_odt; i++){
	data->daqlist.daq[k].odt[i].datasize = 0;
      }
      //Disable selected flag
      data->daqlist.daq[k].selected = 0;
    }
  }
}

/** Signal stop daq in local data structures
 */
/**************************************************************************/
void _xcp_daq_stop( 
		   cs_xcp_t *data,
		   int daq
		    )
/**************************************************************************/
{
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return;
  }
  data->daqlist.daq[daq].running = 0;
  data->daqlist.daq[daq].selected = 0;
}

/** Signal select daq in local data structures
 */
/**************************************************************************/
void _xcp_daq_select( 
		     cs_xcp_t *data,
		     int daq
		      )
/**************************************************************************/
{
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return;
  }
  data->daqlist.daq[daq].selected = 1;
}

/** Is requested feature supported
    \retval 0 feature is supported, or status unknown
    \retval 1 feature is not supported by master
*/
/**************************************************************************/
int _xcp_daq_feature_supported( 
			       cs_xcp_t *data,
			       uint8_t cmd
				)
/**************************************************************************/
{

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return 0;
  }
  
  if(data->xcpa2l.commode.optfeature[cmd] == A2L_FEATURE_DISABLED)
    return 1;
  
  return 0;
}

/**************************************************************************/
/**************************************************************************/
/***** External API-Functions**********************************************/ 
/**************************************************************************/
/**************************************************************************/

/********************************************************************/
uint16_t cs_xcp_getversion()
/********************************************************************/
{
  return CS_XCP_VERSION_MAGIC;
}



/**************************************************************************/
/**************************************************************************/
/***** Error API-Functions*************************************************/ 
/**************************************************************************/
/**************************************************************************/

/**************************************************************************/
int cs_xcp_get_lerror(
		      cs_xcp_t *data
		      )
/**************************************************************************/
{
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL; //general error
  }
  
  return data->lerror;
}

/**************************************************************************/
char *cs_xcp_get_lerrorstr(
			   cs_xcp_t *data
			   )
/**************************************************************************/
{
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return "Invalid Data pointer";
  }
  
  switch(data->lerror){
  case XCP_ERROR_CMD_BUSY:
    return "XCP Protocol Error: Command was not executed";
  case XCP_ERROR_DAQ_ACTIVE:
    return "XCP Protocol Error: Command rejected because DAQ is running.";
  case XCP_ERROR_PGM_ACTIVE:
    return "XCP Protocol Error: Command rejected because PGM is running.";
  case XCP_ERROR_CMD_UNKNOWN:
    return "XCP Protocol Error: Unknown command or not implemented optional command";
  case XCP_ERROR_CMD_SYNTAX:
    return "XCP Protocol Error: Command syntax valid but command";
  case XCP_ERROR_OUT_OF_RANGE:
    return "XCP Protocol Error: Command syntax valid but command parameter(s) out of range";
  case XCP_ERROR_WRITE_PROTECTED:
    return "XCP Protocol Error: The memory location is write protected";
  case XCP_ERROR_ACCESS_DENIED:
    return "XCP Protocol Error: The memory location is not accessible";
  case XCP_ERROR_ACCESS_LOCKED:
    return "XCP Protocol Error: Access denied, Seed & Key is required";
  case XCP_ERROR_PAGE_NOT_VALID:
    return "XCP Protocol Error: Selected page mode not available";
  case XCP_ERROR_MODE_NOT_VALID:
    return "XCP Protocol Error: Selected page mode not available";
  case XCP_ERROR_SEGMENT_NOT_VALID:
    return "XCP Protocol Error: Selected segment not valid";
  case XCP_ERROR_SEQUENCE:
    return "XCP Protocol Error: Sequence error";
  case XCP_ERROR_DAQ_CONFIG:
    return "XCP Protocol Error: DAQ configuration not valid";
  case XCP_ERROR_MEMORY_OVERFLOW:
    return "XCP Protocol Error: Memory overflow error";
  case XCP_ERROR_GENERIC:
    return "XCP Protocol Error: Generic error.";
  case XCP_ERROR_VERIFY:
    return "XCP Protocol Error: The slave internal program verify routine detects an error.";
  case XCP_ERROR_NOERROR:
    return "XCP API: No Failure";
  case XCP_ERROR_ERROR:
    return "XCP API: General unspecific error";
  case XCP_ERROR_INTERFACE:
    return "XCP API: Interface error";
  case XCP_ERROR_API:
    return "XCP API: API call failure";
  case XCP_ERROR_MALLOC:
    return "XCP API: Malloc failure";
  case XCP_ERROR_SEND:
    return "XCP API: Send failure";
  case XCP_ERROR_RECV:
    return "XCP API: Receive failure";
  case XCP_ERROR_UNSUPPORTED:
    return "XCP API: Unsupported feature";
  case XCP_ERROR_BUFFER:
    return "XCP API: Buffer overflow";
  case XCP_ERROR_UNEXPECTED:
    return "XCP API: Receive unexpected data";
  case XCP_ERROR_TIMEOUT:
    return "XCP API: General Timeout failure";
  case XCP_ERROR_TIMEOUTT1:
    return "XCP API: Timeout T1 failure";
  case XCP_ERROR_TIMEOUTT6:
    return "XCP API: Timeout T6 failure";
  case XCP_ERROR_CTO:
    return "XCP API Error: Package request greater than max_cto";
  case XCP_ERROR_ALLOC:
    return "XCP API Error: DAQ alloc sequence failure";
  case XCP_ERROR_DAQ:
    return "XCP API Error: Invalid DAQ request";
  case XCP_ERROR_SLAVE_UNSUPPORTED:
    return "XCP API Error: Function is not supported by Slave";
  case XCP_ERROR_DAQ_RANGE:
    return "XCP API Error: Invalid DAQ index request";
  case XCP_ERROR_EV_RANGE:
    return "XCP API Error: Invalid Event channel index request";
  case XCP_ERROR_DAQ_PTR:
    return "XCP API Error: Invalid DAQ pointer request";
  case XCP_ERROR_ODT_RANGE:
    return "XCP API Error: Invalid ODT index request";
  case XCP_ERROR_ENTRY_RANGE:
    return "XCP API Error: Invalid ENTRY index request";
  case XCP_ERROR_DAQ_SEQUENCE:
    return "XCP API Error: Invalid DAQ data structure request";
  case XCP_ERROR_DAQ_PROCESS:
    return "XCP API Error: Failure in DAQ processing";
  case XCP_ERROR_ODT_SIZE:
    return "XCP API Error: Invalid ODT recv size";
  case XCP_ERROR_ODT_UNKNOWN:
    return "XCP API Error: Unkndown ODT received";
  case XCP_ERROR_BUFFER_RANGE:
    return "XCP API Error: Buffer index out of range";
  case XCP_ERROR_API_VERSION:
    return "XCP API Error: ASAP Version is not supported";
  case XCP_ERROR_API_PARSE:
    return "XCP API Error: A2L parsing failure";
  case XCP_ERROR_API_FILE:
    return "XCP API Error: Cannot open file";
  case XCP_ERROR_API_PARAM:
    return "XCP API Error: Invalid parameter argument";
  case XCP_ERROR_API_SIGNALRANGE:
    return "XCP API Error: Too many signals configured";
  default:
    return "Unknown error code";
  }
}

/**************************************************************************/
/**************************************************************************/
/***** Data handling API-Functions*****************************************/ 
/**************************************************************************/
/**************************************************************************/


/**************************************************************************/
int cs_xcp_data_set_timeout(
			    cs_xcp_t *data,
			    int *timeout,
			    int len
			    )
/**************************************************************************/
{
  int i;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }

  if((timeout == NULL) || 
     (len !=  XCP_A2L_TO_T7)){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;
  }

  for(i = 0; i < len; i++){
    data->xcpa2l.commode.timeout[i] = (uint32_t)timeout[i];
  }

  return CS_OK;
}


/**************************************************************************/
cs_xcp_resources_t cs_xcp_data_get_resources(
					     cs_xcp_t *data
					     )
/**************************************************************************/
{
  cs_xcp_resources_t ldata;

  memset(&ldata,0,sizeof(cs_xcp_resources_t));
    
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return ldata;
  }

  return data->resources;
}

/**************************************************************************/
cs_xcp_daq_t *cs_xcp_data_get_daq(
				  cs_xcp_t *data,
				  int daq
				  )
/**************************************************************************/
{
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return NULL;
  }

  if(daq > data->daqlist.no_daq){
    _xcp_error("Parameter failure\n");
    return NULL;
  }

  return &data->daqlist.daq[daq];
}

/**************************************************************************/
uint64_t cs_xcp_data_get_daq_counter(
				     cs_xcp_t *data,
				     int daq
				     )
/**************************************************************************/
{
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return 0;
  }
  
  if(daq > data->daqlist.no_daq){
    _xcp_error("Parameter failure\n");
    return 0;
  }	

  return data->daqlist.daq[daq].counter;
}

/**************************************************************************/
int cs_xcp_data_set_no_daq(
			   cs_xcp_t *data,
			   int no_daq
			   )
/**************************************************************************/
{
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  if(no_daq <= 0){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;
  }

  if(data->daqlist.daq){
    data->lerror = XCP_ERROR_DAQ_SEQUENCE;
    return CS_FAIL;
  }

  _xcp_verbose("Set no DAQ, %i\n",no_daq);
  //If no failure alloc daq structure
  _xcp_reset_daqlist(data);
  data->daqlist.no_daq = no_daq;
  data->daqlist.daq = (cs_xcp_daq_t*) malloc(sizeof(cs_xcp_daq_t)*data->daqlist.no_daq);
  
  if(!data->daqlist.daq){
    data->lerror = XCP_ERROR_MALLOC;
    data->daqlist.no_daq = 0;
    return CS_FAIL;
  }
  //Reset data structure
  memset(data->daqlist.daq,0,(sizeof(cs_xcp_daq_t)*data->daqlist.no_daq));
  return CS_OK;
}

/**************************************************************************/
int cs_xcp_data_set_no_odt(
			   cs_xcp_t *data,
			   int daq,
			   int no_odt
			   )
/**************************************************************************/
{
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }

  if(no_odt <= 0){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;
  }

  if(daq > data->daqlist.no_daq){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_DAQ_RANGE;
    return CS_FAIL;
  }

  if(data->daqlist.daq[daq].odt){
    data->lerror = XCP_ERROR_DAQ_SEQUENCE;
    return CS_FAIL;
  }

  data->daqlist.daq[daq].no_odt = no_odt;
  data->daqlist.daq[daq].odt = (cs_xcp_odt_t*) malloc(sizeof(cs_xcp_odt_t)*data->daqlist.daq[daq].no_odt);
  
  if(!data->daqlist.daq[daq].odt){
    data->lerror = XCP_ERROR_MALLOC;
    data->daqlist.daq[daq].no_odt = 0;
    return CS_FAIL;
  }
  //Reset data structure
  memset(data->daqlist.daq[daq].odt,0,(sizeof(cs_xcp_odt_t)*data->daqlist.daq[daq].no_odt));

  return CS_OK;
}


/**************************************************************************/
int cs_xcp_data_set_no_odt_entry(
				 cs_xcp_t *data,
				 int daq,
				 int odt,
				 int no_entry
				 )
/**************************************************************************/
{
  int i;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }

  if(no_entry <= 0){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;
  }

  if(daq > data->daqlist.no_daq){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_DAQ_RANGE;
    return CS_FAIL;
  }

  if(odt > data->daqlist.daq[daq].no_odt){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_ODT_RANGE;
    return CS_FAIL;
  }

  if(data->daqlist.daq[daq].odt[odt].entry){
    data->lerror = XCP_ERROR_DAQ_SEQUENCE;
    return CS_FAIL;
  }

  data->daqlist.daq[daq].odt[odt].no_entry = no_entry;
  data->daqlist.daq[daq].odt[odt].entry = (cs_xcp_odt_entry_t*) malloc(sizeof(cs_xcp_odt_entry_t)*data->daqlist.daq[daq].odt[odt].no_entry);
  
  if(!data->daqlist.daq[daq].odt[odt].entry){
    data->lerror = XCP_ERROR_MALLOC;
    data->daqlist.daq[daq].odt[odt].no_entry = 0;
    return CS_FAIL;
  }

  for(i = 0; i < data->daqlist.daq[daq].odt[odt].no_entry; i++){
    _xcp_reset_odt_entry(&data->daqlist.daq[daq].odt[odt].entry[i]);
  }

  return CS_OK;
}

/**************************************************************************/
/**************************************************************************/
/***** DAQ Data accessing API-Functions************************************/ 
/**************************************************************************/
/**************************************************************************/

/**************************************************************************/
int cs_xcp_entry_value_raw(
			   cs_xcp_t *data,
			   int daq,
			   int odt,
			   int entry,
			   uint8_t *udata,
			   int *size
			   )
/**************************************************************************/
{
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  if(udata == NULL){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;
  }

  if(size == NULL){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;
  }

  if((daq > data->daqlist.no_daq) || (daq < 0)){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_DAQ_RANGE;
    return CS_FAIL;
  }

  if((odt > data->daqlist.daq[daq].no_odt) || (odt < 0)){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_ODT_RANGE;
    return CS_FAIL;
  }

  if((entry > data->daqlist.daq[daq].odt[odt].no_entry) || (entry < 0)){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_ENTRY_RANGE;
    return CS_FAIL;
  }

  if(data->daqlist.daq[daq].odt[odt].entry == NULL){
    data->lerror = XCP_ERROR_DAQ_SEQUENCE;
    return CS_FAIL;
  }
  
  if(*size < (data->daqlist.daq[daq].odt[odt].entry[entry].me.size * cs_a2l_get_ag_bytes(data->xcpa2l.commode.AG))){
    data->lerror = XCP_ERROR_BUFFER_RANGE;
    return CS_FAIL;
  }

  //Copy data
  *size = data->daqlist.daq[daq].odt[odt].entry[entry].me.size * cs_a2l_get_ag_bytes(data->xcpa2l.commode.AG);
  memcpy(udata,&data->daqlist.daq[daq].odt[odt].entry[entry].data,*size);

  return CS_OK;
}

/**************************************************************************/
int cs_xcp_convert_to_double(			
			     cs_xcp_t *data,
			     uint8_t *udata,
			     int size,
			     cs_a2l_compu_method_t cm,
			     cs_a2l_data_type_t dtype,
			     double *value
						)
/**************************************************************************/
{
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }	
  
  if(udata == NULL){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;
  }

  if(value == NULL){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;
  }

  if(size == 0){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;
  }

  if(size > 8){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;
  }

  return cs_a2l_convert_to_double(data->xcpa2l.commode.BO,udata,size,cm,dtype,value);
}

/**************************************************************************/
int cs_xcp_entry_reg_value(			
			   cs_xcp_t *data,
			   int daq,
			   int odt,
			   int entry,
			   double *value,
			   cs_a2l_compu_method_t cm,
			   cs_a2l_data_type_t dtype,
			   double *ts,
			   void (*extcallback)(double value, double ts, void *priv),
			   void *priv
						)
/**************************************************************************/
{
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  if((daq > data->daqlist.no_daq) || (daq < 0)){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_DAQ_RANGE;
    return CS_FAIL;
  }

  if((odt > data->daqlist.daq[daq].no_odt) || (odt < 0)){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_ODT_RANGE;
    return CS_FAIL;
  }

  if((entry > data->daqlist.daq[daq].odt[odt].no_entry) || (entry < 0)){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_ENTRY_RANGE;
    return CS_FAIL;
  }

  if(data->daqlist.daq[daq].odt[odt].entry == NULL){
    data->lerror = XCP_ERROR_DAQ_SEQUENCE;
    return CS_FAIL;
  }
  
  data->daqlist.daq[daq].odt[odt].entry[entry].me.exvalue = value;
  data->daqlist.daq[daq].odt[odt].entry[entry].me.cm = cm;
  data->daqlist.daq[daq].odt[odt].entry[entry].me.dtype = dtype;
  data->daqlist.daq[daq].odt[odt].entry[entry].me.exts = ts;
  data->daqlist.daq[daq].odt[odt].entry[entry].me.extcallback = extcallback;
  data->daqlist.daq[daq].odt[odt].entry[entry].me.extcallbackpriv = priv;

  return CS_OK;
}

/**************************************************************************/
double cs_xcp_entry_value_double(			
				 cs_xcp_t *data,
				 int daq,
				 int odt,
				 int entry,
				 cs_a2l_compu_method_t cm,
				 cs_a2l_data_type_t dtype,
				 int *flag
							)
/**************************************************************************/
{
  double rvalue = 0.0;
  uint8_t buffer[8];
  int len = 8;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }

  if(flag == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
    
  *flag = cs_xcp_entry_value_raw(data,daq,odt,entry,buffer,&len);

  if(*flag)
    return rvalue;

  *flag = cs_xcp_convert_to_double(data,buffer,len,cm,dtype,&rvalue);
  
  if(*flag)
    return rvalue = 0.0;
    
  return rvalue;
}

/**************************************************************************/
int cs_xcp_daq_istimeout(			
			 cs_xcp_t *data,
			 double fac_timeout
						)
/**************************************************************************/
{
  int i, j;
  int rv = CS_OK;
  double now;
  struct timespec ts_now;
    
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
  //Get actual time
  cs_ts_gettime(&ts_now);
  now = cs_ts_timespec2d(ts_now);
  
  //Loop over all DAQ, to check ODTs for receive timeout
  for(i = 0; i < data->daqlist.no_daq; i++){
    //Only DAQ direction
    if(data->daqlist.daq[i].dir !=XCP_DAQ_DIR_DAQ)
      continue;
    //Is DAQ running
    if(!data->daqlist.daq[i].running)
      continue;

    for(j = 0; j < data->daqlist.daq[i].no_odt; j++){
      //Do we have already received an ODT
      if(data->daqlist.daq[i].odt[j].last <= 0.0)
	continue;

      if((data->daqlist.daq[i].odt[j].last + 
	  (data->daqlist.daq[i].period * fac_timeout)) < data->daqlist.daq[i].ts_expect){
	//Check if timeout check is deactivated
	if(fac_timeout > 1.0){
	  //Detect timeout
	  _xcp_error("DAQ timeout detected, DAQ %i, ODT %i, Period %f s Last recv %f s, Expected %f s\n",
		     i,i, data->daqlist.daq[i].period,
		     data->daqlist.daq[i].odt[j].last,
		     data->daqlist.daq[i].ts_expect);
	  rv = CS_FAIL;
	}
      }
    }
    //Now calc new expected recv timestamps,
    //Notice we do not know when this function is called, so we have to loop
    if(data->daqlist.daq[i].period > 0.0){
      while(data->daqlist.daq[i].ts_expect <= now){
	data->daqlist.daq[i].ts_expect += data->daqlist.daq[i].period;
      }
    } else {
      //if period is zero, set expected time to actual time
      data->daqlist.daq[i].ts_expect = now;
    }
        
  }
    
  return rv;
}


/**************************************************************************/
/**************************************************************************/
/***** Debugging API-Functions*********************************************/ 
/**************************************************************************/
/**************************************************************************/

/**************************************************************************/
void cs_xcp_print_resources(
			    cs_xcp_resources_t *resources
			    )
/**************************************************************************/
{
  if(resources == NULL){
    _xcp_error("Parameter failure\n");
    return;
  }

  _xcp_info("Resources: CALPAG %s\n",resources->CALPAG?"available":"not available");
  _xcp_info("Resources: DAQ %s\n",resources->DAQ?"available":"not available");
  _xcp_info("Resources: STIM %s\n",resources->STIM?"available":"not available");
  _xcp_info("Resources: PGM %s\n",resources->PGM?"available":"not available");
}

/**************************************************************************/
void cs_xcp_print_status(
			 cs_xcp_status_t *status
			 )
/**************************************************************************/
{
  if(status == NULL){
    _xcp_error("Parameter failure\n");
    return;
  }
    
  _xcp_info("Status: Store CalPag %s\n",status->store_cal_req?"On":"Off");
  _xcp_info("Status: Store DAQ %s\n",status->store_daq_req?"On":"Off");
  _xcp_info("Status: Clear DAQ %s\n",status->clear_daq_req?"On":"Off");
  _xcp_info("Status: DAQ Running %s\n",status->daq_running?"On":"Off");
  _xcp_info("Status: Resume mode %s\n",status->resume?"On":"Off");
  _xcp_info("Status: Protection CalPag %s\n",status->prot_calpag?"On":"Off");
  _xcp_info("Status: Protection DAQ %s\n",status->prot_daq?"On":"Off");
  _xcp_info("Status: Protection STIM %s\n",status->prot_stim?"On":"Off");
  _xcp_info("Status: Protection PGM %s\n",status->prot_pgm?"On":"Off");
  _xcp_info("Status: Reserved %X\n",status->reserved);
  _xcp_info("Status: Session-ID %X\n",status->session_id);

}



/**************************************************************************/
/**************************************************************************/
/***** Interface API-Functions*********************************************/ 
/**************************************************************************/
/**************************************************************************/

/**************************************************************************/
int cs_xcp_iface_can_configure(
			       cs_xcp_t *data,          
			       unsigned int cmd_id,
			       unsigned int resp_id,
			       unsigned int bcast_id
			       )
/**************************************************************************/
{
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  data->iftype = XCP_CAN;
  data->ifmax = CS_XCP_CAN_DLC;
  
  //Configure interface
  cs_xcp_can_configure(data->ifcan,cmd_id,resp_id,bcast_id);
  
  return CS_OK;
}

/**************************************************************************/
int cs_xcp_iface_can_setup(
			   cs_xcp_t *data,          
			   cs_xcp_can_t *iface
			   )
/**************************************************************************/
{
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  data->iftype = XCP_CAN;
  data->ifmax = CS_XCP_CAN_DLC;

  data->ifcan = iface;
  
  return CS_OK;
}

/**************************************************************************/
int cs_xcp_iface_can_open(
			  cs_xcp_t *data,          
			  char *iface
			  )
/**************************************************************************/
{
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  if(iface == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  if(cs_xcp_can_open(data->ifcan, iface)){
    data->lerror = XCP_ERROR_INTERFACE;
    return CS_FAIL;
  }
  return CS_OK;
}

/**************************************************************************/
int cs_xcp_init(
		cs_xcp_t *data
		)
/**************************************************************************/
{
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  //Preinit DAQ lists
  data->daqlist.daq = NULL;
  data->daqlist.no_daq = 0;
  data->work_entry = NULL;
  data->isConnected = 0;
  data->stim = 0;
  data->daq_fac_recvtimeout = 2.0;
  
  //First reset structure
  cs_xcp_a2l_init(&data->xcpa2l);
  //Build up error table
  _xcp_init_error_handling(data);
  //Reset daq processor
  cs_xcp_a2l_init_daqproc(&data->xcpa2l.daq_proc);
  //Reset Error
  _xcp_reset_error(data);
  //Reset resources
  _xcp_reset_resources(data);
  //reset commode
  cs_xcp_a2l_data_reset_commode(&data->xcpa2l.commode);
  //Reset daq lists
  _xcp_reset_daqlist(data);

  //Set interface to undef
  data->iftype = XCP_IFUNDEF;
  //Set default interface transfer length
  data->ifmax = 0;
  //Init CAN iface structure
  data->ifcan = cs_xcp_can_init();

  if(data->ifcan == NULL){
    return CS_FAIL;
  }

  _xcp_info("XCP Init API version %i\n",CS_XCP_VERSION_MAGIC);

  return CS_OK;
}

/**************************************************************************/
int cs_xcp_close(
		 cs_xcp_t *data
		 )
/**************************************************************************/
{
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }

  switch(data->iftype){
  case XCP_CAN:
    data->iftype = XCP_IFUNDEF;
    cs_xcp_can_close(data->ifcan);
    return CS_OK;
    break;
  case XCP_USB:
    //Not supported
    return CS_FAIL;
    break;
  case XCP_ETHERNET:
    //Not supported
    return CS_FAIL;
    break;
  default:
    return CS_FAIL;
  }
  return CS_FAIL;
}

/**************************************************************************/
int cs_xcp_receive(
		   cs_xcp_t *data
		   )
/**************************************************************************/
{
  cs_xcp_pkg_t pkg_recv;
  int counter = 0;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  _xcp_reset_pkg(&pkg_recv);
  //Note only dummy receive pkg, DTO handling in low level routine
  while(!_xcp_receive(data,&pkg_recv,0,1)){
    //process data
    //_xcp_verbose("Looping: %i\n",counter);
    counter++;
  }
  if(data->lerror != XCP_ERROR_NOERROR){
    _xcp_error("Receive Failure: %i, DL:%i\n",counter,pkg_recv.dl);
    return CS_FAIL;
  }else{
    return CS_OK;
  }
}


/**************************************************************************/
/**************************************************************************/
/***** Protocol API-Functions**********************************************/ 
/***** Transport layer specific********************************************/ 
/**************************************************************************/
/**************************************************************************/

/**************************************************************************/
int cs_xcp_cmd_get_slave_id(
			    cs_xcp_t *data,
			    int mode,
			    uint32_t **respid,
			    int *len
			    )
/**************************************************************************/
{
  uint8_t *sdata = NULL;
  int ldata = 0;
  int rlen = 255;
  int recv_counter = 0;
  cs_xcp_pkg_t recv_pkg;
  cs_cu_data32_t l32;
    
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }

  if(data->iftype != XCP_CAN){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;
  }

  if((mode > 1) || (mode < 0)){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;
  }

  if((respid == NULL) || (len == NULL)){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;
  }

  _xcp_debug("XCP API Call Get Slave ID, Mode %s\n",mode?"confirm by inverse echo":"identify by echo");


  //Malloc return value array
  *respid = (uint32_t *)malloc(sizeof(uint32_t)*rlen);
  if(*respid == NULL){
    data->lerror = XCP_ERROR_MALLOC;
    return CS_FAIL;
  }
  memset(*respid,0,sizeof(uint32_t)*rlen);
  *len = 0;

  //Build up raw package

  sdata = (uint8_t*)malloc(sizeof(uint8_t)*data->ifmax);
  if(sdata == NULL){
    data->lerror = XCP_ERROR_MALLOC;
    return CS_FAIL;
  }
  memset(sdata,0,sizeof(uint8_t)*data->ifmax);
  ldata = 6;

  _xcp_reset_pkg(&recv_pkg);

  sdata[0] = XCP_CMD_TRANSPORT_LAYER_CMD;
  sdata[1] = XCP_CMD_TRANSPORT_CAN_GET_SLAVE_ID;
  sdata[2] = 0x58; //X
  sdata[3] = 0x43; //C
  sdata[4] = 0x50; //P
  sdata[5] = (uint8_t)mode;
  
  if(cs_xcp_can_send(data->ifcan,cs_xcp_can_getBcastId(data->ifcan),sdata,ldata)){
    data->lerror = XCP_ERROR_SEND;
    if(sdata)
      free(sdata);
    return CS_FAIL;
  }
  while(!_xcp_receive(data,&recv_pkg,data->xcpa2l.commode.timeout[XCP_A2L_TO_T1], 0)){
    if(recv_pkg.PID != XCP_PKG_RESP){
      data->lerror = XCP_ERROR_UNEXPECTED;
      return CS_FAIL;
    }
    if(recv_pkg.dl != 7){
      data->lerror = XCP_ERROR_UNEXPECTED;
      return CS_FAIL;
    }
    if((recv_pkg.data[0] != 0x58) ||
       (recv_pkg.data[1] != 0x43) ||
       (recv_pkg.data[2] != 0x50)){
      data->lerror = XCP_ERROR_UNEXPECTED;
      return CS_FAIL;
    }
    //Get id
    l32.u8[0] = recv_pkg.data[3];
    l32.u8[1] = recv_pkg.data[4];
    l32.u8[2] = recv_pkg.data[5];
    l32.u8[3] = recv_pkg.data[6];
    (*respid)[recv_counter] = cs_cu_swap32(data->xcpa2l.commode.BO,l32.u32);
    _xcp_debug("Got Response, Mode %s, Index %i, ID %X\n",mode?"by echo":"inverse echo",recv_counter,(*respid)[recv_counter]);
    recv_counter++;
    _xcp_reset_pkg(&recv_pkg);
  }

  if(sdata)
    free(sdata);
  
  return CS_OK;
}

/**************************************************************************/
int cs_xcp_cmd_get_daq_id(
			  cs_xcp_t *data,     
			  int daq
			  )
/**************************************************************************/
{
  cs_xcp_pkg_t pkg_send;
  cs_xcp_pkg_t pkg_recv;
  uint32_t timeout;
  uint16_t timeout_error;
  cs_cu_data16_t l16;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }

  if((daq > data->daqlist.no_daq) || (daq < 0)){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_DAQ;
    return CS_FAIL;
  }

  _xcp_verbose("XCP API Call Get DAQ ID\n");
  
  _xcp_reset_pkg(&pkg_send);
  _xcp_reset_pkg(&pkg_recv);

  pkg_send.pkgtype =  XCP_PKG_CMD;
  pkg_send.PID = XCP_CMD_TRANSPORT_LAYER_CMD;
  l16.u16 = cs_cu_swap16(data->xcpa2l.commode.BO,(uint16_t)daq);
  pkg_send.data[0] = XCP_CMD_TRANSPORT_CAN_GET_DAQ_ID;
  pkg_send.data[1] = l16.u8[0];
  pkg_send.data[2] = l16.u8[1];
  pkg_send.dl = 3;

  timeout = data->xcpa2l.commode.timeout[XCP_A2L_TO_T1];
  timeout_error = XCP_ERROR_TIMEOUTT1;

  //Expect 7 bytes
  pkg_recv.exp_dl = 7;

  //Send Recv and retry for special failures
  if(_xcp_send_recv_action(data,&pkg_send,&pkg_recv,pkg_send.PID,pkg_send.data[0],timeout,timeout_error)){
    return CS_FAIL;
  }

  //Handle data
  //Check data length
  if(pkg_recv.dl < 7){
    data->lerror = XCP_ERROR_UNEXPECTED;
    return CS_FAIL;
  }

  return CS_OK;
}

/**************************************************************************/
/**************************************************************************/
/***** Protocol API-Functions**********************************************/ 
/**************************************************************************/
/**************************************************************************/

/**************************************************************************/
int cs_xcp_cmd_connect(
		       cs_xcp_t *data,     
		       int mode
		       )
/**************************************************************************/
{
  cs_xcp_pkg_t pkg_send;
  cs_xcp_pkg_t pkg_recv;
  uint32_t timeout;
  uint16_t timeout_error;
  cs_cu_data16_t lvalue;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }

  if((mode > 1) || (mode < 0)){
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;
  }

  _xcp_verbose("XCP API Call connect, Mode %s\n",mode?"User defined":"Standard");
  
  _xcp_reset_pkg(&pkg_send);
  _xcp_reset_pkg(&pkg_recv);

  pkg_send.pkgtype =  XCP_PKG_CMD;
  pkg_send.PID = XCP_CMD_CONNECT;
  pkg_send.data[0] = (uint8_t)mode;
  pkg_send.dl = 1;

  if(mode){
    timeout = data->xcpa2l.commode.timeout[XCP_A2L_TO_T1];
    timeout_error = XCP_ERROR_TIMEOUTT1;
  }else{
    timeout = data->xcpa2l.commode.timeout[XCP_A2L_TO_T6];
    timeout_error = XCP_ERROR_TIMEOUTT6;
  }

  //Expect 7 bytes
  pkg_recv.exp_dl = 7;

  //Send Recv and retry for special failures
  if(_xcp_send_recv_action(data,&pkg_send,&pkg_recv,pkg_send.PID,pkg_send.data[0],timeout,timeout_error)){
    return CS_FAIL;
  }

  //Handle data
  //Check data length
  if(pkg_recv.dl < 7){
    data->lerror = XCP_ERROR_UNEXPECTED;
    return CS_FAIL;
  }

  //Get resource information
  data->resources.CALPAG = (pkg_recv.data[0] & XCP_BM_CALPAG)?1:0;
  data->resources.DAQ = (pkg_recv.data[0] & XCP_BM_DAQ)?1:0;
  data->resources.STIM = (pkg_recv.data[0] & XCP_BM_STIM)?1:0;
  data->resources.PGM = (pkg_recv.data[0] & XCP_BM_PGM)?1:0;

  //Get communication information
  data->xcpa2l.commode.BO = (pkg_recv.data[1] & XCP_BM_BYTE_ORDER)?A2L_BO_MOTOROLLA:A2L_BO_INTEL;

  data->xcpa2l.commode.AG = ((pkg_recv.data[1] & XCP_BM_ADDRESS_GRANULARITY_HIGH)<<2) | ((pkg_recv.data[1] & XCP_BM_ADDRESS_GRANULARITY_LOW)<<1);
  data->xcpa2l.commode.SBM = (pkg_recv.data[1] & XCP_BM_SLAVE_BLOCK_MODE)?1:0;
  data->xcpa2l.commode.optional = (pkg_recv.data[1] & XCP_BM_OPTIONAL)?1:0;


  //Get Max values for CTO and DTO
  data->xcpa2l.commode.max_CTO = (int)pkg_recv.data[2];
  lvalue.u8[0] = pkg_recv.data[3];
  lvalue.u8[1] = pkg_recv.data[4];
  data->xcpa2l.commode.max_DTO = cs_cu_swap16(data->xcpa2l.commode.BO,lvalue.u16);

  //Get version information
  data->xcpa2l.commode.prot_version = (int)pkg_recv.data[5];
  data->xcpa2l.commode.trans_version = (int)pkg_recv.data[6];

  return CS_OK;
}


/**************************************************************************/
int cs_xcp_cmd_disconnect(
			  cs_xcp_t *data
			  )
/**************************************************************************/
{
  cs_xcp_pkg_t pkg_send;
  cs_xcp_pkg_t pkg_recv;
  uint32_t timeout;
  uint16_t timeout_error;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }

  _xcp_verbose("XCP API Call disconnect\n");
  
  _xcp_reset_error(data);
  _xcp_reset_pkg(&pkg_send);
  _xcp_reset_pkg(&pkg_recv);

  pkg_send.pkgtype =  XCP_PKG_CMD;
  pkg_send.PID = XCP_CMD_DISCONNECT;
  pkg_send.dl = 0;

  timeout = data->xcpa2l.commode.timeout[XCP_A2L_TO_T1];
  timeout_error = XCP_ERROR_TIMEOUTT1;

  //Expect 0 data bytes
  pkg_recv.exp_dl = 0;

  //Send Recv and retry for special failures
  if(_xcp_send_recv_action(data,&pkg_send,&pkg_recv,pkg_send.PID,0,timeout,timeout_error)){
    return CS_FAIL;
  }

  return CS_OK;
}

/**************************************************************************/
int cs_xcp_cmd_get_status(
			  cs_xcp_t *data,
			  cs_xcp_status_t *status
			  )
/**************************************************************************/
{
  cs_xcp_pkg_t pkg_send;
  cs_xcp_pkg_t pkg_recv;
  uint32_t timeout;
  uint16_t timeout_error;
  cs_cu_data16_t lvalue;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }

  if(status == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  _xcp_verbose("XCP API Call Get status\n");

  _xcp_reset_error(data);
  _xcp_reset_pkg(&pkg_send);
  _xcp_reset_pkg(&pkg_recv);

  memset(status,0,sizeof(cs_xcp_status_t));

  pkg_send.pkgtype =  XCP_PKG_CMD;
  pkg_send.PID = XCP_CMD_GET_STATUS;
  pkg_send.dl = 0;

  timeout = data->xcpa2l.commode.timeout[XCP_A2L_TO_T1];
  timeout_error = XCP_ERROR_TIMEOUTT1;

  //Expect 5 data bytes
  pkg_recv.exp_dl = 5;

  //Send Recv and retry for special failures
  if(_xcp_send_recv_action(data,&pkg_send,&pkg_recv,pkg_send.PID,0,timeout,timeout_error)){
    return CS_FAIL;
  }

  //Handle data
  //Check data length
  if(pkg_recv.dl < 5){
    data->lerror = XCP_ERROR_UNEXPECTED;
    return CS_FAIL;
  }


  //Decode data
  status->store_cal_req = (pkg_recv.data[0] & XCP_BM_STATUS_STORE_CALPAG)?1:0; 
  status->store_daq_req = (pkg_recv.data[0] & XCP_BM_STATUS_STORE_DAQ)?1:0;
  status->clear_daq_req = (pkg_recv.data[0] & XCP_BM_STATUS_CLEAR_DAQ)?1:0;
  status->daq_running = (pkg_recv.data[0] & XCP_BM_STATUS_DAQ_RUNNING)?1:0;
  status->resume = (pkg_recv.data[0] & XCP_BM_STATUS_RESUME)?1:0;
  status->prot_calpag = (pkg_recv.data[1] & XCP_BM_PROT_CALPAG)?1:0;
  status->prot_daq = (pkg_recv.data[1] & XCP_BM_PROT_DAQ)?1:0;
  status->prot_stim = (pkg_recv.data[1] & XCP_BM_PROT_STIM)?1:0;
  status->prot_pgm = (pkg_recv.data[1] & XCP_BM_PROT_PGM)?1:0;
  status->reserved = pkg_recv.data[2];
  lvalue.u8[0] = pkg_recv.data[3];
  lvalue.u8[1] = pkg_recv.data[4];
  status->session_id = cs_cu_swap16(data->xcpa2l.commode.BO,lvalue.u16);

  return CS_OK;
}

/**************************************************************************/
int cs_xcp_cmd_synch(
		     cs_xcp_t *data
		     )
/**************************************************************************/
{
  cs_xcp_pkg_t pkg_send;
  cs_xcp_pkg_t pkg_recv;
  uint32_t timeout;
  uint16_t timeout_error;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  _xcp_verbose("XCP API Call Synch\n");
  _xcp_reset_error(data);
  _xcp_reset_pkg(&pkg_send);
  _xcp_reset_pkg(&pkg_recv);

  pkg_send.pkgtype =  XCP_PKG_CMD;
  pkg_send.PID = XCP_CMD_SYNCH;
  pkg_send.dl = 0;

  timeout = data->xcpa2l.commode.timeout[XCP_A2L_TO_T1];
  timeout_error = XCP_ERROR_TIMEOUTT1;

  //Expect 0 data bytes
  pkg_recv.exp_dl = 0;

  //Send Recv and retry for special failures
  if(_xcp_send_recv_action(data,&pkg_send,&pkg_recv,pkg_send.PID,0,timeout,timeout_error)){
    return CS_FAIL;
  }

  return CS_OK;
}

/**************************************************************************/
int cs_xcp_cmd_get_commode_info(
				cs_xcp_t *data
				)
/**************************************************************************/
{
  cs_xcp_pkg_t pkg_send;
  cs_xcp_pkg_t pkg_recv;
  uint32_t timeout;
  uint16_t timeout_error;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  _xcp_verbose("XCP API Call Get Comm mode info\n");

  if(_xcp_daq_feature_supported(data,XCP_CMD_GET_COMM_MODE_INFO)){
    data->lerror = XCP_ERROR_SLAVE_UNSUPPORTED;
    return CS_FAIL;
  }

  _xcp_reset_error(data);
  _xcp_reset_pkg(&pkg_send);
  _xcp_reset_pkg(&pkg_recv);

  pkg_send.pkgtype =  XCP_PKG_CMD;
  pkg_send.PID = XCP_CMD_GET_COMM_MODE_INFO;
  pkg_send.dl = 0;

  timeout = data->xcpa2l.commode.timeout[XCP_A2L_TO_T1];
  timeout_error = XCP_ERROR_TIMEOUTT1;

  //Expect 7 data bytes
  pkg_recv.exp_dl = 7;

  //Send Recv and retry for special failures
  if(_xcp_send_recv_action(data,&pkg_send,&pkg_recv,pkg_send.PID,0,timeout,timeout_error)){
    return CS_FAIL;
  }

  //Handle data
  //Check data length
  if(pkg_recv.dl < 7){
    data->lerror = XCP_ERROR_UNEXPECTED;
    return CS_FAIL;
  }

  //Decode data
  data->xcpa2l.commode.IBM = (pkg_recv.data[1] & XCP_BM_COMMODE_IBM)?1:0;
  data->xcpa2l.commode.MBM = (pkg_recv.data[1] & XCP_BM_COMMODE_MBM)?1:0;
  data->xcpa2l.commode.max_bs = pkg_recv.data[3];
  data->xcpa2l.commode.min_st = pkg_recv.data[4];
  data->xcpa2l.commode.queque = pkg_recv.data[5];
  data->xcpa2l.commode.driver_version = pkg_recv.data[6];
 
  return CS_OK;
}


/**************************************************************************/
int cs_xcp_cmd_set_mta(
		       cs_xcp_t *data,
		       uint32_t address,
		       uint8_t extension
		       )
/**************************************************************************/
{
  cs_xcp_pkg_t pkg_send;
  cs_xcp_pkg_t pkg_recv;
  uint32_t timeout;
  uint16_t timeout_error;
  cs_cu_data32_t ldata;
  
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
    
  _xcp_verbose("XCP API Call Set MTA\n");

  if(_xcp_daq_feature_supported(data,XCP_CMD_SET_MTA)){
    data->lerror = XCP_ERROR_SLAVE_UNSUPPORTED;
    return CS_FAIL;
  }

  _xcp_reset_error(data);
  _xcp_reset_pkg(&pkg_send);
  _xcp_reset_pkg(&pkg_recv);

  pkg_send.pkgtype =  XCP_PKG_CMD;
  pkg_send.PID = XCP_CMD_SET_MTA;
  pkg_send.dl = 7;
  pkg_send.data[2] = extension;
  ldata.u32 = cs_cu_swap32(data->xcpa2l.commode.BO,address);
  pkg_send.data[3] = ldata.u8[0];
  pkg_send.data[4] = ldata.u8[1];
  pkg_send.data[5] = ldata.u8[2];
  pkg_send.data[6] = ldata.u8[3];

  //Store MTA for preaction
  data->mta.address = address;
  data->mta.extension = extension;

  timeout = data->xcpa2l.commode.timeout[XCP_A2L_TO_T1];
  timeout_error = XCP_ERROR_TIMEOUTT1;

  //Expect 0 data bytes
  pkg_recv.exp_dl = 0;

  //Send Recv and retry for special failures
  if(_xcp_send_recv_action(data,&pkg_send,&pkg_recv,pkg_send.PID,0,timeout,timeout_error)){
    return CS_FAIL;
  }

  //No data to handle

  return CS_OK;
}

/**************************************************************************/
int cs_xcp_cmd_upload(
		      cs_xcp_t *data,
		      uint32_t *recv,
		      int *len
		      )
/**************************************************************************/
{
  cs_xcp_pkg_t pkg_send;
  cs_xcp_pkg_t pkg_recv;
  uint32_t timeout;
  //uint16_t timeout_error;
  int store_len;
  int data_start = 0;
  int data_length = 0;
  int ag_length = 0;
  int ag_size = 1;
  int ag_max = 0;
  int i,j;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }

  if(recv == NULL){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;
  }

  if(len == NULL){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;
  }

  //Limit Data element length
  if((*len <= 0) || (*len > 255)){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;
  }

  //Check if slave does not support block mode,
  //if no check if length fits in one package
  if(!data->xcpa2l.commode.SBM){
    if(data->xcpa2l.commode.AG != A2L_AG_BYTE){
      if((*len * data->xcpa2l.commode.AG) > (data->xcpa2l.commode.max_CTO -1)){
	data->lerror = XCP_ERROR_CTO;
	return CS_FAIL;
      }
    } else {
      if(*len > (data->xcpa2l.commode.max_CTO -1)){
	data->lerror = XCP_ERROR_CTO;
	return CS_FAIL;
      }
    }
  }

  _xcp_verbose("XCP API Call Upload\n");

  if(_xcp_daq_feature_supported(data,XCP_CMD_UPLOAD)){
    data->lerror = XCP_ERROR_SLAVE_UNSUPPORTED;
    return CS_FAIL;
  }

  _xcp_reset_error(data);
  _xcp_reset_pkg(&pkg_send);


  pkg_send.pkgtype =  XCP_PKG_CMD;
  pkg_send.PID = XCP_CMD_UPLOAD;
  pkg_send.dl = 1;
  pkg_send.data[0] = (uint8_t)*len;

  timeout = data->xcpa2l.commode.timeout[XCP_A2L_TO_T1];
  //timeout_error = XCP_ERROR_TIMEOUTT1;

  //Send request
  if(_xcp_send(data,&pkg_send))
    return CS_FAIL;

  //store len
  store_len = *len;
  //reset recv buffer
  memset(recv,0,*len*sizeof(uint32_t));
  //Reset len
  *len = 0;

  while(*len < store_len){
    _xcp_reset_pkg(&pkg_recv);
    if(_xcp_receive(data,&pkg_recv,timeout,0)){
      _xcp_error("Receive negative response: %s\n",cs_xcp_get_lerrorstr(data));
      return CS_FAIL;
    }
    //Copy data segment to output buffer, keep in mind AG
    if(data->xcpa2l.commode.AG != A2L_AG_BYTE){
      ag_length = data->xcpa2l.commode.AG -1;
      ag_size = data->xcpa2l.commode.AG;
      data_start = data->xcpa2l.commode.AG -1;
      //Calculate maximal data length, -1 because PID is not part of data
      data_length = data->xcpa2l.commode.max_CTO - 1 - ag_length;
      //Maximal number of data contend with respect to AG
      ag_max = data_length / ag_size;
    } else { 
      ag_length = 0;
      data_start = 0;
      ag_size = 1;
      data_length = data->xcpa2l.commode.max_CTO - 1;
      ag_max = data_length;
    }

    //Check received length, if it is smaller than fix contend
    if(pkg_recv.dl < ag_length){
      data->lerror = XCP_ERROR_UNEXPECTED;
      return CS_FAIL;
    }
    
    //Check if received data is smaller than max size, normally this will be the last block
    if((pkg_recv.dl - ag_length) < data_length)
      data_length = pkg_recv.dl - ag_length;

    //copy contend to buffer
    for(i = 0; i < ag_max; i++){
      if(i >= store_len)
	break;
      for(j = 0; j < ag_size; j++){
	//From uint8 to spezific AG
	recv[i + *len] |= ((uint32_t)pkg_recv.data[data_start + i* ag_size +j]) << ((ag_size -1) - j );
	_xcp_verbose("Add data Pos: %i, Data %X, packet pos: %i, data: %X\n",
		     (i+ *len),
		     recv[i+*len],
		     data_start + i* ag_size +j,
		     pkg_recv.data[data_start + i* ag_size +j]);
      }
    }
    *len += i;

    _xcp_verbose("Upload loop: Pkg DL: %i, AG len: %i, Data start: %i, Data len: %i, Len: %i, Stored Len: %i\n",
		 pkg_recv.dl,
		 ag_length,
		 data_start,
		 data_length,
		 *len,
		 store_len);

  }

  return CS_OK;
}

/**************************************************************************/
int cs_xcp_cmd_get_id(
		      cs_xcp_t *data,
		      int mode,
		      uint32_t *recvdata,
		      int *recvdatalen
		      )
/**************************************************************************/
{
  cs_xcp_pkg_t pkg_send;
  cs_xcp_pkg_t pkg_recv;
  uint32_t timeout;
  //uint16_t timeout_error;
  uint8_t lmode;
  cs_cu_data32_t ldata;
  uint32_t dl;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  if((mode < 0) || (mode > 255)){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;
  }

  if((recvdata == NULL) || (recvdatalen == NULL)){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;
  }

  _xcp_verbose("XCP API Call Get ID\n");

  if(_xcp_daq_feature_supported(data,XCP_CMD_GET_ID)){
    data->lerror = XCP_ERROR_SLAVE_UNSUPPORTED;
    return CS_FAIL;
  }

  _xcp_reset_error(data);
  _xcp_reset_pkg(&pkg_send);
  _xcp_reset_pkg(&pkg_recv);

  pkg_send.pkgtype =  XCP_PKG_CMD;
  pkg_send.PID = XCP_CMD_GET_ID;
  pkg_send.dl = 1;
  pkg_send.data[0] = (uint8_t)mode;

  timeout = data->xcpa2l.commode.timeout[XCP_A2L_TO_T1];
  //timeout_error = XCP_ERROR_TIMEOUTT1;

  //Send request
  if(_xcp_send(data,&pkg_send))
    return CS_FAIL;
 
  _xcp_reset_pkg(&pkg_recv);
  if(_xcp_receive(data,&pkg_recv,timeout,0)){
    _xcp_error("Receive negative response: %s\n",cs_xcp_get_lerrorstr(data));
    return CS_FAIL;
  }

  //Handle data
  //Check data length
  if(pkg_recv.dl < 7){
    data->lerror = XCP_ERROR_UNEXPECTED;
    return CS_FAIL;
  }

  //Get information
  //Get receiving mode
  lmode = pkg_recv.data[0];

  ldata.u8[0] = pkg_recv.data[3];
  ldata.u8[1] = pkg_recv.data[4];
  ldata.u8[2] = pkg_recv.data[5];
  ldata.u8[3] = pkg_recv.data[5];
  //Get data length
  dl = cs_cu_swap32(data->xcpa2l.commode.BO,ldata.u32);

  if(dl > 255){
    data->lerror = XCP_ERROR_UNEXPECTED;
    return CS_FAIL;
  }

  *recvdatalen = (int) dl;

  if(lmode == 1){
    _xcp_verbose("Receive mode 0\n");
    //TODO read rest of data
  } else {
    _xcp_verbose("Upload data\n");
    if(*recvdatalen == 0){
      _xcp_verbose("No ID available\n");
      return CS_OK;
    }
    if(cs_xcp_cmd_upload(data,recvdata,recvdatalen))
      return CS_FAIL;
  }
  return CS_OK;
}


/**************************************************************************/
int cs_xcp_cmd_get_daq_processor_info(
				      cs_xcp_t *data,
				      cs_xcp_a2l_daq_proc_t *rv
				      )
/**************************************************************************/
{
  cs_xcp_pkg_t pkg_send;
  cs_xcp_pkg_t pkg_recv;
  uint32_t timeout;
  uint16_t timeout_error;
  cs_cu_data16_t lvalue;
  cs_xcp_a2l_daq_proc_t ldata;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }	

  _xcp_verbose("XCP API Get DAQ Processor info\n");

  if(_xcp_daq_feature_supported(data,XCP_CMD_GET_DAQ_PROCESSOR_INFO)){
    data->lerror = XCP_ERROR_SLAVE_UNSUPPORTED;
    return CS_FAIL;
  }

  _xcp_reset_error(data);
  _xcp_reset_pkg(&pkg_send);
  _xcp_reset_pkg(&pkg_recv);

  pkg_send.pkgtype =  XCP_PKG_CMD;
  pkg_send.PID = XCP_CMD_GET_DAQ_PROCESSOR_INFO;
  pkg_send.dl = 0;

  timeout = data->xcpa2l.commode.timeout[XCP_A2L_TO_T1];
  timeout_error = XCP_ERROR_TIMEOUTT1;

  //Expect 7 data bytes
  pkg_recv.exp_dl = 7;

  //Send Recv and retry for special failures
  if(_xcp_send_recv_action(data,&pkg_send,&pkg_recv,pkg_send.PID,0,timeout,timeout_error)){
    return CS_FAIL;
  }

  //Check data length
  if(pkg_recv.dl < 7){
    data->lerror = XCP_ERROR_UNEXPECTED;
    return CS_FAIL;
  }

  //Get existing data
  memcpy(&ldata,&data->xcpa2l.daq_proc,sizeof(cs_xcp_a2l_daq_proc_t));

  //Decode data
  ldata.daq_static = (pkg_recv.data[0] & XCP_BM_DAQ_PROP_CONFIG)?0:1;
  ldata.prescaler_supported = (pkg_recv.data[0] & XCP_BM_DAQ_PROP_PRESCALER)?1:0;
  ldata.resume_supported = (pkg_recv.data[0] & XCP_BM_DAQ_PROP_RESUME)?1:0;
  ldata.bit_stim_supported = (pkg_recv.data[0] & XCP_BM_DAQ_PROP_BIT_STIM)?1:0;
  ldata.ts_supported = (pkg_recv.data[0] & XCP_BM_DAQ_PROP_TIMESTAMP)?1:0;
  ldata.pid_off_supported = (pkg_recv.data[0] & XCP_BM_DAQ_PROP_PID_OFF)?1:0;
  ldata.overload_mode = (pkg_recv.data[0] & XCP_BM_DAQ_PROP_OVERLOAD) >> 6;
  lvalue.u8[0] = pkg_recv.data[1];
  lvalue.u8[1] = pkg_recv.data[2];
  ldata.daq_max = cs_cu_swap16(data->xcpa2l.commode.BO,lvalue.u16);
  lvalue.u8[0] = pkg_recv.data[3];
  lvalue.u8[1] = pkg_recv.data[4];
  ldata.ev_max = cs_cu_swap16(data->xcpa2l.commode.BO,lvalue.u16);
  ldata.daq_min = (uint16_t)pkg_recv.data[5];
  ldata.odt_ident = (pkg_recv.data[6] & XCP_BM_DAQ_KEY_IDENT) >> 6;
  ldata.odt_ext = (pkg_recv.data[6] & XCP_BM_DAQ_KEY_ADDR_EXT) >> 4;
  ldata.odt_opt = (pkg_recv.data[6] & XCP_BM_DAQ_KEY_OPT);

  if(ldata.ev_max > CS_XCP_A2L_EVENT_MAX){
    _xcp_error("Event buffer overrange\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;
  }

  if(rv){
    cs_xcp_a2l_init_daqproc(rv);
    memcpy(rv,&ldata,sizeof(cs_xcp_a2l_daq_proc_t));
  }else{
    memcpy(&data->xcpa2l.daq_proc,&ldata,sizeof(cs_xcp_a2l_daq_proc_t));
  }

  return CS_OK;
}


/**************************************************************************/
int cs_xcp_cmd_get_daq_resolution_info(
				       cs_xcp_t *data,
				       cs_xcp_a2l_daq_proc_t *rv
				       )
/**************************************************************************/
{
  cs_xcp_pkg_t pkg_send;
  cs_xcp_pkg_t pkg_recv;
  uint32_t timeout;
  uint16_t timeout_error;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }

  _xcp_verbose("XCP API Get DAQ Resolution info\n");

  if(_xcp_daq_feature_supported(data,XCP_CMD_GET_DAQ_RESOLUTION_INFO)){
    data->lerror = XCP_ERROR_SLAVE_UNSUPPORTED;
    return CS_FAIL;
  }

  _xcp_reset_error(data);
  _xcp_reset_pkg(&pkg_send);
  _xcp_reset_pkg(&pkg_recv);

  pkg_send.pkgtype =  XCP_PKG_CMD;
  pkg_send.PID = XCP_CMD_GET_DAQ_RESOLUTION_INFO;
  pkg_send.dl = 0;

  timeout = data->xcpa2l.commode.timeout[XCP_A2L_TO_T1];
  timeout_error = XCP_ERROR_TIMEOUTT1;

  //Expect 7 data bytes
  pkg_recv.exp_dl = 7;

  //Send Recv and retry for special failures
  if(_xcp_send_recv_action(data,&pkg_send,&pkg_recv,pkg_send.PID,0,timeout,timeout_error)){
    return CS_FAIL;
  }

  //Check data length
  if(pkg_recv.dl < 7){
    data->lerror = XCP_ERROR_UNEXPECTED;
    return CS_FAIL;
  }

  //Decode data
  data->xcpa2l.daq_proc.daq_odt_entry_gr_size = pkg_recv.data[0];
  data->xcpa2l.daq_proc.daq_odt_entry_max_size = pkg_recv.data[1];
  data->xcpa2l.daq_proc.stim_odt_entry_gr_size = pkg_recv.data[2];
  data->xcpa2l.daq_proc.stim_odt_entry_max_size = pkg_recv.data[3];
  data->xcpa2l.daq_proc.ts_unit = (pkg_recv.data[4] & XCP_BM_TS_UNIT) >> 4;
  data->xcpa2l.daq_proc.ts_size = (pkg_recv.data[4] & XCP_BM_TS_SIZE);
  data->xcpa2l.daq_proc.ts_fixed = (pkg_recv.data[4] & XCP_BM_TS_FIXED)?1:0;

  if(rv)
    memcpy(rv,&data->xcpa2l.daq_proc,sizeof(cs_xcp_a2l_daq_proc_t));

  return CS_OK;
}

/**************************************************************************/
int cs_xcp_cmd_clear_daq_list(
			      cs_xcp_t *data,
			      int daq
			      )
/**************************************************************************/
{
  cs_xcp_pkg_t pkg_send;
  cs_xcp_pkg_t pkg_recv;
  uint32_t timeout;
  uint16_t timeout_error;
  cs_cu_data16_t ldaq;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }

  if(daq < 0){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_DAQ;
    return CS_FAIL;
  }

  if(daq > data->daqlist.no_daq){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_DAQ;
    return CS_FAIL;
  }

  _xcp_verbose("XCP API Call Clear DAQ list\n");

  _xcp_reset_error(data);
  _xcp_reset_pkg(&pkg_send);
  _xcp_reset_pkg(&pkg_recv);

  pkg_send.pkgtype =  XCP_PKG_CMD;
  pkg_send.PID = XCP_CMD_CLEAR_DAQ_LIST;
  pkg_send.dl = 3;
  pkg_send.data[0] = 0; //Reserved
  ldaq.u16 = cs_cu_swap16(data->xcpa2l.commode.BO,daq);
  pkg_send.data[1] = ldaq.u8[0];
  pkg_send.data[2] = ldaq.u8[1];

  timeout = data->xcpa2l.commode.timeout[XCP_A2L_TO_T1];
  timeout_error = XCP_ERROR_TIMEOUTT1;

  //Expect 0 data bytes
  pkg_recv.exp_dl = 0;

  //Send Recv and retry for special failures
  if(_xcp_send_recv_action(data,&pkg_send,&pkg_recv,pkg_send.PID,0,timeout,timeout_error)){
    return CS_FAIL;
  }

  //No data to handle

  //If no failure release daq structure
  _xcp_reset_daq(&data->daqlist.daq[daq]);
  
  return CS_OK;
}

/**************************************************************************/
int cs_xcp_cmd_set_daq_ptr(
			   cs_xcp_t *data,
			   int daq,
			   int odt,
			   int odt_entry
			   )
/**************************************************************************/
{
  cs_xcp_pkg_t pkg_send;
  cs_xcp_pkg_t pkg_recv;
  uint32_t timeout;
  uint16_t timeout_error;
  cs_cu_data16_t ldata;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }

  if((daq < 0) || (odt < 0) || (odt_entry < 0)){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_DAQ;
    return CS_FAIL;
  }

  if(daq > data->daqlist.no_daq){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_DAQ;
    return CS_FAIL;
  }

  if(odt > data->daqlist.daq[daq].no_odt){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_DAQ;
    return CS_FAIL;
  }

  if(odt_entry > data->daqlist.daq[daq].odt[odt].no_entry){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_DAQ;
    return CS_FAIL;
  }

  _xcp_verbose("XCP API Call Set DAQ PTR\n");

  _xcp_reset_error(data);
  _xcp_reset_pkg(&pkg_send);
  _xcp_reset_pkg(&pkg_recv);

  pkg_send.pkgtype =  XCP_PKG_CMD;
  pkg_send.PID = XCP_CMD_SET_DAQ_PTR;
  pkg_send.dl = 5;
  pkg_send.data[0] = 0; //Reserved
  ldata.u16 = cs_cu_swap16(data->xcpa2l.commode.BO,daq);
  pkg_send.data[1] = ldata.u8[0];
  pkg_send.data[2] = ldata.u8[1];
  pkg_send.data[3] = (uint8_t)odt;
  pkg_send.data[4] = (uint8_t)odt_entry;
  timeout = data->xcpa2l.commode.timeout[XCP_A2L_TO_T1];
  timeout_error = XCP_ERROR_TIMEOUTT1;

  //Expect 0 data bytes
  pkg_recv.exp_dl = 0;

  //Send Recv and retry for special failures
  if(_xcp_send_recv_action(data,&pkg_send,&pkg_recv,pkg_send.PID,0,timeout,timeout_error)){
    return CS_FAIL;
  }

  //No data to handle

  //Set working entry pointer in master
  data->work_entry = &data->daqlist.daq[daq].odt[odt].entry[odt_entry];

  return CS_OK;
}


/**************************************************************************/
int cs_xcp_cmd_write_daq(
			 cs_xcp_t *data,
			 uint8_t bit_offset,
			 int size,
			 uint8_t addr_extension,
			 uint32_t address
			 )
/**************************************************************************/
{
  cs_xcp_pkg_t pkg_send;
  cs_xcp_pkg_t pkg_recv;
  uint32_t timeout;
  uint16_t timeout_error;
  cs_cu_data32_t ldata;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }

  //Check for valid work pointer
  if(!data->work_entry){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_DAQ_PTR;
    return CS_FAIL;
  }

  if((size < 0) || (size > 255)){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;
  }


  if(size > data->xcpa2l.daq_proc.daq_odt_entry_max_size){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_DAQ_RANGE;
    return CS_FAIL;
  }

  _xcp_verbose("XCP API Call Write DAQ\n");

  _xcp_reset_error(data);
  _xcp_reset_pkg(&pkg_send);
  _xcp_reset_pkg(&pkg_recv);

  pkg_send.pkgtype =  XCP_PKG_CMD;
  pkg_send.PID = XCP_CMD_WRITE_DAQ;
  pkg_send.dl = 7;
  pkg_send.data[0] = bit_offset;
  pkg_send.data[1] = (uint8_t) size;
  pkg_send.data[2] = addr_extension;
  ldata.u32 = cs_cu_swap32(data->xcpa2l.commode.BO,address);
  pkg_send.data[3] = ldata.u8[0];
  pkg_send.data[4] = ldata.u8[1];
  pkg_send.data[5] = ldata.u8[2];
  pkg_send.data[6] = ldata.u8[3];

  timeout = data->xcpa2l.commode.timeout[XCP_A2L_TO_T1];
  timeout_error = XCP_ERROR_TIMEOUTT1;

  //Expect 0 data bytes
  pkg_recv.exp_dl = 0;

  //Send Recv and retry for special failures
  if(_xcp_send_recv_action(data,&pkg_send,&pkg_recv,pkg_send.PID,0,timeout,timeout_error)){
    return CS_FAIL;
  }

  //No data to handle

  //Store entry data
  data->work_entry->me.bitoffset = bit_offset;
  data->work_entry->me.extension = addr_extension;
  data->work_entry->me.address = address;
  data->work_entry->me.size = size;
  //Reset data
  memset( data->work_entry->data,0,sizeof(uint8_t)*CS_XCP_DATALENGTH);

  return CS_OK;
}

/**************************************************************************/
int cs_xcp_cmd_set_daq_list_mode(
				 cs_xcp_t *data,
				 int daq,
				 int ev,
				 cs_xcp_daq_dir_t direction,
				 int pid_off,
				 int timestamp,
				 int prescaler,
				 uint8_t priority
				 )
/**************************************************************************/
{
  cs_xcp_pkg_t pkg_send;
  cs_xcp_pkg_t pkg_recv;
  uint32_t timeout;
  uint16_t timeout_error;
  cs_cu_data16_t ldata;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }

  if((daq < 0) || (ev < 0)){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_DAQ;
    return CS_FAIL;
  }

  if(daq > data->daqlist.no_daq){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_DAQ_RANGE;
    return CS_FAIL;
  }

  if((ev >= CS_XCP_A2L_EVENT_MAX)){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_EV_RANGE;
    return CS_FAIL;
  }

  if(!data->xcpa2l.daq_proc.pid_off_supported && (pid_off == 1)){
    data->lerror = XCP_ERROR_SLAVE_UNSUPPORTED;
    return CS_FAIL;
  }

  if(!data->xcpa2l.daq_proc.ts_supported && (timestamp == 1)){
    data->lerror = XCP_ERROR_SLAVE_UNSUPPORTED;
    return CS_FAIL;
  }

  if(data->xcpa2l.daq_proc.ts_fixed && (timestamp == 0)){
    data->lerror = XCP_ERROR_SLAVE_UNSUPPORTED;
    return CS_FAIL;
  }

  if(!data->xcpa2l.daq_proc.prescaler_supported && (prescaler != 1)){
    data->lerror = XCP_ERROR_SLAVE_UNSUPPORTED;
    return CS_FAIL;
  }

  _xcp_verbose("XCP API Call Set DAQ List mode\n");

  if(_xcp_daq_feature_supported(data,XCP_CMD_SET_DAQ_LIST_MODE)){
    data->lerror = XCP_ERROR_SLAVE_UNSUPPORTED;
    return CS_FAIL;
  }

  _xcp_reset_error(data);
  _xcp_reset_pkg(&pkg_send);
  _xcp_reset_pkg(&pkg_recv);

  pkg_send.pkgtype =  XCP_PKG_CMD;
  pkg_send.PID = XCP_CMD_SET_DAQ_LIST_MODE;
  pkg_send.dl = 7;
  pkg_send.data[0] = 0;
  pkg_send.data[0] |= pid_off?1<<5:0;
  pkg_send.data[0] |= timestamp?1<<4:0;
  pkg_send.data[0] |= timestamp?1<<4:0;
  pkg_send.data[0] |= direction == XCP_DAQ_DIR_STIM?1<<1:0;
  ldata.u16 = cs_cu_swap16(data->xcpa2l.commode.BO,(uint16_t)daq);
  pkg_send.data[1] = ldata.u8[0];
  pkg_send.data[2] = ldata.u8[1];
  ldata.u16 = cs_cu_swap16(data->xcpa2l.commode.BO,(uint16_t)ev);
  pkg_send.data[3] = ldata.u8[0];
  pkg_send.data[4] = ldata.u8[1];
  pkg_send.data[5] = (uint8_t)prescaler;
  pkg_send.data[6] = priority;

  timeout = data->xcpa2l.commode.timeout[XCP_A2L_TO_T1];
  timeout_error = XCP_ERROR_TIMEOUTT1;

  //Send Recv and retry for special failures
  if(_xcp_send_recv_action(data,&pkg_send,&pkg_recv,pkg_send.PID,0,timeout,timeout_error)){
    return CS_FAIL;
  }

  //No data to handle

  //Store settings
  data->daqlist.daq[daq].dir = direction;
  if(data->iftype == XCP_CAN){
    //depending on direction set CAN id
    switch(data->daqlist.daq[daq].dir){
    case XCP_DAQ_DIR_DAQ:
      data->daqlist.daq[daq].canid = cs_xcp_can_getRespId(data->ifcan);
    case XCP_DAQ_DIR_STIM:
      data->daqlist.daq[daq].canid = cs_xcp_can_getCmdId(data->ifcan);
    default:
      data->daqlist.daq[daq].canid = 0;
    }
  } else {
    data->daqlist.daq[daq].canid = 0;
  }
  data->daqlist.daq[daq].prescaler = prescaler;
  data->daqlist.daq[daq].ev = ev;
  data->daqlist.daq[daq].priority = priority;
  data->daqlist.daq[daq].timestamp = timestamp;
  data->daqlist.daq[daq].pid_off = pid_off;

  //Store event period for recv timeout detection
  data->daqlist.daq[daq].period = data->xcpa2l.daq_proc.event[ev].period;

  return CS_OK;
}

/**************************************************************************/
int cs_xcp_cmd_get_daq_list_mode(
				 cs_xcp_t *data,
				 int daq,
				 cs_xcp_daq_t *rdaq
				 )
/**************************************************************************/
{
  cs_xcp_pkg_t pkg_send;
  cs_xcp_pkg_t pkg_recv;
  uint32_t timeout;
  uint16_t timeout_error;
  cs_cu_data16_t ldata;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }

  if(rdaq == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }

  if(daq < 0){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_DAQ;
    return CS_FAIL;
  }

  if(daq > data->daqlist.no_daq){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_DAQ_RANGE;
    return CS_FAIL;
  }

  _xcp_verbose("XCP API Call Get DAQ List mode\n");

  _xcp_reset_error(data);
  _xcp_reset_pkg(&pkg_send);
  _xcp_reset_pkg(&pkg_recv);

  pkg_send.pkgtype =  XCP_PKG_CMD;
  pkg_send.PID = XCP_CMD_GET_DAQ_LIST_MODE;
  pkg_send.dl = 3;
  pkg_send.data[0] = 0; //Reserved
  ldata.u16 = cs_cu_swap16(data->xcpa2l.commode.BO,(uint16_t)daq);
  pkg_send.data[1] = ldata.u8[0];
  pkg_send.data[2] = ldata.u8[1];

  timeout = data->xcpa2l.commode.timeout[XCP_A2L_TO_T1];
  timeout_error = XCP_ERROR_TIMEOUTT1;

  //Send Recv and retry for special failures
  if(_xcp_send_recv_action(data,&pkg_send,&pkg_recv,pkg_send.PID,0,timeout,timeout_error)){
    return CS_FAIL;
  }

  //Store settings
  data->daqlist.daq[daq].selected = (pkg_recv.data[0] & (1 << 0))? 1: 0;
  data->daqlist.daq[daq].dir = (pkg_recv.data[0] & (1 << 1))? XCP_DAQ_DIR_STIM: XCP_DAQ_DIR_DAQ;
  data->daqlist.daq[daq].timestamp = (pkg_recv.data[0] & (1 << 4))? 1: 0;
  data->daqlist.daq[daq].pid_off = (pkg_recv.data[0] & (1 << 5))? 1: 0;
  data->daqlist.daq[daq].running = (pkg_recv.data[0] & (1 << 6))? 1: 0;
  data->daqlist.daq[daq].resume = (pkg_recv.data[0] & (1 << 7))? 1: 0;
  ldata.u8[0] = pkg_recv.data[3];
  ldata.u8[0] = pkg_recv.data[4];
  data->daqlist.daq[daq].ev = cs_cu_swap16(data->xcpa2l.commode.BO,ldata.u16);
  data->daqlist.daq[daq].prescaler = pkg_recv.data[5];
  data->daqlist.daq[daq].priority = pkg_recv.data[6];

  return CS_OK;
}


/**************************************************************************/
int cs_xcp_cmd_free_daq(
			cs_xcp_t *data
			)
/**************************************************************************/
{
  cs_xcp_pkg_t pkg_send;
  cs_xcp_pkg_t pkg_recv;
  uint32_t timeout;
  uint16_t timeout_error;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }

  _xcp_verbose("XCP API Call Free DAQ\n");

  _xcp_reset_error(data);
  _xcp_reset_pkg(&pkg_send);
  _xcp_reset_pkg(&pkg_recv);

  pkg_send.pkgtype =  XCP_PKG_CMD;
  pkg_send.PID = XCP_CMD_FREE_DAQ;
  pkg_send.dl = 0;

  timeout = data->xcpa2l.commode.timeout[XCP_A2L_TO_T1];
  timeout_error = XCP_ERROR_TIMEOUTT1;

  //Expect 0 data bytes
  pkg_recv.exp_dl = 0;

  //Send Recv and retry for special failures
  if(_xcp_send_recv_action(data,&pkg_send,&pkg_recv,pkg_send.PID,0,timeout,timeout_error)){
    return CS_FAIL;
  }

  //No data to handle

  //If no failure release daq structure
  _xcp_reset_daqlist(data);
  
  return CS_OK;
}

/**************************************************************************/
int cs_xcp_cmd_alloc_daq(
			 cs_xcp_t *data,
			 int count
			 )
/**************************************************************************/
{
  cs_xcp_pkg_t pkg_send;
  cs_xcp_pkg_t pkg_recv;
  uint32_t timeout;
  uint16_t timeout_error;
  cs_cu_data16_t lcount;
  
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  _xcp_verbose("XCP API Call Alloc DAQ\n");
  
  
  if(count > 255){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;   
  }
  
  _xcp_reset_error(data);
  _xcp_reset_pkg(&pkg_send);
  _xcp_reset_pkg(&pkg_recv);
  
  pkg_send.pkgtype =  XCP_PKG_CMD;
  pkg_send.PID = XCP_CMD_ALLOC_DAQ;
  pkg_send.dl = 3;
  lcount.u16 = cs_cu_swap16(data->xcpa2l.commode.BO,(uint16_t) count);
  pkg_send.data[0] = 0; // Reserved
  pkg_send.data[1] = lcount.u8[0];
  pkg_send.data[2] = lcount.u8[1];

  timeout = data->xcpa2l.commode.timeout[XCP_A2L_TO_T1];
  timeout_error = XCP_ERROR_TIMEOUTT1;
  
  //Expect 0 data bytes
  pkg_recv.exp_dl = 0;

  //Send Recv and retry for special failures
  if(_xcp_send_recv_action(data,&pkg_send,&pkg_recv,pkg_send.PID,0,timeout,timeout_error)){
    return CS_FAIL;
  }
  
  //No data to handle
  
  //Prepare local structure
  return cs_xcp_data_set_no_daq(data,count);
}

/**************************************************************************/
int cs_xcp_cmd_alloc_odt(
			 cs_xcp_t *data,
			 int daq,
			 int count
			 )
/**************************************************************************/
{
  cs_xcp_pkg_t pkg_send;
  cs_xcp_pkg_t pkg_recv;
  uint32_t timeout;
  uint16_t timeout_error;
  cs_cu_data16_t ldaq;
  
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
      
  _xcp_verbose("XCP API Call Alloc ODT\n");
  
  
  if((daq < 0) || (daq > 255)){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;   
  }
  
  if((count <= 0) || (count > 255)){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;   
  }
  
  if(daq >= data->daqlist.no_daq){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_DAQ_RANGE;
    return CS_FAIL;   
  }
  
  if(data->daqlist.daq[daq].odt){
    data->lerror = XCP_ERROR_DAQ_SEQUENCE;
    return CS_FAIL;   
  }
  
  _xcp_reset_error(data);
  _xcp_reset_pkg(&pkg_send);
  _xcp_reset_pkg(&pkg_recv);
  
  pkg_send.pkgtype =  XCP_PKG_CMD;
  pkg_send.PID = XCP_CMD_ALLOC_ODT;
  pkg_send.dl = 4;
  pkg_send.data[0] = 0; //Reserved
  ldaq.u16 = cs_cu_swap16(data->xcpa2l.commode.BO,daq);
  pkg_send.data[1] = ldaq.u8[0];
  pkg_send.data[2] = ldaq.u8[1];
  pkg_send.data[3] = (uint8_t)count;
  
  timeout = data->xcpa2l.commode.timeout[XCP_A2L_TO_T1];
  timeout_error = XCP_ERROR_TIMEOUTT1;
  
  //Expect 0 data bytes
  pkg_recv.exp_dl = 0;
  
  //Send Recv and retry for special failures
  if(_xcp_send_recv_action(data,&pkg_send,&pkg_recv,pkg_send.PID,0,timeout,timeout_error)){
    return CS_FAIL;
  }
  
  //No data to handle
  
  //If no failure alloc odt structure
  return cs_xcp_data_set_no_odt(data,daq,count);
}


/**************************************************************************/
int cs_xcp_cmd_alloc_odt_entry(
			       cs_xcp_t *data,
			       int daq,
			       int odt,
			       int count
			       )
/**************************************************************************/
{
  cs_xcp_pkg_t pkg_send;
  cs_xcp_pkg_t pkg_recv;
  uint32_t timeout;
  uint16_t timeout_error;
  cs_cu_data16_t ldaq;
  
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
    
  _xcp_verbose("XCP API Call Alloc ODT Entry\n");
  
  
  if((daq < 0) || (daq > 255)){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;   
  }

  if((odt < 0) || (odt > 255)){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;   
  }
  
  if((count <= 0) || (count > 255)){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;   
  }

  if(daq >= data->daqlist.no_daq){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_DAQ_RANGE;
    return CS_FAIL;   
  }  

  if(odt >= data->daqlist.daq[daq].no_odt){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_ODT_RANGE;
    return CS_FAIL;   
  }  
    
  if(data->daqlist.daq[daq].odt[odt].entry){
    data->lerror = XCP_ERROR_DAQ_SEQUENCE;
    return CS_FAIL;   
  }
  
  _xcp_reset_error(data);
  _xcp_reset_pkg(&pkg_send);
  _xcp_reset_pkg(&pkg_recv);
  
  pkg_send.pkgtype =  XCP_PKG_CMD;
  pkg_send.PID = XCP_CMD_ALLOC_ODT_ENTRY;
  pkg_send.dl = 5;
  pkg_send.data[0] = 0; //Reserved
  ldaq.u16 = cs_cu_swap16(data->xcpa2l.commode.BO,daq);
  pkg_send.data[1] = ldaq.u8[0];
  pkg_send.data[2] = ldaq.u8[1];
  pkg_send.data[3] = (uint8_t)odt;
  pkg_send.data[4] = (uint8_t)count;
  
  timeout = data->xcpa2l.commode.timeout[XCP_A2L_TO_T1];
  timeout_error = XCP_ERROR_TIMEOUTT1;
  
  //Expect 0 data bytes
  pkg_recv.exp_dl = 0;
  
  //Send Recv and retry for special failures
  if(_xcp_send_recv_action(data,&pkg_send,&pkg_recv,pkg_send.PID,0,timeout,timeout_error)){
    return CS_FAIL;
  }
  
  //No data to handle
  
  //If no failure alloc odt entry structure
  return cs_xcp_data_set_no_odt_entry(data,daq,odt,count);
}

/**************************************************************************/
int cs_xcp_cmd_start_stop_daq_list(
				   cs_xcp_t *data,
				   int daq,
				   int mode
				   )
/**************************************************************************/
{
  cs_xcp_pkg_t pkg_send;
  cs_xcp_pkg_t pkg_recv;
  uint32_t timeout;
  uint16_t timeout_error;
  cs_cu_data16_t ldaq;
  
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  _xcp_verbose("XCP API Call Start Stop DAQ list\n");
  
  
  if((mode < 0) || (mode > 3)){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;   
  }
  
  if((daq < 0) || (daq > data->daqlist.no_daq)){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_DAQ_RANGE;
    return CS_FAIL;   
  }

  if((mode == 1) && data->daqlist.daq[daq].running){
    data->lerror = XCP_ERROR_DAQ_SEQUENCE;
    return CS_FAIL;   
  }

  _xcp_reset_error(data);
  _xcp_reset_pkg(&pkg_send);
  _xcp_reset_pkg(&pkg_recv);
  
  pkg_send.pkgtype =  XCP_PKG_CMD;
  pkg_send.PID = XCP_CMD_START_STOP_DAQ_LIST;
  pkg_send.dl = 3;
  pkg_send.data[0] = (uint8_t)mode;
  ldaq.u16 = cs_cu_swap16(data->xcpa2l.commode.BO,(uint16_t) daq);
  pkg_send.data[1] = ldaq.u8[0];
  pkg_send.data[2] = ldaq.u8[1];

  timeout = data->xcpa2l.commode.timeout[XCP_A2L_TO_T1];
  timeout_error = XCP_ERROR_TIMEOUTT1;
  
  //Expect 1 data bytes
  pkg_recv.exp_dl = 1;

  //Send Recv and retry for special failures
  if(_xcp_send_recv_action(data,&pkg_send,&pkg_recv,pkg_send.PID,0,timeout,timeout_error)){
    return CS_FAIL;
  }
  
  //No data to handle
  
  switch(mode){
  case 0://Stop specific
    _xcp_daq_stop(data,daq);
    break;
  case 1: //Start specific
    _xcp_daq_start(data,daq,pkg_recv.data[0]);
    break;
  case 2: //Select specific
    _xcp_daq_select(data,daq);
    break;
  }

  return CS_OK;
}

/**************************************************************************/
int cs_xcp_cmd_start_stop_synch(
				cs_xcp_t *data,
				int mode
				)
/**************************************************************************/
{
  int i;
  cs_xcp_pkg_t pkg_send;
  cs_xcp_pkg_t pkg_recv;
  uint32_t timeout;
  uint16_t timeout_error;
  
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  _xcp_verbose("XCP API Call Start Stop synch\n");
  
  
  if((mode < 0) || (mode > 3)){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;   
  }
  
  _xcp_reset_error(data);
  _xcp_reset_pkg(&pkg_send);
  _xcp_reset_pkg(&pkg_recv);
  
  pkg_send.pkgtype =  XCP_PKG_CMD;
  pkg_send.PID = XCP_CMD_START_STOP_SYNCH;
  pkg_send.dl = 1;
  pkg_send.data[0] = (uint8_t)mode;

  timeout = data->xcpa2l.commode.timeout[XCP_A2L_TO_T1];
  timeout_error = XCP_ERROR_TIMEOUTT1;
  
  //Expect 0 data bytes
  pkg_recv.exp_dl = 0;

  //Send Recv and retry for special failures
  if(_xcp_send_recv_action(data,&pkg_send,&pkg_recv,pkg_send.PID,0,timeout,timeout_error)){
    return CS_FAIL;
  }
  
  //No data to handle
  
  switch(mode){
  case 0://Stop all
    for(i = 0; i< data->daqlist.no_daq; i++)
      _xcp_daq_stop(data,i);
    break;
  case 1: //Start selected
    _xcp_daq_start_selected(data);
    break;
  case 2: //Stop selected
    _xcp_daq_stop_selected(data);
    break;
  }

  return CS_OK;
}


/**************************************************************************/
int cs_xcp_cmd_get_daq_clock(
			     cs_xcp_t *data,
			     double *jitter
			     )
/**************************************************************************/
{
  cs_xcp_pkg_t pkg_send;
  cs_xcp_pkg_t pkg_recv;
  uint32_t timeout;
  uint16_t timeout_error;
  cs_cu_data32_t l32;
  double ts_recv;
  double ts_slave;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  _xcp_verbose("XCP API Call Get DAQ clock\n");

  if(_xcp_daq_feature_supported(data,XCP_CMD_GET_DAQ_CLOCK)){
    _xcp_error("Feature is not supported\n");
    data->lerror = XCP_ERROR_SLAVE_UNSUPPORTED;
    return CS_FAIL;
  }
    
  if(!data->xcpa2l.daq_proc.ts_supported){
    _xcp_error("Timestamps are not supoorted\n");
    data->lerror = XCP_ERROR_SLAVE_UNSUPPORTED;
    return CS_FAIL;
  }

  _xcp_reset_error(data);
  _xcp_reset_pkg(&pkg_send);
  _xcp_reset_pkg(&pkg_recv);
  
  pkg_send.pkgtype =  XCP_PKG_CMD;
  pkg_send.PID = XCP_CMD_GET_DAQ_CLOCK;
  pkg_send.dl = 0;

  timeout = data->xcpa2l.commode.timeout[XCP_A2L_TO_T1];
  timeout_error = XCP_ERROR_TIMEOUTT1;
  
  //Expect 7 data bytes
  pkg_recv.exp_dl = 7;
  
  //Send Recv and retry for special failures
  
  if(_xcp_send_recv_action(data,&pkg_send,&pkg_recv,pkg_send.PID,0,timeout,timeout_error)){
    return CS_FAIL;
  }

  //We have already got an send and receive timestamp, now get XCP timestamp
  ts_recv = cs_ts_timeval2d(pkg_recv.tv);

  l32.u8[0] = pkg_recv.data[3];
  l32.u8[1] = pkg_recv.data[4];
  l32.u8[2] = pkg_recv.data[5];
  l32.u8[3] = pkg_recv.data[6];
  
  l32.u32 = cs_cu_swap32(data->xcpa2l.commode.BO,l32.u32);
    
  //Convert timestamps
  ts_slave = cs_ts_timespec2d(cs_cu_ts_timestamp2ts(l32.u32,(uint8_t)data->xcpa2l.daq_proc.ts_unit));

  //Calc difference
  data->daqlist.clock_shift = ts_recv - ts_slave;
  data->daqlist.clock_base = ts_slave;

  if(jitter)
    *jitter = data->daqlist.clock_shift;

  return CS_OK;
}


/**************************************************************************/
int cs_xcp_cmd_get_daq_event_info(
				  cs_xcp_t *data,
				  int event,
				  cs_xcp_a2l_event_t *ev
				  )
/**************************************************************************/
{
  cs_xcp_pkg_t pkg_send;
  cs_xcp_pkg_t pkg_recv;
  uint32_t timeout;
  uint16_t timeout_error;
  cs_cu_data16_t l16;
  
  
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  _xcp_error("XCP API Call Get DAQ event info\n");

  if(_xcp_daq_feature_supported(data,XCP_CMD_GET_DAQ_EVENT_INFO)){
    data->lerror = XCP_ERROR_SLAVE_UNSUPPORTED;
    return CS_FAIL;
  }
  
  if((event < 0) || (event >= data->xcpa2l.daq_proc.ev_max) || (event >= CS_XCP_A2L_EVENT_MAX)){
    data->lerror = XCP_ERROR_API;
    return CS_FAIL;       
  }
  
  _xcp_reset_error(data);
  _xcp_reset_pkg(&pkg_send);
  _xcp_reset_pkg(&pkg_recv);
  
  pkg_send.pkgtype =  XCP_PKG_CMD;
  pkg_send.PID = XCP_CMD_GET_DAQ_EVENT_INFO;
  pkg_send.dl = 3;
  
  pkg_send.data[0] = 0; //Reserved
  l16.u16 = cs_cu_swap16(data->xcpa2l.commode.BO,(uint16_t)event);
  pkg_send.data[1] = l16.u8[0];
  pkg_send.data[2] = l16.u8[1];
  
  timeout = data->xcpa2l.commode.timeout[XCP_A2L_TO_T1];
  timeout_error = XCP_ERROR_TIMEOUTT1;
  
  //Expect 6 data bytes
  pkg_recv.exp_dl = 6;
  
  //Send Recv and retry for special failures
  
  if(_xcp_send_recv_action(data,&pkg_send,&pkg_recv,pkg_send.PID,0,timeout,timeout_error)){
    return CS_FAIL;
  }
  
  data->xcpa2l.daq_proc.event[event].consistency = ((pkg_recv.data[0] & 0xC0) >> 6);
  data->xcpa2l.daq_proc.event[event].STIM_supported = (pkg_recv.data[0] & 0x08)?1:0;
  data->xcpa2l.daq_proc.event[event].DAQ_supported = (pkg_recv.data[0] & 0x04)?1:0;
  data->xcpa2l.daq_proc.event[event].max_daq_list = (int)pkg_recv.data[1];
  data->xcpa2l.daq_proc.event[event].name_length = (int)pkg_recv.data[2];
  data->xcpa2l.daq_proc.event[event].period_units = pkg_recv.data[3];
  data->xcpa2l.daq_proc.event[event].ts_unit = pkg_recv.data[4];
  data->xcpa2l.daq_proc.event[event].priority = pkg_recv.data[5];
  
  //Precalc period in double
  data->xcpa2l.daq_proc.event[event].period = cs_cu_ts_unit2double(data->xcpa2l.daq_proc.event[event].ts_unit,data->xcpa2l.daq_proc.event[event].period_units);
  
  
  if(ev)
    *ev = data->xcpa2l.daq_proc.event[event];
  
  return CS_OK;
}
