/*  XCP CAN Stack for SocketCAN
    Copyright (C) 2014 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/mman.h>
#include <unistd.h>

#include "cs_calibration_utils.h"
#include "cs_defines.h"
#include "cs_time.h"
#include "cs_xcp.h"
#include "cs_xcp_private.h"
#include "cs_xcp_api.h"

#define XCP_API_DEBUG 0

/**************************************************************************/
int cs_xcp_api_load(
		    cs_xcp_t *data,
		    char * filename,
		    char* project,
		    char* module,
		    char *iface,
		    uint32_t stim
		    )
/**************************************************************************/
{
  int asap2major = 0;
  int asap2minor = 0;

  if((data == NULL) ||
     (filename == NULL) ||
     (project == NULL) ||
     (module == NULL) ||
     (iface == NULL)){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }

  //Reset protocol stack
  cs_xcp_init(data);

  //Try to open file
  if(cs_xcp_a2l_open(&data->xcpa2l,filename,project,module,iface)){
    data->lerror = XCP_ERROR_API_FILE;
    return CS_FAIL;
  }

  //Store stim time
  data->stim = stim;
  //Set Timeout to default 0, deactivated
  data->daq_fac_recvtimeout = 0.0;

  //Get ASAP-Version
  if(cs_a2l_getversion(&data->xcpa2l.a2l,&asap2major,&asap2minor)){
    data->lerror = XCP_ERROR_API_PARSE;
    return CS_FAIL;
  }
  _xcp_verbose("Get version: major %i minor %i\n",asap2major,asap2minor);

  if((asap2major > 1) || (asap2minor > 60)){
    data->lerror = XCP_ERROR_API_VERSION;
    return CS_FAIL;
  }

  //Get Protocol settings
  if(cs_xcp_a2l_GetProtocoll(&data->xcpa2l,&data->xcpa2l.commode)){
    data->lerror = XCP_ERROR_API_PARSE;
    return CS_FAIL;
  }

  //Get DAQ settings
  if(cs_xcp_a2l_GetDAQ(&data->xcpa2l,&data->xcpa2l.daq_proc)){
    data->lerror = XCP_ERROR_API_PARSE;
    return CS_FAIL;
  }

  return CS_OK;
}


/**************************************************************************/
int cs_xcp_api_open(
		    cs_xcp_t *data,
		    cs_xcp_ifacetype_t iftype,
		    char *iface
		    )
/**************************************************************************/
{
  cs_xcp_can_t *ifdata;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }


  switch(iftype){
  case XCP_IFUNDEF:
    _xcp_error("Undefined interface type\n");
    data->lerror = XCP_ERROR_API_PARAM;
    return CS_FAIL;    
    break;
  case XCP_CAN:
    if(iface == NULL){
      _xcp_error("(CAN) No interface name defined\n");
      data->lerror = XCP_ERROR_API_PARAM;
      return CS_FAIL;    
    }
    
    ifdata = cs_xcp_can_init();
    
    if(ifdata == NULL){
      _xcp_error("(CAN) Init CAN interface failed\n");
      data->lerror = XCP_ERROR_API_PARAM;
      return CS_FAIL;    
    }
    
    //Get Physical Interface CAN settings
    if(cs_xcp_a2l_GetPhysInterfaceCAN(&data->xcpa2l,ifdata)){
      _xcp_error("(CAN) Failed to parse Interface CAN section\n");
      data->lerror = XCP_ERROR_API_PARSE;
      return CS_FAIL;    
    }

    //Configure interface
    if(cs_xcp_iface_can_setup(data,ifdata)){
      _xcp_error("(CAN) Failed to configure interface\n");
      data->lerror = XCP_ERROR_API_PARAM;
      return CS_FAIL;    
    }

    //Try to open interface
    if(cs_xcp_iface_can_open(data,iface)){
      _xcp_error("(CAN) Open interface failed\n");
      return CS_FAIL;
    }
    
    break;
  case XCP_USB:
  case XCP_ETHERNET:
    _xcp_error("Interface is not supported\n");
    data->lerror = XCP_ERROR_API_PARAM;
    return CS_FAIL;    
    break;
  }
  
  return CS_OK;
}

/**************************************************************************/
int cs_xcp_api_close(
		     cs_xcp_t *data
		     )
/**************************************************************************/
{
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  switch(data->iftype){
  case XCP_IFUNDEF:
    _xcp_error("Undefined interface type\n");
    data->lerror = XCP_ERROR_INTERFACE;
    return CS_FAIL;    
    break;
  case XCP_CAN:
    //Send disconnect
    if(data->isConnected && cs_xcp_cmd_disconnect(data)){
      _xcp_error("Send disconnect failed\n"); 
      return CS_FAIL;
    }
    //Close physical interface
    if(cs_xcp_close(data)){
      _xcp_error("Failed to close physical interface\n");
      data->lerror = XCP_ERROR_INTERFACE;
      return CS_FAIL;
    }
    break;
  case XCP_USB:
  case XCP_ETHERNET:
    _xcp_error("Interface is not supported\n");
    data->lerror = XCP_ERROR_INTERFACE;
    return CS_FAIL;    
    break;
  }
  
  return CS_OK;
}

/**************************************************************************/
int cs_xcp_api_addSignalDAQ(
			    cs_xcp_t *data,
			    char* signalname,
			    char* ev_name,
			    int prescaler,
			    double *value,
			    double *ts,
			    void (*extcallback)(double value, double ts, void *priv),
			    void *priv
			    )
/**************************************************************************/
{
  int i;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }

  if((signalname == NULL) || (ev_name == NULL)){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API_PARAM;
    return CS_FAIL;
  }

  //Check for already configured signal number
  if((data->xcpa2l.a2l.signalcounter + 1) >= CS_A2L_SIGNALMAX){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API_SIGNALRANGE;
    return CS_FAIL;
  }

  //Check for valid prescaler
  if(prescaler < 1){
    _xcp_error("Parameter failure\n");
    data->lerror = XCP_ERROR_API_PARAM;
    return CS_FAIL;
  }

  //Check if we support prescaler
  if(!data->xcpa2l.daq_proc.prescaler_supported &&
     (prescaler > 1)){
    data->lerror = XCP_ERROR_SLAVE_UNSUPPORTED;
    return CS_FAIL;
  }

  //Search for signal
  if(cs_a2l_GetMeasurement(&data->xcpa2l.a2l,&data->xcpa2l.a2l.signals[data->xcpa2l.a2l.signalcounter],signalname)){
    data->lerror = XCP_ERROR_API_PARSE;
    return CS_FAIL;
  }

  //Look if Event exists
  for(i = 0; i < data->xcpa2l.daq_proc.ev_max; i++){
    if(strcmp(data->xcpa2l.daq_proc.event[i].name,ev_name) == 0){
      //found Event, store event number
      data->xcpa2l.a2l.signals[data->xcpa2l.a2l.signalcounter].ev_idx = i;
      break;
    }
  }
  if(i >= data->xcpa2l.daq_proc.ev_max){
    data->lerror = XCP_ERROR_API_PARAM;
    return CS_FAIL;
  }

  //Set prescaler
  data->xcpa2l.a2l.signals[data->xcpa2l.a2l.signalcounter].prescaler = prescaler;

  //Store feedback values
  data->xcpa2l.a2l.signals[data->xcpa2l.a2l.signalcounter].exvalue = value;
  data->xcpa2l.a2l.signals[data->xcpa2l.a2l.signalcounter].exts = ts;
  data->xcpa2l.a2l.signals[data->xcpa2l.a2l.signalcounter].extcallback = extcallback;
  data->xcpa2l.a2l.signals[data->xcpa2l.a2l.signalcounter].extcallbackpriv = priv;
  
  //Increment signalcounter, because we have a new valid system
  data->xcpa2l.a2l.signalcounter++;

  return CS_OK;
}

/**************************************************************************/
int cs_xcp_api_connect(
		       cs_xcp_t *data,
		       cs_xcp_status_t *status
		       )
/**************************************************************************/
{
  cs_xcp_a2l_daq_proc_t ldata;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }

  //Try to connect
  if(cs_xcp_cmd_connect(data,0)){
    _xcp_error("Failed to connect to slave\n");
    return CS_FAIL;
  }

  if(status != NULL){
    cs_ts_sleep(data->stim);
    //Get status
    if(cs_xcp_cmd_get_status(data,status)){
      _xcp_error("Failed to get status\n");
      return CS_FAIL;
    }
  }
  cs_ts_sleep(data->stim);
  //Read out DAQ processor information print it in debug mode
  if(cs_xcp_cmd_get_daq_processor_info(data,&ldata)){
    _xcp_error("Failed to Get DAQ processor info\n");
    return CS_FAIL;
  }
 
  //Set connected flag
  data->isConnected = 1;
  return CS_OK;
}


/**************************************************************************/
int cs_xcp_api_configureDAQ(
			    cs_xcp_t *data
			    )
/**************************************************************************/
{
  cs_xcp_status_t status;
  int i,j;
  int daqcounter = 0;
  int daq_idx = -1;
  int dsizeodtfirst = 0; //data segment size of first odt (timestamp, pid)
  int dsizeodtsecond = 0; //data segement size of following odts (pid)
  int maxodt = 0xFF; //Maximal number of ODTs per DAQ
  int flag_insert = 0; //Flag storage for insert signal in ODT

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }

  if(data->xcpa2l.daq_proc.daq_max == 0){
    _xcp_error("DAQ max is zero\n");
    data->lerror = XCP_ERROR_DAQ_PROCESS;
    return CS_FAIL;
  }

  //Check if slave is able to be configured dynamically
  if(data->xcpa2l.daq_proc.daq_static == 0){
    _xcp_error("Only static DAQ configuration is available\n");
    data->lerror = XCP_ERROR_SLAVE_UNSUPPORTED;
    return CS_FAIL;
  }

  //Check if there is a running configuration
  if(cs_xcp_cmd_get_status(data,&status)){
    return CS_FAIL;
  }

  if(status.daq_running){
    _xcp_error("There is a running DAQ configuration\n");
    data->lerror = XCP_ERROR_DAQ_PROCESS;
    return CS_FAIL;
  }

  //Precalculate data segment size of ODTs
  //set to maximum
  dsizeodtfirst = data->xcpa2l.commode.max_DTO;
  dsizeodtsecond = data->xcpa2l.commode.max_DTO;

  //Reduce by PID
  //note: we always configure with PID transfer, actually DTO identification via CAN id is not supported
  switch(data->xcpa2l.daq_proc.odt_ident){
  case XCP_A2L_ODT_IDENT_ABS:
    dsizeodtfirst -= 1; //Only 1 Byte
    dsizeodtsecond -= 1;
    break;
  case XCP_A2L_ODT_IDENT_RAB:
    dsizeodtfirst -= 2;//Skip PID, and DAQ byte
    dsizeodtsecond -= 2;
    break;
  case XCP_A2L_ODT_IDENT_RAW:
    dsizeodtfirst -= 3;//Skip PID, and DAQ word
    dsizeodtsecond -= 3;
    break;
  case XCP_A2L_ODT_IDENT_RAWA:
    dsizeodtfirst -= 4;//Skip PID, and DAQ word and fill byte
    dsizeodtsecond -= 4;
    break;
  }
  
  //Now reduce timestamp
  //note: if timestamp is available we will transfer it
  if(data->xcpa2l.daq_proc.ts_supported){
    //Only reduce first odt by timestamp size
    dsizeodtfirst -= data->xcpa2l.daq_proc.ts_size;
  }

  _xcp_verbose("Data netto size in bytes first ODT: %i\n",dsizeodtfirst);
  _xcp_verbose("Data netto size in bytes 1+n ODT: %i\n",dsizeodtsecond);

  //check sizes
  if((dsizeodtfirst <= 0) || (dsizeodtsecond <= 0)){
    _xcp_error("DAQ netto data segment size is zero\n");
    data->lerror = XCP_ERROR_DAQ_PROCESS;
    return CS_FAIL;
  }

  //We will build up a structure

  //helperstructure to reorder signals on ODT level
  struct odtdata{
    int num_entry; //Number of signals to register
    int used;      //Used bytes
  };
  //helperstructure to reorder signals on DAQ level
  struct{
    struct odtdata odt[maxodt]; //ODT storage structure
    int num_odt;   //Number of ODTs to register
    int prescaler; //Prescaler for DAQ
    int ev_idx;    //Event index
  }daq[data->xcpa2l.daq_proc.daq_max];

  //Init structure
  for(i = 0; i< data->xcpa2l.daq_proc.daq_max; i++){
    for(j = 0; j < maxodt; j++){
      daq[i].odt[j].used = 0;
      daq[i].odt[j].num_entry = 0;
    }
    daq[i].prescaler = -1;
    daq[i].num_odt = 0;
    daq[i].ev_idx = -1;
  }

  //Start working on signals
  for(i = 0; i < data->xcpa2l.a2l.signalcounter; i++){
    daq_idx = -1;
    //First search for existing daq
    for(j = 0; j < daqcounter; j++){
      if((daq[j].ev_idx == data->xcpa2l.a2l.signals[i].ev_idx) && 
	 (daq[j].prescaler == data->xcpa2l.a2l.signals[i].prescaler)){
	daq_idx = j;
	break;
      }
    }

    if(daqcounter >= data->xcpa2l.daq_proc.daq_max){
      _xcp_error("Can not find DAQ for signal\n");
      data->lerror = XCP_ERROR_DAQ_PROCESS;
      return CS_FAIL;
    }

    if(daq_idx == -1){
      daq_idx = daqcounter;
      daqcounter++;
    }    

    //Now search for odt
    for(j = 0; j < maxodt; j++){
      flag_insert = 0;
      if((j == 0) && ((daq[daq_idx].odt[j].used + data->xcpa2l.a2l.signals[i].size * cs_a2l_get_ag_bytes(data->xcpa2l.commode.AG)) <= dsizeodtfirst)){
	flag_insert = 1;
      }
      if((j != 0) && ((daq[daq_idx].odt[j].used + data->xcpa2l.a2l.signals[i].size * cs_a2l_get_ag_bytes(data->xcpa2l.commode.AG)) <= dsizeodtsecond)){
	flag_insert = 1;
      }

      if(!flag_insert){
	//No space in actual ODT, try next ODT
	continue;
      }

      //Increment used odt, if we use a new one
      if(daq[daq_idx].odt[j].used == 0)
	daq[daq_idx].num_odt++;

      //Register data
      daq[daq_idx].ev_idx = data->xcpa2l.a2l.signals[i].ev_idx;
      daq[daq_idx].prescaler = data->xcpa2l.a2l.signals[i].prescaler;
      data->xcpa2l.a2l.signals[i].daq_idx = daq_idx;
      data->xcpa2l.a2l.signals[i].odt_idx = j;
      data->xcpa2l.a2l.signals[i].entry_idx = daq[daq_idx].odt[j].num_entry;
      daq[daq_idx].odt[j].num_entry++;
      daq[daq_idx].odt[j].used += data->xcpa2l.a2l.signals[i].size * cs_a2l_get_ag_bytes(data->xcpa2l.commode.AG);

      _xcp_verbose("Register Signal %s:\n",data->xcpa2l.a2l.signals[i].name);
      _xcp_verbose("  Signal size in Bytes %i\n",data->xcpa2l.a2l.signals[i].size * cs_a2l_get_ag_bytes(data->xcpa2l.commode.AG));
      _xcp_verbose("  Event %s \n",data->xcpa2l.daq_proc.event[data->xcpa2l.a2l.signals[i].ev_idx].name);
      _xcp_verbose("  Prescaler %i\n",data->xcpa2l.a2l.signals[i].prescaler);
      _xcp_verbose("  DAQ idx %i\n",data->xcpa2l.a2l.signals[i].daq_idx);
      _xcp_verbose("  ODT idx %i\n",data->xcpa2l.a2l.signals[i].odt_idx);
      _xcp_verbose("  Entry idx %i\n",data->xcpa2l.a2l.signals[i].entry_idx);
      _xcp_verbose("  ODT num entries %i\n",daq[daq_idx].odt[j].num_entry);
      _xcp_verbose("  ODT bytes used %i\n",daq[daq_idx].odt[j].used);
      break;
    }

    if(!flag_insert){
      _xcp_error("Can not find ODT for signal\n");
      data->lerror = XCP_ERROR_DAQ_PROCESS;
      return CS_FAIL;
    }

  }

  //Free DAQ lists
  _xcp_verbose("Free DAQ\n");
  if(cs_xcp_cmd_free_daq(data)){
    return CS_FAIL;
  }
  cs_ts_sleep(data->stim);

  //Alloc DAQ lists
  _xcp_verbose("Allocate DAQ, %i\n",daqcounter);
  if(cs_xcp_cmd_alloc_daq(data,daqcounter)){
    _xcp_error("Failed to alloc DAQ lists\n");
    return CS_FAIL;
  }
  cs_ts_sleep(data->stim);

  for(i = 0; i < daqcounter; i++){
    //Configure DAQ lists
    _xcp_verbose("Configure DAQ list, DAQ %i, Event %i, Direction %i, Prescaler %i, Timestamp %s\n",
		 i, daq[i].ev_idx, XCP_DAQ_DIR_DAQ, daq[i].prescaler, data->xcpa2l.daq_proc.ts_supported?"supported":"unsupported");
    if(cs_xcp_cmd_set_daq_list_mode(data,i,daq[i].ev_idx,XCP_DAQ_DIR_DAQ,
				    0, //always with PID
				    data->xcpa2l.daq_proc.ts_supported, //enable timestamp if supported
				    daq[i].prescaler,
				    0 //always lowest priority
				    )){
      _xcp_error("Failed to configure DAQ list\n");
      return CS_FAIL;
    }
    cs_ts_sleep(data->stim);
    //Alloc ODT in DAQ
    _xcp_verbose("Allocate ODT, DAQ %i, num ODT %i\n",i,daq[i].num_odt);
    if(cs_xcp_cmd_alloc_odt(data,i,daq[i].num_odt)){
      _xcp_error("Failed to alloc ODT\n");
      return CS_FAIL;
    }
    cs_ts_sleep(data->stim);
    //Alloc ODT Entry
    for(j = 0; j < daq[i].num_odt; j++){
      _xcp_verbose("Allocate ODT Entry, DAQ %i, ODT %i, num Entries %i\n",i,j,daq[i].odt[j].num_entry);
      if(cs_xcp_cmd_alloc_odt_entry(data,i,j,daq[i].odt[j].num_entry)){
	_xcp_error("Failed to alloc ODT Entry\n");
	return CS_FAIL;
      }
      cs_ts_sleep(data->stim);
    }
  }

  //Register signals
  for(i = 0; i < data->xcpa2l.a2l.signalcounter; i++){
    //Set working pointer
    _xcp_verbose("Set DAQ PTR, DAQ %i, ODT %i, Entry %i\n",
		 data->xcpa2l.a2l.signals[i].daq_idx,
		 data->xcpa2l.a2l.signals[i].odt_idx,
		 data->xcpa2l.a2l.signals[i].entry_idx);
    if(cs_xcp_cmd_set_daq_ptr(data,
			      data->xcpa2l.a2l.signals[i].daq_idx,
			      data->xcpa2l.a2l.signals[i].odt_idx,
			      data->xcpa2l.a2l.signals[i].entry_idx)){ 
      //always first entry
      _xcp_error("Failed to set DAQ pointer\n");
      return CS_FAIL;
    }
    cs_ts_sleep(data->stim);
    //Store memory addr in ODT Entry
    _xcp_verbose("Write DAQ, Bitoffset: 0x%X, Size %i, Extension 0x%X, Address 0x%X\n",
		 data->xcpa2l.a2l.signals[i].bitoffset,
		 data->xcpa2l.a2l.signals[i].size,
		 data->xcpa2l.a2l.signals[i].extension,
		 data->xcpa2l.a2l.signals[i].address);
    if(cs_xcp_cmd_write_daq(data,
			    data->xcpa2l.a2l.signals[i].bitoffset,
			    data->xcpa2l.a2l.signals[i].size,
			    data->xcpa2l.a2l.signals[i].extension,
			    data->xcpa2l.a2l.signals[i].address)){
      _xcp_error("Failed to write DAQ\n");
      return CS_FAIL;
    }
    cs_ts_sleep(data->stim);
    //Register feedback values
    if(cs_xcp_entry_reg_value(data,
			      data->xcpa2l.a2l.signals[i].daq_idx,
			      data->xcpa2l.a2l.signals[i].odt_idx,
			      data->xcpa2l.a2l.signals[i].entry_idx,
			      data->xcpa2l.a2l.signals[i].exvalue,
			      data->xcpa2l.a2l.signals[i].cm,
			      data->xcpa2l.a2l.signals[i].dtype,
			      data->xcpa2l.a2l.signals[i].exts,
			      data->xcpa2l.a2l.signals[i].extcallback,
			      data->xcpa2l.a2l.signals[i].extcallbackpriv
			      )){
      _xcp_error("Failed to register feedback\n");
      return CS_FAIL;
    }
  }

  return CS_OK;
}

/**************************************************************************/
int cs_xcp_api_startDAQ(
			cs_xcp_t *data
			)
/**************************************************************************/
{
  int i;

  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }

  //Select DAQ list
  for(i = 0; i < data->daqlist.no_daq; i++){
    if(cs_xcp_cmd_start_stop_daq_list(data,i,1)){
      _xcp_error("Failed to start DAQ list\n");
      return CS_FAIL;
    }
    cs_ts_sleep(data->stim);
  }

  return CS_OK;
}

/**************************************************************************/
int cs_xcp_api_stopDAQ(
		       cs_xcp_t *data
		       )
/**************************************************************************/
{
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }

  //Synch start
  if(cs_xcp_cmd_start_stop_synch(data,0)){
    _xcp_error("Failed to stop DAQ lists\n");
    return CS_FAIL;
  }

  return CS_OK;
}

/**************************************************************************/
int cs_xcp_api_process(
		       cs_xcp_t *data
		       )
/**************************************************************************/
{
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
  
  if(cs_xcp_receive(data)){
    _xcp_error("Failed to receive data\n");
    return CS_FAIL;
  }
  return CS_OK;
}

/**************************************************************************/
int cs_xcp_api_setDAQTimeout(
			     cs_xcp_t *data,
			     double fac
			     )
/**************************************************************************/
{
  if(data == NULL)
    return CS_FAIL;
    
  data->daq_fac_recvtimeout = fac;
    
  return CS_OK;
}

/**************************************************************************/
int cs_xcp_api_hasDAQTimeout(
			     cs_xcp_t *data
			     )
/**************************************************************************/
{
  if(data == NULL){
    _xcp_error("Parameter failure\n");
    return CS_FAIL;
  }
  //Check for timeouts
  if(cs_xcp_daq_istimeout(data,data->daq_fac_recvtimeout)){
    _xcp_debug("DAQ Timeout detected\n");
    return CS_FAIL;
  }
  return CS_OK;
}
