/*  XCP CAN Stack for SocketCAN
    Copyright (C) 2013 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <string.h>

#include <inttypes.h>
#include <stdlib.h>
#include "config.h"
#include "cs_defines.h"
#include "cs_socketcan.h"
#include "cs_xcp_can.h"
#include "cs_xcp_private.h"


/** \struct cs_xcp_can_t
    CAN Interface data storage structure
*/
struct cs_xcp_can{
  int version;                     /**< Layer version*/
  //  int baudrate;                /**< Baudrat*/
  char name[CS_XCP_CAN_CHARBUF];   /*Interface name*/
  cs_socketcan_t *socket;          /**< Socket handle*/
  uint32_t BCST_id;                /**< Broadcast CAN id*/
  uint32_t CMD_id;                 /**< Default Cmd target CAN id*/
  uint32_t RES_id;                 /**< Default Resp CAN id*/
  uint32_t DAQ_id[CS_XCP_CAN_MAXDAQ];   /**< Optional DAQ CAN id*/
  uint8_t DAQ_id_mode[CS_XCP_CAN_MAXDAQ];   /**< DAQ CAN id mode, 0=>variable, 1=> fixed*/
  cs_xcp_daq_dir_t DAQ_dir[CS_XCP_CAN_MAXDAQ]; /** DAQ direction*/
  int max_dlc;                     /**< Max DLC*/
  uint8_t max_dlc_required;         /**< Max DLC required, Vector Spec? Maybe fill up frame up to max DLC*/
};

/*************************************************************************************************/
void _cs_xcp_can_reset(cs_xcp_can_t *iface)
/*************************************************************************************************/
{
  int i;
  if(iface == NULL)
    return;


  memset(iface,0,sizeof(cs_xcp_can_t));


  for(i = 0; i < CS_XCP_CAN_MAXDAQ; i++){
    iface->DAQ_id[i] = CS_XCP_CAN_UNDEF;
  }

}

/*************************************************************************************************/
cs_xcp_can_t *cs_xcp_can_init()
/*************************************************************************************************/
{
  cs_xcp_can_t *rv = NULL;

  rv = (cs_xcp_can_t*)malloc(sizeof(cs_xcp_can_t));

  if(rv == NULL)
    return NULL;

  rv->socket = cs_socketcan_init();

  if(rv->socket == NULL){
    free(rv);
    return NULL;
  }
  
  _cs_xcp_can_reset(rv);

  return rv;
}

/*************************************************************************************************/
int cs_xcp_can_configure(cs_xcp_can_t *iface, uint32_t cmd_id, uint32_t res_id, uint32_t bcast_id)
/*************************************************************************************************/
{
  if(iface == NULL)
    return CS_FAIL;

  _cs_xcp_can_reset(iface);

  //Store
  iface->BCST_id = bcast_id; 
  iface->CMD_id = cmd_id;
  iface->RES_id = res_id;
  iface->max_dlc = CS_XCP_CAN_DLC;

  return CS_OK;
}

/*************************************************************************************************/
int cs_xcp_can_getVersion(cs_xcp_can_t *iface)
/*************************************************************************************************/
{
  if(iface == NULL)
    return 0;

  return iface->version;
}

/*************************************************************************************************/
void cs_xcp_can_setVersion(cs_xcp_can_t *iface, int version)
/*************************************************************************************************/
{
  if(iface == NULL)
    return ;

  iface->version = version;
}

/*************************************************************************************************/
uint32_t cs_xcp_can_getRespId(cs_xcp_can_t *iface)
/*************************************************************************************************/
{
  if(iface == NULL)
    return 0;

  return iface->RES_id;
}

/*************************************************************************************************/
void cs_xcp_can_setRespId(cs_xcp_can_t *iface, uint32_t id)
/*************************************************************************************************/
{
  if(iface == NULL)
    return;

  iface->RES_id = id;
}

/*************************************************************************************************/
uint32_t cs_xcp_can_getCmdId(cs_xcp_can_t *iface)
/*************************************************************************************************/
{
  if(iface == NULL)
    return 0;

  return iface->CMD_id;
}

/*************************************************************************************************/
void cs_xcp_can_setCmdId(cs_xcp_can_t *iface, uint32_t id)
/*************************************************************************************************/
{
  if(iface == NULL)
    return;

  iface->CMD_id = id;
}

/*************************************************************************************************/
uint32_t cs_xcp_can_getBcastId(cs_xcp_can_t *iface)
/*************************************************************************************************/
{
  if(iface == NULL)
    return 0;

  return iface->BCST_id;
}

/*************************************************************************************************/
char* cs_xcp_can_getName(cs_xcp_can_t *iface)
/*************************************************************************************************/
{
  if(iface == NULL)
    return NULL;

  return iface->name;
}

/*************************************************************************************************/
void cs_xcp_can_setBcastId(cs_xcp_can_t *iface, uint32_t id)
/*************************************************************************************************/
{
  if(iface == NULL)
    return;

  iface->BCST_id = id;
}
  
/*************************************************************************************************/
int cs_xcp_can_open(cs_xcp_can_t *iface, char *ifcan)
/*************************************************************************************************/
{
  if((iface == NULL) ||
     ifcan == NULL)
    return CS_FAIL;

  strncpy(iface->name,ifcan,CS_XCP_CAN_CHARBUF);

  //Open raw socket
  if(cs_socketcan_config_native(iface->socket, iface->name,0)){
    _xcp_error("Failed to open socket: %s\n",iface->name);
    return CS_FAIL;
  }


  //Set only response and bcast filter id
  if(cs_xcp_can_setdaqid(iface, 0, CS_XCP_CAN_UNDEF, XCP_DAQ_DIR_NONE)){
    _xcp_error("Failed to set filter id\n");
    return CS_FAIL;
  }

  if(cs_socketcan_open(iface->socket)){
    _xcp_error("Failed to open interface\n");
    return CS_FAIL;
  }
  
  return CS_OK;
}

/*************************************************************************************************/
int cs_xcp_can_close(cs_xcp_can_t *iface)
/*************************************************************************************************/
{
  if(iface == NULL)
    return CS_FAIL;
  
  if(iface->socket){
    if(cs_socketcan_close(iface->socket)){
      _xcp_error("Failed to close socket\n");
      return CS_FAIL;
    }
  }
  
  _cs_xcp_can_reset(iface);

  cs_socketcan_free(iface->socket);
  
  free(iface);

  return CS_OK;
}

/*************************************************************************************************/
int cs_xcp_can_isConnected(cs_xcp_can_t *iface)
/*************************************************************************************************/
{
  if(iface == NULL)
    return CS_FAIL;
  
  if(cs_socketcan_isconnected(iface->socket)){
    return CS_OK;
  }
  
  return CS_FAIL;
}

/*************************************************************************************************/
int cs_xcp_can_setdaqid(cs_xcp_can_t *iface, int daq, uint32_t daq_id, cs_xcp_daq_dir_t dir)
/*************************************************************************************************/
{
  //TODO 29 bit id
  uint32_t filter[CS_XCP_CAN_MAXDAQ];
  int i;
  int filtercounter = 1;

  if(iface == NULL)
    return CS_FAIL;

  if(daq >= CS_XCP_CAN_CONFDAQ){
    _xcp_error("DAQ Id out of range\n");
    return CS_FAIL;
  }

  if(daq_id != CS_XCP_CAN_UNDEF){
    iface->DAQ_id[daq] = daq_id;
    iface->DAQ_dir[daq] = dir;
    if(dir != XCP_DAQ_DIR_DAQ)
      return CS_OK;
  }

  //Build up new filter
  filter[0] = (uint32_t)iface->RES_id;

  //if broadcast id is 0 (undefined) set filter to RES_id
  if(iface->BCST_id != 0){
    filtercounter++;
    filter[1] = (uint32_t)iface->BCST_id;
  }

  for(i = 0; i < CS_XCP_CAN_MAXDAQ; i++){
    if((iface->DAQ_id[i] != CS_XCP_CAN_UNDEF) &&
       (iface->DAQ_id[i] != iface->RES_id) &&
       (iface->DAQ_dir[i] == XCP_DAQ_DIR_DAQ)){
      filter[filtercounter] = (uint32_t)iface->DAQ_id[i];
      filtercounter++;
    }
  }

  if(cs_socketcan_set_filter(iface->socket,filter,filtercounter)){
    _xcp_error("Failed to create filter\n");
    return CS_FAIL;
  } 

  return CS_OK;
}


/*************************************************************************************************/
int cs_xcp_can_send(cs_xcp_can_t *iface, uint32_t id, uint8_t *data, int dl)
/*************************************************************************************************/
{
  uint8_t ldata[CS_XCP_CAN_DLC];
  int rv;

  if((iface == NULL) || 
     (data == NULL))
    return CS_FAIL;

  if((dl >  CS_XCP_CAN_DLC) || 
     (iface->max_dlc >  CS_XCP_CAN_DLC) ||
     (dl > iface->max_dlc)){
    _xcp_error("Invalid Interface data length %i, max_dlc %i\n",dl,iface->max_dlc);
    return CS_FAIL;
  }

  memset(ldata,0,CS_XCP_CAN_DLC);
  memcpy(ldata,data,iface->max_dlc);
  _xcp_verbose("Send buffer\n");
  _xcp_verbose_buffer(ldata,iface->max_dlc);
  rv = cs_socketcan_send(iface->socket,id,ldata,iface->max_dlc,SOCKETCAN_SEND_CAN);

  switch(rv){
  case CS_SOCKETCAN_FAIL:
    return CS_FAIL;
    break;
  case CS_SOCKETCAN_OK:
    return CS_OK;
    break;
  case CS_SOCKETCAN_TXFULL:
    return CS_XCP_CAN_TXFULL;
    break;
  default:
    return CS_FAIL;
  }
  return CS_OK;
}



/*************************************************************************************************/
int cs_xcp_can_recv(cs_xcp_can_t *iface, uint32_t *id, uint8_t *data, int *dl, uint32_t timeout, struct timeval *tv_recv)
/*************************************************************************************************/
{

  uint8_t buffer[CS_XCP_CAN_DLC];
  int len = CS_XCP_CAN_DLC;
  uint32_t canid;
  int rv;
  int loop = 1;

  if((iface == NULL) || 
     (data == NULL) || 
     (id == NULL) ||
     (tv_recv == NULL) || 
     (dl == NULL))
    return CS_FAIL;

  if(*dl < CS_XCP_CAN_DLC){
    _xcp_error("Invalid received buffer length:  %i\n",*dl);
    return CS_FAIL;
  }

  _xcp_verbose("XCP CAN Receive blocking\n");
  memset(buffer,0,sizeof(uint8_t)*len);

  while(loop){
    rv = cs_socketcan_recv_block(iface->socket,&canid,buffer,&len,tv_recv,timeout);
    switch(rv){
    case CS_SOCKETCAN_OK:
      loop = 0;
      break;
    case CS_SOCKETCAN_FAIL:
      loop = 0;
      return CS_FAIL;
      break;
    case CS_SOCKETCAN_RTR:
    case CS_SOCKETCAN_ERROR:
      //continue recv
      loop++;
      break;
    default:
      return CS_FAIL;
    }
    if(loop > 10)
      return CS_FAIL;
  }

  //Get ID
  *id = canid;
  //copy buffer
  memcpy(data,buffer,CS_XCP_CAN_DLC);
  //set length
  *dl = len;
  _xcp_verbose("Received buffer, CAN id hex:%X, Data length  %i\n",canid,*dl);
  _xcp_verbose_buffer(data,*dl);
  return CS_OK;
}

/*************************************************************************************************/
int cs_xcp_can_recv_nonblock(cs_xcp_can_t *iface, uint32_t *id, uint8_t *data, int *dl, struct timeval *tv_recv)
/*************************************************************************************************/
{

  uint8_t buffer[CS_XCP_CAN_DLC];
  int len = CS_XCP_CAN_DLC;
  uint32_t canid;

  if((iface == NULL) || 
     (data == NULL) || 
     (id == NULL) ||
     (tv_recv == NULL) || 
     (dl == NULL))
    return CS_FAIL;

  _xcp_verbose("XCP CAN Receive non blocking\n");

  if(*dl < CS_XCP_CAN_DLC){
    _xcp_error("Invalid received buffer length:  %i\n",*dl);
    return CS_FAIL;
  }

  memset(buffer,0,sizeof(uint8_t)*len);

  if(!cs_socketcan_recv_nonblock(iface->socket,&canid,buffer,&len,tv_recv)){
    //Check ID
    *id = canid;
    
    //copy buffer
    memcpy(data,buffer,CS_XCP_CAN_DLC);
    //set length
    *dl = len;
    _xcp_verbose("Received buffer, CAN id hex:%X\n",canid);
    _xcp_verbose_buffer(data,*dl);
    return CS_OK;
  }
  
  //Timeout or error

  return CS_FAIL;
}
