/*----------------------------------------------------------------------------
| File:
|   applmain.c
|
| Project:
|   CCP driver example
|   CANape Calibration Tool
|
| Description:
|   Mini CCP application for C16x
|
|-----------------------------------------------------------------------------
| Copyright (c) 1998 by Vector Informatik GmbH.  All rights reserved.
 ----------------------------------------------------------------------------*/
#include <stdio.h>
#include <unistd.h>
#include <getopt.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "ccp.h"
#include "ecu.h"
#include "applmain.h"
#include "cs_socketcan.h"
#include "cs_ccp_a2l.h"

//------------------------------------------------------------------------------
// Global
#define MEMSIMSIZE 0xFF

unsigned long gTimer;
unsigned long gCounter;
cs_socketcan_t *cansocket = NULL;
uint32_t dto_id;
uint32_t crm_id;
unsigned char gDebugLevel = 3;
unsigned short CCP_STATION_ADDR;
uint8_t *memsim;

// Current Calibration page
CCP_BYTE ccpCalPage = 0; // Default ROM


void help(void)
{
  printf(
          "Usage: ccpsim [OPTION]\n"
          "Test CCP slave stack.\n"
          "\n"
          "Options:\n"
          "  -i, --iface    CAN-Interface\n"
	  "  -a, --a2l      A2L file to use\n"
	  "  -p, --project  Project name\n"
	  "  -m, --module   Module name\n"
          "      --help     display this help and exit\n");
}

/********************************************************************/
void CCPPRINT(const char *fmt, ...)
/********************************************************************/
{
  char buffer[255];
  va_list ap;
  memset(buffer,0,sizeof(buffer));
  va_start(ap, fmt);
  vsnprintf(buffer,255, fmt, ap);
  va_end(ap);
  printf(buffer);
}

/*----------------------------------------------------------------------------*/
/* Transmit the CCP message */
/* Id is CCP_DTO_ID, which is configured at compile time in CAN.C */

void ccpSend( CCP_BYTEPTR msg )
{
  if(cs_socketcan_send(cansocket,dto_id,(uint8_t*)msg,8,SOCKETCAN_SEND_CAN)){
    printf("Failed to send msg\n");
  }
}


/*----------------------------------------------------------------------------*/
/* Convert a memory address from CCP 8/32bit into a C pointer */

CCP_MTABYTEPTR ccpGetPointer( CCP_BYTE addr_ext, CCP_DWORD addr )
{
  //printf("Acces pointer 0x%lX , Extension 0x%x\n",addr,addr_ext);
  return (CCP_MTABYTEPTR) &memsim[0];
}


/*----------------------------------------------------------------------------*/
// CCP Callbacks

void ccpUserBackground( void )
{
}

CCP_BYTE ccpDisableNormalOperation( CCP_MTABYTEPTR a, CCP_WORD s )
{
  return 1;
}


/*----------------------------------------------------------------------------*/
/* Calibration RAM/ROM Selection */


CCP_DWORD ccpGetCalPage( void )
{
  return (CCP_DWORD)ccpCalPage;
}

void ccpSetCalPage( CCP_DWORD a )
{
  ccpCalPage = (CCP_BYTE)a;

}

void ccpInitCalPage( void ) {

    #define CALROM_ADDR 0x14000
    #define CALRAM_ADDR 0x44000
    unsigned char *p1 = (unsigned char *)CALROM_ADDR;
    unsigned char *p2 = (unsigned char *)CALRAM_ADDR;
    unsigned int i;
    for (i=0;i<0x4000;i++) {
      *p2++ = *p1++;
    }
}


//------------------------------------------------------------------------------
// 10ms Timer Isr

void timer_function( void ) {

  // ms Timer
  gTimer += 1;

  // ECU Simulation
  ecuCyclic();

  ccpDaq(0);
  ccpDaq(1);
  // 10 ms
  // Data Acquisition on Channel 2
  ccpDaq(2);

}


//------------------------------------------------------------------------------
// Main Loop

int main(int argc, char **argv)
{
  uint32_t id;
  uint8_t buffer[8];
  int len;
  struct timeval txRqTime;
  int c;
  int i;
  cs_ccp_a2l_t a2l;
  char * filename = NULL;
  char * project = NULL;
  char * module = NULL;
  char * iface = NULL;

  while (1) {
    static struct option long_options[] = {
      /* These options set a flag. */
      {"iface", required_argument, NULL, 'i'},
      {"a2l", required_argument, NULL, 'a'},
      {"project", required_argument, NULL, 'p'},
      {"module", required_argument, NULL, 'm'},
      {"help",    no_argument, NULL, 'h'},
      {0, 0, 0, 0}
    };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    c = getopt_long (argc, argv, "i:ha:p:m:",
                     long_options, &option_index);

    /* Detect the end of the options. */
    if (c == -1)
      break;

    switch (c) {
    case 0:
      /* If this option set a flag, do nothing else now. */
      if (long_options[option_index].flag != 0)
        break;
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
      break;
    case 'i': iface = optarg;     break;
    case 'a': filename = optarg;     break;
    case 'p': project = optarg;     break;
    case 'm': module = optarg;     break;
    case 'h': help(); exit(0);    break;
    case '?':
      /* getopt_long already printed an error message. */
      printf( "Typ `ccpsim --help' for more information\n");
      return -1;
      break;
    default:
      return -1;
      break;
    }
  }

  /*Check inputs*/
  if((iface==NULL)||(filename == NULL)||(project == NULL)||(module == NULL)){
    printf( "Please check input arguments\n");
    printf( "Typ `ccpsim --help' for more information\n");
    return -1;
  }

  //Try to open file
  if(cs_ccp_a2l_open(&a2l,filename,project,module,"ASAP1B_CCP",NULL)){
    printf("Failed to open file %s\n",filename);
    return -1;
  }
  dto_id  = a2l.can_recv_id;
  crm_id  = a2l.can_send_id;
  CCP_STATION_ADDR = a2l.station_id;
  // Initialize global variables
  gTimer = 0;
  gCounter = 0;

  // Initialize CAN driver
  if((cansocket = cs_socketcan_init(cansocket)) == NULL){
    printf("Failed to init socket: %s\n",iface);
    return -1;
  }

  if(cs_socketcan_config_native(cansocket, iface,0)){
    printf("Failed to configure socket: %s\n",iface);
    return -1;
  }
  
  if(cs_socketcan_open(cansocket)){
    printf("Failed to open socket: %s\n",iface);
    return -1;
  }
  printf("Recv filter CAN id 0x%X\n",crm_id);
  if(cs_socketcan_set_filter(cansocket,&crm_id,1)){
    printf("Failed to create filter\n");
    return -1;
  }

  memsim = (uint8_t*)malloc(MEMSIMSIZE);
  
  if(memsim == NULL){
    printf("Malloc failure\n");
    return -1;
  }

  //Init Memory simulation
  for(i = 0; i< MEMSIMSIZE; i++)
    memsim[i] = i;

  // initialize Calibration RAM
  //ccpInitCalPage();

  // Initialize the ECU sample
  ecuInit();

  // Initialize CCP driver
  ccpInit();

  // Mainloop
  for (;;) {

    gCounter++;

    // Check for incomming CCP receive message (CRO)
    len = 8;
    //read and block for 1ms
    if(!cs_socketcan_recv_block(cansocket,&id,buffer,&len,&txRqTime,1)){
      //printf("Recieve frame\n");
      ccpCommand(buffer);
    }

    timer_function();

    // Check for pending CCP transmit messages (DTO)
    ccpSendCallBack();



    // Perform any CCP background processing
    ccpBackground();
    //if((gCounter%1000)  == 0)
    //printf("Loop %lu\n",gCounter);
  } // Mainloop

  cs_socketcan_close(cansocket);
  cs_socketcan_free(cansocket);
  
  return 0;
}



