/*  
    Copyright (C) 2014 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @example main_a2l.c
 * @author Andreas Stewering-Bone
 * @date 9 Oct 2014
 * @brief Simple A2L file parsing application for testing the A2L parsing API.
 *
 * Simple A2L file parsing application for testing the A2L parsing API
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <time.h>

#include "cs_a2l.h"


void help(void)
{
  printf(
          "Usage: testa2l [OPTION]\n"
          "Test A2L parser stack.\n"
          "\n"
          "Options:\n"
          "  -a, --a2l        A2L input file\n"
	  "  -p, --project    Project to search for\n"
	  "  -i, --interface  Interface to search for\n"
	  "  -m, --module     Module to use\n"
	  "  -o, --option     Optiona section to use \n"
          "      --help       display this help and exit\n");
}

int main(int argc, char **argv)
{
  int c;
  char * filename = NULL;
  char * project = NULL;
  char * module = NULL;
  char * interface = NULL;
  char * option = NULL;
  cs_a2l_t data;
  int asap2major = 0;
  int asap2minor = 0;
  cs_parser_result_t result;
  cs_a2l_list_t list;
  cs_a2l_measurement_t me;
  cs_a2l_characteristic_t ch;
  int i;

  while (1) {
    static struct option long_options[] = {
      /* These options set a flag. */
      {"a2l", required_argument, NULL, 'a'},
      {"project", required_argument, NULL, 'p'},
      {"module", required_argument, NULL, 'm'},
      {"interface", required_argument, NULL, 'i'},
      {"option", required_argument, NULL, 'o'},
      {"help",    no_argument, NULL, 'h'},
      {0, 0, 0, 0}
    };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    c = getopt_long (argc, argv, "a:p:m:i:o:h",
                     long_options, &option_index);

    /* Detect the end of the options. */
    if (c == -1)
      break;

    switch (c) {
    case 0:
      /* If this option set a flag, do nothing else now. */
      if (long_options[option_index].flag != 0)
        break;
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
      break;
    case 'a': filename = optarg;     break;
    case 'p': project = optarg;     break;
    case 'm': module = optarg;     break;
    case 'i': interface = optarg;     break;
    case 'o': option = optarg;     break;
    case 'h': help(); exit(0);    break;
    case '?':
      /* getopt_long already printed an error message. */
      printf( "Typ `testxcpa2l --help' for more information\n");
      return -1;
      break;
    default:
      return -1;
      break;
    }
  }

  

  //First check Library version
  if(CS_A2L_CHECKVERSION){
    printf("Incompatible Library interface, Application was compiled with version %i, Library has version %i\n",
	   CS_A2L_VERSION_MAGIC,cs_a2l_libgetversion());
    return -1;
  }

  /*Check inputs*/
  if((filename==NULL)){
    printf( "Please check input arguments\n");
    printf( "Typ `testa2l --help' for more information\n");
    return -1;
  }


  //Try to open file
  if(cs_a2l_open(&data,filename)){
    printf("Failed to open file %s\n",filename);
    return -1;
  }

  //Get ASAP-Version
  if(cs_a2l_getversion(&data,&asap2major,&asap2minor)){
    printf("Failed to get ASAP2 version\n");
    return -1;
  }
  printf("ASAP2 Version Major %i Minor %i\n",asap2major,asap2minor);

  //Try to find a section
  result.searchin = (char*)data.buffer;
  result.searchlen = data.filelen; 
  if(cs_a2l_search_section(&result,"PROJECT",NULL)){
    printf("Cannot find section PROJECT\n");
    return -1;
  }

  //Set serach area
  result.searchin = (char*)data.buffer;
  result.searchlen = data.filelen; 
  if(cs_a2l_list_sections(&result,&list,"PROJECT")){
    printf("Failed to build up section list\n");
    cs_a2l_list_free(&list);
    return -1;
  }
  printf("Available PROJECTs:\n");
  cs_a2l_printlist(&list);
  cs_a2l_list_free(&list);

  if(project == NULL){
    //Close file
    if(cs_a2l_close(&data)){
      printf("Failed to close file\n");
      return -1;
    }
    return 0;
  }
  //Get Module
  if(cs_a2l_GetProject(&data,project)){
    printf("Cannot get Project %s\n",project);
    return -1;
  }

  //Set serach area
  result.searchin = (char*)data.buffer;
  result.searchlen = data.filelen; 
  if(cs_a2l_list_sections(&result,&list,"MODULE")){
    printf("Failed to build up section list\n");
    cs_a2l_list_free(&list);
    return -1;
  }
  printf("Available MODULES:\n");
  cs_a2l_printlist(&list);
  cs_a2l_list_free(&list);

  if(module == NULL){
    //Close file
    if(cs_a2l_close(&data)){
      printf("Failed to close file\n");
      return -1;
    }
    return 0;
  }

  //Get Module
  if(cs_a2l_GetModule(&data,module)){
    printf("Cannot get Module\n");
    return -1;
  }
  
  //Get common parameter if exists
  if(cs_a2l_GetModCommon(&data)){
    printf("Cannot handle common data\n");
  }

  //Set serach area
  result = data.module;
  if(cs_a2l_list_sections(&result,&list,"IF_DATA")){
    printf("Failed to build up section list\n");
    cs_a2l_list_free(&list);
    return -1;
  }
  printf("Available Interfaces:\n");
  cs_a2l_printlist(&list);
  cs_a2l_list_free(&list);  


  if(interface == NULL){
    //Close file
    if(cs_a2l_close(&data)){
      printf("Failed to close file\n");
      return -1;
    }
    return 0;
  }

  if(cs_a2l_GetInterface(&data,interface)){
    printf("Cannot get Interface\n");
    return -1;
  }

  //Set serach area
  result.searchin = (char*)data.buffer;
  result.searchlen = data.filelen; 
  if(cs_a2l_list_sections(&result,&list,"CHARACTERISTIC")){
    printf("Failed to build up section list\n");
    cs_a2l_list_free(&list);
    return -1;
  }
  for(i = 0; i< list.no_result; i++){
    if(cs_a2l_GetCharacteristic(&data,&ch,list.resultname[i])){
      printf("Failed to parse Characteristic %s\n",list.resultname[i]);
    } else{
      //printf("Success parsing Characteristic %s\n",list.resultname[i]);
      cs_a2l_print_characteristic(&ch);
    }
  }
  cs_a2l_printlist(&list);
  cs_a2l_list_free(&list);

  //Set serach area
  result.searchin = (char*)data.buffer;
  result.searchlen = data.filelen; 
  if(cs_a2l_list_sections(&result,&list,"MODULE")){
    cs_a2l_list_free(&list);
    printf("Failed to build up section list\n");
    return -1;
  }
  cs_a2l_printlist(&list);
  cs_a2l_list_free(&list);
  
  //Set serach area
  result.searchin = (char*)data.buffer;
  result.searchlen = data.filelen; 
  if(cs_a2l_list_sections(&result,&list,"MEASUREMENT")){
    printf("Failed to build up section list\n");
    cs_a2l_list_free(&list);
    return -1;
  }
  //Parse signals
  for(i = 0; i< list.no_result; i++){
    if(cs_a2l_GetMeasurement(&data,&me,list.resultname[i])){
      printf("Failed to parse signal %s\n",list.resultname[i]);
    } else{
      //printf("Success parsing signal %s\n",list.resultname[i]);
      cs_a2l_print_measurement(&me);
    }
  }
  cs_a2l_printlist(&list);
  cs_a2l_list_free(&list);

  //Set serach area
  result = data.module;
  if(cs_a2l_list_sections(&result,&list,"MOD_PAR")){
    printf("Failed to build up section list\n");
    cs_a2l_list_free(&list);
    return -1;
  }
  printf("Available MOD_PAR:\n");
  cs_a2l_printlist(&list);
  cs_a2l_list_free(&list);  

  if(option != NULL){
    if(cs_a2l_GetModPar(&data,option)){
      printf("Failed to get MOD_PAR Section %s\n",option);
      return -1;
    }
    cs_a2l_print_modpar(&data.modpar);
  }

  //Close file
  if(cs_a2l_close(&data)){
    printf("Failed to close file\n");
    return -1;
  }

  return 0;
}
