/*  
    Copyright (C) 2015 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @example main_ccpmaster.c
 * @author Andreas Stewering-Bone
 * @date 8 June 2015
 * @brief Simple CCP master application for testing the CCP low level API.
 *
 * Simple CCP master application for testing the CCP low level API.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <time.h>

#include "cs_ccp_api.h"

#define MAXSIGNALS 0xFF
#define MAXCHAR 0xFF
#define MAXBUFFER 0xFFF1

void help(void)
{
  printf(
          "Usage: testccpapi [OPTION]\n"
          "Test CCP API stack.\n"
          "\n"
          "Options:\n"
          "  -i, --iface    CAN-Interface\n"
	  "  -a, --a2l      A2L file to use\n"
	  "  -p, --project  Project name\n"
	  "  -m, --module   Module name\n"
	  "  -o, --option   Optional MOD_PAR section\n"
	  "  -s, --signal   Signal name for DAQ transmission, can be set multiple times\n"
	  "  -e, --event    Event name for DAQ transmission\n"
	  "  -c, --characteristic \"id, value\" Characteristic name,value (double), can be set multiple times\n"
	  "  -f, --force    Test upload download area function with dummy address\n"
          "      --help     display this help and exit\n");
}

void valuecb(double value, double ts, void *priv)
{
  printf("DAQ Transmission Signal %s TS %f: %f\n",(char*)priv,ts,value);
}

int main(int argc, char **argv)
{
  int c;
  int i;
  cs_ccp_t data;
  char * filename = NULL;
  char * project = NULL;
  char * module = NULL;
  char * iface = NULL;
  char * option = NULL;
  char * signal[MAXSIGNALS];
  char * event = NULL;
  char characteristic[MAXCHAR][CS_CHARBUF];
  double rvalue[MAXSIGNALS];
  double tsrvalue[MAXSIGNALS];
  int lt_signal = 0;
  int lt_characteristic = 0;
  double cvalue[MAXCHAR];
  double uvalue;
  int forcearea = 0;
  uint8_t mbuffer[MAXBUFFER];
  int ulen = 1;
  char * split = NULL;

  memset(signal,0,sizeof(char*)*MAXSIGNALS);


  while (1) {
    static struct option long_options[] = {
      /* These options set a flag. */
      {"iface", required_argument, NULL, 'i'},
      {"a2l", required_argument, NULL, 'a'},
      {"project", required_argument, NULL, 'p'},
      {"module", required_argument, NULL, 'm'},
      {"option", required_argument, NULL, 'o'},
      {"signal", required_argument, NULL, 's'},
      {"event", required_argument, NULL, 'e'},
      {"force", required_argument, NULL, 'f'},
      {"characteristic", required_argument, NULL, 'c'},
      {"help",    no_argument, NULL, 'h'},
      {0, 0, 0, 0}
    };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    c = getopt_long (argc, argv, "i:ha:p:m:s:e:c:o:f",
                     long_options, &option_index);

    /* Detect the end of the options. */
    if (c == -1)
      break;

    switch (c) {
    case 0:
      /* If this option set a flag, do nothing else now. */
      if (long_options[option_index].flag != 0)
        break;
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
      break;
    case 'i': iface = optarg;     break;
    case 'f': forcearea = 1;     break;
    case 'a': filename = optarg;     break;
    case 'p': project = optarg;     break;
    case 'm': module = optarg;     break;
    case 'o': option = optarg;     break;
    case 's': 
      signal[lt_signal] = optarg; 
      lt_signal++;
      if(lt_signal >= MAXSIGNALS){
	printf("Invalid number of signal desfintion\n");
	return -1;
      }
      break;
    case 'e': event = optarg;     break;
    case 'c':
      printf("Optarg: %s\n",optarg);
      split = strstr(optarg,",");
      memcpy(characteristic[lt_characteristic],optarg,(split-optarg));
      if(sscanf(split,", %lf",&cvalue[lt_characteristic])<0){
	printf("Failed to parse characterist argument\n");
	return -1;
      }
      printf("%s, %f\n",characteristic[lt_characteristic],cvalue[lt_characteristic]);
      lt_characteristic++;
      if(lt_characteristic >= MAXCHAR){
	printf("Invalid number of characteristic desfintion\n");
	return -1;
      }
      break;
    case 'h': help(); exit(0);    break;
    case '?':
      /* getopt_long already printed an error message. */
      printf( "Typ `testccpapi --help' for more information\n");
      return -1;
      break;
    default:
      return -1;
      break;
    }
  }

  //First check Library version
  if(CS_CCP_CHECKVERSION){
    printf("Incompatible Library interface, Application was compiled with version %i, Library has version %i\n",
	   CS_CCP_VERSION_MAGIC,cs_ccp_getversion());
    return -1;
  }
  
  /*Check inputs*/
  if((iface==NULL)||(filename == NULL)||(project == NULL)||(module == NULL)){
    printf( "Please check input arguments\n");
    printf( "Typ `testccpmaster --help' for more information\n");
    return -1;
  }

  //Try to open A2L file
  if(cs_ccp_api_load(&data,filename,project,module,option,1)){
    printf("Failed to open file %s\n",filename);
    return -1;
  }


  //Open CAN interface
  if(cs_ccp_api_open(&data,iface)){
    printf("Failed to open CAN interface: %s\n",cs_ccp_getlasterrorstring(&data));
    return -1;
  }

  if(lt_characteristic > 0){
    for(i = 0; i < lt_characteristic; i++){
      printf("Try to write CHARACTERISTIC %s\n",characteristic[i]);
      //Try to Download Parameters
      printf("Download value %f\n",cvalue[i]);
      if(cs_ccp_api_dnload_value(&data,characteristic[i],&cvalue[i],1)){
	printf("Failed to download characteristic: %s\n",cs_ccp_getlasterrorstring(&data));
	return -1;
      }
      
      //Upload it from slave
      if(cs_ccp_api_upload_value(&data,characteristic[i],&uvalue,&ulen)){
	printf("Failed to upload characteristic: %s\n",cs_ccp_getlasterrorstring(&data));
	return -1;
      }
      printf("Upload value %f\n",uvalue);
    }
  }

  if((lt_signal > 0)&&(event != NULL)){
    for(i = 0; i < lt_signal; i++){
      //Register signal
      printf("Register signal %s\n",signal[i]);
      if(cs_ccp_api_addSignalDAQ(&data,signal[i],event,1,&rvalue[i],&tsrvalue[i],valuecb,(void*)signal[i])){
	printf("Register signal failed: %s\n",cs_ccp_getlasterrorstring(&data));
	return -1;
      }
    }
    
    //Prepare configuration
    if(cs_ccp_api_configureDAQ(&data)){
      printf("Fail to configure DAQ transmission: %s\n",cs_ccp_getlasterrorstring(&data));
      return -1;
    }

    //Start DAQ transmission
    if(cs_ccp_api_startDAQ(&data)){
      printf("Failed to start DAQ transmission: %s\n",cs_ccp_getlasterrorstring(&data));
      return -1;
    }
   
    for(i=0; i< 10;i++){
      sleep(1);
      //Test if slave is available
      if(cs_ccp_api_process(&data)){
	printf("Receive: %s\n",cs_ccp_getlasterrorstring(&data));
      }
    }
      
    //Stop DAQ transmission
    if(cs_ccp_api_stopDAQ(&data)){
      printf("Cmd StartStop failed: %s\n",cs_ccp_getlasterrorstring(&data));
      return -1;
    }
  }

  if(forcearea){
    //Download memory area
    for(i = 0; i < MAXBUFFER; i++)
      mbuffer[i] = i; // I know there is an overflow
    
    if(cs_ccp_api_dnload_area(&data,0x12345678,0,mbuffer,MAXBUFFER)){
      printf("Download memory area failed: %s\n",cs_ccp_getlasterrorstring(&data));
      return -1;
    }
    
    //Uploadload memory area
    for(i = 0; i < MAXBUFFER; i++)
      mbuffer[i] = i; // I know there is an overflow
    
    if(cs_ccp_api_upload_area(&data,0x12345678,0,mbuffer,MAXBUFFER)){
      printf("Upload memory area failed: %s\n",cs_ccp_getlasterrorstring(&data));
      return -1;
    }
  }
  //Close stack
  if(cs_ccp_api_close(&data)){
    printf("Failed to close CCP stack: %s\n",cs_ccp_getlasterrorstring(&data));
    return -1;
  }
  printf("Close CCP stack succesfully: %s\n",cs_ccp_getlasterrorstring(&data));


  return 0;
}
