/*  
    Copyright (C) 2015 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @example main_ccpmaster.c
 * @author Andreas Stewering-Bone
 * @date 8 June 2015
 * @brief Simple CCP master application for testing the CCP low level API.
 *
 * Simple CCP master application for testing the CCP low level API.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <time.h>

#include "cs_ccp.h"


#define DOSLEEP 0

#if DOSLEEP
#define WAIT(X) sleep(X)
#else
#define WAIT(X)
#endif

void help(void)
{
  printf(
          "Usage: testccpmaster [OPTION]\n"
          "Test CCP master stack.\n"
          "\n"
          "Options:\n"
          "  -i, --iface    CAN-Interface\n"
	  "  -a, --a2l      A2L file to use\n"
	  "  -p, --project  Project name\n"
	  "  -m, --module   Module name\n"
	  "  -o, --option   Optional MOD_PAR name\n"
	  "  -s, --signal   Signal name for DAQ transmission\n"
          "      --help     display this help and exit\n");
}

void valuecb(double value, double ts, void *priv)
{
  printf("DAQ Transmission TS %f: %f\n",ts,value);
}

int main(int argc, char **argv)
{
  int c;
  int i;
  cs_ccp_t data;
  char * filename = NULL;
  char * project = NULL;
  char * module = NULL;
  char * iface = NULL;
  char * signal = NULL;
  char * option = NULL;
  uint8_t ccpmajor;
  uint8_t ccpminor;
  uint8_t exchangeidlength;
  uint8_t tdata[8];
  uint8_t daq_size;
  uint8_t pid_first;
  uint8_t status;
  double rvalue;
  double tsrvalue;
  cs_a2l_measurement_t me;

  while (1) {
    static struct option long_options[] = {
      /* These options set a flag. */
      {"iface", required_argument, NULL, 'i'},
      {"a2l", required_argument, NULL, 'a'},
      {"project", required_argument, NULL, 'p'},
      {"module", required_argument, NULL, 'm'},
      {"signal", required_argument, NULL, 's'},
      {"option", required_argument, NULL, 'o'},
      {"help",    no_argument, NULL, 'h'},
      {0, 0, 0, 0}
    };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    c = getopt_long (argc, argv, "i:ha:p:m:s:o:",
                     long_options, &option_index);

    /* Detect the end of the options. */
    if (c == -1)
      break;

    switch (c) {
    case 0:
      /* If this option set a flag, do nothing else now. */
      if (long_options[option_index].flag != 0)
        break;
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
      break;
    case 'i': iface = optarg;     break;
    case 'a': filename = optarg;     break;
    case 'p': project = optarg;     break;
    case 'm': module = optarg;     break;
    case 's': signal = optarg;     break;
    case 'o': option = optarg;     break;
    case 'h': help(); exit(0);    break;
    case '?':
      /* getopt_long already printed an error message. */
      printf( "Typ `testccpmaster --help' for more information\n");
      return -1;
      break;
    default:
      return -1;
      break;
    }
  }

  //First check Library version
  if(CS_CCP_CHECKVERSION){
    printf("Incompatible Library interface, Application was compiled with version %i, Library has version %i\n",
	   CS_CCP_VERSION_MAGIC,cs_ccp_getversion());
    return -1;
  }
  
  /*Check inputs*/
  if((iface==NULL)||(filename == NULL)||(project == NULL)||(module == NULL)){
    printf( "Please check input arguments\n");
    printf( "Typ `testccpmaster --help' for more information\n");
    return -1;
  }

  //Try to open file
  if(cs_ccp_a2l_open(&data.ccpa2l,filename,project,module,"ASAP1B_CCP",option)){
    printf("Failed to open file %s\n",filename);
    return -1;
  }

  //Try to init stack
  if(cs_ccp_init(&data)){
    printf("Failed to init CCP stack\n");
    return 1;
  }

  //Open CAN interface
  if(cs_ccp_can_open(&data,iface)){
    printf("Failed to open CAN interface: %s\n",cs_ccp_getlasterrorstring(&data));
    return -1;
  }

  //Test if slave is available
  if(cs_ccp_cmd_test(&data)){
    printf("Cmd Test failed: %s\n",cs_ccp_getlasterrorstring(&data));
  } else {
    printf("Get test response from slave\n");
  }

  //Connect
  if(cs_ccp_cmd_connect(&data)){
    printf("Cmd Connect failed: %s\n",cs_ccp_getlasterrorstring(&data));
    return -1;
  } else {
    printf("Get Connect ack from slave\n");
  }

  //Get CCP Version
  if(cs_ccp_cmd_ccpversion(&data,&ccpmajor,&ccpminor)){
    printf("Cmd CCPVersion failed: %s\n", cs_ccp_getlasterrorstring(&data));
    return -1;
  } else {
    printf("Slave CCP Version Major: 0x%X Minor: 0x%X\n",ccpmajor,ccpminor);
  }

  //Get Exchange ID
  if(cs_ccp_cmd_exchangeid(&data,&exchangeidlength,NULL,NULL,NULL)){
    printf("Cmd Exchange id failed: %s\n", cs_ccp_getlasterrorstring(&data));
  } else {
    cs_ccp_print_resource(&data);
  }

  //SetMTA
  if(cs_ccp_cmd_setmta(&data,0,cs_a2l_CalcAddress(&data.ccpa2l.a2l,0x10000000),0)){
    printf("Cmd SetMTA failed: %s\n",cs_ccp_getlasterrorstring(&data));
    return -1;
  } else {
    printf("Get SetMTA ack from slave\n");
  }
  tdata[0] = 0xFF;
  tdata[1] = 0xFE;
  tdata[2] = 0xFD;
  if(cs_ccp_cmd_dnload(&data,tdata,3)){
    printf("Cmd Dnload failed: %s\n",cs_ccp_getlasterrorstring(&data));
    return -1;
  } else {
    printf("Get Dnload ack from slave\n");
  }


  //Try to do communication with ignore frame
  printf("Start communication with ignore ack\n");
  cs_ccp_set_ingnoreresponse(&data,1);

  //SetMTA
  if(cs_ccp_cmd_setmta(&data,0,cs_a2l_CalcAddress(&data.ccpa2l.a2l,0x10000000),0)){
    printf("Cmd SetMTA failed: %s\n",cs_ccp_getlasterrorstring(&data));
    return -1;
  }
  tdata[0] = 0xFF;
  tdata[1] = 0xFE;
  tdata[2] = 0xFD;
  if(cs_ccp_cmd_dnload(&data,tdata,3)){
    printf("Cmd Dnload failed: %s\n",cs_ccp_getlasterrorstring(&data));
    return -1;
  }

  //SetMTA
  if(cs_ccp_cmd_setmta(&data,0,cs_a2l_CalcAddress(&data.ccpa2l.a2l,0x10000000),0)){
    printf("Cmd SetMTA failed: %s\n",cs_ccp_getlasterrorstring(&data));
    return -1;
  }
  tdata[0] = 0xFF;
  tdata[1] = 0xFE;
  tdata[2] = 0xFD;
  if(cs_ccp_cmd_dnload(&data,tdata,3)){
    printf("Cmd Dnload failed: %s\n",cs_ccp_getlasterrorstring(&data));
    return -1;
  }

  printf("Back to normal communication\n");
  cs_ccp_set_ingnoreresponse(&data,0);

  //SetMTA
  if(cs_ccp_cmd_setmta(&data,0,cs_a2l_CalcAddress(&data.ccpa2l.a2l,0x10000000),0)){
    printf("Cmd SetMTA failed: %s\n",cs_ccp_getlasterrorstring(&data));
    return -1;
  } else {
    printf("Get SetMTA ack from slave\n");
  }
  tdata[0] = 0xFF;
  tdata[1] = 0xFE;
  tdata[2] = 0xFD;
  if(cs_ccp_cmd_dnload(&data,tdata,3)){
    printf("Cmd Dnload failed: %s\n",cs_ccp_getlasterrorstring(&data));
    return -1;
  } else {
    printf("Get Dnload ack from slave\n");
  }

  //SetMTA
  if(cs_ccp_cmd_setmta(&data,0,cs_a2l_CalcAddress(&data.ccpa2l.a2l,0x10000000),0)){
    printf("Cmd SetMTA failed: %s\n",cs_ccp_getlasterrorstring(&data));
    return -1;
  } else {
    printf("Get SetMTA ack from slave\n");
  }
  tdata[0] = 0x00;
  tdata[1] = 0x00;
  tdata[2] = 0x00;
  if(cs_ccp_cmd_upload(&data,tdata,3)){
    printf("Cmd Upload failed: %s\n",cs_ccp_getlasterrorstring(&data));
    return -1;
  } else {
    printf("Get Upload ack from slave\n");
    printf("Data: (%i Bytes) 0x%X 0x%X 0x%X\n",3,tdata[0],tdata[1],tdata[2]);
  }

  if(signal != NULL){
    //Get measurement signal from A2L file
    if(cs_a2l_GetMeasurement(&data.ccpa2l.a2l,&me,signal)){
      printf("Failed to parse signal %s\n",signal);
    } else{
      printf("Success parsing signal %s\n",signal);
      
      
      //GetDAQSize, reset existing configutration
      if(cs_ccp_cmd_getdaqsize(&data,0,data.ccpa2l.can_recv_id,&daq_size,&pid_first)){
	printf("Cmd GetDAQSize failed: %s\n",cs_ccp_getlasterrorstring(&data));
	return -1;
      } else {
	printf("Get GetDAQSize ack from slave\n");
	printf("DAQ %i size %i\n",0,daq_size);
	printf("DAQ %i first PID 0x%X\n",0,pid_first);
      }
      
      //Register signal
      if(cs_ccp_entry_reg_value(&data,0,0,0,me.size,&rvalue,
				me.BO,
				me.cm,
				me.dtype,
				&tsrvalue,
				valuecb,NULL)){
	printf("Failed to register signal\n");
	return -1;
      }

      //Set DAQ pointer
      if(cs_ccp_cmd_setdaqptr(&data,0,0,0)){
	printf("Cmd SetDAQPtr failed: %s\n",cs_ccp_getlasterrorstring(&data));
	return -1;
      } else {
	printf("Get SetDAQPtr ack from slave\n");
      }
      
      //Write DAQ element, signal data
      if(cs_ccp_cmd_writedaq(&data,me.size,me.extension,me.address)){
	printf("Cmd WriteDAQ failed: %s\n",cs_ccp_getlasterrorstring(&data));
	return -1;
      } else {
	printf("Get WriteDAQ ack from slave\n");
      }
      
      //Get optional session status
      if(cs_ccp_cmd_getsessionstatus(&data,&status)){
	printf("Cmd GetSessionStatus failed: %s\n",cs_ccp_getlasterrorstring(&data));
      } else {
	printf("Get GetSessionStatus ack from slave\n");
	cs_ccp_print_status(status);
      }
      
      //Start DAQ transmission
      if(cs_ccp_cmd_startstop(&data,1,0,0,0,1)){
	printf("Cmd StartStop failed: %s\n",cs_ccp_getlasterrorstring(&data));
	return -1;
      } else {
	printf("Get StartStop ack from slave\n");
      }
      
      for(i=0; i< 10;i++){
	sleep(1);
	//Test if slave is available
	if(cs_ccp_receive(&data)){
	  printf("Receive: %s\n",cs_ccp_getlasterrorstring(&data));
	}
      }
      
      //Stop DAQ transmission
      if(cs_ccp_cmd_startstop(&data,0,0,0,0,1)){
	printf("Cmd StartStop failed: %s\n",cs_ccp_getlasterrorstring(&data));
	return -1;
      } else {
	printf("Get StartStop ack from slave\n");
      }
    }
  }
  //Disconnect
  if(cs_ccp_cmd_disconnect(&data,1)){
    printf("Cmd Disconnect failed: %s\n",cs_ccp_getlasterrorstring(&data));
    return -1;
  } else {
    printf("Get Disconnect ack from slave\n");
  }

  //Close CAN interface
  if(cs_ccp_can_close(&data)){
    printf("Failed to close CAN interface\n");
    return -1;
  }

  //Close stack
  if(cs_ccp_close(&data)){
    printf("Failed to close CCP stack: %s\n",cs_ccp_getlasterrorstring(&data));
    return -1;
  }
  printf("Close CCP stack succesfully: %s\n",cs_ccp_getlasterrorstring(&data));

  //Close A2L file
  if(cs_ccp_a2l_close(&data.ccpa2l)){
    printf("Failed to close file\n");
    return -1;
  }

  return 0;
}
