/*  
    Copyright (C) 2019 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @example main_dbc_simulate.c
 * @author Andreas Stewering-Bone
 * @date 10 Jan 2019
 * @brief Test DBC read API to simulate a node.
 *
 * Test DBC read API to simulate a node.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>

#include "cs_dbc.h"
#include "cs_process.h"
#include "cs_schedule.h"
#include "cs_message.h"
#include "cs_signal.h"

#define MAX_ENTRIES 255
#define SAMPLETIME 1000 // 1000 ms

void help(void)
{
  printf(
	 "Usage: testdbcsimulate [OPTION]\n"
	 "Test DBC API Stack to simulate CAN node.\n"
	 "\n"
	 "Options:\n"
	 "  -i, --interface NAME   CAN interface name\n"
	 "  -d, --dbc FILE         DBC file\n"
	 "  -t, --target node      Simulate and generate a all messages which are needed by this node,\n"
	 "                         Can be set multiple times\n"
	 "                         Will take all messages marked as cyclic in DBC file\n"
	 "  -g, --generate CAN-ID  In addition to option -t,\n"
	 "                         Add message to the generated messages\n"
	 "                         Can be set multiple times\n"
	 "  -v, --verbose          Verbose\n"
	 "                         Can be set multiple times\n"
	 "      --help             Display this help and exit\n");
}


int main(int argc, char **argv)
{
  int c;
  int i,j;
  cs_dbc_t *data;
  char * filename = NULL;
  char * iface = NULL;
  int verbose = 0;
  char *targets[MAX_ENTRIES];
  int num_targets = 0;
  uint32_t g_id[MAX_ENTRIES];
  int num_gid = 0;

  cs_schedule_t *sc = NULL;
  cs_process_t *p = NULL;
  
  while (1) {
    static struct option long_options[] = {
      /* These options set a flag. */
      {"interface", required_argument, NULL, 'i'},
      {"dbc", required_argument, NULL, 'd'},
      {"target", required_argument, NULL, 't'},
      {"generate", required_argument, NULL, 'g'},
      {"verbose",   no_argument, NULL, 'v'},
      {"help",    no_argument, NULL, 'h'},
      {0, 0, 0, 0}
    };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    c = getopt_long (argc, argv, "i:d:t:g:vh",
                     long_options, &option_index);

    /* Detect the end of the options. */
    if (c == -1)
      break;

    switch (c) {
    case 0:
      /* If this option set a flag, do nothing else now. */
      if (long_options[option_index].flag != 0)
        break;
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
      break;
    case 'i': iface = optarg;     break;
    case 'd': filename = optarg;     break;
    case 't': targets[num_targets] = optarg;
      num_targets++;
      if(num_targets >= MAX_ENTRIES){
	printf("Maximum number of targets reached\n");
	return -1;
      }
      break;
    case 'g':  g_id[num_gid] = strtoul(optarg,(char**)NULL,16);
      num_gid++;
      if(num_gid >= MAX_ENTRIES){
	printf("Maximum number of generator ids reached\n");
	return -1;
      }
      break;
    case 'v': verbose += 1;     break;
    case 'h': help(); exit(0);    break;
    case '?':
      /* getopt_long already printed an error message. */
      printf( "Typ `testdbcsimulate --help' for more information\n");
      return -1;
      break;
    default:
      return -1;
      break;
    }
  }

  if(iface == NULL){
    printf("Please select a CAN interface name\n");
    return -1;
  }
  
  if(filename == NULL){
    printf("Please select a DBC file\n");
    return -1;
  }

  if(num_targets == 0){
    printf("Please select minimum one target node\n");
    return -1;
  }
  
  
  if(verbose > 0){
    cs_dbc_debuglevel(CS_DEBUG_LEVEL_DEBUG);
    cs_schedule_debuglevel(CS_DEBUG_LEVEL_DEBUG);
  }

  if(verbose > 1){
    cs_dbc_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
    cs_schedule_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
    cs_message_debuglevel(CS_DEBUG_LEVEL_DEBUG);
    cs_signal_debuglevel(CS_DEBUG_LEVEL_DEBUG);
    cs_process_debuglevel(CS_DEBUG_LEVEL_DEBUG);
  }

  if(verbose > 3){
    cs_dbc_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
    cs_schedule_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
    cs_message_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
    cs_signal_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
    cs_process_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
  }

  if((data = cs_dbc_init())==NULL){
    printf("Failed to init DBC structure\n");
    return -1;
  }


  if(cs_dbc_read(data,filename)){
    printf("Failed to parse DBC file: %s\n",filename);
    return -1;
  }

  sc = cs_schedule_init();
  if(sc == NULL){
    printf("Failed to init scheduler\n");
    return -1;
  }

  for(i = 0; i < num_targets; i++){
    cs_dbc_node_t* tnode = NULL;
    tnode = cs_dbc_getnode_byname(data,targets[i]);
    if(tnode == NULL){
      printf("Can not find target node %s in DBC file\n",targets[i]);
      return -1;
    }
    for(j = 0; j < cs_dbc_node_no_requested_messages(tnode); j++){
      cs_message_t *tm = NULL; //DBC related message structure
      
      tm = cs_dbc_node_get_requested_message_iter(tnode,j);
      if(tm == NULL){
	printf("Failed to get message from DBC structure\n");
      }
      if(cs_message_get_sendtype(tm) != MESSAGE_STYPE_CYCLIC){
	//Skip because no cyclic message
	printf("Skip non cyclic message %s\n",cs_message_get_name(tm));
	continue;
      }

      //Set message direction, we want to sent to target
      if(cs_message_set_direction(tm,MESSAGE_DIR_SEND)){
	printf("Failed to set message %s direction\n",cs_message_get_name(tm));
	return -1;
      }
      
      if(cs_schedule_add_message(sc,tm)){
	printf("Failed to add message %s to scheduler.\n Maybe already added\n",cs_message_get_name(tm));
      }
      printf("Add message %s to the generated messages\n",cs_message_get_name(tm));
    }

    for(j = 0; j < cs_dbc_node_no_generated_messages(tnode); j++){
      cs_message_t *tm = NULL; //DBC related message structure
      
      tm = cs_dbc_node_get_generated_message_iter(tnode,j);
      if(tm == NULL){
	printf("Failed to get message from DBC structure\n");
      }
      if(cs_message_get_sendtype(tm) != MESSAGE_STYPE_CYCLIC){
	//Skip because no cyclic message
	printf("Skip non cyclic message %s\n",cs_message_get_name(tm));
	continue;
      }

      //Set message direction, we want to receive from target
      if(cs_message_set_direction(tm,MESSAGE_DIR_RECEIVE)){
	printf("Failed to set message %s direction\n",cs_message_get_name(tm));
	return -1;
      }
      
      if(cs_schedule_add_message(sc,tm)){
	printf("Failed to add message %s to scheduler.\n Maybe already added\n",cs_message_get_name(tm));
      }
      printf("Add message %s to the requested messages\n",cs_message_get_name(tm));
    }
  }

  for(i = 0; i < num_gid; i++){
    cs_message_t *tm = NULL;
    if((tm = cs_dbc_getmessage(data,g_id[i])) == NULL){
      printf("Failed to get message 0x%X from DBC file\n",g_id[i]);
      return -1;
    }

    //Set message direction, we want to send target
    if(cs_message_set_direction(tm,MESSAGE_DIR_SEND)){
      printf("Failed to set message %s direction\n",cs_message_get_name(tm));
      return -1;
    }

    if(cs_schedule_add_message(sc,tm)){
      printf("Failed to add message %s to scheduler.\n Maybe already added\n",cs_message_get_name(tm));
    }
    printf("Add message %s to the generated messages\n",cs_message_get_name(tm));
    
  }

  //Prepare scheduler
  cs_schedule_set_cycletime(sc,SAMPLETIME);
  cs_schedule_set_maxpcycle(sc,1);    //Max 1 messages per cycle

  if(cs_schedule_checkmessages(sc)){
    printf("Message check failed\n");
    return -1;
  }

  //Open CAN interface
  if(cs_schedule_open(sc,iface,SOCKETCAN_MODE_FLEX)){
    printf("Failed to open CAN interface\n");
    return -1;
  }

  if((p = cs_process_init()) == NULL){
    printf("Failed to init process\n");
    return -1;
  }

  if(cs_process_prepare(p,99,1.0 / SAMPLETIME,-1)){
    printf("Failed to init process\n");
    return -1;
  }

  printf("Number messages, Shifted Messages, General Error, TX-Full Error\n");
  while(1){
    cs_process_cycle_start(p);
    cs_schedule_schedule(sc,SCHEDULE_MODE_RECIEVE);

    //Handle signals
    //Here we do nothing, because it is only a generic example
    
    cs_schedule_schedule(sc,SCHEDULE_MODE_SEND);

    printf("%i , %i, %i, %i\n",
	   cs_schedule_get_send_msgcounter(sc),
	   cs_schedule_get_send_msgshiftedcounter(sc),
	   cs_schedule_get_send_errorcounter(sc),
	   cs_schedule_get_send_txfullerrorcounter(sc)
	   );
    
    cs_process_cycle_end(p);
  }

  
  //Close CAN interface
  cs_schedule_close(sc);

  cs_process_free(p);
  cs_dbc_free(data);
  cs_schedule_free(sc);
  return 0;
}
