/*  
    Copyright (C) 2015 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @example main_isotpsend.c
 * @author Andreas Stewering-Bone
 * @date 13 July 2015
 * @brief Simple Isotp Send application to test library isotp stack.
 *
 * Simple Isotp Send application to test library isotp stack.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <getopt.h>
#include <unistd.h>
#include <time.h>

#include "cs_isotp.h"

void help(void)
{
  printf(
          "Usage: testisotpsend [OPTION]\n"
          "Test library ISOTP stack.\n"
          "\n"
          "Options:\n"
	  "  -m, --mode       Set Running mode, (0 => CAN2.0 (TX-DL 8 Byte) default,\n" 
          "                                      1=> CAN-FD (TX-DL 8 Byte),\n"
	  "                                      2=> CAN-FD (TX-DL 12 Byte),\n"
	  "                                      3=> CAN-FD (TX-DL 16 Byte),\n"
	  "                                      4=> CAN-FD (TX-DL 24 Byte),\n"
	  "                                      5=> CAN-FD (TX-DL 32 Byte),\n"
	  "                                      6=> CAN-FD (TX-DL 48 Byte),\n"
	  "                                      7=> CAN-FD (TX-DL 64 Byte))\n"
          "  -r, --rx         Enable RX-Padding check\n"
          "  -p, --tx         Enable TX-Padding\n"
          "  -l, --len        Enable exact RX-Length check\n"
          "  -s, --source     Source CAN-ID\n"
	  "  -t, --target     Target CAN-ID\n"
	  "  -i, --interface  CAN Interface\n"
	  "  -d, --server     SocketCANd server hostname\n"
	  "  -k, --port       SocketCANd server port\n"
	  "  -v  --verbose    Enable ISOTP-Runtime debugging\n"
          "      --help       display this help and exit\n");
}

void transfercb(void *ctx,uint8_t mode,uint32_t length, uint32_t counter, uint32_t actbytes)
 {
   printf("Transfer package Mode 0x%X (%i bytes), Counter %i, Act %i bytes\n",mode,length,counter,actbytes);
 }

int main(int argc, char **argv)
{
  int c;
  uint32_t sa = 0;
  uint32_t ta = 0;
  char * interface = NULL;
  cs_isotp_t *data;
  uint8_t isotp_option = ISOTP_OPTION_NONE;
  int isotp_debug = 0;
  int i;
  char * server = NULL;
  int port = 0;
  uint8_t mode = 0;
  int rv;
  
  while (1) {
    static struct option long_options[] = {
      /* These options set a flag. */
      {"source", required_argument, NULL, 's'},
      {"target", required_argument, NULL, 't'},
      {"len",    no_argument, NULL, 'l'},
      {"rx",    no_argument, NULL, 'r'},
      {"tx",    no_argument, NULL, 'p'},
      {"mode",    no_argument, NULL, 'm'},
      {"server", required_argument, NULL, 'd'},
      {"port",  required_argument, NULL, 'k'},
      {"help",    no_argument, NULL, 'h'},
      {"verbose",  no_argument, NULL, 'v'},
      {0, 0, 0, 0}
    };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    c = getopt_long (argc, argv, "s:t:i:d:k:m:hrlpv",
                     long_options, &option_index);

    /* Detect the end of the options. */
    if (c == -1)
      break;

    switch (c) {
    case 0:
      /* If this option set a flag, do nothing else now. */
      if (long_options[option_index].flag != 0)
        break;
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
      break;
    case 's': sa = strtoul(optarg,(char**)NULL,16); break;
    case 't': ta = strtoul(optarg,(char**)NULL,16); break;
    case 'i': interface = optarg;     break;
    case 'v': isotp_debug++;     break;
    case 'm': mode = (uint8_t)strtol(optarg,(char**)NULL,10); break;
    case 'd': server = optarg; break;
    case 'k': port = strtoul(optarg,(char**)NULL,10); break;
    case 'r': isotp_option |= ISOTP_OPTION_RXPADDING;    break;
    case 'p': isotp_option |= ISOTP_OPTION_TXPADDING;    break;
    case 'l': isotp_option |= ISOTP_OPTION_RXLENGTH;    break;
    case 'h': help(); exit(0);    break;
    case '?':
      /* getopt_long already printed an error message. */
      printf( "Typ `testisotpsend --help' for more information\n");
      return -1;
      break;
    default:
      return -1;
      break;
    }
  }

  if((interface == NULL)||
     (sa == 0)||
     (ta == 0)){
    printf("Please check arguments\n");
    return -1;
  }

  if(isotp_debug == 1){
    cs_isotp_debuglevel(CS_DEBUG_LEVEL_DEBUG);
  }

  if(isotp_debug == 2){
    cs_isotp_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
  }

  if(isotp_debug == 3){
    cs_isotp_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
    cs_socketcan_debuglevel(CS_DEBUG_LEVEL_DEBUG);
  }

  if(isotp_debug > 3){
    cs_isotp_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
    cs_socketcan_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
  }
  
  //Init with default values
  if((data = cs_isotp_init(sa,ta))==NULL){
    printf("Init failed\n");
    return -1;
  }

  //Set Options
  if(cs_isotp_setoptions(data,isotp_option)){
    printf("Failed to set options: %s\n",cs_isotp_getlasterrorstring(data));
    return -1;
  }

  //Set Pad byte
  if(cs_isotp_setpadbyte(data,0xCC)){
    printf("Failed to set PAD byte: %s\n",cs_isotp_getlasterrorstring(data));
    return -1;
  }
  
  switch(mode){
  case 0:
  default:
    //CAN2.0
    rv = cs_isotp_setoptions_link_layer(data,ISOTP_LL_MODE_CAN,ISOTP_LL_TX_DL_8);
    break;
  case 1:
    //CAN-FD 8 Byte
    rv = cs_isotp_setoptions_link_layer(data,ISOTP_LL_MODE_CANFD,ISOTP_LL_TX_DL_8);
    break;
  case 2:
    //CAN-FD 12 Byte
    rv = cs_isotp_setoptions_link_layer(data,ISOTP_LL_MODE_CANFD,ISOTP_LL_TX_DL_12);
    break;
  case 3:
    //CAN-FD 16 Byte
    rv = cs_isotp_setoptions_link_layer(data,ISOTP_LL_MODE_CANFD,ISOTP_LL_TX_DL_16);
    break;
  case 4:
    //CAN-FD 24 Byte
    rv = cs_isotp_setoptions_link_layer(data,ISOTP_LL_MODE_CANFD,ISOTP_LL_TX_DL_24);
    break;
  case 5:
    //CAN-FD 32 Byte
    rv = cs_isotp_setoptions_link_layer(data,ISOTP_LL_MODE_CANFD,ISOTP_LL_TX_DL_32);
    break;
  case 6:
    //CAN-FD 48 Byte
    rv = cs_isotp_setoptions_link_layer(data,ISOTP_LL_MODE_CANFD,ISOTP_LL_TX_DL_48);
    break;
  case 7:
    //CAN-FD 64 Byte
    rv = cs_isotp_setoptions_link_layer(data,ISOTP_LL_MODE_CANFD,ISOTP_LL_TX_DL_64);
    break;
  }
  if(rv){
    printf("Failed to data link layer options: %s\n",cs_isotp_getlasterrorstring(data));
    return -1;
  }
  
  //Set transfer callback pointer
  cs_isotp_transfer_setcallback(data,NULL,transfercb);
  
  int maxbytes = cs_isotp_getmaxbytes(data);
  uint8_t buffer[maxbytes];
  uint8_t rbuffer[maxbytes];
  int loops = 100;
  uint32_t len = 8;
  
  //Init send data
  for(i = 0; i < maxbytes; i++){
    buffer[i] = (uint8_t)i;
  }

  for(i = 0; i < loops; i++){
  
    if(cs_isotp_open(data,interface,server,port)){
      printf("Failed to open interface: %s\n",cs_isotp_getlasterrorstring(data));
      return -1;
    }
    
    //First try to send only a SF frame
    if(cs_isotp_send(data,buffer,6)){
      printf("Failed to send SF Frame: %s\n",cs_isotp_getlasterrorstring(data));
      return -1;
    }
    
    //Next try a big data paket
    if(cs_isotp_send(data,buffer,513)){
      printf("Failed to send Frame: %s\n",cs_isotp_getlasterrorstring(data));
      return -1;
    }
    
    //Next try
    if(cs_isotp_send(data,buffer,4)){
      printf("Failed to send Frame: %s\n",cs_isotp_getlasterrorstring(data));
      return -1;
    }
    
    //Next try
    if(cs_isotp_send(data,buffer,maxbytes)){
      printf("Failed to send Frame: %s\n",cs_isotp_getlasterrorstring(data));
      return -1;
    }
    
    //Next try
    //this is the signal for receiver to send a packet
    if(cs_isotp_send(data,buffer,1)){
      printf("Failed to send Frame: %s\n",cs_isotp_getlasterrorstring(data));
      return -1;
    }
    //One time recv
    len = cs_isotp_getmaxbytes(data);
    if(cs_isotp_recv(data,rbuffer,&len,10000)){
      printf("Recv failed: %s\n",cs_isotp_getlasterrorstring(data));
      return -1;
    } 
    
    if(cs_isotp_close(data)){
      printf("Failed to close interface\n");
      return -1;
    }
  }
  cs_isotp_free(data);
  
  
  //Init with force txstim
  if((data = cs_isotp_init(sa,ta))==NULL){
    printf("Init failed\n");
    return -1;
  }

  //Set Options
  if(cs_isotp_setoptions(data,isotp_option)){
    printf("Failed to set options: %s\n",cs_isotp_getlasterrorstring(data));
    return -1;
  }

  if(cs_isotp_open(data,interface,server,port)){
    printf("Failed to open interface: %s\n",cs_isotp_getlasterrorstring(data));
    return -1;
  }

  //First try to send only a SF frame
  if(cs_isotp_send(data,buffer,6)){
    printf("Failed to send SF Frame: %s\n",cs_isotp_getlasterrorstring(data));
    return -1;
  }

  //Next try a big data paket
  if(cs_isotp_send(data,buffer,513)){
    printf("Failed to send Frame: %s\n",cs_isotp_getlasterrorstring(data));
    return -1;
  }

  if(cs_isotp_close(data)){
    printf("Failed to close interface: %s\n",cs_isotp_getlasterrorstring(data));
    return -1;
  }

  cs_isotp_free(data);
  
  return 0;
}
