/*  
    Copyright (C) 2018 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @example main_lin_bit_scanner.c
 * @author Andreas Stewering-Bone
 * @date 16 May 2018
 * @brief Simple STDIN Lin Bit scanner.
 *
 * Simple STDIN Lin Bit scanner.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <time.h>

#include "cs_lin.h"
#include "cs_lin_bit.h"

//Global variables
int verbose = 0;

void help(void)
{
  printf(
	 "Usage: testlinbitscanner [OPTION]\n"
	 "Test STDIN Lin Bit scanner.\n"
	 "Example: cat file | testlinbitscanner -r 1000000 -b 192000\n"
	 "file should be a text file with 1 and 0\n"
	 "\\n is allowed in the text file\n"
	 "\n"
	 "Options:\n"
	 " -i   --input FILE  Input file, if not defined wait for STDIN\n"
	 " -r   --rate VALUE  RAW bit rate, default 1.000.000 Bits/s\n"
	 " -b   --baud VALUE  LIN baud rate, default 19.200 Bits/s\n"
	 " -v   --verbose     Verbose output\n"
	 "      --help        Display this help and exit\n");
}

int main(int argc, char **argv)
{
  int c;
  char *filename = NULL;
  FILE *f = NULL;
  uint32_t rawrate = 1000000;
  uint32_t linrate = 19200;
  int r;
  uint8_t v;
  int rv;
  uint32_t failure_counter = 0;
  uint64_t bit_counter = 0;
  int loffset = 0;
  uint8_t scan_pid = 0;
  uint8_t scan_data[CS_LIN_SIGNAL_MAX_BYTE];
  int scan_length = CS_LIN_SIGNAL_MAX_BYTE;
  uint8_t scan_crc = 0;  
  cs_lin_phys_bit_t* bt;
  
  while (1) {
    static struct option long_options[] = {
      /* These options set a flag. */
      {"input",  required_argument, NULL, 'i'},
      {"rate",  required_argument, NULL, 'r'},
      {"baud",  required_argument, NULL, 'b'},
      {"help",    no_argument, NULL, 'h'},
      {"verbose",    no_argument, NULL, 'v'},
      {0, 0, 0, 0}
    };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    c = getopt_long (argc, argv, "i:r:b:hv",
                     long_options, &option_index);

    /* Detect the end of the options. */
    if (c == -1)
      break;

    switch (c) {
    case 0:
      /* If this option set a flag, do nothing else now. */
      if (long_options[option_index].flag != 0)
        break;
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
      break;
    case 'i': filename = optarg; break;
    case 'r': rawrate = strtoul(optarg,(char**)NULL,10); break;
    case 'b': linrate = strtoul(optarg,(char**)NULL,10); break;
    case 'v': verbose = 1;    break;
    case 'h': help(); exit(0);    break;
    case '?':
      /* getopt_long already printed an error message. */
      printf( "Typ `testlinbitscanner --help' for more information\n");
      return -1;
      break;
    default:
      return -1;
      break;
    }
  }  
  //Set Debug level
  if(verbose == 1){
    cs_lin_bit_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
  } else {
    cs_lin_bit_debuglevel(CS_DEBUG_LEVEL_NONE);
  }

  if(filename != NULL){
    f = fopen(filename, "r");
    if(f == NULL){
      printf("Failed to open %s\n",filename);
      return -1;
    }
  } else {
    f = stdin;
  }

  
  bt = cs_lin_phys_bit_init(linrate,rawrate);
  if(bt == NULL){
    printf("Failed to init bit interface\n");
    return -1;
  }

  while((r = getc(f)) != EOF){
    switch(r){
    case 48: // 0
    case 49: // 1
      bit_counter++;
      v = ((uint8_t)r)-48;
      loffset = 0;
      scan_length = CS_LIN_SIGNAL_MAX_BYTE;
      rv = cs_lin_phys_lin_scan(bt,&v,1,&loffset,&scan_pid,scan_data,&scan_length,&scan_crc);
      
      switch(rv){
      case -1: /*Failure*/
	printf("Lin scan failure, Bit: %ld\n",bit_counter);
	failure_counter++;
	break;
      case 0: /*Idle*/
	/*Nothing to do*/
	break;
      case 2: /*Header but no content*/
	/*Nothing to do*/
	printf("Scan header without data, Bit: %ld,  Id 0x%X\n",bit_counter,scan_pid);
	break;
      case 1: /*Frame in progress */
	/*Nothing to do*/
	break;
      case 4: /*Bit failures*/
	printf("Detect bit failure, Bit: %ld\n",bit_counter);
	failure_counter++;
	break;
      case 3: /*Valid frame*/
	printf("Received a valid frame, Bit: %ld,  Id 0x%X\n",bit_counter,scan_pid);
	break;
      default:
	printf("Lin scan, Unknwon return content, Bit: %ld\n",bit_counter);
	failure_counter++;
	break;
      }
      break;
    default:
      //Ignore
      break;
    }
  }    
  cs_lin_phys_bit_free(bt);

  fclose(f);
  return 0;
}
