/*  
    Copyright (C) 2017 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @example main_lin_sllin_linener.c
 * @author Andreas Stewering-Bone
 * @date 07 May 2018
 * @brief Simple lin sllin lister.
 *
 * Simple lin sllin lister.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <time.h>

#include "cs_lin.h"
#include "cs_lin_bit.h"
#include "cs_socketcan.h"

//Global variables
int verbose = 0; //Default only debug output

void help(void)
{
  printf(
	 "Usage: testlinsllinlistener [OPTION]\n"
	 "Test lin sllin listener stack.\n"
	 "\n"
	 "Options:\n"
	 " -i   --interface <name> CAN interface for testing\n"
	 " -v   --verbose  Verbose output\n"
	 "      --help     display this help and exit\n");
}

int main(int argc, char **argv)
{
  int c;
  char *interface = NULL;
  cs_socketcan_t *sc = NULL;
  cs_lin_t *sk = NULL;
  cs_lin_ldf_t *ldf = NULL;
  cs_lin_enc_t* enc = NULL;
  cs_lin_enc_representation_t* rep = NULL;
  cs_lin_signal_def_t* def = NULL;
  cs_lin_frame_t* frame = NULL;
  cs_lin_schedule_table_t* table = NULL;
  cs_lin_nodeattribute_t* attr = NULL;
  
  while (1) {
    static struct option long_options[] = {
      /* These options set a flag. */
      {"interface",  required_argument, NULL, 'i'},
      {"help",    no_argument, NULL, 'h'},
      {"verbose",    no_argument, NULL, 'v'},
      {0, 0, 0, 0}
    };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    c = getopt_long (argc, argv, "i:hv",
                     long_options, &option_index);

    /* Detect the end of the options. */
    if (c == -1)
      break;

    switch (c) {
    case 0:
      /* If this option set a flag, do nothing else now. */
      if (long_options[option_index].flag != 0)
        break;
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
      break;
    case 'v': verbose = 1;    break;
    case 'i': interface = optarg;    break;
    case 'h': help(); exit(0);    break;
    case '?':
      /* getopt_long already printed an error message. */
      printf( "Typ `testlinsllinlistener --help' for more information\n");
      return -1;
      break;
    default:
      return -1;
      break;
    }
  }

  if(interface == NULL){
    printf("No interface defined\n");
    help();
    return -1;
  }
  
  //Set Debug level
  if(verbose == 1){
    cs_lin_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
    cs_socketcan_debuglevel(CS_DEBUG_LEVEL_DEBUG);
  } else {
    cs_lin_debuglevel(CS_DEBUG_LEVEL_DEBUG);
  }

  //Init structure
  ldf = cs_lin_ldf_init();

  if(ldf == NULL){
    printf("Failed to init structure\n");
    return -1;
  }

  //Set general informations
  if(cs_lin_ldf_set_general(ldf,2.1,2.1,19200)){
    printf("Failed to set LDF general information\n");
    return -1;
  }


  /*Prepare encodings*/
  enc = cs_lin_ldf_get_encoding(ldf,"DigitEncoding",LIN_ENC_PHYSICAL);
  if(enc == NULL){
    printf("Failed to init encoding");
    return -1;
  }
  if(cs_lin_ldf_encoding_add_physical(enc,0,255,1.0,0.0,"-")){
    printf("Failed to add physical");
    return -1;
  }
  
  enc = cs_lin_ldf_get_encoding(ldf,"Angle",LIN_ENC_PHYSICAL);
  if(enc == NULL){
    printf("Failed to init encoding");
    return -1;
  }
  if(cs_lin_ldf_encoding_add_physical(enc,0,255,0.5,-10,"°")){
    printf("Failed to add physical");
    return -1;
  }

  enc = cs_lin_ldf_get_encoding(ldf,"Position Command",LIN_ENC_PHYSICAL);
  if(enc == NULL){
    printf("Failed to init encoding");
    return -1;
  }
  if(cs_lin_ldf_encoding_add_physical(enc,0,1024,0.136852,-15,"°")){
    printf("Failed to add physical");
    return -1;
  }

  
  /*Add Signal representations (signals with same encoding)*/
  rep = cs_lin_ldf_get_signalrepresentation(ldf,"DigitEncoding");
  if(rep == NULL){
    printf("Failed to init signal representation");
    return -1;
  }
  if(cs_lin_ldf_signalrepresentation_add_signal(rep,"Comm Error")){
    printf("Failed to add signal to representation");
    return -1;
  }
  if(cs_lin_ldf_signalrepresentation_add_signal(rep,"Alive Counter")){
    printf("Failed to add signal to representation");
    return -1;
  }
  if(cs_lin_ldf_signalrepresentation_add_signal(rep,"Actuator Faulted")){
    printf("Failed to add signal to representation");
    return -1;
  }
  if(cs_lin_ldf_signalrepresentation_add_signal(rep,"Calib Command Run")){
    printf("Failed to add signal to representation");
    return -1;
  }
  
  
  rep = cs_lin_ldf_get_signalrepresentation(ldf,"Angle");
  if(rep == NULL){
    printf("Failed to init signal representation");
    return -1;
  }
  if(cs_lin_ldf_signalrepresentation_add_signal(rep,"Angle")){
    printf("Failed to add signal to representation");
    return -1;
  }

  rep = cs_lin_ldf_get_signalrepresentation(ldf,"Position Command");
  if(rep == NULL){
    printf("Failed to init signal representation");
    return -1;
  }
  if(cs_lin_ldf_signalrepresentation_add_signal(rep,"Position Command")){
    printf("Failed to add signal to representation");
    return -1;
  }

  
  
  /*Order signal representation to node and low level definition*/
  cs_lin_raw_t siginit[8];
  memset(siginit,0,sizeof(cs_lin_raw_t)*8);

  def = cs_lin_ldf_get_signaldefinition(ldf,"Comm Error",siginit,1,2,"BRV");
  if(def == NULL){
    printf("Failed to add signal definition Comm Error");
    return -1;
  }
  if(cs_lin_ldf_signaldefinition_add_subscriber(def,"Master")){
    printf("Failed to add signal subscriber");
    return -1;
  }

  def = cs_lin_ldf_get_signaldefinition(ldf,"Alive Counter",siginit,1,2,"BRV");
  if(def == NULL){
    printf("Failed to add signal definition Alive Counter");
    return -1;
  }
  if(cs_lin_ldf_signaldefinition_add_subscriber(def,"Master")){
    printf("Failed to add signal subscriber");
    return -1;
  }

  def = cs_lin_ldf_get_signaldefinition(ldf,"Actuator Faulted",siginit,1,2,"BRV");
  if(def == NULL){
    printf("Failed to add signal definition Actuator Faulted");
    return -1;
  }
  if(cs_lin_ldf_signaldefinition_add_subscriber(def,"Master")){
    printf("Failed to add signal subscriber");
    return -1;
  }

  def = cs_lin_ldf_get_signaldefinition(ldf,"Angle",siginit,1,8,"BRV");
  if(def == NULL){
    printf("Failed to add signal definition Angle");
    return -1;
  }
  if(cs_lin_ldf_signaldefinition_add_subscriber(def,"Master")){
    printf("Failed to add signal subscriber");
    return -1;
  }

  def = cs_lin_ldf_get_signaldefinition(ldf,"Position Command",siginit,1,10,"Master");
  if(def == NULL){
    printf("Failed to add signal definition Position Command");
    return -1;
  }
  if(cs_lin_ldf_signaldefinition_add_subscriber(def,"BRV")){
    printf("Failed to add signal subscriber");
    return -1;
  }

  def = cs_lin_ldf_get_signaldefinition(ldf,"Calib Command Run",siginit,1,1,"Master");
  if(def == NULL){
    printf("Failed to add signal definition Calib Command Run");
    return -1;
  }
  if(cs_lin_ldf_signaldefinition_add_subscriber(def,"BRV")){
    printf("Failed to add signal subscriber");
    return -1;
  }

  
  /*Add frame definition*/
  frame = cs_lin_ldf_get_frame(ldf,"x28",0x28,2,"BRV");
  if(frame == NULL){
    printf("Failed to add frame definition");
    return -1;
  }
  if(cs_lin_ldf_frame_add_signal(frame,"Position Command",0) ||
     cs_lin_ldf_frame_add_signal(frame,"Calib Command Run",10)){
    printf("Failed to add signals to frame definition");
    return -1;
  }

  frame = cs_lin_ldf_get_frame(ldf,"x29",0x29,2,"BRV");
  if(frame == NULL){
    printf("Failed to add frame definition");
    return -1;
  }
  if(cs_lin_ldf_frame_add_signal(frame,"Comm Error",0) ||
     cs_lin_ldf_frame_add_signal(frame,"Alive Counter",1) ||
     cs_lin_ldf_frame_add_signal(frame,"Actuator Faulted",3) ||
     cs_lin_ldf_frame_add_signal(frame,"Angle",8)){
    printf("Failed to add signals to frame definition");
    return -1;
  }


  /*Create scheduler table, 1=search and create*/
  table = cs_lin_ldf_get_schedulertable(ldf,"DEFAULT",1);
  if(table == NULL){
    printf("Failed to add scheduler table definition");
    return -1;
  }
  
  /*Add frames to scheduler table
    Frame, every 50 ms*/
  if(cs_lin_ldf_schedulertable_add_entry(table,"x28",0.05) ||
     cs_lin_ldf_schedulertable_add_entry(table,"x29",0.05)){
    printf("Failed to add frames to scheduler table definition");
    return -1;
  }
  
  /*Add nodes
    Master and Node*/
  if(cs_lin_ldf_add_node(ldf,1,"Master",0.01,0.005) ||
     cs_lin_ldf_add_node(ldf,0,"BRV",0,0)){
    printf("Failed to add node definition");
    return -1;
  }
  
  /*Configure nodes, add addional infermation
    Only used for LDF creation*/
  attr = cs_lin_ldf_get_nodeattribute(ldf,"Master",2.2,0,0);
  if(attr == NULL){
    printf("Failed to add node attribute definition");
    return -1;
  }
  attr = cs_lin_ldf_get_nodeattribute(ldf,"BRV",2.2,0,0);
  if(attr == NULL){
    printf("Failed to add node attribute definition");
    return -1;
  }
  
  //Configuration is finished, look if it is valid
  if(cs_lin_ldf_apply_config(ldf)){
    printf("Failed to apply configuration\n");
    return -1;
  }




  
  
  sk = cs_lin_init();

  if(sk == NULL){
    printf("Failed to init lin stack\n");
    return -1;
  }

  if((sc = cs_socketcan_init()) == NULL){
    printf("Failed to init socketcan stack\n");
    return -1;
  }
    
  if(cs_socketcan_config_native(sc,interface,0)){
    printf("Failed to configure socketcan interface: %s\n",interface);
    return -1;
  }
    
  if(cs_socketcan_open(sc)){
    printf("Failed to open socketcan interface: %s\n",interface);
    return -1;
  }
   
  if(cs_lin_init_listener(sk,ldf)){
    printf("Failed to init slave listener\n");
    return -1;
  }

  //Store frame pointer to handle them in appication
  cs_lin_frame_t* f28 = cs_lin_get_frame_by_id(sk,0x28);
  cs_lin_frame_t* f29 = cs_lin_get_frame_by_id(sk,0x29);

  cs_lin_signal_t *s_28_poscmd = cs_lin_get_frame_signal(f28,"Position Command");
  cs_lin_signal_t *s_28_calibcmd = cs_lin_get_frame_signal(f28,"Calib Command Run");
  cs_lin_signal_t *s_29_commerror = cs_lin_get_frame_signal(f29,"Comm Error");
  cs_lin_signal_t *s_29_alivecounter = cs_lin_get_frame_signal(f29,"Alive Counter");
  cs_lin_signal_t *s_29_actuatorfaulted = cs_lin_get_frame_signal(f29,"Actuator Faulted");
  cs_lin_signal_t *s_29_angle = cs_lin_get_frame_signal(f29,"Angle");
  double v_28_poscmd = 0.0;
  double v_28_calibcmd = 0.0;
  double v_29_commerror = 0.0;
  double v_29_alivecounter = 0.0;
  double v_29_actuatorfaulted = 0.0;
  double v_29_angle = 0.0;
  
  while(1){
    uint32_t id = 0;
    uint8_t data[cs_socketcan_maxdlc(sc)];
    int len = 0;
    struct timeval tv;
    int loop = 1;
    int rv = 0;
    double time = cs_lin_get_time();
      
    if(cs_lin_process_start(sk,time)){
      printf("Process start error: %s\n",cs_lin_getlasterrorstring(sk));
    }
    
    //Receive loop
    while(loop){
      len = cs_socketcan_maxdlc(sc);
      rv = cs_socketcan_recv_nonblock(sc,&id,data,&len,&tv);
      
      switch (rv){
      case CS_SOCKETCAN_OK:
	if(cs_lin_process_msg(sk,id,data,len)){
	  printf("Failure during frame processing\n");
	}
	break;
      case CS_SOCKETCAN_ERROR:
	printf("Receive error: %s\n", cs_socketcan_getlasterrorstring(sc));
	loop = 0;
	break;
      case CS_SOCKETCAN_FAIL:
	loop = 0;
	break;
      case CS_SOCKETCAN_RTR:
	printf("%f: Receive Master request for frame 0x%X\n",time,id);
	break;
      case CS_SOCKETCAN_TIMEOUT:
	loop = 0;
	break;
      }
    }


    //Here application content
    printf("%f: F28 state: %i\n",time,cs_lin_get_frame_state(f28));
    printf("%f: F29 state: %i\n",time,cs_lin_get_frame_state(f29));
  
    
    if(cs_lin_get_frame_state(f28) == LIN_STATE_RECEIVED){
      if(cs_lin_get_signal_physical(sk,s_28_poscmd,&v_28_poscmd,1)){
	printf("%f: Failed to convert signal Position Cmd\n",time);
      }
      if(cs_lin_get_signal_physical(sk,s_28_calibcmd,&v_28_calibcmd,1)){
	printf("%f: Failed to convert signal Calibration Run  Cmd\n",time);
      }
      printf("%f: Frame 0x28: PositionCmd: %f, CalibrationRunCmd:%f\n",time,v_28_poscmd,v_28_calibcmd);
    }
    
    if(cs_lin_get_frame_state(f29) == LIN_STATE_RECEIVED){
      if(cs_lin_get_signal_physical(sk,s_29_commerror,&v_29_commerror,1)){
	printf("%f: Failed to convert signal CommError\n",time);
      }

      if(cs_lin_get_signal_physical(sk,s_29_alivecounter,&v_29_alivecounter,1)){
	printf("%f: Failed to convert signal AliveCounter\n",time);
      }

      if(cs_lin_get_signal_physical(sk,s_29_actuatorfaulted,&v_29_actuatorfaulted,1)){
	printf("%f: Failed to convert signal ActuatorFaulted\n",time);
      }

      if(cs_lin_get_signal_physical(sk,s_29_angle,&v_29_angle,1)){
	printf("%f: Failed to convert signal Angle\n",time);
      }
      printf("%f: Frame 0x29: CommError: %f, AliveCounter:%f, ActuatorFaulted: %f, Angle: %f\n",time,v_29_commerror,v_29_alivecounter,v_29_actuatorfaulted,v_29_angle);     
    }
      
    if(cs_lin_process_end(sk)){
      printf("Failure during process end\n");
    }
    usleep(10000);
  }

  cs_socketcan_close(sc);
  
  //All done free config
  cs_socketcan_free(sc);
  cs_lin_free(sk);
  cs_lin_ldf_free(ldf);
  
  return 0;
}
