/*  
    Copyright (C) 2017 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @example main_linldf.c
 * @author Andreas Stewering-Bone
 * @date 20 April 2017
 * @brief Simple ldf layer test programm.
 *
 * Simple ldf layer test programm.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <time.h>

#include "cs_lin.h"
#include "cs_lin_bit.h"
#include "cs_socketcan.h"

//Global variables
int ismaster = 1; //Default Is master
int sllin = 0; //Default do not use sllin interface
int verbose = 0; //Default only debug output
int use_bit = 0; //Use bitinterface

struct bit_ctx{
  cs_lin_phys_bit_t* bt;
  uint8_t *ecinbuffer;
  int ecinbufferlen;
  uint8_t ecoutcontenttosent;
  uint8_t *ecoutrawbuffer;
  int ecoutrawbufferlen;
  int ecoutrawworklen;
  uint8_t *ecoutbuffer;
  int ecoutbufferlen;
  int ecoutindex;
};


void ec_write(struct bit_ctx * btx)
{
  //Loop eoutrawbuffer and send each time a buffer of size ecoutbufferlen

  if(btx->ecoutcontenttosent == 0){
    //Nothing to send, reset out buffer, note to high bits
    memset(btx->ecoutbuffer,1,sizeof(uint8_t)*btx->ecoutbufferlen);
  } else {
    memcpy(btx->ecoutbuffer,&btx->ecoutrawbuffer[btx->ecoutindex],sizeof(uint8_t)*btx->ecoutbufferlen);
    btx->ecoutindex += btx->ecoutbufferlen;
  }

  //Check index position
  if(btx->ecoutindex > btx->ecoutrawbufferlen){
    //Reset
    btx->ecoutindex = 0;
    btx->ecoutcontenttosent = 0; //Send is finished
  }
  
  //Write btx->ecoutbuffer to EtherCAT

  int loffset = 0;
  uint8_t scan_pid = 0;
  uint8_t scan_data[CS_LIN_SIGNAL_MAX_BYTE];
  int scan_length = CS_LIN_SIGNAL_MAX_BYTE;
  uint8_t scan_crc = 0;
  
  while(cs_lin_phys_lin_scan(btx->bt,btx->ecoutbuffer,btx->ecoutbufferlen,
			     &loffset,&scan_pid,scan_data,&scan_length,&scan_crc) == 3){
    scan_length = CS_LIN_SIGNAL_MAX_BYTE;
  }  
}


void help(void)
{
  printf(
	 "Usage: testlinldf [OPTION]\n"
	 "Test lin LDF stack.\n"
	 "\n"
	 "Options:\n"
	 " -m   --master   Act as master (default)\n"
	 " -s   --slave    Act as slave\n"
	 " -i   --interface <name> CAN interface for testing\n"
	 " -b   --bit      Emulate bit generation interface\n"
	 " -l   --sllin    Use sslin interface\n"
	 " -v   --verbose  Verbose output\n"
	 "      --help     display this help and exit\n");
}


void socketcan_native_send(void *ctx, uint32_t id, uint8_t *data, int len, int enhanced)
{
  cs_socketcan_t *sc = (cs_socketcan_t *)ctx;
  uint32_t pid = 0; //LIN id with parity
  //uint8_t checksum = 0;
  
  if(sllin == 0){
    //Calculated the id parity
    cs_lin_id_parity(id,&pid);

    //Calc checksum
    //checksum = cs_lin_checksum(pid,data,len,enhanced);

    //Note in this example we can not use the checksum,
    //because of CAN limitation of max 8 bytes in frame.
    //So there is no space for checksum
  
    if(cs_socketcan_send(sc,pid,data,len,SOCKETCAN_SEND_CAN)){
      printf("Failed to send frame via socketcan\n");
    }
  } else {
    if(cs_socketcan_send(sc,id,data,len,SOCKETCAN_SEND_CAN)){
      printf("Failed to send frame via socketcan\n");
    }
  }
}

void bit_native_send(void *ctx, uint32_t id, uint8_t *data, int len, int enhanced)
{
  struct bit_ctx * btx = (struct bit_ctx *)ctx;
  cs_lin_phys_bit_t *bt = btx->bt;
  uint32_t pid = 0; //LIN id with parity
  uint8_t checksum = 0;
  int i;


  if(btx->ecoutcontenttosent == 1){
    printf("Collision at send\n");
  }
  
  //Calculated the id parity
  cs_lin_id_parity(id,&pid);
  
  //Calc checksum
  checksum = cs_lin_checksum(pid,data,len,enhanced);

  //Reset buffer length to maximum
  btx->ecoutrawworklen = btx->ecoutrawbufferlen;
  
  //Get bit field
  if(cs_lin_phys_bit_send(bt,btx->ecoutrawbuffer,&btx->ecoutrawworklen,pid,data,len,checksum)){
    printf("Failed to calculate bit field\n");    
  }
  
  
  btx->ecoutcontenttosent = 1; //There is something to send
  
  printf("Send data PID: 0x%X checksum: 0x%X len %d Bytes, Bitlen %i Bits:",pid,checksum,len,btx->ecoutrawworklen);
  for(i = 0; i < len; i++){
    printf("|%X",data[i]);
  }
  printf("\n");

}

void bit_native_send_req(void *ctx, uint32_t id, int enhanced)
{
  struct bit_ctx * btx = (struct bit_ctx *)ctx;
  cs_lin_phys_bit_t *bt = btx->bt;
  uint32_t pid = 0; //LIN id with parity
  

  if(btx->ecoutcontenttosent == 1){
    printf("Collision at send\n");
  }
  
  //Calculated the id parity
  cs_lin_id_parity(id,&pid);

  //Reset buffer length to maximum
  btx->ecoutrawworklen = btx->ecoutrawbufferlen;
  
  //Get bit field
  if(cs_lin_phys_bit_send_req(bt,btx->ecoutrawbuffer,&btx->ecoutrawworklen,pid)){
    printf("Failed to calculate bit field\n");    
  }

  
  btx->ecoutcontenttosent = 1; //There is something to send
  
  printf("Send request PID: 0x%X Bitlen %i Bits:\n",pid,btx->ecoutrawworklen);
}

void socketcan_native_send_req(void *ctx, uint32_t id, int enhanced)
{
  cs_socketcan_t *sc = (cs_socketcan_t *)ctx;
  uint32_t pid = 0; //LIN id with parity
  //uint8_t checksum = 0;
  
  if(sllin == 0){
    //Calculated the id parity
    cs_lin_id_parity(id,&pid);
      
    if(cs_socketcan_send_rtr(sc,pid)){
      printf("Failed to send RTR frame via socketcan\n");
    }
  } else {
    if(cs_socketcan_send_rtr(sc,id)){
      printf("Failed to send RTR frame via socketcan\n");
    }
  }
  
}

void slave_sllin_prepcache(void *ctx, uint32_t id, uint8_t *data, int len, int enhanced)
{
  //cs_socketcan_t *sc = (cs_socketcan_t *)ctx;
  
}

void socketcan_prepfilt(void *ctx, uint32_t *ids, int len)
{
  cs_socketcan_t *sc = (cs_socketcan_t *)ctx;
  int i;
  
  if(sllin == 0){
    for(i=0; i < len; i++){
      cs_lin_id_parity(ids[i],&ids[i]);
    }
    cs_socketcan_set_filter(sc,ids,len);
  } else {
    cs_socketcan_set_filter(sc,ids,len);
  }
}

int main(int argc, char **argv)
{
  int c;
  char *interface = NULL;
  cs_socketcan_t *sc = NULL;
  cs_lin_t *sk = NULL;
  cs_lin_ldf_t *ldf = NULL;
  cs_lin_enc_t* enc = NULL;
  cs_lin_enc_representation_t* rep = NULL;
  cs_lin_signal_def_t* def = NULL;
  cs_lin_frame_t* frame = NULL;
  cs_lin_schedule_table_t* table = NULL;
  cs_lin_nodeattribute_t* attr = NULL;
  const int bufferlen = 255; 
  char buffer[bufferlen];
  struct bit_ctx btx;
  
  while (1) {
    static struct option long_options[] = {
      /* These options set a flag. */
      {"interface",  required_argument, NULL, 'i'},
      {"master",  no_argument, NULL, 'm'},
      {"bit",  no_argument, NULL, 'b'},
      {"slave",   no_argument, NULL, 's'},
      {"sllin",   no_argument, NULL, 'l'},
      {"help",    no_argument, NULL, 'h'},
      {"verbose",    no_argument, NULL, 'v'},
      {0, 0, 0, 0}
    };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    c = getopt_long (argc, argv, "i:bhlmsv",
                     long_options, &option_index);

    /* Detect the end of the options. */
    if (c == -1)
      break;

    switch (c) {
    case 0:
      /* If this option set a flag, do nothing else now. */
      if (long_options[option_index].flag != 0)
        break;
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
      break;
    case 'm': ismaster = 1;    break;
    case 's': ismaster = 0;    break;
    case 'l': sllin = 1;    break;
    case 'b': use_bit = 1;    break;
    case 'v': verbose = 1;    break;
    case 'i': interface = optarg;    break;
    case 'h': help(); exit(0);    break;
    case '?':
      /* getopt_long already printed an error message. */
      printf( "Typ `testlinldf --help' for more information\n");
      return -1;
      break;
    default:
      return -1;
      break;
    }
  }

  if((use_bit == 0) && (interface == NULL)){
    printf("No interface defined\n");
    help();
    return -1;
  }
  
  //Set Debug level
  if(verbose == 1){
    cs_lin_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
    cs_socketcan_debuglevel(CS_DEBUG_LEVEL_DEBUG);
    cs_lin_bit_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
  } else {
    cs_lin_debuglevel(CS_DEBUG_LEVEL_DEBUG);
    cs_lin_bit_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
  }

  //Init structure
  ldf = cs_lin_ldf_init();

  if(ldf == NULL){
    printf("Failed to init structure\n");
    return -1;
  }

  //Set general informations
  if(cs_lin_ldf_set_general(ldf,2.1,2.1,19200)){
    printf("Failed to set LDF general information\n");
    return -1;
  }

  
  //Prepare configuration maually
  enc = cs_lin_ldf_get_encoding(ldf,"MyEncodingDynamic",LIN_ENC_DYNAMIC);
  if(enc == NULL){
    printf("Failed to init encoding\n");
    return -1;
  }
  if(cs_lin_ldf_encoding_add_physical(enc,10,255,0.1,-5,"V")){
    printf("Failed to add physical\n");
    return -1;
  }
  if(cs_lin_ldf_encoding_add_logical(enc,0,"Invalid")){
    printf("Failed to add logical\n");
    return -1;
  }

  enc = cs_lin_ldf_get_encoding(ldf,"MyEncodingDynamic2",LIN_ENC_DYNAMIC);
  if(enc == NULL){
    printf("Failed to init encoding\n");
    return -1;
  }
  if(cs_lin_ldf_encoding_add_physical(enc,10,255,0.1,-5,"V")){
    printf("Failed to add physical\n");
    return -1;
  }
  if(cs_lin_ldf_encoding_add_logical(enc,0,"Invalid")){
    printf("Failed to add logical\n");
    return -1;
  }
  
  enc = cs_lin_ldf_get_encoding(ldf,"MyEncodingASCII",LIN_ENC_ASCII);
  if(enc == NULL){
    printf("Failed to init encoding\n");
    return -1;
  }

  enc = cs_lin_ldf_get_encoding(ldf,"MyBitFlag",LIN_ENC_DYNAMIC);
  if(enc == NULL){
    printf("Failed to init encoding\n");
    return -1;
  }
  if(cs_lin_ldf_encoding_add_logical(enc,0,"OFF")){
    printf("Failed to add logical\n");
    return -1;
  }
  if(cs_lin_ldf_encoding_add_logical(enc,1,"ON")){
    printf("Failed to add logical\n");
    return -1;
  }
  
  //Add Signal representations
  rep = cs_lin_ldf_get_signalrepresentation(ldf,"MyEncodingDynamic");
  if(rep == NULL){
    printf("Failed to init signal representation\n");
    return -1;
  }
  if(cs_lin_ldf_signalrepresentation_add_signal(rep,"SignalPhys")){
    printf("Failed to add signal to representation\n");
    return -1;
  }
  if(cs_lin_ldf_signalrepresentation_add_signal(rep,"SignalPhys3")){
    printf("Failed to add signal to representation\n");
    return -1;
  }
  if(cs_lin_ldf_signalrepresentation_add_signal(rep,"SignalPhys4")){
    printf("Failed to add signal to representation\n");
    return -1;
  }
  rep = cs_lin_ldf_get_signalrepresentation(ldf,"MyEncodingDynamic2");
  if(rep == NULL){
    printf("Failed to init signal representation\n");
    return -1;
  }
  if(cs_lin_ldf_signalrepresentation_add_signal(rep,"SignalPhys2")){
    printf("Failed to add signal to representation\n");
    return -1;
  }
  rep = cs_lin_ldf_get_signalrepresentation(ldf,"MyEncodingASCII");
  if(rep == NULL){
    printf("Failed to init signal representation\n");
    return -1;
  }
  if(cs_lin_ldf_signalrepresentation_add_signal(rep,"SignalASCII")){
    printf("Failed to add signal to representation\n");
    return -1;
  }
  rep = cs_lin_ldf_get_signalrepresentation(ldf,"MyBitFlag");
  if(rep == NULL){
    printf("Failed to init signal representation\n");
    return -1;
  }
  if(cs_lin_ldf_signalrepresentation_add_signal(rep,"SignalBit")){
    printf("Failed to add signal to representation\n");
    return -1;
  }

  //Add signal definition
  cs_lin_raw_t siginit[8];
  memset(siginit,0,sizeof(cs_lin_raw_t)*8);
  def = cs_lin_ldf_get_signaldefinition(ldf,"SignalPhys",siginit,1,8,"MyNode");
  if(def == NULL){
    printf("Failed to add signal definition\n");
    return -1;
  }
  def = cs_lin_ldf_get_signaldefinition(ldf,"SignalPhys2",siginit,1,8,"MyNode");
  if(def == NULL){
    printf("Failed to add signal definition\n");
    return -1;
  }
  if(cs_lin_ldf_signaldefinition_add_subscriber(def,"MyMaster")){
    printf("Failed to add signal subscriber\n");
    return -1;
  }
  def = cs_lin_ldf_get_signaldefinition(ldf,"SignalPhys3",siginit,4,32,"MyNode3");
  if(def == NULL){
    printf("Failed to add signal definition\n");
    return -1;
  }
  if(cs_lin_ldf_signaldefinition_add_subscriber(def,"MyMaster")){
    printf("Failed to add signal subscriber\n");
    return -1;
  }
  def = cs_lin_ldf_get_signaldefinition(ldf,"SignalPhys4",siginit,2,16,"MyNode3");
  if(def == NULL){
    printf("Failed to add signal definition\n");
    return -1;
  }
  if(cs_lin_ldf_signaldefinition_add_subscriber(def,"MyMaster")){
    printf("Failed to add signal subscriber\n");
    return -1;
  }
  def = cs_lin_ldf_get_signaldefinition(ldf,"SignalASCII",siginit,1,4,"MyNode");
  if(def == NULL){
    printf("Failed to add signal definition\n");
    return -1;
  }
  if(cs_lin_ldf_signaldefinition_add_subscriber(def,"MyMaster")){
    printf("Failed to add signal subscriber\n");
    return -1;
  }
  if(cs_lin_ldf_signaldefinition_add_subscriber(def,"MyNode")){
    printf("Failed to add signal subscriber\n");
    return -1;
  }

  def = cs_lin_ldf_get_signaldefinition(ldf,"SignalBit",siginit,8,64,"MyNode2");
  if(def == NULL){
    printf("Failed to add signal definition\n");
    return -1;
  }
  if(cs_lin_ldf_signaldefinition_add_subscriber(def,"MyMaster")){
    printf("Failed to add signal subscriber\n");
    return -1;
  }
  if(cs_lin_ldf_signaldefinition_add_subscriber(def,"MyNode2")){
    printf("Failed to add signal subscriber\n");
    return -1;
  }

  //Add frame definition
  frame = cs_lin_ldf_get_frame(ldf,"MyFrame1",0x02,3,"MyNode");
  if(frame == NULL){
    printf("Failed to add frame definition\n");
    return -1;
  }
  if(cs_lin_ldf_frame_add_signal(frame,"SignalPhys",0) ||
     cs_lin_ldf_frame_add_signal(frame,"SignalPhys2",8) ||
     cs_lin_ldf_frame_add_signal(frame,"SignalASCII",16)){
    printf("Failed to add signals to frame definition\n");
    return -1;
  }
  //New frame but same signals, only shifted
  frame = cs_lin_ldf_get_frame(ldf,"MyFrame2",0x03,4,"MyNode");
  if(frame == NULL){
    printf("Failed to add frame definition\n");
    return -1;
  }
  if(cs_lin_ldf_frame_add_signal(frame,"SignalPhys",3) ||
     cs_lin_ldf_frame_add_signal(frame,"SignalPhys2",11) ||
     cs_lin_ldf_frame_add_signal(frame,"SignalASCII",21)){
    printf("Failed to add signals to frame definition\n");
    return -1;
  }

  //New frame but same signals, but send by master
  frame = cs_lin_ldf_get_frame(ldf,"MyFrame3",0x04,3,"MyMaster");
  if(frame == NULL){
    printf("Failed to add frame definition\n");
    return -1;
  }
  if(cs_lin_ldf_frame_add_signal(frame,"SignalPhys",0) ||
     cs_lin_ldf_frame_add_signal(frame,"SignalPhys2",9) ||
     cs_lin_ldf_frame_add_signal(frame,"SignalASCII",17)){
    printf("Failed to add signals to frame definition\n");
    return -1;
  }

  frame = cs_lin_ldf_get_frame(ldf,"MyFrame4",0x10,8,"MyNode2");
  if(frame == NULL){
    printf("Failed to add frame definition\n");
    return -1;
  }
  if(cs_lin_ldf_frame_add_signal(frame,"SignalBit",0)){
    printf("Failed to add signals to frame definition\n");
    return -1;
  }

  frame = cs_lin_ldf_get_frame(ldf,"MyFrame5",0x11,6,"MyNode3");
  if(frame == NULL){
    printf("Failed to add frame definition\n");
    return -1;
  }
  if(cs_lin_ldf_frame_add_signal(frame,"SignalPhys3",0)){
    printf("Failed to add signals to frame definition\n");
    return -1;
  }
  if(cs_lin_ldf_frame_add_signal(frame,"SignalPhys4",32)){
    printf("Failed to add signals to frame definition\n");
    return -1;
  }
  
  table = cs_lin_ldf_get_schedulertable(ldf,"DEFAULT",1);
  if(table == NULL){
    printf("Failed to add scheduler table definition\n");
    return -1;
  }
  if(cs_lin_ldf_schedulertable_add_entry(table,"MyFrame1",0.05) ||
     cs_lin_ldf_schedulertable_add_entry(table,"MyFrame2",0.1) ||
     cs_lin_ldf_schedulertable_add_entry(table,"MyFrame3",0.15) ||
     cs_lin_ldf_schedulertable_add_entry(table,"MyFrame4",0.05) ||
     cs_lin_ldf_schedulertable_add_entry(table,"MyFrame5",0.05)){
    printf("Failed to add frames to scheduler table definition\n");
    return -1;
  }

  //Add nodes
  if(cs_lin_ldf_add_node(ldf,1,"MyMaster",0.1,0.005) ||
     cs_lin_ldf_add_node(ldf,0,"MyNode",0,0) ||
     cs_lin_ldf_add_node(ldf,0,"MyNode2",0,0) ||
     cs_lin_ldf_add_node(ldf,0,"MyNode3",0,0)){
    printf("Failed to add node definition\n");
    return -1;
  }

  //Configure nodes
  attr = cs_lin_ldf_get_nodeattribute(ldf,"MyMaster",2.2,0,0);
  if(attr == NULL){
    printf("Failed to add node attribute definition\n");
    return -1;
  }
  attr = cs_lin_ldf_get_nodeattribute(ldf,"MyNode",1.3,0,0);
  if(attr == NULL){
    printf("Failed to add node attribute definition\n");
    return -1;
  }
  attr = cs_lin_ldf_get_nodeattribute(ldf,"MyNode2",2.1,0,0);
  if(attr == NULL){
    printf("Failed to add node attribute definition\n");
    return -1;
  }
   attr = cs_lin_ldf_get_nodeattribute(ldf,"MyNode3",1.3,0,0);
  if(attr == NULL){
    printf("Failed to add node attribute definition\n");
    return -1;
  }

  //Configuration is finished, look if it is valid
  if(cs_lin_ldf_apply_config(ldf)){
    printf("Failed to apply configuration\n");
    return -1;
  }




  
  
  sk = cs_lin_init();

  if(sk == NULL){
    printf("Failed to init lin stack\n");
    return -1;
  }

  if(use_bit == 0){
    if((sc = cs_socketcan_init()) == NULL){
      printf("Failed to init socketcan stack\n");
      return -1;
    }
    
    if(cs_socketcan_config_native(sc,interface,0)){
      printf("Failed to configure socketcan interface: %s\n",interface);
      return -1;
    }
    
    if(cs_socketcan_open(sc)){
      printf("Failed to open socketcan interface: %s\n",interface);
      return -1;
    }
  } else {
    btx.bt = cs_lin_phys_bit_init(ldf->speed,1000000);
    if(btx.bt == NULL){
      printf("Failed to init bit interface\n");
    }
    btx.ecinbufferlen = 1000;
    btx.ecinbuffer = NULL;
    btx.ecinbuffer = (uint8_t*)malloc(sizeof(uint8_t)*btx.ecinbufferlen);
    if(btx.ecinbuffer == NULL){
      printf("Failed to malloc bit ec in buffer\n");
    }
    btx.ecoutbufferlen = 1000; //1000 Bits
    btx.ecoutbuffer = NULL;
    btx.ecoutbuffer = (uint8_t*)malloc(sizeof(uint8_t)*btx.ecoutbufferlen);
    if(btx.ecoutbuffer == NULL){
      printf("Failed to malloc bit ec out buffer\n");
    }
    btx.ecoutindex = 0;
    btx.ecoutcontenttosent = 0; //Nothing to send
    
    btx.ecoutrawbufferlen = cs_lin_phys_bit_get_buffer_size(btx.bt)+btx.ecoutbufferlen; //This is a hack to get a fast sub copy
    btx.ecoutrawbuffer = NULL;
    btx.ecoutrawbuffer = (uint8_t*)malloc(sizeof(uint8_t)*btx.ecoutrawbufferlen);
    if(btx.ecoutrawbuffer == NULL){
      printf("Failed to malloc bit ec out raw buffer\n");
    }
    btx.ecoutbufferlen = 1000; //1000 Bits
    btx.ecoutbuffer = NULL;
    btx.ecoutbuffer = (uint8_t*)malloc(sizeof(uint8_t)*btx.ecoutbufferlen);
    if(btx.ecoutbuffer == NULL){
      printf("Failed to malloc bit ec out buffer\n");
    }
    btx.ecoutindex = 0;
  }
    
  if(ismaster){
    //master

    if(use_bit == 0){
      if(sllin == 0){
	cs_lin_set_callback(sk,(void*)sc,0,socketcan_prepfilt,socketcan_native_send,socketcan_native_send_req,NULL);
      } else {
	cs_lin_set_callback(sk,(void*)sc,0,socketcan_prepfilt,socketcan_native_send,socketcan_native_send_req,NULL);
      }
    } else {
      cs_lin_set_callback(sk,(void*)&btx,0,NULL,bit_native_send,bit_native_send_req,NULL);
    }
      
    if(cs_lin_init_master(sk,ldf,"DEFAULT")){
      printf("Failed to init master with scheduling table: %s\n","DEFAULT");
      return -1;
    }

    //Store frame pointer to handle them in appication
    cs_lin_frame_t* f1 = cs_lin_get_frame_by_id(sk,0x02);
    cs_lin_frame_t* f2 = cs_lin_get_frame_by_id(sk,0x03);
    cs_lin_frame_t* f3 = cs_lin_get_frame_by_id(sk,0x04);
    cs_lin_frame_t* f4 = cs_lin_get_frame_by_id(sk,0x10);
    cs_lin_frame_t* f5 = cs_lin_get_frame_by_id(sk,0x11);
    
    //Start cycle
    while(1){
      uint32_t id = 0;
      uint8_t data[cs_socketcan_maxdlc(sc)];
      int len = 0;
      struct timeval tv;
      int loop = 1;
      int rv = 0;
      cs_lin_frame_t *lf = NULL;
      int i;
      
      if((rv = cs_lin_process_start(sk,cs_lin_get_time()))> 0){
	switch(rv){
	case 1:
	  printf("Process start error: %s\n",cs_lin_getlasterrorstring(sk));
	  break;
	case 2:
	  printf("Process start frame timeout\n");
	  break;
	}
      }

      
      if(use_bit == 0){
	//Receive loop
	while(loop){
	  len = cs_socketcan_maxdlc(sc);
	  rv = cs_socketcan_recv_nonblock(sc,&id,data,&len,&tv);
	  switch (rv){
	  case CS_SOCKETCAN_OK:
	    if(sllin == 0){
	      //Emulate checksum
	      lf = cs_lin_get_frame_by_id(sk,CS_LIN_GET_ID(id));
	      if(lf == NULL){
		break;
	      }
	    
	      if(cs_lin_process_msg_raw(sk,(uint8_t)id,data,len,cs_lin_checksum(id,data,len,cs_lin_frame_is_enhanced(lf)))){
		printf("Failure during frame processing\n");
	      }
	    } else {
	      if(cs_lin_process_msg(sk,id,data,len)){
		printf("Failure during frame processing\n");
	      }
	    }
	    break;
	  case CS_SOCKETCAN_ERROR:
	    printf("Receive error: %s\n", cs_socketcan_getlasterrorstring(sc));
	    loop = 0;
	    break;
	  case CS_SOCKETCAN_FAIL:
	    loop = 0;
	    break;
	  case CS_SOCKETCAN_RTR:
	    printf("Receive RTR frame\n");
	    break;
	  case CS_SOCKETCAN_TIMEOUT:
	    loop = 0;
	    break;
	  }
	}
      } else {
	//Here EtherCAT receive
      }
      //Here application content
      //Receive
      //printf("F1 state: %i\n",cs_lin_get_frame_state(f1));
      //printf("F2 state: %i\n",cs_lin_get_frame_state(f2));
      //Send
      //printf("F3 state: %i\n",cs_lin_get_frame_state(f3));

      if(cs_lin_get_frame_state(f1) == LIN_STATE_RECEIVED){
	printf("Received signal frame1 from slave\n");
	cs_lin_reset_frame_state(f1);
      }
      
      if(cs_lin_get_frame_state(f2) == LIN_STATE_RECEIVED){
	printf("Received signal frame2 from slave\n");
	cs_lin_reset_frame_state(f2);
	cs_lin_signal_t *s = NULL;
	cs_lin_signal_t *s2 = NULL;
	double v = -1.0;
	double v2 = -1.0;
	s = cs_lin_get_frame_signal(f2,"SignalPhys");
	s2 = cs_lin_get_frame_signal(f2,"SignalPhys2");
	if((s == NULL) || (s2 == NULL)){
	  printf("Invalid signal\n");
	}
	if(cs_lin_get_signal_physical(sk,s,&v,1)){
	  printf("Failed to convert signal physically, try logical\n");
	  if(cs_lin_get_signal_logical(sk,s,buffer,bufferlen,0)){
	    printf("Failed to convert signal logically\n");
	  } else {
	    if(strncmp(buffer,"Invalid",bufferlen) == 0){
	      printf("Signal is marked invalid\n");
	    }
	  }
	} else {
	  printf("Updated signal SignalPhys: %f\n",v);
	  v+=0.2;
	  printf("Incremented signal SignalPhys: %f\n",v);
	}

	if(cs_lin_get_signal_physical(sk,s2,&v2,1)){
	  printf("Failed to convert signal physically, try logical\n");
	  if(cs_lin_get_signal_logical(sk,s2,buffer,bufferlen,0)){
	    printf("Failed to convert signal logically\n");
	  } else {
	    if(strncmp(buffer,"Invalid",bufferlen) == 0){
	      printf("Signal is marked invalid\n");
	    }
	  }
	} else {
	  printf("Updated signal SignalPhys2: %f\n",v2);
	  v2+=0.2;
	  printf("Incremented signal SignalPhys2: %f\n",v2);
	}
	  
	s = cs_lin_get_frame_signal(f3,"SignalPhys");
	s2 = cs_lin_get_frame_signal(f3,"SignalPhys2");
	if((s == NULL)||(s2 == NULL)){
	  printf("Invalid signal\n");
	}
	if(cs_lin_update_signal_physical(sk,s,&v,1)){
	  printf("Failed to update signal physical, set invalid\n");
	  v=-1.0;
	  if(cs_lin_update_signal_logical(sk,s,"Invalid",0)){
	    printf("Failed to update signal logical\n");
	  }
	}
	if(cs_lin_update_signal_physical(sk,s2,&v2,1)){
	  printf("Failed to update signal physical, set invalid\n");
	  v2=-1.0;
	  if(cs_lin_update_signal_logical(sk,s2,"Invalid",0)){
	    printf("Failed to update signal logical\n");
	  }
	}
      }
      if(cs_lin_get_frame_state(f4) == LIN_STATE_RECEIVED){
	cs_lin_signal_t *s4 = cs_lin_get_frame_signal(f4,"SignalBit");
	printf("Received SignalBit vector:");
	for(i = 0; i < 8; i++){ 
	  if(cs_lin_get_signal_logical(sk,s4,buffer,bufferlen,i)){
	    printf("Failed to convert signal logically\n");
	  }
	  printf("%s,",buffer);
	}
	printf("\n");
      }
      if(cs_lin_get_frame_state(f5) == LIN_STATE_RECEIVED){
	cs_lin_signal_t *s5 = cs_lin_get_frame_signal(f5,"SignalPhys3");
	int i;
	double vl[4];
	cs_lin_get_signal_physical(sk,s5,vl,4);
	printf("Frame5 SignalPhys3: ");
	for(i = 0; i<4; i++ ){
	  printf("%f,",vl[i]);
	}
	printf("\n");
	cs_lin_signal_t *s6 = cs_lin_get_frame_signal(f5,"SignalPhys4");
	cs_lin_get_signal_physical(sk,s6,vl,2);
	printf("Frame5 SignalPhys4: ");
	for(i= 0; i<2; i++ ){
	  printf("%f,",vl[i]);
	}
	printf("\n");
      }
      
      if(cs_lin_process_end(sk)){
	printf("Failure during process end\n");
      }

      //Process send buffer (EtherCAT)
      ec_write(&btx);
      
      usleep(1000);
    }
    
  } else {
    //slave

    if(use_bit == 0){
      if(sllin == 0){
	cs_lin_set_callback(sk,(void*)sc,0,socketcan_prepfilt,socketcan_native_send,socketcan_native_send_req,NULL);
      } else {
	cs_lin_set_callback(sk,(void*)sc,0,socketcan_prepfilt,socketcan_native_send,socketcan_native_send_req,slave_sllin_prepcache);
      }
    } else {
      cs_lin_set_callback(sk,(void*)&btx,0,NULL,bit_native_send,bit_native_send_req,NULL);
    }
    
    const char* nodes[] = {"MyNode","MyNode2","MyNode3"};
    if(cs_lin_init_slave(sk,ldf,nodes,3)){
      printf("Failed to init slave configuration\n");
      return -1;
    }

    //Store frame pointer to handle them in appication
    cs_lin_frame_t* f1 = cs_lin_get_frame_by_id(sk,0x02);
    cs_lin_frame_t* f2 = cs_lin_get_frame_by_id(sk,0x03);
    cs_lin_frame_t* f3 = cs_lin_get_frame_by_id(sk,0x04);
    cs_lin_frame_t* f4 = cs_lin_get_frame_by_id(sk,0x10);
    cs_lin_frame_t* f5 = cs_lin_get_frame_by_id(sk,0x11);
    uint8_t counter = 0;
    
    //Start cycle
    while(1){
      uint32_t id = 0;
      uint8_t data[cs_socketcan_maxdlc(sc)];
      int len = 0;
      struct timeval tv;
      int loop = 1;
      int rv = 0;
      cs_lin_frame_t *lf = NULL;
      
      if(cs_lin_process_start(sk,cs_lin_get_time())){
	printf("Process start error: %s\n",cs_lin_getlasterrorstring(sk));
      }

      if(use_bit == 0){
	//Receive loop
	while(loop){
	  len = cs_socketcan_maxdlc(sc);
	  rv = cs_socketcan_recv_nonblock(sc,&id,data,&len,&tv);
	  
	  switch (rv){
	  case CS_SOCKETCAN_OK:
	    if(sllin == 0){
	      //Emulate checksum
	      lf = cs_lin_get_frame_by_id(sk,CS_LIN_GET_ID(id));
	      if(lf == NULL){
		break;
	      }
	      
	      if(cs_lin_process_msg_raw(sk,(uint8_t)id,data,len,cs_lin_checksum(id,data,len,cs_lin_frame_is_enhanced(lf)))){
		printf("Failure during frame processing\n");
	      }
	      loop = 0;
	    } else {
	      if(cs_lin_process_msg(sk,id,data,len)){
		printf("Failure during frame processing\n");
	      }
	    }
	    break;
	  case CS_SOCKETCAN_ERROR:
	    printf("Receive error: %s\n", cs_socketcan_getlasterrorstring(sc));
	    loop = 0;
	    break;
	  case CS_SOCKETCAN_FAIL:
	    loop = 0;
	    break;
	  case CS_SOCKETCAN_RTR:
	    if(sllin == 0){
	      //Emulate checksum
	      lf = cs_lin_get_frame_by_id(sk,CS_LIN_GET_ID(id));
	      if(lf == NULL){
		break;
	      }
	      
	      if(cs_lin_process_msg_raw(sk,(uint8_t)id,data,len,cs_lin_checksum(id,data,len,cs_lin_frame_is_enhanced(lf)))){
		printf("Failure during frame processing\n");
	      }
	      loop = 0;
	    } else {
	      if(cs_lin_process_msg(sk,id,data,len)){
		printf("Failure during frame processing\n");
	      }
	    }
	    break;
	  case CS_SOCKETCAN_TIMEOUT:
	    loop = 0;
	    break;
	  }
	}
      } else {

      }

      //Here application content
      //printf("F1 state: %i\n",cs_lin_get_frame_state(f1));
      //printf("F2 state: %i\n",cs_lin_get_frame_state(f2));
      //printf("F3 state: %i\n",cs_lin_get_frame_state(f3));

      
      if(cs_lin_get_frame_state(f3) == LIN_STATE_RECEIVED){
	printf("Received signal frame from master\n");
	cs_lin_get_frame_state(f3);
	cs_lin_signal_t *s = NULL;
	cs_lin_signal_t *s2 = NULL;
	double v = -3.0;
	double v2 = -3.0;
	s = cs_lin_get_frame_signal(f3,"SignalPhys");
	s2 = cs_lin_get_frame_signal(f3,"SignalPhys2");
	if((s == NULL)||(s2 == NULL)){
	  printf("Invalid signal\n");
	}
	if(cs_lin_get_signal_physical(sk,s,&v,1)){
	  printf("Failed to convert signal\n");
	  if(cs_lin_get_signal_logical(sk,s,buffer,bufferlen,0)){
	    printf("Failed to convert signal logically\n");
	  }
	  if(strncmp(buffer,"Invalid",bufferlen) == 0){
	    printf("Signal is marked invalid\n");
	  }	  
	} else {
	  printf("Updated signal SignalPhys: %f\n",v);
	  v-=0.1;
	  printf("Incremented signal SignalPhys: %f\n",v);
	}
	if(cs_lin_get_signal_physical(sk,s2,&v2,1)){
	  printf("Failed to convert signal\n");
	  if(cs_lin_get_signal_logical(sk,s2,buffer,bufferlen,0)){
	    printf("Failed to convert signal logically\n");
	  }
	  if(strncmp(buffer,"Invalid",bufferlen) == 0){
	    printf("Signal is marked invalid\n");
	  }	  
	} else {
	  printf("Updated signal SignalPhys2: %f\n",v2);
	  v2-=0.1;
	  printf("Incremented signal SignalPhys2: %f\n",v2);
	}
	  
	s = cs_lin_get_frame_signal(f2,"SignalPhys");
	s2 = cs_lin_get_frame_signal(f2,"SignalPhys2");
	if((s == NULL)||(s2 == NULL)){
	  printf("Invalid signal\n");
	}
	if(cs_lin_update_signal_physical(sk,s,&v,1)){
	  printf("Failed to update signal physically, set invalid\n");
	  v=-3;
	  if(cs_lin_update_signal_logical(sk,s,"Invalid",0)){
	    printf("Failed to update signal logical\n");
	  }
	}
	if(cs_lin_update_signal_physical(sk,s2,&v2,1)){
	  printf("Failed to update signal physically, set invalid\n");
	  v2=-3;
	  if(cs_lin_update_signal_logical(sk,s2,"Invalid",0)){
	    printf("Failed to update signal logical\n");
	  }
	}
      }
      if(cs_lin_get_frame_state(f4) == LIN_STATE_RECEIVED){
	printf("Received signal frame Frame4 request from master\n");
	cs_lin_signal_t *s4 = cs_lin_get_frame_signal(f4,"SignalBit");
	int i;
	for(i = 0; i < 8; i++){
	  counter++;
	  printf("Bit %d: 0x%X \n",i,(counter%3)>=1);
	  if(cs_lin_update_signal_logical(sk,s4,(counter%3)>=1?"ON":"OFF",i)){
	    printf("Failed to update signal logical\n");
	  }
	}
      }
      if(cs_lin_get_frame_state(f5) == LIN_STATE_RECEIVED){
	printf("Received signal frame Frame5 request from master\n");
	cs_lin_signal_t *s5 = cs_lin_get_frame_signal(f5,"SignalPhys3");
	cs_lin_signal_t *s6 = cs_lin_get_frame_signal(f5,"SignalPhys4");
	double vl[4];
	vl[0] = 1.1;
	vl[1] = 5.1;
	vl[2] = 13.1;
	vl[3] = 17.1;
	if(cs_lin_update_signal_physical(sk,s5,vl,4)){
	  printf("Failed to update SignalPhys3\n");
	}
	if(cs_lin_update_signal_physical(sk,s6,vl,2)){
	  printf("Failed to update SignalPhys4\n");
	}
      }
    
      if(cs_lin_process_end(sk)){
	printf("Failure during process end\n");
      }
      usleep(1000);
    }
  }

  if(use_bit == 0){
    cs_socketcan_close(sc);
    
    //All done free config
    cs_socketcan_free(sc);
  } else {
    cs_lin_phys_bit_free(btx.bt);
  }
  cs_lin_free(sk);
  cs_lin_ldf_free(ldf);
  
  return 0;
}
