/*  
    Copyright (C) 2018 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @example main_signal_encode_decode.c
 * @author Andreas Stewering-Bone
 * @date 20 March 2018
 * @brief Simple Signal en- and decoding to test library stack.
 *
 * Simple Signal en- and decoding to test library stack.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <time.h>
#include <math.h>

#include "cs_signal.h"
#include "cs_socketcan.h"

/*Libefence setting*/
/*
extern int EF_PROTECT_BELOW = 1;
extern int EF_PROTECT_FREE = 1;
extern int EF_FILL = 1; 
*/

#define NUM_SIGNAL 4

#define BYTE_TO_BINARY_PATTERN "%c%c%c%c%c%c%c%c "
#define BYTE_TO_BINARY(byte)  \
  (byte & 0x80 ? '1' : '0'), \
  (byte & 0x40 ? '1' : '0'), \
  (byte & 0x20 ? '1' : '0'), \
  (byte & 0x10 ? '1' : '0'), \
  (byte & 0x08 ? '1' : '0'), \
  (byte & 0x04 ? '1' : '0'), \
  (byte & 0x02 ? '1' : '0'), \
  (byte & 0x01 ? '1' : '0')

#define BYTE_TO_BINARY_REVERTED(byte)  \
  (byte & 0x01 ? '1' : '0'), \
  (byte & 0x02 ? '1' : '0'), \
  (byte & 0x04 ? '1' : '0'), \
  (byte & 0x08 ? '1' : '0'), \
  (byte & 0x10 ? '1' : '0'), \
  (byte & 0x20 ? '1' : '0'), \
  (byte & 0x40 ? '1' : '0'), \
  (byte & 0x80 ? '1' : '0')

int debug = 0;
int binary = 0;
int listenonly=0;
int maxdlc = 8; //Note for CAN 2.0

void help(void)
{
  printf(
          "Usage: testsignalencodedecode [OPTION]\n"
          "Test library signal en- and decoding stack.\n"
          "\n"
          "Options:\n"
	  "  -i  --interface NAME Optional: Name of SocketCAN interface to send data\n"
	  "  -l  --listen     Listen on SocketCAN interface, -i is required\n"
	  "  -b  --binary     Binary output\n"
	  "  -v  --verbose    Enable Runtime debugging, can be set multimple times\n"
          "      --help       display this help and exit\n");
}

void print_bitmask(double time, cs_signal_t *s)
{
  int i;
  int mb =  maxdlc * 8;
  printf("%f: Signal: %s Bitmask:",time,cs_signal_get_name(s));
  for(i = 0; i < mb; i++){
    if((i < cs_signal_get_bitoffset(s)) || (i >= (cs_signal_get_bitoffset(s) + cs_signal_get_bitlength(s)))){
      if((i % 8) == 0){
	printf(" 0");
      } else {
	printf("0");
      }
    } else {
      if((i % 8) == 0){
	printf(" 1");
      } else {
	printf("1");
      }
    }
  }
  printf("\n");
}

void print_buffer(double time, uint8_t *buffer, int len)
{
  int i;
  printf("%f: ",time);
  for(i = 0; i < len; i++){
    if(binary == 1){
      printf(BYTE_TO_BINARY_PATTERN,BYTE_TO_BINARY(buffer[i]));
    } else {
      printf("%02X ",buffer[i]);
    }
  }
  printf("\n");
  
}

int check_range(cs_signal_t *s,uint8_t *buffer, int len)
{
  int start_offset = (int)(cs_signal_get_bitoffset(s) & 7);
  int start_byte   = (int)(cs_signal_get_bitoffset(s) / 8);
  int end_offset = (int)((cs_signal_get_bitoffset(s) + cs_signal_get_bitlength(s))& 7);
  int end_byte   = (int)((cs_signal_get_bitoffset(s) + cs_signal_get_bitlength(s)) / 8);
  int i,j;

  if((start_byte >= maxdlc) ||
     (end_byte >= maxdlc)){
    printf("Detect range failure: Byte\n");
    return 1;
  }
  
  for(i = 0; i< start_byte; i++){
    if(buffer[i] != 0){
      printf("Detect range failure: Byte start\n");
      return 1;
    }
  }

  for(i = start_byte; i<= end_byte; i++){
    if(i == start_byte){
      for(j = 0; j < start_offset; j++){
	if(buffer[i] & (1<< j)){
	  printf("Detect range failure: Bit start\n");
	  return 1;
	}
      }
    }
    if(i == end_byte){
      for(j = end_offset; j < 8; j++){
	if(buffer[i] & (1<< j)){
	  printf("Detect range failure: Bit stop\n");
	  return 1;
	}
      }
    }
  }
  
  for(i = end_byte +1; i< maxdlc; i++){
    if(buffer[i] != 0){
      printf("Detect range failure: Byte end\n");
      return 1;
    }
  }  

  return 0;
}

int main(int argc, char **argv)
{
  int c;
  uint8_t *mbuffer = NULL;
  cs_signal_t **msignal = NULL;
  double *rsinv = NULL;
  char *iface = NULL;
  cs_socketcan_t *sk = NULL;
  double time_step = 0.1;
  useconds_t utime_step = (useconds_t)(time_step*1e6);
  double time = -100.0;
  double ampl = 10.0;
  double freq = 1.0;
  double scale = 0.1;
  double offset = 0.0;
  int i;
  double sinv;
  
  while (1) {
    static struct option long_options[] = {
      /* These options set a flag. */
      {"interface",    required_argument, NULL, 'i'},
      {"listen",    no_argument, NULL,'l'},
      {"binary",    no_argument, NULL, 'b'},
      {"help",    no_argument, NULL, 'h'},
      {"verbose",  no_argument, NULL, 'v'},
      {0, 0, 0, 0}
    };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    c = getopt_long (argc, argv, "i:lbhv",
                     long_options, &option_index);

    /* Detect the end of the options. */
    if (c == -1)
      break;


    switch (c) {
    case 0:
      /* If this option set a flag, do nothing else now. */
      if (long_options[option_index].flag != 0)
        break;
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
      break;
    case 'i': iface = optarg;     break;
    case 'b': binary = 1;     break;
    case 'l': listenonly = 1;     break;
    case 'v': debug +=1;     break;
    case 'h': help(); exit(0);    break;
    case '?':
      /* getopt_long already printed an error message. */
      printf( "Typ `testsignalencodedecode --help' for more information\n");
      return -1;
      break;
    default:
      return -1;
      break;
    }
  }

  printf("Debuglevel %d\n",debug);
  
  /*Do alloctaion for memory checking*/
  mbuffer = (uint8_t *)calloc(sizeof(char), maxdlc);
  msignal= (cs_signal_t**)calloc(sizeof(cs_signal_t*),NUM_SIGNAL);
  rsinv = (double*)calloc(sizeof(double),NUM_SIGNAL);

  if((mbuffer == NULL)||
     (msignal == NULL) ||
     (rsinv == NULL)){
    printf("Malloc failure\n");
    return -1;
  }
  

  if(debug > 1){
    cs_signal_debuglevel(CS_DEBUG_LEVEL_DEBUG);
    if(iface != NULL){
      cs_socketcan_debuglevel(CS_DEBUG_LEVEL_DEBUG);
    }
  }
  if(debug > 2){
    cs_signal_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
    if(iface != NULL){
      cs_socketcan_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
    }
  }

  for(i = 0; i < NUM_SIGNAL; i++){
    msignal[i] = cs_signal_init();
  }
  
  if(
     cs_signal_set(msignal[0],"Testsignalmb","mm",55,16,SIGNAL_BT_BIG,SIGNAL_MUX_SIGNAL,SIGNAL_TYPE_UNSIGNED,scale,offset,-100000.0,100000.0)||
     cs_signal_set(msignal[1],"Testsignalml","mm",16,16,SIGNAL_BT_LITTLE,SIGNAL_MUX_SIGNAL,SIGNAL_TYPE_UNSIGNED,scale,offset,-100000.0,100000.0)||
     cs_signal_set(msignal[2],"Testsignalmb1","mm",39,15,SIGNAL_BT_BIG,SIGNAL_MUX_SIGNAL,SIGNAL_TYPE_UNSIGNED,scale,offset,-100000.0,100000.0)||
     cs_signal_set(msignal[1],"Testsignalml","mm",16,16,SIGNAL_BT_LITTLE,SIGNAL_MUX_SIGNAL,SIGNAL_TYPE_UNSIGNED,scale,offset,-100000.0,100000.0)||
     cs_signal_set(msignal[1],"Time","s",0,16,SIGNAL_BT_LITTLE,SIGNAL_MUX_SIGNAL,SIGNAL_TYPE_UNSIGNED,0.1,0.0,-3200.0,3200.0)
     ){
    printf("Failure during signal configuration\n");
    return -1;
  }

  for(i = 0; i < NUM_SIGNAL; i++){
    if(cs_signal_check(msignal[i],0)){
      printf("Message check failed for signal %s\n",cs_signal_get_name(msignal[i]));
      return -1;
    }
  }
  
  if((listenonly == 1) && (iface == NULL)){
    printf("Listen only is only possible with a defined SocketCAN interface\n");
    return -1;
  }

  
  if(iface != NULL){
    sk = cs_socketcan_init();
    if(sk == NULL){
      printf("Failed to init SocketCAN\n");
      return -1;
    }
    if(cs_socketcan_config_native(sk,iface,0)){
      printf("Failed to config SocketCAN interface\n");
      return -1;
    }

    if(cs_socketcan_open(sk)){
      printf("Failed to open SocketCAN interface\n");
      return -1;
    }
  }

  sinv=6553.6 / 2;
  if(listenonly == 0){
    while(1){
      //sinv = ampl * sin(2*M_PI*freq*time);
      sinv = sinv - 0.1;
      memset(mbuffer,0,sizeof(uint8_t)*maxdlc);
      for(i = 0; i < NUM_SIGNAL; i++){
	if(i == (NUM_SIGNAL -1 )){
	  cs_signal_runtime_encode(msignal[i],time,mbuffer,maxdlc,0);
	} else {
	  cs_signal_runtime_encode(msignal[i],sinv,mbuffer,maxdlc,0);
	}
      }
      if(iface != NULL){
	if(cs_socketcan_send(sk,5,mbuffer,maxdlc,SOCKETCAN_SEND_CAN)){
	  printf("%f: Failed to send SocketCAN frame\n",time);
	}
	usleep(utime_step);
      }
      printf("%f: Sin: %f\n",time,sinv);
      time +=time_step;
      if(sinv <= 0.0){
	sinv = 6553.6 / 2;
      }
    }
  } else {
    uint32_t rid;
    struct timeval rtv;
    int rlen;
    while(1){
      rlen = cs_socketcan_maxdlc(sk);
      memset(mbuffer,0,sizeof(uint8_t)*maxdlc);
      if(cs_socketcan_recv_nonblock(sk,&rid,mbuffer,&rlen,&rtv) == CS_SOCKETCAN_OK){
	//printf("Receive frame ID: 0x%X DLC: %i\n",rid,rlen);
	if((rid == 5) && (rlen == maxdlc)){
	  if (debug >= 2) print_buffer(time,mbuffer,maxdlc);
	  for(i = 0; i < NUM_SIGNAL; i++){
	    if(cs_signal_runtime_decode(msignal[i],&rsinv[i],NULL,mbuffer,maxdlc)){
	      printf("%f: Failure during signal decoding for %s\n",time,cs_signal_get_name(msignal[i]));
	    }
	  }
	  if(debug >= 1){
	    printf("Frame 5: ");
	    for(i = 0; i < NUM_SIGNAL; i++){
	      printf(", %d: %f",i,rsinv[i]);
	    }
	    printf("\n");
	  } 
	}
      }
      usleep(100000); //0.001s
    }
  }

  for(i = 0; i < NUM_SIGNAL; i++){
    cs_signal_free(msignal[i]);
  }
  
  free(mbuffer);
  free(rsinv);
  if(iface != NULL){
    cs_socketcan_close(sk);
    cs_socketcan_free(sk);
  }
  
  return 0;
}
