/*  
    Copyright (C) 2018 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @example main_signal_reset.c
 * @author Andreas Stewering-Bone
 * @date 11 July 2018
 * @brief Simple Signal reset to test library stack.
 *
 * Simple Signal reset to test library stack.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>

#include "cs_signal.h"
#include "cs_socketcan.h"


#define BYTE_TO_BINARY_PATTERN "%c%c%c%c%c%c%c%c "
#define BYTE_TO_BINARY(byte)  \
  (byte & 0x80 ? '1' : '0'), \
  (byte & 0x40 ? '1' : '0'), \
  (byte & 0x20 ? '1' : '0'), \
  (byte & 0x10 ? '1' : '0'), \
  (byte & 0x08 ? '1' : '0'), \
  (byte & 0x04 ? '1' : '0'), \
  (byte & 0x02 ? '1' : '0'), \
  (byte & 0x01 ? '1' : '0')

#define BYTE_TO_BINARY_REVERTED(byte)  \
  (byte & 0x01 ? '1' : '0'), \
  (byte & 0x02 ? '1' : '0'), \
  (byte & 0x04 ? '1' : '0'), \
  (byte & 0x08 ? '1' : '0'), \
  (byte & 0x10 ? '1' : '0'), \
  (byte & 0x20 ? '1' : '0'), \
  (byte & 0x40 ? '1' : '0'), \
  (byte & 0x80 ? '1' : '0')

int debug = 0;
int binary = 0;

void help(void)
{
  printf(
          "Usage: testsignalreset [OPTION]\n"
          "Test library signal reset stack.\n"
          "\n"
          "Options:\n"
	  "  -b  --binary     Binary output\n"
	  "  -v  --verbose    Enable Runtime debugging, can be set multimple times\n"
          "      --help       display this help and exit\n");
}

void print_buffer(double time, uint8_t *buffer, int len)
{
  int i;
  printf("%f: ",time);
  for(i = 0; i < len; i++){
    if(binary == 1){
      printf(BYTE_TO_BINARY_PATTERN,BYTE_TO_BINARY(buffer[i]));
    } else {
      printf("%02X ",buffer[i]);
    }
  }
  printf("\n");
  
}


int main(int argc, char **argv)
{
  int c;
  int maxdlc = 8; //for CAN 2.0
  uint8_t mbuffer[maxdlc];
  cs_signal_t* bigsignal = NULL;
  cs_signal_t* littlesignal = NULL;
  
  while (1) {
    static struct option long_options[] = {
      /* These options set a flag. */
      {"binary",    no_argument, NULL, 'b'},
      {"help",    no_argument, NULL, 'h'},
      {"verbose",  no_argument, NULL, 'v'},
      {0, 0, 0, 0}
    };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    c = getopt_long (argc, argv, "bhv",
                     long_options, &option_index);

    /* Detect the end of the options. */
    if (c == -1)
      break;


    switch (c) {
    case 0:
      /* If this option set a flag, do nothing else now. */
      if (long_options[option_index].flag != 0)
        break;
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
      break;
    case 'b': binary = 1;     break;
    case 'v': debug +=1;     break;
    case 'h': help(); exit(0);    break;
    case '?':
      /* getopt_long already printed an error message. */
      printf( "Typ `testsignalreset --help' for more information\n");
      return -1;
      break;
    default:
      return -1;
      break;
    }
  }

 
  if(debug > 1){
    cs_signal_debuglevel(CS_DEBUG_LEVEL_DEBUG);
  }
  if(debug > 2){
    cs_signal_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
  }

  bigsignal = cs_signal_init();
  littlesignal = cs_signal_init();
  
  if(
     cs_signal_set(bigsignal,"Testsignalmb","mm",43,10,SIGNAL_BT_BIG,SIGNAL_MUX_SIGNAL,SIGNAL_TYPE_UNSIGNED,1.0,0.0,-100000.0,100000.0)||
     cs_signal_set(littlesignal,"Testsignalml","mm",7,12,SIGNAL_BT_BIG,SIGNAL_MUX_SIGNAL,SIGNAL_TYPE_UNSIGNED,1.0,0.0,-100000.0,100000.0)||
     cs_signal_check(bigsignal,0)||
     cs_signal_check(littlesignal,0)
     ){
    printf("Failure during signal configuration\n");
    return -1;
  }


  memset(mbuffer,0xFF,sizeof(uint8_t)*maxdlc);
  cs_signal_runtime_encode(bigsignal,0.0,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(littlesignal,0.0,mbuffer,maxdlc,0);
  print_buffer(0.0,mbuffer,maxdlc);
  
  
  return 0;
}
