/*  
    Copyright (C) 2018 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @example main_signal_test_bits.c
 * @author Andreas Stewering-Bone
 * @date 07 December 2018
 * @brief Simple Signal en- and decoding to test library stack for bit signals.
 *
 * Simple Signal en- and decoding to test library stack for bit signal.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <time.h>
#include <math.h>

#include "cs_signal.h"
#include "cs_socketcan.h"


void help(void)
{
  printf(
	 "Usage: testsignalbits [OPTION]\n"
	 "Test library signal en- and decoding stack for bit signals.\n"
	 "\n"
	 "Options:\n"
	 "  -b  --binary     Binary output\n"
	 "  -v  --verbose    Enable Runtime debugging, can be set multimple times\n"
	 "      --help       display this help and exit\n");
}


#define BYTE_TO_BINARY_PATTERN "%c%c%c%c%c%c%c%c "
#define BYTE_TO_BINARY(byte)			\
  (byte & 0x80 ? '1' : '0'),			\
    (byte & 0x40 ? '1' : '0'),			\
    (byte & 0x20 ? '1' : '0'),			\
    (byte & 0x10 ? '1' : '0'),			\
    (byte & 0x08 ? '1' : '0'),			\
    (byte & 0x04 ? '1' : '0'),			\
    (byte & 0x02 ? '1' : '0'),			\
    (byte & 0x01 ? '1' : '0')

#define BYTE_TO_BINARY_REVERTED(byte)		\
  (byte & 0x01 ? '1' : '0'),			\
    (byte & 0x02 ? '1' : '0'),			\
    (byte & 0x04 ? '1' : '0'),			\
    (byte & 0x08 ? '1' : '0'),			\
    (byte & 0x10 ? '1' : '0'),			\
    (byte & 0x20 ? '1' : '0'),			\
    (byte & 0x40 ? '1' : '0'),			\
    (byte & 0x80 ? '1' : '0')

int debug = 0;
int binary = 0;
#define NUM_SIGNAL 18

void print_buffer(double time, uint8_t *buffer, int len)
{
  int i;
  printf("%f: ",time);
  for(i = 0; i < len; i++){
    if(binary == 1){
      printf(BYTE_TO_BINARY_PATTERN,BYTE_TO_BINARY(buffer[i]));
    } else {
      printf("%02X ",buffer[i]);
    }
  }
  printf("\n");
  
}

int main(int argc, char **argv)
{
  int c;
  int maxdlc = 8; // for CAN 2.0
  uint8_t mbuffer[maxdlc];
  uint8_t mbuffer2[maxdlc];
  cs_signal_t **msignal = NULL;
  int i;
  
  while (1) {
    static struct option long_options[] = {
      /* These options set a flag. */
      {"binary",    no_argument, NULL, 'b'},
      {"help",    no_argument, NULL, 'h'},
      {"verbose",  no_argument, NULL, 'v'},
      {0, 0, 0, 0}
    };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    c = getopt_long (argc, argv, "bhv",
                     long_options, &option_index);

    /* Detect the end of the options. */
    if (c == -1)
      break;


    switch (c) {
    case 0:
      /* If this option set a flag, do nothing else now. */
      if (long_options[option_index].flag != 0)
        break;
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
      break;
    case 'b': binary = 1;     break;
    case 'v': debug +=1;     break;
    case 'h': help(); exit(0);    break;
    case '?':
      /* getopt_long already printed an error message. */
      printf( "Typ `testsignalbits --help' for more information\n");
      return -1;
      break;
    default:
      return -1;
      break;
    }
  }

  printf("Debuglevel %d\n",debug);
  
  /*Do alloctaion for memory checking*/
  msignal= (cs_signal_t**)calloc(sizeof(cs_signal_t*),NUM_SIGNAL);

  if((msignal == NULL)){
    printf("Malloc failure\n");
    return -1;
  }
  

  if(debug > 1){
    cs_signal_debuglevel(CS_DEBUG_LEVEL_DEBUG);
  }
  if(debug > 2){
    cs_signal_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
  }

  for(i = 0; i < NUM_SIGNAL; i++){
    msignal[i] = cs_signal_init();
  }
  
  if(
     cs_signal_set(msignal[0],"Test1","-",0,1,SIGNAL_BT_LITTLE,SIGNAL_MUX_SIGNAL,SIGNAL_TYPE_UNSIGNED,1.0,0.0,0,1)||
     cs_signal_set(msignal[1],"Test2","-",1,1,SIGNAL_BT_LITTLE,SIGNAL_MUX_SIGNAL,SIGNAL_TYPE_UNSIGNED,1.0,0.0,0,1)||
     cs_signal_set(msignal[2],"Test3","-",2,1,SIGNAL_BT_LITTLE,SIGNAL_MUX_SIGNAL,SIGNAL_TYPE_UNSIGNED,1.0,0.0,0,1)||
     cs_signal_set(msignal[3],"Test4","-",3,1,SIGNAL_BT_LITTLE,SIGNAL_MUX_SIGNAL,SIGNAL_TYPE_UNSIGNED,1.0,0.0,0,1)||
     cs_signal_set(msignal[4],"Test5","-",8,2,SIGNAL_BT_LITTLE,SIGNAL_MUX_SIGNAL,SIGNAL_TYPE_UNSIGNED,1.0,0.0,0,3)||
     cs_signal_set(msignal[5],"Test6","-",10,2,SIGNAL_BT_LITTLE,SIGNAL_MUX_SIGNAL,SIGNAL_TYPE_UNSIGNED,1.0,0.0,0,3)||
     cs_signal_set(msignal[6],"Test7","-",12,2,SIGNAL_BT_LITTLE,SIGNAL_MUX_SIGNAL,SIGNAL_TYPE_UNSIGNED,1.0,0.0,0,3)||
     cs_signal_set(msignal[7],"Test8","-",14,2,SIGNAL_BT_LITTLE,SIGNAL_MUX_SIGNAL,SIGNAL_TYPE_UNSIGNED,1.0,0.0,0,3)||

     cs_signal_set(msignal[8],"Test9","-",0,1,SIGNAL_BT_BIG,SIGNAL_MUX_SIGNAL,SIGNAL_TYPE_UNSIGNED,1.0,0.0,0,1)||
     cs_signal_set(msignal[9],"Test10","-",1,1,SIGNAL_BT_BIG,SIGNAL_MUX_SIGNAL,SIGNAL_TYPE_UNSIGNED,1.0,0.0,0,1)||
     cs_signal_set(msignal[10],"Test11","-",2,1,SIGNAL_BT_BIG,SIGNAL_MUX_SIGNAL,SIGNAL_TYPE_UNSIGNED,1.0,0.0,0,1)||
     cs_signal_set(msignal[11],"Test12","-",3,1,SIGNAL_BT_BIG,SIGNAL_MUX_SIGNAL,SIGNAL_TYPE_UNSIGNED,1.0,0.0,0,1)||
     cs_signal_set(msignal[12],"Test13","-",8,2,SIGNAL_BT_BIG,SIGNAL_MUX_SIGNAL,SIGNAL_TYPE_UNSIGNED,1.0,0.0,0,3)||
     cs_signal_set(msignal[13],"Test14","-",10,2,SIGNAL_BT_BIG,SIGNAL_MUX_SIGNAL,SIGNAL_TYPE_UNSIGNED,1.0,0.0,0,3)||
     cs_signal_set(msignal[14],"Test15","-",12,2,SIGNAL_BT_BIG,SIGNAL_MUX_SIGNAL,SIGNAL_TYPE_UNSIGNED,1.0,0.0,0,3)||
     cs_signal_set(msignal[15],"Test16","-",14,2,SIGNAL_BT_BIG,SIGNAL_MUX_SIGNAL,SIGNAL_TYPE_UNSIGNED,1.0,0.0,0,3)||

     cs_signal_set(msignal[16],"Test17","-",21,9,SIGNAL_BT_LITTLE,SIGNAL_MUX_SIGNAL,SIGNAL_TYPE_UNSIGNED,1.0,0.0,0,511)||
     cs_signal_set(msignal[17],"Test18","-",21,9,SIGNAL_BT_BIG,SIGNAL_MUX_SIGNAL,SIGNAL_TYPE_UNSIGNED,1.0,0.0,0,511)
     ){
    printf("Failure during signal configuration\n");
    return -1;
  }


  for(i = 0; i < NUM_SIGNAL; i++){
    if(cs_signal_check(msignal[i],0)){
      printf("Message check failed for signal %s\n",cs_signal_get_name(msignal[i]));
      return -1;
    }
  }
  
  //Reset buffer
  memset(mbuffer,0,sizeof(uint8_t)*maxdlc);
  memset(mbuffer2,0,sizeof(uint8_t)*maxdlc);

  //Fill buffer
  cs_signal_runtime_encode(msignal[0],1,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[1],1,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[2],1,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[3],1,mbuffer,maxdlc,0);
  
  cs_signal_runtime_encode(msignal[4],3,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[5],3,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[6],3,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[7],3,mbuffer,maxdlc,0);
  
  cs_signal_runtime_encode(msignal[8],1,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[9],1,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[10],1,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[11],1,mbuffer2,maxdlc,0);
  
  cs_signal_runtime_encode(msignal[12],3,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[13],3,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[14],3,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[15],3,mbuffer2,maxdlc,0);

  cs_signal_runtime_encode(msignal[16],511,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[17],511,mbuffer2,maxdlc,0);
  
  print_buffer(0.0, mbuffer,maxdlc);
  print_buffer(0.0, mbuffer2,maxdlc);
  
  //Reset buffer
  memset(mbuffer,0,sizeof(uint8_t)*maxdlc);
  memset(mbuffer2,0,sizeof(uint8_t)*maxdlc);
  
  //Fill buffer
  cs_signal_runtime_encode(msignal[0],1,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[1],0,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[2],1,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[3],0,mbuffer,maxdlc,0);
  
  cs_signal_runtime_encode(msignal[4],3,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[5],0,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[6],3,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[7],0,mbuffer,maxdlc,0);
  
  cs_signal_runtime_encode(msignal[8],1,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[9],0,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[10],1,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[11],0,mbuffer2,maxdlc,0);
  
  cs_signal_runtime_encode(msignal[12],3,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[13],0,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[14],3,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[15],0,mbuffer2,maxdlc,0);

  cs_signal_runtime_encode(msignal[16],511,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[17],511,mbuffer2,maxdlc,0);
  
  print_buffer(0.0, mbuffer,maxdlc);
  print_buffer(0.0, mbuffer2,maxdlc);
    
  //Reset buffer
  memset(mbuffer,0,sizeof(uint8_t)*maxdlc);
  memset(mbuffer2,0,sizeof(uint8_t)*maxdlc);
  
  //Fill buffer
  cs_signal_runtime_encode(msignal[0],0,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[1],1,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[2],0,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[3],1,mbuffer,maxdlc,0);
  
  cs_signal_runtime_encode(msignal[4],0,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[5],3,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[6],0,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[7],3,mbuffer,maxdlc,0);
  
  cs_signal_runtime_encode(msignal[8],0,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[9],1,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[10],0,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[11],1,mbuffer2,maxdlc,0);
  
  cs_signal_runtime_encode(msignal[12],0,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[13],3,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[14],0,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[15],3,mbuffer2,maxdlc,0);

  cs_signal_runtime_encode(msignal[16],511,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[17],511,mbuffer2,maxdlc,0);
  
  print_buffer(0.0, mbuffer,maxdlc);
  print_buffer(0.0, mbuffer2,maxdlc);
  
  //Reset buffer
  memset(mbuffer,0,sizeof(uint8_t)*maxdlc);
  memset(mbuffer2,0,sizeof(uint8_t)*maxdlc);
  
  //Fill buffer

  cs_signal_runtime_encode(msignal[3],1,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[2],1,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[1],1,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[0],1,mbuffer,maxdlc,0);
  
  cs_signal_runtime_encode(msignal[7],3,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[6],3,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[5],3,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[4],3,mbuffer,maxdlc,0);
  
  cs_signal_runtime_encode(msignal[11],1,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[10],1,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[9],1,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[8],1,mbuffer2,maxdlc,0);
  
  cs_signal_runtime_encode(msignal[15],3,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[14],3,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[13],3,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[12],3,mbuffer2,maxdlc,0);

  cs_signal_runtime_encode(msignal[16],511,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[17],511,mbuffer2,maxdlc,0);
  
  print_buffer(0.0, mbuffer,maxdlc);
  print_buffer(0.0, mbuffer2,maxdlc);



  //Reset buffer
  memset(mbuffer,0,sizeof(uint8_t)*maxdlc);
  memset(mbuffer2,0,sizeof(uint8_t)*maxdlc);
  
  //Fill buffer

  cs_signal_runtime_encode(msignal[0],1,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[3],1,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[2],1,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[1],1,mbuffer,maxdlc,0);
    
  cs_signal_runtime_encode(msignal[4],3,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[7],3,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[6],3,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[5],3,mbuffer,maxdlc,0);
  
  cs_signal_runtime_encode(msignal[8],1,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[11],1,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[10],1,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[9],1,mbuffer2,maxdlc,0);
  
  cs_signal_runtime_encode(msignal[12],3,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[15],3,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[14],3,mbuffer2,maxdlc,0);
  cs_signal_runtime_encode(msignal[13],3,mbuffer2,maxdlc,0);

  cs_signal_runtime_encode(msignal[16],511,mbuffer,maxdlc,0);
  cs_signal_runtime_encode(msignal[17],511,mbuffer2,maxdlc,0);

  print_buffer(0.0, mbuffer,maxdlc);
  print_buffer(0.0, mbuffer2,maxdlc);

  for(i = 0; i < NUM_SIGNAL; i++){
    cs_signal_free(msignal[i]);
  }
  
  return 0;
}
