/*  
    Copyright (C) 2017 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @example main_socketcand_receive.c
 * @author Andreas Stewering-Bone
 * @date 20 March 2017
 * @brief Simple SocketCAN Receive application to test library stack.
 *
 * Simple SocketCAN Receive application to test library stack.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <time.h>

#include "cs_socketcan.h"


void help(void)
{
  printf(
          "Usage: testsocketcandreceive [OPTION]\n"
          "Test library SocketCANd stack.\n"
          "\n"
          "Options:\n"
          "  -s, --server     Server hostname\n"
          "  -p, --port       Server port\n"
	  "  -i, --interface  Server CAN interface name\n"
	  "  -v  --verbose    Enable Runtime debugging\n"
          "      --help       display this help and exit\n");
}

int main(int argc, char **argv)
{
  int c;
  char * interface = NULL;
  char * server = NULL;
  int port = 0;
  cs_socketcan_t *data;
  uint32_t id;
  struct timeval tv;
  int len;
  int debug = 0;
  int i;
  int rv;
  
  while (1) {
    static struct option long_options[] = {
      /* These options set a flag. */
      {"server", required_argument, NULL, 's'},
      {"port", required_argument, NULL, 'p'},
      {"interface", required_argument, NULL, 'i'},
      {"help",    no_argument, NULL, 'h'},
      {"verbose",  no_argument, NULL, 'v'},
      {0, 0, 0, 0}
    };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    c = getopt_long (argc, argv, "s:p:i:hv",
                     long_options, &option_index);

    /* Detect the end of the options. */
    if (c == -1)
      break;

    switch (c) {
    case 0:
      /* If this option set a flag, do nothing else now. */
      if (long_options[option_index].flag != 0)
        break;
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
      break;
    case 's': server = optarg; break;
    case 'p': port = strtoul(optarg,(char**)NULL,10); break;
    case 'i': interface = optarg;     break;
    case 'v': debug = 1;     break;
    case 'h': help(); exit(0);    break;
    case '?':
      /* getopt_long already printed an error message. */
      printf( "Typ `testsocketcandreceive --help' for more information\n");
      return -1;
      break;
    default:
      return -1;
      break;
    }
  }

  if((interface == NULL)||
     (interface == NULL)
     ){
    printf("Please check arguments\n");
    help();
    return -1;
  }

  if(debug){
    cs_socketcan_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
  }
  
  //Init with default values
  if((data = cs_socketcan_init())==NULL){
    printf("Init failed\n");
    return -1;
  }

  if((server == NULL)|| (port == 0)){
    printf("Configure native socket\n");
    if(cs_socketcan_config_native(data,interface,0)){
      printf("Config native failed\n");
      return -1;
    }
  } else {
    printf("Configure socketcand socket\n");
    if(cs_socketcan_config_socketcand(data,interface,server,port)){
      printf("Config socketcand failed\n");
      return -1;
    }
  }

  
  if(cs_socketcan_open(data)){
    printf("Open failed\n");
    return -1;
  }
  
  uint8_t buffer[cs_socketcan_maxdlc(data)];

    //Init send data
  for(i = 0; i < cs_socketcan_maxdlc(data); i++){
    buffer[i] = (uint8_t)i;
  }
  

  while(1){  
    rv = cs_socketcan_recv_block(data,&id,buffer,&len,&tv,5000);
    if( rv == CS_SOCKETCAN_TIMEOUT){
      printf("Receive timout\n");
      continue;
    }
    if ( rv == CS_SOCKETCAN_ERROR){
      printf("Receive error frame\n");
      continue;
    }
    if ( rv == CS_SOCKETCAN_RTR){
      printf("Receive RTR frame\n");
      continue;
    }
    if(rv){
      printf("Failed to receive Frame via SoketCANd: %s\n",cs_socketcan_getlasterrorstring(data));
    } else {
      printf("Receive frame Id: 0x%X Ts: %ld.%lds Len: %i, Data: ",id,(long int)tv.tv_sec,(long int)tv.tv_usec,len);
      for(i=0;i < len;i++){
	printf("0x%X ",buffer[i]);
      }
      printf("\n");
    }
    
  }

  if(cs_socketcan_close(data)){
    printf("Failed to close interface\n");
    return -1;
  }
  cs_socketcan_free(data);
  
  return 0;
}
