/*  
    Copyright (C) 2017 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @example main_socketcand_send.c
 * @author Andreas Stewering-Bone
 * @date 20 March 2017
 * @brief Simple SocketCAN Send application to test library stack.
 *
 * Simple SocketCAN Send application to test library stack.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <time.h>

#include "cs_socketcan.h"


void help(void)
{
  printf(
          "Usage: testsocketcandsend [OPTION]\n"
          "Test library SocketCANd stack.\n"
          "\n"
          "Options:\n"
          "  -s, --server     Server hostname\n"
          "  -p, --port       Server port\n"
	  "  -i, --interface  Server CAN interface name\n"
	  "  -v  --verbose    Enable Runtime debugging\n"
          "      --help       display this help and exit\n");
}

int main(int argc, char **argv)
{
  int c;
  char * interface = NULL;
  char * server = NULL;
  int port = 0;
  cs_socketcan_t *data;
  int debug = 0;
  int i;
  struct timespec lsleep;
  lsleep.tv_sec = 0;
  lsleep.tv_nsec = 1e6;
  int rv;

  while (1) {
    static struct option long_options[] = {
      /* These options set a flag. */
      {"server", required_argument, NULL, 's'},
      {"port", required_argument, NULL, 'p'},
      {"interface", required_argument, NULL, 'i'},
      {"help",    no_argument, NULL, 'h'},
      {"verbose",  no_argument, NULL, 'v'},
      {0, 0, 0, 0}
    };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    c = getopt_long (argc, argv, "s:p:i:hv",
                     long_options, &option_index);

    /* Detect the end of the options. */
    if (c == -1)
      break;

    switch (c) {
    case 0:
      /* If this option set a flag, do nothing else now. */
      if (long_options[option_index].flag != 0)
        break;
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
      break;
    case 's': server = optarg; break;
    case 'p': port = strtoul(optarg,(char**)NULL,10); break;
    case 'i': interface = optarg;     break;
    case 'v': debug = 1;     break;
    case 'h': help(); exit(0);    break;
    case '?':
      /* getopt_long already printed an error message. */
      printf( "Typ `testsocketcandsend --help' for more information\n");
      return -1;
      break;
    default:
      return -1;
      break;
    }
  }

  if((interface == NULL)||
     (interface == NULL)
     ){
    printf("Please check arguments\n");
    help();
    return -1;
  }

  if(debug){
    cs_socketcan_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
  }
  
  //Init with default values
  if((data = cs_socketcan_init())==NULL){
    printf("Init failed\n");
    return -1;
  }

  if((server == NULL)|| (port == 0)){
    printf("Configure native socket\n");
    if(cs_socketcan_config_native(data,interface,0)){
      printf("Config native failed\n");
      return -1;
    }
  } else {
    printf("Configure socketcand socket\n");
    if(cs_socketcan_config_socketcand(data,interface,server,port)){
      printf("Config socketcand failed\n");
      return -1;
    }
  }

  if(cs_socketcan_open(data)){
    printf("Open failed\n");
    return -1;
  }

  int dlc = cs_socketcan_maxdlc(data);
  uint8_t buffer[dlc];
  //Init send data
  for(i = 0; i < dlc; i++){
    buffer[i] = (uint8_t)i;
  }

  
  while(1){
    rv = cs_socketcan_send(data,0x231,buffer,dlc,SOCKETCAN_SEND_CAN);
    if(rv == CS_SOCKETCAN_TXFULL){
      printf("Failed to send Frame because TX-Full\n");
      return -1;
    }
    if(rv){
      printf("Failed to send Frame via SoketCANd: %s\n",cs_socketcan_getlasterrorstring(data));
      return -1;
    }
    nanosleep(&lsleep,NULL);
  }

  if(cs_socketcan_close(data)){
    printf("Failed to close interface\n");
    return -1;
  }
  cs_socketcan_free(data);
  
  return 0;
}
