/*  
    Copyright (C) 2013 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @example main_udsclient.c
 * @author Andreas Stewering-Bone
 * @date 9 Oct 2014
 * @brief Simple UDS client application for testing the UDS API.
 *
 * Simple UDS client application for testing the UDS API.
*/

#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include <stdlib.h>

#include "cs_uds.h"
#include "cs_isotp.h"
#include "cs_socketcan.h"

#include "seed.h"

#define TRANSFERLEN 100
#define TRANSFERBIGLEN 1234


void help(void)
{
  printf(
          "Usage: testudsclient [OPTION]\n"
          "Test udsclient stack.\n"
          "\n"
          "Options:\n"
          "  -i, --iface    CAN-Interface\n"
	  "  -s, --sa       hex source adress\n"
	  "  -t, --ta       hex target adress\n"
	  "  -v, --verbose  Debugging output\n"
          "      --help     display this help and exit\n");
}

int main(int argc, char **argv)
{
  int c,i;

  char *iface = NULL;

  cs_uds_service_t *data;
  cs_uds_diagsession_t diag_data;
  uint8_t didbuffer[CS_UDS_MAXBYTE];
  uint8_t didbufferw[CS_UDS_MAXBYTE];
  uint8_t routinebuffer[CS_UDS_MAXBYTE];
  cs_uds_dtc_reqdata_t reqdata;
  cs_uds_dtc_respdata_t respdata;
  uint8_t seedbuffer[SEEDBUFFERLEN];
  int seedlen = SEEDBUFFERLEN;
  uint8_t keybuffer[SEEDBUFFERLEN];
  int keylen = SEEDBUFFERLEN;
  int readb;
  int j;
  unsigned long int sa = 0;
  unsigned long int ta = 0;
  uint64_t maxblocklen = 0;
  uint8_t transferdata[TRANSFERLEN];
  uint64_t transferlen = TRANSFERLEN;
  uint8_t transferbigdata[TRANSFERBIGLEN];
  uint8_t transferexitdata[2];
  int transferexitlen = 2;
  cs_uds_client_flash_t flash;
  int rv;
  int verbose = 0;

  for(i = 0; i < TRANSFERLEN; i++){
    transferdata[i]=i;
  }

  for(i = 0; i < TRANSFERBIGLEN; i++){
    transferbigdata[i]=i;
  }

  while (1) {
    static struct option long_options[] = {
      /* These options set a flag. */
      {"iface", required_argument, NULL, 'i'},
      {"sa", required_argument, NULL, 's'},
      {"verbose", required_argument, NULL, 'v'},
      {"ta", required_argument, NULL, 't'},
      {"help",    no_argument, NULL, 'h'},
      {0, 0, 0, 0}
    };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    c = getopt_long (argc, argv, "i:ht:s:v",
                     long_options, &option_index);

    /* Detect the end of the options. */
    if (c == -1)
      break;

    switch (c) {
    case 0:
      /* If this option set a flag, do nothing else now. */
      if (long_options[option_index].flag != 0)
        break;
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
      break;
    case 'i': iface = optarg;     break;
    case 'v': verbose += 1; if(verbose > 2){ verbose = 2;}    break;
    case 's': sa = strtoul(optarg,(char**)NULL,16);     break;
    case 't': ta = strtoul(optarg,(char**)NULL,16);     break;
    case 'h': help(); exit(0);    break;
    case '?':
      /* getopt_long already printed an error message. */
      printf( "Typ `testudsclient --help' for more information\n");
      return -1;
      break;
    default:
      return -1;
      break;
    }
  }

  /*Check inputs*/

  if(iface==NULL){
    printf( "Please check input arguments\n");
    printf( "Typ `testudsclient --help' for more information\n");
    return -1;
  }

  /*First check Library version*/
  if(CS_UDS_CHECKVERSION){
    printf("Incompatible Library interface, Application was compiled with version %i, Library has version %i\n",
	   CS_UDS_VERSION_MAGIC,cs_uds_getversion());
    return -1;
  }

  //Set verbose level
  printf("Verbose level %i\n",verbose);
  switch (verbose){
  case 0:
    cs_uds_debuglevel(CS_DEBUG_LEVEL_NONE);
    cs_isotp_debuglevel(CS_DEBUG_LEVEL_NONE);
    cs_socketcan_debuglevel(CS_DEBUG_LEVEL_NONE);
    break;
  case 1:
    cs_uds_debuglevel(CS_DEBUG_LEVEL_DEBUG);
    cs_isotp_debuglevel(CS_DEBUG_LEVEL_DEBUG);
    cs_socketcan_debuglevel(CS_DEBUG_LEVEL_DEBUG);
    break;
  case 2:
    cs_uds_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
    cs_isotp_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
    cs_socketcan_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
    break;
  }
  

  /*Start UDS-Stack*/
  if((data = cs_uds_init()) == NULL){
    printf("Open interface failed\n");
    return -1;
  }
  
  if(cs_uds_can_init(data,iface, sa, ta)){
    printf("Init CAN layer failed\n");
    return -1;
  }

  if(cs_uds_can_open(data)){
    printf("Open CAN layer failed\n");
    return -1;
  }

  printf("Request diagnostic session\n");
  if(cs_uds_client_diagsession(data,CS_UDS_SI_DSC_SUP, RESPREQ, &diag_data)){
    printf("Diagnostic request failed: %s\n",cs_uds_strlerror(data));
    return -1;
  }
  /*try to set timing*/
  if(cs_uds_channel_settiming(data, diag_data.tserver[0], 11)){
    printf("Settiming failed: %s\n",cs_uds_strlerror(data));
    return -1;
  }

  for(i = 0; i < 10; i++){
    //Loop to check wait time
    if(cs_uds_client_testerpresent(data, RESPSUP)){
      printf("Test present failed: %s\n",cs_uds_strlerror(data));
      return -1;
    }
  }
  
  
  /*Security Access, first get seed*/
  if(cs_uds_client_securityaccess(data,CS_UDS_SI_SA_RSD, NULL,0,seedbuffer,&seedlen)){
    printf("SecurityAccess: get seed failed, %s\n",cs_uds_strlerror(data));
    return -1;
  }    
  
  /*Calculate seed algo*/
  seed_algo(CS_UDS_SI_SA_RSD,seedbuffer,seedlen,keybuffer,&keylen);
  
  
  /*Security Access, next send key*/
  if(cs_uds_client_securityaccess(data,CS_UDS_SI_SA_SK, keybuffer,keylen,NULL,NULL)){
    printf("SecurityAccess: get seed failed, %s\n",cs_uds_strlerror(data));
    return -1;
  }    
  
  //Flashing in manualy way
  printf("Transfer data, download, via low level functions\n");
  if(cs_uds_client_requestDownload(data,0,0,0x2345,transferlen,&maxblocklen)){
    printf("Request download failed: %s\n",cs_uds_strlerror(data));
    return -1;
  }
  printf("Request download: Transfer block size 0x%llX\n", (unsigned long long)maxblocklen);
  
  if(cs_uds_client_transferData(data,CS_UDS_SI_RD,1,maxblocklen,transferdata,&transferlen)){
    printf("Transfer data upload failed: %s\n",cs_uds_strlerror(data));
    return -1;
  }
  printf("Upload %lli Bytes\n",(unsigned long long)transferlen);
  
  if(cs_uds_client_transferExit(data,NULL,0,transferexitdata,&transferexitlen)){
    printf("Transfer Exit failed: %s\n",cs_uds_strlerror(data));
    return -1;
  }
  printf("TransferExit: Recv data %X %X\n",transferexitdata[0],transferexitdata[1]);
  
  printf("Transfer data, upload, via low level functions\n");
  if(cs_uds_client_requestUpload(data,0,0,0x2346,25,&maxblocklen)){
    printf("Request upload failed: %s\n",cs_uds_strlerror(data));
    return -1;
  }
  printf("Request upload: Transfer block size 0x%llX\n",(unsigned long long)maxblocklen);
  
  if(cs_uds_client_transferData(data,CS_UDS_SI_RU,1,maxblocklen,transferdata,&transferlen)){
    printf("Transfer data upload failed: %s\n",cs_uds_strlerror(data));
    return -1;
  }
  printf("Upload %lli Bytes\n",(long long int)transferlen);
  
  if(cs_uds_client_transferExit(data,NULL,0,transferexitdata,&transferexitlen)){
    printf("Transfer Exit failed: %s\n",cs_uds_strlerror(data));
    return -1;
  }
  printf("TransferExit: Recv data %X %X\n",transferexitdata[0],transferexitdata[1]);
  
  printf("Transfer data via API in blocking style\n");
  transferexitlen = 2;
  if(cs_uds_client_api_transfer(data,CS_UDS_SI_RD,0,0,0x12345,transferbigdata,TRANSFERBIGLEN,
				NULL,0,transferexitdata,&transferexitlen)){
    printf("API Transfer failed: %s\n",cs_uds_strlerror(data));
    return -1;
  }
  printf("TransferExit: Recv data %X %X\n",transferexitdata[0],transferexitdata[1]);
  
  transferexitlen = 2;
  if(cs_uds_client_api_transfer(data,CS_UDS_SI_RU,0,0,0x12345,transferbigdata,TRANSFERBIGLEN,
				NULL,0,transferexitdata,&transferexitlen)){
    printf("API Transfer failed: %s\n",cs_uds_strlerror(data));
    return -1;
  }
  printf("TransferExit: Recv data %X %X\n",transferexitdata[0],transferexitdata[1]);
  
  
  printf("Transfer data in non blocking style\n");
  if(cs_uds_client_api_flash_init(&flash,data,CS_UDS_SI_RD,0,0,0x2345,transferbigdata,TRANSFERBIGLEN)){
    printf("Failed to init flash process\n");
    return -1;
  }
  //Transfer
  while((rv = cs_uds_client_api_flash_run(&flash)) == 0){
    printf("Transfer block, actually %lli Bytes of %lli Bytes\n",(long long)flash.transfered,(long long)flash.transferlen);
    cs_uds_client_testerpresent(data, RESPSUP);
  }
  if(rv == 1){
    printf("Transfer block failure\n");
    return -1;
  }
  printf("Transfer block, finished %lli Bytes of %lli Bytes\n",(long long)flash.transfered,(long long)flash.transferlen);
  transferexitlen = 2;
  if(cs_uds_client_transferExit(data,NULL,0,transferexitdata,&transferexitlen)){
    printf("Transfer Exit failed: %s\n",cs_uds_strlerror(data));
    return -1;
  }
  printf("TransferExit: Recv data %X %X\n",transferexitdata[0],transferexitdata[1]);
  
  /*ECU reset*/
  if(cs_uds_client_ecureset(data, CS_UDS_SI_ER_HR, RESPREQ, NULL)){
    printf("ECU Reset failed: %s\n",cs_uds_strlerror(data));
    return -1;
  }
  
  /*DTC settings*/
  if(cs_uds_client_controldtcsettings(data, CS_UDS_SI_CDTCS_ON, RESPREQ)){
    printf("Control DTC settings ON failed: %s\n",cs_uds_strlerror(data));
    return -1;
  }
  if(cs_uds_client_controldtcsettings(data, CS_UDS_SI_CDTCS_OFF, RESPREQ)){
    printf("Control DTC settings OFF failed: %s\n",cs_uds_strlerror(data));
    return -1;
  }
  if(cs_uds_client_testerpresent(data, RESPSUP)){
    printf("Testet present failed: %s\n",cs_uds_strlerror(data));
    return -1;
  }
  
  if(cs_uds_client_linkcontrol(data, CS_UDS_SI_LC_VMTWFP, RESPREQ, CS_UDS_SI_LC_PC9600)){
    printf("Link control VMTWFP failed: %s\n",cs_uds_strlerror(data));
    return -1;
    }
  if(cs_uds_client_linkcontrol(data, CS_UDS_SI_LC_VMTWSP, RESPREQ, 115000)){
    printf("Link control VMTWSP failed: %s\n",cs_uds_strlerror(data));
    return -1;
  }
  if(cs_uds_client_linkcontrol(data, CS_UDS_SI_LC_TM, RESPREQ, 0)){
    printf("Link control TM failed: %s\n",cs_uds_strlerror(data));
    return -1;
  }
  if(cs_uds_client_testerpresent(data, RESPSUP)){
    printf("Test present failed: %s\n",cs_uds_strlerror(data));
    return -1;
  }
  //Try to read DID
  memset(didbuffer,0,CS_UDS_MAXBYTE);
  readb = -1; //Unspecific length
  if(cs_uds_client_readDatabyIdentifier(data, 0x1001, &readb, didbuffer)){
    printf("Read DID failed: %s\n",cs_uds_strlerror(data));
    return -1;
  } else {
    printf("Read DID %x: ",0x1001);
    for(j=0; j<readb; j++)
      printf("%x",didbuffer[j]);
    printf("\n");
  }
  for(i = 0; i < 100; i++){
    //Try to write DID
    didbufferw[0]= 0xff;
    didbufferw[1]= 0x01;
    didbufferw[2]= 0x11;
    didbufferw[3]= 0x44;
    readb = 4;
    if(cs_uds_client_writeDatabyIdentifier(data, 0x1002 + i, readb, didbufferw)){
      printf("Write DID failed: %s\n",cs_uds_strlerror(data));
      return -1;
    }
  }
  //reread DID
  memset(didbuffer,0,CS_UDS_MAXBYTE);
  readb = 4; //Set length, failure if received length does not equal
  if(cs_uds_client_readDatabyIdentifier(data, 0x1002, &readb, didbuffer)){
    printf("Read DID failed: %s\n",cs_uds_strlerror(data));
  } else {
    int j;
    printf("Read DID %x: ",0x1002);
    for(j=0; j<readb; j++)
      printf("%x",didbuffer[j]);
    printf("\n");
  } 
  cs_uds_client_testerpresent(data, RESPSUP);
  if(cs_uds_client_routinecontrol(data, CS_UDS_SI_RC_STR, RESPREQ, 0x2001,
				  NULL,0,NULL,NULL, 0))
    printf("RoutineControl: 0x2001  Start failed: %s\n",cs_uds_strlerror(data));
  
    cs_uds_client_testerpresent(data, RESPSUP);
    routinebuffer[0]= 0x01;
    routinebuffer[1]= 0x02;
    routinebuffer[2]= 0x03;
    routinebuffer[3]= 0x04;
    int len = 10;
    if(cs_uds_client_routinecontrol(data, CS_UDS_SI_RC_STR, RESPREQ, 0x2002,
				    routinebuffer,4,NULL,routinebuffer, &len))
      printf("RoutineControl: 0x2002  Start failed: %s\n",cs_uds_strlerror(data));
    cs_uds_client_ClearDiagnosticInformation(data, RESPREQ, 0x00FFFF33);
    
    cs_uds_client_testerpresent(data, RESPSUP);
    
    
    //reportNumberOfDTCByStatusMask
    memset(&reqdata,0,sizeof(reqdata));
    reqdata.status_mask = 0xFF;
    if(cs_uds_client_ReadDTCInformation(data, RESPREQ,CS_UDS_SI_RDTCI_RNODTCBSM, 
					&reqdata, &respdata)){
      printf("ReadDTCInformation: reportNumberOfDTCByStatusMask: Failure: %s\n",cs_uds_strlerror(data));
      return -1;
    } else {
      printf("ReadDTCInformation: reportNumberOfDTCByStatusMask: Avail-Mask %x, Format: %x, Number of DTC: %i\n",
	     respdata.avail_mask,respdata.format,respdata.count);
    }
    
    
    //reportDTCByStatusMask
    memset(&reqdata,0,sizeof(reqdata));
    reqdata.status_mask = 0xFF;
    if(cs_uds_client_ReadDTCInformation(data, RESPREQ,CS_UDS_SI_RDTCI_RDTCBSM, 
					&reqdata, &respdata)){
      printf("ReadDTCInformation: reportDTCByStatusMask: Failure: %s\n",cs_uds_strlerror(data));
      return -1;
    } else {
      printf("ReadDTCInformation: reportDTCByStatusMask: Avail-Mask %x\n",respdata.avail_mask);
      for(i = 0; i < respdata.lt_statusdtcrec; i++){
	printf("  Entry %i: DTC %x, Statusmask %x\n",i,respdata.statusdtcrec[i].dtc,respdata.statusdtcrec[i].status);
      }
    }
    
    //reportDTCSnapshotIdentification
    memset(&reqdata,0,sizeof(reqdata));
    if(cs_uds_client_ReadDTCInformation(data, RESPREQ,CS_UDS_SI_RDTCI_RDTCSSI, 
					&reqdata, &respdata)){
      printf("ReadDTCInformation: reportDTCSnapshotIdentification: Failure: %s\n",cs_uds_strlerror(data));
      return -1;
    } else {
      printf("ReadDTCInformation: reportDTCSnapshotIdentification: \n");
      for(i = 0; i < respdata.lt_statusdtcrec; i++){
	printf("  Entry %i: DTC %x, Snapshot record number %i\n",i,respdata.statusdtcrec[i].dtc,respdata.statusdtcrec[i].snaprecno);
      }
    }    
    //reportDTCBySeverityMaskRecord
    memset(&reqdata,0,sizeof(reqdata));
    reqdata.severeity_mask = 0xFF; 
    reqdata.status_mask = 0xFF; 
    if(cs_uds_client_ReadDTCInformation(data, RESPREQ,CS_UDS_SI_RDTCI_RDTCBSMR, 
					&reqdata, &respdata)){
      printf("ReadDTCInformation: reportDTCBySeverityMaskRecord: Failure: %s\n",cs_uds_strlerror(data));
      return -1;
    } else {
      printf("ReadDTCInformation: reportDTCBySeverityMaskRecord: \n");
      for(i = 0; i < respdata.lt_statusdtcrec; i++){
	printf("  Entry %i: DTC %x, Statusmask 0x%X Severity mask 0x%X Functional unit 0x%X\n",i,
	       respdata.statusdtcrec[i].dtc,
	       respdata.statusdtcrec[i].status,
	       respdata.statusdtcrec[i].severity,
	       respdata.statusdtcrec[i].functionalunit);
      }
    }
    
    //reportSeverityInformationOfDTC
    memset(&reqdata,0,sizeof(reqdata));
    reqdata.dtc_mask = 0xFF123; 
    if(cs_uds_client_ReadDTCInformation(data, RESPREQ,CS_UDS_SI_RDTCI_RSIODTC, 
					&reqdata, &respdata)){
      printf("ReadDTCInformation: reportSeverityInformationOfDTC: Failure: %s\n",cs_uds_strlerror(data));
      return -1;
    } else {
      printf("ReadDTCInformation: reportSeverityInformationOfDTC: \n");
      for(i = 0; i < respdata.lt_statusdtcrec; i++){
	printf("  Entry %i: DTC %x, Statusmask 0x%X Severity mask 0x%X Functional unit 0x%X\n",i,
	       respdata.statusdtcrec[i].dtc,
	       respdata.statusdtcrec[i].status,
	       respdata.statusdtcrec[i].severity,
	       respdata.statusdtcrec[i].functionalunit);
      }
    }
    
    //reportSupportedDTC
    memset(&reqdata,0,sizeof(reqdata));
    
    if(cs_uds_client_ReadDTCInformation(data, RESPREQ,CS_UDS_SI_RDTCI_RSUPDTC, 
					&reqdata, &respdata)){
      printf("ReadDTCInformation: reportSupportedDTCs: Failure: %s\n",cs_uds_strlerror(data));
      return -1;
    } else {
      printf("ReadDTCInformation: reportSupportedDTCs: Avail-Mask %x\n",respdata.avail_mask);
      for(i = 0; i < respdata.lt_statusdtcrec; i++){
	printf("  Entry %i: DTC %x, Statusmask %x\n",i,respdata.statusdtcrec[i].dtc,respdata.statusdtcrec[i].status);
      }
    }

  cs_uds_can_close(data);
    
  cs_uds_free(data);
  return 0;
}
