/*  
    Copyright (C) 2019 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @example main_udsclient_endlesstest.c
 * @author Andreas Stewering-Bone
 * @date 09 Jan 2019
 * @brief UDS client application for testing the UDS and IsoTP API in an endless test.
 *
 * UDS client application for testing the UDS and IsoTP API in an endless test.
*/

#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include <stdlib.h>
#include <inttypes.h>

#include "cs_uds.h"
#include "cs_isotp.h"
#include "cs_process.h"
#include "cs_socketcan.h"


void help(void)
{
  printf(
          "Usage: testudsclientendless [OPTION]\n"
          "Test udsclient stack.\n"
          "\n"
          "Options:\n"
          "  -i, --iface    CAN-Interface\n"
	  "  -s, --sa       hex source adress\n"
	  "  -t, --ta       hex target adress\n"
	  "  -v, --verbose  Debugging output\n"
          "      --help     display this help and exit\n");
}

int myrand(int range){
  return rand()%range;
}

int main(int argc, char **argv)
{
  int c,i,j;

  char *iface = NULL;

  cs_uds_service_t *data;
  cs_uds_diagsession_t diag_data;
  uint8_t didbuffer[CS_UDS_MAXBYTE];
  int readb = 0;
  cs_uds_dtc_reqdata_t reqdata;
  cs_uds_dtc_respdata_t respdata;
  unsigned long int sa = 0;
  unsigned long int ta = 0;
  int verbose = 0;
  uint64_t loop = 0;

  int rv;
  
  while (1) {
    static struct option long_options[] = {
      /* These options set a flag. */
      {"iface", required_argument, NULL, 'i'},
      {"sa", required_argument, NULL, 's'},
      {"verbose", required_argument, NULL, 'v'},
      {"ta", required_argument, NULL, 't'},
      {"help",    no_argument, NULL, 'h'},
      {0, 0, 0, 0}
    };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    c = getopt_long (argc, argv, "i:ht:s:v",
                     long_options, &option_index);

    /* Detect the end of the options. */
    if (c == -1)
      break;

    switch (c) {
    case 0:
      /* If this option set a flag, do nothing else now. */
      if (long_options[option_index].flag != 0)
        break;
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
      break;
    case 'i': iface = optarg;     break;
    case 'v': verbose += 1; if(verbose > 2){ verbose = 2;}    break;
    case 's': sa = strtoul(optarg,(char**)NULL,16);     break;
    case 't': ta = strtoul(optarg,(char**)NULL,16);     break;
    case 'h': help(); exit(0);    break;
    case '?':
      /* getopt_long already printed an error message. */
      printf( "Typ `testudsclientendless --help' for more information\n");
      return -1;
      break;
    default:
      return -1;
      break;
    }
  }

  /*Check inputs*/

  if(iface==NULL){
    printf( "Please check input arguments\n");
    printf( "Typ `testudsclientendless --help' for more information\n");
    return -1;
  }

  /*First check Library version*/
  if(CS_UDS_CHECKVERSION){
    printf("Incompatible Library interface, Application was compiled with version %i, Library has version %i\n",
	   CS_UDS_VERSION_MAGIC,cs_uds_getversion());
    return -1;
  }

  //Set verbose level
  printf("Verbose level %i\n",verbose);
  switch (verbose){
  case 0:
    cs_uds_debuglevel(CS_DEBUG_LEVEL_NONE);
    cs_isotp_debuglevel(CS_DEBUG_LEVEL_NONE);
    cs_socketcan_debuglevel(CS_DEBUG_LEVEL_NONE);
    break;
  case 1:
    cs_uds_debuglevel(CS_DEBUG_LEVEL_DEBUG);
    cs_isotp_debuglevel(CS_DEBUG_LEVEL_DEBUG);
    cs_socketcan_debuglevel(CS_DEBUG_LEVEL_DEBUG);
    break;
  case 2:
    cs_uds_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
    cs_isotp_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
    cs_socketcan_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
    break;
  }
  

  /*Start UDS-Stack*/
  if((data = cs_uds_init()) == NULL){
    printf("Open interface failed\n");
    return -1;
  }
  
  if(cs_uds_can_init(data,iface, sa, ta)){
    printf("Init CAN layer failed\n");
    return -1;
  }

  if(cs_uds_can_open(data)){
    printf("Open CAN layer failed\n");
    return -1;
  }

  int num_did = 4;
  uint16_t didlist[num_did];
  didlist[0] = 0xD100;
  didlist[1] = 0xD113;
  didlist[2] = 0xF163;
  didlist[3] = 0xF162;
 
  
  typedef enum {
    WAITFOR = 0,     //Wait for testitem, via tester present
    DIAGSESSION_REQUEST,     //Request diagnostig session
    DIAGSESSION_LEAVE,       //Leave diagnostic session
    DID_READ,                //Read several DID
    DTC_READ,                //Read DTCs
    ECU_RESET,               //Request ECU Reset
    WAIT_INACTIVE,           //Wait without interaction
  }state_t;


  state_t state = WAITFOR;
  uint16_t error;
  uint32_t wtime = 0;

  while(1){
    switch(state){
    case WAITFOR:
      loop++;
      printf("State %i: Loop %llu\n",state,loop);
      //Call tester present until we got a feedback
      while(1){
	printf("State %i: Tester present\n",state);
	rv = cs_uds_client_testerpresent(data, RESPREQ);
	if(rv == 0){
	  //Got positive response
	  printf("State %i: Testitem is alive\n",state);
	  break;
	} else {
	  error = cs_uds_getlerror(data);
	  if(error !=  CS_UDS_ERROR_CLIENT_STIMEOUT){
	    printf("State %i: Got failure %s\n",state,cs_uds_strlerror(data));
	    return -1;
	  }
	  cs_process_sleep(5000); //5s
	}
      }
      state = DIAGSESSION_REQUEST;
      break;
    case DIAGSESSION_REQUEST:
      //Extended session
      printf("State %i: Request extended diagnostic session\n",state);
      if(cs_uds_client_diagsession(data,CS_UDS_SI_DSC_EXTDS, RESPREQ, &diag_data)){
	printf("State %i: Diagnostic request failed: %s\n",state,cs_uds_strlerror(data));
	state = WAIT_INACTIVE;
	break;
      }
      printf("Got timing:\n");
      printf("P2CANServer: %i ms\n",diag_data.tserver[0].P2CANServer);
      printf("P2CANServerE: %i ms\n",diag_data.tserver[0].P2CANServerE);
      //try to set timing
      if(cs_uds_channel_settiming(data, diag_data.tserver[0], 0)){
	printf("State %i: Settiming failed: %s\n",state,cs_uds_strlerror(data));
      }
      state = DID_READ; 
      break;
    case DID_READ:
      printf("State %i: DID read\n",state);
      //Try to read DIDs
      for(i = 0; i < num_did; i++){
	printf("State %i: Read DID 0x%X\n",state,didlist[i]);
	memset(didbuffer,0,CS_UDS_MAXBYTE);
	readb = -1; //Unspecific length
	if(cs_uds_client_readDatabyIdentifier(data, didlist[i], &readb, didbuffer)){
	  printf("State %i: Read DID failed: %s\n",state,cs_uds_strlerror(data));
	} else {
	  printf("State %i: Read DID 0x%X: ",state, didlist[i]);
	  for(j=0; j<readb; j++){
	    printf("%x",didbuffer[j]);
	  }
	  printf("\n");
	}
      }
      state=DTC_READ;
      break;
    case DTC_READ:
      //reportDTCByStatusMask
      memset(&reqdata,0,sizeof(reqdata));
      reqdata.status_mask = 0xFF;
      if(cs_uds_client_ReadDTCInformation(data, RESPREQ,CS_UDS_SI_RDTCI_RDTCBSM, 
					  &reqdata, &respdata)){
	printf("State %i: reportDTCByStatusMask: Failure: %s\n",state,cs_uds_strlerror(data));
      } /*else {
	printf("State %i: reportDTCByStatusMask: Avail-Mask %x\n",state,respdata.avail_mask);
	for(i = 0; i < respdata.lt_statusdtcrec; i++){
	  printf("  Entry %i: DTC %x, Statusmask %x\n",i,respdata.statusdtcrec[i].dtc,respdata.statusdtcrec[i].status);
	}
	}*/
      state=DIAGSESSION_LEAVE;
      break;
    case ECU_RESET:
      printf("State %i: ECU hard reset\n",state);
      if(cs_uds_client_ecureset(data, CS_UDS_SI_ER_HR, RESPREQ, NULL)){
	printf("State %i: ECU Reset failed: %s\n",state,cs_uds_strlerror(data));
	state = WAIT_INACTIVE;
      }
      wtime = 10000; //10s
      state = WAIT_INACTIVE;
      break;
    case DIAGSESSION_LEAVE:
      printf("State %i: Leave extended diagnostic session\n",state);
      //Back to default session
      if(cs_uds_client_diagsession(data,CS_UDS_SI_DSC_DS, RESPREQ, &diag_data)){
	printf("State %i:Diagnostic request failed: %s\n",state,cs_uds_strlerror(data));
	state = WAITFOR;
	break;
      }
      cs_uds_channel_reset_timing(data);
      switch(myrand(3)){
      case 0:
	wtime = 5000; //5s
	state = WAIT_INACTIVE;
	break;
      case 1:
	state = ECU_RESET;
	break;
      case 2:
	wtime = 5000; //5s
	state = WAIT_INACTIVE;
	break;
      }
      break;
    case WAIT_INACTIVE:
      printf("State %i: Wait inactive for %u ms\n",state,wtime);
      //Sleep, no CAN activity, force testitem to shut down
      cs_process_sleep(wtime);
      state = WAITFOR;
      break;
    }
  }

  cs_uds_can_close(data);
    
  cs_uds_free(data);
  return 0;
}
