/* 
    Copyright (C) 2013 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @example main_udsserver.c
 * @author Andreas Stewering-Bone
 * @date 9 Oct 2014
 * @brief Simple UDS server application for testing.
 *
 * Simple UDS server application for testing.
*/

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <getopt.h>
#include "cs_uds.h"
#include "cs_isotp.h"
#include "cs_socketcan.h"

//Verbose level
int verbose = 0;

/********************************************************************/
void lverbose(int level,const char *fmt, ...)
/********************************************************************/
{
  if(verbose >= level){
    char buffer[CS_CHARBUF];
    va_list ap;
    memset(buffer,0,sizeof(buffer));
    va_start(ap, fmt);
    vsnprintf(buffer,CS_CHARBUF, fmt, ap);
    va_end(ap);
    printf(buffer);
  }
}

/********************************************************************/
int cb_diagsession(void *srv_ctx,      /**< Server context pointer*/
		   uint8_t subfunc     /**< Subfunction called by client,  CS_UDS_SI_DSC_* (cs_uds_defines.h)*/
		   )
/********************************************************************/
{
  switch(subfunc){ /*Subfunction*/
  case CS_UDS_SI_DSC_DS:
    lverbose(1,"Default session request\n");
    break;
  case CS_UDS_SI_DSC_PRGS:
    lverbose(1,"Programming session request\n");
    break;
  case CS_UDS_SI_DSC_EXTDS:
    lverbose(1,"Extended diagnostic session request\n");
    break;
  case CS_UDS_SI_DSC_SSDS:
    lverbose(1,"Safety diagnostic session request\n");
    break;
  case CS_UDS_SI_DSC_EOL:
    lverbose(1,"EOL diagnostic session request\n");
    break;
  case CS_UDS_SI_DSC_SUP:
    lverbose(1,"Supplier diagnostic session request\n");
    break;
  default:
    lverbose(1,"Diagsession: Subfunction is not supported\n");
    return CS_UDS_ERROR_SFNS;
  }
  return CS_UDS_ERROR_PR;
}

/*****************************************************************************/
int cb_ecureset(void *srv_ctx,         /**< Server context pointer*/
		uint8_t subfunc,       /**< Subfunction called by client,  CS_UDS_SI_ER_* (cs_uds_defines.h)*/
		uint8_t *powerDownTime /**< Return value, power down time*/
		)
/*****************************************************************************/
{
  switch(subfunc){
  case CS_UDS_SI_ER_HR:
    lverbose(1,"ECUReset: Hard reset requested\n");
    break;
  case CS_UDS_SI_ER_KOFFONR:
    lverbose(1,"ECUReset: Key off on Reset requested\n");
    break;
  case CS_UDS_SI_ER_SR:
    lverbose(1,"ECUReset: Soft reset requested\n");
    break;
  case CS_UDS_SI_ER_ERPSD:
    lverbose(1,"ECUReset: Enable rapid power shutdown requested\n");
    break;
  case CS_UDS_SI_ER_DRPSD:
    lverbose(1,"ECUReset: Disable rapid power shutdown requested\n");
    *powerDownTime = 0x15;
    break;
  default:
    return CS_UDS_ERROR_SFNS;
  }
  return CS_UDS_ERROR_PR;
}

/*****************************************************************************/
int cb_securityaccess_seed(void *srv_ctx,      /**< Server context pointer*/
			   uint8_t subfunc,    /**< Subfunction and requesteted seed level*/
			   uint8_t *seed,      /**< Seed storage buffer*/
			   int *seedlen        /**< Seed buffer length pointer, initialize on max allowed buffer len*/
			   )
/*****************************************************************************/
{
  int len = 3;
  uint8_t seedbuffer[len];

  seedbuffer[0] = 0x01;
  seedbuffer[1] = 0x02;
  seedbuffer[2] = 0x03;

  memcpy(seed,seedbuffer,len);
  *seedlen = len;
  return CS_UDS_ERROR_PR;
}

/*****************************************************************************/
int cb_securityaccess_key(void *srv_ctx,      /**< Server context pointer*/
			  uint8_t subfunc,    /**< Subfunction and key level*/
			  uint8_t *key,       /**< Key buffer*/
			  int keylen          /**< Key buffer length*/
			  )
/*****************************************************************************/
{
  //Always ok
  return CS_UDS_ERROR_PR;
}
  
/*****************************************************************************/
int cb_controldtcsetting(void *srv_ctx,       /**< Server context pointer*/
			 uint8_t subfunc      /**< Subfunction called, CS_UDS_SI_CDTCS_* (cs_uds_defines.h)*/
			 )
/*****************************************************************************/
{
  switch(subfunc){ /*Subfunction*/
  case CS_UDS_SI_CDTCS_ON:
    lverbose(1,"ControlDTCSettings: Switch on\n");
    break;
  case CS_UDS_SI_CDTCS_OFF:
    lverbose(1,"ControlDTCSettings: Switch off\n");
    break;
  default:
    lverbose(1,"ControlDTCSettings: Subfunction is not supported\n");
    return CS_UDS_ERROR_SFNS;
  }
  return CS_UDS_ERROR_PR;
}

/*****************************************************************************/
int cb_linkcontrol(void *srv_ctx,             /**< Server context pointer*/
		   uint8_t subfunc,           /**< Subfunction called CS_UDS_SI_LC_* (cs_uds_defines.h)*/
		   uint32_t data              /**< Optional data, subfunction depended*/
		   )
/*****************************************************************************/
{
  //Always ok
  return CS_UDS_ERROR_PR;
}

/*****************************************************************************/
 int cb_testerpresent(void *srv_ctx           /**< Server context pointer*/
		      )
/*****************************************************************************/
{
  return CS_UDS_ERROR_PR;
}

/*****************************************************************************/
int cb_readdatabyidentifier(void *srv_ctx,    /**< Server context pointer*/
			    uint16_t did,     /**< DID requested*/
			    uint8_t *data,    /**< Raw data buffer*/
			    int *datalen      /**< Raw data buffer length pointer, initialized on max allowd value*/ 
			    )
/*****************************************************************************/
{
  int i;
  lverbose(1,"ReadDatabyIdentifier: Recv DID request %x \n",did);
  *datalen = rand()%0xFF;
  for( i = 0; i < *datalen; i++){
    data[i] = rand()%0xFE;
  }
  return CS_UDS_ERROR_PR;
}

/*****************************************************************************/
int cb_writedatabyidentifier(void *srv_ctx,   /**< Server context pointer*/
			     uint16_t did,    /**< DID to write*/
			     uint8_t *data,   /**< Raw data buffer*/
			     int datalen      /**< Raw data buffer length*/
			     )
/*****************************************************************************/
{
  return CS_UDS_ERROR_PR;
}

/*****************************************************************************/
int cb_routinecontrol(void *srv_ctx,         /**< Server context pointer*/
		      uint8_t subfunction,   /**< Subfunction called CS_UDS_SI_RC_* (cs_uds_defines.h)*/
		      uint16_t routine,      /**< Routine to call*/
		      uint8_t *reqbuffer,    /**< Request buffer data*/
		      int reqlen,            /**< Request buffer length*/
		      uint8_t *respbuffer,   /**< Reponse buffer*/
		      int *resplen,          /**< Response bufferlength pointer, initialized on max allowed value*/
		      uint8_t *routineinfo,  /**< Optional Routine info data*/
		      uint8_t *send_routineinfo/**< Flag point send info data, if value != 0, sen routine info*/
		      )
/*****************************************************************************/
{
  respbuffer[0] = 0x01;
  respbuffer[1] = 0x02;
  *resplen = 2;
  *routineinfo = 0x13;
  *send_routineinfo = 0x01;
  return CS_UDS_ERROR_PR;
}

/*****************************************************************************/
int cb_cleardtcinformation(void *srv_ctx,     /**< Server context pointer*/
			   uint32_t mask      /**< DTC mask*/
			   )
/*****************************************************************************/
{
  return CS_UDS_ERROR_PR;
}

/*****************************************************************************/
int cb_readdtcinformation_getnumberofdtcbymask(void *srv_ctx,       /**< Server context pointer*/
					       uint8_t subfunction,   /**< Subfunction called CS_UDS_SI_RDTCI_* (cs_uds_defines.h)*/
					       uint8_t dtcstatusmask,/**< DTC status mask*/
					       uint8_t dtcseveritymask,/**< DTC severity mask*/
					       uint8_t *dtcavailmask,/**< DTC status available mask*/
					       uint8_t *format,     /**< Response format CS_UDS_DTC_FORMAT_* (cs_uds_defines.h)*/
					       uint16_t *count      /**< DTC count*/
					       )
/*****************************************************************************/
{
  *dtcavailmask = dtcstatusmask;
  *format = CS_UDS_DTC_FORMAT_J2012_DADTCF00;
  *count = 16;
  return CS_UDS_ERROR_PR;
}

/*****************************************************************************/
int cb_readdtcinformation_getdtcbymask(void *srv_ctx,       /**< Server context pointer*/
				       uint8_t subfunction,   /**< Subfunction called CS_UDS_SI_RDTCI_* (cs_uds_defines.h)*/
				       uint8_t dtcstatusmask,/**< DTC status mask*/
				       uint8_t *dtcavailmask,/**< DTC status available mask*/
				       uint32_t *dtc,       /**< DTC buffer*/
				       uint8_t *status,     /**< Status buffer*/
				       int *count           /**< Length of buffer dtc and status, initialized on max length*/
				       )
/*****************************************************************************/
{
  int i;
  *dtcavailmask = dtcstatusmask;
  *count = 12;
  for( i = 0; i < *count; i++){
    dtc[i] = i;
    status[i] = i;
  }
  return CS_UDS_ERROR_PR;
}

/*****************************************************************************/
int cb_readdtcinformation_reportsnapshotidentification(void *srv_ctx,       /**< Server context pointer*/
						       uint32_t *dtc,       /**< DTC buffer*/
						       uint8_t *status,     /**< Status buffer*/
						       int *count           /**< Length of buffer dtc and status, initialized on max length*/
						     )
/*****************************************************************************/
{
  int i;
  *count = 21;
  for( i = 0; i < *count; i++){
    dtc[i] = i;
    status[i] = i;
  }
  return CS_UDS_ERROR_PR;
}

/*****************************************************************************/
int cb_readdtcinformation_reportdtcmaskrecord(void *srv_ctx,       /**< Server context pointer*/
					      uint8_t subfunction,   /**< Subfunction called CS_UDS_SI_RDTCI_* (cs_uds_defines.h)*/
					      uint8_t dtcstatusmask,/**< DTC status mask*/
					      uint8_t dtcseveritymask,/**< DTC severity mask*/
					      uint32_t dtcmask,    /**< DTC mask*/
					      uint8_t *dtcavailmask,/**< DTC status available mask*/
					      uint8_t *serverity,  /**< Severity buffer*/
					      uint8_t *func,       /**< Functional buffer*/
					      uint32_t *dtc,       /**< DTC buffer*/
					      uint8_t *status,     /**< Status buffer*/
					      int *count           /**< Length of buffer, initialized on max length*/
					      )
/*****************************************************************************/
{
  int i;
  *count = 27;
  for( i = 0; i < *count; i++){
    dtc[i] = i;
    status[i] = i;
    serverity[i] = i;
    func[i] = i;
  }
  *dtcavailmask = dtcstatusmask;
  return CS_UDS_ERROR_PR;
}


void help(void)
{
  printf(
          "Usage: testudserver [OPTION]\n"
          "Test uds stack.\n"
          "\n"
          "Options:\n"
	  "  -i, --iface    CAN_Interface\n"
	  "  -s, --sa       hex source adress\n"
	  "  -t, --ta       hex target adress\n"
          "  -v  --verbose  Verbose output\n"
          "      --help     display this help and exit\n");
}

int main(int argc, char **argv)
{
  int c;
  char *iface = NULL;
  cs_uds_service_t *data = NULL;
  struct timespec waittime;
  unsigned long int sa = 0;
  unsigned long int ta = 0;
  
  while (1) {
    static struct option long_options[] = {
      /* These options set a flag. */
      {"iface", required_argument, NULL, 'i'},
      {"sa", required_argument, NULL, 's'},
      {"ta", required_argument, NULL, 't'},
      {"verbose",    no_argument, NULL, 'v'},
      {"help",    no_argument, NULL, 'h'},
      {0, 0, 0, 0}
    };

    /* getopt_long stores the option index here. */
    int option_index = 0;

    c = getopt_long (argc, argv, "i:hs:t:v",
                     long_options, &option_index);
    
    /* Detect the end of the options. */
    if (c == -1)
      break;
    
    switch (c) {
    case 0:
      /* If this option set a flag, do nothing else now. */
      if (long_options[option_index].flag != 0)
        break;
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
      break;
    case 'i': iface = optarg;     break;
    case 's': sa = strtoul(optarg,(char**)NULL,16);     break;
    case 't': ta = strtoul(optarg,(char**)NULL,16);     break;
    case 'v': verbose = verbose + 1; if(verbose > 2){verbose = 2;}    break;
    case 'h': help(); exit(0);    break;
    case '?':
      /* getopt_long already printed an error message. */
      printf( "Typ `testudsserver --help' for more information\n");
      return -1;
      break;
    default:
      return -1;
      break;
    }
  }

  /*Check inputs*/

  if(iface == NULL){
    printf( "Please check input arguments\n");
    printf( "Typ `testudsserver --help' for more information\n");
    return -1;
  }

  waittime.tv_sec = 0;
  waittime.tv_nsec = 1000000L;

  if((sa == 0)|| (ta == 0)){
    printf("Invalid sa or ta adress\n");
    return -1;
  }

  /*First check Library version*/
  if(CS_UDS_CHECKVERSION){
    printf("Incompatible Library interface, Application was compiled with version %i, Library has version %i\n",
	   CS_UDS_VERSION_MAGIC,cs_uds_getversion());
    return -1;
  }
  //Set verbose level
  switch (verbose){
  case 0:
    cs_uds_debuglevel(CS_DEBUG_LEVEL_NONE);
    cs_isotp_debuglevel(CS_DEBUG_LEVEL_NONE);
    cs_socketcan_debuglevel(CS_DEBUG_LEVEL_NONE);
    break;
  case 1:
    cs_uds_debuglevel(CS_DEBUG_LEVEL_DEBUG);
    cs_isotp_debuglevel(CS_DEBUG_LEVEL_DEBUG);
    cs_socketcan_debuglevel(CS_DEBUG_LEVEL_DEBUG);
    break;
  case 2:
    cs_uds_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
    cs_isotp_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
    cs_socketcan_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
    break;
  }

  //Prepare callback table
  //server callback structure
  cs_uds_server_cb_t cb;
  memset(&cb,0,sizeof(cs_uds_server_cb_t));
  cb.server_cb_diagsession = cb_diagsession;
  cb.server_cb_ecureset = cb_ecureset;
  cb.server_cb_securityaccess_seed = cb_securityaccess_seed;
  cb.server_cb_securityaccess_key = cb_securityaccess_key;
  cb.server_cb_controldtcsetting = cb_controldtcsetting;
  cb.server_cb_linkcontrol = cb_linkcontrol;
  cb.server_cb_testerpresent = cb_testerpresent;
  cb.server_cb_readdatabyidentifier = cb_readdatabyidentifier;
  cb.server_cb_writedatabyidentifier = cb_writedatabyidentifier;
  cb.server_cb_routinecontrol = cb_routinecontrol;
  cb.server_cb_cleardtcinformation = cb_cleardtcinformation;
  cb.server_cb_readdtcinformation_getnumberofdtcbymask = cb_readdtcinformation_getnumberofdtcbymask;
  cb.server_cb_readdtcinformation_getdtcbymask = cb_readdtcinformation_getdtcbymask;
  cb.server_cb_readdtcinformation_reportsnapshotidentification = cb_readdtcinformation_reportsnapshotidentification;
  cb.server_cb_readdtcinformation_reportdtcmaskrecord = cb_readdtcinformation_reportdtcmaskrecord;


  if((data = cs_uds_init()) == NULL){
    printf("Open interface failed\n");
    return -1;
  }
  
  if(cs_uds_can_init(data,iface, sa, ta)){
    printf("Init CAN layer failed\n");
    return -1;
  }

  if(cs_uds_can_open(data)){
    printf("Open CAN layer failed\n");
    return -1;
  }

  //Set table
  cs_uds_server_set_context(data,NULL,&cb);
  
  while(1){
    cs_uds_server_process(data);
    //printf("Loop\n");
    nanosleep(&waittime, NULL);
  }
  
  
  cs_uds_can_close(data);
  cs_uds_free(data);
  return 0;
}


