/*  
    Copyright (C) 2014 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @example main_xcpa2l.c
 * @author Andreas Stewering-Bone
 * @date 9 Oct 2014
 * @brief Simple XCP A2L file parsing application for testing the XCP A2L parsing API.
 *
 * Simple XCP A2L file parsing application for testing the XCP A2L parsing API
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <time.h>

#include "cs_xcp_a2l.h"

#define MAXSIGNALS 0xFF

void help(void)
{
  printf(
          "Usage: testxcpa2l [OPTION]\n"
          "Test XCP A2L parser stack.\n"
          "\n"
          "Options:\n"
          "  -i, --input    A2L input file\n"
	  "  -p, --project  A2L project name\n"
	  "  -m, --module   A2L module name\n"
	  "  -s, --signal   Signal name, can be set multiple times\n"
          "      --help     display this help and exit\n");
}

int main(int argc, char **argv)
{
  int c;
  char * filename = NULL;
  char * project = NULL;
  char * module = NULL;
  char * signals[MAXSIGNALS];
  int lt_signals = 0;
  cs_xcp_a2l_t data;
  cs_xcp_a2l_commode_t mode;
  cs_xcp_can_t *ifcan = NULL;
  cs_xcp_a2l_daq_proc_t daq;
  int asap2major = 0;
  int asap2minor = 0;
  cs_a2l_measurement_t me;
  int i;

  while (1) {
    static struct option long_options[] = {
      /* These options set a flag. */
      {"input", required_argument, NULL, 'i'},
      {"project", required_argument, NULL, 'p'},
      {"module", required_argument, NULL, 'm'},
      {"signal", required_argument, NULL, 's'},
      {"help",    no_argument, NULL, 'h'},
      {0, 0, 0, 0}
    };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    c = getopt_long (argc, argv, "i:p:m:s:h",
                     long_options, &option_index);

    /* Detect the end of the options. */
    if (c == -1)
      break;

    switch (c) {
    case 0:
      /* If this option set a flag, do nothing else now. */
      if (long_options[option_index].flag != 0)
        break;
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
      break;
    case 'i': filename = optarg;     break;
    case 'p': project = optarg;     break;
    case 'm': module = optarg;     break;
    case 's':
      signals[lt_signals] = optarg; 
      lt_signals++;
      if(lt_signals >= MAXSIGNALS){
	printf("Invalid number of signal defintion\n");
	return -1;
      }     
      break;
    case 'h': help(); exit(0);    break;
    case '?':
      /* getopt_long already printed an error message. */
      printf( "Typ `testxcpa2l --help' for more information\n");
      return -1;
      break;
    default:
      return -1;
      break;
    }
  }

  /*Check inputs*/
  if((filename==NULL)||(project == NULL)||(module == NULL)){
    printf( "Please check input arguments\n");
    printf( "Typ `testxcpa2l --help' for more information\n");
    return -1;
  }

  //Try to open file
  if(cs_xcp_a2l_open(&data,filename,project,module,"XCP")){
    printf("Failed to open file %s\n",filename);
    return -1;
  }

  //Get ASAP-Version
  if(cs_a2l_getversion(&data.a2l,&asap2major,&asap2minor)){
    printf("Failed to get ASAP2 version\n");
    return -1;
  }
  printf("ASAP2 Version Major %i Minor %i\n",asap2major,asap2minor);

  
  //Get Protocol settings
  if(cs_xcp_a2l_GetProtocoll(&data,&mode)){
    printf("Failed to parse protocol section\n");
    return -1;
  }
  cs_xcp_a2l_print_commode(&mode);

  //Get DAQ settings
  if(cs_xcp_a2l_GetDAQ(&data,&daq)){
    printf("Failed to parse DAQ section\n");
    return -1;
  }
  cs_xcp_a2l_print_processor_info(&daq);

  //Get Physical Interface CAN settings
  if(cs_xcp_a2l_GetPhysInterfaceCAN(&data,ifcan)){
    printf("Failed to parse Interface CAN section\n");
    return -1;
  }
  cs_xcp_a2l_print_PhysInterfaceCAN(ifcan);

  //Try to get Signal
  for (i = 0; i < lt_signals; i++){
    if(cs_a2l_GetMeasurement(&data.a2l,&me,signals[i])){
      printf("Failed to parse signal %s\n",signals[i]);
      return -1;
    }
    cs_a2l_print_measurement(&me);
  }

  //Close file
  if(cs_xcp_a2l_close(&data)){
    printf("Failed to close file\n");
    return -1;
  }

  return 0;
}
