/*  
    Copyright (C) 2014 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

/**
 * @example main_xcpapi.c
 * @author Andreas Stewering-Bone
 * @date 9 Oct 2014
 * @brief Simple XCP master application for testing the XCP high level API.
 *
 * Simple XCP master application for testing the XCP high level API.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <time.h>

#include "cs_xcp_api.h"


#define MAXSIGNALS 0xFF

/** \struct cbdata
    Callback private data structure
*/
struct cbdata{
    double rvalue;     /**< Feedback value*/
    double rvaluets;   /**< Feedback timestamp*/
    int idx;           /**< Index of feddback, only for readable output*/
    int recv;          /**< Callback increment, to check for overrun and synch*/
};

void help(void)
{
    printf(
        "Usage: testxcpapi [OPTION]\n"
        "Test XCP high level stack.\n"
        "\n"
        "Options:\n"
        "  -i, --input    A2L input file\n"
        "  -d, --device   CAN device\n"
        "  -p, --project  Project name\n"
        "  -m, --project  Module name\n"
        "  -s, --signal   Signal name for DAQ transmission, can be set multiple times\n"
	"  -e, --event    Event name for DAQ transmission\n"
        "      --help     display this help and exit\n");
}

/** Callback function wich is called by XCP stack*/
void mycallback(
    double value,   /**< Feedback value*/
    double ts,      /**< Feedback timestamp*/
    void *priv      /**< Private data pointer*/
    )
{
    struct cbdata *ldata = (struct cbdata*) priv;
    printf("Callback Value %i: TS: %f Value: %f\n",ldata->idx,ldata->rvaluets,ldata->rvalue);
    ldata->rvalue = value;
    ldata->rvaluets = ts;
    ldata->recv++;
}


int main(int argc, char **argv)
{
    struct cbdata rdata[MAXSIGNALS];
    int c;
    int i,j;
    char * filename = NULL;
    char * project = NULL;
    char * module = NULL;
    char * device = NULL;
    char * signals[MAXSIGNALS];
    char * event = NULL;
    int lt_signals = 0;
    cs_xcp_t data;
    struct timespec ts_sleep;
    int printout = 0;

    ts_sleep.tv_sec = 0;
    ts_sleep.tv_nsec = 1000000; //1000nues


    while (1) {
        static struct option long_options[] = {
            /* These options set a flag. */
            {"input", required_argument, NULL, 'i'},
            {"device", required_argument, NULL, 'd'},
            {"project", required_argument, NULL, 'p'},
            {"module", required_argument, NULL, 'm'},
            {"signal", required_argument, NULL, 's'},
            {"event", required_argument, NULL, 'e'},
            {"help",    no_argument, NULL, 'h'},
            {0, 0, 0, 0}
        };
        /* getopt_long stores the option index here. */
        int option_index = 0;

        c = getopt_long (argc, argv, "i:d:p:m:s:e:h",
                         long_options, &option_index);

        /* Detect the end of the options. */
        if (c == -1)
            break;

        switch (c) {
        case 0:
            /* If this option set a flag, do nothing else now. */
            if (long_options[option_index].flag != 0)
                break;
            printf ("option %s", long_options[option_index].name);
            if (optarg)
                printf (" with arg %s", optarg);
            printf ("\n");
            break;
        case 'i': filename = optarg;     break;
        case 'p': project = optarg;     break;
        case 'm': module = optarg;     break;
        case 'e': event = optarg;     break;
        case 's': 
	  signals[lt_signals] = optarg; 
	  lt_signals++;
	  if(lt_signals >= MAXSIGNALS){
	    printf("Invalid number of signal defintion\n");
	    return -1;
	  }     
	  break;
        case 'd': device = optarg;     break;
        case 'h': help(); exit(0);    break;
        case '?':
            /* getopt_long already printed an error message. */
            printf( "Typ `testxcpapi --help' for more information\n");
            return -1;
            break;
        default:
            return -1;
            break;
        }
    }

    //First check Library version
    if(CS_XCP_CHECKVERSION){
      printf("Incompatible Library interface, Application was compiled with version %i, Library has version %i\n",
	     CS_XCP_VERSION_MAGIC,cs_xcp_getversion());
      return -1;
    }

    //Prepare callback structure
    for(i = 0; i < MAXSIGNALS; i++){
      rdata[i].idx = i;
      rdata[i].rvalue = 0;
      rdata[i].rvaluets = 0;
      rdata[i].recv = 0;
    }
    
    /*Check inputs*/
    if((filename==NULL) || (device == NULL) || (project == NULL) || (module == NULL)){
      printf( "Please check input arguments\n");
      printf( "Typ `testxcpapi --help' for more information\n");
      return -1;
    }
    
    if(cs_xcp_api_load(&data,filename,project,module,"XCP",100)){
      printf("Failed to load basic data, Error: %s\n",cs_xcp_get_lerrorstr(&data));
      return -1;
    }
    
    if(cs_xcp_api_open(&data,XCP_CAN,device)){
      printf("Failed to open CAN interface, Error: %s\n",cs_xcp_get_lerrorstr(&data));
      return -1;
    }

    if(cs_xcp_api_connect(&data,NULL)){
      printf("Failed to connect to slave, Error: %s\n",cs_xcp_get_lerrorstr(&data));
      return -1;
    }
    
    if(lt_signals > 0){
      if(event == NULL){
	printf("No DAQ event defined\n");
	return -1;
      }
      for(i = 0; i < lt_signals; i++){
        if(cs_xcp_api_addSignalDAQ(&data,signals[i],event,1,NULL,NULL,&mycallback,&rdata[i])){
            printf("Failed to find Signal %s, Error: %s\n",signals[i],cs_xcp_get_lerrorstr(&data));
            return -1;
        }
      }

      if(cs_xcp_api_configureDAQ(&data)){
	printf("Failed to configure DAQ transmission, Error: %s\n",cs_xcp_get_lerrorstr(&data));
	return -1;
      }

      //Set timeout for DAQ receive timeout detection,
      // here receive period * 1.95  is allowed
      cs_xcp_api_setDAQTimeout(&data,1.95);
      
      if(cs_xcp_api_startDAQ(&data)){
	printf("Failed to start DAQ transmission, Error: %s\n",cs_xcp_get_lerrorstr(&data));
	return -1;
      }
      printf("Start loop\n");
      //Do cylic message handling
      for(i = 0; i < 100L; i++){
	printout = 1;
	nanosleep(&ts_sleep,NULL);
	if(cs_xcp_api_process(&data)){
	  printf("Failed to receive data: %X, %s\n",cs_xcp_get_lerror(&data),cs_xcp_get_lerrorstr(&data));
	  return -1;
	}
	
	//Test for ODT recv timeout
	if(cs_xcp_api_hasDAQTimeout(&data)){
	  printf("DAQ ODT recv timeout detected\n");
	}
	
	//Try to sync output
	for(j = 0; j < lt_signals; j++){
	  if(rdata[j].recv > 1){
	    printf("Overrun detected, Rvalue %i\n",j);
	  }
	  if(rdata[j].recv == 0){
	    printout = 0;
	  }
	}
	if(printout){
	  printf("Receive %llu pakets for DAQ %i\n",
		 (long long unsigned int)cs_xcp_data_get_daq_counter(&data,0),0);
	  for(j = 0; j < lt_signals; j++)
	    rdata[j].recv = 0;
	}
	if(cs_xcp_get_indicator(&data) & XCP_INDICATOR_OVERLOAD){
	  printf("Get an overload indication\n");
	  cs_xcp_reset_indicator(&data,XCP_INDICATOR_MASK_ALL);
	}
      }
      
      if(cs_xcp_api_stopDAQ(&data)){
        printf("Failed to stop DAQ transmission, Error: %s\n",cs_xcp_get_lerrorstr(&data));
        return -1;
      }
    }
    if(cs_xcp_api_close(&data)){
      printf("Failed to close interface, Error: %s\n",cs_xcp_get_lerrorstr(&data));
      return -1;
    }
    
    return 0;
}
