/*  
    Copyright (C) 2014 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @example main_xcpmaster.c
 * @author Andreas Stewering-Bone
 * @date 9 Oct 2014
 * @brief Simple XCP master application for testing the XCP low level API.
 *
 * Simple XCP master application for testing the XCP low level API.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <time.h>

#include "cs_xcp.h"


#define DOSLEEP 0

#if DOSLEEP
#define WAIT(X) sleep(X)
#else
#define WAIT(X)
#endif

void help(void)
{
  printf(
          "Usage: testxcpmaster [OPTION]\n"
          "Test XCP master stack.\n"
          "\n"
          "Options:\n"
          "  -i, --iface    CAN-Interface\n"
	  "  -s, --sa       hex source adress\n"
	  "  -t, --ta       hex target adress\n"
          "      --help     display this help and exit\n");
}

int main(int argc, char **argv)
{
  int c;
  cs_xcp_t data;
  cs_xcp_status_t status;
  cs_xcp_a2l_commode_t commode;
  char * iface = NULL;
  uint32_t *resp_getslaveid = NULL;
  int resplen_getslaveid = 0;
  unsigned long int sa = 0;
  unsigned long int ta = 0;
  int i;
  double feedback = 0.0;
  double tsfeedback;
  double feedback1 = 0.0;
  double feedback2 = 0.0;
  double feedback3 = 0.0;
  double feedback4 = 0.0;
  double feedbacktimer = 0.0;
  double feedbackchannel1 = 0.0;
  double jitter;
  struct timespec ts_sleep;
  cs_a2l_compu_method_t cm;

  ts_sleep.tv_sec = 0;
  ts_sleep.tv_nsec = 500000; //500nues

  //Create default compu method object
  cs_a2l_compu_method_linear_create(&cm,"Default",NULL,NULL,NULL,1.0,0.0);


  while (1) {
    static struct option long_options[] = {
      /* These options set a flag. */
      {"iface", required_argument, NULL, 'i'},
      {"sa", required_argument, NULL, 's'},
      {"ta", required_argument, NULL, 't'},
      {"help",    no_argument, NULL, 'h'},
      {0, 0, 0, 0}
    };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    c = getopt_long (argc, argv, "i:ht:s:",
                     long_options, &option_index);

    /* Detect the end of the options. */
    if (c == -1)
      break;

    switch (c) {
    case 0:
      /* If this option set a flag, do nothing else now. */
      if (long_options[option_index].flag != 0)
        break;
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
      break;
    case 'i': iface = optarg;     break;
    case 's': sa = strtoul(optarg,(char**)NULL,16);     break;
    case 't': ta = strtoul(optarg,(char**)NULL,16);     break;
    case 'h': help(); exit(0);    break;
    case '?':
      /* getopt_long already printed an error message. */
      printf( "Typ `testxcpmaster --help' for more information\n");
      return -1;
      break;
    default:
      return -1;
      break;
    }
  }

  //First check Library version
  if(CS_XCP_CHECKVERSION){
    printf("Incompatible Library interface, Application was compiled with version %i, Library has version %i\n",
	   CS_XCP_VERSION_MAGIC,cs_xcp_getversion());
    return -1;
  }
  
  /*Check inputs*/
  if(iface==NULL){
    printf( "Please check input arguments\n");
    printf( "Typ `testxcpmaster --help' for more information\n");
    return -1;
  }

  /*Init XCP-Stack*/
  if(cs_xcp_init(&data)){
    printf("Open XCP stack failed\n");
    return -1;
  }

  /*Configure CAN interface*/
  if(cs_xcp_iface_can_configure(&data,ta,sa,0x100)){
    printf("Configure interface failed: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  /*Start SocketCAN interface*/
  if(cs_xcp_iface_can_open(&data,iface)){
    printf("Open interface failed: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  /*Search for XCP slaves, first get CMD/STIM ID*/
  if(cs_xcp_cmd_get_slave_id(&data,0,&resp_getslaveid,&resplen_getslaveid)){
    printf("Failed to send broadcast request: %s\n",cs_xcp_get_lerrorstr(&data));
  } else {
    printf("Got slave response:\n");
    for(i = 0; i < resplen_getslaveid; i++){
      printf("Slave %i CAN ID: %X\n",i,resp_getslaveid[i]);
    }
  }

  //Try to connect
  if(cs_xcp_cmd_connect(&data,0)){
    printf("Failed to connect: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }
  sleep(1);

  //Get status
  if(cs_xcp_cmd_get_status(&data,&status)){
    printf("Failed to get status: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  
  //Get additional communication information
  commode = cs_xcp_a2l_data_get_commode(&data.xcpa2l);
  if(commode.optional){
    printf("Get additional communication information\n");
    if(cs_xcp_cmd_get_commode_info(&data)){
      printf("Failed to get additional communication information: %s\n",cs_xcp_get_lerrorstr(&data));
      return -1;
    }
  }

  //Send a manually synch
  if(cs_xcp_cmd_synch(&data)){
    printf("Failed to synch: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  //Try to set MTA
  if(cs_xcp_cmd_set_mta(&data,0x8051D32,0x00)){
    printf("Failed to set MTA: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  //Try to upload 7 Byte
  uint32_t uploadbuffer[7];
  int uploadlen = 7;
  if(cs_xcp_cmd_upload(&data,uploadbuffer,&uploadlen)){
    printf("Failed to Upload: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  //Try to get ASAM id
  uint32_t getidbuffer[255];
  int getidlen = 255;
  if(cs_xcp_cmd_get_id(&data,0,getidbuffer,&getidlen)){
    printf("Failed to Get ID: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }
  if(getidlen > 0){
    printf("GET ID: ");
    for(i = 0; i < getidlen; i++){
      printf("%X ",getidbuffer[i]);
    }
    printf("\n");
  }

  //Get DAQ processor info
  if(cs_xcp_cmd_get_daq_processor_info(&data,NULL)){
    printf("Failed to Get DAQ processor info: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  //Get DAQ resolution info
  if(cs_xcp_cmd_get_daq_resolution_info(&data,NULL)){
    printf("Failed to Get DAQ resolution info: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  //Get information about event channel
  for(i = 0; i < data.xcpa2l.daq_proc.ev_max; i++){
    if(cs_xcp_cmd_get_daq_event_info(&data,i,NULL)){
      if(cs_xcp_get_lerror(&data) == XCP_ERROR_CMD_UNKNOWN){
	printf("Sorry, get event information is not supported: %s\n",cs_xcp_get_lerrorstr(&data));	
	break;
      } else {
	printf("Failed to get event information: %s\n",cs_xcp_get_lerrorstr(&data));
	return -1;
      }
    }
  }

  //Free DAQ lists
  if(cs_xcp_cmd_free_daq(&data)){
    printf("Failed to free DAQ lists: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  //Alloc DAQ lists
  if(cs_xcp_cmd_alloc_daq(&data,1)){
    printf("Failed to alloc DAQ lists: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  //Get DAQ processor info
  if(cs_xcp_cmd_get_daq_processor_info(&data,NULL)){
    printf("Failed to Get DAQ processor info: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  if(cs_xcp_cmd_set_daq_list_mode(&data,
			      0, // DAQ 0
			      3, //Event channel 3
			      XCP_DAQ_DIR_DAQ, //Direction
			      0, //Transmit with PID
			      0, //enable timestamp
			      1000, //Prescaler
			      0xff //Priority
			      )){
    printf("Failed to setup DAQ : %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }


  //Clear DAQ list
  /*if(cs_xcp_clear_daq_list(&data,0)){
    printf("Failed to clear DAQ lists: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
    }*/

  //Alloc two ODT in DAQ
  if(cs_xcp_cmd_alloc_odt(&data,0,5)){
    printf("Failed to alloc ODT: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  //Alloc ODT Entry
  if(cs_xcp_cmd_alloc_odt_entry(&data,0,0,2)){
    printf("Failed to alloc ODT Entry: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  //Alloc ODT Entry
  if(cs_xcp_cmd_alloc_odt_entry(&data,0,1,1)){
    printf("Failed to alloc ODT Entry: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  //Alloc ODT Entry
  if(cs_xcp_cmd_alloc_odt_entry(&data,0,2,1)){
    printf("Failed to alloc ODT Entry: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  //Alloc ODT Entry
  if(cs_xcp_cmd_alloc_odt_entry(&data,0,3,1)){
    printf("Failed to alloc ODT Entry: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  //Alloc ODT Entry
  if(cs_xcp_cmd_alloc_odt_entry(&data,0,4,1)){
    printf("Failed to alloc ODT Entry: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  //Set working pointer, to first ODT entry
  if(cs_xcp_cmd_set_daq_ptr(&data,0,0,0)){
    printf("Failed to set DAQ pointer: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  //Store memory addr in ODT Entry
  if(cs_xcp_cmd_write_daq(&data,0xFF,2,0,0x8051D31)){
    printf("Failed to write DAQ: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  //Set working pointer, to second ODT entry
  if(cs_xcp_cmd_set_daq_ptr(&data,0,0,1)){
    printf("Failed to set DAQ pointer: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  //Store memory addr in ODT Entry
  if(cs_xcp_cmd_write_daq(&data,0xFF,2,0,0x8051C30)){
    printf("Failed to write DAQ: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }


  //Set working pointer, to first ODT entry
  if(cs_xcp_cmd_set_daq_ptr(&data,0,2,0)){
    printf("Failed to set DAQ pointer: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  //Store memory addr in ODT Entry
  if(cs_xcp_cmd_write_daq(&data,0xFF,3,0,0x8051C01)){
    printf("Failed to write DAQ: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  //Set working pointer, to first ODT entry
  if(cs_xcp_cmd_set_daq_ptr(&data,0,1,0)){
    printf("Failed to set DAQ pointer: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  //Store memory addr in ODT Entry
  if(cs_xcp_cmd_write_daq(&data,0xFF,4,0,0x8051D01)){
    printf("Failed to write DAQ: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  //Set working pointer, to first ODT entry
  if(cs_xcp_cmd_set_daq_ptr(&data,0,3,0)){
    printf("Failed to set DAQ pointer: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  //Store memory addr in ODT Entry
  if(cs_xcp_cmd_write_daq(&data,0xFF,4,0,0x8052F6C)){
    printf("Failed to write DAQ: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  //Set working pointer, to first ODT entry
  if(cs_xcp_cmd_set_daq_ptr(&data,0,4,0)){
    printf("Failed to set DAQ pointer: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  //Store memory addr in ODT Entry
  if(cs_xcp_cmd_write_daq(&data,0xFF,4,0,0x8052F78)){
    printf("Failed to write DAQ: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  //Register feedback value
  if(cs_xcp_entry_reg_value(&data,0,0,0,&feedback,cm,A2L_UINT,&tsfeedback,NULL,NULL)){
    printf("Failed to register feedback: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  if(cs_xcp_entry_reg_value(&data,0,0,1,&feedback1,cm,A2L_UINT,NULL,NULL,NULL)){
    printf("Failed to register feedback: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  if(cs_xcp_entry_reg_value(&data,0,1,0,&feedback3,cm,A2L_UINT,NULL,NULL,NULL)){
    printf("Failed to register feedback: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  if(cs_xcp_entry_reg_value(&data,0,2,0,&feedback4,cm,A2L_UINT,NULL,NULL,NULL)){
    printf("Failed to register feedback: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  if(cs_xcp_entry_reg_value(&data,0,3,0,&feedbacktimer,cm,A2L_FLOAT,NULL,NULL,NULL)){
    printf("Failed to register feedback: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  if(cs_xcp_entry_reg_value(&data,0,4,0,&feedbackchannel1,cm,A2L_FLOAT,NULL,NULL,NULL)){
    printf("Failed to register feedback: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  //Select DAQ list
  if(cs_xcp_cmd_start_stop_daq_list(&data,0,1)){
    printf("Failed to select DAQ list: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  //Get slave time jitter
  if(cs_xcp_cmd_get_daq_clock(&data,&jitter)){
    if(cs_xcp_get_lerror(&data) == XCP_ERROR_SLAVE_UNSUPPORTED){
      printf("Sorry, get daq clock is not supported: %s\n",cs_xcp_get_lerrorstr(&data)); 
    } else {
      printf("Failed to get DAQ clock: %s\n",cs_xcp_get_lerrorstr(&data));
      return -1;
    }
  } else {
    printf("Clock jitter: %f\n",jitter);
  }

  //Select DAQ list
  /*if(cs_xcp_start_stop_daq_list(&data,1,1)){
    printf("Failed to select DAQ list: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
    }*/

  //Start selected daq lists
  /*if(cs_xcp_start_stop_synch(&data,1)){
    printf("Failed to start DAQ lists: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
    }*/

  //Only receive data, cyclic mode, no blocking
  for(i = 0; i < 10000; i++){
    nanosleep(&ts_sleep,NULL);
    if(cs_xcp_receive(&data)){
      printf("Failed to receive data: %X, %s\n",cs_xcp_get_lerror(&data),cs_xcp_get_lerrorstr(&data));
      return -1;
    }
    if((i%1000) == 1){
      printf("Feedback value(DAQ 0, ODT 0, Entry 0):%f Timestamp: %f\n",feedback,tsfeedback);
      printf("Feedback value(DAQ 0, ODT 0, Entry 1):%f\n",feedback1);
      printf("Feedback value(DAQ 0, ODT 0, Entry 2):%f\n",feedback2);
      printf("Feedback value(DAQ 0, ODT 1, Entry 0):%f\n",feedback3);
      printf("Feedback value(DAQ 0, ODT 2, Entry 0):%f\n",feedback4);
      printf("Feedback value(DAQ 0, ODT 3, Entry 0):%f\n",feedbacktimer);
      printf("Feedback value(DAQ 0, ODT 4, Entry 0):%f\n",feedbackchannel1);
    }
  }
 
  //Start DAQ list
  /*if(cs_xcp_start_stop_daq_list(&data,0,0)){
    printf("Failed to stop DAQ list: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
    }*/

  //Stop all daq lists
  if(cs_xcp_cmd_start_stop_synch(&data,0)){
    printf("Failed to stop DAQ lists: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }
  
  //Try to disconnect
  if(cs_xcp_cmd_disconnect(&data)){
    printf("Failed to disconnect: %s\n",cs_xcp_get_lerrorstr(&data));
    return -1;
  }

  cs_xcp_close(&data);
  return 0;
}
