/*----------------------------------------------------------------------------
| File:
|   XCPSIM.C
|
| Project:
|   PC Emulation of a ECU using XCP on Ethernet or XCP on CAN
|
 ----------------------------------------------------------------------------*/

#include <stdlib.h>
#include <time.h>
#include <pthread.h>
#include <getopt.h>
#include <sys/socket.h>
#include <sys/mman.h>

// XCP driver
#include "XcpBasic.h"

// ECU simulation
#include "ecu.h"

// XCP handler
#include "XCPsim.h"


#define MAX_SAFE_STACK (1*1024*1024)
#define PREGETMEM (1*1024*1024)

///////////////////////////////////////////////////////////////////////
// Globals


int tcpMode                  = 0; // Ethernet or CAN, default is CAN
int mainRunning              = 0; // Status
unsigned short gTimer        = 0; // DAQ clock 
vuint8 gDebugLevel           = 1; // Debug output 
extern BYTE __xc_a;

V_MEMROM0 vuint8 MEMORY_ROM kXcpStationId[]="XCPSIM";


#if defined ( XCP_ENABLE_VECTOR_MAPNAMES )

#define MAP_FORMAT 29
#define MAP_NAME   "xcpsim"

unsigned char MapTest[500];
unsigned int MapTestSize;

vuint32 ApplXcpGetIdData( vuint8 **pData) {

  MapTestSize = sprintf((char*)MapTest,"%c%c%s.map",MAP_FORMAT,0,MAP_NAME);
  *pData = MapTest;
  return MapTestSize;
}

#endif


/****************************************************************************/
struct timespec addtimespec(struct timespec time1, struct timespec time2)
/****************************************************************************/
{
  struct timespec result;
  if((time1.tv_nsec + time2.tv_nsec) >= 1e9)
    {
      result.tv_sec = time1.tv_sec + time2.tv_sec + 1;
      result.tv_nsec = time1.tv_nsec +time2.tv_nsec -1e9;
    }
  else
    {
      result.tv_sec = time1.tv_sec + time2.tv_sec ;
      result.tv_nsec = time1.tv_nsec +time2.tv_nsec;
    }
  return result;
}


int set_thread_rt_opt(int priority, int policy,pthread_attr_t *attr)
{
  struct sched_param thread_param;

  /*Set realtime stuff*/
  mlockall(MCL_CURRENT|MCL_FUTURE);
  
  //Configure Thread Parameter
  if(pthread_attr_init(attr) !=0){
    printf("Thread Attribut init failed\n");
    return 0;
  }
  if(pthread_attr_setinheritsched(attr, PTHREAD_EXPLICIT_SCHED)!=0) {
    printf("Setting explicit scheduling inheritance failed\n");
    return 0;
  }
  if(pthread_attr_setschedpolicy(attr, policy) !=0) {
    printf("Thread Attribut Policy failed\n");
    return 0;
  }
  if((priority <= sched_get_priority_max(policy))&&(priority >= sched_get_priority_min(policy))) {
    thread_param.sched_priority = priority; 
  }
  else {
    printf("Priority value not valid, select highest priority\n");
    thread_param.sched_priority = sched_get_priority_max(policy); //highest 
  }
  if(pthread_attr_setschedparam(attr, &thread_param) !=0) {
    printf("Thread Attribut Priority failed\n");
    return 0;
  }
  if(pthread_attr_setstacksize(attr, MAX_SAFE_STACK)){
    printf("Thread Attribut Stacksize failed\n");
    return 0;
  }
  return 1;
}


/*
  Configuration depended functions
 */

//Read Byte from memory
vuint8 ApplXcpRead( vuint32 addr )
{
  //return *((vuint8*)addr);
  return 0x34;
}

//Write byte to memory
void XCP_FAR ApplXcpWrite( vuint32 addr, vuint8 data )
{
  //*((uint8_t*)addr) = (uint8_t)data;
}

/**************************************************************************/
// ApplXcpSend()
// ApplXcpSendFlush()
// ApplXcpGetPointer() 
// ApplXcpGetSeed()
// ApplXcpUnlock()
// Platform and implementation specific functions for the XCP driver
/**************************************************************************/

// Flush the tranmit queue
void ApplXcpSendFlush(void) {

  if (tcpMode) {

    tcpSendFlush();

  } 
}


// Transmit a message
void ApplXcpSend( vuint8 len, MEMORY_ROM BYTEPTR msg  ) {

  if (tcpMode) {

    tcpSend( len, msg );
    XcpSendCallBack();

  } 
  else {
    //printf("Send len %i Pointer %X\n",len,msg);
    canSend( len, msg );

  }


}


// Convert a XCP (BYTE,DWORD) address to a C pointer
BYTEPTR ApplXcpGetPointer( BYTE addr_ext, DWORD addr ) 
{
  uint8_t *a = (uint8_t*)addr;
  return (BYTEPTR)a;
}


// Seed & Key
#if defined(XCP_ENABLE_SEED_KEY) || defined(XCP_ENABLE_SEED_KEY)

BYTE xcpResourceMask = 0;

BYTE ApplXcpGetSeed( BYTE resourceMask, BYTE *seed )
{
  
  xcpResourceMask = resourceMask;

  /* Generate the seed */
  seed[0] = 0;
  seed[1] = 1;
  seed[2] = 2;
  seed[3] = 3;
  seed[4] = 4;
  seed[5] = 5;

  return 0;
}

vuint8 ApplXcpUnlock( MEMORY_CONST vuint8 *key, vuint8 length )
{
  /* Ckeck the key */
  /* key[0],key[1],key[2],key[3],key[4],key[5] */

  xcp.ProtectionStatus &= ~xcpResourceMask; /* Reset the appropriate resource protection mask bit */
  return 1;
}

#endif




///////////////////////////////////////////////////////////////////////
// main()
//---------------------------------------------------------------------
//

// help
static void usage( void ) {

  printf(
    "\n"
    "Usage:\n"
    "  XCPsim [options]\n"
    "\n"
    "  Options:\n"
    "    -s       Run in soft realtime, does not require root rights\n"
    "    -v <x>   Set output verbosity to x (0=Off,1=Cmd,2=+Res,3=+Raw CRO/DTO,4=+Raw CAN)\n"
    "\n"
    "  XCP Options:\n"
    "    -cro <x> Set CRO identifier to x (Default 1)\n"
    "    -dto <x> Set DTO identifier to x (Default 2)\n"
    "    -x       Use extended ids (Default standard)\n"
    "\n"
    "  Ethernet Options:\n"
    "    -udp     Use UDP\n"
    "    -tcp     Use TCP/IP\n"
    "    -port <x> Use port x (Default 5555)\n"
    "\n"
    "  CAN Options (Defaults from CAN-Hw Applet):\n"
    "    -i <name> Set CAN interface name\n"
    "\n"
    "  Keyboard Commands:\n"
    "    ESC      Exit\n"
    "\n"
    );

    
}

// Keyboard handler
int handleKey( int key ) {

  switch (key) {

    // ESC
    case 27:
      return 1;
      break;

    default:
      break;
  }

  return 0;
}


static pthread_t threadHandle;
static pthread_attr_t thread_attr;
static int threadRunning     = 0;
static struct timespec proc_time;
static uint32_t time_counter = 0;

//Get timestamp
uint32_t tmTimer(void)
{
  //printf("Get Timestamp: %X\n",time_counter);
  if(time_counter >= (1<<(kXcpDaqTimestampSize*8)))
    time_counter = 0;
  return time_counter;
}

// Task Sheduler
void* mainthread(void *data){
  
  uint32_t i = 0;
  struct timespec ts_sleep;
  printf("Start Main thread\n");

  ts_sleep.tv_sec = 0;
  ts_sleep.tv_nsec = 1000000; //1ms

  threadRunning = 1;
  while (threadRunning) {

    proc_time = addtimespec(proc_time,ts_sleep);
    /*Sleep*/
    clock_nanosleep(CLOCK_MONOTONIC,TIMER_ABSTIME,&proc_time,NULL);
    //Increment time counter
    time_counter++;

    if (mainRunning) {
      //printf("Loop\n");
      /* 1ms Task ECU Simulation */
      ecuCyclic();

#ifdef XCP_ENABLE_DAQ
      gTimer += 1;

      /* Event Channel 0 is cyclic 1 ms */
      XcpEvent(0);

      /* Event Channel 1 is cyclic 2 ms */
      if (i%2==0) {
        XcpEvent(1);
      }
      
      /* Event Channel 2 is cyclic 5 ms */
      if (i%5 ==0) {
	XcpEvent(2);
      }
      
      /* Event Channel 3 is cyclic 10 ms */
      if (i%10 ==0) {
	XcpEvent(3);
      }
#endif
 
      if (i%100 ==0) {
        /* Flush every 100ms */
        ApplXcpSendFlush();  
      }
      i++;
      i = i%100;
      /* XCP driver background processing */
      #ifdef XCP_ENABLE_CHECKSUM
        XcpBackground();
      #endif  

    }
  }

  return 0;
}


// Mainloop
int main( int argc, char *argv[])
{
  int c;
  int soft = 0;

  printf(
    "XCPsim ECU simulator for Ethernet or CAN (Basic Version) \n"
    "Vector Informatik GmbH, 2004\n\n"
    "Build " __DATE__ " " __TIME__ "\n"
  );


  // Get the channel assignment from the CAN control panel
  canGetDefaults();
  
  
  while (1) {
    static struct option long_options[] = {
      /* These options set a flag. */
      {"dto", required_argument, NULL, 'd'},
      {"cro", required_argument, NULL, 'c'},
      {"iface", required_argument, NULL, 'i'},
      {"extended", no_argument, NULL, 'x'},
      {"udp", no_argument, NULL, 'u'},
      {"tcp", no_argument, NULL, 't'},
      {"port", required_argument, NULL, 'p'},
      {"verbose", required_argument, NULL, 'v'},
      {"soft", no_argument, NULL, 's'},
      {"help",    no_argument, NULL, 'h'},
      {0, 0, 0, 0}
    };
    /* getopt_long stores the option index here. */
    int option_index = 0;
    
    c = getopt_long (argc, argv, "d:c:i:xutp:v:sh",
                     long_options, &option_index);
    
    /* Detect the end of the options. */
    if (c == -1)
      break;
    
    switch (c) {
    case 0:
      /* If this option set a flag, do nothing else now. */
      if (long_options[option_index].flag != 0)
        break;
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
      break;
    case 'i': caniface = optarg;
      printf("Using CAN interface %s\n",caniface);
      break;
    case 'c': gCroId = strtoul(optarg,(char**)NULL,16);
      printf("Use CRO CAN id 0x%X\n",gCroId);
      break;
    case 'd': gDtoId = strtoul(optarg,(char**)NULL,16);
      printf("Use DTO CAN id 0x%X\n",gDtoId);
      break;
    case 'p': 
      gPort =strtoul(optarg,(char**)NULL,16);  
      printf("Using port %u\n",gPort); 
      break;
    case 't': 
      tcpMode = 1;
      gProto = SOCK_STREAM;
      printf("Using TCP/IP\n");  
      break;
    case 'u': 
      tcpMode = 1;
      gProto = SOCK_DGRAM;
      printf("Using UDP\n");  
      break;
    case 'v':
      gDebugLevel = strtoul(optarg,(char**)NULL,16);    
      printf("Set debug level to %u\n",gDebugLevel);
      break;
    case 's': soft = 1;    break;
    case 'h': usage(); exit(0);    break;
    case '?':
      /* getopt_long already printed an error message. */
      printf( "Typ `testudsclient --help' for more information\n");
      return -1;
      break;
    default:
      printf("Invalid Option %c \n",c);
      return -1;
      break;
    }
  }
  
  if(caniface == 0){
    printf("Please configure interface name\n");
    usage();
    exit(-1);
  }
  
  // Initialize XCP driver
  XcpInit();
  
  
  // Initialize ECU simulation
  ecuInit();
  

  //Get start timestamp
  clock_gettime(CLOCK_MONOTONIC, &proc_time);
  
  // Create the DAQ thread
  if(soft){
    printf("Starting thread in soft realtime.!\n");
    if(pthread_create(&threadHandle, NULL, mainthread,NULL)!=0){
      printf("Can not create Thread. Exiting!\n");
      exit(0);
    }
  } else {
    printf("Starting thread in hard realtime.!\n");
    set_thread_rt_opt(99, SCHED_RR, &thread_attr);
    if(pthread_create(&threadHandle, &thread_attr, mainthread,NULL)!=0){
      printf("Can not create Thread. Exiting!\n");
      exit(0);
    }
  }

  // Mainloop
  if (tcpMode) {
    tcpMain();
  } else {
    canMain();
  }

  // Stop DAQ thread
  if (threadHandle) {
    threadRunning = 0;
  }

  return 0;
}

