/*  
    Copyright (C) 2014 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @example lsa2l.c
 * @author Andreas Stewering-Bone
 * @date 29 June 2015
 * @brief Simple A2L Tool to list characteristics or mesasurement data.
 *
 * Simple A2L Tool to list characteristics or mesasurement data. 
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <time.h>

#include "cs_a2l.h"

#define MAXCHAR 0xFF
#define MAXSIGNALS 0XFF
#define EXECNAME "can_lsa2l"

void help(void)
{
  printf(
          "Usage: %s [OPTION]\n"
          "List A2L characteristis or measurement fiels.\n"
          "\n"
          "Options:\n"
          "  -a, --a2l        A2L input file\n"
	  "  -p, --project    Project to search for\n"
	  "  -m, --module     Module to use\n"
	  "  -c, --characteristic Characteristic name for details, can be set multiple times\n"
	  "  -C, --listcharacteristic List characteristic list and quit\n"
	  "  -s, --signal     Measurement name for details, can be set multiple times\n"
	  "  -S, --listsignal List measurement list and quit\n"
          "      --help       display this help and exit\n",EXECNAME);
}

int main(int argc, char **argv)
{
  int c;
  char * filename = NULL;
  char * project = NULL;
  char * module = NULL;
  char * characteristic[MAXCHAR];
  int lt_characteristic = 0;
  char * signals[MAXSIGNALS];
  int lt_signals = 0;
  int flag_lchar = 0;
  int flag_lsignal = 0;
  cs_a2l_t data;
  int asap2major = 0;
  int asap2minor = 0;
  cs_parser_result_t result;
  cs_a2l_list_t list;
  cs_a2l_measurement_t me;
  cs_a2l_characteristic_t ch;
  int i;

  while (1) {
    static struct option long_options[] = {
      /* These options set a flag. */
      {"a2l", required_argument, NULL, 'a'},
      {"project", required_argument, NULL, 'p'},
      {"module", required_argument, NULL, 'm'},
      {"characteristic", required_argument, NULL, 'c'},
      {"listcharacteristic", required_argument, NULL, 'C'},
      {"signal", required_argument, NULL, 's'},
      {"listsignal", required_argument, NULL, 'S'},
      {"help",    no_argument, NULL, 'h'},
      {0, 0, 0, 0}
    };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    c = getopt_long (argc, argv, "a:p:m:c:Cs:Sh",
                     long_options, &option_index);

    /* Detect the end of the options. */
    if (c == -1)
      break;

    switch (c) {
    case 0:
      /* If this option set a flag, do nothing else now. */
      if (long_options[option_index].flag != 0)
        break;
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
      break;
    case 'a': filename = optarg;     break;
    case 'p': project = optarg;     break;
    case 'm': module = optarg;     break;
    case 's':
      signals[lt_signals] = optarg; 
      lt_signals++;
      if(lt_signals >= MAXSIGNALS){
	printf("Invalid number of signal defintion\n");
	return -1;
      }     
      break;
    case 'c':
      characteristic[lt_characteristic] = optarg; 
      lt_characteristic++;
      if(lt_characteristic >= MAXCHAR){
	printf("Invalid number of characteristic defintion\n");
	return -1;
      }     
      break;
    case 'C': flag_lchar = 1;     break;
    case 'S': flag_lsignal = 1;     break;
    case 'h': help(); exit(0);    break;
    case '?':
      /* getopt_long already printed an error message. */
      printf( "Typ `%s --help' for more information\n",EXECNAME);
      return -1;
      break;
    default:
      return -1;
      break;
    }
  }

  

  //First check Library version
  if(CS_A2L_CHECKVERSION){
    printf("Incompatible Library interface, Application was compiled with version %i, Library has version %i\n",
	   CS_A2L_VERSION_MAGIC,cs_a2l_libgetversion());
    return -1;
  }

  /*Check inputs*/
  if((filename==NULL)){
    printf( "Please check input arguments\n");
    printf( "Typ %s --help' for more information\n",EXECNAME);
    return -1;
  }


  //Try to open file
  if(cs_a2l_open(&data,filename)){
    printf("Failed to open file %s\n",filename);
    return -1;
  }

  //Get ASAP-Version
  if(cs_a2l_getversion(&data,&asap2major,&asap2minor)){
    printf("Failed to get ASAP2 version\n");
    return -1;
  }
  printf("ASAP2 Version Major %i Minor %i\n",asap2major,asap2minor);

  if(project == NULL){
    //Try to find a section
    result.searchin = (char*)data.buffer;
    result.searchlen = data.filelen; 
    if(cs_a2l_search_section(&result,"PROJECT",NULL)){
      printf("Cannot find section PROJECT\n");
      return -1;
    }
    
    //Set serach area
    result.searchin = (char*)data.buffer;
    result.searchlen = data.filelen; 
    if(cs_a2l_list_sections(&result,&list,"PROJECT")){
      printf("Failed to build up section list\n");
      cs_a2l_list_free(&list);
      return -1;
    }
    printf("Available PROJECTs:\n");
    cs_a2l_printlist(&list);
    cs_a2l_list_free(&list);
    //Close file
    if(cs_a2l_close(&data)){
      printf("Failed to close file\n");
      return -1;
    }
    return 0;
  }
  //Get Module
  if(cs_a2l_GetProject(&data,project)){
    printf("Cannot get Project %s\n",project);
    return -1;
  }

  if(module == NULL){    
    //Set serach area
    result.searchin = (char*)data.project.result;
    result.searchlen = data.project.resultlen; 
    if(cs_a2l_list_sections(&result,&list,"MODULE")){
      printf("Failed to build up section list\n");
      cs_a2l_list_free(&list);
      return -1;
    }
    printf("Available MODULES:\n");
    cs_a2l_printlist(&list);
    cs_a2l_list_free(&list);
    
    //Close file
    if(cs_a2l_close(&data)){
      printf("Failed to close file\n");
      return -1;
    }
    return 0;
  }

  //Get Module
  if(cs_a2l_GetModule(&data,module)){
    printf("Cannot get Module\n");
    return -1;
  }
  
  if(flag_lchar){
    //Set serach area
    result.searchin = (char*)data.module.result;
    result.searchlen = data.module.resultlen; 
    if(cs_a2l_list_sections(&result,&list,"CHARACTERISTIC")){
      printf("Failed to build up section list\n");
      cs_a2l_list_free(&list);
      return -1;
    }
    printf("Available Characteristics:\n");
    for(i = 0; i< list.no_result; i++){
      printf("%s\n",list.resultname[i]);
    }
    cs_a2l_list_free(&list);
  }

  if(flag_lsignal){
    //Set serach area
    result.searchin = (char*)data.module.result;
    result.searchlen = data.module.resultlen; 
    if(cs_a2l_list_sections(&result,&list,"MEASUREMENT")){
      printf("Failed to build up section list\n");
      cs_a2l_list_free(&list);
      return -1;
    }
    printf("Available Measurements:\n");
    for(i = 0; i< list.no_result; i++){
      printf("%s\n",list.resultname[i]);
    }
    cs_a2l_list_free(&list);
  }

  if(lt_characteristic > 0){
    for(i = 0; i < lt_characteristic; i++){
      if(cs_a2l_GetCharacteristic(&data,&ch,characteristic[i])){
	printf("Failed to parse characteristic :%s\n",characteristic[i]);
      } else {
	cs_a2l_print_characteristic(&ch);
      }
    }
  }

  if(lt_signals > 0){
    for(i = 0; i < lt_signals; i++){
      if(cs_a2l_GetMeasurement(&data,&me,signals[i])){
	printf("Failed to parse Measurement :%s\n",signals[i]);
      } else {
	cs_a2l_print_measurement(&me);
      }
    }
  }


  //Close file
  if(cs_a2l_close(&data)){
    printf("Failed to close file\n");
    return -1;
  }

  return 0;
}
