/*  
    Copyright (C) 2018 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @example main_dbc.c
 * @author Andreas Stewering-Bone
 * @date 29 June 2018
 * @brief Decode DBC human readable.
 *
 * Decode DBC human readable.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>

#include "cs_dbc.h"
#include "cs_message.h"
#include "cs_signal.h"

void help(void)
{
  printf(
	 "Usage: can_dbcls [OPTION]\n"
	 "Decode DBC hum readable.\n"
	 "\n"
	 "Options:\n"
	 "  -d, --dbc FILE  DBC file\n"
	 "  -v, --verbose   Verbose\n"
	 "      --help      display this help and exit\n");
}

void print_signal(cs_signal_t *s)
{
  int k;
  const int bufsize = 512;
  char buffer[bufsize];
  double value;
  int nolabels;
  
  printf("    %s: start: %d len: %d, scale:%f offset: %f, min: %f, max %f, signedness: %s, byteorder: %s, unit: %s, startvalue: %f\n",
	 cs_signal_get_name(s),
	 cs_signal_get_bitoffset(s),
	 cs_signal_get_bitlength(s),
	 cs_signal_get_scale(s),
	 cs_signal_get_offset(s),
	 cs_signal_get_min(s),
	 cs_signal_get_max(s),
	 cs_signal_get_signal_type(s)==SIGNAL_TYPE_SIGNED?"signed":"unsigned",
	 cs_signal_get_byteorder(s)==SIGNAL_BT_BIG?"BIG":"LITTLE",
	 cs_signal_get_unit(s),
	 cs_signal_get_startvalue(s));
  if((nolabels = cs_signal_getnolabel(s)) >= 0){
    for(k = 0; k < nolabels; k++){
      if(cs_signal_getlabel_iter(s,k,&value,buffer,bufsize)){
	printf("Failed to get label data\n");
      } else {
	printf("    Label %s (%.3f)\n",buffer,value);
      }
    }
  } else {
    printf("Failed to get number of labels\n");
  }
}

int main(int argc, char **argv)
{
  int c;
  int i,j;
  cs_dbc_t *data;
  char * filename = NULL;
  int verbose = 0;


  while (1) {
    static struct option long_options[] = {
      /* These options set a flag. */
      {"dbc", required_argument, NULL, 'd'},
      {"verbose",   no_argument, NULL, 'v'},
      {"help",    no_argument, NULL, 'h'},
      {0, 0, 0, 0}
    };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    c = getopt_long (argc, argv, "d:vh",
                     long_options, &option_index);

    /* Detect the end of the options. */
    if (c == -1)
      break;

    switch (c) {
    case 0:
      /* If this option set a flag, do nothing else now. */
      if (long_options[option_index].flag != 0)
        break;
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
      break;
    case 'd': filename = optarg;     break;
    case 'v': verbose += 1;     break;
    case 'h': help(); exit(0);    break;
    case '?':
      /* getopt_long already printed an error message. */
      printf( "Typ `can_dbcls --help' for more information\n");
      return -1;
      break;
    default:
      return -1;
      break;
    }
  }


  if(filename == NULL){
    printf("Please select a DBC file\n");
    return -1;
  }
  
  
  if(verbose > 0){
    cs_dbc_debuglevel(CS_DEBUG_LEVEL_DEBUG);
    cs_message_debuglevel(CS_DEBUG_LEVEL_DEBUG);
    cs_signal_debuglevel(CS_DEBUG_LEVEL_DEBUG);
  }

  if(verbose > 1){
    cs_dbc_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
  }

  if((data = cs_dbc_init())==NULL){
    printf("Failed to init DBC structure\n");
    return -1;
  }


  if(cs_dbc_read(data,filename)){
    printf("Failed to parse DBC file: %s\n",filename);
    cs_dbc_free(data);
    return -1;
  }


  for(i = 0; i < cs_dbc_getnonodes(data); i++){
    cs_dbc_node_t *n = cs_dbc_getnode_iter(data,i);
    if(n == NULL){
      printf("Failed to get node\n");
      break;
    }
    printf("Bus node %s\n",cs_dbc_node_get_name(n));
    printf("Generated messages:\n");
    for(j = 0; j < cs_dbc_node_no_generated_messages(n); j++){
      cs_message_t *m = cs_dbc_node_get_generated_message_iter(n,j);
      if(m == NULL){
	printf("Failed to get generated message\n");
	break;
      }
      printf("   Message ID 0x%X:%s Mode: %d\n",cs_message_get_id(m), cs_message_get_name(m), cs_message_get_sendtype(m));
    }
    printf("Requested messages:\n");
    for(j = 0; j < cs_dbc_node_no_requested_messages(n); j++){
      cs_message_t *m = cs_dbc_node_get_requested_message_iter(n,j);
      if(m == NULL){
	printf("Failed to get requested message\n");
	break;
      }
      printf("   Message ID 0x%X:%s Producer: %s Mode :%d\n",cs_message_get_id(m), cs_message_get_name(m),cs_message_get_producer(m), cs_message_get_sendtype(m));
    }
  }
  
  for(i = 0; i < cs_dbc_getnummessages(data); i++){
    cs_message_t *m = cs_dbc_getmessage_iter(data,i);
    if(m == NULL){
      printf("Invalid message pointer\n");
      continue;
    }
    printf("0x%04X: %s DLC:%d, Cyclitime: %f s, Offset: %f s, Sender: %s\n",cs_message_get_id(m),cs_message_get_name(m),cs_message_get_dlc(m),1.0e-3*cs_message_get_interval(m),1.0e-3*cs_message_get_offset(m),cs_message_get_producer(m));
    printf("Send type %d\n",cs_message_get_sendtype(m));
    printf("Send interval %i us\n",cs_message_get_interval(m));
    printf("Send interval offset %i us\n",cs_message_get_offset(m));
    printf("Send start offset %i us\n",cs_message_get_startoffset(m));
    if((cs_message_get_sendtype(m) == MESSAGE_STYPE_CYCLIC) && (cs_message_get_interval(m) == 0)){
      printf("WARNING: Cyclic message but no interval defined.\n");
    }
    for(j = 0; j < cs_message_get_no_signals(m); j++){
      cs_signal_t *s = cs_message_get_signal_iter(m,j);
      if(s == NULL){
	printf("Invalid signal pointer\n");
	continue;
      }
      print_signal(s);
    }
  }
  
  
  cs_dbc_free(data);
  return 0;
}
