/*  
    Copyright (C) 2015 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @example main_decoder.c
 * @author Andreas Stewering-Bone
 * @date 30 June 2015
 * @brief Simple CAN decoder application.
 *
 * Simple CAN decoding. Prints results to stdout.
 * Reads SocketCAN interface, ASC or TRC file for input.
*/
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <time.h>
#include <limits.h>
#include <signal.h>

#include "cs_dbc.h"
#include "cs_lists.h"
#include "cs_ccp_decoder.h"
#include "cs_isotp_decoder.h"
#include "cs_uds_decoder.h"
#include "cs_file_parser.h"
#include "cs_message.h"
#include "cs_signal.h"
#include "cs_socketcan.h"

#define BUFSIZE 1024
#define PARSE_UNKNOWN 0
#define PARSE_ASC 1
#define PARSE_TRC 2
#define PARSE_SOCKETCAN 3
#define FILTERMAX 0x10
#define EXECNAME "can_decoder"

typedef struct{
  double *signals;
  double *signals_old;
  cs_message_t *mdata;
  uint8_t first_call;
  uint8_t *skip_signal;
} messagestore_t;

typedef struct{
  uint32_t sa;
  uint32_t ta;
} canpair_t;

typedef struct{
  cs_message_t **messages;
  int only_update;
  messagestore_t *fmessages;
  int lt_messages;
  uint32_t include[FILTERMAX];
  int lt_include;
  int lt_isotp;
  canpair_t isotp_id[FILTERMAX];
  int lt_uds;
  canpair_t uds_id[FILTERMAX];
  int lt_sig;
  char *esc_sig[FILTERMAX];
  cs_socketcan_t *socket;
  cs_dbc_t* dbc;
} snifferdata_t;

//global variable for signal handler
int run;
int verbose = 0;

void signal_handler(int signum)
{
  switch (signum) {
  case SIGINT:
  case SIGTERM:
    printf("Receive stop signal %i\n",signum);
    run = 0;
    break;
  default:
    break;
  }
}

double timeval2d(struct timeval time)
{
    return (double)time.tv_sec + (double)time.tv_usec * 1.0e-6;
}

void help(void)
{
  printf(
          "Usage: %s [OPTION]\n"
          "Simple CAN decoder.\n"
          "\n"
          "Options:\n"
          "  -i, --input    SocketCAN-Iface, ASC or TRC file\n"
	  "  -a, --a2l      A2L file to use, Optional\n"
	  "  -d, --dbc      DBC file for signal decoding.Optional\n"
	  "  -c, --change       output DBC based signals only on signal change\n"
	  "  -f, --filter 0xXX Include filter for CAN decoding, can be set multiple times\n"
	  "  -s, --skip   SIGNAL Skip signal with name or part of name, can be set multiple times\n"
	  "  -p, --project  A2L project. Optional\n"
	  "  -m, --module   A2L module. Optional\n"
	  "  -t 0xID1,0xID2 , --isotp 0xID1,0xID2 ISOTP SA and TA ID, can be set multiple times\n"
	  "  -u 0xID1,0xID2 , --uds 0xID1,0xID2 UDS SA and TA ID, can be set multiple times\n"
	  "  -v, --verbose      Increase verbose level\n"
          "      --help     display this help and exit\n",EXECNAME);
}

int dbc_prepare(snifferdata_t *data, char *filename)
{
  cs_message_t *dummymessage = NULL;
  cs_signal_t *signal = NULL;
  int i,j,k;

  if(filename == NULL){
    printf("DBC-Name is empty\n");
    return -1;
  }
    
  if(cs_dbc_read(data->dbc,filename)){
    printf("Cannot read DBC-File: %s\n",filename);
    return -1;
  }
  
  data->lt_messages = cs_dbc_getnummessages(data->dbc);
  
  if(data->lt_messages < 0){
    printf("Cannot get number of messages from DBC-File\n");
    return -1;
  }
  
  if(data->lt_messages == 0){
    printf("No messages defined in DBC-File\n");
    return -1;
  }

  data->messages = (cs_message_t**)malloc(sizeof(cs_message_t*)*data->lt_messages);
  
  if(data->messages == NULL){
    printf("Malloc failure\n");
    return -1;
  }
  
  data->fmessages = (messagestore_t*)malloc(sizeof(messagestore_t)*data->lt_messages);
  
  if(data->fmessages == NULL){
    printf("Malloc failure\n");
    return -1;
  }
  memset(data->fmessages,0,sizeof(messagestore_t)*data->lt_messages);
  
  /*Set all messages from dbc file*/
  for(i = 0; i < data->lt_messages; i++){
    if((dummymessage = cs_dbc_getmessage_iter(data->dbc, i)) == NULL){
      printf("Cannot get message handle from DBC-File\n");
      return -1;
    }
    data->messages[i] = cs_message_cp(dummymessage);
    if(data->messages[i] == NULL){
      printf("Cannot set Message in restbus API\n");
      return -1;
    }
    //Set message direction
    if(cs_message_set_direction(data->messages[i],MESSAGE_DIR_RECEIVE)){
      printf("Cannot set message direction\n");
      return -1;
    }
    
    //Store message handle
    data->fmessages[i].mdata = data->messages[i];
    
    //Malloc signal feedback space
    data->fmessages[i].signals = (double*)malloc(sizeof(double)*cs_message_get_no_signals(data->messages[i]));
    data->fmessages[i].signals_old = (double*)malloc(sizeof(double)*cs_message_get_no_signals(data->messages[i]));
    data->fmessages[i].skip_signal = (uint8_t*)malloc(sizeof(uint8_t)*cs_message_get_no_signals(data->messages[i]));
    
    if(
       (data->fmessages[i].signals == NULL) ||
       (data->fmessages[i].signals_old == NULL) ||
       (data->fmessages[i].skip_signal == NULL)
       ){
      printf("Cannot allocate internal data\n");
      return -1;
    }
    
    memset(data->fmessages[i].signals,0,sizeof(double)*cs_message_get_no_signals(data->messages[i]));
    memset(data->fmessages[i].signals_old,0,sizeof(double)*cs_message_get_no_signals(data->messages[i]));
    memset(data->fmessages[i].skip_signal,0,sizeof(uint8_t)*cs_message_get_no_signals(data->messages[i]));
    data->fmessages[i].first_call = 1;
    for(j = 0; j < cs_message_get_no_signals(data->messages[i]); j++){
      signal = cs_message_get_signal_iter(data->messages[i],j);
      for(k = 0; k < data->lt_sig; k++){
	if(strstr((const char*)cs_signal_get_name(signal),(const char*)data->esc_sig[k]) != NULL){
	  //printf("Skip Signal %s\n",signal->name);
	  data->fmessages[i].skip_signal[j] = 1;
	  continue;
	}
      }
      //Register feedback value
      if(cs_signal_set_mux_type(signal,SIGNAL_MUX_SIGNAL)||
	 cs_signal_set_rvalue(signal,
			      &data->fmessages[i].signals[j],
			      NULL,NULL)){			
	printf("Failed to register signal return value: %s\n",cs_signal_get_name(signal)); 
	return -1;
      }
    }
    //Check and precalc message settings
    if(cs_message_check(data->messages[i])){
      printf("Cannot check message\n");
      return -1;
    }
    
  }

  return 0;
}

void print_data(snifferdata_t *data, int idx, double tv)
{
  int i;
  uint8_t fault = 0;
  char label[CS_CHARBUF];
  cs_message_t *m = data->messages[idx];
  cs_signal_t *s = NULL;
  int do_print = 0;

  if(data->only_update == 1){
    if(data->fmessages[idx].first_call == 1){
      do_print = 1;
    } else {
      //Check if we have to print the message
      for(i = 0; i < cs_message_get_no_signals(m); i++){
	if(data->fmessages[idx].signals[i] != data->fmessages[idx].signals_old[i]){
	  do_print = 1;
	  break;
	}
      }
    }
    if(do_print == 0){
      //Nothing todo
      return;
    }
  }
  
  printf("%f: ID: 0x%X: CAN Message %s",tv,cs_message_get_id(m),cs_message_get_name(m));
  for(i = 0; i < cs_message_get_no_signals(m); i++){
    if(data->fmessages[idx].skip_signal[i] == 1){
      //Do not print signal
      continue;
    }
    if((data->only_update == 0) ||
       (data->fmessages[idx].first_call == 1) ||
       (data->fmessages[idx].signals[i] != data->fmessages[idx].signals_old[i])){
      s = cs_message_get_signal_iter(m,i);
      printf(",%s,%f",cs_signal_get_name(s),data->fmessages[idx].signals[i]);
      if(cs_signal_runtime_decodelabel(s,data->fmessages[idx].signals[i],&fault,label)){
	printf(",()");
	continue;
      }
      printf(",(%s)",label);
    }
  }
  printf("\n");
  data->fmessages[idx].first_call = 0;
}

int message_decode(cs_ccp_decoder_t *ddata,snifferdata_t *sdata,uint32_t canid, int dlc, uint8_t *rdata, double tv)
{
  int i,j;
  int flag_decode = 0;
  if(sdata->lt_include > 0){
    for(i = 0; i < sdata->lt_include; i++){
      if((canid == sdata->include[i]) && 
	 (canid != ddata->ccpa2l.can_send_id) &&
	 (canid != ddata->ccpa2l.can_recv_id)){
	flag_decode = 1;	
      }
    }
  } else {
    //Escape CCP
    if((canid != ddata->ccpa2l.can_send_id) &&
       (canid != ddata->ccpa2l.can_recv_id)){
      flag_decode = 1;
    }
  }

  if(flag_decode == 1){
    for(i = 0; i < sdata->lt_messages; i++){
      if(canid == cs_message_get_id(sdata->messages[i])){
	if(dlc < cs_message_get_dlc(sdata->messages[i])){
	  printf("Error: Received invalid DLC for CAN id 0x%X, got %i should %i Bytes\n",canid,dlc,cs_message_get_dlc(sdata->messages[i]));
	  return CS_FAIL;
	}
	if(dlc > cs_message_get_dlc(sdata->messages[i])){
	  printf("Warning: Received invalid DLC for CAN id 0x%X, got %i should %i Bytes\n",canid,dlc,cs_message_get_dlc(sdata->messages[i]));
	}
	//Store old values before update
	for(j = 0; j < cs_message_get_no_signals(sdata->messages[i]); j++){
	  sdata->fmessages[i].signals_old[j] = sdata->fmessages[i].signals[j];
	}
	if(cs_message_runtime_decode(sdata->messages[i],rdata,dlc)){
	  printf("Failed to decode Message %s\n",cs_message_get_name(sdata->messages[i]));
	  return CS_FAIL;
	}
	print_data(sdata, i, tv);
	break;
      }
    }
  }
  return CS_OK;
}

int split_ids(canpair_t *pair, char *string)
{
  char delimiter[] = ",";
  char *ptr;
  uint32_t entry;

  // initialisieren und ersten Abschnitt erstellen
  ptr = strtok(string, delimiter);
  if (ptr == NULL){
    return 0;
  }
  entry = strtoul(ptr, NULL, 16);
  
  if( entry == ULONG_MAX){
    return 0;
  }
  
  pair->sa = entry;
  ptr = strtok(NULL, delimiter);
  if (ptr == NULL){
    return 0;
  }
  entry = strtoul(ptr, NULL, 16);
  
  if( entry == ULONG_MAX){
    return 0;
  }
  
  pair->ta = entry;

  return 1;
}

void isotp_print(void *priv_data,
		 double ts,
		 uint32_t id,
		 const char *fmt, ...)
{
  char buffer[BUFSIZE];
  va_list ap;
  printf("%f: ID: 0x%X: ",ts,id);
  memset(buffer,0,sizeof(buffer));
  va_start(ap, fmt);
  vsnprintf(buffer,BUFSIZE, fmt, ap);
  va_end(ap);
  printf("%s",buffer);
  printf("\n");
}

void isotp_data_dummy(void *priv_data,
		      double ts,
		      uint32_t id,
		      uint8_t *data,
		      int len,
		      int error)
{
  //Nothing to do
}
  
int main(int argc, char **argv)
{
  struct sigaction sa;
  int c;
  cs_ccp_decoder_t data;
  cs_isotp_decoder_t **tpdata = NULL;
  cs_uds_decoder_t **udsdata = NULL;
  cs_file_parser_t *parser = NULL;
  cs_file_parser_type_t parser_type = FILE_PARSER_TYPE_ASC;
  cs_file_parser_entry_t pentry;
  snifferdata_t sdata;
  char buffer[BUFSIZE];
  int flag_parse = PARSE_UNKNOWN;
  uint32_t entry;
  int i;
  uint32_t canid;
  int rlen;
  uint8_t rdata[CS_DATA];
  double rts;
  struct timeval rtv;

  char * filename = NULL;
  char * dbcfile = NULL;
  char * project = NULL;
  char * module = NULL;
  char * input = NULL;
  char * tpids = NULL;
  char * udsids = NULL;

  run = 1;
  
  memset(&sa, 0x00, sizeof(sa));
  sa.sa_handler = &signal_handler;
  if (sigaction(SIGTERM, &sa, NULL) == -1) {
    printf("Failed to install signal handler!\n");
    return -1;
  }
  if (sigaction(SIGINT, &sa, NULL) == -1) {
    printf("Failed to install signal handler!\n");
    return -1;
  }

  memset(&sdata,0,sizeof(snifferdata_t));
  sdata.lt_include = -1;
  sdata.lt_isotp = -1;
  sdata.lt_uds = -1;
  sdata.lt_sig = -1;

  while (1) {
    static struct option long_options[] = {
      /* These options set a flag. */
      {"input", required_argument, NULL, 'i'},
      {"a2l", required_argument, NULL, 'a'},
      {"dbc", required_argument, NULL, 'd'},
      {"change", no_argument, NULL, 'c'},
      {"filter", required_argument, NULL, 'f'},
      {"skip", required_argument, NULL, 's'},
      {"project", required_argument, NULL, 'p'},
      {"module", required_argument, NULL, 'm'},
      {"isotp", required_argument, NULL, 't'},
      {"uds", required_argument, NULL, 'u'},
      {"help",    no_argument, NULL, 'h'},
      {"verbose",    no_argument, NULL, 'v'},
      {0, 0, 0, 0}
    };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    c = getopt_long (argc, argv, "i:a:d:cf:s:p:m:ht:u:v",
                     long_options, &option_index);

    /* Detect the end of the options. */
    if (c == -1)
      break;

    switch (c) {
    case 0:
      /* If this option set a flag, do nothing else now. */
      if (long_options[option_index].flag != 0)
        break;
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
      break;
    case 'i': 
      input = optarg;     
      if((strstr(input,".asc")!=NULL)||
	 (strstr(input,".ASC")!=NULL)){
	flag_parse = PARSE_ASC;
	parser_type = FILE_PARSER_TYPE_ASC;
	break;
      }
      if((strstr(input,".trc")!=NULL)||
	 (strstr(input,".TRC")!=NULL)){
	flag_parse = PARSE_TRC;
	parser_type = FILE_PARSER_TYPE_TRC;
	break;
      }
      if(strstr(input,"can")!=NULL){
	flag_parse = PARSE_SOCKETCAN;
	break;
      }
      break;
    case 'a': filename = optarg;  break;
    case 'd': dbcfile = optarg;  break;
    case 'c': sdata.only_update = 1;  break;
    case 'f':
      if(sdata.lt_include >= FILTERMAX){
	printf("Invalid number of filter.\n");
	return -1;
      }
      if(sdata.lt_include == -1){
	sdata.lt_include = 0;
      }
      entry = strtoul(optarg, NULL, 16);
      
      if( entry == ULONG_MAX){
	printf("Invalid filter entry: %s\n",optarg);
	return -1;
      }
      sdata.include[sdata.lt_include] = (uint32_t) entry;
      sdata.lt_include++;
      break;
    case 's':
      if(sdata.lt_sig >= FILTERMAX){
	printf("Invalid number of skip signals.\n");
	return -1;
      }
      if(sdata.lt_sig == -1){
	sdata.lt_sig = 0;
      }
      sdata.esc_sig[sdata.lt_sig] = (char*)malloc(sizeof(char)*(strlen(optarg) +1));

      if(sdata.esc_sig[sdata.lt_sig] == NULL){
	printf("Malloc failure\n");
	return -1;
      }
      memset(sdata.esc_sig[sdata.lt_sig],0,sizeof(char)*(strlen(optarg) +1));
      strncpy(sdata.esc_sig[sdata.lt_sig],optarg,strlen(optarg));
      sdata.lt_sig++;
      break;
    case 'p': project = optarg;  break;
    case 'm': module = optarg;  break;
    case 't': tpids = optarg;
      if(sdata.lt_isotp >= FILTERMAX){
	printf("Invalid number of IsoTP pairs.\n");
	return -1;
      }
      if(sdata.lt_isotp == -1){
	sdata.lt_isotp = 0;
      }
      if(!split_ids(&sdata.isotp_id[sdata.lt_isotp],tpids)){
	printf("Invalid Can ID pair for ISOTP\n");
	return -1;
      }
      sdata.lt_isotp++;
      break;
    case 'u': udsids = optarg;
      if(sdata.lt_uds >= FILTERMAX){
	printf("Invalid number of UDS pairs.\n");
	return -1;
      }
      if(sdata.lt_uds == -1){
	sdata.lt_uds = 0;
      }
      if(!split_ids(&sdata.uds_id[sdata.lt_uds],udsids)){
	printf("Invalid Can ID pair for UDS\n");
	return -1;
      }
      sdata.lt_uds++;
      break;
    case 'v': verbose++;    break;
    case 'h': help(); exit(0);    break;
    case '?':
      /* getopt_long already printed an error message. */
      printf( "Typ `ccpsniffer --help' for more information\n");
      return -1;
      break;
    default:
      return -1;
      break;
    }
  }


  if(verbose > 0){
    cs_dbc_debuglevel(CS_DEBUG_LEVEL_DEBUG);
  }

  if(verbose > 1){
    cs_dbc_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
    cs_message_debuglevel(CS_DEBUG_LEVEL_DEBUG);
    cs_signal_debuglevel(CS_DEBUG_LEVEL_DEBUG);
  }

  if(verbose > 3){
    cs_dbc_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
    cs_message_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
    cs_signal_debuglevel(CS_DEBUG_LEVEL_VERBOSE);
  }
  
  /*Check inputs*/
  if(input==NULL){
    printf( "Please check input arguments\n");
    printf( "Typ `%s --help' for more information\n",EXECNAME);
    return -1;
  }

  if((filename != NULL)&&((project == NULL)||(module == NULL))){
    printf("If a A2L file is defined, you have to enter a project and a module name\n");
    printf( "Please check input arguments\n");
    printf( "Typ `%s --help' for more information\n",EXECNAME);
    return -1;
  }

  if((sdata.dbc = cs_dbc_init())==NULL){
    printf("Failed to init DBC structure\n");
    return -1;
  }
  
  if(dbcfile != NULL){
    if(dbc_prepare(&sdata,dbcfile)){
      printf("Failed to Prepare DBC information\n");
      return -1;
    }
  }

  if(sdata.lt_isotp > 0){
    tpdata = (cs_isotp_decoder_t**)malloc(sizeof(cs_isotp_decoder_t*)*sdata.lt_isotp);
    if(tpdata == NULL){
      	printf("Failed to Malloc\n");
	return -1;
    }
    for(i = 0; i < sdata.lt_isotp; i++){
      tpdata[i] = cs_isotp_decoder_init(sdata.isotp_id[i].sa,sdata.isotp_id[i].ta,
					isotp_print,isotp_print,isotp_data_dummy,NULL,NULL);
      if(tpdata[i] == NULL){
	printf("Failed to Prepare ISOTP decoder\n");
	return -1;
      }
    }
  }

  if(sdata.lt_uds > 0){
    udsdata = (cs_uds_decoder_t**)malloc(sizeof(cs_uds_decoder_t*)*sdata.lt_uds);
    if(udsdata == NULL){
      printf("Failed to Malloc\n");
      return -1;
    }
    for(i = 0; i < sdata.lt_uds; i++){
      udsdata[i] = cs_uds_decoder_init(sdata.uds_id[i].sa,
				       sdata.uds_id[i].ta,
				       isotp_print,
				       isotp_print,
				       isotp_data_dummy,
				       NULL,
				       UDS_DECODER_ISOTP_ERROR);
      if(udsdata[i] == NULL){
	printf("Failed to Prepare UDS decoder\n");
	return -1;
      }
    }
  }
  
  switch(flag_parse){
  case PARSE_TRC:
  case PARSE_ASC:
    //Prepare parser
    if((parser = cs_file_parser_open(input,parser_type))==NULL){
      printf("Failed to open input file: %s\n",input);
      return -1;
    }
    
    if(filename != NULL){
      //Prepare decoder
      if(cs_ccp_decoder_open(&data,filename,project,module)){
	printf("Failed to open decoder\n");
	return -1;
      }
    }

    while(run){
      if(cs_file_parser_getentry(parser,&pentry)){
	break;
      }
      if(dbcfile != NULL){
	message_decode(&data,&sdata,pentry.canid,pentry.dlc, pentry.data, pentry.ts);
      }
      if(filename != NULL){
	if(!cs_ccp_decoder_buffer(&data,
				  pentry.ts,
				  pentry.canid,
				  pentry.data,
				  pentry.dlc,
				  buffer,
				  BUFSIZE)){
	  printf("%s",buffer);
	}
      }
      if(sdata.lt_isotp > 0){
	for(i = 0; i < sdata.lt_isotp; i++){
	  cs_isotp_decoder_data(tpdata[i],
				pentry.ts,
				pentry.canid,
				pentry.data,
				pentry.dlc);
	}
      }
      if(sdata.lt_uds > 0){
	for(i = 0; i < sdata.lt_uds; i++){
	  cs_uds_decoder_data(udsdata[i],
			      pentry.ts,
			      pentry.canid,
			      pentry.data,
			      pentry.dlc);
	}
      }
    }
    
    cs_file_parser_close(parser);
    if(filename != NULL)
      cs_ccp_decoder_close(&data);
    break;
  case PARSE_SOCKETCAN:
    //Open CAN socket

    if((sdata.socket = cs_socketcan_init()) == NULL){
      printf("Failed to init CAN interface\n");
      return -1;
    }

    if(cs_socketcan_config_native(sdata.socket, input,SOCKETCAN_MODE_CAN)){
      printf("Failed to configure CAN interface\n");
      return -1;
    }
    
    if(cs_socketcan_open(sdata.socket)){
      printf("Failed to open CAN interface\n");
      return -1;
    }
        
    while(run){
      //Recv blocking
      rlen = 8;
      while(cs_socketcan_recv_block(sdata.socket,&canid,rdata,&rlen,&rtv,1000) == CS_SOCKETCAN_OK){
	rts = timeval2d(rtv);
	if(dbcfile != NULL){
	  message_decode(&data,&sdata,canid,rlen, rdata, rts);
	}
	if(filename != NULL){
	  if(!cs_ccp_decoder_buffer(&data,
				    rts,
				    canid,
				    rdata,
				    rlen,
				    buffer,
				    BUFSIZE)){
	    printf("%s",buffer);
	  }
	}
	if(sdata.lt_isotp > 0){
	  for(i = 0; i < sdata.lt_isotp; i++){
	    cs_isotp_decoder_data(tpdata[i],
				  rts,
				  canid,
				  rdata,
				  rlen);
	  }
	}
	
	if(sdata.lt_uds > 0){
	  for(i = 0; i < sdata.lt_uds; i++){
	    cs_uds_decoder_data(udsdata[i],
				rts,
				canid,
				rdata,
				rlen);
	  }
	}
	rlen = 8;
      }
    }
    for(i = 0; i < sdata.lt_isotp; i++){
      cs_isotp_decoder_close(tpdata[i]);
    }
    free(tpdata);
    for(i = 0; i < sdata.lt_uds; i++){
      cs_uds_decoder_close(udsdata[i]);
    }
    free(udsdata);
    cs_socketcan_close(sdata.socket);
    cs_socketcan_free(sdata.socket);
    break;
  default:
    printf("Cannot detect file type\n");
    return -1;
  }

  //Free DBC structure
  cs_dbc_free(sdata.dbc);
  return 0;
}
