/*  
    Copyright (C) 2015 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @example main_replay.c
 * @author Andreas Stewering-Bone
 * @date 6 July 2015
 * @brief Simple CAN resend application.
 *
 * Simple CAN resend application. Send content to given can interface.
 * Reads ASC or TRC file for input.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <time.h>
#include <limits.h>

#include "cs_file_parser.h"
#include "cs_socketcan.h"


#define BUFSIZE 1024
#define MAXID 0xFF 
#define PARSE_UNKNOWN 0
#define PARSE_ASC 1
#define PARSE_TRC 2
#define EXECNAME "can_replay"

void help(void)
{
  printf(
          "Usage: %s [OPTION]\n"
          "Simple CAN ASCor TRC replay.\n"
          "\n"
          "Options:\n"
          "  -i, --input    ASC or TRC file\n"
	  "  -d, --device   CAN device to use\n"
	  "  -l, --loop     Endless loop\n"
	  "  -f, --filter   Include CAN ids for sending, can be set multiple times\n"
	  "  -s, --skip     Exclude CAN ids for sending, can be set multiple times\n"
          "      --help     display this help and exit\n",EXECNAME);
}

struct timespec d2timespec(double ts)
{
  struct timespec rv;
  rv.tv_sec = (long int)ts;
  rv.tv_nsec = (long int)((ts-rv.tv_sec)*1e9);
  return rv;
}

void send(cs_socketcan_t *socket,double ts_last, double ts, uint32_t id, uint8_t *data, int len)
{
  struct timespec lsleep;
  double diff = ts-ts_last;
  if(diff > 0.0001){
    //Wait until timeout
    lsleep = d2timespec(diff);
    nanosleep(&lsleep,NULL);
  }
  //Send
  cs_socketcan_send(socket,id,data,len,SOCKETCAN_SEND_CAN);
}

int main(int argc, char **argv)
{
  int c,i;
  cs_file_parser_t *parser = NULL;
  cs_file_parser_entry_t pentry;
  cs_file_parser_type_t ptype= FILE_PARSER_TYPE_UNKNOWN;
  double ts_last = -1.0;
  uint32_t filter[MAXID];
  int lt_filter = -1;
  uint32_t skip[MAXID];
  int lt_skip = -1;
  uint32_t entry;
  cs_socketcan_t *socket = NULL;
  int flag_parse = PARSE_UNKNOWN;
  int flag_send;
  int flag_endless = 0;

  char * filename = NULL;
  char * caniface = NULL;

  while (1) {
    static struct option long_options[] = {
      /* These options set a flag. */
      {"input", required_argument, NULL, 'i'},
      {"device", required_argument, NULL, 'd'},
      {"filter", required_argument, NULL, 'f'},
      {"skip", required_argument, NULL, 's'},
      {"help",    no_argument, NULL, 'h'},
      {0, 0, 0, 0}
    };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    c = getopt_long (argc, argv, "i:d:f:s:lh",
                     long_options, &option_index);

    /* Detect the end of the options. */
    if (c == -1)
      break;

    switch (c) {
    case 0:
      /* If this option set a flag, do nothing else now. */
      if (long_options[option_index].flag != 0)
        break;
      printf ("option %s", long_options[option_index].name);
      if (optarg)
        printf (" with arg %s", optarg);
      printf ("\n");
      break;
    case 'i': 
      filename = optarg;     
      if((strstr(filename,".asc")!=NULL)||
	 (strstr(filename,".ASC")!=NULL)){
	flag_parse = PARSE_ASC;
	ptype = FILE_PARSER_TYPE_ASC;
      }
      if((strstr(filename,".trc")!=NULL)||
	 (strstr(filename,".TRC")!=NULL)){
	flag_parse = PARSE_TRC;
	ptype = FILE_PARSER_TYPE_TRC;
      }
      break;
    case 'd': caniface = optarg;  break;
    case 'f':
      if(lt_filter >= MAXID){
	printf("Invalid number of filter.\n");
	return -1;
      }
      if(lt_filter == -1){
	lt_filter = 0;
      }
      entry = strtoul(optarg, NULL, 16);

      if( entry == ULONG_MAX){
	printf("Invalid filter entry: %s\n",optarg);
	return -1;
      }
      filter[lt_filter] = (uint32_t) entry;
      lt_filter++;
      break;
    case 's':
      if(lt_skip >= MAXID){
	printf("Invalid number of skip ids.\n");
	return -1;
      }
      if(lt_skip == -1){
	lt_skip = 0;
      }
      entry = strtoul(optarg, NULL, 16);

      if( entry == ULONG_MAX){
	printf("Invalid filter entry: %s\n",optarg);
	return -1;
      }
      skip[lt_skip] = (uint32_t) entry;
      lt_skip++;
      break;
    case 'l': flag_endless = 1;    break;
    case 'h': help(); exit(0);    break;
    case '?':
      /* getopt_long already printed an error message. */
      printf( "Type %s --help' for more information\n",EXECNAME);
      return -1;
      break;
    default:
      return -1;
      break;
    }
  }

  if(flag_parse == PARSE_UNKNOWN){
    printf( "Cannot detect file type\n");
    printf( "Type %s --help' for more information\n",EXECNAME);
    return -1;
  }
  
  /*Check inputs*/
  if((filename==NULL)||(caniface == NULL)){
    printf( "Please check input arguments\n");
    printf( "Type %s --help' for more information\n",EXECNAME);
    return -1;
  }

 //Open CAN socket
  if((socket = cs_socketcan_init()) ==  NULL){
    printf("Failed to init CAN interface\n");
    return -1;
  }

  if(cs_socketcan_config_native(socket, caniface,0)){
    printf("Failed to open CAN interface\n");
    return -1;
  }
  
  if(cs_socketcan_open(socket)){
    printf("Failed to open CAN interface\n");
    return -1;
  }

  while(1){
      //Prepare parser
    if((parser = cs_file_parser_open(filename,ptype))==NULL){
      printf("Failed to open input file: %s\n",filename);
      return -1;
    }    
    while(!cs_file_parser_getentry(parser,&pentry)){
      if(ts_last < 0.0)
	ts_last = pentry.ts;
      flag_send = 0;
      if(lt_filter > 0){
	for(i = 0; i < lt_filter; i++){
	  if(pentry.canid == filter[i]){
	    flag_send = 1;
	    continue;
	  }
	}
      } else {
	flag_send = 1;
      }
      if(lt_skip > 0){
	for(i = 0; i < lt_skip; i++){
	  if(pentry.canid == skip[i]){
	    flag_send = 0;
	    continue;
	  }
	}
      }
      if(flag_send){
	send(socket,ts_last, pentry.ts, pentry.canid, pentry.data, pentry.dlc);
	ts_last = pentry.ts;
      }
    }
    cs_file_parser_close(parser);
    
    if(!flag_endless)
      break;
  }
  cs_socketcan_close(socket);
  cs_socketcan_free(socket);
  return 0;
}
